SystemOrganization addCategory: #'HTML-DOM'!
SystemOrganization addCategory: #'HTML-DOM-Errors'!
SystemOrganization addCategory: #'HTML-Tests'!
SystemOrganization addCategory: #'HTML-Validator'!
SystemOrganization addCategory: #'HTML-CSS'!
SystemOrganization addCategory: #'HTML-CSS-Tests'!
SystemOrganization addCategory: #'HTML-CSS-Errors'!
SystemOrganization addCategory: #'HTML-PList'!
SystemOrganization addCategory: #OSProcess!
SystemOrganization addCategory: #'OSProcess-AIO'!
SystemOrganization addCategory: #'OSProcess-Base'!
SystemOrganization addCategory: #'OSProcess-Mac'!
SystemOrganization addCategory: #'OSProcess-OS2'!
SystemOrganization addCategory: #'OSProcess-RiscOS'!
SystemOrganization addCategory: #'OSProcess-Tests'!
SystemOrganization addCategory: #'OSProcess-Unix'!
SystemOrganization addCategory: #'OSProcess-Win32'!

!String methodsFor: '*html' stamp: 'tb 1/17/2008 14:57'!
content

	^self! !

!String methodsFor: '*html' stamp: 'tb 1/17/2008 14:21'!
splitOn: aString

	| idx lastIdx |
	idx := lastIdx := 1.

	^Array streamContents: 
	[:s |
		[idx <= self size] whileTrue:
		[
			idx := self findString: aString startingAt: lastIdx.
			idx = 0 ifTrue: [idx := self size + 1].
			s nextPut: (self copyFrom: lastIdx to: idx-1).
			lastIdx := idx + aString size.
		]
	]
! !

!String methodsFor: '*html' stamp: 'tb 11/28/2005 18:14'!
unquoted

	^self withoutQuoting! !

Model subclass: #AioEventHandler
	instanceVariableNames: 'semaphore semaIndex handlerProc descriptor'
	classVariableNames: 'AioPluginPresent'
	poolDictionaries: ''
	category: 'OSProcess-AIO'!

!AioEventHandler commentStamp: 'dtl 11/25/2006 15:55' prior: 0!
AioEventHandler responds to external IO events, such as data available on a file descriptor. When an external IO event is received, an instance of AioEventHandler sends #changed to itself to notify its dependents that the event has occurred.!

!AioEventHandler class methodsFor: 'testing' stamp: 'dtl 2/11/2013 19:17'!
aioPluginPresent
	"Answer true if an AIO plugin is available. The value of AioPluginPresent
	is cleared at startup time, and is reestablished once for each Squeak session.
	See initializeAioPluginPresent to initialize after an image startup if warning
	dialog is not desired."

	"AioEventHandler aioPluginPresent"

	^ AioPluginPresent
		ifNil: [AioPluginPresent := self basicNew primAioModuleVersionString notNil.
			AioPluginPresent ifFalse:
				[self notify: 'AioPlugin not present, AioEventHandler will use polling input'].
			^ AioPluginPresent]! !

!AioEventHandler class methodsFor: 'instance creation' stamp: 'dtl 7/5/2003 09:16'!
for: aSocketOrFileStream

	"self for: SourceFiles first"
	"self for: Socket new"
	"self for: OSProcess thisOSProcess stdIn"

	^ self new for: aSocketOrFileStream! !

!AioEventHandler class methodsFor: 'instance creation' stamp: 'dtl 7/5/2003 11:14'!
for: aSocketOrFileStream exceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag
	"Flag parameters are true or false, indicating types of events to be handled."

	^ self new for: aSocketOrFileStream
			exceptions: exceptionEventFlag
			readEvents: readEventFlag
			writeEvents: writeEventFlag! !

!AioEventHandler class methodsFor: 'instance creation' stamp: 'dtl 7/4/2003 18:47'!
forFileStream: aFileStream

	"self forFileStream: SourceFiles first"

	^ self new forFileStream: aFileStream

! !

!AioEventHandler class methodsFor: 'instance creation' stamp: 'dtl 7/5/2003 11:14'!
forFileStream: aFileStream exceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag
	"Flag parameters are true or false, indicating types of events to be handled."

	"self forFileStream: SourceFiles first"

	^ self new forFileStream: aFileStream
			exceptions: exceptionEventFlag
			readEvents: readEventFlag
			writeEvents: writeEventFlag
! !

!AioEventHandler class methodsFor: 'instance creation' stamp: 'dtl 7/4/2003 19:01'!
forSocket: aSocket

	"self forSocket: Socket new"

	^ self new forSocket: aSocket

! !

!AioEventHandler class methodsFor: 'instance creation' stamp: 'dtl 7/5/2003 11:14'!
forSocket: aSocket exceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag
	"Flag parameters are true or false, indicating types of events to be handled."

	"self forSocket: Socket new"

	^ self new forSocket: aSocket
			exceptions: exceptionEventFlag
			readEvents: readEventFlag
			writeEvents: writeEventFlag
! !

!AioEventHandler class methodsFor: 'testing' stamp: 'dtl 10/3/2012 20:24'!
initializeAioPluginPresent
	"Initialize the AioPluginPresent flag silently without invoking a warning
	if the plugin is not present, and answer the value of the flag. Send this
	to prevent a warning dialog from being presented after image startup,
	as may be preferred if the image is to be run headless."

	"AioEventHandler initializeAioPluginPresent"

	^ AioPluginPresent
		ifNil: [AioPluginPresent := self basicNew primAioModuleVersionString notNil]! !

!AioEventHandler class methodsFor: 'system startup' stamp: 'dtl 11/25/2006 10:04'!
startUp: resuming
	"Clear the value of AioPluginPresent. The value will be set once when
	#aioPluginPresent is first sent, and will remain set to that value for the
	duration of this Squeak session.
	This method is called by ThisOSProcess>>startUp. AioEventHandler does
	not need to be registered in the system startup list."

	resuming ifTrue: [AioPluginPresent := nil]
! !

!AioEventHandler methodsFor: 'aio event forwarding' stamp: 'dtl 7/4/2003 19:10'!
aioDisable: ioDescriptor
	"Definitively disable asynchronous event notification for a descriptor. The ioDescriptor
	parameter is an object representing a low level OS file or socket descriptor."

	^ self primAioDisable: ioDescriptor
! !

!AioEventHandler methodsFor: 'aio event forwarding' stamp: 'dtl 9/1/2003 16:58'!
aioEnable: ioDescriptor forSemaphore: semaphoreIndex externalObject: trueOrFalse
	"Enable asynchronous notification for a descriptor. Send this message one time
	prior to beginning event handling for ioDescriptor. The ioDescriptor parameter is an
	object representing a low level OS file or socket descriptor. The semaphoreIndex is
	the index of a Semaphore to be notified, and the third parameter is a flag indicating
	that ioDescriptor represents an external object which should not be closed on termination
	of aio handling."

	^ (self primAioEnable: ioDescriptor forSemaphore: semaphoreIndex externalObject: trueOrFalse)
			ifNil: [self notify: 'aio event forwarding not supported']
! !

!AioEventHandler methodsFor: 'aio event forwarding' stamp: 'dtl 7/4/2003 18:52'!
aioHandle: ioDescriptor exceptionEvents: exceptionFlag readEvents: readFlag writeEvents: writeFlag
	"Enable asynchronous notification for a descriptor. Send this message one time to
	enable a single event notification. Send it again after each event has been received
	and handled (in other words, the process which waits on the event semaphore is
	responsible for re-enabling the handler by calling this method each time an event
	is handled). The ioDescriptor parameter is an object representing a low level OS file or
	socket descriptor. The remaining three parameters are Boolean flags representing the
	types of events for which notification is being requested: handle exceptions, handle
	for read, and handle for write. It is common to watch for read events and exception
	events, or to watch for write events and exception events."

	^ self primAioHandle: ioDescriptor
		exceptionEvents: exceptionFlag
		readEvents: readFlag
		writeEvents: writeFlag
! !

!AioEventHandler methodsFor: 'aio event forwarding' stamp: 'dtl 7/4/2003 19:10'!
aioSuspend: ioDescriptor exceptionEvents: exceptionFlag readEvents: readFlag writeEvents: writeFlag

	"Temporarily suspend asynchronous event notification for a descriptor. The
	ioDescriptor parameter is an object representing a low level OS file or socket
	descriptor. The remaining three parameters are Boolean flags representing the
	types of events for which notification is being requested: handle exceptions,
	handle for read, and handle for write."

	^ self primAioSuspend: ioDescriptor
		exceptionEvents: exceptionFlag
		readEvents: readFlag
		writeEvents: writeFlag
! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 8/20/2006 18:43'!
close
	"When the FileStream or Socket handled by this aio handler is closed,
	it should send #close to this handler."

	| p |
	self breakDependents.
	Smalltalk unregisterExternalObject: semaphore.
	self aioDisable: self descriptor.
	semaphore := nil.
	semaIndex := nil.
	p := handlerProc.
	handlerProc := nil.
	p ifNotNil: [p terminate] "p may be the active process, do this last"
! !

!AioEventHandler methodsFor: 'handler process' stamp: 'dtl 7/5/2003 10:50'!
defaultHandlerProcess
	"Generate a #changed notification whenever an external aio event occurs"

	^ self handleReadAndExceptionsEvents! !

!AioEventHandler methodsFor: 'accessing' stamp: 'dtl 3/30/2003 19:05'!
descriptor

	^ descriptor! !

!AioEventHandler methodsFor: 'accessing' stamp: 'dtl 3/30/2003 19:05'!
descriptor: aLowLevelIODescriptor

	descriptor := aLowLevelIODescriptor! !

!AioEventHandler methodsFor: 'finalization' stamp: 'dtl 9/4/2003 06:54'!
finalize
	"Note: An aio handler will not be garbage collected until the semaphore
	is unregistered. When the FileStream or Socket handled by this aio
	handler is closed, it should send #close to this handler."

	self close
! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 7/5/2003 09:15'!
for: aSocketOrFileStream
	"Answer an event handler for any kind of IO stream that can be associated
	with an OS handle for aio events. Currently, subclasses of FileStream and
	Socket can have aio event handlers."

	(aSocketOrFileStream isKindOf: Socket)
		ifTrue: [^ self forSocket: aSocketOrFileStream].
	(aSocketOrFileStream isKindOf: FileStream)
		ifTrue: [^ self forFileStream: aSocketOrFileStream].
	self error: 'expected a FileStream or Socket subclass'! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 7/5/2003 11:03'!
for: aSocketOrFileStream exceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag
	"Answer an event handler for any kind of IO stream that can be associated
	with an OS handle for aio events. Currently, subclasses of FileStream and
	Socket can have aio event handlers."

	(aSocketOrFileStream isKindOf: Socket)
		ifTrue: [^ self forSocket: aSocketOrFileStream
					exceptions: exceptionEventFlag
					readEvents: readEventFlag
					writeEvents: writeEventFlag].
	(aSocketOrFileStream isKindOf: FileStream)
		ifTrue: [^ self forFileStream: aSocketOrFileStream
					exceptions: exceptionEventFlag
					readEvents: readEventFlag
					writeEvents: writeEventFlag].
	self error: 'expected a FileStream or Socket subclass'! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 7/9/2005 14:00'!
forFileStream: aFileStream

	self descriptor: (self handleForFile: aFileStream).
	self setDefaultEventMask.
! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 7/5/2003 11:01'!
forFileStream: aFileStream exceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag

	self descriptor: (self handleForFile: aFileStream).
	self initializeForExceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag

! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 7/9/2005 14:00'!
forSocket: aSocket
	"Any existing event handling for aSocket will be disabled. Note that this
	will make the socket useless for any applications that expect the prior
	event handling behavior."

	self descriptor: (self handleForSocket: aSocket).
	self aioDisable: self descriptor.
	self setDefaultEventMask.
! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 7/5/2003 12:46'!
forSocket: aSocket exceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag
	"Any existing event handling for aSocket will be disabled. Note that this
	will make the socket useless for any applications that expect the prior
	event handling behavior."

	self descriptor: (self handleForSocket: aSocket).
	self aioDisable: self descriptor.
	self initializeForExceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag
! !

!AioEventHandler methodsFor: 'handler process' stamp: 'dtl 11/12/2011 21:25'!
handleExceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag
	"Generate a #changed notification whenever the requested type of external aio event occurs."

	| p sema handler |
	sema := Semaphore new.
	handler := [[self hasValidHandler] whileTrue:
		[self aioHandle: self descriptor
			exceptionEvents: exceptionEventFlag
			readEvents: readEventFlag
			writeEvents: writeEventFlag.
		sema signal. "event handler is ready"
		self semaphore wait.
		self changed]].
	p := handler forkAt: Processor userBackgroundPriority.
	sema wait. "until process has been started and events are being handled"
	^p! !

!AioEventHandler methodsFor: 'private' stamp: 'dtl 9/4/2003 06:40'!
handleForFile: aFileStream

	"self new handleForFile: SourceFiles first"

	| ioHandle |
	ioHandle := self useIOHandle
		ifTrue: [aFileStream ioHandle]
		ifFalse: [aFileStream fileID].
	^ self primOSFileHandle: ioHandle
! !

!AioEventHandler methodsFor: 'private' stamp: 'dtl 7/4/2003 15:16'!
handleForSocket: aSocket

	"self new handleForSocket: Socket newTCP"

	| ioHandle |
	ioHandle := self useIOHandle
		ifTrue: [aSocket ioHandle]
		ifFalse: [aSocket socketHandle].
	^ self primOSSocketHandle: ioHandle
! !

!AioEventHandler methodsFor: 'handler process' stamp: 'dtl 7/5/2003 10:49'!
handleReadAndExceptionsEvents
	"Generate a #changed notification whenever data is available for reading or an exception
	occurs on the external IO channel."

	^ self handleExceptions: true readEvents: true writeEvents: false
! !

!AioEventHandler methodsFor: 'handler process' stamp: 'dtl 7/5/2003 10:49'!
handleReadEvents
	"Generate a #changed notification whenever data is available for reading"

	^ self handleExceptions: false readEvents: true writeEvents: false
! !

!AioEventHandler methodsFor: 'accessing' stamp: 'dtl 3/30/2003 19:04'!
handlerProc

	^ handlerProc! !

!AioEventHandler methodsFor: 'handler process' stamp: 'dtl 4/23/2012 06:45'!
hasValidHandler
	"True if the event handler is running, and if it refers to the correct external
	object semaphore. For protection following an image restart."

	^ semaIndex notNil
		and: [Smalltalk externalObjects size >= semaIndex
			and: [(Smalltalk externalObjects at: semaIndex) == semaphore]]! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 9/3/2010 16:48'!
initializeForExceptions: exceptionEventFlag readEvents: readEventFlag writeEvents: writeEventFlag

	semaphore := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: semaphore.
	([self aioEnable: self descriptor forSemaphore: self semaIndex externalObject: true]
		on: Warning
		do: [:e |
			self close. "unregister the semaphore"
			self notify: e messageText. nil])
		ifNotNil:
			[handlerProc := self
				handleExceptions: exceptionEventFlag
				readEvents: readEventFlag
				writeEvents: writeEventFlag]
! !

!AioEventHandler methodsFor: 'primitive access' stamp: 'dtl 9/1/2003 17:53'!
primAioDisable: aDescriptor
	"Definitively disable asynchronous event notification for a descriptor. The descriptor
	parameter is an object representing a low level OS file or socket descriptor."

	<primitive: 'primitiveAioDisable' module: 'AioPlugin'>
	^ nil
! !

!AioEventHandler methodsFor: 'primitive access' stamp: 'dtl 9/1/2003 17:53'!
primAioEnable: aDescriptor forSemaphore: semaphoreIndex externalObject: trueOrFalse
	"Enable asynchronous notification for a descriptor. The descriptor parameter is an
	object representing a low level OS file or socket descriptor. The semaphoreIndex
	is the index of a Semaphore to be notified, and the third parameter is a flag indicating
	that descriptor represents an external object which should not be closed on termination
	of aio handling."

	<primitive: 'primitiveAioEnable' module: 'AioPlugin'>
	^ nil
! !

!AioEventHandler methodsFor: 'primitive access' stamp: 'dtl 9/1/2003 17:53'!
primAioHandle: aDescriptor exceptionEvents: exceptionFlag readEvents: readFlag writeEvents: writeFlag
	"Enable asynchronous notification for a descriptor. The descriptor parameter is an
	object representing a low level OS file or socket descriptor. The second parameter is
	the index of a Semaphore to be notified, the remaining three parameters are Boolean
	flags representing the types of events for which notification is being requested:
	handle exceptions, handle for read, and handle for write. It is common to watch for
	read events and exception events, or to watch for write events and exception events."

	<primitive: 'primitiveAioHandle' module: 'AioPlugin'>
	^ nil
! !

!AioEventHandler methodsFor: 'primitive access' stamp: 'dtl 9/1/2003 17:26'!
primAioModuleName
	"Module name of the installed plugin, if any."

	"self new primAioModuleName"

	<primitive: 'primitiveModuleName' module: 'AioPlugin'>
	^ nil
! !

!AioEventHandler methodsFor: 'primitive access' stamp: 'dtl 9/1/2003 17:28'!
primAioModuleVersionString
	"Module name of the installed plugin, if any."

	"self new primAioModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'AioPlugin'>
	^ nil
! !

!AioEventHandler methodsFor: 'primitive access' stamp: 'dtl 9/1/2003 17:54'!
primAioSuspend: aDescriptor exceptionEvents: exceptionFlag readEvents: readFlag writeEvents: writeFlag
	"Temporarily suspend asynchronous event notification for a descriptor. The
	descriptor parameter is an object representing a low level OS file or socket
	descriptor. The remaining three parameters are Boolean flags representing the
	types of events for which notification is being requested: handle exceptions,
	handle for read, and handle for write."

	<primitive: 'primitiveAioSuspend' module: 'AioPlugin'>
	^ nil
! !

!AioEventHandler methodsFor: 'primitive access' stamp: 'dtl 9/1/2003 17:54'!
primOSFileHandle: sqFile
	"Answer the low level file descriptor for a file IO handle."

	<primitive: 'primitiveOSFileHandle' module: 'AioPlugin'>
	^ nil
! !

!AioEventHandler methodsFor: 'primitive access' stamp: 'dtl 9/1/2003 17:54'!
primOSSocketHandle: sqFile
	"Answer the low level socket descriptor for a socket IO handle."

	<primitive: 'primitiveOSSocketHandle' module: 'AioPlugin'>
	^ nil
! !

!AioEventHandler methodsFor: 'accessing' stamp: 'dtl 3/30/2003 19:04'!
semaIndex

	^ semaIndex! !

!AioEventHandler methodsFor: 'accessing' stamp: 'dtl 3/30/2003 19:04'!
semaphore

	^ semaphore! !

!AioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 7/9/2005 14:00'!
setDefaultEventMask
	"Default initialization for read events and exception events"

	^ self initializeForExceptions: true readEvents: true writeEvents: false

! !

!AioEventHandler methodsFor: 'private' stamp: 'dtl 3/30/2003 19:26'!
useIOHandle

	^ Smalltalk hasClassNamed: #IOHandle! !

Model subclass: #OSProcessAccessor
	instanceVariableNames: 'sessionIdentifier canObtainSessionIdentifierFromPlugin hasTestEof'
	classVariableNames: 'EmulateWin32FileLocking FileLockRegistry ThisOSProcessAccessor UseIOHandle'
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!OSProcessAccessor commentStamp: 'dtl 7/7/2010 07:58' prior: 0!
I am an abstract class whose subclasses provide access to an operating system process, such as the process in which the Squeak VM is currently running. My subclasses collaborate with instances of OSProcess subclasses.

The single instance ThisOSProcessAccessor provides access to the OS process in which the virtual machine is running. On Unix, this instance sets a signal handler to respond to externally generated sigchld signals. This must be done after each image restart in order to call a primitive which informs the VM of the  identity of the semaphore to signal. A similar mechanism is used for Windows to obtain exit status of external OS processes. ThisOSProcessor maintains a process that waits on child exit events and updates a dependent OSProcess of changes to external OS processes.

When an image is restarted on a different kind of platform, a new instance is assigned to ThisOSProcessAccessor to provide access to the virtual machine OS process.
!

OSProcessAccessor subclass: #MacOSProcessAccessor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Mac'!

!MacOSProcessAccessor commentStamp: '<historical>' prior: 0!
I provide access to an operating system process, such as the process in which the Squeak VM is currently running. I am based on the Macintosh process model.!

!MacOSProcessAccessor class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:44'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for the
	Squeak VM running on the current platform."

	^ OSProcess isNonUnixMac
! !

!MacOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 2/23/2002 05:57'!
primGetSession
	"Answer the unique identifier for this session of Smalltalk running in this OS Process."

	^ nil
! !

OSProcessAccessor subclass: #OS2OSProcessAccessor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-OS2'!

!OS2OSProcessAccessor commentStamp: '<historical>' prior: 0!
I provide access to an operating system process, such as the process in which the Squeak VM is currently running. I am based on the OS2 process model.!

!OS2OSProcessAccessor class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:47'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for the
	Squeak VM running on the current platform."

	^ OSProcess isOS2
! !

!OS2OSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 2/23/2002 05:56'!
primGetSession
	"Answer the unique identifier for this session of Smalltalk running in this OS Process."

	^ nil
! !

!OSProcessAccessor class methodsFor: 'concrete subclasses' stamp: 'dtl 3/5/2005 12:04'!
concreteClass

	"OSProcessAccessor concreteClass"

	^ self subclasses
		detect: [:c | c isResponsibleForThisPlatform]
		ifNone: [self notify: self printString,
					': No concrete class implementation available for system type ',
					OSProcess platformName printString.
				nil]

! !

!OSProcessAccessor class methodsFor: 'initialize-release' stamp: 'dtl 3/5/2005 13:24'!
emulateWin32FileLocking: trueOrFalse
	"This is a preference that controls whether file locking will attempt to emulation
	Win32 behavior, in which a lock request will fail if the requested region overlaps
	a region for which there is an existing lock. This behavior is valid only for locks
	managed within a single Squeak image, and will not produce the expected results
	for a Squeak image cooperating with another Squeak image, or with some other
	external program.

	Use of the Win32 emulation may result in performance penalties for an application
	that performs a large number of lock requests, such as a database."

	"self emulateWin32FileLocking: true"
	"self emulateWin32FileLocking: false"

	EmulateWin32FileLocking := trueOrFalse
! !

!OSProcessAccessor class methodsFor: 'instance creation' stamp: 'dtl 12/14/2013 09:15'!
forThisOSProcess
	"Answer a single instance corresponding to the OS process in which this 
	Smalltalk image is executing."

	"OSProcessAccessor forThisOSProcess"

	ThisOSProcessAccessor
		ifNotNil: [ThisOSProcessAccessor isResponsibleForThisPlatform
					ifTrue:
						["Common case, platform has not changed"
						^ThisOSProcessAccessor ]
					ifFalse:
						["We are running on a different platform, so start a new accessor"
						ThisOSProcessAccessor changed: #invalidProcessAccessor.
						ThisOSProcessAccessor finalize]].
	^ ThisOSProcessAccessor := self concreteClass basicNew initialize
! !

!OSProcessAccessor class methodsFor: 'initialize-release' stamp: 'dtl 12/14/2013 09:15'!
initialize

	"OSProcessAccessor initialize"

	UseIOHandle := (Smalltalk hasClassNamed: #IOHandle).
	ThisOSProcessAccessor := nil.
	self emulateWin32FileLocking: false.
	self allSubInstances do: [:e | e finalize]
! !

!OSProcessAccessor class methodsFor: 'platform identification' stamp: 'dtl 8/24/2003 09:17'!
isResponsibleForThisPlatform
	"Answer true if an instance of this class is responsible for representing the
	OS process for the Squeak VM running on the current platform."

	^ self subclassResponsibility! !

!OSProcessAccessor class methodsFor: 'instance creation' stamp: 'dtl 3/2/2002 08:29'!
new

	self inform: 'use OSProcessAccessor>>forThisOSProcess to create or obtain the OSProcess instance for this Smalltalk session.'.
	^ nil! !

!OSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:41'!
aioModuleName
	"Answer a string containing the module name string for the AIO plugin."

	"OSProcess accessor aioModuleName"

	^ self subclassResponsibility! !

!OSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:41'!
aioVersionString
	"Answer a string containing the version string for the AIO plugin."

	"OSProcess accessor aioVersionString"

	^ self subclassResponsibility! !

!OSProcessAccessor methodsFor: 'testing' stamp: 'dtl 10/18/2001 22:58'!
canAccessSystem
	"Answer true if it is possible to access the external process. Concrete subclasses should
	know how to answer true."

	^ false
! !

!OSProcessAccessor methodsFor: 'file lock registry' stamp: 'dtl 4/9/2005 22:44'!
canAcquireLock: anOSFileLockDescriptor
	"Answer true if the file lock cache will permit fileLock to be acquired. This method
	does not guarantee that the underlying OS will grant the lock."

	^ (self fileLockRegistry anySatisfy:
			[:ld | ld isActive and: [ld conflictsWith: anOSFileLockDescriptor]]) not! !

!OSProcessAccessor methodsFor: 'testing' stamp: 'dtl 2/14/2004 11:43'!
canForwardExternalSignals
	"Answer true if it is possible to forward OS signals to a Smalltalk Semaphore."

	^ false
! !

!OSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 4/7/2007 10:19'!
canObtainSessionIdentifierFromPlugin

	^ canObtainSessionIdentifierFromPlugin
		ifNil: [canObtainSessionIdentifierFromPlugin := self primGetSession notNil]! !

!OSProcessAccessor methodsFor: 'file lock registry' stamp: 'dtl 3/5/2005 13:21'!
emulateWin32FileLocking
	"Answer the current value of this preference"

	^ EmulateWin32FileLocking! !

!OSProcessAccessor methodsFor: 'file lock registry' stamp: 'dtl 9/24/2009 21:36'!
fileLockRegistry

	^ FileLockRegistry ifNil: [FileLockRegistry := Set new]
! !

!OSProcessAccessor methodsFor: 'session identification' stamp: 'dtl 8/6/2003 21:39'!
getSessionIdentifier
	"Call a primitive to obtain the unique identifier for this Squeak session. If the
	primitive fails, try to deduce the session identifier from an instance of
	StandardFileStream. Some versions of the OSProcessPlugin may not be able to
	obtain a session ID, so this provides a mechanism for obtaining the session ID
	indirectly if necessary."

	"OSProcess accessor getSessionIdentifier"

	| session |
	session := self primGetSession.
	session ifNil: [session :=  self getSessionIdentifierFromSourcesFile].
	session ifNil: [session :=  self getSessionIdentifierFromFileInSqueakDirectory].
	^ session
! !

!OSProcessAccessor methodsFor: 'session identification' stamp: 'dtl 1/23/2013 21:30'!
getSessionIdentifierFromFileInSqueakDirectory
	"Deduce the session identifier using method of last resort.
	This is an unreliable method, because it assumes knowledge of the internal structure
	of the SQFile data structure.

	Deprecated:
	As of approximately Squeak 3.8 and beyond, the session id has been moved to the
	first slot of the data structure. This method will not work for a Squeak VM beyond
	that point, and will not work for any 64 bit VM. However, an reliable means of
	obtaining sessionID is now available (#getSessionIdentifier), so this method is retained
	only for backwards compatibility to allow OSPP to be built on an older VMMaker."

	"OSProcess accessor getSessionIdentifierFromFileInSqueakDirectory"

	| id s |
	(OSProcess directoryEntryNames: OSProcess defaultPathString) do:
		[:n | (s := FileStream readOnlyFileNamed: n) ifNotNil:
			[(Smalltalk hasClassNamed: #IOHandle)
				ifTrue: [id := s ioHandle getHandle copyFrom: 5 to: 8]
				ifFalse: [(id := s fileID) ifNotNil: [id copyFrom: 5 to: 8]].
			s close.
			^ id]].
	^ nil
! !

!OSProcessAccessor methodsFor: 'session identification' stamp: 'dtl 8/7/2005 22:02'!
getSessionIdentifierFromSourcesFile
	"Deduce the session identifier from an existing open FileStream on the sources file.
	This is an unreliable method, because it assumes knowledge of the internal structure
	of the SQFile data structure.

	Deprecated:
	As of approximately Squeak 3.8 and beyond, the session id has been moved to the
	first slot of the data structure. This method will not work for a Squeak VM beyond
	that point, and will not work for any 64 bit VM. However, an reliable means of
	obtaining sessionID is now available (#getSessionIdentifier), so this method is retained
	only for backwards compatibility to allow OSPP to be built on an older VMMaker."

	"OSProcess accessor getSessionIdentifierFromSourcesFile"

	| s id |
	s := SourceFiles first.
	s ifNil: [^ nil].
	^ (Smalltalk hasClassNamed: #IOHandle)
		ifTrue: [s ioHandle getHandle copyFrom: 5 to: 8]
		ifFalse: [(id := s fileID) ifNotNil: [id copyFrom: 5 to: 8]]
! !

!OSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/25/2005 16:07'!
getStdErrHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard error for the
	OS process in which I am currently executing."

	^ self subclassResponsibility
! !

!OSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/25/2005 16:07'!
getStdInHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard input for the
	OS process in which I am currently executing."

	^ self subclassResponsibility
! !

!OSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/25/2005 16:06'!
getStdOutHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard output for the
	OS process in which I am currently executing."

	^ self subclassResponsibility
! !

!OSProcessAccessor methodsFor: 'private - IOHandle' stamp: 'dtl 3/1/2002 15:19'!
handleFromAccessor: aByteArrayOrIOAccessor

	UseIOHandle
		ifTrue: [aByteArrayOrIOAccessor isNil
			ifTrue: [^ nil]
			ifFalse: [^ aByteArrayOrIOAccessor asSQFileStruct]]
		ifFalse: [^ aByteArrayOrIOAccessor]! !

!OSProcessAccessor methodsFor: 'private - IOHandle' stamp: 'dtl 1/3/2004 21:21'!
handleFromFileStream: aFileStream

	^ UseIOHandle
		ifTrue: [self handleFromAccessor: aFileStream ioHandle]
		ifFalse: [aFileStream fileID]
! !

!OSProcessAccessor methodsFor: 'testing' stamp: 'dtl 8/8/2002 15:13'!
handlesOsSignals
	"True if OS signals can be handled and forwarded to the image"

	^ false! !

!OSProcessAccessor methodsFor: 'initialize - release' stamp: 'dtl 4/8/2007 11:09'!
initialize

	canObtainSessionIdentifierFromPlugin := nil.
	self canObtainSessionIdentifierFromPlugin.
	sessionIdentifier := nil.
	hasTestEof := nil.
	self sessionIdentifier.
! !

!OSProcessAccessor methodsFor: 'private - IOHandle' stamp: 'dtl 7/12/2003 12:17'!
ioAccessorFromSQFile: aByteArray
	"Answer an object which represents an IO channel. If IOHandle is present in
	this image, use it; otherwise just answer aByteArray."

	UseIOHandle
		ifTrue: [^ (Smalltalk at: #IOHandle) newFromSqFileStruct: aByteArray]
		ifFalse: [^ aByteArray]! !

!OSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 11/28/2010 14:04'!
isAtEndOfFile: anIOHandle
	"Answer whether the file represented by anIOHandle is at end of file, as determined
	by a call to feof(). This is different from StandardFileStream>>primAtEnd: which answers
	true if the file pointer is at the end of the file, but which does not call feof() to
	determine that an end of file condition has occurred. The difference is significant
	if aSqFileStruct represents a pipe or a device file, which may not be positionable
	in the sense of a conventional disk file."

	^ self primTestEndOfFileFlag: (self handleFromAccessor: anIOHandle)
! !

!OSProcessAccessor methodsFor: 'platform identification' stamp: 'dtl 8/24/2003 09:18'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ self class isResponsibleForThisPlatform! !

!OSProcessAccessor methodsFor: 'pipe open' stamp: 'dtl 11/28/2010 12:28'!
makePipeHandles
	"Create a pipe, and answer an array of two IO accessors for the pipe 
	reader and writer."

	"OSProcess accessor makePipeHandles"

	^self primCreatePipe
		ifNotNilDo: [:p | ^ p collect: [:e | self ioAccessorFromSQFile: e]]
! !

!OSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:41'!
osppModuleName
	"Answer a string containing the module name string for the OSPP plugin."

	"OSProcess accessor osppModuleName"

	^ self subclassResponsibility! !

!OSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:41'!
osppModuleVersionString
	"Answer a string containing the version string for the OSPP plugin."

	"OSProcess accessor osppModuleVersionString"

	^ self subclassResponsibility! !

!OSProcessAccessor methodsFor: 'session identification' stamp: 'dtl 3/2/2002 09:07'!
primGetSession
	"Subclasses should override if they know how to obtain the session identifier."

	^ nil! !

!OSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 4/8/2007 10:49'!
primTestEndOfFileFlag: aSqFileStruct
	"Answer whether the file represented by aSqFileStruct is at end of file, as determined
	by a call to feof(). This is different from StandardFileStream>>primAtEnd: which answers
	true if the file pointer is at the end of the file, but which does not call feof() to
	determine that an end of file condition has occurred. The difference is significant
	if aSqFileStruct represents a pipe or a device file, which may not be positionable
	in the sense of a conventional disk file."

	^ self subclassResponsibility
! !

!OSProcessAccessor methodsFor: 'file lock registry' stamp: 'dtl 3/5/2005 13:06'!
register: fileRegionLock
	"If an object equal to fileRegionLock exists in the registry, answer it. Otherwise, add
	fileRegionLock to the registry and answer fileRegionLock."

	^ (self fileLockRegistry like: fileRegionLock)
		ifNil: [self fileLockRegistry add: fileRegionLock]
! !

!OSProcessAccessor methodsFor: 'file lock registry' stamp: 'dtl 4/10/2005 15:05'!
registeredLocksForFile: aFileStream
	"Answer all lock descriptors associated with aFileStream"

	^ self fileLockRegistry select: [:ea | ea fileStream = aFileStream]
! !

!OSProcessAccessor methodsFor: 'file lock registry' stamp: 'jf 2/26/2004 18:06'!
removeInactiveLocks
	"Go through the lock cache and remove any that have been left
	behind after their streams were closed."

	^ self fileLockRegistry copy do: [:ea | ea isActive ifFalse: [self fileLockCache remove: ea]]! !

!OSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 3/2/2002 08:32'!
sessionIdentifier

	^ sessionIdentifier ifNil: [sessionIdentifier := self getSessionIdentifier]
! !

!OSProcessAccessor methodsFor: 'file control' stamp: 'dtl 2/11/2001 15:37'!
setNonBlocking: anIOHandle
	"Convert anIOHandle to an SQFile data structure and call primitive to set it non-blocking."

	^ self subclassResponsibility
! !

!OSProcessAccessor methodsFor: 'file lock registry' stamp: 'dtl 3/5/2005 13:07'!
unregister: fileRegionLock
	"If an object equal to fileRegionLock exists in the registry, remove it and
	answer the object. Otherwise answer nil."


	^ self fileLockRegistry remove: fileRegionLock ifAbsent: [nil]
! !

!OSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:41'!
xdcpModuleName
	"Answer a string containing the module name string for the display control plugin."

	"OSProcess accessor xdcpModuleName"

	^ self subclassResponsibility! !

!OSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:42'!
xdcpVersionString
	"Answer a string containing the version string for the display control plugin."

	"OSProcess accessor xdcpVersionString"

	^ self subclassResponsibility! !

OSProcessAccessor subclass: #RiscOSProcessAccessor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-RiscOS'!

!RiscOSProcessAccessor commentStamp: '<historical>' prior: 0!
I provide access to the operating system process in which the Squeak VM is currently running. I am based on the RiscOS task model. There is only one instance of me, and instances of RiscOSProcess depend on me to provide access to the operating system process which they represent. I know how to create child processes. I use a semaphore to receive signals when child processes die, and I notify my dependents (instances ofRiscOSProcess) when these events occur.
!

!RiscOSProcessAccessor class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:48'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for the
	Squeak VM running on the current platform."

	^ OSProcess isRiscOS
! !

!RiscOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 1/6/2001 23:15'!
primForkAndExec: executableFile
	withArgs: anArrayOfArgumentStrings
	argCount: numberOfArgumentStrings
	withEnv: anArrayOfEnvironmentStrings
	envCount: numberOfEnvironmentStrings
	stdIn: inputFileHandle
	stdOut: outputFileHandle
	stdErr: errorFileHandle
	"Parameters are expected to have been properly prepared by the caller,
	including string values which are to be null terminated strings.
	In other words, all strings should have (Character value: 0) as the
	last element in the string."! !

!RiscOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 2/23/2002 05:56'!
primGetSession
	"Answer the unique identifier for this session of Smalltalk running in this OS Process."

	^ nil
! !

OSProcessAccessor subclass: #UnixOSProcessAccessor
	instanceVariableNames: 'grimReaper sigChldSemaphore'
	classVariableNames: 'ThisProcessPid'
	poolDictionaries: ''
	category: 'OSProcess-Unix'!

!UnixOSProcessAccessor commentStamp: '<historical>' prior: 0!
I provide access to the operating system process in which the Squeak VM is currently running. I am based on the Unix process model. There is only one instance of me, and instances of UnixOSProcess depend on me to provide access to the operating system process which they represent.

I know how to create child processes. I use a semaphore to receive signals when child processes die, and I notify my dependents (instances of UnixOSProcess) when these events occur.

!

!UnixOSProcessAccessor class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:49'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for the
	Squeak VM running on the current platform."

	^ OSProcess isUnix
! !

!UnixOSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:43'!
aioModuleName
	"Answer a string containing the module name string for the AIO plugin."

	"OSProcess accessor aioModuleName"

	^ self primAioPluginModuleName
! !

!UnixOSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:44'!
aioVersionString
	"Answer a string containing the version string for the AIO plugin."

	"OSProcess accessor aioVersionString"

	^ self primAioPluginVersionString
! !

!UnixOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 10/7/2000 13:39'!
canAccessChildProcess: anExternalProcess
	"Is the child process still there? Maybe not if we have restarted the image
	and anExternalProcess refers to a process which died while we were not
	watching."

	^ self primCanReceiveSignals: anExternalProcess pid! !

!UnixOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 9/9/2000 15:45'!
canAccessSystem
	"Answer true if it is possible to access the external process, else false. Failure
	to access the external process is probably due to lack of a UnixOSProcessPlugin
	module."

	^ self primGetPid notNil
! !

!UnixOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 8/30/2003 18:43'!
canControlXDisplay
	"True if the XDisplayControlPlugin is accessible. Older versions of OSProcess relied on
	the X display control to be embedded in the OSProcessPlugin module. This has been
	moved to a separate display control plugin to allow support of non-X platforms and
	other display media on Unix platforms (OS X)."

	"OSProcess accessor canControlXDisplay"

	^ self primXDisplayControlPluginModuleName notNil
! !

!UnixOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 2/14/2004 12:17'!
canForwardExternalSignals
	"Answer true if it is possible to forward OS signals to a Smalltalk Semaphore."

	^ true
! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/31/2001 15:07'!
chDir: pathString
	"Change current working directory. The $PWD environment variable is not updated (but
	see UnixProcess>>chDir:). Answer nil for success, or an error message."

	"OSProcess accessor chDir: '/tmp'"
	"OSProcess accessor chDir: '/no/such/path'"
	"OSProcess accessor chDir: FileDirectory default name"

	| result message |
	result := self primChdir: pathString.
	result isNil
		ifTrue:
			[^ nil]
		ifFalse:
			[message := self primErrorMessageAt: result.
			^ message]
! !

!UnixOSProcessAccessor methodsFor: 'signal print stack trace' stamp: 'dtl 7/20/2010 00:06'!
clearPrintAllStacksOnSigUsr1
	"Clear any signal handler in the VM that was set to print all call stacks on
	console output when a SIGUSR1 is received. Restores default SIGUSR1
	handler, which may result in exiting the VM on this signal. Answer an
	identifier for the previously defined signal handler or nil if no handler
	had been set."

	^self primForwardSignal: self primSigUsr1Number toSemaphore: nil	
! !

!UnixOSProcessAccessor methodsFor: 'file control' stamp: 'dtl 3/2/2009 22:42'!
closeUnixFileNumber: unixFileDescriptor
	"Close the file associated with unixFileDescriptor, a small integer value."

	<primitive: 'primitiveUnixFileClose' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed
! !

!UnixOSProcessAccessor methodsFor: 'debugging' stamp: 'dtl 11/19/2006 09:02'!
currentSigHandlerAddress: signalNumber
	"Answer the current machine address of the signal handler for signalNumber,
	expressed as a ByteArray.

	Temporarily set a signal forwarded for signalNumber, and remember the
	machine address of the prior signal handler. Restore the signal handler
	to its previous value, and answer the machine address of the handler."

	"OSProcess accessor currentSigHandlerAddress: OSProcess accessor primSigIntNumber"

	| sema index previousHandlerAddress |
	sema := Semaphore new.
	index := Smalltalk registerExternalObject: sema.
	(previousHandlerAddress := self primForwardSignal: signalNumber toSemaphore: index)
		ifNil:
			[Smalltalk unregisterExternalObject: sema.
			^ self error: 'could not forward signal number ', signalNumber asString].
	self restoreSignal: signalNumber.
	^ previousHandlerAddress! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 7/2/2000 19:59'!
environmentAt: aSymbolOrString
	"Get an environment variable from the external OS process."

	"OSProcess thisOSProcess environmentAt: 'PATH'"

	^ self primEnvironmentAtSymbol: aSymbolOrString
! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 10/7/2001 01:18'!
environmentAt: aSymbolOrString put: aString
	"Add or update an environment variable in the external OS process. Convert aSymbol
	and aString into a KEY=value string and pass this to the OS process environment.
	Standard C libraries provide a putenv() function for this purpose, with a parameter
	in the form KEY=value. Note: Maintain a reference to the return value, see note in
	primitivePutEnv."

	"OSProcess accessor environmentAt: 'AAAA' put: 'this is the value of AAAA'"
	"OSProcess accessor environmentAt: 'AAAA' put: nil"
	"OSProcess accessor environmentAt: 'AAAA'"

	aString isNil
		ifTrue:
			[^ self primUnsetEnv: aSymbolOrString, (Character value: 0) asString]
		ifFalse:
			[^ self environmentPut: (aSymbolOrString asString, '=', aString)]
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 9/10/2000 09:48'!
environmentPut: aString
	"Add or update an environment variable in the external OS process using a 'KEY=value'
	string. Create a null terminated string for use by the external putenv() call in a pluggable
	primitive."

	"OSProcess thisOSProcess processAccessor environmentPut: 'SOMEKEY=somevalue'"

	| cString |
	cString := aString, (Character value: 0) asString.
	^ self primPutEnv: cString
! !

!UnixOSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 5/1/2001 05:49'!
fileProtectionMask: aPathString
	"Call stat(2) to obtain the file protection mask for a file."

	"OSProcess accessor fileProtectionMask: '/bin/sh'"
	"OSProcess accessor fileProtectionMask: '/etc/hosts'"
	"OSProcess accessor fileProtectionMask: '/bin/su'"
	"OSProcess accessor fileProtectionMask: '/bin/NOSUCHFILE'"
	"OSProcess accessor fileProtectionMask: 12345"

	| mask |
	(aPathString isKindOf: String) ifFalse: [self error: 'expected a path string'. ^ nil].
	mask := self primFileProtectionMask: aPathString.
	(mask == nil) ifTrue: [^ nil].
	(mask isKindOf: Integer)
		ifTrue: [^ self primErrorMessageAt: mask].
	^ mask
! !

!UnixOSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 5/1/2001 05:49'!
fileStat: aPathString
	"Call stat(2) to obtain the UID, GID, and file protection mask for a file."

	"OSProcess accessor fileStat: '/var/spool/news'"
	"OSProcess accessor fileStat: '/etc/hosts'"
	"OSProcess accessor fileStat: '/bin/su'"
	"OSProcess accessor fileStat: '/bin/NOSUCHFILE'"
	"OSProcess accessor fileStat: 12345"

	| mask |
	(aPathString isKindOf: String) ifFalse: [self error: 'expected a path string'. ^ nil].
	mask := self primFileStat: aPathString.
	(mask == nil) ifTrue: [^ nil].
	(mask isKindOf: Integer)
		ifTrue: [^ self primErrorMessageAt: mask].
	^ mask
! !

!UnixOSProcessAccessor methodsFor: 'initialize - release' stamp: 'dtl 7/6/2010 21:22'!
finalize
	"Clean up grimReaper and associated semaphore."

	self grimReaper ifNotNil:
			[grimReaper terminate.
			grimReaper := nil].
	self sigChldSemaphore ifNotNil:
			[self restoreSigChld.
			self sigChldSemaphore: nil]
! !

!UnixOSProcessAccessor methodsFor: 'file control' stamp: 'dtl 11/28/2010 12:13'!
flushExternalStream: anIOHandle
	"Convert anIOHandle to an SQFile data structure and call primitive to flush the
	external I/O stream."

	^ self primSQFileFlush: (self handleFromAccessor: anIOHandle)
! !

!UnixOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 1/25/2004 21:43'!
forkAndExec: executableFile
	stdIn: inputFileHandle
	stdOut: outputFileHandle
	stdErr: errorFileHandle
	argBuf: argVec
	argOffsets: argOffsets
	envBuf: envVec
	envOffsets: envOffsets
	workingDir: pathString

	"Parameters are expected to have been properly prepared by the caller, including string
	values which are to be null terminated strings. In other words, all strings should have
	(Character value: 0) as the last element in the string.

	Parameters should be:
		executableFile: a string with the name of a file to execute.
		stdIn: a fileID ByteArray (struct SQFile in C) to be used as standard input.
		stdOut: a fileID ByteArray to be used as standard output.
		stdErr: a fileID ByteArray to be used as standard error.
		argVec: a String arranged to look more or less like a char **, but with the addresses not yet fixed.
		argOffsets: an Array of offsets for fixing up the argVec addresses.
		envVec: a String arranged to look more or less like a char **, but with the addresses not yet fixed.
		envOffsets: an Array of offsets for fixing up the envVec addresses.
		workingDir: a null terminated path name String, or nil.
	The envVec parameter may be nil, in which case envOffsets is ignored. workingDir may be nil. The
	other parameters are required. Parameters with nil value indicate that current values for this process
	should be used."

	^ self primForkExec: executableFile
			stdIn: inputFileHandle
			stdOut: outputFileHandle
			stdErr: errorFileHandle
			argBuf: argVec
			argOffsets: argOffsets
			envBuf: envVec
			envOffsets: envOffsets
			workingDir: pathString
! !

!UnixOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 4/27/2011 21:01'!
forkSqueak
	"Clone this Squeak Smalltalk image in a child OSProcess. The child is the same as the
	parent, except for its new X session connection, and the return value of this method,
	which is zero for the child process, and a positive integer equal to the pid of the child
	for the parent process.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	self safeToForkSqueak
		ifTrue: [^self primForkSqueak].
	[self notify: 'forkSqueak requires XDisplayControlPlugin']
		ensure: [^nil]
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:51'!
forwardSigChld
	"Set a signal handler for SIGCHLD. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigChld"

	^ self forwardSignal: self primSigChldNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:51'!
forwardSigHup
	"Set a signal handler for SIGHUP. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigHup"

	^ self forwardSignal: self primSigHupNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:51'!
forwardSigInt
	"Set a signal handler for SIGINT. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigInt"

	^ self forwardSignal: self primSigIntNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:51'!
forwardSigKill
	"Set a signal handler for SIGKILL. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigKill"

	self notify: 'SIGKILL and SIGSTOP signals cannot be caught, see man signal(2)'.
	^ self forwardSignal: self primSigKillNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:52'!
forwardSigPipe
	"Set a signal handler for SIGPIPE. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigPipe"

	^ self forwardSignal: self primSigPipeNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:52'!
forwardSigQuit
	"Set a signal handler for SIGQUIT. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigQuit"

	^ self forwardSignal: self primSigQuitNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:52'!
forwardSigTerm
	"Set a signal handler for SIGTERM. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigTerm"

	^ self forwardSignal: self primSigTermNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:52'!
forwardSigUsr1
	"Set a signal handler for SIGUSR1. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigUsr1"

	^ self forwardSignal: self primSigUsr1Number
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:52'!
forwardSigUsr2
	"Set a signal handler for SIGUSR2. Answer a new Semaphore, or nil if unable
	to set the handler (possibly because it has already been set)."

	"OSProcess accessor forwardSigUsr1"

	^ self forwardSignal: self primSigUsr2Number
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 12/6/2002 18:28'!
forwardSignal: signalNumber 
	"Set a signal handler in the VM which will signal a Smalltalk semaphore 
	at semaphoreIndex whenever an external signal sigNum is received.
	Answer a new Semaphore, or nil if unable to set the handler (possibly
	because it has already been set). A Smalltalk process can wait on the
	Semaphore, and take action when a signal is detected. See man(7) signal
	for signal number definitions on your unix system."

	"OSProcess accessor forwardSignal: OSProcess accessor primSigIntNumber"

	| sema index |
	sema := Semaphore new.
	index := Smalltalk registerExternalObject: sema.
	(self primForwardSignal: signalNumber toSemaphore: index)
		ifNil:
			[Smalltalk unregisterExternalObject: sema.
			^ nil].
	^ sema! !

!UnixOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 11/28/2010 12:07'!
getStdErrHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard error for the
	OS process in which I am currently executing."

	^ self primGetStdErrHandle
		ifNotNilDo: [:handle | self ioAccessorFromSQFile: handle]! !

!UnixOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 11/28/2010 12:07'!
getStdInHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard input for the
	OS process in which I am currently executing."

	^ self primGetStdInHandle
		ifNotNilDo: [:handle | self ioAccessorFromSQFile: handle]! !

!UnixOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 11/28/2010 12:08'!
getStdOutHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard output for the
	OS process in which I am currently executing."

	^ self primGetStdOutHandle
		ifNotNilDo: [:handle | self ioAccessorFromSQFile: handle]! !

!UnixOSProcessAccessor methodsFor: 'pthreads' stamp: 'dtl 3/17/2007 23:14'!
getThreadID
	"Answer the ID of the pthread that is currently executing (the interpreter thread).
	A thread ID may be a 64 bit value on some platforms, so answer a byte array
	containing the value in machine-dependent byte order."

	"OSProcess accessor getThreadID"

	| osppVer |
	[osppVer := OSProcess accessor osppModuleVersionString asNumber]
		on: Error
		do: [^ nil].
	(osppVer notNil and: [osppVer >= 4.2])
		ifTrue: [^ self primGetThreadID]
		ifFalse: [^ nil]
! !

!UnixOSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 1/25/2004 21:29'!
grimReaper
	"Answer the value of grimReaper"

	^ grimReaper! !

!UnixOSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 2/14/2004 11:45'!
grimReaper: anObject
	"Set the value of grimReaper"

	grimReaper := anObject! !

!UnixOSProcessAccessor methodsFor: 'initialize - release' stamp: 'dtl 9/3/2010 12:49'!
grimReaperProcess
	"This is a process which waits for the death of a child OSProcess, and 
	informs any dependents of the change. Use SIGCHLD events if possible,
	otherwise a Delay to poll for exiting child processes."

	| event processSynchronizationDelay |
	^ self canAccessSystem
		ifTrue:
			[event := (self canAccessSystem and: [self canForwardExternalSignals])
				ifTrue: [self sigChldSemaphore]
				ifFalse: [Delay forMilliseconds: 200].
			processSynchronizationDelay := Delay forMilliseconds: 20.
			grimReaper ifNil:
				[grimReaper :=
					[[event wait.
					processSynchronizationDelay wait. "Avoids lost signals in heavy process switching"
					self changed: #childProcessStatus] repeat] newProcess.
					grimReaper resume.
					"name selected to look reasonable in the process browser"
					grimReaper name: ((ReadStream on: grimReaper hash asString) next: 5)
							, ': the child OSProcess watcher']]
		ifFalse:
			[nil]
! !

!UnixOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 8/8/2002 15:13'!
handlesOsSignals
	"True if OS signals can be handled and forwarded to the image"

	^ true! !

!UnixOSProcessAccessor methodsFor: 'initialize - release' stamp: 'dtl 7/7/2010 07:46'!
initialize
	"Call this method when an instance is first created, or to refresh after an image
	restart to clean up from the previous session. Notify dependents of my singleton
	instance if the image has restarted in a different OS process (this is not the case
	when #startUp is called after a simple image save)."

	self finalize.
	super initialize.
	ThisProcessPid ~~ self primGetPid
		ifTrue: 
			["Image has been restarted and is now running in a different OS process"
			ThisProcessPid := self primGetPid.
			self changed: #startUp].
	self changed.
	self restartChildWatcherProcess.

! !

!UnixOSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 10/6/2001 12:08'!
isExecutable: aPathName
	"Answer true if file at aPathName has execute permission for this process."

	"OSProcess accessor isExecutable: '/bin/sh'"
	"OSProcess accessor isExecutable: '/no/such/file'"

	^ self isExecutable: aPathName forUser: self primGetUid inGroup: self primGetGid

! !

!UnixOSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 5/1/2001 22:53'!
isExecutable: aPathName forUser: uid inGroup: gid
	"Answer true if file at aPathName has execute permission for a user
	identified by user uid and group gid."

	"OSProcess accessor
		isExecutable: '/bin/sh'
		forUser: OSProcess accessor primGetUid
		inGroup: OSProcess accessor primGetGid"

	| fStat suid sgid user group protectionMask |
	fStat := self fileStat: aPathName.
	(fStat isKindOf: String)
		ifTrue: [self inform: aPathName, ': ', fStat. ^ nil].
	protectionMask := fStat at: 3.
	((protectionMask at: 4) allMask: 1)
		ifTrue: [^ true]. "Test executable by any user ID"
	sgid := (protectionMask at: 1) allMask: 2.
	sgid
		ifTrue: [group := gid]
		ifFalse: [group := fStat at: 2].
	((gid == group) and: [(protectionMask at: 3) allMask: 1])
		ifTrue: [^ true]. "Test executable by my group ID"
	suid := (protectionMask at: 1) allMask: 4.
	suid
		ifTrue: [user := uid]
		ifFalse: [user := fStat at: 1].
	((uid == user) and: [(protectionMask at: 2) allMask: 1])
		ifTrue: [^ true]. "Test executable by my user ID"
	^ false

! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/14/2004 15:10'!
isLockableFile: aFileStream
	"Check for ability to place an exclusive lock on the entire file represented by
	aFileStream. An exclusive lock (write lock) permits only one OS process to hold
	the lock. Answer true if the region is lockable."

	^ self isLockableFile: aFileStream exclusive: true
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'jf 2/22/2004 17:30'!
isLockableFile: aFileStream exclusive: flag
	"Check for ability to place an exclusive lock on the entire file represented by
	aFileStream. An exclusive lock (write lock) permits only one OS process to hold
	the lock. Answer true if the region is lockable."

	| fileLock |
	aFileStream ifNil: [^ false].
	fileLock := OSFileLock onFile: aFileStream exclusive: flag.
	^ self isLockableFileRegion: fileLock
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/14/2004 15:09'!
isLockableFile: aFileStream from: start to: end
	"Pass a struct SQFile on the stack, and check for ability to lock the specified region.
	If the exclusive flag is true, then specify an exclusive (F:=WRLCK) lock on the
     file. Otherwise, specify a shared (F:=RDLCK) lock. Any number of Unix processes
     may hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock).

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	Answer true if the region is lockable."

	^ self isLockableFile: aFileStream from: start to: end exclusive: true
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/22/2004 15:10'!
isLockableFile: aFileStream from: start to: end exclusive: flag
	"Pass a struct SQFile on the stack, and check for ability to lock the specified region.
	If the exclusive flag is true, then specify an exclusive (F:=WRLCK) lock on the
     file. Otherwise, specify a shared (F:=RDLCK) lock. Any number of Unix processes
     may hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock).

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	Answer true if the region is lockable."

	| fileRegion |
	aFileStream ifNil: [^ false].
	fileRegion := OSFileRegionLock onFile: aFileStream from: start to: end exclusive: flag.
	^ self isLockableFileRegion: fileRegion
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'jf 2/22/2004 18:42'!
isLockableFileRegion: aFileLock
	"Pass a struct SQFile on the stack, and check for ability to lock the specified region.
	If the exclusive flag is true, then specify an exclusive (F:=WRLCK) lock on the
     file. Otherwise, specify a shared (F:=RDLCK) lock. Any number of Unix processes
     may hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock).

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	Answer true if the region is lockable."

	aFileLock ifNil: [^ false].
	(self canAcquireLock: aFileLock) ifFalse: [^ false].
	
	^ aFileLock test
! !

!UnixOSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 5/1/2001 23:09'!
isReadable: aPathName
	"Answer true if file at aPathName has read permission for this process."

	"OSProcess accessor isWritable: '/bin/sh'"

	^ self isReadable: aPathName forUser: self primGetUid inGroup: self primGetGid

! !

!UnixOSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 5/1/2001 23:00'!
isReadable: aPathName forUser: uid inGroup: gid
	"Answer true if file at aPathName has read permission for a user
	identified by user uid and group gid."

	"OSProcess accessor
		isReadable: '/bin/sh'
		forUser: OSProcess accessor primGetUid
		inGroup: OSProcess accessor primGetGid"

	| fStat user group protectionMask |
	fStat := self fileStat: aPathName.
	(fStat isKindOf: String)
		ifTrue: [self inform: aPathName, ': ', fStat. ^ nil].
	protectionMask := fStat at: 3.
	((protectionMask at: 4) allMask: 4)
		ifTrue: [^ true]. "Test readable by any user ID"
	group := fStat at: 2.
	((gid == group) and: [(protectionMask at: 3) allMask: 4])
		ifTrue: [^ true]. "Test readable by my group ID"
	user := fStat at: 1.
	((uid == user) and: [(protectionMask at: 2) allMask: 4])
		ifTrue: [^ true]. "Test readable by my user ID"
	^ false

! !

!UnixOSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 5/1/2001 23:09'!
isWritable: aPathName
	"Answer true if file at aPathName has write permission for this process."

	"OSProcess accessor isWritable: '/bin/sh'"

	^ self isWritable: aPathName forUser: self primGetUid inGroup: self primGetGid

! !

!UnixOSProcessAccessor methodsFor: 'file testing' stamp: 'dtl 5/1/2001 23:08'!
isWritable: aPathName forUser: uid inGroup: gid
	"Answer true if file at aPathName has read permission for a user
	identified by user uid and group gid."

	"OSProcess accessor
		isWritable: '/bin/sh'
		forUser: OSProcess accessor primGetUid
		inGroup: OSProcess accessor primGetGid"

	| fStat user group protectionMask |
	fStat := self fileStat: aPathName.
	(fStat isKindOf: String)
		ifTrue: [self inform: aPathName, ': ', fStat. ^ nil].
	protectionMask := fStat at: 3.
	((protectionMask at: 4) allMask: 2)
		ifTrue: [^ true]. "Test writable by any user ID"
	group := fStat at: 2.
	((gid == group) and: [(protectionMask at: 3) allMask: 2])
		ifTrue: [^ true]. "Test writable by my group ID"
	user := fStat at: 1.
	((uid == user) and: [(protectionMask at: 2) allMask: 2])
		ifTrue: [^ true]. "Test writable by my user ID"
	^ false

! !

!UnixOSProcessAccessor methodsFor: 'VM atexit' stamp: 'dtl 3/17/2007 18:39'!
kill: listOfPids withSignal: signumOrNil
	"On exit of the VM process, send signal to the external processes
	identified by listOfPids. If signumOrNil is nil, the default SIGTERM
	signal will be sent to listOfPids."

	^ self primKill: listOfPids withSignal: signumOrNil
! !

!UnixOSProcessAccessor methodsFor: 'VM atexit' stamp: 'dtl 3/18/2007 10:39'!
killOnVmExit: proxies withSignal: signumOrNil
	"When Squeak uses OSProcess to start a long running external process, it
	may be useful to guarantee that one or more child processes is killed when
	the Squeak VM exits, regardless of whether the normal Squeak shutdown
	processing has occurred. This method arranges for a collection of external
	process proxies to receive a signal when the Squeak VM exits. If signumOrNil
	is nil, the default SIGTERM will be used. Each invocation of this method will
	override the effects of previous calls."

	"OSProcess accessor killOnVmExit: OSProcess thisOSProcess allMyChildren withSignal: nil"

	| pids |
	pids := (proxies select: [:p | p isRunning] thenCollect: [:e | e pid]) asArray.
	^ self kill: pids withSignal: signumOrNil
! !

!UnixOSProcessAccessor methodsFor: 'debugging' stamp: 'dtl 11/19/2006 09:05'!
listSigHandlerAddressesOnConsole

	"OSProcess accessor listSigHandlerAddressesOnConsole"

	(0 to: 67) do: [:sigNum | | prev |
		[prev := OSProcess accessor currentSigHandlerAddress: sigNum.
		OSProcess debugMessage: 'signal ', sigNum asString,' hander is ', prev printString]
			on: Error
			do: [:ex | OSProcess debugMessage: ex printString]]
! !

!UnixOSProcessAccessor methodsFor: 'debugging' stamp: 'dtl 11/19/2006 09:07'!
listSigHandlerAddressesOnTranscript

	"OSProcess accessor listSigHandlerAddressesOnTranscript"

	Transcript cr.
	(0 to: 67) do: [:sigNum | | prev |
		[prev := OSProcess accessor currentSigHandlerAddress: sigNum.
		Transcript show: 'signal ', sigNum asString,' hander is ', prev printString; cr]
			on: Error
			do: [:ex | Transcript show: ex printString; cr]]
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/14/2004 15:01'!
lockFile: aFileStream
	"Request an exclusive lock on the entire file represented by aFileStream. The
	exclusive lock (write lock) permits only one OS process to hold the lock. Answer
	a descriptor for the locked file region, an Array of file handle and region interval;
	or answer nil on error.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self lockFile: aFileStream exclusive: true
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/11/2006 23:40'!
lockFile: aFileStream exclusive: flag
	"Request an exclusive lock on the entire file represented by aFileStream. The
	exclusive lock (write lock) permits only one OS process to hold the lock. Answer
	a descriptor for the locked file region, an Array of file handle and region interval;
	or answer nil on error.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self lockFile: aFileStream exclusive: flag ifFail: nil
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/11/2006 23:39'!
lockFile: aFileStream exclusive: flag ifFail: failBlock
	"Request an exclusive lock on the entire file represented by aFileStream. The
	exclusive lock (write lock) permits only one OS process to hold the lock. Answer
	a descriptor for the locked file region, an Array of file handle and region interval;
	or answer nil on error.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	| fileLock |
	aFileStream ifNil: [^ failBlock value].
	fileLock := OSFileLock onFile: aFileStream exclusive: flag.
	^ self lockFileRegion: fileLock ifFail: failBlock
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/14/2004 15:01'!
lockFile: aFileStream from: start to: end
	"Pass a struct SQFile on the stack, and request a lock on the specified region.
	If the exclusive flag is true, then request an exclusive (F:=WRLCK) lock on the
     file. Otherwise, request a shared (F:=RDLCK) lock. Any number of Unix processes
     may hold a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock). Answer a descriptor for the locked file region,
	an Array of file handle and region interval; or answer nil on error.

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self lockFile: aFileStream from: start to: end exclusive: true
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/22/2004 15:10'!
lockFile: aFileStream from: start to: end exclusive: flag
	"Pass a struct SQFile on the stack, and request a lock on the specified region.
	If the exclusive flag is true, then request an exclusive (F:=WRLCK) lock on the
     file. Otherwise, request a shared (F:=RDLCK) lock. Any number of Unix processes
     may hold a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock). Answer a descriptor for the locked file region,
	an Array of file handle and region interval; or answer nil on error.

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	| fileRegion |
	aFileStream ifNil: [^ nil].
	fileRegion := OSFileRegionLock onFile: aFileStream from: start to: end exclusive: flag.
	^ self lockFileRegion: fileRegion
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 8/30/2009 14:51'!
lockFile: fileStream from: start to: end exclusive: writeLockFlag ifFail: failBlock

	| lock |
	lock := OSFileRegionLock onFile: fileStream from: start to: end exclusive: writeLockFlag.
	^ self lockFileRegion: lock ifFail: [failBlock value]! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 11/28/2010 19:07'!
lockFile: fileStream from: start to: end ifFail: failBlock

	^ self lockFile: fileStream from: start to: end exclusive: true ifFail: failBlock
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/11/2006 23:41'!
lockFile: aFileStream ifFail: failBlock
	"Request an exclusive lock on the entire file represented by aFileStream. The
	exclusive lock (write lock) permits only one OS process to hold the lock. Answer
	a descriptor for the locked file region, an Array of file handle and region interval;
	or answer nil on error.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self lockFile: aFileStream exclusive: true ifFail: failBlock
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/8/2006 07:04'!
lockFileRegion: aFileLock
	"Pass a struct SQFile on the stack, and request a lock on the specified region.
	If the exclusive flag is true, then request an exclusive (F:=WRLCK) lock on the
     file. Otherwise, request a shared (F:=RDLCK) lock. Any number of Unix processes
     may hold a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock). Answer a descriptor for the locked file region,
	an Array of file handle and region interval; or answer nil on error.

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self lockFileRegion: aFileLock ifFail: [nil]
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/8/2006 07:03'!
lockFileRegion: aFileLock ifFail: failBlock
	"Pass a struct SQFile on the stack, and request a lock on the specified region.
	If the exclusive flag is true, then request an exclusive (F:=WRLCK) lock on the
     file. Otherwise, request a shared (F:=RDLCK) lock. Any number of Unix processes
     may hold a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock). Answer a descriptor for the locked file region,
	an Array of file handle and region interval; or answer nil on error.

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	aFileLock ifNil: [^ failBlock value].
	"Check region lock overlap for Win32 compatibility"
	(self canAcquireLock: aFileLock) ifFalse: [^ failBlock value].
	
	aFileLock lock
		ifFalse: [^ failBlock value]
		ifTrue:
			[self register: aFileLock.
			^ aFileLock]
! !

!UnixOSProcessAccessor methodsFor: 'pipe open' stamp: 'dtl 1/31/2004 19:08'!
makePipeHandles
	"Create a pipe, and answer an array of two IO accessors for the pipe 
	reader and writer. Set a signal handler to cause SIGPIPE signals to be
	ignored. This will register a Semaphore to receive the SIGPIPE events,
	but we will simply ignore the semaphore, effectively ignoring the
	external OS signals. This is done instead of explicitly telling the OS to
	ignore the signals because it allows the use of a uniform signal
	forwarding mechanism in Squeak, even for signals that are ultimately
	ignored."

	self forwardSigPipe.
	^ super makePipeHandles
! !

!UnixOSProcessAccessor methodsFor: 'temp files and directories' stamp: 'dtl 6/3/2013 22:05'!
mkdtemp: templateString
	"Call mkdtemp() to generate a uniquely named temporary directory from a template
	string. The last six or more characters of template must be XXXXXX and these are
	replaced with characters that make the directory name unique. The directory is then
	created with permissions 0700. Answer the new directory name."

	"OSProcess accessor mkdtemp: 'foodir.XXXXXX' "

	^self primMkdtemp: templateString
! !

!UnixOSProcessAccessor methodsFor: 'temp files and directories' stamp: 'dtl 6/3/2013 22:36'!
mkstemp: templateString

	"Call mkstemp() to generate a unique temporary file from a template string.
	The last six or more characters of template must be XXXXXX and these are
	replaced with a string that make the filename unique. Contents of the template
	string are modified. The file is created with permissions 0600, that is, read plus
	write for owner only. Answer an opened file stream on the newly created file."

	"OSProcess accessor mkstemp: 'foo.XXXXXX' "

	" | s | { s := 'foo.XXXXXX' . OSProcess accessor mkstemp: s } "

	| fileName sqFile |
	fileName := templateString copy.
	sqFile := self primMkstemp: fileName.
	^ AttachableFileStream
		name: fileName
		attachTo: sqFile
		writable: true
! !

!UnixOSProcessAccessor methodsFor: 'temp files and directories' stamp: 'dtl 6/3/2013 21:59'!
mktemp: templateString
	"Call mktemp() to generate a unique temporary filename from a template string.
	The last six or more characters of template must be XXXXXX and these are replaced
	with a string that make the filename unique. The templateString must be a mutable
	string, because the primitive will change its contents.
	
	Use primitiveMkstemp to avoid file creation race condition, see man 3 mktemp."

	"OSProcess accessor mktemp: 'foo.XXXXXX' "
	"OSProcess accessor mktemp: '/tmp/foo.XXXXXX' "

	^ self primMktemp: templateString
! !

!UnixOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 10/8/2005 09:55'!
nice: inc
	"Change the scheduling priority of this OS process by the given nice increment.
	A positive increment decreases the priority. Only the superuser can specify
	a negative value to increase the priority. A typical use is to increase the
	nice value by 1 in order to make the Squeak VM run at lower priority. This
	may be useful for a background Squeak doing an image save or other non-
	interactive process."

	"OSProcess accessor nice: 1"

	^ self primNice: inc
! !

!UnixOSProcessAccessor methodsFor: 'display handling - old plugin compatibility' stamp: 'dtl 8/3/2003 20:34'!
oldPrimCanConnectToXDisplay: xDisplayName

	"Deprecated - install the XDisplayControlPlugin to eliminate the need to call this method"

	"Open and close a connection to displayName. It the connection was successfully
	opened, answer true; otherwise false. This is intended to check for the ability
	to open an X display prior to actually making the attempt."

	"OSProcess accessor primCanConnectToXDisplay: ':0.0' "
	"OSProcess accessor primCanConnectToXDisplay: ':1' "

	<primitive: 'primitiveCanConnectToXDisplay' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'display handling - old plugin compatibility' stamp: 'dtl 8/3/2003 20:34'!
oldPrimGetXDisplayName

	"Deprecated - install the XDisplayControlPlugin to eliminate the need to call this method"

	"Answer a string containing the name for the X display, or nil if the display was opened
	using the $DISPLAY environment variable. This answers the name of the X display as of
	the time it was last opened, which may be different from the current setting of $DISPLAY."

	"OSProcess accessor primGetXDisplayName"

	<primitive: 'primitiveGetXDisplayName' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'display handling - old plugin compatibility' stamp: 'dtl 8/3/2003 20:35'!
oldPrimIsConnectedToXServer

	"Deprecated - install the XDisplayControlPlugin to eliminate the need to call this method"

	"Answer true if VM is currently connected to an X server."

	"OSProcess accessor primIsConnectedToXServer inspect"
	"| x |
	OSProcess accessor primKillDisplay.
	x := OSProcess accessor primIsConnectedToXServer.
	OSProcess accessor primOpenXDisplay.
	x inspect"

	<primitive: 'primitiveIsConnectedToXServer' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'display handling - old plugin compatibility' stamp: 'dtl 8/3/2003 20:35'!
oldPrimKillDisplay

	"Deprecated - install the XDisplayControlPlugin to eliminate the need to call this method"

	"Call an internal function which will disconnect the X display session."

	"OSProcess thisOSProcess processAccessor primKillDisplay"

	<primitive: 'primitiveKillDisplay' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'display handling - old plugin compatibility' stamp: 'dtl 8/3/2003 20:35'!
oldPrimOpenXDisplay

	"Deprecated - install the XDisplayControlPlugin to eliminate the need to call this method"

	"Call an internal function which will open the X display session."

	"OSProcess thisOSProcess processAccessor primKillDisplay.
	(Delay forSeconds: 5) wait.
	OSProcess thisOSProcess processAccessor primOpenXDisplay"

	<primitive: 'primitiveOpenXDisplay' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'display handling - old plugin compatibility' stamp: 'dtl 8/3/2003 20:35'!
oldPrimSetXDisplayName: aStringOrNil

	"Deprecated - install the XDisplayControlPlugin to eliminate the need to call this method"

	"Set the name for the X display for use in the next call to primitiveOpenXDisplay.
	aStringOrNil must be either a String (such as 'myhost:0') or nil, indicating that
	the current value of $DISPLAY should be used."

	"OSProcess accessor primSetXDisplayName: ':0.0' "
	"OSProcess accessor primSetXDisplayName: nil "

	<primitive: 'primitiveSetXDisplayName' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 11:48'!
osppModuleName
	"Answer a string containing the module name string for the OSPP plugin."

	"OSProcess accessor osppModuleName"

	^ self primOSProcessPluginModuleName
! !

!UnixOSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 11:49'!
osppModuleVersionString
	"Answer a string containing the version string for the OSPP plugin."

	"OSProcess accessor osppModuleVersionString"

	^ self primOSProcessPluginModuleVersionString
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/1/2005 10:37'!
primAioPluginModuleName
	"Answer a string containing the module name string for the AIO plugin."

	"OSProcess accessor primAioPluginModuleName"

	<primitive: 'primitiveModuleName' module: 'AioPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/1/2005 10:38'!
primAioPluginVersionString
	"Answer a string containing the version string for the AIO plugin."

	"OSProcess accessor primAioPluginVersionString"

	<primitive: 'primitiveVersionString' module: 'AioPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 7/22/2000 17:19'!
primArgumentAt: index
	"Answer the argument string in the argument OS process argument list at position index.
	In Unix, the first element of the list is the program name, and any additional elements
	of the list are optional command line arguments passed to the program. This convention
	may be simulated by the C runtime libraries on other operating systems, but argument
	list handling should be assumed to be operating system dependent."

	<primitive: 'primitiveArgumentAt' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'display handling' stamp: 'dtl 8/3/2003 18:26'!
primCanConnectToXDisplay: xDisplayName
	"Open and close a connection to displayName. It the connection was successfully
	opened, answer true; otherwise false. This is intended to check for the ability
	to open an X display prior to actually making the attempt."

	"OSProcess accessor primCanConnectToXDisplay: ':0.0' "
	"OSProcess accessor primCanConnectToXDisplay: ':1' "

	<primitive: 'primitiveCanConnectToDisplay' module: 'XDisplayControlPlugin'>
	^ self oldPrimCanConnectToXDisplay: xDisplayName

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 2/24/2001 14:28'!
primCanReceiveSignals: anIntegerPid
	"Answer true if an external OS process can receive signals. In most
	cases, if the process identified by anIntegerPid cannot receive signals, it is
	because the process does not exist and anIntegerPid is a stale reference (possibly
	left over from a previous Squeak session). Answer nil if the primitive does
	not exist (possibly because the VM is using an older version of the plugin)."

	<primitive: 'primitiveCanReceiveSignals' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 3/31/2001 11:18'!
primChdir: pathString
	"Change current working directory. Does not update the $PWD environment variable."

	<primitive: 'primitiveChdir' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 11/28/2010 12:31'!
primCreatePipe
	"Create a pipe, and answer an array of two file handles (SQFile data structures in interp.c)
	for the pipe reader and writer."

	<primitive: 'primitiveCreatePipe' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 11/28/2010 12:32'!
primCreatePipeWithSessionIdentifier: aByteArray
	"Create a pipe, and answer an array of two file handles (SQFile data structures in interp.c)
	for the pipe reader and writer."

	<primitive: 'primitiveCreatePipeWithSessionIdentifier' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'display handling' stamp: 'dtl 8/6/2003 06:24'!
primDisconnectDisplay
	"Disconnect the X display session. The actual Squeak window on the X server is not
	effected, but this instance of Squeak will not have any further interaction with it."

	"OSProcess thisOSProcess processAccessor primDisconnectDisplay"

	<primitive: 'primitiveDisconnectDisplay' module: 'XDisplayControlPlugin'>
	^ self oldPrimKillDisplay

! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 9/10/2011 23:28'!
primDup: fileDescriptor
	"Call dup2 to duplicate a file descriptor to the next available descriptor. Answer
	the new file descriptor or -1 on failure."

	<primitive: 'primitiveDup' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 5/17/2009 21:57'!
primDup: oldFileDescriptor to: newFileDescriptor
	"Call dup2() to duplicate a file descriptor. Answer the duplicated file descriptor
	or -1 on failure."

	<primitive: 'primitiveDupTo' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed
! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 7/22/2000 17:20'!
primEnvironmentAt: index
	"Answer the environment string at index position in the OS process environment list.
	This returns a 'KEY=value' string, which the caller is expected to parse into #KEY
	and 'value' to be stored an environment dictionary."

	<primitive: 'primitiveEnvironmentAt' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:28'!
primEnvironmentAtSymbol: aSymbol
	"Answer the value of an environment variable in the external OS process."

	<primitive: 'primitiveEnvironmentAtSymbol' module: 'UnixOSProcessPlugin'>
	^ nil ! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/31/2001 11:03'!
primErrorMessageAt: anInteger
	"Answer an error message string from the sys:=errlist array, indexed by anInteger."

	"OSProcess accessor primErrorMessageAt: 0"
	"OSProcess accessor primErrorMessageAt: 100"
	"OSProcess accessor primErrorMessageAt: 1000"
	"OSProcess accessor primErrorMessageAt: -1"

	<primitive: 'primitiveErrorMessageAt' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 4/30/2001 05:49'!
primFileProtectionMask: aPathString
	"Call stat(2) to obtain the file protection mask for a file. Answer an Array of
	four integers representing the protection mask, or answer errno on failure. The
	protection mask is four Integers, each of which may be considered an octal digit
	(0-7), with bit values 4, 2, and 1. The first digit selects the set user ID (4) and set
	group ID (2) and save text image (1) attributes. The second digit selects permissions
	for the user who owns the file: read (4), write (2), and execute (1); the third
	selects permissions for other users in the file's group, with the same values; and
	the fourth for other users not in the file's group, with the same values."

	<primitive: 'primitiveFileProtectionMask' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 4/30/2001 20:56'!
primFileStat: aPathString
	"Call stat(2) to obtain the file protection mask for a file. Answer errno on failure,
	or on success answer an array with: UID with: GID with: protectionMask. The	
	protectionMask is an Array of four integers representing the protection mask, or
	answer errno on failure. The protection mask is four Integers, each of which may
	be considered an octal digit (0-7), with bit values 4, 2, and 1. The first digit selects
	the set user ID (4) and set group ID (2) and save text image (1) attributes. The second
	digit selects permissions for the user who owns the file: read (4), write (2), and
	execute (1); the third selects permissions for other users in the file's group, with
	the same values; and the fourth for other users not in the file's group, with the
	same values."

	<primitive: 'primitiveFileStat' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'display handling' stamp: 'dtl 8/24/2003 10:13'!
primFlushXDisplay
	"Call an internal function to synchronize output to the X display."

	"OSProcess thisOSProcess processAccessor primFlushXDisplay"

	<primitive: 'primitiveFlushDisplay' module: 'XDisplayControlPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 11/28/2010 12:35'!
primForkExec: executableFile
	stdIn: inputFileHandle
	stdOut: outputFileHandle
	stdErr: errorFileHandle
	argBuf: argVec
	argOffsets: argOffsets
	envBuf: envVec
	envOffsets: envOffsets
	workingDir: pathString

	"Parameters are expected to have been properly prepared by the caller, including string
	values which are to be null terminated strings. In other words, all strings should have
	(Character value: 0) as the last element in the string.

	Parameters should be:
		executableFile: a string with the name of a file to execute.
		stdIn: a fileID ByteArray (struct SQFile in C) to be used as standard input.
		stdOut: a fileID ByteArray to be used as standard output.
		stdErr: a fileID ByteArray to be used as standard error.
		argVec: a String arranged to look more or less like a char **, but with the addresses not yet fixed.
		argOffsets: an Array of offsets for fixing up the argVec addresses.
		envVec: a String arranged to look more or less like a char **, but with the addresses not yet fixed.
		envOffsets: an Array of offsets for fixing up the envVec addresses.
		workingDir: a null terminated path name String, or nil.
	The envVec parameter may be nil, in which case envOffsets is ignored. workingDir may be nil. The
	other parameters are required. Parameters with nil value indicate that current values for this process
	should be used."

	<primitive: 'primitiveForkExec' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 11/28/2010 12:37'!
primForkSqueak
	"Clone this Squeak Smalltalk image in a child OSProcess. The child is the same as the
	parent, except for its new X session connection, and the return value of this method,
	which is zero for the child process, and a positive integer equal to the pid of the child
	for the parent process.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	<primitive: 'primitiveForkSqueakWithoutSigHandler' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 8/3/2002 20:29'!
primForwardSignal: signalNumber toSemaphore: semaphoreIndex
	"Set a signal handler in the VM which will signal a Smalltalk semaphore at
	semaphoreIndex whenever an external signal sigNum is received. Answer the
	prior value of the signal handler. If semaphoreIndex is zero, the handler is
	unregistered, and the VM returns to its default behavior for handling that
	signal.

	The Smalltalk semaphore is expected to be kept at the same index location
	indefinitely during the lifetime of a Squeak session. If that is not the case, the
	handler must be unregistered prior to unregistering the Smalltalk semaphore."

	<primitive: 'primitiveForwardSignalToSemaphore' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 10/14/2001 11:15'!
primGetChildExitStatus: childPid
	"Clean up after the death of a child process, and answer the exit status of the child process."

	<primitive: 'primitiveReapChildProcess' module: 'UnixOSProcessPlugin'>
	^ Array with: childPid with: nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 7/22/2000 17:20'!
primGetCurrentWorkingDirectory
	"Call getcwd() to get the current working directory."

	"OSProcess thisOSProcess processAccessor primGetCurrentWorkingDirectory"

	<primitive: 'primitiveGetCurrentWorkingDirectory' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 4/30/2001 21:07'!
primGetEGid
	"Answer the effective group ID for the OS process in which I am currently executing."

	"OSProcess accessor primGetEGid"

	<primitive: 'primitiveGetEGid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 4/30/2001 21:01'!
primGetEUid
	"Answer the effective user ID for the OS process in which I am currently executing."

	"OSProcess accessor primGetEUid"

	<primitive: 'primitiveGetEUid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 4/30/2001 21:01'!
primGetGid
	"Answer the group ID for the OS process in which I am currently executing."

	"OSProcess accessor primGetGid"

	<primitive: 'primitiveGetGid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/15/2007 19:43'!
primGetPGid: pid
	"Answer the process group ID of the process identified by pid"

	<primitive: 'primitiveGetPGid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/15/2007 19:39'!
primGetPGrp
	"Answer the process group ID of this OS process"

	<primitive: 'primitiveGetPGrp' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 7/22/2000 17:20'!
primGetPPid
	"Answer the OS process ID for the parent process of the OS process in which I am
	currently executing."

	<primitive: 'primitiveGetPPid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 7/22/2000 17:20'!
primGetPid
	"Answer the OS process ID for the OS process in which I am currently executing."

	<primitive: 'primitiveGetPid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/2/2002 08:15'!
primGetSession
	"Answer the unique identifier for this session of Smalltalk running in this OS Process."

	"OSProcess accessor primGetSession"

	<primitive: 'primitiveGetSession' module: 'UnixOSProcessPlugin'>

	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:28'!
primGetStdErrHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard error for the
	OS process in which I am currently executing."

	<primitive: 'primitiveGetStdErrHandle' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:28'!
primGetStdInHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard input for
	the OS process in which I am currently executing."

	<primitive: 'primitiveGetStdInHandle' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:28'!
primGetStdOutHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard output for the
	OS process in which I am currently executing."

	<primitive: 'primitiveGetStdOutHandle' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'pthreads' stamp: 'dtl 3/5/2006 10:51'!
primGetThreadID
	"Answer the ID of the pthread that is currently executing (the interpreter thread).
	A thread ID may be a 64 bit value on some platforms, so answer a byte array
	containing the value in machine-dependent byte order."

	"OSProcess accessor primGetThreadID"

	<primitive: 'primitiveGetThreadID' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 4/30/2001 21:00'!
primGetUid
	"Answer the user ID for the OS process in which I am currently executing."

	"OSProcess accessor primGetUid"

	<primitive: 'primitiveGetUid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'display handling' stamp: 'dtl 8/3/2003 18:27'!
primGetXDisplayName
	"Answer a string containing the name for the X display, or nil if the display was opened
	using the $DISPLAY environment variable. This answers the name of the X display as of
	the time it was last opened, which may be different from the current setting of $DISPLAY."

	"OSProcess accessor primGetXDisplayName"

	<primitive: 'primitiveGetDisplayName' module: 'XDisplayControlPlugin'>
	^ self oldPrimGetXDisplayName

! !

!UnixOSProcessAccessor methodsFor: 'display handling' stamp: 'dtl 8/3/2003 18:29'!
primIsConnectedToXServer
	"Answer true if VM is currently connected to an X server."

	"OSProcess accessor primIsConnectedToXServer inspect"
	"| x |
	OSProcess accessor primKillDisplay.
	x := OSProcess accessor primIsConnectedToXServer.
	OSProcess accessor primOpenXDisplay.
	x inspect"

	<primitive: 'primitiveIsConnectedToDisplay' module: 'XDisplayControlPlugin'>
	^ self oldPrimIsConnectedToXServer

! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 3/18/2007 10:35'!
primKill: listOfPids withSignal: signumOrNil
	"Set a list of pids to kill with signum when VM exits. If the signum
	parameter is nil, the default value of SIGTERM will be used."

	<primitive: 'primitiveKillOnExit' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed! !

!UnixOSProcessAccessor methodsFor: 'display handling' stamp: 'dtl 8/6/2003 06:22'!
primKillDisplay
	"Disconnect the X display session and destroy the Squeak window on the X display."

	"OSProcess thisOSProcess processAccessor primKillDisplay"

	<primitive: 'primitiveKillDisplay' module: 'XDisplayControlPlugin'>
	^ self oldPrimKillDisplay

! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 1/3/2004 19:59'!
primLockFileRegion: aSQFileStruct offset: begin length: len exclusive: flag
	"Pass a struct SQFile on the stack, and request a lock on the specified region.
	If the exclusive flag is true, then request an exclusive (F:=WRLCK) lock on the
     file. Otherwise, request a shared (F:=RDLCK) lock. Any number of Unix processes
     may  hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock). Answer the result of the call to fcntl()."

	<primitive: 'primitiveLockFileRegion' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 6/1/2013 17:14'!
primMkdtemp: templateString
	"Call mkdtemp() to generate a uniquely named temporary directory from a template
	string. The last six characters of template must be XXXXXX and these are replaced
	with characters that make the directory name unique. The directory is then created
	with permissions 0700. Answer the new directory name."

	<primitive: 'primitiveMkdtemp' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 6/3/2013 22:33'!
primMkstemp: templateString
	"Call mkstemp() to generate a unique temporary file from a template string.
	The last six characters of template must be XXXXXX and these are replaced with
	a string that make the filename unique. Answer a SQFile descriptor for the file stream.
	
	Contents of the template string are modified, and provide the name of the newly
	created file.

	The  file  is  created  with permissions 0600, that is, read plus write for owner only.

	Sender must convert the file descriptor to a FileStream before it can be used."

	<primitive: 'primitiveMkstemp' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 6/1/2013 17:14'!
primMktemp: templateString
	"Call mktemp() to generate a unique temporary filename from a template string.
	The last six characters of template must be XXXXXX and these are replaced with
	a string that make the filename unique. The templateString must be a mutable
	string, because the primitive will change its contents.
	
	Use primitiveMkstemp to avoid file creation race condition, see man 3 mktemp."

	<primitive: 'primitiveMktemp' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed
! !

!UnixOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 10/8/2005 09:55'!
primNice: inc
	"Change the scheduling priority of this OS process by the given nice increment."

	<primitive: 'primitiveNice' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/1/2005 11:48'!
primOSProcessPluginModuleName
	"Answer a string containing the module name string for the OSPP plugin."

	"OSProcess accessor primOSProcessPluginModuleName"

	<primitive: 'primitiveModuleName' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/1/2005 11:48'!
primOSProcessPluginModuleVersionString
	"Answer a string containing the version string for the OSPP plugin."

	"OSProcess accessor primOSProcessPluginModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'display handling' stamp: 'dtl 8/3/2003 18:31'!
primOpenXDisplay
	"Call an internal function which will open the X display session."

	"OSProcess thisOSProcess processAccessor primKillDisplay.
	(Delay forSeconds: 5) wait.
	OSProcess thisOSProcess processAccessor primOpenXDisplay"

	<primitive: 'primitiveOpenDisplay' module: 'XDisplayControlPlugin'>
	^ self oldPrimOpenXDisplay

! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 7/19/2010 23:38'!
primPrintAllStacksOnSignal: signalNumber
	"Set a signal handler in the VM which will print all call stacks on the console
	output whenever an external signal sigNum is received. Answer the
	prior value of the signal handler."

	"OSProcess accessor primPrintAllStacksOnSignal: OSProcess accessor primSigUsr1Number"

	"OSProcess accessor primForwardSignal: OSProcess accessor primSigUsr1Number toSemaphore: nil"

	<primitive: 'primitivePrintAllStacksOnSignal' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:28'!
primPutEnv: aString
	"Add or update an environment variable in the external OS process using a 'KEY=value'
	string."

	<primitive: 'primitivePutEnv' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 3/31/2001 15:44'!
primRealpath: pathString
	"Resolve pathString into a real path if possible, or answer nil."

	<primitive: 'primitiveRealpath' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:28'!
primSQFileFlush: aSQFileStruct
	"Pass a struct SQFile on the stack, flush the external file stream."

	<primitive: 'primitiveSQFileFlush' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:29'!
primSQFileSetBlocking: aSQFileStruct
	"Pass a struct SQFile on the stack, and call fcntl() to set the file non-blocking."

	<primitive: 'primitiveSQFileSetBlocking' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:29'!
primSQFileSetNonBlocking: aSQFileStruct
	"Pass a struct SQFile on the stack, and call fcntl() to set the file non-blocking."

	<primitive: 'primitiveSQFileSetNonBlocking' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 7/22/2000 17:29'!
primSQFileSetUnbuffered: aSQFileStruct
	"Pass a struct SQFile on the stack, set the file non-blocking."

	<primitive: 'primitiveSQFileSetUnbuffered' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 8/4/2002 01:13'!
primSemaIndexFor: sigNum
	"Answer the registration index of the semaphore currently associated with the
	signal handler for sigNum."

	<primitive: 'primitiveSemaIndexFor' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:24'!
primSendSigabrtTo: anIntegerPid
	"Send SIGABRT (abort) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigabrtTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:24'!
primSendSigalrmTo: anIntegerPid
	"Send SIGALRM (alarm) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigalrmTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:24'!
primSendSigchldTo: anIntegerPid
	"Send SIGCHLD (child status has changed, usually death of child) to the OS process
	identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigchldTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:25'!
primSendSigcontTo: anIntegerPid
	"Send SIGCONT (continue) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigcontTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:25'!
primSendSighupTo: anIntegerPid
	"Send SIGHUP (hangup) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSighupTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:25'!
primSendSigintTo: anIntegerPid
	"Send SIGINT (interrupt) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigintTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:25'!
primSendSigkillTo: anIntegerPid
	"Send SIGKILL (kill, unblockable) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigkillTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:25'!
primSendSigpipeTo: anIntegerPid
	"Send SIGPIPE (broken pipe) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigpipeTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:25'!
primSendSigquitTo: anIntegerPid
	"Send SIGQUIT (quit) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigquitTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:25'!
primSendSigstopTo: anIntegerPid
	"Send SIGSTOP (stop, unblockable) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigstopTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:26'!
primSendSigtermTo: anIntegerPid
	"Send SIGTERM (termination) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigtermTo' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:26'!
primSendSigusr1To: anIntegerPid
	"Send SIGUSR1 (User-defined signal 1) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigusr1To' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'OS signal sending' stamp: 'dtl 7/22/2000 17:26'!
primSendSigusr2To: anIntegerPid
	"Send SIGUSR2 (User-defined signal 2) to the OS process identified by anIntegerPid.
	Answer 0 on success, -1 on failure, and nil if the pluggable primitive is not present."

	<primitive: 'primitiveSendSigusr2To' module: 'UnixOSProcessPlugin'>
	^ nil

! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/15/2007 19:47'!
primSet: pid pGid: processGroupId
	"Set the process group ID of the process identified by pid to a new process
	group ID."

	<primitive: 'primitiveSetPGid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/15/2007 19:40'!
primSetPGrp
	"Set a new process group for this OS process. Newly created child processes
	will be members of the new process group."

	<primitive: 'primitiveSetPGrp' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/18/2007 10:50'!
primSetSid
	"Quoted from Linux man pages:
	setsid() creates a new session if the calling process is not a process group leader.
	The calling process is the leader of the new session, the process  group  leader  of
	the new process group, and has no controlling tty.  The process group ID and session
	ID of the calling process are set to the PID of the calling  process.   The  calling
	process  will be the only process in this new process group and in this new session."

	<primitive: 'primitiveSetSid' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'display handling' stamp: 'dtl 8/3/2003 18:32'!
primSetXDisplayName: aStringOrNil
	"Set the name for the X display for use in the next call to primitiveOpenXDisplay.
	aStringOrNil must be either a String (such as 'myhost:0') or nil, indicating that
	the current value of $DISPLAY should be used."

	"OSProcess accessor primSetXDisplayName: ':0.0' "
	"OSProcess accessor primSetXDisplayName: nil "

	<primitive: 'primitiveSetDisplayName' module: 'XDisplayControlPlugin'>
	^ self oldPrimSetXDisplayName: aStringOrNil

! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 8/3/2002 20:36'!
primSigChldNumber
	"Integer value corresponding to SIGCHLD"

	<primitive: 'primitiveSigChldNumber' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 12/6/2002 18:04'!
primSigHupNumber
	"Integer value corresponding to SIGHUP"

	<primitive: 'primitiveSigHupNumber' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 8/4/2002 00:19'!
primSigIntNumber
	"Integer value corresponding to SIGINT"

	"OSProcess accessor primSigIntNumber"

	<primitive: 'primitiveSigIntNumber' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 12/6/2002 18:05'!
primSigKillNumber
	"Integer value corresponding to SIGKILL"

	"OSProcess accessor primSigKillNumber"

	<primitive: 'primitiveSigKillNumber' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 8/3/2002 20:35'!
primSigPipeNumber
	"Integer value corresponding to SIGPIPE"

	<primitive: 'primitiveSigPipeNumber' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 12/6/2002 18:06'!
primSigQuitNumber
	"Integer value corresponding to SIGQUIT"

	<primitive: 'primitiveSigQuitNumber' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 12/6/2002 18:06'!
primSigTermNumber
	"Integer value corresponding to SIGTERM"

	<primitive: 'primitiveSigTermNumber' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 11/4/2005 06:41'!
primSigUsr1Number
	"Integer value corresponding to SIGUSR1"

	<primitive: 'primitiveSigUsr1Number' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - signal forwarding' stamp: 'dtl 11/4/2005 06:42'!
primSigUsr2Number
	"Integer value corresponding to SIGUSR2"

	<primitive: 'primitiveSigUsr2Number' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 8/7/2005 12:33'!
primSizeOfInt
	"Size of an integer for this C compiler on this machine."

	<primitive: 'primitiveSizeOfInt' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 8/7/2005 12:33'!
primSizeOfPointer
	"Size of a void pointer for this C compiler on this machine."

	<primitive: 'primitiveSizeOfPointer' module: 'UnixOSProcessPlugin'>
	^ self primitiveFailed
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 11/28/2010 14:06'!
primTestEndOfFileFlag: aSqFileStruct
	"Answer whether the file represented by aSqFileStruct is at end of file, as determined
	by a call to feof(). This is different from StandardFileStream>>primAtEnd: which answers
	true if the file pointer is at the end of the file, but which does not call feof() to
	determine that an end of file condition has occurred. The difference is significant
	if aSqFileStruct represents a pipe or a device file, which may not be positionable
	in the sense of a conventional disk file."

	<primitive: 'primitiveTestEndOfFileFlag' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 1/4/2004 15:29'!
primTestLockableFileRegion: aSQFileStruct offset: begin length: len exclusive: flag
	"Pass a struct SQFile on the stack, and check for ability to lock the specified region.
	If the exclusive flag is true, then specify an exclusive (F:=WRLCK) lock on the
     file. Otherwise, specify a shared (F:=RDLCK) lock. Any number of Unix processes
     may hold  a read lock (shared lock) on a file region, but only one process may
     hold a write lock (exclusive lock).

	If length is zero, then the request is for the entire file to be locked, including
	region extents that have not yet been allocated for the file.

	If the fcntl() call fails, answer -1 (the result of the failed call). Otherwise,
	answer an array with the following six fields:
		lockable (true or false)
		l:=pid (pid of the process preventing this lock request, or nil)
		l:=type (request type F:=WRLCK or F:=RDLOCK of the process preventing this lock request)
		l:=whence (the SEEK:=SET, SEEK:=CUR, or SEEK:=END value of the lock preventing this lock request).
		l:=start (offset of the region lock preventing this lock request)
		l:=len (length of the region lock preventing this lock request)"

	<primitive: 'primitiveTestLockableFileRegion' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 5/17/2009 21:07'!
primUnixFileClose: integerFileNumber
	"Close a file handle at the close(2) level, using a handle returned by
	#primUnixFileNumber."

	<primitive: 'primitiveUnixFileClose' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 2/11/2001 16:23'!
primUnixFileNumber: aSQFileStruct
	"Pass a struct SQFile on the stack, and answer the corresponding Unix file number."

	<primitive: 'primitiveUnixFileNumber' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 1/3/2004 23:08'!
primUnlockFileRegion: aSQFileStruct offset: begin length: len
	"Pass a struct SQFile on tthe stack, and unlock the specified region.
	Answer the result of the call to fcntl(). If the region is in the file lock cache,
	remove it, but otherwise ignore the cache. The cache supports Win32 semantics
	within a single Squeak image, but not across separate images, therefore the
	unlock should be attempted regardless of whether this image thinks that the
	region has previously been locked. Answer the result of the call to fcntl()."

	<primitive: 'primitiveUnlockFileRegion' module: 'UnixOSProcessPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/7/2001 00:37'!
primUnsetEnv: aKeyString
	"Remove an environment variable from the external OS process environment."

	<primitive: 'primitiveUnsetEnv' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 8/30/2003 18:29'!
primXDisplayControlPluginModuleName
	"Answer a string containing the module name string for the display control plugin."

	"OSProcess accessor primXDisplayControlPluginModuleName"

	<primitive: 'primitiveModuleName' module: 'XDisplayControlPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 8/30/2003 18:32'!
primXDisplayControlPluginModuleVersionString
	"Answer a string containing the version string for the display control plugin."

	"OSProcess accessor primXDisplayControlPluginModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'XDisplayControlPlugin'>
	^ nil
! !

!UnixOSProcessAccessor methodsFor: 'printing' stamp: 'dtl 9/10/2000 10:16'!
printOn: aStream
	"In English, say 'a Unix' rather than 'an Unix'. Therefore do not use super printOn, which
	treats $U as a vowel."

	aStream
		nextPutAll: 'a ';
		nextPutAll: self class name;
		nextPutAll: ' on pid ';
		nextPutAll: self primGetPid printString
! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 9/10/2000 11:59'!
putPath: aString
	"Convenience method. Set the environment PATH variable to aString."

	| pathString |
	pathString := 'PATH=', aString, ((Character value: 0) asString). 
	^ self environmentPut: pathString




! !

!UnixOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 1/25/2013 18:59'!
realpath: pathString
	"Get the real path for pathString from the external OS process."

	"OSProcess accessor realpath: '/tmp'"
	"OSProcess accessor realpath: OSProcess defaultPathString"
	"OSProcess accessor realpath: '/bogus/path/name'"

	^ self primRealpath: pathString
! !

!UnixOSProcessAccessor methodsFor: 'file lock registry' stamp: 'dtl 3/5/2005 13:10'!
register: fileRegionLock
	"If an object equal to fileRegionLock exists in the registry, answer it. Otherwise, add
	fileRegionLock to the registry and answer fileRegionLock. Caching is enabled when
	EmulateWin32FileLocking is true."

	^ self emulateWin32FileLocking
		ifTrue: [super register: fileRegionLock]
		ifFalse: [fileRegionLock]
! !

!UnixOSProcessAccessor methodsFor: 'initialize - release' stamp: 'dtl 8/15/2010 15:30'!
restartChildWatcherProcess
	"OSProcess accessor restartChildWatcherProcess"

	self finalize.
	^ self grimReaperProcess
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:45'!
restoreSigChld
	"Unset a SIGCHLD signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigChld"

	^ self restoreSignal: self primSigChldNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:47'!
restoreSigHup
	"Unset a SIGHUP signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigHup"

	^ self restoreSignal: self primSigHupNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:46'!
restoreSigInt
	"Unset a SIGINT signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigInt"

	^ self restoreSignal: self primSigIntNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:48'!
restoreSigKill
	"Unset a SIGKILL signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigKill"

	self notify: 'SIGKILL and SIGSTOP signals cannot be caught, see man signal(2)'.
	^ self restoreSignal: self primSigIntNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:48'!
restoreSigPipe
	"Unset a SIGPIPE signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigPipe"

	^ self restoreSignal: self primSigPipeNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:49'!
restoreSigQuit
	"Unset a SIGQUIT signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigQuit"

	^ self restoreSignal: self primSigQuitNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:49'!
restoreSigTerm
	"Unset a SIGTERM signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigTerm"

	^ self restoreSignal: self primSigTermNumber
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:49'!
restoreSigUsr1
	"Unset a SIGUSR1 signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigUsr1"

	^ self restoreSignal: self primSigUsr1Number
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 11/4/2005 06:50'!
restoreSigUsr2
	"Unset a SIGUSR2 signal handler and unregister the Smalltalk semaphore.
	Answer the unregistered Semaphore, or nil if unable to restore the signal
	(possibly because no handler had been set)."

	"OSProcess accessor restoreSigUsr2"

	^ self restoreSignal: self primSigUsr2Number
! !

!UnixOSProcessAccessor methodsFor: 'signal forwarding' stamp: 'dtl 12/28/2002 15:33'!
restoreSignal: signalNumber 
	"Unset a signal handler and unregister the Smalltalk semaphore. Answer
	the unregistered Semaphore, or nil if unable to restore the signal (possibly
	because no handler had been set)."

	"OSProcess accessor restoreSignal: OSProcess accessor primSigIntNumber"

	| semaphoreIndex sema |
	semaphoreIndex := self primSemaIndexFor: signalNumber.
	semaphoreIndex ifNotNil:
		[sema := Smalltalk externalObjects at: semaphoreIndex ifAbsent: [].
		sema ifNotNil:
			[self primForwardSignal: signalNumber toSemaphore: nil.
			Smalltalk unregisterExternalObject: sema]].
	^ sema
! !

!UnixOSProcessAccessor methodsFor: 'fork and exec' stamp: 'dtl 11/21/2010 12:10'!
safeToForkSqueak
	"The forkSqueak method is currently supported only for X11 displays with
	XDisplayControlPlugin available. If the plugin is not available, the child Squeak
	VM will continue interacting with the X display, which leads to errors in the
	parent Squeak, including VM crashes."

	^self canControlXDisplay! !

!UnixOSProcessAccessor methodsFor: 'file control' stamp: 'dtl 11/28/2010 12:15'!
setBlocking: anIOHandle
	"Convert anIOHandle to an SQFile data structure and call primitive to set for blocking I/O."

	^ self primSQFileSetBlocking: (self handleFromAccessor: anIOHandle)
! !

!UnixOSProcessAccessor methodsFor: 'file control' stamp: 'dtl 11/28/2010 12:16'!
setNonBlocking: anIOHandle
	"Convert anIOHandle to an SQFile data structure and call primitive to set it non-blocking."

	^ self primSQFileSetNonBlocking: (self handleFromAccessor: anIOHandle)! !

!UnixOSProcessAccessor methodsFor: 'signal print stack trace' stamp: 'dtl 7/20/2010 00:07'!
setPrintAllStacksOnSigUsr1
	"Set a signal handler in the VM that will print all call stacks on console output
	whenever a SIGUSR1 is received. Answer an identifier for the previously
	defined signal handler or nil if this signal handler has already been set."
	
	"OSProcess accessor setPrintAllStacksOnSigUsr1"
	"OSProcess accessor clearPrintAllStacksOnSigUsr1"

	^self primPrintAllStacksOnSignal: self primSigUsr1Number	
! !

!UnixOSProcessAccessor methodsFor: 'file control' stamp: 'dtl 11/28/2010 12:17'!
setUnbuffered: anIOHandle
	"Convert anIOHandle to an SQFile data structure and call primitive to set unbuffered I/O."

	^ self primSQFileSetUnbuffered: (self handleFromAccessor: anIOHandle)! !

!UnixOSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 1/25/2004 21:31'!
sigChldSemaphore
	"Answer the value of sigChldSemaphore"

	^ sigChldSemaphore ifNil: [sigChldSemaphore := self forwardSigChld].
! !

!UnixOSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 1/25/2004 21:29'!
sigChldSemaphore: anObject
	"Set the value of sigChldSemaphore"

	sigChldSemaphore := anObject! !

!UnixOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 3/25/2001 15:34'!
sizeOfInt
	"Size of an integer on this machine with this C compiler."

	^ self primSizeOfInt! !

!UnixOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 3/25/2001 20:55'!
sizeOfPointer
	"Size of a void pointer on this machine with this C compiler."

	^ self primSizeOfPointer! !

!UnixOSProcessAccessor methodsFor: 'file control' stamp: 'dtl 10/6/2001 07:27'!
unixFileNumber: anIOHandle
	"Answer the integer Unix file number corresponding to anIOHandle."

	^ anIOHandle ifNotNil: [self primUnixFileNumber: (self handleFromAccessor: anIOHandle)]
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 4/10/2005 15:17'!
unlockAllForFile: aFileStream
	"Unlock and uncache all locks associated with aFileStream.  This could be called
	before closing a stream, for example."

	(self registeredLocksForFile: aFileStream) do: [:ea | self unlockFileRegion: ea]
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/22/2004 15:02'!
unlockFile: aFileStream
	"Unlock the file represented by aFileStream.  Answer a descriptor for the unlocked
	file region, an Array of file handle and region interval; or answer nil on error
	or if the region did not appear in the cache.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self unlockFile: aFileStream exclusive: true
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'jf 2/22/2004 17:30'!
unlockFile: aFileStream exclusive: flag
	"Unlock the file represented by aFileStream.  Answer a descriptor for the unlocked
	file region, an Array of file handle and region interval; or answer nil on error
	or if the region did not appear in the cache.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	| fileLock |
	aFileStream ifNil: [^ nil].
	fileLock := OSFileLock onFile: aFileStream exclusive: flag.
	^ self unlockFileRegion: fileLock
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/12/2006 07:05'!
unlockFile: aFileStream exclusive: flag ifFail: failBlock
	"Unlock the file represented by aFileStream.  Answer a descriptor for the unlocked
	file region, an Array of file handle and region interval; or answer nil on error
	or if the region did not appear in the cache.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	| fileLock |
	aFileStream ifNil: [^ failBlock value].
	fileLock := OSFileLock onFile: aFileStream exclusive: flag.
	^ self unlockFileRegion: fileLock ifFail: failBlock

! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/22/2004 15:01'!
unlockFile: aFileStream from: start to: end
	"Pass a struct SQFile on the stack, and unlock the specified region.
	Answer the result of the call to fcntl(). If the region is in the file lock cache,
	remove it, but otherwise ignore the cache. The cache supports Win32 semantics
	within a single Squeak image, but not across separate images, therefore the
	unlock should be attempted regardless of whether this image thinks that the
	region has previously been locked. Answer a descriptor for the unlocked file
	region, an Array of file handle and region interval; or answer nil on error
	or if the region did not appear in the cache.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self unlockFile: aFileStream from: start to: end exclusive: true
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 2/22/2004 15:09'!
unlockFile: aFileStream from: start to: end exclusive: flag
	"Pass a struct SQFile on the stack, and unlock the specified region.
	Answer the result of the call to fcntl(). If the region is in the file lock cache,
	remove it, but otherwise ignore the cache. The cache supports Win32 semantics
	within a single Squeak image, but not across separate images, therefore the
	unlock should be attempted regardless of whether this image thinks that the
	region has previously been locked. Answer a descriptor for the unlocked file
	region, an Array of file handle and region interval; or answer nil on error
	or if the region did not appear in the cache.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	| fileRegion |
	aFileStream ifNil: [^ nil].
	fileRegion := OSFileRegionLock onFile: aFileStream from: start to: end exclusive: flag.
	^ self unlockFileRegion: fileRegion
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 8/30/2009 14:54'!
unlockFile: fileStream from: start to: end exclusive: writeLockFlag ifFail: failBlock

	| lock |
	lock := OSFileRegionLock onFile: fileStream from: start to: end exclusive: writeLockFlag.
	^ self unlockFileRegion: lock ifFail: [failBlock value]! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 8/30/2009 16:25'!
unlockFile: fileStream from: start to: end ifFail: failBlock

	^ self unlockFile: fileStream from: start to: end exclusive: true ifFail: failBlock
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/12/2006 07:06'!
unlockFile: aFileStream ifFail: failBlock
	"Unlock the file represented by aFileStream.  Answer a descriptor for the unlocked
	file region, an Array of file handle and region interval; or answer nil on error
	or if the region did not appear in the cache.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self unlockFile: aFileStream exclusive: true ifFail: failBlock
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/8/2006 07:12'!
unlockFileRegion: aFileLock
	"Pass a struct SQFile on the stack, and unlock the specified region.
	Answer the result of the call to fcntl(). If the region is in the file lock cache,
	remove it, but otherwise ignore the cache. The cache supports Win32 semantics
	within a single Squeak image, but not across separate images, therefore the
	unlock should be attempted regardless of whether this image thinks that the
	region has previously been locked. Answer a descriptor for the unlocked file
	region, an Array of file handle and region interval; or answer nil on error
	or if the region did not appear in the cache.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	^ self unlockFileRegion: aFileLock ifFail: [nil]
! !

!UnixOSProcessAccessor methodsFor: 'file locking' stamp: 'dtl 5/20/2006 12:46'!
unlockFileRegion: aFileLock ifFail: failBlock
	"Pass a struct SQFile on the stack, and unlock the specified region.
	Answer the result of the call to fcntl(). If the region is in the file lock cache,
	remove it, but otherwise ignore the cache. The cache supports Win32 semantics
	within a single Squeak image, but not across separate images, therefore the
	unlock should be attempted regardless of whether this image thinks that the
	region has previously been locked. Answer a descriptor for the unlocked file
	region, an Array of file handle and region interval; or answer nil on error
	or if the region did not appear in the cache.

	Warning: The registry permits compatibility with Win32 file locking semantics,
	but only within a single Squeak image. Multiple cooperating images must not
	rely on the overlap checking, because the registry is local to this image and
	cannot be shared across images in different OS process contexts."

	aFileLock ifNil: [^ failBlock value].
	"Check region lock overlap for Win32 compatibility"
	self emulateWin32FileLocking
		ifTrue: [ | unregisteredLock |
				(unregisteredLock := self unregister: aFileLock)
					ifNil: [^ failBlock value]
					ifNotNil: [unregisteredLock unlock
						ifTrue: [^ unregisteredLock]
						ifFalse: [^ failBlock value]]]
		ifFalse: [aFileLock unlock
					ifTrue: [^ aFileLock]
					ifFalse: [^ failBlock value]]
! !

!UnixOSProcessAccessor methodsFor: 'file lock registry' stamp: 'dtl 3/5/2005 13:10'!
unregister: fileRegionLock
	"If an object equal to fileRegionLock exists in the registry, remove it and
	answer the object. Otherwise answer nil. Caching is enabled when
	EmulateWin32FileLocking is true."

	^ self emulateWin32FileLocking
		ifTrue: [super unregister: fileRegionLock]
		ifFalse: [nil]
! !

!UnixOSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:45'!
xdcpModuleName
	"Answer a string containing the module name string for the display control plugin."

	"OSProcess accessor xdcpModuleName"

	^ self primXDisplayControlPluginModuleName
! !

!UnixOSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 10:46'!
xdcpVersionString
	"Answer a string containing the version string for the display control plugin."

	"OSProcess accessor xdcpVersionString"

	^ self primXDisplayControlPluginModuleVersionString
! !

OSProcessAccessor subclass: #WindowsOSProcessAccessor
	instanceVariableNames: 'sigChldSemaphore semaIndex grimReaper childWatcherThread'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Win32'!

!WindowsOSProcessAccessor commentStamp: '<historical>' prior: 0!
I provide access to an operating system process, such as the process in which the Squeak VM is currently running. I am based on the Win32 process model for Windows and Windows NT.!

!WindowsOSProcessAccessor class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:50'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for the
	Squeak VM running on the current platform."

	^ OSProcess isWindows
! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 2/26/2002 16:11'!
activeHandles
	"Answer an Array of handles for all children that are believed to be running."

	^ OSProcess thisOSProcess activeHandles
! !

!WindowsOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 9/10/2002 09:24'!
canAccessChildProcess: anExternalProcess
	"Is the child process still there? Maybe not if we have restarted the image
	and anExternalProcess refers to a process which died while we were not
	watching."

	| handle |
	^ (handle := anExternalProcess handle) notNil
		ifTrue: [self primCanAccessChildProcess: handle]
		ifFalse: [false]
! !

!WindowsOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 2/22/2002 22:17'!
canAccessSystem
	"Answer true if it is possible to access the external process, else false. Failure
	to access the external process is probably due to lack of a UnixOSProcessPlugin
	module."

	^ self primGetPid notNil
! !

!WindowsOSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 2/27/2002 11:56'!
childWatcherThread
	"A thread which signals my sigChldSemaphore when any child process exits."

	^ childWatcherThread! !

!WindowsOSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 2/27/2002 11:44'!
childWatcherThread: aThreadObject
	"A thread which signals my sigChldSemaphore when any child process exits."

	childWatcherThread := aThreadObject
! !

!WindowsOSProcessAccessor methodsFor: 'initialize - release' stamp: 'dtl 2/26/2002 08:43'!
finalize
	"Clean up grimReaper and associated semaphore."

	grimReaper ifNotNil:
			[grimReaper terminate.
			grimReaper := nil].
	sigChldSemaphore ifNotNil:
			[Smalltalk unregisterExternalObject: sigChldSemaphore.
			sigChldSemaphore := nil].
	semaIndex := nil
! !

!WindowsOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 2/27/2002 13:22'!
getMainThread
	"Answer the main thread of this OS process. The handle for this thread is a
	pseudo-handle, and cannot be used to terminate the thread."

	"OSProcess accessor getMainThread"

	^ WindowsThread threadID: self primGetPid handle: self primGetPidHandle running: true
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/7/2002 20:09'!
getStdErr
	"Answer an IO handle (representing a SQFile data structure in interp.c) for the standard
	error for the OS process in which I am currently executing, or nil if the IO handle
	cannot be obtained."

	"OSProcess accessor getStdErr"

	| error |
	error := self primGetStdErrorForSession: self sessionIdentifier.
	^ (error notNil and: [error last])
		ifTrue: [self ioAccessorFromSQFile: error first]
		ifFalse: [nil]
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/25/2005 16:11'!
getStdErrHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard error for the
	OS process in which I am currently executing."

	^ self ioAccessorFromSQFile: self getStdErr
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/7/2002 20:11'!
getStdIn
	"Answer an IO handle (representing a SQFile data structure in interp.c) for the standard
	input for the OS process in which I am currently executing, or nil if the IO handle
	cannot be obtained."

	"OSProcess accessor getStdIn"

	| input |
	input := self primGetStdInputForSession: self sessionIdentifier.
	^ (input notNil and: [input last])
		ifTrue: [self ioAccessorFromSQFile: input first]
		ifFalse: [nil]
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/25/2005 16:10'!
getStdInHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard input for the
	OS process in which I am currently executing."

	^ self ioAccessorFromSQFile: self getStdIn
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/7/2002 20:11'!
getStdOut
	"Answer an IO handle (representing a SQFile data structure in interp.c) for the standard
	output for the OS process in which I am currently executing, or nil if the IO handle
	cannot be obtained."

	"OSProcess accessor getStdOut"

	| output |
	output := self primGetStdOutputForSession: self sessionIdentifier.
	^ (output notNil and: [output last])
		ifTrue: [self ioAccessorFromSQFile: output first]
		ifFalse: [nil]
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/25/2005 16:09'!
getStdOutHandle
	"Answer the handle (a SQFile data structure in interp.c) for the standard output for the
	OS process in which I am currently executing."

	^ self ioAccessorFromSQFile: self getStdOut
! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 9/3/2010 13:13'!
grimReaperProcess
	"This is a process which waits for the death of a child OSProcess, and 
	informs any dependents of the change."

	grimReaper ifNil: [grimReaper := [
				[self sigChldSemaphoreSet wait.
				self changed: #childProcessStatus] repeat] newProcess.
				grimReaper resume.
				"name selected to look reasonable in the process browser"
				grimReaper name: ((ReadStream on: grimReaper hash asString) next: 5)
						, ': the child OSProcess watcher'].
	^ grimReaper! !

!WindowsOSProcessAccessor methodsFor: 'initialize - release' stamp: 'dtl 3/2/2002 08:33'!
initialize
	"Create and register a semaphore to be used for signaling external process exits"

	super initialize.
	self sigChldSemaphoreSet.
	self grimReaperProcess
! !

!WindowsOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 9/26/2005 07:45'!
isExecutable: aPathName
	"Answer true if file at aPathName has execute permission for this process."

	"FIXME: Default to true for Windows"

	^ true

! !

!WindowsOSProcessAccessor methodsFor: 'nonblocking read' stamp: 'dtl 10/1/2005 09:26'!
lastReadFor: aSemaphoreIndex
	"A character has been read into an external buffer corresponding to aSemaphoreIndex,
	and is now available. Answer integer value of the character, or nil if no character
	was read, or -1 if an error occurred on the read."

	| c readResult |
	readResult := Array new: 3.
	c := self primLastReadFor: aSemaphoreIndex storeIn: readResult.
	(c < 1) ifTrue: [self error: 'primLastReadFor: error, returned negative value'].
	^ c
! !

!WindowsOSProcessAccessor methodsFor: 'nonblocking read' stamp: 'dtl 10/1/2005 09:26'!
lastReadFor: aSemaphoreIndex storeIn: aThreeElementArray
	"A character has been read into an external buffer corresponding to aSemaphoreIndex,
	and is now available. Answer integer value of the character, or nil if no character
	was read, or -1 if an error occurred on the read. The results of the read call are
	stored in aThreeElementArray as a side effect."

	^ self primLastReadFor: aSemaphoreIndex storeIn: aThreeElementArray
! !

!WindowsOSProcessAccessor methodsFor: 'nonblocking read' stamp: 'dtl 4/1/2002 11:02'!
nextFrom: aFileStream signaling: aSemaphoreIndex
	"Read the next character from aFileStream into a buffer in the VM. When the read completes,
	signal the specified Semaphore to notify that the character is available."

	| sqFile |
	sqFile := UseIOHandle
				ifTrue: [aFileStream ioHandle handle]
				ifFalse: [aFileStream fileID].
	^ self primNextFrom: sqFile signaling: aSemaphoreIndex
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/7/2002 20:03'!
openStdErr
	"Answer an IO handle (representing a SQFile data structure in interp.c) for the standard
	error for the OS process in which I am currently executing. Open a console if necessary
	to make standard error available."

	"OSProcess accessor openStdErr"

	| error |
	error := self primGetStdErrorForSession: self sessionIdentifier.
	error ifNil: [^ nil].
	error last
		ifFalse:
			[self primAllocConsole.
			error := self primGetStdOutputForSession: self sessionIdentifier].
	^ self ioAccessorFromSQFile: error first
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/7/2002 20:04'!
openStdIn
	"Answer an IO handle (representing a SQFile data structure in interp.c) for the standard
	input for the OS process in which I am currently executing. Open a console if necessary
	to make standard input available."

	"OSProcess accessor openStdIn"

	| input |
	input := self primGetStdInputForSession: self sessionIdentifier.
	input ifNil: [^ nil].
	input last
		ifFalse:
			[self primAllocConsole.
			input := self primGetStdOutputForSession: self sessionIdentifier].
	^ self ioAccessorFromSQFile: input first
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/7/2002 20:05'!
openStdOut
	"Answer an IO handle (representing a SQFile data structure in interp.c) for the standard
	output for the OS process in which I am currently executing. Open a console if necessary
	to make standard input available."

	"OSProcess accessor openStdOut"

	| output |
	output := self primGetStdOutputForSession: self sessionIdentifier.
	output ifNil: [^ nil].
	output last
		ifFalse:
			[self primAllocConsole.
			output := self primGetStdOutputForSession: self sessionIdentifier].
	^ self ioAccessorFromSQFile: output first
! !

!WindowsOSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 11:46'!
osppModuleName
	"Answer a string containing the module name string for the OSPP plugin."

	"OSProcess accessor osppModuleName"

	^ self primOSProcessPluginModuleName
! !

!WindowsOSProcessAccessor methodsFor: 'plugin identification' stamp: 'dtl 10/1/2005 11:46'!
osppModuleVersionString
	"Answer a string containing the version string for the OSPP plugin."

	"OSProcess accessor osppModuleVersionString"

	^ self primOSProcessPluginModuleVersionString
! !

!WindowsOSProcessAccessor methodsFor: 'console' stamp: 'dtl 3/25/2002 06:28'!
primAllocConsole
	"Allocate a console if not already allocated."

	"OSProcess accessor primAllocConsole"

	<primitive: 'primitiveAllocConsole' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 8/17/2002 12:46'!
primBufferValuesAt: aSemaphoreIndex
	"For debugging only. Answer the current values of readCharBufferArray,
	readCharCountArray, and readCharStatusArray at index, an integer corresponding
	to a semaphore for one read handler thread. Answer an Array with the buffered
	character, the character count, and the status value."

	"OSProcess accessor primBufferValuesAt: 1"

	<primitive: 'primitiveBufferValuesAt' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 9/10/2002 09:19'!
primCanAccessChildProcess: handleObject
	"Answer true if the OS process represented by a HANDLE can be accessed by this OS process."

	<primitive: 'primitiveCanAccessChildProcess' module: 'Win32OSProcessPlugin'>

	^ false! !

!WindowsOSProcessAccessor methodsFor: 'handles' stamp: 'dtl 2/25/2002 07:37'!
primCloseHandle: handleObject
	"Close the specified handle, which may refer to a process, a thread, or
	some other Win32 object."

	"| procInfo |
	procInfo := OSProcess accessor primCommand: 'SOL'.
	OSProcess accessor primCloseHandle: procInfo first"

	<primitive: 'primitiveCloseHandle' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'deprecated' stamp: 'dtl 9/9/2002 09:24'!
primCommand: commandString
	"Run a command in a new external process. Answer a result array with
	hProcess, hThread, dwProcessId, dwThreadId.

	This primitive has been replaced by #primCommand:stdIn:stdOut:stdErr:, and
	will be removed in future versions of OSProcess."

	"OSProcess accessor primCommand: 'C:\WINDOWS\SOL'"
	"OSProcess accessor primCommand: 'SOL'"
	"OSProcess accessor primCommand: 'NoSuchProgram'"

	<primitive: 'primitiveCommand' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'process creation' stamp: 'dtl 9/9/2002 09:19'!
primCommand: commandString stdIn: inputFileIoHandle stdOut: outputFileIoHandle stdErr: errorFileIoHandle

	"Run a command in a new external process. The standard input, output and error
	stream handles are sqFile byte arrays (not Win32 HANDLE arrays), and may be nil.
	Answer a result array with hProcess, hThread, dwProcessId, dwThreadId."

	"OSProcess accessor primCommand: 'C:\WINDOWS\SOL' stdIn: nil stdOut: nil stdErr: nil"
	"OSProcess accessor primCommand: 'SOL' stdIn: nil stdOut: nil stdErr: nil"
	"OSProcess accessor primCommand: 'NoSuchProgram' stdIn: nil stdOut: nil stdErr: nil"
	"OSProcess accessor primCommand: 'SOL' stdIn: (FileStream fileNamed: 'output.tmp') fileID stdOut: nil stdErr: nil"

	<primitive: 'primitiveCommandWithInputOutputError' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'pipe open' stamp: 'dtl 11/28/2010 12:51'!
primCreatePipe
	"Create a pipe, and answer an array of two file handles (SQFile data structures in interp.c)
	for the pipe reader and writer."

	<primitive: 'primitiveCreatePipe' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'pipe open' stamp: 'dtl 11/28/2010 12:52'!
primCreatePipeWithSessionIdentifier: aByteArray
	"Create a pipe, and answer an array of two file handles (SQFile data structures in interp.c)
	for the pipe reader and writer."

	<primitive: 'primitiveCreatePipeWithSessionIdentifier' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'console' stamp: 'dtl 3/25/2002 06:27'!
primFreeConsole
	"Deallocate the console if allocated."

	"OSProcess accessor primFreeConsole"

	<primitive: 'primitiveFreeConsole' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 9/6/2002 14:58'!
primGetCurrentWorkingDirectory
	"Call getcwd() to get the current working directory."

	"OSProcess accessor primGetCurrentWorkingDirectory"

	<primitive: 'primitiveGetCurrentWorkingDirectory' module: 'Win32OSProcessPlugin'>
	^ nil
! !

!WindowsOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 2/22/2002 18:23'!
primGetEnvironmentStrings
	"Answer the environment block in the form of an Array of Strings. The
	caller is expected to parse the strings into a dictionary of keys and values."

	<primitive: 'primitiveGetEnvironmentStrings' module: 'Win32OSProcessPlugin'>
	^ nil
! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 2/25/2002 08:23'!
primGetExitStatusForHandle: handleObject
	"Answer the exit status for the process represented by a HANDLE. Fail if
	the process is still active, or if the GetExitCodeProcess call fails."

	"| procInfo |
	procInfo := OSProcess accessor primCommand: 'SOL'.
	(Delay forSeconds: 5) wait.
	OSProcess accessor primGetExitStatusForHandle: procInfo first"

	<primitive: 'primitiveGetExitStatusForHandle' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 2/27/2002 13:20'!
primGetMainThreadHandle
	"Answer a pseudo-handle for my main thread."

	"OSProcess accessor primGetMainThreadHandle"

	<primitive: 'primitiveGetMainThreadHandle' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 3/4/2006 17:57'!
primGetMainThreadID
	"Answer the ID of my main thread."

	"OSProcess accessor primGetMainThreadID"

	<primitive: 'primitiveGetMainThreadID' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 2/22/2002 16:06'!
primGetPid
	"Answer the OS process ID for the OS process in which I am currently executing."

	<primitive: 'primitiveGetPid' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'external process access' stamp: 'dtl 2/27/2002 13:21'!
primGetPidHandle
	"Answer the pseudo-handle for my OS process"

	"OSProcess accessor primGetPidHandle"

	<primitive: 'primitiveGetPidHandle' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'session identification' stamp: 'dtl 3/2/2002 08:15'!
primGetSession
	"Answer the unique identifier for this session of Smalltalk running in this OS Process."

	"OSProcess accessor primGetSession"

	<primitive: 'primitiveGetSession' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 9/7/2002 14:55'!
primGetStdErrorForSession: sessionIdentifierByteArray
	"Answer a two element array containing the sqFile data structure representing
	standard error stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard error
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	"OSProcess accessor primGetStdError"

	<primitive: 'primitiveGetStdError' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 9/7/2002 14:55'!
primGetStdInputForSession: sessionIdentifierByteArray
	"Answer a two element array containing the sqFile data structure representing
	standard input stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard input
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	"OSProcess accessor primGetStdInput"

	<primitive: 'primitiveGetStdInput' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 9/7/2002 14:55'!
primGetStdOutputForSession: sessionIdentifierByteArray
	"Answer a two element array containing the sqFile data structure representing
	standard output stream for my OS process, and a flag (true or false) to indicate
	whether the sqFile data structure contains a valid HANDLE. If no standard output
	stream is available for this OS process, the sqFile data structure will contain an
	invalid HANDLE value, which will result in failures on subsequent accesses."

	"OSProcess accessor primGetStdOutput"

	<primitive: 'primitiveGetStdOutput' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/1/2005 09:26'!
primLastReadFor: aSemaphoreIndex
	"A character has been read into an external buffer corresponding to aSemaphoreIndex,
	and is now available. Answer integer value of the character, or nil if no character
	was read, or -1 if an error occurred on the read."

	<primitive: 'primitiveLastReadFor' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/1/2005 09:27'!
primLastReadFor: aSemaphoreIndex storeIn: aThreeElementArray
	"A character has been read into an external buffer corresponding to aSemaphoreIndex,
	and is now available. Answer integer value of the character, or nil if no character
	was read, or -1 if an error occurred on the read. Contents of the aThreeElementArray
	will be status of the read call, character read, and character count (which should
	always be 1)."

	<primitive: 'primitiveLastReadForStoreIn' module: 'Win32OSProcessPlugin'>
	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 4/1/2002 10:58'!
primNextFrom: anIOHandle signaling: aSemaphoreIndex
	"Read the next character from anIOHandle (a SQFile struct) into a buffer in the VM. When
	the read completes, signal the specified Semaphore to notify that the character is available."

	<primitive: 'primitiveNextFromSignaling' module: 'Win32OSProcessPlugin'>

	^ nil

! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/1/2005 11:50'!
primOSProcessPluginModuleName
	"Answer a string containing the module name string for the OSPP plugin."

	"OSProcess accessor primOSProcessPluginModuleName"

	<primitive: 'primitiveModuleName' module: 'Win32OSProcessPlugin'>
	^ nil
! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 10/1/2005 11:51'!
primOSProcessPluginModuleVersionString
	"Answer a string containing the version string for the OSPP plugin."

	"OSProcess accessor primOSProcessPluginModuleVersionString"

	<primitive: 'primitiveVersionString' module: 'Win32OSProcessPlugin'>
	^ nil
! !

!WindowsOSProcessAccessor methodsFor: 'testing' stamp: 'dtl 10/10/2012 08:05'!
primOneShot
	"Answer true the first time this is called in a Squeak session, and false thereafter."

	"OSProcess accessor primOneShot"

	<primitive: 'primitiveOneShot' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'handles' stamp: 'dtl 3/29/2002 17:47'!
primSetStdErr: anIOHandle
	"Set the standard error handle to that of anIOHandle, where anIOHandle is a
	ByteArray representation of a SQFile structure."

	<primitive: 'primitiveSetStdErr' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'handles' stamp: 'dtl 3/29/2002 17:48'!
primSetStdIn: anIOHandle
	"Set the standard input handle to that of anIOHandle, where anIOHandle is a
	ByteArray representation of a SQFile structure."

	<primitive: 'primitiveSetStdIn' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'handles' stamp: 'dtl 3/29/2002 17:48'!
primSetStdOut: anIOHandle
	"Set the standard output handle to that of anIOHandle, where anIOHandle is a
	ByteArray representation of a SQFile structure."

	<primitive: 'primitiveSetStdOut' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 2/26/2002 16:14'!
primSetWaitForAnyProcessExit: arrayOfProcessHandleObjects thenSignalSemaphoreWithIndex: index
	"Set up a thread to wait for a process HANDLE to exit, then signal the Semaphore
	at index. This provides asychronous notification of an external process exit."

	<primitive: 'primitiveSetWaitForAnyProcessExitThenSignalSemaphoreWithIndex' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 2/27/2002 11:34'!
primTerminateThread: aThreadHandle
	"Kill the thread. No cleanup is performed, so use with caution for a thread which
	is (for example) manipulating a mutex. Answer true for success, else false."

	<primitive: 'primitiveTerminateThread' module: 'Win32OSProcessPlugin'>

	^ nil! !

!WindowsOSProcessAccessor methodsFor: 'private - primitive access' stamp: 'dtl 4/8/2007 10:54'!
primTestEndOfFileFlag: aSqFileStruct
	"Answer whether the file represented by aSqFileStruct is at end of file, as determined
	by a call to feof(). This is different from StandardFileStream>>primAtEnd: which answers
	true if the file pointer is at the end of the file, but which does not call feof() to
	determine that an end of file condition has occurred. The difference is significant
	if aSqFileStruct represents a pipe or a device file, which may not be positionable
	in the sense of a conventional disk file."

	<primitive: 'primitiveTestEndOfFileFlag' module: 'Win32OSProcessPlugin'>
	self flag: 'FIXME'. "not yet implemented in OSPP for Windows"
	^ self primitiveFailed
! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 1/13/2007 09:48'!
restartChildWatcherThread: arrayOfProcessHandleObjects
	"Start a new child watcher thread. If a thread is alread active, terminate
	it before starting a new one."

	self childWatcherThread ifNotNil: [childWatcherThread terminate].
	self childWatcherThread: (self setWaitForAnyProcessExit: arrayOfProcessHandleObjects).
	^ childWatcherThread
! !

!WindowsOSProcessAccessor methodsFor: 'accessing' stamp: 'dtl 2/25/2002 21:17'!
semaIndex
	"Index of the registered Semaphore"

	semaIndex ifNil: [self initialize].
	^ semaIndex! !

!WindowsOSProcessAccessor methodsFor: 'file control' stamp: 'dtl 9/26/2005 07:35'!
setBlocking: anIOHandle
	"Convert anIOHandle to an SQFile data structure and call primitive to set for blocking I/O."

	"FIXME: need to implement this for Win32"
	"self notify: 'there is no general mechanism to set blocking IO on Win32'"
! !

!WindowsOSProcessAccessor methodsFor: 'file control' stamp: 'dtl 9/26/2005 07:36'!
setNonBlocking: anIOHandle
	"Convert anIOHandle to an SQFile data structure and call primitive to set it non-blocking."

	"FIXME: need to implement this for Win32"
	"self notify: 'there is no general mechanism to set nonblocking IO on Win32'"

! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/9/2002 07:22'!
setStdErr: aFileStream
	"Set the standard error handle for this OSProcess to be that of aFileStream"

	"| fs |
	fs := FileStream fileNamed: 'stdError.tmp'.
	OSProcess accessor setStdErr: fs"

	| sqFile |
	aFileStream ifNil: [^ false].
	sqFile := UseIOHandle
				ifTrue: [aFileStream ioHandle handle]
				ifFalse: [aFileStream fileID].
	^ self primSetStdErr: sqFile
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/9/2002 07:22'!
setStdIn: aFileStream
	"Set the standard input handle for this OSProcess to be that of aFileStream"

	"| fs |
	fs := FileStream fileNamed: 'stdInput.tmp'.
	fs nextPutAll: 'this is a line of text'; cr.
	fs position: 0.
	OSProcess accessor setStdIn: fs"

	| sqFile |
	aFileStream ifNil: [^ false].
	sqFile := UseIOHandle
				ifTrue: [aFileStream ioHandle handle]
				ifFalse: [aFileStream fileID].
	^ self primSetStdIn: sqFile
! !

!WindowsOSProcessAccessor methodsFor: 'standard IO handles' stamp: 'dtl 9/9/2002 07:22'!
setStdOut: aFileStream
	"Set the standard output handle for this OSProcess to be that of aFileStream"

	"| fs |
	fs := FileStream fileNamed: 'stdOutput.tmp'.
	OSProcess accessor setStdOut: fs"

	| sqFile |
	aFileStream ifNil: [^ false].
	sqFile := UseIOHandle
				ifTrue: [aFileStream ioHandle handle]
				ifFalse: [aFileStream fileID].
	^ self primSetStdOut: sqFile
! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 2/26/2002 19:08'!
setWaitForAnyProcessExit: arrayOfProcessHandleObjects
	"Set up a thread to wait for a process HANDLE to exit, then signal the Semaphore
	at index. This provides asychronous notification of an external process exit. The
	caller should close the thread handle when it is no longer needed."

	"OSProcess command: 'SOL'.
	OSProcess accessor setWaitForAnyProcessExit: OSProcess thisOSProcess activeHandles."

	| threadInfo |
	arrayOfProcessHandleObjects isEmpty ifTrue: [^ nil].
	threadInfo := self
		primSetWaitForAnyProcessExit: arrayOfProcessHandleObjects
		thenSignalSemaphoreWithIndex: self semaIndex.
	^ WindowsThread
		threadID: threadInfo last
		handle: threadInfo first
		running: true

! !

!WindowsOSProcessAccessor methodsFor: 'child process management' stamp: 'dtl 2/25/2002 21:12'!
sigChldSemaphoreSet

	sigChldSemaphore ifNil:
		[sigChldSemaphore := Semaphore new.
		semaIndex := Smalltalk registerExternalObject: sigChldSemaphore].
	^ sigChldSemaphore! !

Model subclass: #PseudoAioEventHandler
	instanceVariableNames: 'eventGenerator'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-AIO'!

!PseudoAioEventHandler commentStamp: 'dtl 11/25/2006 10:42' prior: 0!
PseudoAioEventHandler is a replacement for AioEventHandler for use when an AioPlugin is not present. It creates a polling loop by generating #changed: events periodically. With a real AioEventHandler, events are generated only when actual IO activity occurs, while the PseudoAioEventHandler produces regularly timed events regardless of whether any actual IO changes have happened.!

!PseudoAioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 11/25/2006 10:49'!
close
	"When the FileStream or Socket handled by this aio handler is closed,
	it should send #close to this handler."

	eventGenerator ifNotNil: [eventGenerator terminate]! !

!PseudoAioEventHandler methodsFor: 'accessing' stamp: 'dtl 11/25/2006 13:17'!
eventGenerator
	"Answer the value of eventGenerator"

	^ eventGenerator ifNil: [eventGenerator := self eventGeneratorProcess]! !

!PseudoAioEventHandler methodsFor: 'initialize-release' stamp: 'RB 9/3/2010 11:52'!
eventGeneratorProcess
	"A process that generates periodic #changed events"

	| d p|
	d := Delay forMilliseconds: 125.
	p := [[self changed.
	d wait] repeat] newProcess.
	^p resume.! !

!PseudoAioEventHandler methodsFor: 'initialize-release' stamp: 'dtl 11/25/2006 13:15'!
initialize

	self eventGenerator.
	^ super initialize! !

!SequenceableCollection methodsFor: '*html' stamp: 'tb 9/14/2006 20:37'!
joinOn: aString

	^ String streamContents:
		[:s |
		self do: [:ea | s nextPutAll: ea asString] separatedBy: [s nextPutAll: aString]]
! !

!HttpUrl methodsFor: '*html' stamp: 'tb 1/9/2006 15:02'!
retrieveContentsArgs: args
	^self retrieveContentsArgs: args accept: (MIMEDocument guessTypeFromName: self)! !

TestCase subclass: #AbstractUnixProcessFileLockingTestCase
	instanceVariableNames: 'accessor fileStream delay remoteProcess initialCompatibilitySetting'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Tests'!

!AbstractUnixProcessFileLockingTestCase commentStamp: 'dtl 4/30/2006 14:02' prior: 0!
Test file locking with the UnixOSProcessPlugin. The test suite requires that OSProcess and CommandShell be loaded in the image.

These tests rely on a remote Squeak image to test file locks between cooperating Unix processes. This may be timing dependent (see #delay, set in #setUp, and cleanup in #tearDown). In case of intermittent failures, try running the failed test individually. In some cases it may be necessary to restart Squeak in order to clear leftover file locks from previous failed tests.!

!AbstractUnixProcessFileLockingTestCase class methodsFor: 'testing' stamp: 'dtl 4/30/2006 13:49'!
isAbstract
	"Override to true if a TestCase subclass is Abstract and should not have
	TestCase instances built from it"

	^ true! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'accessing' stamp: 'dtl 4/30/2006 13:47'!
accessor
	"Answer the value of accessor"

	^ accessor! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'accessing' stamp: 'dtl 4/30/2006 13:47'!
accessor: anObject
	"Set the value of accessor"

	accessor := anObject! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'accessing' stamp: 'dtl 4/30/2006 13:47'!
delay
	"Answer the value of delay"

	^ delay! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'accessing' stamp: 'dtl 4/30/2006 13:47'!
delay: anObject
	"Set the value of delay"

	delay := anObject! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'remote image commands' stamp: 'dtl 12/15/2013 15:50'!
doRemote: remoteBlock doLocal: localBlock
	"Evaluate remoteBlock in a remote Squeak image. While the remote Squeak is
	still running, evaluate localBlock. When done, terminate the remote Squeak
	image and answer the result of evaluating localBlock."

	"self new
		doRemote: [OSProcess thisOSProcess stdOut nextPutAll: 'hello from child Squeak';
						nextPut: Character lf]
		doLocal: [(Delay forSeconds: 1) wait.
						OSProcess thisOSProcess stdOut nextPutAll: 'hello from parent Squeak';
					nextPut: Character lf]"

	| result |
	self shouldnt: [remoteProcess := OSProcess thisOSProcess forkHeadlessSqueak]
		raise: Warning.
	self assert: remoteProcess notNil. "Fail if fork did not succeed"
	(remoteProcess == OSProcess thisOSProcess)
		ifTrue:
			["remote child Squeak"
			remoteBlock value]
		ifFalse:
			["parent Squeak"
			[[remoteProcess isRunning] whileFalse: [self delay wait].
			self delay wait.
			result := localBlock value]
				ensure:	[remoteProcess terminate.
					[remoteProcess isComplete] whileFalse: [self delay wait].
					self delay wait.
					result]].
	^ result
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'failures' stamp: 'eem 3/27/2014 11:28'!
expectedFailures
	^UnixOSProcessAccessor basicNew safeToForkSqueak
		ifTrue: [#()]
		ifFalse: [#(	testCooperatingProcesses01
					testCooperatingProcesses02
					testCooperatingProcesses03
					testCooperatingProcesses04
					testCooperatingProcesses05
					testFailFileLockOnLockedFile
					testFailLockOnLockedOverlappedRegion
					testFailLockOnLockedRegion
					testFailLockOnLockedSupersetRegion
					testFailRegionLockOnLockedFile
					testLockEntireFileForWrite01
					testLockEntireFileForWrite02
					testLockEntireFileForWrite03
					testLockEntireFileForWrite04
					testLockEntireFileForWrite05
					testLockEntireFileForWrite06
					testLockRegionForRead01
					testLockRegionForRead02
					testLockRegionForWrite01
					testLockRegionForWrite02
					testLockRegionForWrite03
					testLockRegionForWrite04
					testLockRegionForWrite05
					testLockRegionForWrite06
					testLockRegionForWrite07
					testLockRegionForWrite08
					testNoFailLockOnAdjacentLockedRegions
					testNoFailLockOnDifferentLockedRegion)]! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'accessing' stamp: 'dtl 4/30/2006 13:47'!
fileStream
	"Answer the value of fileStream"

	^ fileStream! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'accessing' stamp: 'dtl 4/30/2006 13:47'!
fileStream: anObject
	"Set the value of fileStream"

	fileStream := anObject! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'accessing' stamp: 'dtl 4/30/2006 13:57'!
initialCompatibilitySetting
	"Answer the value of initialCompatibilitySetting"

	^ initialCompatibilitySetting! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'accessing' stamp: 'dtl 4/30/2006 13:57'!
initialCompatibilitySetting: anObject
	"Set the value of initialCompatibilitySetting"

	initialCompatibilitySetting := anObject! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'initialize' stamp: 'dtl 4/30/2006 13:58'!
initialize

	initialCompatibilitySetting
		ifNil: [self initialCompatibilitySetting: OSProcess accessor emulateWin32FileLocking]! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/13/2006 09:36'!
isValidUnlockResult: result

	self subclassResponsibility
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'remote image commands' stamp: 'dtl 4/30/2006 13:47'!
quitImage

	fileStream close.
	OSProcess snapshot: false andQuit: true
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'running' stamp: 'dtl 7/30/2010 16:38'!
setUp

	(self respondsTo: #timeout: ) "Recent Squeak images with test case timeout"
		ifTrue: [self perform: #timeout: with: 30].
	self delay: (Delay forMilliseconds: 150).
	self accessor: ThisOSProcess accessor.
	self fileStream: (FileStream fileNamed: 'junkfile').
	1000 timesRepeat: (fileStream nextPutAll: 'hello world ').
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'running' stamp: 'dtl 9/3/2010 16:02'!
tearDown

	| d |
	OSProcessAccessor emulateWin32FileLocking: self initialCompatibilitySetting.
	d := Delay forMilliseconds: 50.
	self fileStream close.
	remoteProcess ifNotNil:
		[remoteProcess terminate.
		[remoteProcess isComplete] whileFalse: [d wait].
		remoteProcess := nil]
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 4/30/2006 13:47'!
testCooperatingProcesses01
	"Two cooperating Squeak processes using file region locking to coordinate their
	writes to a shared file."

	"(self selector: #testCooperatingProcesses01) run"

	| result |
	result := self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 120) notNil]) ifNil: [self quitImage].
			fileStream position: 100.
			fileStream nextPutAll: 'THIS '.
			accessor unlockFile: fileStream from: 100 to: 120.
			delay wait.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			fileStream position: 108.
			fileStream nextPutAll: 'A '.
			accessor unlockFile: fileStream from: 108 to: 120.
			delay wait.
			self quitImage]
		doLocal:
			[self assert: (self tryUntil: [(accessor lockFile: fileStream from: 104 to: 120) notNil]).
			fileStream position: 105.
			fileStream nextPutAll: 'IS '.
			accessor unlockFile: fileStream from: 104 to: 120.
			delay wait.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			fileStream position: 110.
			fileStream nextPutAll: 'TEST '.
			accessor unlockFile: fileStream from: 108 to: 120.
			delay wait].
	self tryUntil: [(accessor lockFile: fileStream from: 100 to: 140) notNil].
	fileStream position: 100.
	result := fileStream next: 14.
	accessor unlockFile: fileStream from: 100 to: 140.
	self assert: result = 'THIS IS A TEST'
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 4/30/2006 13:47'!
testCooperatingProcesses02
	"Two cooperating Squeak processes using file region locking to coordinate their
	writes to a shared file."

	"(self selector: #testCooperatingProcesses02) run"

	| result |
	result := self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 120) notNil]) ifNil: [self quitImage].
			fileStream position: 100.
			fileStream nextPutAll: 'THIS 11111111111111111111111'.
			accessor unlockFile: fileStream from: 100 to: 120.
			delay wait; wait; wait.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			fileStream position: 108.
			fileStream nextPutAll: 'A 3333333333333333333333333'.
			accessor unlockFile: fileStream from: 108 to: 120.
			self quitImage]
		doLocal:
			[self assert: (self tryUntil: [(accessor lockFile: fileStream from: 104 to: 120) notNil]).
			fileStream position: 105.
			fileStream nextPutAll: 'IS 2222222222222222222222222'.
			accessor unlockFile: fileStream from: 104 to: 120.
			delay wait; wait; wait.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			fileStream position: 110.
			fileStream nextPutAll: 'TEST 4444444444444444444444'.
			accessor unlockFile: fileStream from: 108 to: 120].
	self tryUntil: [(accessor lockFile: fileStream from: 100 to: 140) notNil].
	fileStream position: 100.
	result := fileStream next: 17.
	accessor unlockFile: fileStream from: 100 to: 140.
	self assert: result = 'THIS IS A TEST 44'
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 4/30/2006 13:47'!
testCooperatingProcesses03
	"Two cooperating Squeak processes using file region locking to coordinate their
	writes to a shared file."

	"(self selector: #testCooperatingProcesses03) run"

	| result |
	result := self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 120) notNil]) ifNil: [self quitImage].
			fileStream position: 100.
			fileStream nextPutAll: 'THIS 11111111111111111111111'.
			accessor unlockFile: fileStream from: 100 to: 120.
			delay wait; wait; wait.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			fileStream position: 108.
			fileStream nextPutAll: 'A 3333333333333333333333333'.
			accessor unlockFile: fileStream from: 108 to: 120.
			self quitImage]
		doLocal:
			[self assert: (self tryUntil: [(accessor lockFile: fileStream from: 104 to: 120) notNil]).
			fileStream position: 105.
			fileStream nextPutAll: 'IS 2222222222222222222222222'.
			accessor unlockFile: fileStream from: 104 to: 120.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			fileStream position: 110.
			fileStream nextPutAll: 'TEST 4444444444444444444444'.
			accessor unlockFile: fileStream from: 108 to: 120].
	self tryUntil: [(accessor lockFile: fileStream from: 100 to: 140) notNil].
	fileStream position: 100.
	result := fileStream next: 17.
	accessor unlockFile: fileStream from: 100 to: 140.
	self assert: result = 'THIS IS 22TEST 44'
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 4/30/2006 13:47'!
testCooperatingProcesses04
	"Two cooperating Squeak processes using file region locking to coordinate their
	writes to a shared file."

	"(self selector: #testCooperatingProcesses04) run"

	| result |
	result := self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 120) notNil]) ifNil: [self quitImage].
			delay wait.
			fileStream position: 100.
			fileStream nextPutAll: 'THIS 11111111111111111111111'.
			accessor unlockFile: fileStream from: 100 to: 120.
			delay wait.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			delay wait.
			fileStream position: 108.
			fileStream nextPutAll: 'A 3333333333333333333333333'.
			accessor unlockFile: fileStream from: 108 to: 120.
			delay wait.
			self quitImage]
		doLocal:
			[self assert: (self tryUntil: [(accessor lockFile: fileStream from: 104 to: 120) notNil]).
			delay wait.
			fileStream position: 105.
			fileStream nextPutAll: 'IS 2222222222222222222222222'.
			accessor unlockFile: fileStream from: 104 to: 120.
			delay wait.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			fileStream position: 110.
			fileStream nextPutAll: 'TEST 4444444444444444444444'.
			accessor unlockFile: fileStream from: 108 to: 120].
	self tryUntil: [(accessor lockFile: fileStream from: 100 to: 140) notNil].
	fileStream position: 100.
	result := fileStream next: 17.
	accessor unlockFile: fileStream from: 100 to: 140.
	self assert: result = 'THIS IS A TEST 44'
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 4/30/2006 13:47'!
testCooperatingProcesses05
	"Two cooperating Squeak processes using file region locking to coordinate their
	writes to a shared file."

	"(self selector: #testCooperatingProcesses05) run"

	| result |
	result := self
		doRemote:
			[OSProcess thisOSProcess stdOut nextPutAll: 'starting remote squeak requesting lock on 100 to 120';
				nextPut: Character lf.
			(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 120) notNil]) ifNil: [self quitImage].
			OSProcess thisOSProcess stdOut nextPutAll: 'remote squeak lock acquired on 100 to 120'; nextPut: Character lf.
			fileStream position: 100.
			fileStream nextPutAll: 'THIS '.
			OSProcess thisOSProcess stdOut nextPutAll: 'remote squeak releasing lock on 100 to 120'; nextPut: Character lf.
			accessor unlockFile: fileStream from: 100 to: 120.
			OSProcess thisOSProcess stdOut nextPutAll: 'remote squeak lock released on 100 to 120'; nextPut: Character lf.
			delay wait.
			OSProcess thisOSProcess stdOut nextPutAll: 'remote squeak requesting lock on 108 to: 120'; nextPut: Character lf.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			OSProcess thisOSProcess stdOut nextPutAll: 'remote squeak lock acquired on 108 to: 120'; nextPut: Character lf.
			fileStream position: 108.
			fileStream nextPutAll: 'A '.
			OSProcess thisOSProcess stdOut nextPutAll: 'remote squeak releasing lock on 108 to: 120'; nextPut: Character lf.
			accessor unlockFile: fileStream from: 108 to: 120.
			OSProcess thisOSProcess stdOut nextPutAll: 'remote squeak lock released on 108 to: 120'; nextPut: Character lf.
			delay wait.
			self quitImage]
		doLocal:
			[OSProcess thisOSProcess stdOut nextPutAll: 'starting local squeak requesting lock on 104 to: 120';
				nextPut: Character lf.
			self assert: (self tryUntil: [(accessor lockFile: fileStream from: 104 to: 120) notNil]).
			OSProcess thisOSProcess stdOut nextPutAll: 'local squeak lock acquired on 104 to: 120'; nextPut: Character lf.
			fileStream position: 105.
			fileStream nextPutAll: 'IS '.
			OSProcess thisOSProcess stdOut nextPutAll: 'local squeak releasing lock on 104 to: 120'; nextPut: Character lf.
			accessor unlockFile: fileStream from: 104 to: 120.
			OSProcess thisOSProcess stdOut nextPutAll: 'local squeak lock released on 104 to: 120'; nextPut: Character lf.
			delay wait.
			OSProcess thisOSProcess stdOut nextPutAll: 'local squeak requesting lock on 108 to: 120'; nextPut: Character lf.
			self tryUntil: [(accessor lockFile: fileStream from: 108 to: 120) notNil].
			OSProcess thisOSProcess stdOut nextPutAll: 'local squeak lock acquired on 108 to: 120'; nextPut: Character lf.
			fileStream position: 110.
			fileStream nextPutAll: 'TEST '.
			OSProcess thisOSProcess stdOut nextPutAll: 'local squeak releasing lock on 108 to: 120'; nextPut: Character lf.
			accessor unlockFile: fileStream from: 108 to: 120.
			OSProcess thisOSProcess stdOut nextPutAll: 'local squeak lock released on 108 to: 120'; nextPut: Character lf.
			delay wait].
	self tryUntil: [(accessor lockFile: fileStream from: 100 to: 140) notNil].
	fileStream position: 100.
	result := fileStream next: 14.
	accessor unlockFile: fileStream from: 100 to: 140.
	self assert: result = 'THIS IS A TEST'
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - exclusive locking' stamp: 'dtl 8/30/2009 15:55'!
testFailFileLockOnLockedFile
	"Two cooperating Squeak processes using mandatory file locking. When
	child Squeak takes a lock on a region, lock attempt in parent Squeak will
	fail."

	"(self selector: #testFailFileLockOnLockedFile) run"

	self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream) notNil]) ifNil: [self quitImage].
			delay wait; wait; wait.
			accessor unlockFile: fileStream.
			self quitImage]
		doLocal:
			[self should: [
				10 timesRepeat: [ "lock and unlock until failure due to child taking the lock"
					accessor lockFile: fileStream
							exclusive: true
							ifFail: [self error: '#lockFile failed because child Squeak now holds a lock'].
					accessor unlockFile: fileStream. "release lock to allow child to acquire it"
					delay wait]]
				raise: Error
				description: '#lockFile failed because child Squeak now holds a lock']
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - exclusive locking' stamp: 'dtl 8/30/2009 16:00'!
testFailLockOnLockedOverlappedRegion
	"Two cooperating Squeak processes using mandatory file locking. When
	child Squeak takes a lock on a region, lock attempt in parent Squeak will
	fail. Locked regions overlap by one character."

	"(self selector: #testFailLockOnLockedOverlappedRegion) run"

	self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 104) notNil]) ifNil: [self quitImage].
			delay wait; wait; wait.
			accessor unlockFile: fileStream from: 100 to: 104.
			self quitImage]
		doLocal:
			[self should: [
				10 timesRepeat: [ "lock and unlock until failure due to child taking the lock"
					accessor lockFile: fileStream
							from: 104
							to: 120
							exclusive: true
							ifFail: [self error: '#lockFile failed because child Squeak now holds a lock'].
					accessor unlockFile: fileStream from: 104 to: 120. "release lock to allow child to acquire it"
					delay wait]]
				raise: Error
				description: '#lockFile failed because child Squeak now holds a lock']
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - exclusive locking' stamp: 'dtl 8/30/2009 15:51'!
testFailLockOnLockedRegion
	"Two cooperating Squeak processes using mandatory file locking. When
	child Squeak takes a lock on a region, lock attempt in parent Squeak will
	fail."

	"(self selector: #testFailLockOnLockedRegion) run"

	self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 120) notNil]) ifNil: [self quitImage].
			delay wait; wait; wait.
			accessor unlockFile: fileStream from: 100 to: 120.
			self quitImage]
		doLocal:
			[self should: [
				10 timesRepeat: [ "lock and unlock until failure due to child taking the lock"
					accessor lockFile: fileStream
							from: 100
							to: 120
							exclusive: true
							ifFail: [self error: '#lockFile failed because child Squeak now holds a lock'].
					accessor unlockFile: fileStream from: 100 to: 120. "release lock to allow child to acquire it"
					delay wait]]
				raise: Error
				description: '#lockFile failed because child Squeak now holds a lock']
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - exclusive locking' stamp: 'dtl 8/30/2009 16:01'!
testFailLockOnLockedSupersetRegion
	"Two cooperating Squeak processes using mandatory file locking. When
	child Squeak takes a lock on a region, lock attempt in parent Squeak will
	fail. Fail when requesting a lock on a subset of the locked region."

	"(self selector: #testFailLockOnLockedSupersetRegion) run"

	self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 140) notNil]) ifNil: [self quitImage].
			delay wait; wait; wait.
			accessor unlockFile: fileStream from: 100 to: 140.
			self quitImage]
		doLocal:
			[self should: [
				10 timesRepeat: [ "lock and unlock until failure due to child taking the lock"
					accessor lockFile: fileStream
							from: 104
							to: 120
							exclusive: true
							ifFail: [self error: '#lockFile failed because child Squeak now holds a lock'].
					accessor unlockFile: fileStream from: 104 to: 120. "release lock to allow child to acquire it"
					delay wait]]
				raise: Error
				description: '#lockFile failed because child Squeak now holds a lock']
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - exclusive locking' stamp: 'dtl 8/30/2009 16:02'!
testFailRegionLockOnLockedFile
	"Two cooperating Squeak processes using mandatory file locking. When
	child Squeak takes a lock on a file, the region lock attempt in parent
	Squeak will fail."

	"(self selector: #testFailRegionLockOnLockedFile) run"

	self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream) notNil]) ifNil: [self quitImage].
			delay wait; wait; wait.
			accessor unlockFile: fileStream.
			self quitImage]
		doLocal:
			[self should: [
				10 timesRepeat: [ "lock and unlock until failure due to child taking the lock"
					accessor lockFile: fileStream
							from: 100
							to: 120
							exclusive: true
							ifFail: [self error: '#lockFile failed because child Squeak now holds a lock'].
					accessor unlockFile: fileStream from: 100 to: 120. "release lock to allow child to acquire it"
					delay wait]]
				raise: Error
				description: '#lockFile failed because child Squeak now holds a lock']
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 22:44'!
testLockEntireFileForWrite01
	"Local lock request should succeed"

	"(self selector: #testLockEntireFileForWrite01) run"

	| result |
	result := self
		doRemote: [nil]
		doLocal: [accessor lockFile: fileStream].
	self delay wait; wait.
	self assert: (result isKindOf: OSFileLock).
	result := accessor unlockFile: fileStream.
	self assert: (result isKindOf: OSFileLock)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:40'!
testLockEntireFileForWrite02
	"Local lock request should fail"

	"(self selector: #testLockEntireFileForWrite02) run"

	| result |
	result := self
		doRemote:
			[accessor lockFile: fileStream]
		doLocal: [accessor lockFile: fileStream].
	self delay wait; wait.
	self assert: result isNil.
	"Region is in the cache as a result of the (failed) lock request. This is legitimate
	because the lock exists in some other image, so we want to mark it as locked
	locally. However, it is not quite correct because the remote lock may pertain to
	a different but overlapping region. Nevertheless, this seems like reasonable
	behavior."
	result := accessor unlockFile: fileStream.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 22:45'!
testLockEntireFileForWrite03
	"Local lock request should succeed"

	"(self selector: #testLockEntireFileForWrite03) run"

	| result |
	result := self
		doRemote:
			[accessor lockFile: fileStream.
			accessor unlockFile: fileStream]
		doLocal:
			[accessor lockFile: fileStream].
	self delay wait; wait.
	self assert: (result isKindOf: OSFileLock).
	result := accessor unlockFile: fileStream.
	self assert: (result isKindOf: OSFileLock)! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:48'!
testLockEntireFileForWrite04
	"Local lock should be available"

	"(self selector: #testLockEntireFileForWrite04) run"

	| result |
	result := self
		doRemote: [nil]
		doLocal: [accessor isLockableFile: fileStream].
	self delay wait; wait.
	self should: result.
	"No lock on the region, so the unlock should fail."
	result := accessor unlockFile: fileStream.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:50'!
testLockEntireFileForWrite05
	"Local lock should be available"

	"(self selector: #testLockEntireFileForWrite05) run"

	| result |
	result := self
		doRemote:
			[accessor lockFile: fileStream]
		doLocal: [accessor isLockableFile: fileStream].
	self delay wait; wait.
	self shouldnt: result.
	"No local lock in the cache for this file, so the unlock call will appear to fail.
	Actually, it will have passed the unlock request through to the operating system,
	and it will actually unlock the file."
	result := accessor unlockFile: fileStream.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:49'!
testLockEntireFileForWrite06
	"Local lock should be available"

	"(self selector: #testLockEntireFileForWrite06) run"

	| result |
	result := self
		doRemote:
			[accessor lockFile: fileStream.
			accessor unlockFile: fileStream]
		doLocal:
			[accessor isLockableFile: fileStream].
	self delay wait; wait.
	self should: result.
	"No local lock in the cache for this file, so the unlock call will appear to fail.
	Actually, it will have passed the unlock request through to the operating system,
	but it will have no effect because the file was never locked."
	result := accessor unlockFile: fileStream.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'fail blocks' stamp: 'dtl 5/11/2006 07:22'!
testLockFileIfFail

	"(self selector: #testLockFileIfFail) run"

	| fail |
	fail := false.
	accessor lockFile: fileStream ifFail: [fail := true].
	self shouldnt: fail.
	fileStream close.
	accessor lockFile: fileStream ifFail: [fail := true].
	self should: fail
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'fail blocks' stamp: 'dtl 11/28/2010 19:05'!
testLockFileRegionIfFail

	"(self selector: #testLockFileRegionIfFail) run"

	| fail |
	fail := false.
	accessor
		lockFile: fileStream from: 100 to: 199 exclusive: true
		ifFail: [fail := true].
	self shouldnt: fail.
	fileStream close.
	accessor
		lockFile: fileStream from: 100 to: 199 exclusive: true
		ifFail: [fail := true].
	self should: fail.
	accessor
		lockFile: fileStream from: 100 to: 199
		ifFail: [fail := true].
	self should: fail
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/10/2006 07:04'!
testLockPreviouslyLockedFile

	self subclassResponsibility! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/10/2006 07:10'!
testLockPreviouslyLockedFileRegion

	self subclassResponsibility! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 22:46'!
testLockRegionForRead01
	"Read lock, the local lock request should succeed"

	"(self selector: #testLockRegionForRead01) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: false]
		doLocal: [accessor lockFile: fileStream from: 100 to: 199 exclusive: false].
	self delay wait; wait.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199)).
	result := accessor unlockFile: fileStream from: 100 to: 199 exclusive: false.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199))! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:53'!
testLockRegionForRead02
	"Read lock, the local lock request should fail when requesting a write lock"

	"(self selector: #testLockRegionForRead02) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true]
		doLocal: [accessor lockFile: fileStream from: 100 to: 199 exclusive: false].
	self delay wait; wait.
	self assert: result isNil.
	result := accessor unlockFile: fileStream from: 100 to: 199.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 22:46'!
testLockRegionForWrite01
	"Local lock request should succeed"

	"(self selector: #testLockRegionForWrite01) run"

	| result |
	result := self
		doRemote: [nil]
		doLocal: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true].
	self delay wait; wait.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199)).
	result := accessor unlockFile: fileStream from: 100 to: 199.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199))! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 4/30/2006 13:47'!
testLockRegionForWrite02
	"Write lock, the local lock request should fail"

	"(self selector: #testLockRegionForWrite02) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true]
		doLocal: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true].
	self delay wait; wait.
	self assert: result isNil.
	"Lock is not in the local cache, so unlock will appear to fail. Actually, the region
	will be unlocked."
	accessor unlockFile: fileStream from: 100 to: 199.
	self assert: result isNil.
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:55'!
testLockRegionForWrite03
	"Write lock, nonoverlapping regions, the lock check should succeed"

	"(self selector: #testLockRegionForWrite03) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true]
		doLocal: [accessor isLockableFile: fileStream from: 200 to: 299 exclusive: true].
	self delay wait; wait.
	self assert: result.
	result := accessor unlockFile: fileStream from: 200 to: 299.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:53'!
testLockRegionForWrite04
	"Write lock, overlapping regions, the local lock request should fail"

	"(self selector: #testLockRegionForWrite04) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true]
		doLocal: [accessor lockFile: fileStream from: 150 to: 249 exclusive: true].
	self delay wait; wait.
	self assert: result isNil.
	result := accessor unlockFile: fileStream from: 150 to: 249.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 22:47'!
testLockRegionForWrite05
	"Write lock, nonoverlapping regions, the local lock request should succeed"

	"(self selector: #testLockRegionForWrite05) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true]
		doLocal: [accessor lockFile: fileStream from: 200 to: 299 exclusive: true].
	self delay wait; wait.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (200 to: 299)).
	result := accessor unlockFile: fileStream from: 200 to: 299.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (200 to: 299))! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:53'!
testLockRegionForWrite06
	"Write lock, overlapping regions, the local lock request should fail"

	"(self selector: #testLockRegionForWrite06) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true]
		doLocal: [accessor lockFile: fileStream from: 199 to: 199 exclusive: true].
	self delay wait; wait.
	self assert: result isNil.
	result := accessor unlockFile: fileStream from: 199 to: 199.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:55'!
testLockRegionForWrite07
	"Write lock, overlapping regions, the lock check should fail"

	"(self selector: #testLockRegionForWrite07) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true]
		doLocal: [accessor isLockableFile: fileStream from: 100 to: 199 exclusive: true].
	self delay wait; wait.
	self deny: result.
	result := accessor unlockFile: fileStream from: 100 to: 199.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/13/2006 09:55'!
testLockRegionForWrite08
	"Write lock, overlapping regions, the lock check should fail"

	"(self selector: #testLockRegionForWrite08) run"

	| result |
	result := self
		doRemote: [accessor lockFile: fileStream from: 100 to: 199 exclusive: true]
		doLocal: [accessor isLockableFile: fileStream from: 150 to: 299 exclusive: true].
	self delay wait; wait.
	self deny: result.
	result := accessor unlockFile: fileStream from: 150 to: 299.
	self assert: (self isValidUnlockResult: result)
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - exclusive locking' stamp: 'dtl 8/30/2009 16:03'!
testNoFailLockOnAdjacentLockedRegions
	"Two cooperating Squeak processes using mandatory file locking. When
	child Squeak takes a lock on a region, lock attempt in parent Squeak will
	fail. No failure occurs when the regions are adjacent but do not overlap."

	"(self selector: #testNoFailLockOnAdjacentLockedRegions) run"

	self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 100 to: 103) notNil]) ifNil: [self quitImage].
			delay wait; wait; wait.
			accessor unlockFile: fileStream from: 100 to: 103.
			self quitImage]
		doLocal:
			[self shouldnt: [
				10 timesRepeat: [ "lock and unlock until failure due to child taking the lock"
					accessor lockFile: fileStream
							from: 104
							to: 120
							exclusive: true
							ifFail: [self error: '#lockFile failed because child Squeak now holds a lock'].
					accessor unlockFile: fileStream from: 104 to: 120. "release lock to allow child to acquire it"
					delay wait]]
				raise: Error]
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - exclusive locking' stamp: 'dtl 8/30/2009 16:03'!
testNoFailLockOnDifferentLockedRegion
	"Two cooperating Squeak processes using mandatory file locking. When
	child Squeak takes a lock on a region, lock attempt in parent Squeak will
	fail. No failure occurs if the regions are different and do not overlap."

	"(self selector: #testNoFailLockOnDifferentLockedRegion) run"

	self
		doRemote:
			[(self tryUntil: [(accessor lockFile: fileStream from: 50 to: 80) notNil]) ifNil: [self quitImage].
			delay wait; wait; wait.
			accessor unlockFile: fileStream from: 50 to: 80.
			self quitImage]
		doLocal:
			[self shouldnt: [
				10 timesRepeat: [ "lock and unlock until failure due to child taking the lock"
					accessor lockFile: fileStream
							from: 104
							to: 120
							exclusive: true
							ifFail: [self error: '#lockFile failed because child Squeak now holds a lock'].
					accessor unlockFile: fileStream from: 104 to: 120. "release lock to allow child to acquire it"
					delay wait]]
				raise: Error]
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'fail blocks' stamp: 'dtl 5/11/2006 07:24'!
testUnlockFileIfFail

	"(self selector: #testUnlockFileIfFail) run"

	| fail |
	fail := false.
	accessor lockFile: fileStream ifFail: [fail := true].
	self shouldnt: fail.
	accessor unlockFile: fileStream ifFail: [fail := true].
	self shouldnt: fail.
	fileStream close.
	accessor unlockFile: fileStream ifFail: [fail := true].
	self should: fail
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'fail blocks' stamp: 'dtl 11/28/2010 19:11'!
testUnlockFileRegionIfFail

	"(self selector: #testUnlockFileRegionIfFail) run"

	| fail |
	fail := false.
	accessor
		lockFile: fileStream from: 100 to: 199 exclusive: true
		ifFail: [fail := true].
	self shouldnt: fail.
	accessor
		unlockFile: fileStream from: 100 to: 199 exclusive: true
		ifFail: [fail := true].
	self shouldnt: fail.
	fileStream close.
	accessor
		unlockFile: fileStream from: 100 to: 199 exclusive: true
		ifFail: [fail := true].
	self should: fail.
	accessor
		unlockFile: fileStream from: 100 to: 199
		ifFail: [fail := true].
	self should: fail

! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 06:53'!
testUnlockPreviouslyLockedFile

	self subclassResponsibility! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 06:54'!
testUnlockPreviouslyLockedFileRegion

	self subclassResponsibility! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 12:56'!
testWin32LockEmulation01
	"Make sure the cache does what is expected, and that it can be disabled
	Expect non-exclusive read locks to to succeed."

	"(self selector: #testWin32LockEmulation01) run"

	| result |
	accessor class emulateWin32FileLocking: true.
	result := accessor lockFile: fileStream from: 100 to: 199 exclusive: false.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199)).
	result := accessor lockFile: fileStream from: 100 to: 199 exclusive: false.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199)).
	result := accessor unlockFile: fileStream from: 100 to: 199 exclusive: false.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199))
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 12:56'!
testWin32LockEmulation02
	"Make sure the cache does what is expected, and that it can be disabled.
	Expect exclusive read-write locks to be protected by the cache."

	"(self selector: #testWin32LockEmulation02) run"

	| result |
	accessor class emulateWin32FileLocking: true.
	result := accessor lockFile: fileStream from: 100 to: 199 exclusive: true.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199)).
	result := accessor lockFile: fileStream from: 100 to: 199 exclusive: true.
	self assert: result isNil.
	result := accessor unlockFile: fileStream from: 100 to: 199 exclusive: true.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199))
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 12:57'!
testWin32LockEmulation03
	"Make sure the cache does what is expected, and that it can be disabled.
	Expect the cache to protect the file even if we open a second stream"

	"(self selector: #testWin32LockEmulation03) run"

	| result stream2 |
	accessor class emulateWin32FileLocking: true.
	result := accessor lockFile: fileStream from: 100 to: 199 exclusive: true.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199)).
	stream2 := FileStream fileNamed: 'junkfile'.
	[result := accessor lockFile: stream2 from: 100 to: 199 exclusive: true.
	self assert: result isNil]
		ensure: [stream2 close].
	result := accessor unlockFile: fileStream from: 100 to: 199 exclusive: true.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199))
! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'testing' stamp: 'dtl 5/7/2006 12:57'!
testWin32LockEmulation04
	"Make sure the cache does what is expected, and that it can be disabled.
	Expect the cache to allow a lock after the stream has been closed (since the OS will have freed its lock)"

	"(self selector: #testWin32LockEmulation04) run"

	| result stream2 |
	accessor class emulateWin32FileLocking: true.
	stream2 := FileStream fileNamed: 'junkfile'.
	[result := accessor lockFile: stream2 from: 100 to: 199 exclusive: true.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199))]
		ensure: [stream2 close].
	result := accessor lockFile: fileStream from: 100 to: 199 exclusive: true.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199)).
	result := accessor unlockFile: fileStream from: 100 to: 199 exclusive: true.
	self assert: (result isKindOf: OSFileRegionLock).
	self assert: (result interval = (100 to: 199))! !

!AbstractUnixProcessFileLockingTestCase methodsFor: 'private' stamp: 'dtl 4/30/2006 13:47'!
tryUntil: aBlock
	"Repeat until block evaluates true, but do not get in an infinite loop if a primitive is failing"

	| result |
	10 timesRepeat:
		[result := aBlock value.
		result ifTrue: [^ result].
		delay wait].
	^ result
! !

AbstractUnixProcessFileLockingTestCase subclass: #UnixProcessUnixFileLockingTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Tests'!

!UnixProcessUnixFileLockingTestCase commentStamp: 'dtl 4/30/2006 14:03' prior: 0!
Test file locking with the UnixOSProcessPlugin using Unix file locking semantics. The test suite requires that OSProcess and CommandShell be loaded in the image.

These tests rely on a remote Squeak image to test file locks between cooperating Unix processes. This may be timing dependent (see #delay, set in #setUp, and cleanup in #tearDown). In case of intermittent failures, try running the failed test individually. In some cases it may be necessary to restart Squeak in order to clear leftover file locks from previous failed tests.!

!UnixProcessUnixFileLockingTestCase class methodsFor: 'testing' stamp: 'dtl 1/13/2007 15:40'!
isAbstract
	"Override to true if a TestCase subclass is Abstract and should not have
	TestCase instances built from it"

	^self name = #TestCase
! !

!UnixProcessUnixFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/13/2006 09:38'!
isValidUnlockResult: result
	"With no lock cache, the unlock requests will succeed even though there
	is no lock to be removed."	

	^ result isKindOf: OSFileLock
! !

!UnixProcessUnixFileLockingTestCase methodsFor: 'running' stamp: 'dtl 4/30/2006 14:05'!
setUp

	"Set system to use Unix file locking semantics. The tearDown method is expected to restore the setting."

	OSProcessAccessor emulateWin32FileLocking: false.
	^ super setUp
! !

!UnixProcessUnixFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 07:12'!
testLockPreviouslyLockedFile
	"Locking a previously locked file is permitted"

	"(self selector: #testLockPreviouslyLockedFile) run"

	accessor lockFile: fileStream.
	self should: (accessor isLockableFile: fileStream).
	self should: ((accessor lockFile: fileStream) isKindOf: OSFileLock).
	accessor unlockFile: fileStream! !

!UnixProcessUnixFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 07:12'!
testLockPreviouslyLockedFileRegion
	"Locking a previously locked file region is permitted"

	"(self selector: #testLockPreviouslyLockedFileRegion) run"

	accessor
		lockFile: fileStream
		from: 100
		to: 199
		exclusive: true.
	self should: (accessor
		isLockableFile: fileStream
		from: 100
		to: 199
		exclusive: true).
	self should: ((accessor
		lockFile: fileStream
		from: 100
		to: 199
		exclusive: true)
		isKindOf: OSFileLock).
	accessor unlockFile: fileStream! !

!UnixProcessUnixFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 07:07'!
testUnlockPreviouslyLockedFile
	"The unlock operation will succeed even if no lock is in effect"

	"(self selector: #testUnlockPreviouslyLockedFile) run"

	self should: ((accessor unlockFile: fileStream) isKindOf: OSFileLock).
	accessor lockFile: fileStream.
	self should: ((accessor unlockFile: fileStream) isKindOf: OSFileLock)
! !

!UnixProcessUnixFileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 07:07'!
testUnlockPreviouslyLockedFileRegion
	"The unlock operation will succeed even if no lock is in effect"

	"(self selector: #testUnlockPreviouslyLockedFileRegion) run"

	self should: ((accessor
		unlockFile: fileStream
		from: 100
		to: 199
		exclusive: true) isKindOf: OSFileLock).
	accessor
		lockFile: fileStream
		from: 100
		to: 199
		exclusive: true.
	self should: ((accessor
		unlockFile: fileStream
		from: 100
		to: 199
		exclusive: true) isKindOf: OSFileLock)
! !

AbstractUnixProcessFileLockingTestCase subclass: #UnixProcessWin32FileLockingTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Tests'!

!UnixProcessWin32FileLockingTestCase commentStamp: 'dtl 4/30/2006 14:03' prior: 0!
Test file locking with the UnixOSProcessPlugin using Windows file locking semantics. The test suite requires that OSProcess and CommandShell be loaded in the image.

These tests rely on a remote Squeak image to test file locks between cooperating Unix processes. This may be timing dependent (see #delay, set in #setUp, and cleanup in #tearDown). In case of intermittent failures, try running the failed test individually. In some cases it may be necessary to restart Squeak in order to clear leftover file locks from previous failed tests.!

!UnixProcessWin32FileLockingTestCase class methodsFor: 'testing' stamp: 'dtl 1/13/2007 15:40'!
isAbstract
	"Override to true if a TestCase subclass is Abstract and should not have
	TestCase instances built from it"

	^self name = #TestCase
! !

!UnixProcessWin32FileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/13/2006 09:47'!
isValidUnlockResult: result
	"With a lock cache, the unlock requests will fail when if there
	is no lock to be removed."	

	^ result == nil
! !

!UnixProcessWin32FileLockingTestCase methodsFor: 'running' stamp: 'dtl 4/30/2006 14:06'!
setUp

	"Set system to use Windows file locking semantics. The tearDown method is expected to restore the setting."

	OSProcessAccessor emulateWin32FileLocking: true.
	^ super setUp
! !

!UnixProcessWin32FileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 07:13'!
testLockPreviouslyLockedFile
	"A previously locked file cannot be locked."

	"(self selector: #testLockPreviouslyLockedFile) run"

	accessor lockFile: fileStream.
	self shouldnt: (accessor isLockableFile: fileStream).
	self should: (accessor lockFile: fileStream) isNil.
	accessor unlockFile: fileStream! !

!UnixProcessWin32FileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 07:13'!
testLockPreviouslyLockedFileRegion
	"A previously locked file region cannot be locked."

	"(self selector: #testLockPreviouslyLockedFileRegion) run"

	accessor
		lockFile: fileStream
		from: 100
		to: 199
		exclusive: true.
	self shouldnt: (accessor
		isLockableFile: fileStream
		from: 100
		to: 199
		exclusive: true).
	self should: ((accessor
		lockFile: fileStream
		from: 100
		to: 199
		exclusive: true) isNil).
	accessor unlockFile: fileStream! !

!UnixProcessWin32FileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/11/2006 07:08'!
testUnlockPreviouslyLockedFile
	"The unlock operation will fail if lock is in effect"

	"(self selector: #testUnlockPreviouslyLockedFile) run"

	self shouldnt: ((accessor unlockFile: fileStream) isKindOf: OSFileLock).
	accessor lockFile: fileStream.
	self should: ((accessor unlockFile: fileStream) isKindOf: OSFileLock).
	self shouldnt: ((accessor unlockFile: fileStream) isKindOf: OSFileLock)
! !

!UnixProcessWin32FileLockingTestCase methodsFor: 'testing - platform specific' stamp: 'dtl 5/12/2006 07:48'!
testUnlockPreviouslyLockedFileRegion
	"The unlock operation will fail if a lock is in effect"

	"(self selector: #testUnlockPreviouslyLockedFileRegion) run"

	self shouldnt: ((accessor
		unlockFile: fileStream
		from: 100
		to: 199
		exclusive: true) isKindOf: OSFileLock).
	accessor
		lockFile: fileStream
		from: 100
		to: 199
		exclusive: true.
	self should: ((accessor
		unlockFile: fileStream
		from: 100
		to: 199
		exclusive: true) isKindOf: OSFileLock).
	self shouldnt: ((accessor
		unlockFile: fileStream
		from: 100
		to: 199
		exclusive: true) isKindOf: OSFileLock)
! !

TestCase subclass: #AioEventHandlerTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Tests'!

!AioEventHandlerTestCase commentStamp: 'dtl 7/10/2005 15:41' prior: 0!
Test AioEventHandler and AioPlugin. Provides fair coverage of IO readable events, minimal coverage of IO writable events, and no real coverage for IO exception events. The writable events and exception events probably work, but this test suite will not prove it.!

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 8/7/2005 13:02'!
testEnableHandleAndDisable

	| eventHandler anOpenFile fileHandle aioHandleResult sema semaIndex aioEnableResult aioDisableResult |
	eventHandler := AioEventHandler new.
	anOpenFile := SourceFiles at: 1.
	fileHandle := eventHandler handleForFile: anOpenFile.
	sema := Semaphore new.
	[semaIndex := Smalltalk registerExternalObject: sema.
	[aioEnableResult := eventHandler
			aioEnable: fileHandle
			forSemaphore: semaIndex
			externalObject: true]
		on: Warning
		do: ["Suppress warning dialog. Unit test will fail next assert."].
	self assert: semaIndex == aioEnableResult.
	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: true
		readEvents: true
		writeEvents: true.
	self assert: aioHandleResult == 2r0111. "return bitmask, all three bits set"
	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: true
		readEvents: false
		writeEvents: true.
	self assert: aioHandleResult == 2r0101.
	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: false
		writeEvents: true.
	self assert: aioHandleResult == 2r0100.
	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: true
		readEvents: false
		writeEvents: false.
	self assert: aioHandleResult == 2r0001.
	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: false
		writeEvents: false.
	self assert: aioHandleResult == 2r0000.
	aioDisableResult := eventHandler aioDisable: fileHandle.
	self assert: aioDisableResult == fileHandle]
		ensure:
			[Smalltalk unregisterExternalObject: sema.
			(fileHandle isKindOf: Integer) ifTrue: [eventHandler aioDisable: fileHandle]]! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 1/23/2013 19:40'!
testFileReadableEvent

	| eventHandler anOpenFile fileHandle aioHandleResult sema semaIndex aioEnableResult fileName |
	fileName := 'DeleteThisTemporaryFile'.

	eventHandler := AioEventHandler new.
	OSProcess deleteFileNamed: fileName.
	[anOpenFile := FileStream fileNamed: fileName.
	fileHandle := eventHandler handleForFile: anOpenFile.
	sema := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: sema.
	[aioEnableResult := eventHandler
			aioEnable: fileHandle
			forSemaphore: semaIndex
			externalObject: true]
		on: Warning
		do: ["Suppress warning dialog. Unit test will fail next assert."].
	self assert: semaIndex == aioEnableResult.

	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.
	self assert: aioHandleResult == 2r010.
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self assert: sema isSignaled. "File is empty but readable. A signal is sent once,
	presumably in the interest of priming the pump. A listening process would
	begin reading data based on having received this initial signal."
	sema wait. "consume the signal"
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self deny: sema isSignaled. "Nothing has happened to indicate more data available."

	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.
	self assert: aioHandleResult == 2r010.
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self assert: sema isSignaled. "File is empty but readable. A signal is sent once,
	presumably in the interest of priming the pump. A listening process would
	begin reading data based on having received this initial signal."
	sema wait. "consume the signal"
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self deny: sema isSignaled. "Nothing has happened to indicate more data available."

	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.
	self assert: aioHandleResult == 2r010.
	anOpenFile nextPutAll: 'write some stuff to the file'; flush.
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self assert: sema isSignaled. "File is empty but readable. A signal is sent once,
	presumably in the interest of priming the pump. A listening process would
	begin reading data based on having received this initial signal."
	sema wait. "consume the signal"

	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.
	self assert: aioHandleResult == 2r010.
	anOpenFile upToEnd.
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self assert: sema isSignaled. "File is empty but readable. A signal is sent once,
	presumably in the interest of priming the pump. A listening process would
	begin reading data based on having received this initial signal."
	sema wait. "consume the signal"

	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self deny: sema isSignaled. "handler has not been reset"]
		ensure:
			[[anOpenFile close] on: Error do: [].
			Smalltalk unregisterExternalObject: sema.
			(fileHandle isKindOf: Integer) ifTrue: [eventHandler aioDisable: fileHandle]]! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 1/23/2013 19:41'!
testFileWritableEvent

	| eventHandler anOpenFile fileHandle aioHandleResult sema semaIndex aioEnableResult fileName |
	fileName := 'DeleteThisTemporaryFile'.

	eventHandler := AioEventHandler new.
	OSProcess deleteFileNamed: fileName.
	[anOpenFile := FileStream fileNamed: fileName.
	fileHandle := eventHandler handleForFile: anOpenFile.
	sema := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: sema.
	[aioEnableResult := eventHandler
			aioEnable: fileHandle
			forSemaphore: semaIndex
			externalObject: true]
		on: Warning
		do: ["Suppress warning dialog. Unit test will fail next assert."].
	self assert: semaIndex == aioEnableResult.

	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: false
		writeEvents: true.
	self assert: aioHandleResult == 2r0100.
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self assert: sema isSignaled. "File is empty but writable. A signal is sent once,
	presumably in the interest of priming the pump. A listening process would
	begin writing data based on having received this initial signal."
	sema wait. "consume the signal"
	self deny: sema isSignaled.

	"Reset handler"
	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: false
		writeEvents: true.
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self assert: sema isSignaled. "File is empty but writable. A signal is sent once,
	presumably in the interest of priming the pump. A listening process would
	begin writing data based on having received this initial signal."
	sema wait. "consume the signal"
	self deny: sema isSignaled.

	"Reset handler"
	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: false
		writeEvents: true.
	anOpenFile upToEnd.
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self assert: sema isSignaled. "File is empty but writable. A signal is sent once,
	presumably in the interest of priming the pump. A listening process would
	begin writing data based on having received this initial signal."
	sema wait. "consume the signal"
	self deny: sema isSignaled.

	"Reset handler"
	aioHandleResult := eventHandler
		aioHandle: fileHandle
		exceptionEvents: false
		readEvents: false
		writeEvents: true.
	anOpenFile nextPutAll: 'write some stuff to the file'; flush.
	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self assert: sema isSignaled. "File is empty but writable. A signal is sent once,
	presumably in the interest of priming the pump. A listening process would
	begin writing data based on having received this initial signal."
	sema wait. "consume the signal"
	self deny: sema isSignaled.

	(Delay forMilliseconds: 20) wait. "Allow signals to be handled"
	self deny: sema isSignaled. "handler has not been reset"]
		ensure:
			[[anOpenFile close] on: Error do: [].
			Smalltalk unregisterExternalObject: sema.
			(fileHandle isKindOf: Integer) ifTrue: [eventHandler aioDisable: fileHandle]]! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 7/9/2005 17:10'!
testHandleForFile

	| eventHandler anOpenFile fileHandle |
	eventHandler := AioEventHandler new.
	anOpenFile := SourceFiles at: 1.
	fileHandle := eventHandler handleForFile: anOpenFile.
	self assert: fileHandle notNil.
	self assert: (fileHandle isKindOf: Integer)
! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 7/9/2005 17:10'!
testHandleForSocket

	| eventHandler socket socketHandle |
	eventHandler := AioEventHandler new.
	socket := Socket newTCP.
	socketHandle := eventHandler handleForSocket: socket.
	self assert: socketHandle notNil.
	self assert: (socketHandle isKindOf: Integer).
	self assert: (socketHandle ~= 0) "0 is stdin on unix, unlikely to be used on other platforms"
! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 7/9/2005 16:41'!
testPrimAioModuleName

	| eventHandler moduleName |
	eventHandler := AioEventHandler new.
	moduleName := eventHandler primAioModuleName.
	self assert: ('AioPlugin*' match: moduleName)
! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 7/9/2005 16:42'!
testPrimAioModuleVersionString

	| eventHandler versionString |
	eventHandler := AioEventHandler new.
	versionString := eventHandler primAioModuleName.
	self assert: (versionString isKindOf: String)
! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 8/7/2005 12:58'!
testSocketExceptionEvent
	"Close the client socket to generate an event on the server socket"

	| port serverName serverTcpSocket clientTcpSocket sema semaIndex socketHandle eventHandler receivedData |
	port := 8086.
	serverName := '127.0.0.1'.
	Socket initializeNetwork.

	"Create the server (reader) socket"
	serverTcpSocket := Socket newTCP.
	serverTcpSocket listenOn: port.

	"Create the client (writer) socket"
	clientTcpSocket := Socket newTCP.
	clientTcpSocket connectTo: (NetNameResolver addressFromString: serverName) port: port.

	"Esstablish socket connection"
	[serverTcpSocket waitForConnectionFor: 10] fork.
	clientTcpSocket waitForConnectionFor: 10.

	["Set up a read event handler on the server socket"
	sema := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: sema.
	eventHandler := AioEventHandler new.
	socketHandle := eventHandler handleForSocket: serverTcpSocket.

	"An aio handler may already have been established for the socket
	by the SocketPlugin. The following request disables any existing handling
	in order to prevent a warning message on the console stderr."
	eventHandler aioDisable: socketHandle.

	"Now set our own handler."
	[eventHandler
			aioEnable: socketHandle
			forSemaphore: semaIndex
			externalObject: true]
		on: Warning
		do: ["Suppress warning dialog. Unit test will fail next assert."].
	eventHandler
		aioHandle: socketHandle
		exceptionEvents: true
		readEvents: false
		writeEvents: false.
	self deny: sema isSignaled.
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	self deny: sema isSignaled.
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.
	self deny: sema isSignaled.
	receivedData := serverTcpSocket receiveAvailableData.
	clientTcpSocket close.
	(Delay forMilliseconds: 200) wait.
	receivedData := serverTcpSocket receiveAvailableData.
	(Delay forMilliseconds: 200) wait.
"	self assert: sema isSignaled.	FIXME: no error generated; need to find a way to generate an error -dtl"
	self deny: sema isSignaled.
	serverTcpSocket close]
		ensure:
			[Smalltalk unregisterExternalObject: sema.
			(socketHandle isKindOf: Integer) ifTrue: [eventHandler aioDisable: socketHandle]]! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 8/7/2005 13:01'!
testSocketExceptionEvent2
	"Close the client socket to generate an event on the server socket"

	| port serverTcpSocket sema semaIndex socketHandle eventHandler |
	port := 8086.
	Socket initializeNetwork.

	"Create the server (reader) socket"
	serverTcpSocket := Socket newTCP.
	serverTcpSocket listenOn: port.

	["Set up an exception event handler on the server socket"
	sema := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: sema.
	eventHandler := AioEventHandler new.
	socketHandle := eventHandler handleForSocket: serverTcpSocket.

	"An aio handler may already have been established for the socket
	by the SocketPlugin. The following request disables any existing handling
	in order to prevent a warning message on the console stderr."
	eventHandler aioDisable: socketHandle.

	"Now set our own handler."
	[eventHandler
			aioEnable: socketHandle
			forSemaphore: semaIndex
			externalObject: true]
		on: Warning
		do: ["Suppress warning dialog. Unit test will fail next assert."].
	eventHandler
		aioHandle: socketHandle
		exceptionEvents: true
		readEvents: false
		writeEvents: true.
	self deny: sema isSignaled.
	[serverTcpSocket
		primSocket: serverTcpSocket socketHandle
		sendData: 'a string to send'
		startIndex: 1
		count: 10] on: Error do: [self halt].
	serverTcpSocket
		receiveSomeDataInto: (String new: 1000)
		startingAt: 1.

"	self assert: sema isSignaled.	FIXME: no error generated; need to find a way to generate an error -dtl"
	self deny: sema isSignaled.

	serverTcpSocket close]
		ensure:
			[Smalltalk unregisterExternalObject: sema.
			(socketHandle isKindOf: Integer) ifTrue: [eventHandler aioDisable: socketHandle]]! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 8/7/2005 13:02'!
testSocketReadableEvent

	| port serverName serverTcpSocket clientTcpSocket sema semaIndex socketHandle eventHandler receivedData |
	port := 8086.
	serverName := '127.0.0.1'.
	Socket initializeNetwork.

	"Create the server (reader) socket"
	serverTcpSocket := Socket newTCP.
	serverTcpSocket listenOn: port.

	"Create the client (writer) socket"
	clientTcpSocket := Socket newTCP.
	clientTcpSocket connectTo: (NetNameResolver addressFromString: serverName) port: port.

	"Establish socket connection"
	[serverTcpSocket waitForConnectionFor: 10] fork.
	clientTcpSocket waitForConnectionFor: 10.

	["Set up a read event handler on the server socket"
	sema := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: sema.
	eventHandler := AioEventHandler new.
	socketHandle := eventHandler handleForSocket: serverTcpSocket.

	"An aio handler may already have been established for the socket
	by the SocketPlugin. The following request disables any existing handling
	in order to prevent a warning message on the console stderr."
	eventHandler aioDisable: socketHandle.

	"Now set our own handler."
	[eventHandler
			aioEnable: socketHandle
			forSemaphore: semaIndex
			externalObject: true]
		on: Warning
		do: ["Suppress warning dialog. Unit test will fail next assert."].

	eventHandler
		aioHandle: socketHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.
	self deny: sema isSignaled.
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	self assert: sema isSignaled.
	sema wait. "consume the signal"
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.

	eventHandler
		aioHandle: socketHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.
	self deny: sema isSignaled.
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	self assert: sema isSignaled.
	sema wait. "consume the signal"
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.

	eventHandler
		aioHandle: socketHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.
	self deny: sema isSignaled.
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	self assert: sema isSignaled.
	sema wait. "consume the signal"
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.

	clientTcpSocket close.
	serverTcpSocket close]
		ensure:
			[Smalltalk unregisterExternalObject: sema.
			(socketHandle isKindOf: Integer) ifTrue: [eventHandler aioDisable: socketHandle]]! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 8/7/2005 12:59'!
testSocketReadableEventWithMaskNotSet
	"Same as testSocketReadableEvent, except that the mask bit for read events is not set,
	so we expect to not receive the event notification. This is just a check to ensure that
	the event mask is working."

	| port serverName serverTcpSocket clientTcpSocket sema semaIndex socketHandle eventHandler receivedData |
	port := 8086.
	serverName := '127.0.0.1'.
	Socket initializeNetwork.

	"Create the server (reader) socket"
	serverTcpSocket := Socket newTCP.
	serverTcpSocket listenOn: port.

	"Create the client (writer) socket"
	clientTcpSocket := Socket newTCP.
	clientTcpSocket connectTo: (NetNameResolver addressFromString: serverName) port: port.

	"Establish socket connection"
	[serverTcpSocket waitForConnectionFor: 10] fork.
	clientTcpSocket waitForConnectionFor: 10.

	["Set up a read event handler on the server socket"
	sema := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: sema.
	eventHandler := AioEventHandler new.
	socketHandle := eventHandler handleForSocket: serverTcpSocket.

	"An aio handler may already have been established for the socket
	by the SocketPlugin. The following request disables any existing handling
	in order to prevent a warning message on the console stderr."
	eventHandler aioDisable: socketHandle.

	"Now set our own handler."
	[eventHandler
			aioEnable: socketHandle
			forSemaphore: semaIndex
			externalObject: true]
		on: Warning
		do: ["Suppress warning dialog. Unit test will fail next assert."].
	eventHandler
		aioHandle: socketHandle
		exceptionEvents: true
		readEvents: false	"Read event bit not set"
		writeEvents: false.
	self deny: sema isSignaled.
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	self deny: sema isSignaled. "This time we should not receive a signal"
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.
	clientTcpSocket close.
	serverTcpSocket close]
		ensure:
			[Smalltalk unregisterExternalObject: sema.
			(socketHandle isKindOf: Integer) ifTrue: [eventHandler aioDisable: socketHandle]]! !

!AioEventHandlerTestCase methodsFor: 'testing' stamp: 'dtl 8/7/2005 12:59'!
testSuspendAioForSocketReadableEvent

	| port serverName serverTcpSocket clientTcpSocket sema semaIndex socketHandle eventHandler receivedData |
	port := 8086.
	serverName := '127.0.0.1'.
	Socket initializeNetwork.

	"Create the server (reader) socket"
	serverTcpSocket := Socket newTCP.
	serverTcpSocket listenOn: port.

	"Create the client (writer) socket"
	clientTcpSocket := Socket newTCP.
	clientTcpSocket connectTo: (NetNameResolver addressFromString: serverName) port: port.

	"Establish socket connection"
	[serverTcpSocket waitForConnectionFor: 10] fork.
	clientTcpSocket waitForConnectionFor: 10.

	["Set up a read event handler on the server socket"
	sema := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: sema.
	eventHandler := AioEventHandler new.
	socketHandle := eventHandler handleForSocket: serverTcpSocket.

	"An aio handler may already have been established for the socket
	by the SocketPlugin. The following request disables any existing handling
	in order to prevent a warning message on the console stderr."
	eventHandler aioDisable: socketHandle.

	"Now set our own handler."
	[eventHandler
			aioEnable: socketHandle
			forSemaphore: semaIndex
			externalObject: true]
		on: Warning
		do: ["Suppress warning dialog. Unit test will fail next assert."].

	eventHandler
		aioHandle: socketHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.

	self deny: sema isSignaled.

	"Handler generates a signal"
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.
	self assert: sema isSignaled.
	sema wait.
	self deny: sema isSignaled.

	"Handler has not been reset, so no signal this time"
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.
	self deny: sema isSignaled.

	"Reset the handler"
	eventHandler
		aioHandle: socketHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.

	"Handler generates a signal"
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.
	self assert: sema isSignaled.
	sema wait.
	self deny: sema isSignaled.

	"Reset the handler"
	eventHandler
		aioHandle: socketHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.
	"But temporarily disable it"
	eventHandler aioSuspend: socketHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.

	"Handler suspended, so no signal this time"
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.
	self deny: sema isSignaled.

	"Reset the handler"
	eventHandler
		aioHandle: socketHandle
		exceptionEvents: false
		readEvents: true
		writeEvents: false.

	"Handler generates a signal"
	clientTcpSocket sendData: 'some data to send to the socket'.
	(Delay forMilliseconds: 200) wait.
	receivedData := serverTcpSocket receiveAvailableData.
	self assert: 'some data to send to the socket' = receivedData.
	self assert: sema isSignaled.
	sema wait.
	self deny: sema isSignaled.

	clientTcpSocket close.
	serverTcpSocket close]
		ensure:
			[Smalltalk unregisterExternalObject: sema.
			(socketHandle isKindOf: Integer) ifTrue: [eventHandler aioDisable: socketHandle]]! !

TestCase subclass: #OSPipeTestCase
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Tests'!

!OSPipeTestCase commentStamp: '<historical>' prior: 0!
Test operation of OSPipe in blocking and nonBlocking mode.!

!OSPipeTestCase methodsFor: 'private' stamp: 'dtl 11/28/2010 14:05'!
primTestEndOfFileFlag: aSqFileStruct
	"Answer whether the file represented by aSqFileStruct is at end of file, as determined
	by a call to feof(). This is different from StandardFileStream>>primAtEnd: which answers
	true if the file pointer is at the end of the file, but which does not call feof() to
	determine that an end of file condition has occurred. The difference is significant
	if aSqFileStruct represents a pipe or a device file, which may not be positionable
	in the sense of a conventional disk file."

	<primitive: 'primitiveTestEndOfFileFlag' module: 'UnixOSProcessPlugin'>
	^ nil! !

!OSPipeTestCase methodsFor: 'private' stamp: 'dtl 6/4/2006 16:03'!
readFromAndClose: aPipe writingTo: aStream

	| s |
	[aPipe atEndOfFile] whileFalse:
		[s := aPipe next: 10000.
		aStream nextPutAll: s asString.
		(Delay forMilliseconds: 100) wait].
	(aPipe respondsTo: #reader) ifTrue: [aPipe reader close].
	^ aStream
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 4/12/2014 16:07'!
testBlocking

	"(OSPipeTestCase selector: #testBlocking) run"

	| p |
	p := OSPipe blockingPipe.
	(1 to: 10) do: [:i | p nextPutAll: 'this is line ', i printString; cr].
	p flush. "protect against buggy clib that may never flush output"
	self should: ['this is line 1*' match: (p next: 15)].
	self should: ['this is line 2*' match: (p next: 15)].
	self should: ['this is line 3*' match: (p next: 15)].
	self should: ['this is line 4*' match: (p next: 15)].
	self should: ['this is line 5*' match: (p next: 15)].
	self should: ['this is line 6*' match: (p next: 15)].
	self should: ['this is line 7*' match: (p next: 15)].
	self should: ['this is line 8*' match: (p next: 15)].
	self should: ['this is line 9*' match: (p next: 15)].
	p writer close.
	self shouldnt: p atEnd.
	self should: ['this is line 10*' match: (p next: 16)].
	self should: p atEnd.
	p close
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 4/2/2006 21:38'!
testBlockingBufferedPipe

	"(OSPipeTestCase selector: #testBlockingBufferedPipe) run"

	| p |
	p := ExternalPipe bufferedBlockingPipe.
	[(1 to: 10) do: [:i | p nextPutAll: 'this is line ', i printString; cr].
	(Delay forMilliseconds: 100) wait. "allow pipe reads to complete"
	self should: ['this is line 1*' match: (p next: 15)].
	self should: ['this is line 2*' match: (p next: 15)].
	self should: ['this is line 3*' match: (p next: 15)].
	self should: ['this is line 4*' match: (p next: 15)].
	self should: ['this is line 5*' match: (p next: 15)].
	self should: ['this is line 6*' match: (p next: 15)].
	self should: ['this is line 7*' match: (p next: 15)].
	self should: ['this is line 8*' match: (p next: 15)].
	self should: ['this is line 9*' match: (p next: 15)].
	p writer close.
	self shouldnt: p atEnd.
	self should: ['this is line 10*' match: (p next: 16)].
	self should: p atEnd]
		ensure: [p close]
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 4/12/2014 16:09'!
testBlockingPeek

	"(OSPipeTestCase selector: #testBlockingPeek) run"

	| p |
	p := OSPipe blockingPipe.
	p nextPutAll: 'abc'.
	p flush. "protect against buggy clib that may never flush output"
	self assert: (p peek == $a).
	self assert: (p next == $a).
	self assert: (p peek == $b).
	self assert: (p next == $b).
	self assert: (p peek == $c).
	self deny: p atEnd.
	self assert: (p next == $c).
	self deny: p atEnd.
	p closeWriter.
	self assert: p atEnd.
	p close
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 4/2/2006 21:39'!
testBlockingPeekBufferedPipe

	"(OSPipeTestCase selector: #testBlockingPeekBufferedPipe) run"

	| p |
	p := ExternalPipe bufferedBlockingPipe.
	[p nextPutAll: 'abc'.
	(Delay forMilliseconds: 100) wait. "allow pipe reads to complete"
	self assert: (p peek == $a).
	self assert: (p next == $a).
	self assert: (p peek == $b).
	self assert: (p next == $b).
	self assert: (p peek == $c).
	self deny: p atEnd.
	self assert: (p next == $c).
	self deny: p atEnd.
	p closeWriter.
	self assert: p atEnd]
		ensure: [p close]
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 5/17/2006 23:36'!
testBufferedUpToEnd

	"(OSPipeTestCase selector: #testBufferedUpToEnd) debug"

	| s p r1 r2 |
	s := 'some data to put through the pipe'.
	p := OSPipe new.
	self primTestEndOfFileFlag: p reader fileID. "check that OSPP is updated"
	self assert: p setBufferedReader.
	[p nextPutAll: s.
	p closeWriter.
	r1 := p upToEnd. "#atEnd does not detect EOF on a pipe"
	self assert: r1 isEmpty.
	r2 := p upToEndOfFile. "detects EOF correctly, gets remaining data"
	self assert: r2 = s]
		ensure: [p close]
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 4/24/2012 19:24'!
testIsAtEndOfFile

	"(OSPipeTestCase selector: #testIsAtEndOfFile) debug"

	| p |
	p := OSPipe new.
	self primTestEndOfFileFlag: p reader fileID. "check that OSPP is updated"
	self deny: (OSProcess accessor isAtEndOfFile: p reader fileID).
	self deny: p reader atEnd.
	self deny: p atEnd.
	self assert: (p reader basicNext == nil).
	self deny: (OSProcess accessor isAtEndOfFile: p reader fileID).
	self deny: p reader atEnd.
	self deny: p atEnd.
	p writer nextPut: $X.
	self assert: (p reader basicNext == $X).
	self deny: (OSProcess accessor isAtEndOfFile: p reader fileID).
	self deny: p reader atEnd.
	self deny: p atEnd.
	self assert: (p reader basicNext == nil).
	self deny: (OSProcess accessor isAtEndOfFile: p reader fileID).
	self deny: p reader atEnd.
	self deny: p atEnd.

	p writer close.
	self deny: (OSProcess accessor isAtEndOfFile: p reader fileID).
	self deny: p atEndOfFile. "no read yet, so flag has not been set"
	self deny: p reader atEnd.
	self assert: p atEnd. "writer closed and read gives nil"

	self assert: (p reader basicNext == nil). "read operation should set flag"
	self assert: (OSProcess accessor isAtEndOfFile: p reader fileID).
	self deny: p reader atEnd. "#atEnd does not detect EOF on a pipe"
	self assert: p reader atEndOfFile.
	self assert: p atEnd.
	p close
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 4/8/2007 10:15'!
testIsAtEndOfFile2

	"(OSPipeTestCase selector: #testIsAtEndOfFile2) debug"

	| p s |
	p := OSPipe new.
	[p nextPutAll: 'hello'.
	s := p next: 100.
	self assert: s = 'hello'.
	self deny: p atEndOfFile.
	p closeWriter.
	self deny: p atEndOfFile.
	s := p next: 100.
	self assert: s = ''.
	self assert: p atEndOfFile]
		ensure: [p close]
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 3/25/2006 13:44'!
testNonBlocking

	"(OSPipeTestCase selector: #testNonBlocking) run"

	| p ws str |
	p := OSPipe nonBlockingPipe.
	self writeStuffOnThenClose: p.
	ws := self readFromAndClose: p writingTo: (WriteStream on: String new).
	str := (ws contents last: 16).
	self should: ['this is line 10*' match: str].
	p close
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 4/2/2006 21:40'!
testNonBlockingBufferedPipe

	"(OSPipeTestCase selector: #testNonBlockingBufferedPipe) run"

	| p ws str |
	p := ExternalPipe bufferedNonBlockingPipe.
	[self writeStuffOnThenClose: p.
	self assert: (p peek == nil).
	(Delay forMilliseconds: 100) wait. "allow pipe reads to complete"
	ws := self readFromAndClose: p writingTo: (WriteStream on: String new).
	str := (ws contents last: 16).
	self should: ['this is line 10*' match: str]]
		ensure: [p close]
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 3/25/2006 13:50'!
testNonBlockingPeek

	"(OSPipeTestCase selector: #testNonBlockingPeek) run"

	| p |
	p := OSPipe nonBlockingPipe.
	p nextPutAll: 'abc'.
	self assert: (p peek == $a).
	self assert: (p next == $a).
	self assert: (p peek == $b).
	self assert: (p next == $b).
	self assert: (p peek == $c).
	self deny: p atEnd.
	self assert: (p next == $c).
	self deny: p atEnd.
	p closeWriter.
	self assert: p atEnd.
	p close
! !

!OSPipeTestCase methodsFor: 'testing' stamp: 'dtl 4/2/2006 21:41'!
testNonBlockingPeekBufferedPipe

	"(OSPipeTestCase selector: #testNonBlockingPeekBufferedPipe) run"

	| p |
	p := ExternalPipe bufferedNonBlockingPipe.
	[p nextPutAll: 'abc'.
	self assert: (p peek == nil).
	(Delay forMilliseconds: 100) wait. "allow pipe reads to complete"
	self assert: (p peek == $a).
	self assert: (p next == $a).
	self assert: (p peek == $b).
	self assert: (p next == $b).
	self assert: (p peek == $c).
	self deny: p atEnd.
	self assert: (p next == $c).
	self deny: p atEnd.
	p closeWriter.
	self assert: p atEnd]
		ensure: [p close]
! !

!OSPipeTestCase methodsFor: 'private' stamp: 'dtl 3/25/2006 18:29'!
writeStuffOnThenClose: aPipe

	^ [(1 to: 10) do:
		[:i |
		[aPipe nextPutAll: 'this is line ', i printString; cr]
			on: Error
			do: ["Ignore error. Test case will still fail, and throwing an error in this
				process would cause a debugger to be scheduled inconveniently."].
		(Delay forMilliseconds: 50) wait].
	(aPipe respondsTo: #writer)
		ifTrue: [[aPipe writer close]
					on: Error
					do: ["Ignore error to avoid a debugger"]]
		ifFalse: [[aPipe close]
					on: Error
					do: ["Ignore error to avoid a debugger"]]]
		forkAt: Processor userBackgroundPriority! !

TestCase subclass: #UnixProcessAccessorTestCase
	instanceVariableNames: 'accessor'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Tests'!

!UnixProcessAccessorTestCase commentStamp: '<historical>' prior: 0!
Unit tests for the UnixProcessAccessor.!

!UnixProcessAccessorTestCase methodsFor: 'testing - C string arrays' stamp: 'dtl 7/11/2005 19:00'!
primFixPointersInArrayOfStrings: cStringArray withOffsets: offsetArray count: count
	"This primitive call exists only for purposes of testing the
	fixPointersInArrayOfStrings:withOffsets:count: method. I believe it to be
	reasonably machine and compiler independent, but have no way of verifying
	this on a variety of machines, so I'll leave this test method here in case
	someone runs into problems on other hardware or compilers. -dtl"

	"| a |
	a := OSProcess thisOSProcess envAsFlatArrayAndOffsets: UnixProcess env.
	UnixProcessAccessorTestCase new
		primFixPointersInArrayOfStrings: (a at: 1)
		withOffsets: (a at: 2)
		count: (a at: 2) size"

	<primitive: 'primitiveFixPointersInArrayOfStrings' module: 'UnixOSProcessPlugin'>
	^ nil! !

!UnixProcessAccessorTestCase methodsFor: 'running' stamp: 'dtl 3/1/2002 18:02'!
runAll

	"UnixProcessAccessorTestCase new runAll"

	| result suite |
	suite := TestSuite new.
	suite addTest: (UnixProcessAccessorTestCase selector: #testSessionIdentifier).
	suite addTest: (UnixProcessAccessorTestCase selector: #testCanAccessSystem).
	suite addTest: (UnixProcessAccessorTestCase selector: #testMakePipeHandles).
	suite addTest: (UnixProcessAccessorTestCase selector: #testUnixFileNumber).
	suite addTest: (UnixProcessAccessorTestCase selector: #testFileProtectionMask).
	suite addTest: (UnixProcessAccessorTestCase selector: #testFileStat).
	suite addTest: (UnixProcessAccessorTestCase selector: #testIsExecutable).
	suite addTest: (UnixProcessAccessorTestCase selector: #testIsExecutableForUserInGroup).
	suite addTest: (UnixProcessAccessorTestCase selector: #testIsReadable).
	suite addTest: (UnixProcessAccessorTestCase selector: #testIsReadableForUserInGroup).
	suite addTest: (UnixProcessAccessorTestCase selector: #testIsWritable).
	suite addTest: (UnixProcessAccessorTestCase selector: #testIsWritableForUserInGroup).

	suite addTest: (UnixProcessAccessorTestCase selector: #runExternalProcessAccess).
	suite addTest: (UnixProcessAccessorTestCase selector: #runForkAndExec).

	result := suite run.
	self should: [result defects size == 0].
	^ result
! !

!UnixProcessAccessorTestCase methodsFor: 'running' stamp: 'dtl 8/3/2003 18:02'!
runExternalProcessAccess

	"UnixProcessAccessorTestCase new runExternalProcessAccess"

	| result suite |
	suite := TestSuite new.
	suite addTest: (UnixProcessAccessorTestCase selector: #testCanAccessSystem).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetCurrentWorkingDirectory).
	suite addTest: (UnixProcessAccessorTestCase selector: #testChDir).
	suite addTest: (UnixProcessAccessorTestCase selector: #testEnvironmentAt).
	suite addTest: (UnixProcessAccessorTestCase selector: #testEnvironmentAtPut1).
	suite addTest: (UnixProcessAccessorTestCase selector: #testEnvironmentAtPut2).
	suite addTest: (UnixProcessAccessorTestCase selector: #testEnvironmentAtPut).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimUnsetEnv).
	suite addTest: (UnixProcessAccessorTestCase selector: #testGetStdInHandle).
	suite addTest: (UnixProcessAccessorTestCase selector: #testGetStdOutHandle).
	suite addTest: (UnixProcessAccessorTestCase selector: #testGetStdErrHandle).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimArgumentAt).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimErrorMessageAt).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetPid).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetGid).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetEGid).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetPid).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetPPid).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetUid).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetSession).
	suite addTest: (UnixProcessAccessorTestCase selector: #testRealpath).

	result := suite run.
	self should: [result defects size == 0].
	^ result
! !

!UnixProcessAccessorTestCase methodsFor: 'running' stamp: 'dtl 10/6/2001 11:37'!
runForkAndExec
	"Most of this must be tested from class UnixProcess"

	"UnixProcessAccessorTestCase new runForkAndExec"

	| result suite |
	suite := TestSuite new.
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimGetChildExitStatus).
	suite addTest: (UnixProcessAccessorTestCase selector: #testPrimForkAndExec).
	result := suite run.
	self should: [result defects size == 0].
	^ result
! !

!UnixProcessAccessorTestCase methodsFor: 'running' stamp: 'dtl 10/6/2001 08:29'!
setUp

	accessor := ThisOSProcess accessor! !

!UnixProcessAccessorTestCase methodsFor: 'testing - C string arrays' stamp: 'dtl 9/10/2011 14:17'!
testArgsAsFlatArrayAndOffsets
	"Test conversion of an array of strings into the form used to create a C **char
	in UnixOSProcessPlugin."

	"(UnixProcessAccessorTestCase selector: #testArgsAsFlatArrayAndOffsets) debug"

	| a pointerSize lengthOfCStrings totalLength stringsAndOffsets rs s arrayWithMachineDependentPointers |
	pointerSize := OSProcess accessor sizeOfPointer.
	a := { 'one ' . 'two ' . 'three ' }.
	a do: [:e | e at: e size put: (Character value: 0)].
	stringsAndOffsets := OSProcess thisOSProcess argsAsFlatArrayAndOffsets: a.
	self assert: stringsAndOffsets size == 2.
	self assert: stringsAndOffsets second size == 3.
	lengthOfCStrings := (a collect: [:e | e size]) sum.
	totalLength := 4 * pointerSize + lengthOfCStrings.
	self assert: stringsAndOffsets first size == totalLength.
	stringsAndOffsets second do: [:offset |
		(stringsAndOffsets first at: offset) == (Character value: 0)].

	rs := ReadStream on: stringsAndOffsets first.
	rs next: 4 * pointerSize.
	s :=  rs next: 3.
	rs next.
	self assert: s = 'one'.
	s :=  rs next: 3.
	rs next.
	self assert: s = 'two'.
	s :=  rs next: 5.
	rs next.
	self assert: s = 'three'.

	rs := ReadStream on: stringsAndOffsets first.
	rs next: (stringsAndOffsets second at: 1).
	s :=  rs next: 3.
	self assert: s = 'one'.
	rs := ReadStream on: stringsAndOffsets first.
	rs next: (stringsAndOffsets second at: 2).
	s :=  rs next: 3.
	self assert: s = 'two'.
	rs := ReadStream on: stringsAndOffsets first.
	rs next: (stringsAndOffsets second at: 3).
	s :=  rs next: 5.
	self assert: s = 'three'.

	arrayWithMachineDependentPointers := self
		primFixPointersInArrayOfStrings: stringsAndOffsets first
		withOffsets: stringsAndOffsets second
		count: stringsAndOffsets second size.
	self assert: arrayWithMachineDependentPointers size == stringsAndOffsets first size.
	rs := ReadStream on: arrayWithMachineDependentPointers.
	s := rs next: pointerSize.
	self assert: s ~= (String new: pointerSize). "a C pointer"
	s := rs next: pointerSize.
	self assert: s ~= (String new: pointerSize). "a C pointer"
	s := rs next: pointerSize.
	self assert: s ~= (String new: pointerSize). "a C pointer"
	s := rs next: pointerSize.
	self assert: s = (String new: pointerSize). "a NULL pointer"
	self assert: rs upToEnd = (stringsAndOffsets first allButFirst: pointerSize * 4)
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 10/6/2001 08:36'!
testCanAccessSystem

	"(UnixProcessAccessorTestCase selector: #testCanAccessSystem) run"

	self assert: accessor canAccessSystem! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 6/10/2011 15:04'!
testChDir

	"(UnixProcessAccessorTestCase selector: #testChDir) run"

	| cwd new |
	cwd := accessor primGetCurrentWorkingDirectory.
	new := '/bin'. "nb Do not use /tmp because OS X implements it as a sym link to private/tmp"
	accessor chDir: new.
	self should: [new = accessor primGetCurrentWorkingDirectory].
	accessor chDir: cwd.
	self should: [cwd = accessor primGetCurrentWorkingDirectory].
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 1/23/2013 19:41'!
testDupTo
	"Perform a dup2() call to assign IO to a new file stream"

	"(UnixProcessAccessorTestCase selector: #testDupTo) run"

	| f1 f2 fd1 fd2 result contents1 contents2 |
	OSProcess deleteFileNamed: '/tmp/delete-1.me'.	
	OSProcess deleteFileNamed: '/tmp/delete-2.me'.	
	f1 := FileStream newFileNamed: '/tmp/delete-1.me'.
	f2 := FileStream newFileNamed: '/tmp/delete-2.me'.
	fd1 := OSProcess accessor primUnixFileNumber: f1 fileID.
	fd2 := OSProcess accessor primUnixFileNumber: f2 fileID.
	result := OSProcess accessor primDup: fd1 to: fd2.
	"f1 and f2 now both refer to the same IO stream"
	self assert: result = fd2.
	f2 nextPutAll: 'write something to the original f2 file stream that will appear in duped f1 instead'.
	f1 reset.
	contents1 := f1 upToEnd.
	self deny: contents1 isEmpty.
	self assert: ('write something*' match: contents1).
	f2 reset.
	contents2 := f2 upToEnd.
	self deny: contents2 isEmpty.
	self assert: ('write something*' match: contents2).
	f1 close.
	f2 close.
	"check the contents of the actual files"
	contents1 := (FileStream fileNamed: '/tmp/delete-1.me') contentsOfEntireFile.
	self assert: ('write something*' match: contents1).
	contents2 := (FileStream fileNamed: '/tmp/delete-2.me') contentsOfEntireFile.
	self assert: contents2 isEmpty.
	OSProcess deleteFileNamed: '/tmp/delete-1.me'.	
	OSProcess deleteFileNamed: '/tmp/delete-2.me'.	
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - C string arrays' stamp: 'dtl 9/10/2011 14:16'!
testEnvAsFlatArrayAndOffsets
	"Test conversion of a dictionary of key-value environment variables into the
	form used to create a C **char in UnixOSProcessPlugin."

	"(UnixProcessAccessorTestCase selector: #testEnvAsFlatArrayAndOffsets) debug"

	| pointerSize lengthOfCStrings totalLength stringsAndOffsets rs s arrayWithMachineDependentPointers d |
	pointerSize := OSProcess accessor sizeOfPointer.
	d := Dictionary new.
	d at: #ONE put: 'one'.
	d at: #TWO put: 'two'.
	d at: #THREE put: 'three'.
	stringsAndOffsets := OSProcess thisOSProcess envAsFlatArrayAndOffsets: d.
	self assert: stringsAndOffsets size == 2.
	self assert: stringsAndOffsets second size == 3.

	lengthOfCStrings := d keys
		inject: 0
		into: [:sum :e | sum + e size + '=' size + (d at: e) size + (String new: 1) size].

	totalLength := 4 * pointerSize + lengthOfCStrings.
	self assert: stringsAndOffsets first size == totalLength.
	stringsAndOffsets second do: [:offset |
		(stringsAndOffsets first at: offset) == (Character value: 0)].

	arrayWithMachineDependentPointers := self
		primFixPointersInArrayOfStrings: stringsAndOffsets first
		withOffsets: stringsAndOffsets second
		count: stringsAndOffsets second size.
	self assert: arrayWithMachineDependentPointers size == stringsAndOffsets first size.
	rs := ReadStream on: arrayWithMachineDependentPointers.
	s := rs next: pointerSize.
	self assert: s ~= (String new: pointerSize). "a C pointer"
	s := rs next: pointerSize.
	self assert: s ~= (String new: pointerSize). "a C pointer"
	s := rs next: pointerSize.
	self assert: s ~= (String new: pointerSize). "a C pointer"
	s := rs next: pointerSize.
	self assert: s = (String new: pointerSize). "a NULL pointer"
	self assert: rs upToEnd = (stringsAndOffsets first allButFirst: pointerSize * 4)
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 08:53'!
testEnvironmentAt

	"(UnixProcessAccessorTestCase selector: #testEnvironmentAt) run"

	| path |
	path := accessor environmentAt: 'PATH'.
	self should: [path notNil and: [path isEmpty not]]! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 2/24/2013 10:26'!
testEnvironmentAtPut

	"(UnixProcessAccessorTestCase selector: #testEnvironmentAtPut) run"

	| oldPath newPath resetPath s |
	oldPath := accessor environmentAt: 'PATH'.
	newPath := 'this is a test string'.
	accessor environmentAt: 'PATH' put: newPath.
	resetPath := accessor environmentAt: 'PATH'.
	self should: [newPath = resetPath].
	accessor environmentAt: 'PATH' put: oldPath.
	resetPath := accessor environmentAt: 'PATH'.
	self should: [oldPath = resetPath].
	s := WriteStream on: String new.
	10000 timesRepeat: [s nextPutAll: 'Do something to provoke garbage collection'].
	resetPath := accessor environmentAt: 'PATH'.
	self should: [oldPath = resetPath].


! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/7/2001 11:47'!
testEnvironmentAtPut1

	"(UnixProcessAccessorTestCase selector: #testEnvironmentAtPut1) run"

	| oldVal newVal resetVal |
	oldVal := accessor environmentAt: 'AAAA'.
	newVal := 'this is a test string'.
	accessor environmentAt: 'AAAA' put: newVal.
	resetVal := accessor environmentAt: 'AAAA'.
	self should: [resetVal notNil and: [resetVal isEmpty not]].
	self should: [newVal = resetVal].
	accessor environmentAt: 'AAAA' put: oldVal.
	self should: [oldVal = (accessor environmentAt: 'AAAA')]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/7/2001 12:14'!
testEnvironmentAtPut2
	"This looks for a bug in which the enviroment is successfully, but later gets
	invalidated by memory moves or reallocation."

	"(UnixProcessAccessorTestCase selector: #testEnvironmentAtPut2) run"

	| oldVal newVal resetVal count ws goodResults notFound totallyBogus |
	oldVal := accessor environmentAt: 'AAAA'.
	newVal := 'this is a test string'.
	accessor environmentAt: 'AAAA' put: newVal.
	resetVal := accessor environmentAt: 'AAAA'.
	self should: [resetVal notNil and: [resetVal isEmpty not]].
	self should: [newVal = resetVal].

	count := 50000.
	ws := WriteStream on: Array new.
	count timesRepeat: [ws nextPut: (accessor environmentAt: 'AAAA')].
	goodResults := ws contents select: [:e | e = newVal].
	notFound := ws contents select: [:e | e isNil].
	totallyBogus := ws contents select: [:e | e notNil and: [e ~= newVal]].

	"Uncomment the following to see when the problem is happening"
	"(Array with: goodResults with: notFound with: totallyBogus) inspect."

	self should: [goodResults size == count].
	self should: [notFound isEmpty].
	self should: [totallyBogus isEmpty].

	accessor environmentAt: 'AAAA' put: oldVal.
	self should: [oldVal = (accessor environmentAt: 'AAAA')]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 10/6/2001 11:59'!
testFileProtectionMask

	"(UnixProcessAccessorTestCase selector: #testFileProtectionMask) run"

	| mask |
	mask := accessor fileProtectionMask: '/bin/rm'.
	self assert: (mask isKindOf: Array).
	self assert: (mask size == 4)
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 10/6/2001 12:06'!
testFileStat

	"(UnixProcessAccessorTestCase selector: #testFileStat) run"

	| stat |
	stat := accessor fileStat: '/bin/rm'.
	self assert: (stat isKindOf: Array).
	self assert: (stat size == 3).
	self assert: ((stat at: 1) isKindOf: Integer).
	self assert: ((stat at: 2) isKindOf: Integer).
	self assert: ((stat at: 3) isKindOf: Array).
	self assert: ((stat at: 3) size == 4)
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:36'!
testGetStdErrHandle

	"(UnixProcessAccessorTestCase selector: #testGetStdErrHandle) run"

	| handle |
	handle := accessor primGetStdErrHandle.
	self assert: handle notNil
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:36'!
testGetStdInHandle

	"(UnixProcessAccessorTestCase selector: #testGetStdInHandle) run"

	| handle |
	handle := accessor primGetStdInHandle.
	self assert: handle notNil
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:36'!
testGetStdOutHandle

	"(UnixProcessAccessorTestCase selector: #testGetStdOutHandle) run"

	| handle |
	handle := accessor primGetStdOutHandle.
	self assert: handle notNil
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 10/6/2001 12:08'!
testIsExecutable

	"(UnixProcessAccessorTestCase selector: #testIsExecutable) run"

	self should: [accessor isExecutable: '/bin/sh'].
	self shouldnt: [accessor isExecutable: '/etc/hosts']
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 1/23/2013 21:35'!
testIsExecutableForUserInGroup

	"(UnixProcessAccessorTestCase selector: #testIsExecutableForUserInGroup) run"

	self shouldnt: [OSProcess waitForCommand: 'touch /tmp/delete.me']
		raise: Warning.
	OSProcess waitForCommand: 'chmod 550 /tmp/delete.me'.
	"explicitly set file group because OS X may have set it to something other than current gid"
	OSProcess waitForCommand: 'chgrp ', accessor primGetGid asString, ' /tmp/delete.me'.
	self should: [accessor isExecutable: '/bin/sh'
					forUser: accessor primGetUid
					inGroup: accessor primGetGid].
	self shouldnt: [accessor isExecutable: '/etc/hosts'
					forUser: accessor primGetUid
					inGroup: accessor primGetGid].
	self should: [accessor isExecutable: '/tmp/delete.me'
					forUser: accessor primGetUid
					inGroup: 99999].
	self should: [accessor isExecutable: '/tmp/delete.me'
					forUser: 99999
					inGroup: accessor primGetGid].
	self shouldnt: [accessor isExecutable: '/tmp/delete.me'
					forUser: 99999
					inGroup: 99999].
	OSProcess deleteFileNamed: '/tmp/delete.me'
! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 11:13'!
testIsLockableFile

	"(UnixProcessAccessorTestCase selector: #testIsLockableFile) debug"

	| fs result |
	fs := FileStream fileNamed: 'junkfile'.
	[fs nextPutAll: 'ABCDEFG'.

	result := OSProcess accessor isLockableFile: fs.
	self should: result.

	result := OSProcess accessor lockFile: fs.
	self should: (result isKindOf: OSFileLock).
	result := OSProcess accessor isLockableFile: fs.
	self should: result.

	OSProcess accessor unlockFile: fs.
	result := OSProcess accessor isLockableFile: fs.
	self should: result.

	fs close.
	result := OSProcess accessor isLockableFile: fs.
	self shouldnt: result.
	] ensure: [fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 10/20/2001 11:39'!
testIsReadable

	"(UnixProcessAccessorTestCase selector: #testIsReadable) run"

	self should: [accessor isReadable: '/bin/sh'].
	OSProcess waitForCommand: 'touch /tmp/delete.me'.
	OSProcess waitForCommand: 'chmod 440 /tmp/delete.me'.
	(Delay forMilliseconds: 200) wait.
	self should: [accessor isReadable: '/tmp/delete.me'].
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 1/23/2013 21:35'!
testIsReadableForUserInGroup

	"(UnixProcessAccessorTestCase selector: #testIsReadableForUserInGroup) run"

	self shouldnt: [OSProcess waitForCommand: 'touch /tmp/delete.me']
		raise: Warning.
	OSProcess waitForCommand: 'chmod 550 /tmp/delete.me'.
	"explicitly set file group because OS X may have set it to something other than current gid"
	OSProcess waitForCommand: 'chgrp ', accessor primGetGid asString, ' /tmp/delete.me'.
	self should: [accessor isReadable: '/bin/sh'
					forUser: accessor primGetUid
					inGroup: accessor primGetGid].
	self should: [accessor isReadable: '/tmp/delete.me'
					forUser: accessor primGetUid
					inGroup: 99999].
	self should: [accessor isReadable: '/tmp/delete.me'
					forUser: 99999
					inGroup: accessor primGetGid].
	self shouldnt: [accessor isReadable: '/tmp/delete.me'
					forUser: 99999
					inGroup: 99999].
	OSProcess deleteFileNamed: '/tmp/delete.me'
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 10/20/2001 11:38'!
testIsWritable

	"(UnixProcessAccessorTestCase selector: #testIsWritable) run"

	self shouldnt: [accessor isWritable: '/bin/sh'].
	OSProcess waitForCommand: 'touch /tmp/delete.me'.
	OSProcess waitForCommand: 'chmod 770 /tmp/delete.me'.
	(Delay forMilliseconds: 200) wait.
	self should: [accessor isWritable: '/tmp/delete.me'].
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 1/23/2013 21:35'!
testIsWritableForUserInGroup

	"(UnixProcessAccessorTestCase selector: #testIsWritableForUserInGroup) run"

	self shouldnt: [OSProcess waitForCommand: 'touch /tmp/delete.me']
		raise: Warning.
	OSProcess waitForCommand: 'chmod 770 /tmp/delete.me'.
	"explicitly set file group because OS X may have set it to something other than current gid"
	OSProcess waitForCommand: 'chgrp ', accessor primGetGid asString, ' /tmp/delete.me'.
	self shouldnt: [accessor isWritable: '/bin/sh'
					forUser: accessor primGetUid
					inGroup: accessor primGetGid].
	self should: [accessor isWritable: '/tmp/delete.me'
					forUser: accessor primGetUid
					inGroup: 99999].
	self should: [accessor isWritable: '/tmp/delete.me'
					forUser: 99999
					inGroup: accessor primGetGid].
	self shouldnt: [accessor isWritable: '/tmp/delete.me'
					forUser: 99999
					inGroup: 99999].
	OSProcess deleteFileNamed: '/tmp/delete.me'
! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 10:58'!
testLockFile

	"(UnixProcessAccessorTestCase selector: #testLockFile) debug"

	| fs result |
	fs := FileStream fileNamed: 'junkfile'.
	[fs nextPutAll: 'ABCDEFG'.
	result := OSProcess accessor lockFile: fs.
	self should: (result isKindOf: OSFileLock).
	fs close.
	result := OSProcess accessor lockFile: fs.
	self should: (result == nil).
	] ensure: [fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 10/6/2001 11:48'!
testMakePipeHandles

	"(UnixProcessAccessorTestCase selector: #testMakePipeHandles) run"

	| p |
	p := OSPipe new.
	self assert: p reader closed not.
	self assert: p writer closed not.
	p close! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 10:54'!
testOSFileLockLock

	"(UnixProcessAccessorTestCase selector: #testOSFileLockLock) debug"

	| fs fileLock result |
	fs := FileStream fileNamed: 'junkfile'.
	[fs nextPutAll: 'ABCDEFG'.
	fileLock := OSFileLock onFile: fs exclusive: true.
	result := fileLock lock.
	self should: (result == true).
	result := fileLock lock.
	self should: (result == true).
	fs close.
	result := fileLock lock.
	self should: (result == false).
	] ensure: [fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 10:44'!
testOSFileLockTest

	"(UnixProcessAccessorTestCase selector: #testOSFileLockTest) debug"

	| fs fileLock result |
	fs := FileStream fileNamed: 'junkfile'.
	[fs nextPutAll: 'ABCDEFG'.
	fileLock := OSFileLock onFile: fs exclusive: true.
	result := fileLock test.
	self should: (result == true).
	result := fileLock lock.
	self should: (result == true).
	result := fileLock test.
	self should: (result == true).
	result := fileLock unlock.
	self should: (result == true).
	result := fileLock test.
	self should: (result == true).
	fs close.
	result := fileLock test.
	self should: (result == false).
	] ensure: [fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 10:41'!
testOSFileLockUnlock

	"(UnixProcessAccessorTestCase selector: #testOSFileLockUnlock) debug"

	| fs fileLock result |
	fs := FileStream fileNamed: 'junkfile'.
	[fs nextPutAll: 'ABCDEFG'.
	fileLock := OSFileLock onFile: fs exclusive: true.
	result := fileLock unlock.
	self should: (result == true).
	result := fileLock lock.
	self should: (result == true).
	result := fileLock unlock.
	self should: (result == true).
	fs close.
	result := fileLock unlock.
	self should: (result == false).
	] ensure: [fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 10:54'!
testOSFileRegionLockLock

	"(UnixProcessAccessorTestCase selector: #testOSFileRegionLockLock) debug"

	| fs fileLock result |
	fs := FileStream fileNamed: 'junkfile'.
	[fs nextPutAll: 'ABCDEFG'.
	fileLock := OSFileRegionLock onFile: fs from: 2 to: 4 exclusive: true.
	result := fileLock lock.
	self should: (result == true).
	result := fileLock lock.
	self should: (result == true).
	fs close.
	result := fileLock lock.
	self should: (result == false).
	] ensure: [fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 10:46'!
testOSFileRegionLockTest

	"(UnixProcessAccessorTestCase selector: #testOSFileRegionLockTest) debug"

	| fs fileLock result |
	fs := FileStream fileNamed: 'junkfile'.
	[fs nextPutAll: 'ABCDEFG'.
	fileLock := OSFileRegionLock onFile: fs from: 2 to: 4 exclusive: true.
	result := fileLock test.
	self should: (result == true).
	result := fileLock lock.
	self should: (result == true).
	result := fileLock test.
	self should: (result == true).
	result := fileLock unlock.
	self should: (result == true).
	result := fileLock test.
	self should: (result == true).
	fs close.
	result := fileLock test.
	self should: (result == false).
	] ensure: [fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 10:46'!
testOSFileRegionLockUnlock

	"(UnixProcessAccessorTestCase selector: #testOSFileRegionLockUnlock) debug"

	| fs fileLock result |
	fs := FileStream fileNamed: 'junkfile'.
	[fs nextPutAll: 'ABCDEFG'.
	fileLock := OSFileRegionLock onFile: fs from: 2 to: 4 exclusive: true.
	result := fileLock unlock.
	self should: (result == true).
	result := fileLock lock.
	self should: (result == true).
	result := fileLock unlock.
	self should: (result == true).
	fs close.
	result := fileLock unlock.
	self should: (result == false).
	] ensure: [fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:42'!
testPrimArgumentAt

	"(UnixProcessAccessorTestCase selector: #testPrimArgumentAt) run"

	| progName |
	progName := accessor primArgumentAt: 1.
	self should: [progName notNil and: [progName isEmpty not]]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:45'!
testPrimErrorMessageAt

	"(UnixProcessAccessorTestCase selector: #testPrimErrorMessageAt) run"

	| msg |
	msg := accessor primErrorMessageAt: 1.
	self should: [msg notNil and: [msg isEmpty not]].
	msg := accessor primErrorMessageAt: 0.
	self should: [msg notNil and: [msg isEmpty not]]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - fork and exec' stamp: 'dtl 6/11/2011 12:25'!
testPrimForkAndExec
	"These methods can only be tested properly from UnixProcess, which knows how
	to set up the parameters on the stack."

	"(UnixProcessAccessorTestCase selector: #testPrimForkAndExec) run"

	| p |
	self shouldnt: [p := UnixProcess command: 'ls /bin']
		raise: Warning.
	(Delay forMilliseconds: 500) wait.
	self should: [p isComplete].
	self should: [p exitStatus == 0]

! !

!UnixProcessAccessorTestCase methodsFor: 'testing - fork and exec' stamp: 'dtl 10/6/2001 10:44'!
testPrimGetChildExitStatus
	"Cannot really test this here, because it needs to be wired into the interrupt
	handler. Just make sure it returns nil when there is nothing to do."

	"(UnixProcessAccessorTestCase selector: #testPrimGetChildExitStatus) run"

	| stat |
	stat := accessor primGetChildExitStatus: -1.
	self should: [stat isNil].
	stat := accessor primGetChildExitStatus: 1.
	self should: [stat isNil]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 08:45'!
testPrimGetCurrentWorkingDirectory

	"(UnixProcessAccessorTestCase selector: #testPrimGetCurrentWorkingDirectory) run"

	| cwd |
	cwd := accessor primGetCurrentWorkingDirectory.
	self should: [cwd notNil and: [cwd isEmpty not]]! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:49'!
testPrimGetEGid

	"(UnixProcessAccessorTestCase selector: #testPrimGetEGid) run"

	| id |
	id := accessor primGetEGid.
	self should: [id isKindOf: Integer]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:49'!
testPrimGetEUid

	"(UnixProcessAccessorTestCase selector: #testPrimGetEUid) run"

	| id |
	id := accessor primGetEUid.
	self should: [id isKindOf: Integer]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:50'!
testPrimGetGid

	"(UnixProcessAccessorTestCase selector: #testPrimGetGid) run"

	| id |
	id := accessor primGetGid.
	self should: [id isKindOf: Integer]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:50'!
testPrimGetPPid

	"(UnixProcessAccessorTestCase selector: #testPrimGetPPid) run"

	| id |
	id := accessor primGetPPid.
	self should: [id isKindOf: Integer]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:50'!
testPrimGetPid

	"(UnixProcessAccessorTestCase selector: #testPrimGetPid) run"

	| id |
	id := accessor primGetPid.
	self should: [id isKindOf: Integer]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 8/19/2005 06:32'!
testPrimGetSession
	"As of OSPP version 4 and greater, the session ID should always be obtained from
	the interpreter. Previously it was deduced by looking at the handle of an open file."

	"(UnixProcessAccessorTestCase selector: #testPrimGetSession) run"

	| id expectedSessionSize |
	expectedSessionSize := [OSProcess accessor sizeOfInt] on: Error do: [4].
	id := OSProcess accessor primGetSession.
	self assert: (id isNil or: [id isKindOf: ByteArray]).
	id := OSProcess accessor getSessionIdentifier.
	self assert: (id isKindOf: ByteArray).
	self assert: (id size / expectedSessionSize) == 1

! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 09:50'!
testPrimGetUid

	"(UnixProcessAccessorTestCase selector: #testPrimGetUid) run"

	| id |
	id := accessor primGetUid.
	self should: [id isKindOf: Integer]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/7/2001 14:07'!
testPrimUnsetEnv

	"(UnixProcessAccessorTestCase selector: #testPrimUnsetEnv) run"

	| oldVal newVal resetVal |
	oldVal := accessor environmentAt: 'AAAA'.
	newVal := 'this is a test string'.
	accessor environmentAt: 'AAAA' put: newVal.
	resetVal := accessor environmentAt: 'AAAA'.
	self should: [resetVal notNil and: [resetVal isEmpty not]].
	self should: [newVal = resetVal].

	accessor primUnsetEnv: 'AAAA', (Character value: 0) asString.
	resetVal := accessor environmentAt: 'AAAA'.
	self should: [resetVal isNil].

	accessor environmentAt: 'AAAA' put: oldVal.
	self should: [oldVal = (accessor environmentAt: 'AAAA')]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing - external process access' stamp: 'dtl 10/6/2001 10:03'!
testRealpath

	"(UnixProcessAccessorTestCase selector: #testRealpath) run"

	| p |
	p := accessor realpath: '/usr/bin'.
	self should: [p notNil and: [p isEmpty not]].
	p := accessor realpath: '/bogus/path/name'.
	self should: [p isNil]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 1/23/2013 19:42'!
testRedirectStdOutTo
	"Perform a dup2() call on the standard output descriptor to assign it
	to a new file stream"

	"(UnixProcessAccessorTestCase selector: #testRedirectStdOutTo) run"

	| out aFileStream contents outFd saveFd newFileNo saveFs |
	out := OSProcess thisOSProcess stdOut.
	"save file number to allow cleanup at end of test"
	OSProcess deleteFileNamed: '/tmp/delete.2.me'.	
	saveFs := FileStream newFileNamed: '/tmp/delete.2.me'.
	outFd := OSProcess accessor primUnixFileNumber: out fileID.
	saveFd := OSProcess accessor primUnixFileNumber: saveFs fileID.
	OSProcess accessor primDup: outFd to: saveFd.
	"Original stdout file descriptor is saved as safeFd, to be restored at end of test"

	OSProcess deleteFileNamed: '/tmp/delete.me'.	
	aFileStream := FileStream newFileNamed: '/tmp/delete.me'.
	"nb - #contentsOfEntireFile closes the file, do not use"
	aFileStream reset.
	self assert: aFileStream upToEnd isEmpty.
	"note - do not close aFileStream as this affects actual stdout"
	out nextPutAll: 'foo'.
	aFileStream reset.
	self assert: aFileStream upToEnd isEmpty.
	OSProcess thisOSProcess redirectStdOutTo: aFileStream.
	out nextPutAll: 'foo'.
	aFileStream reset.
	contents := aFileStream upToEnd.
	self deny: contents isEmpty.
	self assert: contents = 'foo'. "assume noone else is using stdout right now"
	"ensure the we can still find the real stdout"
	self deny: OSProcess thisOSProcess stdOut closed.
	OSProcess thisOSProcess stdOut nextPutAll: 'bar'.
	aFileStream reset.
	contents := aFileStream upToEnd.
	self deny: contents isEmpty.
	self assert: contents = 'foobar'.
	"Restore the original file number. The stdOut stream is cached in ThisOSProcess
	(and should probably be recalculated following a dup2(), so find the real handle
	with getStdOutHandle."
	newFileNo := OSProcess accessor primUnixFileNumber: OSProcess accessor getStdOutHandle.
	OSProcess accessor primDup: saveFd to: newFileNo.
	OSProcess thisOSProcess stdOut lf; nextPutAll: 'UnixProcessAccessorTestCase>>testRedirectStdOutTo - this should appear on original console stdout'; lf.
	saveFs close.
	aFileStream close.	
	OSProcess deleteFileNamed: '/tmp/delete.me'.	
	OSProcess deleteFileNamed: '/tmp/delete.2.me'.	

! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 10/1/2005 10:33'!
testSessionIdentifier
	"This is an indirect way to make sure that the primitive for obtaining the session ID does
	not get out of kilter with respect to the method for obtaining session ID from an existing
	open file. The dangerous failure mode is if the SQFile data stucture format changes and
	OSProcess does not get updated to reflect the change.

	As of approximately Squeak 3.8 and beyond, the session id has been moved to the
	first slot of the data structure. Deducing the session ID from an open file will not
	work for a Squeak VM beyond that point, and will not work for any 64 bit VM.
	OSPP version 4.0 or higher obtains session ID from the interpreter, so this is no
	longer an issue. Skip this test for OSPP 4.0 and higher."

	"(UnixProcessAccessorTestCase selector: #testSessionIdentifier) debug"

	| installedOSPPMajorVersion |
	self should: [accessor getSessionIdentifier isKindOf: ByteArray].
	installedOSPPMajorVersion := (OSProcess accessor osppModuleVersionString
		ifNil: [-1]) asInteger.
	(installedOSPPMajorVersion >= 4)
		ifTrue: ["ignore this test"]
		ifFalse: [self should:
			[accessor getSessionIdentifierFromSourcesFile = accessor getSessionIdentifier]]
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'eem 3/27/2014 11:32'!
testSignalNumbers
	"Signal numbers as reported by UnixOSProcessPlugin. Note that some signal definitions
	are architecture dependent, with specific number assignments differing between e.g. Intel
	and Sparc. Intel architecture is assumed for this test."

	"NOTE OSPP prior to version 4.4.12 had several incorrect signal number assignments.
	Check the version level of OSPP by evaluating this expression:

		OSProcess accessor primOSProcessPluginModuleVersionString"

	self assert: OSProcess accessor primSigHupNumber = 1.
	self assert: OSProcess accessor primSigIntNumber = 2.
	self assert: OSProcess accessor primSigKillNumber = 9.
	self assert: OSProcess accessor primSigPipeNumber = 13.
	self assert: OSProcess accessor primSigQuitNumber = 3.
	self assert: OSProcess accessor primSigTermNumber = 15.
	OSProcess platformName caseOf:
		{ ['unix']		->
			[self assert: OSProcess accessor primSigUsr1Number = 10.
			 self assert: OSProcess accessor primSigUsr2Number = 12.
			 self assert: OSProcess accessor primSigChldNumber = 17.].
		   ['Mac OS']	->
			[self assert: OSProcess accessor primSigUsr1Number = 30.
			 self assert: OSProcess accessor primSigUsr2Number = 31.
			 self assert: OSProcess accessor primSigChldNumber = 20.]}
	
	
! !

!UnixProcessAccessorTestCase methodsFor: 'testing' stamp: 'dtl 9/10/2011 09:26'!
testUnixFileNumber
	"For purposes of this test, assume that stdin and stderr have not been
	directed."

	"(UnixProcessAccessorTestCase selector: #testUnixFileNumber) run"

	| err errFileNo errHandle in inFileNo inHandle aFileStream fFileNo fHandle |
	err := OSProcess thisOSProcess stdErr.
	errHandle := err fileID.
	errFileNo := OSProcess accessor unixFileNumber: errHandle.
	self assert: errFileNo == 2.
	in := OSProcess thisOSProcess stdIn.
	inHandle := in fileID.
	inFileNo := OSProcess accessor unixFileNumber: inHandle.
	self assert: inFileNo == 0.
	aFileStream := FileStream fileNamed: '/tmp/delete.me'.
	fHandle := aFileStream fileID.
	fFileNo := OSProcess accessor unixFileNumber: fHandle.
	self assert: fFileNo notNil.
	self assert: fFileNo > 2.
	aFileStream close

! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 12:13'!
testUnlockFileUnixCompatible

	"(UnixProcessAccessorTestCase selector: #testUnlockFileUnixCompatible) debug"

	| mode fs result |
	mode := OSProcess accessor emulateWin32FileLocking. "remember setting"

	fs := FileStream fileNamed: 'junkfile'.
	[OSProcessAccessor emulateWin32FileLocking: false. "unix mode"
	fs nextPutAll: 'ABCDEFG'.
	result := OSProcess accessor unlockFile: fs.
	"On Unix, the fcntl call to unlock the file will succeed even if
	the file is not locked."
	self should: (result isKindOf: OSFileLock).

	result := OSProcess accessor lockFile: fs.
	self should: (result isKindOf: OSFileLock).
	result := OSProcess accessor unlockFile: fs.
	self should: (result isKindOf: OSFileLock).

	fs close.
	result := OSProcess accessor unlockFile: fs.
	self should: result == nil] ensure:
		[OSProcessAccessor emulateWin32FileLocking: mode.
		fs close]
! !

!UnixProcessAccessorTestCase methodsFor: 'file locking' stamp: 'dtl 5/7/2006 12:11'!
testUnlockFileWin32Compatible

	"(UnixProcessAccessorTestCase selector: #testUnlockFileWin32Compatible) debug"

	| mode fs result |
	mode := OSProcess accessor emulateWin32FileLocking. "remember setting"

	fs := FileStream fileNamed: 'junkfile'.
	[OSProcessAccessor emulateWin32FileLocking: true. "win32 mode"
	fs nextPutAll: 'ABCDEFG'.
	result := OSProcess accessor unlockFile: fs.
	self should: result == nil.

	result := OSProcess accessor lockFile: fs.
	self should: (result isKindOf: OSFileLock).
	result := OSProcess accessor unlockFile: fs.
	self should: (result isKindOf: OSFileLock).

	fs close.
	result := OSProcess accessor unlockFile: fs.
	self should: result == nil] ensure:
		[OSProcessAccessor emulateWin32FileLocking: mode.
		fs close]
! !

TestCase subclass: #UnixProcessFileLockTestCase
	instanceVariableNames: 'fileStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Tests'!

!UnixProcessFileLockTestCase commentStamp: 'dtl 3/7/2005 21:57' prior: 0!
This test case was provided by Julian Fitzell. It provides more file locking tests in addition to those in UnitProcessFileLockingTestCase.!

!UnixProcessFileLockTestCase methodsFor: 'asserting' stamp: 'jf 2/22/2004 19:07'!
assertLock: lock1 lock: lock2 conflicts: aBoolean
	self assert: (lock1 conflictsWith: lock2) = aBoolean.
	self assert: (lock2 conflictsWith: lock1) = aBoolean.! !

!UnixProcessFileLockTestCase methodsFor: 'data' stamp: 'jf 2/22/2004 19:11'!
entireExclusiveLock
	^ OSFileLock onFile: self fileStream exclusive: true! !

!UnixProcessFileLockTestCase methodsFor: 'data' stamp: 'jf 2/22/2004 19:11'!
entireSharedLock
	^ OSFileLock onFile: self fileStream exclusive: false! !

!UnixProcessFileLockTestCase methodsFor: 'accessing' stamp: 'jf 2/21/2004 18:14'!
fileStream
	^ fileStream! !

!UnixProcessFileLockTestCase methodsFor: 'data' stamp: 'jf 2/22/2004 19:22'!
regionExclusiveLock
	^ self regionLockFrom: 10 to: 20 exclusive: true! !

!UnixProcessFileLockTestCase methodsFor: 'data' stamp: 'jf 2/22/2004 19:22'!
regionLockFrom: from to: to exclusive: exclusive
	^ OSFileRegionLock onFile: self fileStream from: from to: to exclusive: exclusive! !

!UnixProcessFileLockTestCase methodsFor: 'data' stamp: 'jf 2/22/2004 19:22'!
regionSharedLock
	^ self regionLockFrom: 10 to: 20 exclusive: false! !

!UnixProcessFileLockTestCase methodsFor: 'running' stamp: 'jf 2/21/2004 18:12'!
setUp
	fileStream := FileStream fileNamed: 'junkfile'.! !

!UnixProcessFileLockTestCase methodsFor: 'running' stamp: 'dtl 1/23/2013 19:42'!
tearDown
	fileStream close.
	OSProcess deleteFileNamed: fileStream fullName! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:27'!
testDistinctExclusiveRegionWithExclusiveRegion
	self
		assertLock: (self regionLockFrom: 10 to: 20 exclusive: true)
		lock: (self regionLockFrom: 30 to: 40 exclusive: true)
		conflicts: false! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:27'!
testDistinctExclusiveRegionWithSharedRegion
	self
		assertLock: (self regionLockFrom: 10 to: 20 exclusive: true)
		lock: (self regionLockFrom: 30 to: 40 exclusive: false)
		conflicts: false! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:27'!
testDistinctSharedRegionWithExclusiveRegion
	self
		assertLock: (self regionLockFrom: 10 to: 20 exclusive: false)
		lock: (self regionLockFrom: 30 to: 40 exclusive: true)
		conflicts: false! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:26'!
testDistinctSharedRegionWithSharedRegion
	self
		assertLock: (self regionLockFrom: 10 to: 20 exclusive: false)
		lock: (self regionLockFrom: 30 to: 40 exclusive: false)
		conflicts: false! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:12'!
testEntireExclusiveWithEntireExclusive
	self
		assertLock: self entireExclusiveLock
		lock: self entireExclusiveLock
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:16'!
testEntireExclusiveWithExclusiveRegion
	self
		assertLock: self entireExclusiveLock
		lock: self regionExclusiveLock
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:17'!
testEntireExclusiveWithSharedRegion
	self
		assertLock: self entireExclusiveLock
		lock: self regionSharedLock
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:12'!
testEntireSharedWithEntireExclusive
	self
		assertLock: self entireSharedLock
		lock: self entireExclusiveLock
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:12'!
testEntireSharedWithEntireShared
	self
		assertLock: self entireSharedLock
		lock: self entireSharedLock
		conflicts: false! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:14'!
testEntireSharedWithExclusiveRegion
	self
		assertLock: self entireSharedLock
		lock: self regionExclusiveLock
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:15'!
testEntireSharedWithSharedRegion
	self
		assertLock: self entireSharedLock
		lock: self regionSharedLock
		conflicts: false! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:18'!
testExclusiveRegionWithExclusiveRegion
	self
		assertLock: self regionExclusiveLock
		lock: self regionExclusiveLock
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:25'!
testIdenticalExclusiveRegionWithExclusiveRegion
	self
		assertLock: self regionExclusiveLock
		lock: self regionExclusiveLock
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:24'!
testIdenticalSharedRegionWithExclusiveRegion
	self
		assertLock: self regionSharedLock
		lock: self regionExclusiveLock
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:24'!
testIdenticalSharedRegionWithSharedRegion
	self
		assertLock: self regionSharedLock
		lock: self regionSharedLock
		conflicts: false! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:25'!
testOverlappingExclusiveRegionWithExclusiveRegion
	self
		assertLock: (self regionLockFrom: 10 to: 20 exclusive: true)
		lock: (self regionLockFrom: 20 to: 30 exclusive: true)
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:26'!
testOverlappingExclusiveRegionWithSharedRegion
	self
		assertLock: (self regionLockFrom: 10 to: 20 exclusive: true)
		lock: (self regionLockFrom: 20 to: 30 exclusive: false)
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:26'!
testOverlappingSharedRegionWithExclusiveRegion
	self
		assertLock: (self regionLockFrom: 10 to: 20 exclusive: false)
		lock: (self regionLockFrom: 20 to: 30 exclusive: true)
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:26'!
testOverlappingSharedRegionWithSharedRegion
	self
		assertLock: (self regionLockFrom: 10 to: 20 exclusive: false)
		lock: (self regionLockFrom: 20 to: 30 exclusive: false)
		conflicts: false! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:21'!
testSupersetExclusiveRegionWithExclusiveRegion
	self
		assertLock: (self regionLockFrom: 10 to: 40 exclusive: true)
		lock: (self regionLockFrom: 20 to: 30 exclusive: true)
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:23'!
testSupersetExclusiveRegionWithSharedRegion
	self
		assertLock: (self regionLockFrom: 10 to: 40 exclusive: true)
		lock: (self regionLockFrom: 20 to: 30 exclusive: false)
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:23'!
testSupersetSharedRegionWithExclusiveRegion
	self
		assertLock: (self regionLockFrom: 10 to: 40 exclusive: false)
		lock: (self regionLockFrom: 20 to: 30 exclusive: true)
		conflicts: true! !

!UnixProcessFileLockTestCase methodsFor: 'testing' stamp: 'jf 2/22/2004 19:24'!
testSupersetSharedRegionWithSharedRegion
	self
		assertLock: (self regionLockFrom: 10 to: 40 exclusive: false)
		lock: (self regionLockFrom: 20 to: 30 exclusive: false)
		conflicts: false! !

TestCase subclass: #UnixProcessTestCase
	instanceVariableNames: 'thisOSProcess'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Tests'!

!UnixProcessTestCase commentStamp: '<historical>' prior: 0!
Unit tests for the Unix portion of OSProcess.!

!UnixProcessTestCase class methodsFor: 'failures' stamp: 'eem 3/27/2014 11:20'!
expectedFailures
	^UnixOSProcessAccessor basicNew safeToForkSqueak
		ifTrue: [#()]
		ifFalse: [#(	testClassForkHeadlessSqueakAndDo
					testClassForkHeadlessSqueakAndDoThenQuit
					testClassForkSqueak
					testClassForkSqueakAndDo
					testClassForkSqueakAndDoThenQuit
					testForkHeadlessSqueakAndDo
					testForkHeadlessSqueakAndDoThenQuit
					testForkSqueak
					testForkSqueakAndDo
					testForkSqueakAndDoThenQuit
					testHeadlessChild
					testSpawnTenHeadlessChildren)]! !

!UnixProcessTestCase methodsFor: 'failures' stamp: 'eem 3/27/2014 11:29'!
expectedFailures
	^UnixOSProcessAccessor basicNew safeToForkSqueak
		ifTrue: [#()]
		ifFalse: [#(	testClassForkHeadlessSqueakAndDo
					testClassForkHeadlessSqueakAndDoThenQuit
					testClassForkSqueak
					testClassForkSqueakAndDo
					testClassForkSqueakAndDoThenQuit
					testEightLeafSqueakTree
					testForkHeadlessSqueakAndDo
					testForkHeadlessSqueakAndDoThenQuit
					testForkSqueak
					testForkSqueakAndDo
					testForkSqueakAndDoThenQuit
					testHeadlessChild
					testSpawnTenHeadlessChildren)]! !

!UnixProcessTestCase methodsFor: 'private' stamp: 'dtl 1/23/2013 21:30'!
numberOfOpenFiles
	"Answer the number of files currently open for this OS process. This works
	only on a system with a /proc filesystem and file descriptors located in a
	directory called /proc/<pid>/fd. On other systems, just answer 0."

	"UnixProcessTestCase new numberOfOpenFiles"

	| path |
	path := '/proc/' , OSProcess thisOSProcess pid printString, '/fd'.
	^ (OSProcess directoryExists: path)
		ifTrue: [(OSProcess directoryEntryNames: path) size]
		ifFalse: [0]
! !

!UnixProcessTestCase methodsFor: 'running' stamp: 'dtl 3/10/2002 10:52'!
runAll
	"If you get intermittent failures, try doing a garbage collect. Some of these
	tests can fail intermittently on the open file handle count checks"

	"UnixProcessTestCase new runAll"

	| result suite |
	Smalltalk garbageCollect.
	suite := TestSuite new.
	suite addTest: (UnixProcessTestCase selector: #testClassForkSqueak).
	suite addTest: (UnixProcessTestCase selector: #testClassForkSqueakAndDo).
	suite addTest: (UnixProcessTestCase selector: #testClassForkSqueakAndDoThenQuit).
	suite addTest: (UnixProcessTestCase selector: #testClassForkHeadlessSqueakAndDo).
	suite addTest: (UnixProcessTestCase selector: #testClassForkHeadlessSqueakAndDoThenQuit).
	suite addTest: (UnixProcessTestCase selector: #testForkSqueak).
	suite addTest: (UnixProcessTestCase selector: #testForkSqueakAndDo).
	suite addTest: (UnixProcessTestCase selector: #testForkSqueakAndDoThenQuit).
	suite addTest: (UnixProcessTestCase selector: #testForkHeadlessSqueakAndDo).
	suite addTest: (UnixProcessTestCase selector: #testForkHeadlessSqueakAndDoThenQuit).
	suite addTest: (UnixProcessTestCase selector: #testHeadlessChild).
	suite addTest: (UnixProcessTestCase selector: #testSpawnTenHeadlessChildren).
	suite addTest: (UnixProcessTestCase selector: #testEightLeafSqueakTree).
	suite addTest: (UnixProcessTestCase selector: #testCatAFile).
	suite addTest: (UnixProcessTestCase selector: #testCatFromFileToFiles).
	suite addTest: (UnixProcessTestCase selector: #testRunCommand).
	suite addTest: (UnixProcessTestCase selector: #testPipe).
	suite addTest: (UnixProcessTestCase selector: #testPipeLine).

	result := suite run.
	self should: [result defects size == 0].
	^ result
! !

!UnixProcessTestCase methodsFor: 'running' stamp: 'dtl 7/30/2010 16:47'!
setUp

	(self respondsTo: #timeout: ) "Recent Squeak images with test case timeout"
		ifTrue: [self perform: #timeout: with: 60].
	thisOSProcess := OSProcess thisOSProcess
! !

!UnixProcessTestCase methodsFor: 'testing - class examples' stamp: 'dtl 6/11/2011 12:26'!
testCatAFile

	"(UnixProcessTestCase selector: #testCatAFile) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt: [p := UnixProcess catAFile]
		raise: Warning.
	self assert: p notNil.
	self should: [p isRunning].
	[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
	self should: [p isComplete].
	self should: [p exitStatus == 0].
	self should: [p initialStdIn closed]. "Used a file for input, should be closed"
	self shouldnt: [p initialStdOut closed]. "Shared stdout with the VM, should be open"
	self shouldnt: [p initialStdErr closed]. "Shared stderr with the VM, should be open"
	self assert: self numberOfOpenFiles == openFileCount
! !

!UnixProcessTestCase methodsFor: 'testing - class examples' stamp: 'dtl 1/23/2013 21:36'!
testCatFromFileToFiles

	"(UnixProcessTestCase selector: #testCatFromFileToFiles) run"

	| p f openFileCount |
	openFileCount := self numberOfOpenFiles.
	OSProcess deleteFileNamed: '/tmp/deleteMe.out'.
	OSProcess deleteFileNamed: '/tmp/deleteMe.err'.
	self shouldnt: [p := UnixProcess catFromFileToFiles]
		raise: Warning.
	self assert: p notNil.
	self should: [p isRunning].
	[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
	self should: [p isComplete].
	self should: [p exitStatus == 0].
	self should: [p initialStdIn closed].
	self should: [p initialStdOut closed].
	self should: [p initialStdErr closed].
	f := FileStream oldFileNamed: '/tmp/deleteMe.out'.
	self shouldnt: [f upToEnd isEmpty].
	f close.
	f := FileStream oldFileNamed: '/tmp/deleteMe.err'.
	self should: [f upToEnd isEmpty].
	f close.
	self assert: self numberOfOpenFiles == openFileCount
! !

!UnixProcessTestCase methodsFor: 'testing - class side methods' stamp: 'dtl 6/29/2005 14:24'!
testClassForkHeadlessSqueakAndDo

	"(UnixProcessTestCase selector: #testClassForkHeadlessSqueakAndDo) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt:
		[p := UnixProcess forkHeadlessSqueakAndDo:
			[(Delay forMilliseconds: 100) wait.
			OSProcess snapshot: false andQuit: true]]
		raise: Warning. "Catch warning if fork fails"
	self assert: p notNil.
	(p == thisOSProcess)
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - class side methods' stamp: 'dtl 6/29/2005 14:26'!
testClassForkHeadlessSqueakAndDoThenQuit

	"(UnixProcessTestCase selector: #testClassForkHeadlessSqueakAndDoThenQuit) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt:
		[p := UnixProcess forkHeadlessSqueakAndDoThenQuit:
			[(Delay forMilliseconds: 100) wait]]
		raise: Warning. "Catch warning if fork fails"
	self assert: p notNil.
	(p == thisOSProcess)
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - class side methods' stamp: 'dtl 6/10/2011 15:38'!
testClassForkSqueak

	"(UnixProcessTestCase selector: #testClassForkSqueak) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt: [p := UnixProcess forkSqueak]
		raise: Warning.
	self assert: p notNil.
	(p == thisOSProcess)
		ifTrue:
			["Child Squeak"
			p inspect.
			(Delay forMilliseconds: 100) wait.
			OSProcess snapshot: false andQuit: true]
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - class side methods' stamp: 'dtl 6/10/2011 15:43'!
testClassForkSqueakAndDo

	"(UnixProcessTestCase selector: #testClassForkSqueakAndDo) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt:
			[p := UnixProcess forkSqueakAndDo:
					[(Delay forMilliseconds: 100) wait.
					OSProcess snapshot: false andQuit: true]]
		raise: Warning.
	self assert: p notNil.
	(p == thisOSProcess)
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - class side methods' stamp: 'dtl 6/10/2011 15:43'!
testClassForkSqueakAndDoThenQuit

	"(UnixProcessTestCase selector: #testClassForkSqueakAndDoThenQuit) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt:
			[p := UnixProcess forkSqueakAndDoThenQuit:
					[(Delay forMilliseconds: 100) wait]]
		raise: Warning.
	self assert: p notNil.
	(p == thisOSProcess)
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - class examples' stamp: 'dtl 6/10/2011 15:55'!
testEightLeafSqueakTree

	"(UnixProcessTestCase selector: #testEightLeafSqueakTree) run"

	| a openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt: [a := UnixProcess eightLeafSqueakTree]
		raise: Warning.
	self assert: (a isKindOf: Array).
	self assert: a size == 3.
	(a includes: 0)
		ifTrue:
			[Smalltalk quitPrimitive].
	self assert: self numberOfOpenFiles == openFileCount
! !

!UnixProcessTestCase methodsFor: 'testing - child process creation' stamp: 'dtl 6/10/2011 15:47'!
testForkHeadlessSqueakAndDo

	"(UnixProcessTestCase selector: #testForkHeadlessSqueakAndDo) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt:
			[p := thisOSProcess forkHeadlessSqueakAndDo:
					[(Delay forMilliseconds: 100) wait.
					OSProcess snapshot: false andQuit: true]]
		raise: Warning. "Catch warning if fork fails"
	self assert: p notNil.
	(p == thisOSProcess)
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - child process creation' stamp: 'dtl 6/10/2011 15:47'!
testForkHeadlessSqueakAndDoThenQuit

	"(UnixProcessTestCase selector: #testForkHeadlessSqueakAndDoThenQuit) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt:
			[p := thisOSProcess forkHeadlessSqueakAndDoThenQuit:
					[(Delay forMilliseconds: 100) wait]]
		raise: Warning. "Catch warning if fork fails"
	self assert: p notNil.
	(p == thisOSProcess)
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - child process creation' stamp: 'dtl 6/29/2005 14:30'!
testForkSqueak

	"(UnixProcessTestCase selector: #testForkSqueak) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt: [p := thisOSProcess forkSqueak]
		raise: Warning. "Catch warning if fork fails"
	self assert: p notNil.
	(p == thisOSProcess)
		ifTrue:
			["Child Squeak"
			p inspect.
			(Delay forMilliseconds: 100) wait.
			OSProcess snapshot: false andQuit: true]
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - child process creation' stamp: 'dtl 6/10/2011 15:45'!
testForkSqueakAndDo

	"(UnixProcessTestCase selector: #testForkSqueakAndDo) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt:
			[p := thisOSProcess forkSqueakAndDo:
					[(Delay forMilliseconds: 100) wait.
					OSProcess snapshot: false andQuit: true]]
		raise: Warning.
	self assert: p notNil.
	(p == thisOSProcess)
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - child process creation' stamp: 'dtl 6/10/2011 15:46'!
testForkSqueakAndDoThenQuit

	"(UnixProcessTestCase selector: #testForkSqueakAndDoThenQuit) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt:
			[p := thisOSProcess forkSqueakAndDoThenQuit:
					[(Delay forMilliseconds: 100) wait]]
		raise: Warning.
	self assert: p notNil.
	(p == thisOSProcess)
		ifFalse:
			["Parent Squeak process"
			self should: [p isRunning].
			[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
			self should: [p isComplete].
			self should: [p exitStatus == 0].
			self assert: self numberOfOpenFiles == openFileCount]
! !

!UnixProcessTestCase methodsFor: 'testing - class examples' stamp: 'dtl 6/29/2005 14:32'!
testHeadlessChild

	"(UnixProcessTestCase selector: #testHeadlessChild) run"

	| p openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt: [p := UnixProcess headlessChild]
		raise: Warning. "Catch warning if fork fails"
	[p isComplete] whileFalse: [(Delay forMilliseconds: 100) wait].
	self assert: p isComplete.
	self assert: p exitStatus == 0.
	self assert: self numberOfOpenFiles == openFileCount
! !

!UnixProcessTestCase methodsFor: 'testing - class examples' stamp: 'dtl 10/12/2001 08:32'!
testPipe

	"(UnixProcessTestCase selector: #testPipe) run"

	| openFileCount |
	openFileCount := self numberOfOpenFiles.
	self should: [UnixProcess testPipe = 'this is some text to write into the pipe'].
	self assert: self numberOfOpenFiles == openFileCount
! !

!UnixProcessTestCase methodsFor: 'testing - class examples' stamp: 'dtl 10/12/2001 08:32'!
testPipeLine

	"(UnixProcessTestCase selector: #testPipeLine) run"

	| openFileCount |
	openFileCount := self numberOfOpenFiles.
	self should: ['This is the text to write*' match: UnixProcess testPipeLine].
	self assert: self numberOfOpenFiles == openFileCount
! !

!UnixProcessTestCase methodsFor: 'testing - class examples' stamp: 'dtl 1/23/2013 21:36'!
testRunCommand

	"(UnixProcessTestCase selector: #testRunCommand) run"

	| p f openFileCount |
	openFileCount := self numberOfOpenFiles.
	OSProcess deleteFileNamed: '/tmp/deleteMe.out'.
	OSProcess deleteFileNamed: '/tmp/deleteMe.err'.
	self shouldnt: [p := UnixProcess testRunCommand]
		raise: Warning.
	self assert: p notNil.
	self should: [p isRunning].
	[p isRunning] whileTrue: [(Delay forMilliseconds: 100) wait].
	self should: [p isComplete].
	self should: [p exitStatus > 0].	"Forced an error exit status"
	"stdin was shared with the Squeak VM, so it should not have been closed."
	self shouldnt: [p initialStdIn closed].
	"but the output and error streams should have been closed."
	self should: [p initialStdOut closed].
	self should: [p initialStdErr closed].
	f := FileStream oldFileNamed: '/tmp/deleteMe.out'.
	self shouldnt: [f upToEnd isEmpty].
	f close.
	f := FileStream oldFileNamed: '/tmp/deleteMe.err'.
	self shouldnt: [f upToEnd isEmpty].
	f close.
	self assert: self numberOfOpenFiles == openFileCount


! !

!UnixProcessTestCase methodsFor: 'testing - class examples' stamp: 'dtl 6/29/2005 14:33'!
testSpawnTenHeadlessChildren

	"(UnixProcessTestCase selector: #testSpawnTenHeadlessChildren) run"

	| a openFileCount |
	openFileCount := self numberOfOpenFiles.
	self shouldnt: [a := UnixProcess spawnTenHeadlessChildren]
		raise: Warning. "Catch warning if fork fails"
	[(a detect: [:p | p isComplete not] ifNone: []) notNil]
		whileTrue: [(Delay forMilliseconds: 100) wait].
	self should: [(a select: [:p | p isComplete not]) isEmpty].
	self assert: self numberOfOpenFiles == openFileCount
! !

StandardFileStream subclass: #AttachableFileStream
	instanceVariableNames: 'autoClose'
	classVariableNames: 'UseIOHandle'
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!AttachableFileStream commentStamp: '<historical>' prior: 0!
I am a stream on an input or output channel provided by the underlying operating system. I behave like an ordinary file stream, except that I can attach myself to an input or output stream which has already been opened by the underlying operating system.!

AttachableFileStream subclass: #AsyncFileReadStream
	instanceVariableNames: 'eventHandler'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!AsyncFileReadStream commentStamp: 'dtl 7/9/2003 21:04' prior: 0!
AsyncFileReadStream implements event-driven read behavior on a file stream. Whenever data is available, a #changed event is generated. An AsyncFileReadStream expects to have a client object respond immediately to the change notification by reading the available data, otherwise a possibly endless stream of change notifications will be generated.

AsyncFileReadStream requires aio support in the AioPlugin module.!

!AsyncFileReadStream class methodsFor: 'examples' stamp: 'dtl 7/9/2003 20:16'!
stdIn

	"self stdIn"

	^ super stdIn initialize
! !

!AsyncFileReadStream methodsFor: 'finalization' stamp: 'dtl 6/17/2006 08:08'!
actAsExecutor

	super actAsExecutor.
	eventHandler := nil! !

!AsyncFileReadStream methodsFor: 'converting' stamp: 'dtl 7/8/2003 20:53'!
asAsyncFileReadStream

	^ self
! !

!AsyncFileReadStream methodsFor: 'converting' stamp: 'dtl 7/8/2003 19:40'!
asAttachableFileStream
	"Answer a replacement for this object, with no asynchronous event handling.
	Do not close the ioHandle when this object is finalized."

	self keepOpen.
	^ AttachableFileStream
		name: self name
		attachTo: self ioHandle
		writable: self isReadOnly not
! !

!AsyncFileReadStream methodsFor: 'converting' stamp: 'dtl 7/8/2003 21:01'!
asBufferedAsyncFileReadStream
	"Answer a replacement for this object, with buffered output."

	^ BufferedAsyncFileReadStream
		name: self name
		attachTo: self ioHandle
		writable: self isReadOnly not
! !

!AsyncFileReadStream methodsFor: 'initialize-release' stamp: 'dtl 7/8/2003 21:48'!
close

	self disableEventHandling.
	^ super close! !

!AsyncFileReadStream methodsFor: 'read event handling' stamp: 'dtl 9/3/2010 16:50'!
disableEventHandling

	eventHandler
		ifNotNil: [eventHandler removeDependent: self; close].
	self eventHandler: nil! !

!AsyncFileReadStream methodsFor: 'read event handling' stamp: 'dtl 11/25/2006 10:56'!
enableEventHandling

	self eventHandler: (AioEventHandler aioPluginPresent
		ifTrue: [AioEventHandler
				forFileStream: self
				exceptions: true
				readEvents: true
				writeEvents: false]
		ifFalse: [PseudoAioEventHandler new]).
	self eventHandler addDependent: self! !

!AsyncFileReadStream methodsFor: 'accessing' stamp: 'dtl 7/8/2003 19:26'!
eventHandler
	"The aio event handler. Provides notification whenever external data is available."

	^ eventHandler! !

!AsyncFileReadStream methodsFor: 'accessing' stamp: 'dtl 7/8/2003 19:26'!
eventHandler: anAioEventHandler
	"The aio event handler. Provides notification whenever external data is available."

	eventHandler := anAioEventHandler! !

!AsyncFileReadStream methodsFor: 'finalization' stamp: 'dtl 9/3/2010 16:57'!
finalize
	"Shut the the event handler before closing the file, otherwise aio
	functions may reference a closed file handle."

	eventHandler ifNotNil: [eventHandler finalize].
	super finalize
! !

!AsyncFileReadStream methodsFor: 'read event handling' stamp: 'dtl 7/8/2003 19:29'!
hasValidHandler

	^ self eventHandler notNil and: [eventHandler hasValidHandler]
! !

!AsyncFileReadStream methodsFor: 'initialize-release' stamp: 'dtl 3/15/2006 07:45'!
initialize

	super initialize.
	self readOnly.
	OSProcess accessor setNonBlocking: self ioHandle.
	self enableEventHandling
! !

!AsyncFileReadStream methodsFor: 'initialize-release' stamp: 'dtl 7/8/2003 21:48'!
open

	^ super open initialize
! !

!AsyncFileReadStream methodsFor: 'read, write, position' stamp: 'dtl 9/2/2006 09:58'!
upToEndOfFile
	"Answer a subcollection from the current access position through
	the last element of the receiver.  Use #atEndOfFile to determine end
	of file status with feof(), required for reliable end of file test on
	OS pipes. Close the file when end of file is detected."

	| b |
	b := super upToEndOfFile.
	self isReadOnly
		ifTrue: [self close].
	^ b! !

!AsyncFileReadStream methodsFor: 'updating' stamp: 'dtl 7/8/2003 21:22'!
update: aParameter

	(aParameter == self eventHandler)
		ifTrue: [self changed]
! !

AsyncFileReadStream subclass: #BufferedAsyncFileReadStream
	instanceVariableNames: 'nonBlockingMode readBuffer readSyncSemaphore dataAvailableSemaphore'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!BufferedAsyncFileReadStream commentStamp: 'dtl 3/7/2006 06:55' prior: 0!
BufferedAsyncFileReadStream adds output buffering behavior to an event driven file stream, permitting blocking reads without risk of blocking the Squeak VM. This is useful for OS pipes, for which Squeak may wish to read and write the pipe without concern for VM deadlocks.

A BufferedAsyncFileReadStream may be set for either blocking or nonblocking reads. When in blocking mode, a Smalltalk Process that requests a read will be blocked until data is available, but the VM will not be blocked and other Smalltalk Processes can proceed normally.

Whenever data becomes available, a dataAvailableSemaphore is signalled and a #changed event is generated.!

!BufferedAsyncFileReadStream class methodsFor: 'examples' stamp: 'dtl 7/9/2003 20:24'!
stdIn

	"self stdIn"

	^ super stdIn
! !

!BufferedAsyncFileReadStream methodsFor: 'finalization' stamp: 'dtl 6/17/2006 08:10'!
actAsExecutor

	super actAsExecutor.
	nonBlockingMode := nil.
	readBuffer := nil.
	readSyncSemaphore := nil.
	dataAvailableSemaphore := nil
! !

!BufferedAsyncFileReadStream methodsFor: 'read ahead buffer' stamp: 'dtl 10/1/2006 17:12'!
appendAllToBuffer: chars
	"Append all chars to readBuffer, then signal dataAvailableSemaphore to inform
	any blocked reader that the read can proceed. Also generate a #changed event
	to inform any interested objects that new data has become available."

	| pos |
	self readSyncSemaphore critical:
		[(self readBuffer position > self maxReadBufferSize) ifTrue:
			["Read buffer is getting too large. Replace it."
			self readBuffer: (ReadWriteStream on: readBuffer upToEnd)].
		pos := readBuffer position.
		readBuffer setToEnd.
		readBuffer nextPutAll: chars.
		readBuffer position: pos].
	self dataAvailableSemaphore signal.
	self changed! !

!BufferedAsyncFileReadStream methodsFor: 'read ahead buffer' stamp: 'dtl 10/1/2006 17:12'!
appendToBuffer: aCharacter
	"Append aCharacter to readBuffer, then signal dataAvailableSemaphore to inform
	any blocked reader that the read can proceed, and trigger a #dataReady event to
	inform any interested objects that new data has become available."

	| pos |
	self readSyncSemaphore critical:
		[pos := self readBuffer position.
		readBuffer setToEnd.
		readBuffer nextPut: aCharacter.
		readBuffer position: pos].
	self dataAvailableSemaphore signal.
	self changed! !

!BufferedAsyncFileReadStream methodsFor: 'converting' stamp: 'dtl 7/8/2003 20:54'!
asAsyncFileReadStream
	"Answer a replacement for this object, with asynchronous event handling but
	no buffering."

	^ AsyncFileReadStream
		name: self name
		attachTo: self ioHandle
		writable: self isReadOnly not
! !

!BufferedAsyncFileReadStream methodsFor: 'converting' stamp: 'dtl 7/8/2003 20:55'!
asBufferedAsyncFileReadStream

	^ self! !

!BufferedAsyncFileReadStream methodsFor: 'testing' stamp: 'dtl 2/22/2007 08:00'!
atEnd
	"Answer whether the receiver can access any more objects.

	Warning: If this instance represents the reader end of an OS pipe, it
	is possible for the #atEnd test to give a false negative. In particular, after
	closing the writer end of an empty OSPipe, the reader may not appear to
	be atEnd until some time has elapsed, or until an explicit read on the pipe
	causes the status to be updated. To illustrate the problem:

		(OSPipe new setBufferedReader; yourself) closeWriter; atEnd>>false
		(OSPipe new ) closeWriter; next; yourself; atEnd>>true
		(OSPipe new setBufferedReader; yourself) closeWriter; next; yourself; atEnd>>true"

	^ self readSyncSemaphore
		critical: [self readBuffer atEnd
				and: [super atEnd]]! !

!BufferedAsyncFileReadStream methodsFor: 'testing' stamp: 'dtl 2/22/2007 08:01'!
atEndOfFile
	"Answer whether the receiver is at its end based on the result of
	the last read operation. This uses feof() to test the underlying file
	stream status, and can be used as an alternative to #atEnd, which
	does not properly report end of file status for an OSPipe."

	^ self readSyncSemaphore
		critical: [self readBuffer atEnd
				and: [fileID isNil
						or: [OSProcess accessor isAtEndOfFile: fileID]]]! !

!BufferedAsyncFileReadStream methodsFor: 'read, write, position' stamp: 'dtl 2/22/2007 08:02'!
basicNext
	"Answer the next byte from this file, or nil if at the end of the file.
	If the readBuffer is empty, force a basicNext in order to ensure that
	the end of file flag is updated (in stdio stream)."

	(self readSyncSemaphore
			critical: [self readBuffer atEnd])
		ifTrue: [^ super basicNext]
		ifFalse: [[self isBlocking]
				ifTrue: [self waitForDataReady].
			^ self readSyncSemaphore
				critical: [readBuffer next]]! !

!BufferedAsyncFileReadStream methodsFor: 'accessing' stamp: 'dtl 7/6/2003 10:15'!
dataAvailableSemaphore
	"Signalled one or more times when data becomes available. Only one Process
	is permitted to wait on this Semaphore."

	^ dataAvailableSemaphore ifNil: [dataAvailableSemaphore := Semaphore new]
! !

!BufferedAsyncFileReadStream methodsFor: 'testing' stamp: 'dtl 7/9/2003 20:59'!
isBlocking
	"Answer true if set to blocking mode."

	^ self nonBlockingMode not! !

!BufferedAsyncFileReadStream methodsFor: 'defaults' stamp: 'dtl 7/6/2003 10:15'!
maxReadBufferSize
	"Replace readBuffer when position exceeds this."

	^ 40000
! !

!BufferedAsyncFileReadStream methodsFor: 'read ahead buffer' stamp: 'dtl 12/7/2012 20:02'!
moveAvailableDataFrom: sqFile

	| count bufferSize buffer |
	(readBuffer notNil and: [readBuffer size > self readBufferMemoryWarningThreshold])
		ifTrue: [self notify: 'buffer size ', readBuffer size asString.
			^(Delay forSeconds: 2) wait].
	bufferSize := self readAheadChunkSize.
	buffer := String new: bufferSize.
	[count := self primRead: sqFile
				into: buffer
				startingAt: 1
				count: bufferSize]
		on: Error	"Could fail if closed"
		do: [count := 0].
	(count notNil and: [count > 0])
		ifTrue:
			[count == bufferSize
				ifTrue:
					[self appendAllToBuffer: buffer]
				ifFalse:
					[count > 0
						ifTrue:
							[self appendAllToBuffer: (buffer copyFrom: 1 to: count)]]]
		ifFalse:
			[(self closed or: [OSProcess accessor isAtEndOfFile: self ioHandle])
				ifTrue:
					[self disableEventHandling]]
! !

!BufferedAsyncFileReadStream methodsFor: 'read ahead buffer' stamp: 'dtl 7/6/2003 20:30'!
moveAvailableDataToBuffer

	^ self moveAvailableDataFrom: (OSProcess accessor handleFromAccessor: self ioHandle).
! !

!BufferedAsyncFileReadStream methodsFor: 'read, write, position' stamp: 'dtl 9/3/2010 17:01'!
next

	((self readSyncSemaphore
				critical: [self readBuffer atEnd])
			and: [self isBlocking])
		ifTrue: [self waitForDataReady].
	^ self readSyncSemaphore
		critical: [readBuffer next]! !

!BufferedAsyncFileReadStream methodsFor: 'read, write, position' stamp: 'dtl 7/6/2003 10:15'!
next: n into: aString startingAt: startIndex
	"Read n bytes into the given string.
	Return aString or a partial copy if less than n elements have been read."

	| count |
	count := self readInto: aString startingAt: startIndex count: n.
	count = n
		ifTrue: [^ aString]
		ifFalse: [^ aString copyFrom: 1 to: startIndex+count-1]! !

!BufferedAsyncFileReadStream methodsFor: 'accessing' stamp: 'dtl 7/9/2003 20:57'!
nonBlockingMode
	"True if nonblocking read behavior should be implemented"

	^ nonBlockingMode ifNil: [nonBlockingMode := true]
! !

!BufferedAsyncFileReadStream methodsFor: 'accessing' stamp: 'dtl 7/9/2003 20:58'!
nonBlockingMode: trueOrFalse
	"True if nonblocking read behavior should be implemented"

	nonBlockingMode := trueOrFalse
! !

!BufferedAsyncFileReadStream methodsFor: 'read, write, position' stamp: 'dtl 9/3/2010 17:01'!
peek
	"Answer what would be returned if the message next were sent to the
	receiver. If the receiver is at the end, answer nil. "

	((self readSyncSemaphore
				critical: [self readBuffer atEnd])
			and: [self isBlocking])
		ifTrue: [self waitForDataReady].
	^ self readSyncSemaphore
		critical: [readBuffer peek]! !

!BufferedAsyncFileReadStream methodsFor: 'defaults' stamp: 'dtl 7/6/2003 10:15'!
readAheadChunkSize
	"The async read ahead process will read at most this many characters.
	Notes: On my system (dtl), a chunk size of 2000 leads to ExternalCommandShell
	overdriving the stdout stream when doing (for example) a long directory listing.
	I have added error handling to accommodate this, but I do know know how
	reliable it is, so I would prefer to avoid generating the condition. A chunk
	size of 200 is small enough that performance is noticably impacted in a
	CommandShell window. A chunk size of 1000 seems to produce good overall
	behavior."

	^ 1000! !

!BufferedAsyncFileReadStream methodsFor: 'accessing' stamp: 'dtl 2/24/2013 10:22'!
readBuffer
	"Read ahead buffer, filled asynchronously as data becomes available on the IO channel"

	^ readBuffer ifNil: [readBuffer := ReadWriteStream on: String new]
! !

!BufferedAsyncFileReadStream methodsFor: 'accessing' stamp: 'dtl 7/6/2003 10:15'!
readBuffer: aStream

	readBuffer := aStream! !

!BufferedAsyncFileReadStream methodsFor: 'defaults' stamp: 'dtl 12/7/2012 19:54'!
readBufferMemoryWarningThreshold
	"A read buffer larger that this may start causing problems. Issue a warning
	before the object memory becomes too large." 

	^self maxReadBufferSize * 1000
! !

!BufferedAsyncFileReadStream methodsFor: 'read, write, position' stamp: 'dtl 3/21/2007 21:56'!
readInto: byteArray startingAt: startIndex count: count 
	"Read into the given array as specified, and return the count actually
	transferred. "

	| s size |
	self isBlocking
		ifTrue: [self waitForDataReady: count.
			s := self readSyncSemaphore
						critical: [self readBuffer next: count].
			size := count]
		ifFalse: [size := self readBuffer size - readBuffer position min: count.
			s := self readSyncSemaphore
						critical: [self readBuffer next: size]].
	byteArray
		replaceFrom: startIndex
		to: startIndex + size - 1
		with: s.
	^ size! !

!BufferedAsyncFileReadStream methodsFor: 'accessing' stamp: 'dtl 7/6/2003 10:15'!
readSyncSemaphore
	"A semaphore for synchronizing access to readBuffer"

	^ readSyncSemaphore ifNil: [readSyncSemaphore := Semaphore forMutualExclusion]! !

!BufferedAsyncFileReadStream methodsFor: 'initialize-release' stamp: 'dtl 7/9/2003 20:54'!
setBlocking
	"Set for blocking reads. Default is nonblocking mode."

	self nonBlockingMode: false
! !

!BufferedAsyncFileReadStream methodsFor: 'initialize-release' stamp: 'dtl 7/9/2003 20:55'!
setNonBlocking
	"Set for nonblocking reads. This is the default mode."

	self nonBlockingMode: true
! !

!BufferedAsyncFileReadStream methodsFor: 'read, write, position' stamp: 'dtl 3/19/2007 18:28'!
upTo: delim 

	^ self readSyncSemaphore critical: [self readBuffer upTo: delim]
! !

!BufferedAsyncFileReadStream methodsFor: 'read, write, position' stamp: 'dtl 12/21/2007 13:14'!
upToEndOfFile
	"Answer a subcollection from the current access position through
	the last element of the receiver.  Use #atEndOfFile to determine end
	of file status with feof(), required for reliable end of file test on
	OS pipes. Close the file when end of file is detected."

	| newStream buffer |
	buffer := buffer1 species new: 1000.
	newStream := WriteStream on: (buffer1 species new: 100).
	[self atEndOfFile] whileFalse:
		[self moveAvailableDataToBuffer.
		newStream nextPutAll: (self nextInto: buffer)].
	self isReadOnly ifTrue: [self close].
	^ newStream contents
! !

!BufferedAsyncFileReadStream methodsFor: 'updating' stamp: 'dtl 7/8/2003 22:06'!
update: aParameter
	"A #changed event is generated as a side effect of this method"

	(aParameter == self eventHandler)
		ifTrue: [self moveAvailableDataToBuffer]
! !

!BufferedAsyncFileReadStream methodsFor: 'read ahead buffer' stamp: 'dtl 7/6/2003 10:15'!
waitForBufferAvailable
	"Block if the readBuffer has grown too large. No-op for now, but add
	this later if large pipes prove to be a problem."! !

!BufferedAsyncFileReadStream methodsFor: 'read ahead buffer' stamp: 'dtl 7/6/2003 10:15'!
waitForDataReady
	"Block until at least one character is available in the readBuffer. This is not
	thread safe, and only one Process is permitted to send this message."

	self dataAvailableSemaphore wait
! !

!BufferedAsyncFileReadStream methodsFor: 'read ahead buffer' stamp: 'dtl 2/22/2007 08:09'!
waitForDataReady: count 
	"Block until at least count characters are available in the readBuffer"

	[self readSyncSemaphore
		critical: [self readBuffer size - readBuffer position < count]]
		whileTrue: [self waitForDataReady]! !

!AttachableFileStream class methodsFor: 'file creation' stamp: 'dtl 6/12/1999 15:53'!
fileNamed: fileName

	self shouldNotImplement
! !

!AttachableFileStream class methodsFor: 'initialize-release' stamp: 'dtl 10/19/2001 21:53'!
initialize

	"AttachableFileStream initialize"

	UseIOHandle := (Smalltalk hasClassNamed: #IOHandle)
! !

!AttachableFileStream class methodsFor: 'file creation' stamp: 'dtl 6/12/1999 15:53'!
isAFileNamed: fileName

	self shouldNotImplement
! !

!AttachableFileStream class methodsFor: 'instance creation' stamp: 'dtl 3/15/2006 19:50'!
name: aSymbolOrString attachTo: anIOHandle writable: readWriteFlag 
	"Create a new instance attached to anIOHandle, where anIOHandle
	represents an open IO channel. For write streams, this represents two
	Smalltalk streams which write to the same OS file or output stream,
	presumably with interleaved output. The purpose of this method is to
	permit a FileStream to be attached to an existing IOHandle, such as
	the IOHandle for standard input, standard output, and standard error."

	^ (super basicNew
		name: aSymbolOrString
		attachTo: anIOHandle
		writable: readWriteFlag) initialize! !

!AttachableFileStream class methodsFor: 'file creation' stamp: 'dtl 6/12/1999 15:54'!
newFileNamed: fileName

	self shouldNotImplement
! !

!AttachableFileStream class methodsFor: 'file creation' stamp: 'dtl 6/12/1999 15:55'!
oldFileNamed: fileName

	self shouldNotImplement
! !

!AttachableFileStream class methodsFor: 'file creation' stamp: 'dtl 6/12/1999 15:56'!
readOnlyFileNamed: fileName

	self shouldNotImplement
! !

!AttachableFileStream class methodsFor: 'registry' stamp: 'dtl 8/25/2006 18:15'!
register: anObject
	"An attachable file stream is generally either a second reference to an
	existing file stream, or a reference to a transient object such as a pipe
	endpoint. There is no need to register it for finalization."

	^ anObject! !

!AttachableFileStream class methodsFor: 'examples' stamp: 'dtl 3/7/2006 06:42'!
stdIn

	"self stdIn"

	^ self name: 'stdIn' attachTo: OSProcess thisOSProcess stdIn ioHandle writable: false
! !

!AttachableFileStream class methodsFor: 'registry' stamp: 'dtl 8/25/2006 18:15'!
unregister: anObject
	"An attachable file stream is generally either a second reference to an
	existing file stream, or a reference to a transient object such as a pipe
	endpoint. There is no need to register it for finalization."

	^ anObject! !

!AttachableFileStream methodsFor: 'converting' stamp: 'dtl 7/8/2003 21:01'!
asAsyncFileReadStream
	"Answer a replacement for this object, with asynchronous event handling.
	Do not close the ioHandle when this object is finalized."

	self keepOpen.
	^ AsyncFileReadStream
		name: self name
		attachTo: self ioHandle
		writable: self isReadOnly not
! !

!AttachableFileStream methodsFor: 'converting' stamp: 'dtl 9/16/2002 17:59'!
asAttachableFileStream

	^ self
! !

!AttachableFileStream methodsFor: 'converting' stamp: 'dtl 7/8/2003 21:01'!
asBufferedAsyncFileReadStream
	"Answer a replacement for this object, with asynchronous event handling
	and buffered output. Do not close the ioHandle when this object is finalized."

	self keepOpen.
	^ BufferedAsyncFileReadStream
		name: self name
		attachTo: self ioHandle
		writable: self isReadOnly not
! !

!AttachableFileStream methodsFor: 'finalization' stamp: 'dtl 9/17/2002 08:08'!
autoClose
	"Private. Answer true if the file should be automatically closed when 
	this object is finalized."

	^ autoClose
		ifNil: [autoClose := true]! !

!AttachableFileStream methodsFor: 'open/close' stamp: 'dtl 9/3/2010 17:00'!
close
	"Close this file."

	| h |
	(h := self ioHandle) ifNotNil:
		[UseIOHandle
			ifTrue:
				[h close.
				self ioHandle: nil]
			ifFalse:
				[self primCloseNoError: h.
				self unregister.
				fileID := nil]]
! !

!AttachableFileStream methodsFor: 'initialize-release' stamp: 'dtl 4/14/2006 09:34'!
disableEventHandling
	"Subclasses may disable event handling"! !

!AttachableFileStream methodsFor: 'open/close' stamp: 'dtl 6/12/1999 16:00'!
ensureOpen

	self shouldNotImplement
! !

!AttachableFileStream methodsFor: 'finalization' stamp: 'dtl 7/6/2006 22:17'!
finalize

	self autoClose
		ifTrue: [[self primCloseNoError: fileID] on: Error do: []]! !

!AttachableFileStream methodsFor: 'read, write, position' stamp: 'dtl 11/8/2000 21:55'!
flush
	"Flush the external OS stream (the one in the C library)."

	OSProcess accessor flushExternalStream: self ioHandle! !

!AttachableFileStream methodsFor: 'private - IOHandle' stamp: 'dtl 1/29/2000 15:18'!
ioHandle

	UseIOHandle
		ifTrue: [^ super ioHandle]
		ifFalse: [^ fileID]! !

!AttachableFileStream methodsFor: 'testing' stamp: 'dtl 3/26/2006 15:52'!
isPipe

	^ false
! !

!AttachableFileStream methodsFor: 'finalization' stamp: 'dtl 9/17/2002 08:05'!
keepOpen
	"Do not allow the file to be closed when this object is finalized."

	autoClose := false
! !

!AttachableFileStream methodsFor: 'attaching' stamp: 'dtl 11/30/2002 09:22'!
name: aSymbolOrString attachTo: anIOHandle writable: readWriteFlag
	"Attach to an existing file handle, assumed to have been previously opened by the underlying operating system."

	name := aSymbolOrString.
	UseIOHandle
		ifTrue: [self ioHandle: anIOHandle]
		ifFalse: [fileID := anIOHandle].
	readWriteFlag ifTrue: [self readWrite] ifFalse: [self readOnly].
	self ascii.
	UseIOHandle ifFalse: [self register]
! !

!AttachableFileStream methodsFor: 'open/close' stamp: 'dtl 6/12/1999 16:00'!
open

	self shouldNotImplement
! !

!AttachableFileStream methodsFor: 'open/close' stamp: 'dtl 6/12/1999 16:01'!
open: fileName forWrite: writeMode

	self shouldNotImplement
! !

!AttachableFileStream methodsFor: 'open/close' stamp: 'dtl 6/12/1999 16:02'!
openReadOnly

	self shouldNotImplement
! !

!AttachableFileStream methodsFor: 'read, write, position' stamp: 'dtl 5/17/2009 14:11'!
position
	"Return the receiver's current file position. If the stream is not positionable,
	as in the case of a Unix pipe stream, answer 0."

	^ [super position]
		on: Error
		do: [0]
! !

!AttachableFileStream methodsFor: 'open/close' stamp: 'dtl 6/12/1999 16:02'!
reopen

	self shouldNotImplement
! !

!AttachableFileStream methodsFor: 'nonblocking' stamp: 'dtl 2/17/2007 18:10'!
setBlocking

	OSProcess accessor setBlocking: self ioHandle! !

!AttachableFileStream methodsFor: 'nonblocking' stamp: 'dtl 2/17/2007 18:10'!
setNonBlocking

	OSProcess accessor setNonBlocking: self ioHandle! !

!AttachableFileStream methodsFor: 'read, write, position' stamp: 'dtl 9/2/2006 17:17'!
upToEnd
	"Answer a subcollection from the current access position through the last element
	of the receiver. This is slower than the method in StandardFileStream, but it
	works with pipes which answer false to #atEnd when no further input is
	currently available, but the pipe is not yet closed."

	| newStream buffer nextBytes |
	buffer := buffer1 species new: 1000.
	newStream := WriteStream on: (buffer1 species new: 100).
	[self atEnd or: [(nextBytes := self nextInto: buffer) isEmpty]]
		whileFalse: [newStream nextPutAll: nextBytes].
	^ newStream contents
! !

!AttachableFileStream methodsFor: 'read, write, position' stamp: 'dtl 6/4/2006 16:02'!
upToEndOfFile
	"Answer a subcollection from the current access position through
	the last element of the receiver.  Use #atEndOfFile to determine end
	of file status with feof(), required for reliable end of file test on
	OS pipes."

	| newStream buffer |
	buffer := buffer1 species new: 1000.
	newStream := WriteStream on: (buffer1 species new: 100).
	[self atEndOfFile] whileFalse: [newStream nextPutAll: (self nextInto: buffer)].
	^ newStream contents! !

!StandardFileStream methodsFor: '*OSProcess-Base' stamp: 'dtl 6/5/2006 06:59'!
atEndOfFile
	"Answer whether the receiver is at its end based on the result of
	the last read operation. This uses feof() to test the underlying file
	stream status, and can be used as an alternative to #atEnd, which
	does not properly report end of file status for an OSPipe."

	^ fileID isNil or: [OSProcess accessor isAtEndOfFile: fileID]
! !

!StandardFileStream methodsFor: '*OSProcess-Base' stamp: 'dtl 1/13/2007 16:44'!
fileID
	"The contents of fileID can and will change after calling this method.
	The sender should copy the result rather than depending on the result
	to be immutable."

	^ fileID! !

!StandardFileStream methodsFor: '*OSProcess-Base' stamp: 'dtl 6/5/2006 07:18'!
isPipe

	^ false
! !

Stream subclass: #ExternalPipe
	instanceVariableNames: 'writer reader blocking'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!ExternalPipe commentStamp: 'dtl 3/10/2006 11:06' prior: 0!
I represent a pipe provided by the underlying operating system, such as a Unix pipe. I have a reader stream and a writer stream which behave similarly to a read-only FileStream and a writeable FileStream.

Subclasses implement buffering behavior for the reader end of a pipe.!

!ExternalPipe class methodsFor: 'instance creation' stamp: 'dtl 4/2/2006 21:34'!
blockingPipe
	"Warning: a read on a blocking pipe will hang the VM if there is insufficient
	data in the pipe to fulfill the read request."

	"OSPipe blockingPipe"

	^ super basicNew initialize; setBlocking
! !

!ExternalPipe class methodsFor: 'instance creation' stamp: 'dtl 4/2/2006 21:34'!
bufferedBlockingPipe
	"Warning: a read on a blocking pipe will hang the VM if there is insufficient
	data in the pipe to fulfill the read request."

	"OSPipe bufferedBlockingPipe"

	^ (super basicNew initialize; setBlocking) setBufferedReader; yourself
! !

!ExternalPipe class methodsFor: 'instance creation' stamp: 'dtl 4/2/2006 21:33'!
bufferedNonBlockingPipe

	"OSPipe bufferedNonBlockingPipe"

	^ (super basicNew initialize; setNonBlocking) setBufferedReader; yourself
! !

!ExternalPipe class methodsFor: 'instance creation' stamp: 'dtl 4/2/2006 21:35'!
new

	"ExternalPipe new"

	^ self nonBlockingPipe
! !

!ExternalPipe class methodsFor: 'instance creation' stamp: 'dtl 3/25/2006 14:14'!
nonBlockingPipe

	"OSPipe nonBlockingPipe"

	^ super basicNew initialize; setNonBlocking
! !

!ExternalPipe class methodsFor: 'examples' stamp: 'dtl 3/7/2006 19:44'!
testPipe
	"OSPipe testPipe inspect"

	| pipe result |
	pipe := self new.
	pipe nextPutAll: 'string to send through an OSPipe'.
	pipe writer close.
	result := pipe upToEnd.
	pipe close.
	^ result
! !

!ExternalPipe methodsFor: 'testing' stamp: 'dtl 4/2/2006 21:14'!
atEnd
	"Answer whether the receiver can access any more objects."

	^ writer closed and: [self peek == nil]
! !

!ExternalPipe methodsFor: 'testing' stamp: 'dtl 6/4/2006 16:01'!
atEndOfFile
	"Answer whether the receiver is at its end based on the result of
	the last read operation. This uses feof() to test the underlying file
	stream status, and can be used as an alternative to #atEnd, which
	does not properly report end of file status for an OSPipe."

	^ reader atEndOfFile
! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 3/7/2006 20:00'!
blocking
	"True if reader end is set to blocking mode."

	^ blocking ifNil: [blocking := true]! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 3/7/2006 20:00'!
blocking: trueOrFalse
	"True if reader end is set to blocking mode."

	blocking := trueOrFalse! !

!ExternalPipe methodsFor: 'finalization' stamp: 'dtl 3/7/2006 19:44'!
close

	self closeWriter; closeReader
! !

!ExternalPipe methodsFor: 'finalization' stamp: 'dtl 3/7/2006 19:44'!
closeReader

	reader ifNotNil: [reader close]
! !

!ExternalPipe methodsFor: 'finalization' stamp: 'dtl 3/7/2006 19:44'!
closeWriter

	writer ifNotNil: [writer close]
! !

!ExternalPipe methodsFor: 'testing' stamp: 'dtl 9/16/2002 17:35'!
closed

	^ reader closed! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 4/16/2003 06:01'!
contents
	"Answer contents of the pipe, and return the contents to the pipe so it can still be read."

	"ExternalPipe new nextPutAll: 'hello'; contents"

	| s |
	self closed ifTrue:
		[self notify: self printString, ' ', self reader printString, ' closed'.
		^ nil].
	s := self reader upToEnd.
	s isEmpty ifFalse:
		[self writer closed
			ifTrue: [self notify: self printString, ' ', self writer printString,
								' closed, cannot replace contents']
			ifFalse: [self nextPutAll: s]].
	^ s! !

!ExternalPipe methodsFor: 'character writing' stamp: 'dtl 3/7/2006 19:44'!
cr
	"Append a return character to the receiver."

	self writer cr! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 4/12/2014 16:04'!
flush
	^writer flush! !

!ExternalPipe methodsFor: 'initialize-release' stamp: 'dtl 3/25/2006 14:08'!
initialize

	^ self makePipe
! !

!ExternalPipe methodsFor: 'testing' stamp: 'dtl 3/26/2006 15:48'!
isPipe

	^ true
! !

!ExternalPipe methodsFor: 'initialize-release' stamp: 'dtl 3/7/2006 19:44'!
makePipe
	"Create an OS pipe and attach it to my input and output streams."

	| handleArray |
	handleArray := OSProcess accessor makePipeHandles.
	handleArray isNil
		ifTrue:
			[self error: 'cannot create OS pipe']
		ifFalse:
			[self reader: (AttachableFileStream
							name: 'pipeReader'
							attachTo: (handleArray at: 1)
							writable: false).
			self writer: (AttachableFileStream
							name: 'pipeWriter'
							attachTo: (handleArray at: 2)
							writable: true)]
! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 9/16/2002 17:33'!
next
	"Answer the next object accessible by the receiver."

	^ reader next! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 9/16/2002 17:33'!
next: anInteger 
	"Answer the next anInteger elements of my collection."

	^ reader next: anInteger
! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 3/7/2006 19:44'!
nextPut: anObject 
	"Insert the argument, anObject, as the next object accessible by the 
	receiver. Answer anObject."

	^ writer nextPut: anObject! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 3/7/2006 19:44'!
nextPutAll: aCollection 
	"Append the elements of aCollection to the sequence of objects accessible 
	by the receiver. Answer aCollection."

	^ writer nextPutAll: aCollection! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 9/16/2002 17:34'!
peek

	^ reader peek! !

!ExternalPipe methodsFor: 'printing' stamp: 'dtl 4/2/2006 11:40'!
printOn: aStream
	"The implementation of Stream>>printOn: has bad side effects when used
	for OSPipe. This implementation is copied from Object."

	| title |
	title := self class name.
	aStream
		nextPutAll: (title first isVowel ifTrue: ['an '] ifFalse: ['a ']);
		nextPutAll: title! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 3/7/2006 19:44'!
reader
	"Answer a stream on the read end of the pipe."

	^ reader! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 3/7/2006 19:44'!
reader: aReadStream

	reader := aReadStream! !

!ExternalPipe methodsFor: 'initialize-release' stamp: 'dtl 3/9/2006 06:40'!
setBlocking
	"Set the reader side of the pipe for blocking reads."

	reader ifNotNil: [OSProcess accessor setBlocking: reader ioHandle].
	self blocking: true
! !

!ExternalPipe methodsFor: 'initialize-release' stamp: 'dtl 4/2/2006 19:06'!
setBufferedReader
	"Use an event driven AsyncFileReadStream to represent the reader end of the pipe.
	This should be used if the pipe will be read by a Smalltalk process. It should not
	be used if the pipe is to be read by an external OS process."

	reader ifNotNil:
		[reader removeDependent: self.
		reader unregister.
		self reader: reader asBufferedAsyncFileReadStream.
		self setNonBlocking.
		reader addDependent: self.
		^ true].
	^ false
! !

!ExternalPipe methodsFor: 'initialize-release' stamp: 'dtl 3/9/2006 06:40'!
setNonBlocking
	"Set the reader side of the pipe for non-blocking reads."

	reader ifNotNil: [OSProcess accessor setNonBlocking: reader ioHandle].
	self blocking: false
! !

!ExternalPipe methodsFor: 'updating' stamp: 'dtl 3/15/2006 07:17'!
triggerDataReady
	"Notify any object waiting for data ready on the pipe."

	self triggerEvent: #dataReady.
! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 9/18/2002 20:29'!
upToEnd
	"Answer the remaining elements in the string"

	reader closed
		ifTrue: [^ '']
		ifFalse: [^ reader upToEnd]! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 11/11/2011 09:27'!
upToEndOfFile
	"Answer the remaining elements in the pipe.  Use #isAtEndOfFile: to
	determine end of file status with feof(), required for reliable end of
	file test on OS pipes."

	reader closed
		ifTrue: [^ '']
		ifFalse: [^ reader upToEndOfFile]! !

!ExternalPipe methodsFor: 'updating' stamp: 'dtl 1/18/2003 14:31'!
update: aParameter
	"Notify any object waiting for data ready on the pipe."

	self changed.
	self triggerDataReady.
	^ super update: aParameter
! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 3/7/2006 19:44'!
writer
	"Answer a stream on the write end of the pipe."

	^ writer! !

!ExternalPipe methodsFor: 'accessing' stamp: 'dtl 3/7/2006 19:44'!
writer: aWriteStream

	writer := aWriteStream! !

ExternalPipe subclass: #OSPipe
	instanceVariableNames: 'nextChar'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!OSPipe commentStamp: 'dtl 3/8/2006 07:27' prior: 0!
I represent a pipe provided by the underlying operating system, such as a Unix pipe. I have a reader stream and a writer stream which behave similarly to a read-only FileStream and a writeable FileStream.

I use a single-character buffer to implement #peek without losing data from the external OS pipe.!

!OSPipe methodsFor: 'accessing' stamp: 'dtl 10/14/2001 12:16'!
next
	"Answer the next object accessible by the receiver."

	| c |
	nextChar isNil
		ifTrue:
			[^ [reader next]
				on: Error
				do: [nil]]
		ifFalse:
			[c := nextChar.
			nextChar := nil.
			^ c]
! !

!OSPipe methodsFor: 'accessing' stamp: 'dtl 2/24/2013 10:24'!
next: anInteger 
	"Answer the next anInteger elements of my collection."

	| c strm |
	strm := WriteStream on: String new.
	(1 to: anInteger) do:
		[:index |
		c := self next.
		c isNil
			ifTrue: [^ strm contents]	
			ifFalse: [strm nextPut: c. false]].
	^ strm contents
! !

!OSPipe methodsFor: 'accessing' stamp: 'dtl 5/16/2006 06:52'!
peek

	^ nextChar isNil
		ifTrue: [reader closed
				ifFalse: [nextChar := reader next]]
		ifFalse: [nextChar]! !

!OSPipe methodsFor: 'accessing' stamp: 'dtl 2/24/2013 10:24'!
upToEnd
	"Answer the remaining elements in the string. This method is retained for backward
	compatibility with older versions of CommandShell."

	| strm s |
	strm := WriteStream on: String new.
	[(s := self next: 2000) isEmpty
		ifTrue: [^ strm contents]
		ifFalse: [strm nextPutAll: s]] repeat
! !

!OSPipe methodsFor: 'accessing' stamp: 'dtl 2/24/2013 10:24'!
upToEndOfFile
	"Answer the remaining elements in the pipe. Use #atEndOfFile to
	determine end of file status with feof(), required for reliable end of
	file test on OS pipes. Compare #upToEnd, which uses the generic end
	of file test in FilePlugin."

	| strm d s |
	strm := WriteStream on: String new.
	d := Delay forMilliseconds: 200.
	[(s := self next: 2000) isEmpty
		ifTrue: [self atEndOfFile
			ifTrue: [^ strm contents]
			ifFalse: [d wait]]
		ifFalse: [strm nextPutAll: s]] repeat
! !

Object subclass: #AioEventHandlerExample
	instanceVariableNames: 'handler ioStream'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-AIO'!

!AioEventHandlerExample commentStamp: 'dtl 7/5/2003 18:38' prior: 0!
Demonstrate asynchronous read handers for file streams, OS pipes, standard input, and sockets. See class category "examples". Some examples require OSProcess.!

!AioEventHandlerExample class methodsFor: 'examples' stamp: 'dtl 9/4/2003 06:22'!
osPipeExample
	"Demonstrate an asynchronous read hander on an OS pipe. Output will
	be displayed on the Transcript."

	"self osPipeExample"

	| pipe aio |
	(Smalltalk hasClassNamed: #OSProcess)
		ifFalse:
			[self notify: 'this example requires OSProcess']
		ifTrue:
			[pipe := (Smalltalk at: #OSPipe) new.
			aio := super new handler: (AioEventHandler for: pipe reader); ioStream: pipe reader.
			aio handler addDependent: aio.
			(1 to: 10) do:
				[:i |
				pipe nextPutAll: 'this is line ', i asString; cr.
				(Delay forMilliseconds: 500) wait].
			^ aio handler close]
! !

!AioEventHandlerExample class methodsFor: 'examples' stamp: 'dtl 9/4/2003 06:22'!
showTtyOnTranscript: ttyName
	"Enter lines on /dev/tty, and watch them show up on the Transcript. Normally,
	/dev/tty corresponds to standard input for the Squeak process, so if you have
	a serial port connected to something that generates data, try this example with
	/dev/whatever to demonstrate asych input on a serial port.

	Warning: This method does not set the file stream for nonblocking input, so
	it can block the Squeak VM. The #getAvailableData attempts to work around this,
	but save your image before testing with another serial interfaces."

	"self showTtyOnTranscript: '/dev/tty'"

	| inputStream handler example |
	inputStream := FileStream readOnlyFileNamed: '/dev/tty'.
	handler := AioEventHandler for: inputStream
				exceptions: true
				readEvents: true
				writeEvents: false.
	example := super new handler: handler; ioStream: inputStream.
	handler addDependent: example.
	Transcript cr; show: ''.
	self notify: 'Enter lines on ', ttyName, ', watch the Transcript, and select "Proceed" when done'.
	^ example handler close
! !

!AioEventHandlerExample class methodsFor: 'examples' stamp: 'dtl 9/4/2003 06:23'!
standardInputExample
	"Enter lines on stdin, and watch them show up on the Transcript."

	"self standardInputExample"

	| aio stdin |
	(Smalltalk hasClassNamed: #OSProcess)
		ifFalse:
			[self notify: 'this example requires OSProcess']
		ifTrue:
			[Transcript cr; show: ''.
			stdin := (Smalltalk at: #OSProcess) thisOSProcess stdIn.
			aio := super new handler: (AioEventHandler for: stdin); ioStream: stdin.
			aio handler addDependent: aio.
			self notify: 'Enter lines on standard input, watch the Transcript, and select "Proceed" when done'.
			^ aio handler close]
! !

!AioEventHandlerExample class methodsFor: 'examples' stamp: 'dtl 9/4/2003 06:24'!
tcpSocketExample
	"Loosely based on OldSocket>>remoteTestServerTCP. Output is displayed on the Transcript."

	"self tcpSocketExample"

	| port serverTcpSocket serverName clientTcpSocket handler example useOldStyleSockets |
	port := 8086.
	serverName := '127.0.0.1'.

	"The networking code was updated for Squeak 3.6. This checks for which version to use."
	useOldStyleSockets := Socket respondsTo: #initializeNetworkIfFail:.

	Transcript show: 'initializing network ... '.

	useOldStyleSockets
		ifTrue:
			[Socket initializeNetworkIfFail: [^Transcript show:'failed']]
		ifFalse:
			[[Socket initializeNetwork]
				on: Error
				do: [:ex | ^Transcript show:'failed']].
	Transcript show:'ok';cr.

	"Create the server (reader) socket"
	serverTcpSocket := Socket newTCP.
	serverTcpSocket listenOn: port.
	[Transcript show: 'server endpoint created on port ', port asString; cr.
	useOldStyleSockets
		ifTrue:
			[serverTcpSocket waitForConnectionUntil: Socket standardDeadline]
		ifFalse:
			[serverTcpSocket waitForConnectionFor: 10]] fork.
	(Delay forMilliseconds: 1000) wait.

	"Create the client (writer) socket"
	clientTcpSocket := Socket newTCP.
	clientTcpSocket connectTo: (NetNameResolver addressFromString: serverName) port: port.
	useOldStyleSockets
		ifTrue:
			[clientTcpSocket waitForConnectionUntil: Socket standardDeadline]
		ifFalse:
			[clientTcpSocket waitForConnectionFor: 10].
	Transcript show: 'client endpoint connected to ', serverName, ' port ', port asString; cr.

	"Set up a read event handler on the server socket"
	handler := AioEventHandler for: serverTcpSocket
				exceptions: true
				readEvents: true
				writeEvents: false.
	example := super new handler: handler; ioStream: serverTcpSocket.
	handler addDependent: example.
	Transcript show: 'event handler started'; cr.

	"Send a few lines of data to the client socket, waiting briefly between lines.
	The event handler will watch the server socket, and copy data to the Transcript
	each time a new line of data is available to the server."
	(1 to: 10) do:
		[:i |
		clientTcpSocket sendData: 'this is line ', i asString, Character cr asString.
		(Delay forMilliseconds: 500) wait].

	clientTcpSocket closeAndDestroy.
	Transcript show: 'client endpoint closed'; cr.
	serverTcpSocket closeAndDestroy.
	Transcript show: 'server endpoint closed'; cr.
	example close.
	Transcript show: 'event handler stopped'; cr.
	^ Array with: example with: serverTcpSocket with: clientTcpSocket! !

!AioEventHandlerExample methodsFor: 'initialize-release' stamp: 'dtl 9/4/2003 06:20'!
close

	self handler close.
	self handler removeDependent: self
! !

!AioEventHandlerExample methodsFor: 'updating' stamp: 'dtl 2/24/2013 10:22'!
getAvailableData
	"Obtain all available data from ioStream. For a FileStream, keep reading until
	a line terminator is reached. This allows use with a FileStream that has not been
	set for nonblocking input."

	| ws c buffer n |
	buffer := String new: 4000.
	(self ioStream isKindOf: FileStream)
		ifTrue:
			[ws := WriteStream on: String new.
			[c := ioStream next.
			(c == Character lf)
				ifTrue: [ws nextPut: Character cr]
				ifFalse: [ws nextPut: c].
			(c ~= Character lf) and: [c ~= Character cr]] whileTrue.
			^ ws contents]
		ifFalse:
			[ioStream dataAvailable
				ifTrue:
					[n := ioStream receiveDataInto: buffer.
					^ buffer copyFrom: 1 to: n]
				ifFalse:
					[^ '']]
! !

!AioEventHandlerExample methodsFor: 'accessing' stamp: 'dtl 7/5/2003 09:40'!
handler

	^ handler! !

!AioEventHandlerExample methodsFor: 'accessing' stamp: 'dtl 7/5/2003 09:40'!
handler: anAioHandler

	handler := anAioHandler! !

!AioEventHandlerExample methodsFor: 'accessing' stamp: 'dtl 7/5/2003 09:57'!
ioStream

	^ ioStream! !

!AioEventHandlerExample methodsFor: 'accessing' stamp: 'dtl 7/5/2003 09:57'!
ioStream: aFileStream

	ioStream := aFileStream! !

!AioEventHandlerExample methodsFor: 'updating' stamp: 'dtl 7/5/2003 18:23'!
update: anObject

	(anObject isKindOf: AioEventHandler)
		ifTrue: [Transcript show: self getAvailableData]
! !

Object subclass: #CSSPropertyDefinition
	instanceVariableNames: 'appliesTo propertyName initialValue media percentages valueString valueType'
	classVariableNames: 'Definitions'
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSPropertyDefinition class methodsFor: 'as yet unclassified' stamp: 'tb 1/18/2006 20:21'!
initialData

	^'( { 
   initialValue = "depends on user agent and writing direction" ; 
   media = ( visual); 
   appliesTo = "block-level elements" ; 
   valueString = "left | right | center | justify |  <string>  |  inherit" ; 
   percentages = no; 
   propertyName = text-align; 
}, { 
 initialValue = none; 
 media = ( visual); 
 appliesTo = "block-level elements" ; 
 valueString = "none | left | right | both |  inherit" ; 
 percentages = no; 
 propertyName = clear; 
}, { 
 initialValue = XX; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<padding-width> {1,4} |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = padding; 
}, { 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "\''table\''  and \''inline-table\'' elements" ; 
 valueString = "auto | fixed |  inherit" ; 
 percentages = no; 
 propertyName = table-layout; 
}, { 
 initialValue = "depends on user agent" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<color>  |  inherit" ; 
 percentages = no; 
 propertyName = color; 
}, 
{ 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "block-level and replaced elements" ; 
 valueString = "<shape>  | auto |  inherit" ; 
 percentages = no; 
 propertyName = clip; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "normal | small-caps |  inherit" ; 
 percentages = no; 
 propertyName = font-variant; 
}, { 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[  \''border-top-width\''  ||  \''border-style\''  ||  <color>  ] |  inherit" ; 
 percentages = no; 
 propertyName = border-top; 
}, { 
 initialValue = 2; 
 media = ( visual, paged); 
 appliesTo = "block-level elements" ; 
 valueString = "<integer>  |  inherit" ; 
 percentages = no; 
 propertyName = widows; 
}, { 
 initialValue = XX; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[ \''background-color\''  ||  \''background-image\''  ||  \''background-repeat\''  ||  \''background-attachment\''  ||  \''background-position\'' ] |  inherit" ; 
 percentages = "allowed on \''background-position\''" ; 
 propertyName = background; 
}, 
{ 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-style>  |  inherit" ; 
 percentages = no; 
 propertyName = border-top-style; 
}, { 
 initialValue = ltr; 
 media = ( visual); 
 appliesTo = "all elements, but see prose" ; 
 valueString = "ltr | rtl |  inherit" ; 
 percentages = no; 
 propertyName = direction; 
}, { 
 initialValue = medium; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<absolute-size>  |  <relative-size>  |  <length>  |  <percentage>  |  inherit" ; 
 percentages = "refer to parent element\''s font size" ; 
 propertyName = font-size; 
}, { 
 initialValue = "see individual properties" ; 
 media = ( visual, interactive); 
 appliesTo = all; 
 valueString = "[  \''outline-color\''  ||  \''outline-style\''  ||  \''outline-width\''  ] |  inherit" ; 
 percentages = no; 
 propertyName = outline; 
}, { 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "all elements but non-replaced inline elements, table rows, and row groups" ; 
 valueString = "<length>  |  <percentage>  | auto |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = width; 
}, 
{ 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<color> {1,4} | transparent |  inherit" ; 
 percentages = no; 
 propertyName = border-color; 
}, { 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-width> {1,4} |  inherit" ; 
 percentages = no; 
 propertyName = border-width; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "normal | bold | bolder | lighter | 100 | 200 | 300 | 400 | 500 | 600 | 700 | 800 | 900 |  inherit" ; 
 percentages = no; 
 propertyName = font-weight; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "normal |  <number>  |  <length>  |  <percentage>  |  inherit" ; 
 percentages = "refer to the font size of the element itself" ; 
 propertyName = line-height; 
}, { 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<margin-width>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = margin-top; 
}, 
{ 
 initialValue = outside; 
 media = ( visual); 
 appliesTo = "elements with \''display: list-item\''" ; 
 valueString = "inside | outside |  inherit" ; 
 percentages = no; 
 propertyName = list-style-position; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = "block-level elements" ; 
 valueString = "normal | pre | nowrap |  inherit" ; 
 percentages = no; 
 propertyName = white-space; 
}, { 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = "all elements except non-replaced inline elements and table elements" ; 
 valueString = "<length>  |  <percentage>  |  inherit" ; 
 percentages = "refer to height of containing block" ; 
 propertyName = min-height; 
}, { 
 initialValue = "the value of the \''color\'' property" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<color>  |  inherit" ; 
 percentages = no; 
 propertyName = border-bottom-color; 
}, { 
 initialValue = medium; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-width>  |  inherit" ; 
 percentages = no; 
 propertyName = border-bottom-width; 
}, 
{ 
 initialValue = auto; 
 media = ( visual, interactive); 
 appliesTo = all; 
 valueString = "[ [ <uri>  ,]* [ auto | crosshair | default | pointer | move | e-resize | ne-resize | nw-resize | n-resize | se-resize | sw-resize | s-resize | w-resize| text | wait | help ] ] |  inherit" ; 
 percentages = no; 
 propertyName = cursor; 
}, { 
 initialValue = auto; 
 media = ( visual, paged); 
 appliesTo = "block-level elements" ; 
 valueString = "<identifier>  | auto" ; 
 percentages = no; 
 propertyName = page; 
}, { 
 initialValue = disc; 
 media = ( visual); 
 appliesTo = "elements with \''display: list-item\''" ; 
 valueString = "disc | circle | square | decimal | decimal-leading-zero |   lower-roman | upper-roman | lower-greek |  lower-alpha | lower-latin | upper-alpha | upper-latin |   hebrew | armenian | georgian |   cjk-ideographic | hiragana | katakana | hiragana-iroha | katakana-iroha |  none |  inherit" ; 
 percentages = no; 
 propertyName = list-style-type; 
}, { 
 initialValue = inline; 
 media = ( all); 
 appliesTo = all; 
 valueString = "inline | block | list-item | run-in | compact | marker | table | inline-table | table-row-group | table-header-group | table-footer-group | table-row | table-column-group | table-column | table-cell | table-caption | none |  inherit" ; 
 percentages = no; 
 propertyName = display; 
}, { 
 initialValue = auto; 
 media = ( visual, paged); 
 appliesTo = "block-level elements" ; 
 valueString = "auto | always | avoid | left | right |  inherit" ; 
 percentages = no; 
 propertyName = page-break-after; 
}, 
{ 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = "\''table\'' and \''inline-table\'' elements" ; 
 valueString = "[<length> <length>?] |  inherit" ; 
 percentages = no; 
 propertyName = border-spacing; 
}, { 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-style> {1,4} |  inherit" ; 
 percentages = no; 
 propertyName = border-style; 
}, { 
 initialValue = none; 
 media = ( all); 
 appliesTo = all; 
 valueString = "[  <identifier> <integer> ? ]+ | none |  inherit" ; 
 percentages = no; 
 propertyName = counter-increment; 
}, { 
 initialValue = auto; 
 media = ( visual, paged); 
 appliesTo = "block-level elements" ; 
 valueString = "auto | always | avoid | left | right |  inherit" ; 
 percentages = no; 
 propertyName = page-break-before; 
}, { 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "positioned elements" ; 
 valueString = "<length>  |  <percentage>  | auto |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = right; 
}, 
{ 
 initialValue = "the value of the \''color\'' property" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<color>  |  inherit" ; 
 percentages = no; 
 propertyName = border-right-color; 
}, { 
 initialValue = medium; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-width>  |  inherit" ; 
 percentages = no; 
 propertyName = border-right-width; 
}, { 
 initialValue = show; 
 media = ( visual); 
 appliesTo = "\''table-cell\'' elements" ; 
 valueString = "show | hide |  inherit" ; 
 percentages = no; 
 propertyName = empty-cells; 
}, { 
 initialValue = none; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<uri>  | none |  inherit" ; 
 percentages = no; 
 propertyName = cue-after; 
}, { 
 initialValue = none; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "code | none |  inherit" ; 
 percentages = no; 
 propertyName = speak-punctuation; 
}, 
{ 
 initialValue = XX; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<margin-width> {1,4} |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = margin; 
}, { 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<number>  | none |  inherit" ; 
 percentages = no; 
 propertyName = font-size-adjust; 
}, { 
 initialValue = baseline; 
 media = ( visual); 
 appliesTo = "inline-level and \''table-cell\'' elements" ; 
 valueString = "baseline | sub | super | top | text-top | middle |     bottom | text-bottom |  <percentage>  |  <length>  |  inherit" ; 
 percentages = "refer to the \''line-height\'' of the element itself" ; 
 propertyName = vertical-align; 
}, { 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[  \''border-top-width\''  ||  \''border-style\''  ||  <color>  ] |  inherit" ; 
 percentages = no; 
 propertyName = border-right; 
}, { 
 initialValue = 50; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<number>  |  inherit" ; 
 percentages = no; 
 propertyName = pitch-range; 
}, 
{ 
 initialValue = inherit; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "visible | hidden | collapse |  inherit" ; 
 percentages = no; 
 propertyName = visibility; 
}, { 
 initialValue = medium; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<number>  |  <percentage>  | silent | x-soft | soft | medium | loud | x-loud |  inherit" ; 
 percentages = "refer to inherited value" ; 
 propertyName = volume; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "normal | italic | oblique |  inherit" ; 
 percentages = no; 
 propertyName = font-style; 
}, { 
 initialValue = "depends on user agent" ; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<time>  |  <percentage>  |  inherit" ; 
 percentages = "see prose" ; 
 propertyName = pause-after; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "normal |  <length>  |  inherit" ; 
 percentages = no; 
 propertyName = word-spacing; 
}, 
{ 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<margin-width>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = margin-bottom; 
}, { 
 initialValue = "the value of the \''color\'' property" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<color>  |  inherit" ; 
 percentages = no; 
 propertyName = border-top-color; 
}, { 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-style>  |  inherit" ; 
 percentages = no; 
 propertyName = border-right-style; 
}, { 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[  \''border-top-width\''  ||  \''border-style\''  ||  <color>  ] |  inherit" ; 
 percentages = no; 
 propertyName = border-bottom; 
}, { 
 initialValue = medium; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-width>  |  inherit" ; 
 percentages = no; 
 propertyName = border-top-width; 
}, 
{ 
 initialValue = none; 
 media = ( visual); 
 appliesTo = "all elements except non-replaced inline elements and table elements" ; 
 valueString = "<length>  |  <percentage>  | none |  inherit" ; 
 percentages = "refer to height of containing block" ; 
 propertyName = max-height; 
}, { 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = "block-level elements" ; 
 valueString = "<length>  |  <percentage>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = text-indent; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "normal | wider | narrower | ultra-condensed | extra-condensed | condensed | semi-condensed  | semi-expanded | expanded | extra-expanded |     ultra-expanded |  inherit" ; 
 percentages = no; 
 propertyName = font-stretch; 
}, { 
 initialValue = XX; 
 media = ( visual); 
 appliesTo = "elements with \''display: list-item\''" ; 
 valueString = "[  \''list-style-type\''  ||  \''list-style-position\''  ||  \''list-style-image\''  ] |  inherit" ; 
 percentages = no; 
 propertyName = list-style; 
}, { 
 initialValue = auto; 
 media = ( visual, paged); 
 appliesTo = "the page context" ; 
 valueString = "<length> {1,2} | auto | portrait | landscape |  inherit" ; 
 percentages = no; 
 propertyName = size; 
}, 
{ 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<padding-width>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = padding-right; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = "all elements, but see prose" ; 
 valueString = "normal | embed | bidi-override |  inherit" ; 
 percentages = no; 
 propertyName = unicode-bidi; 
}, { 
 initialValue = "the value of the \''color\'' property" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<color>  |  inherit" ; 
 percentages = no; 
 propertyName = border-left-color; 
}, { 
 initialValue = medium; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-width>  |  inherit" ; 
 percentages = no; 
 propertyName = border-left-width; 
}, { 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "all elements but non-replaced inline elements, table columns, and column groups" ; 
 valueString = "<length>  |  <percentage>  | auto |  inherit" ; 
 percentages = "see prose" ; 
 propertyName = height; 
}, 
{ 
 initialValue = center; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<angle>  | [[ left-side | far-left | left | center-left | center | center-right | right | far-right | right-side ] || behind ] | leftwards | rightwards |  inherit" ; 
 percentages = no; 
 propertyName = azimuth; 
}, { 
 initialValue = collapse; 
 media = ( visual); 
 appliesTo = "\''table\'' and \''inline-table\'' elements" ; 
 valueString = "collapse | separate |  inherit" ; 
 percentages = no; 
 propertyName = border-collapse; 
}, { 
 initialValue = medium; 
 media = ( visual, interactive); 
 appliesTo = all; 
 valueString = "<border-width>  |  inherit" ; 
 percentages = no; 
 propertyName = outline-width; 
}, { 
 initialValue = medium; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<frequency>  | x-low | low | medium | high | x-high |  inherit" ; 
 percentages = no; 
 propertyName = pitch; 
}, { 
 initialValue = static; 
 media = ( visual); 
 appliesTo = "all elements, but not to generated content" ; 
 valueString = "static | relative | absolute | fixed |  inherit" ; 
 percentages = no; 
 propertyName = position; 
}, 
{ 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "none | [ <color>  || [ <length>{2,3} ,]]* [ <color>  ||  <length>{2,3}] |  inherit" ; 
 percentages = no; 
 propertyName = text-shadow; 
}, { 
 initialValue = none; 
 media = ( all); 
 appliesTo = all; 
 valueString = "[  <identifier> <integer> ? ]+ | none |  inherit" ; 
 percentages = no; 
 propertyName = counter-reset; 
}, { 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<margin-width>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = margin-right; 
}, { 
 initialValue = "depends on user agent" ; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<time>  |  <percentage>  |  inherit" ; 
 percentages = "see prose" ; 
 propertyName = pause-before; 
}, { 
 initialValue = none; 
 media = ( visual); 
 appliesTo = "all elements except non-replaced inline elements and table  elements" ; 
 valueString = "<length>  |  <percentage>  | none |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = max-width; 
}, 
{ 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-style>  |  inherit" ; 
 percentages = no; 
 propertyName = border-left-style; 
}, { 
 initialValue = scroll; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "scroll | fixed |  inherit" ; 
 percentages = no; 
 propertyName = background-attachment; 
}, { 
 initialValue = none; 
 media = ( visual, interactive); 
 appliesTo = all; 
 valueString = "<border-style>  |  inherit" ; 
 percentages = no; 
 propertyName = outline-style; 
}, { 
 initialValue = top; 
 media = ( visual); 
 appliesTo = "\''table-caption\'' elements" ; 
 valueString = "top | bottom | left | right |  inherit" ; 
 percentages = no; 
 propertyName = caption-side; 
}, { 
 initialValue = "depends on user agent" ; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "[[ <specific-voice>   |  <generic-voice>  ],]* [ <specific-voice>   |  <generic-voice>  ] |  inherit" ; 
 percentages = no; 
 propertyName = voice-family; 
}, 
{ 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "positioned elements" ; 
 valueString = "<length>  |  <percentage>  | auto |  inherit" ; 
 percentages = "refer to height of containing block" ; 
 propertyName = bottom; 
}, { 
 initialValue = normal; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "normal |  <length>  |  inherit" ; 
 percentages = no; 
 propertyName = letter-spacing; 
}, { 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[ [  \''font-style\''  ||  \''font-variant\''  ||  \''font-weight\''  ]?  \''font-size\''  [ /  \''line-height\''  ]?  \''font-family\''  ] | caption | icon | menu | message-box | small-caption | status-bar |  inherit" ; 
 percentages = "allowed on \''font-size\'' and \''line-height\''" ; 
 propertyName = font; 
}, { 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<padding-width>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = padding-top; 
}, { 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<padding-width>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = padding-left; 
}, 
{ 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[  \''border-top-width\''  ||  \''border-style\''  ||  <color>  ] |  inherit" ; 
 percentages = no; 
 propertyName = border-left; 
}, { 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<padding-width>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = padding-bottom; 
}, { 
 initialValue = none; 
 media = ( visual); 
 appliesTo = "all but positioned elements and generated content" ; 
 valueString = "left | right | none |  inherit" ; 
 percentages = no; 
 propertyName = float; 
}, { 
 initialValue = "empty string" ; 
 media = ( all); 
 appliesTo = ":before and :after pseudo-elements" ; 
 valueString = "[  <string>  |  <uri>  |  <counter>  | <attr-x> | open-quote | close-quote | no-open-quote | no-close-quote ]+ |  inherit" ; 
 percentages = no; 
 propertyName = content; 
}, { 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "positioned elements" ; 
 valueString = "<length>  |  <percentage>  | auto |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = left; 
}, 
{ 
 initialValue = once; 
 media = ( aural); 
 appliesTo = "elements that have table header information" ; 
 valueString = "once | always |  inherit" ; 
 percentages = no; 
 propertyName = speak-header; 
}, { 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "elements with \''display: marker\''" ; 
 valueString = "<length>  | auto |  inherit" ; 
 percentages = no; 
 propertyName = marker-offset; 
}, { 
 initialValue = level; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<angle>   | below | level | above | higher | lower |  inherit" ; 
 percentages = no; 
 propertyName = elevation; 
}, { 
 initialValue = "depends on user agent" ; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "[ [ <time>  |  <percentage> ]{1,2} ] |  inherit" ; 
 percentages = "see descriptions of \''pause-before\'' and \''pause-after\''" ; 
 propertyName = pause; 
}, { 
 initialValue = auto; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<uri>  mix? repeat? | auto | none |  inherit" ; 
 percentages = no; 
 propertyName = play-during; 
}, 
{ 
 initialValue = continuous; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "digits | continuous |  inherit" ; 
 percentages = no; 
 propertyName = speak-numeral; 
}, { 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "none | [ underline || overline || line-through || blink ] |  inherit" ; 
 percentages = no; 
 propertyName = text-decoration; 
}, { 
 initialValue = invert; 
 media = ( visual, interactive); 
 appliesTo = all; 
 valueString = "<color>  | invert |  inherit" ; 
 percentages = no; 
 propertyName = outline-color; 
}, { 
 initialValue = "depends on user agent" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[ <string> <string> ]+ | none |  inherit" ; 
 percentages = no; 
 propertyName = quotes; 
}, { 
 initialValue = repeat; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "repeat-x | repeat-y | no-repeat | repeat |  inherit" ; 
 percentages = no; 
 propertyName = background-repeat; 
}, 
{ 
 initialValue = "0% 0%" ; 
 media = ( visual); 
 appliesTo = "block-level and replaced elements" ; 
 valueString = "[ [ <percentage>  |  <length>  ]{1,2} |  [ [top | center | bottom] || [left | center | right] ] ] |  inherit" ; 
 percentages = "refer to the size of the box itself" ; 
 propertyName = background-position; 
}, { 
 initialValue = 0; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<margin-width>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = margin-left; 
}, { 
 initialValue = auto; 
 media = ( visual, paged); 
 appliesTo = "block-level elements" ; 
 valueString = "avoid | auto |  inherit" ; 
 percentages = no; 
 propertyName = page-break-inside; 
}, { 
 initialValue = 50; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<number>  |  inherit" ; 
 percentages = no; 
 propertyName = stress; 
}, { 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "positioned elements" ; 
 valueString = "auto |  <integer>  |  inherit" ; 
 percentages = no; 
 propertyName = z-index; 
}, 
{ 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<uri>  | none |  inherit" ; 
 percentages = no; 
 propertyName = background-image; 
}, { 
 initialValue = visible; 
 media = ( visual); 
 appliesTo = "block-level and replaced elements" ; 
 valueString = "visible | hidden | scroll | auto |  inherit" ; 
 percentages = no; 
 propertyName = overflow; 
}, { 
 initialValue = "see individual properties" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[  \''border-width\''  ||  \''border-style\''  ||  <color>  ] |  inherit" ; 
 percentages = no; 
 propertyName = border; 
}, { 
 initialValue = none; 
 media = ( visual, paged); 
 appliesTo = "page context" ; 
 valueString = "[ crop || cross ] | none |  inherit" ; 
 percentages = no; 
 propertyName = marks; 
}, { 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "capitalize | uppercase | lowercase | none |  inherit" ; 
 percentages = no; 
 propertyName = text-transform; 
}, 
{ 
 initialValue = none; 
 media = ( visual); 
 appliesTo = "elements with \''display: list-item\''" ; 
 valueString = "<uri>  | none |  inherit" ; 
 percentages = no; 
 propertyName = list-style-image; 
}, { 
 initialValue = none; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<uri>  | none |  inherit" ; 
 percentages = no; 
 propertyName = cue-before; 
}, { 
 initialValue = "depends on user agent" ; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "[[  <family-name>   |  <generic-family>  ],]* [ <family-name>   |  <generic-family> ] |  inherit" ; 
 percentages = no; 
 propertyName = font-family; 
}, { 
 initialValue = transparent; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<color>  | transparent |  inherit" ; 
 percentages = no; 
 propertyName = background-color; 
}, { 
 initialValue = medium; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<number>   | x-slow | slow | medium | fast | x-fast | faster | slower |  inherit" ; 
 percentages = no; 
 propertyName = speech-rate; 
}, 
{ 
 initialValue = none; 
 media = ( visual); 
 appliesTo = all; 
 valueString = "<border-style>  |  inherit" ; 
 percentages = no; 
 propertyName = border-bottom-style; 
}, { 
 initialValue = 50; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "<number>  |  inherit" ; 
 percentages = no; 
 propertyName = richness; 
}, { 
 initialValue = "UA dependent" ; 
 media = ( visual); 
 appliesTo = "all elements except non-replaced inline elements and table elements" ; 
 valueString = "<length>  |  <percentage>  |  inherit" ; 
 percentages = "refer to width of containing block" ; 
 propertyName = min-width; 
}, { 
 initialValue = auto; 
 media = ( visual); 
 appliesTo = "positioned elements" ; 
 valueString = "<length>  |  <percentage>  | auto |  inherit" ; 
 percentages = "refer to height of containing block" ; 
 propertyName = top; 
}, { 
 initialValue = XX; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "[  \''cue-before\''  ||  \''cue-after\''  ] |  inherit" ; 
 percentages = no; 
 propertyName = cue; 
}, 
{ 
 initialValue = 2; 
 media = ( visual, paged); 
 appliesTo = "block-level elements" ; 
 valueString = "<integer>  |  inherit" ; 
 percentages = no; 
 propertyName = orphans; 
}, { 
 initialValue = normal; 
 media = ( aural); 
 appliesTo = all; 
 valueString = "normal | none | spell-out |  inherit" ; 
 percentages = no; 
 propertyName = speak; 
})'! !

!CSSPropertyDefinition class methodsFor: 'as yet unclassified' stamp: 'tb 8/1/2006 17:58'!
initialize 

	| list dict |
	"CSSPropertyDefinition initialize"
	"Definitions notNil ifTrue: [^self]."
	
	HtmlPListSerializer initialize. "Solves order of initialization dependency"
	list := (HtmlPListSerializer on: (ReadStream on: self initialData)) next.
	dict := Dictionary new.

	list do: 
	[:ea |
		| def |
		def := CSSPropertyDefinition new.
		ea keysDo: [:k | def perform: ((k,':') asSymbol) with: (ea at: k)].
		dict at: def propertyName put: def
	].	
	Definitions := dict.
	dict values do:[:ea | ea valueType]
	! !

!CSSPropertyDefinition class methodsFor: 'instance creation' stamp: 'tb 12/29/2005 16:35'!
named: aString

	^Definitions at: aString ifAbsent: [nil]! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'tb 12/29/2005 16:02'!
appliesTo
	"Answer the value of appliesTo"

	^ appliesTo! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:17'!
appliesTo: anObject
	"Set the value of appliesTo"

	appliesTo := anObject! !

!CSSPropertyDefinition methodsFor: 'printing' stamp: 'tb 12/29/2005 16:54'!
asPList

	| dict |
	dict := Dictionary new.

	#('appliesTo' 'propertyName' 'initialValue' 'media' 'percentages' 'valueString')
	do:[:ea | dict at: ea put: (self perform: (ea asSymbol))].

	^dict! !

!CSSPropertyDefinition methodsFor: 'printing' stamp: 'tb 12/29/2005 16:26'!
definitionString

	| out |
	out := WriteStream on: String new.
	self printDefinitionOn: out.
	^out contents.! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'tb 12/29/2005 16:02'!
initialValue
	"Answer the value of initialValue"

	^ initialValue! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:17'!
initialValue: anObject
	"Set the value of initialValue"

	initialValue := anObject! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'tb 12/29/2005 16:02'!
media
	"Answer the value of media"

	^ media! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:17'!
media: anObject
	"Set the value of media"

	media := anObject! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'tb 12/29/2005 16:02'!
percentages
	"Answer the value of percentages"

	^ percentages! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:17'!
percentages: anObject
	"Set the value of percentages"

	percentages := anObject! !

!CSSPropertyDefinition methodsFor: 'printing' stamp: 'tb 12/29/2005 16:28'!
printDefinitionOn: aStream

	aStream nextPutAll: 'CSSPropertyDefinition new appliesTo: '. appliesTo printOn: aStream.
	aStream nextPutAll: ';'; cr;
	nextPutAll: 'propertyName: '. propertyName printOn: aStream.
	aStream nextPutAll: ';'; cr;
	nextPutAll: 'initialValue: '. initialValue printOn: aStream.
	aStream nextPutAll: ';'; cr;
	nextPutAll: 'media: '. media printOn: aStream.
	aStream nextPutAll: ';'; cr;
	nextPutAll: 'percentages: '. percentages printOn: aStream.
	aStream nextPutAll: ';'; cr;
	nextPutAll: 'valueString: '. valueString printOn: aStream.
	aStream nextPutAll: '.'; cr; cr.! !

!CSSPropertyDefinition methodsFor: 'printing' stamp: 'tb 12/29/2005 16:12'!
printOn: aStream

	aStream nextPutAll: self propertyName; nextPutAll: ': '; nextPutAll: valueString; nextPutAll: '; '! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'tb 12/29/2005 16:02'!
propertyName
	"Answer the value of propertyName"

	^ propertyName! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:17'!
propertyName: anObject
	"Set the value of propertyName"

	propertyName := anObject! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'tb 12/29/2005 16:02'!
valueString
	"Answer the value of valueString"

	^ valueString! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:17'!
valueString: anObject
	"Set the value of valueString"

	valueString := anObject! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'tb 12/29/2005 16:32'!
valueType
	"Answer the value of valueType"

	valueType isNil ifTrue: [self valueType: (CSSValueType fromString: self valueString)].
	^valueType! !

!CSSPropertyDefinition methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:18'!
valueType: anObject
	"Set the value of valueType"

	valueType := anObject! !

Object subclass: #CSSRuleSet
	instanceVariableNames: 'idSelectors classSelectors tagSelectors otherSelectors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSRuleSet commentStamp: '<historical>' prior: 0!
(1) The selector uses an ID. (Example: #foo)
(2) The selector doesn't have an ID but uses a class. (Example: .foo)
(3) The selector has no class or ID but specifies a tag name. (Example: div)
(4) The selector specifies none of these things. (Example: *[disabled])!

!CSSRuleSet methodsFor: 'accessing' stamp: 'tb 1/18/2006 23:55'!
candidateRulesForNode: aNode

	| id classes list |

	id := aNode id.
	list := OrderedCollection new.
	list addAll: otherSelectors.
	aNode id isNil ifFalse:
	[
		list addAll: (idSelectors at: id ifAbsent:[#()]).
	].
	classes := aNode classes.
	(classes isNil or: [classes isEmpty]) ifFalse:
	[
		classes do: [:ea | list addAll: (classSelectors at: ea ifAbsent: [#()])]. 
	].
	list addAll: (tagSelectors at: aNode tag asLowercase ifAbsent: [#()]).
	^list! !

!CSSRuleSet methodsFor: 'accessing' stamp: 'tb 1/3/2006 02:42'!
classSelectors
	"Answer the value of classSelectors"

	^ classSelectors! !

!CSSRuleSet methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:19'!
classSelectors: anObject
	"Set the value of classSelectors"

	classSelectors := anObject! !

!CSSRuleSet methodsFor: 'accessing' stamp: 'tb 1/3/2006 02:39'!
idSelectors
	"Answer the value of idSelectors"

	^ idSelectors! !

!CSSRuleSet methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:20'!
idSelectors: anObject
	"Set the value of idSelectors"

	idSelectors := anObject! !

!CSSRuleSet methodsFor: 'initialize-release' stamp: 'tb 1/3/2006 03:22'!
initialize

	super initialize.
	idSelectors := Dictionary new.
	classSelectors := Dictionary new.
	tagSelectors := Dictionary new.
	otherSelectors := OrderedCollection new.! !

!CSSRuleSet methodsFor: 'accessing' stamp: 'tb 1/3/2006 02:39'!
otherSelectors
	"Answer the value of otherSelectors"

	^ otherSelectors! !

!CSSRuleSet methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:20'!
otherSelectors: anObject
	"Set the value of otherSelectors"

	otherSelectors := anObject! !

!CSSRuleSet methodsFor: 'accessing' stamp: 'tb 1/3/2006 02:39'!
tagSelectors
	"Answer the value of tagSelectors"

	^ tagSelectors! !

!CSSRuleSet methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:20'!
tagSelectors: anObject
	"Set the value of tagSelectors"

	tagSelectors := anObject! !

Object subclass: #CSSSelector
	instanceVariableNames: 'isValid'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSSelector subclass: #CSSAttributeSelector
	instanceVariableNames: 'attribute value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSAttributeSelector subclass: #CSSAttributeBeginsWithSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAttributeBeginsWithSelector methodsFor: 'as yet unclassified' stamp: 'tb 12/21/2005 08:31'!
matches: aNode

	| attr |

	attr := self privateAttributeKeyInNode: aNode.

	^attr isNil 
		ifTrue: [false]
		ifFalse: 
		[
			((aNode attributes at: attr) findTokens: '-') 
				first = self value unquoted
		]! !

CSSAttributeSelector subclass: #CSSAttributeEqualsSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAttributeEqualsSelector methodsFor: 'as yet unclassified' stamp: 'tb 12/21/2005 08:32'!
matches: aNode

	| attr |

	attr := self privateAttributeKeyInNode: aNode.

	^attr isNil 
		ifFalse: [(aNode attributes at: attr) unquoted = self value unquoted]
		ifTrue: [false]! !

!CSSAttributeEqualsSelector methodsFor: 'as yet unclassified' stamp: 'tb 7/12/2007 12:13'!
printOn: aStream

	aStream nextPut: $[;
	 nextPutAll: self attribute;
	 nextPut: $=;	
	 nextPut: $";
	 nextPutAll: self value;
	 nextPut: $";
	 nextPut: $]! !

CSSAttributeEqualsSelector subclass: #CSSIDSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSIDSelector class methodsFor: 'as yet unclassified' stamp: 'tb 12/21/2005 14:28'!
readFrom: in

	| out | 
	
	out := WriteStream on: String new.
	in peek = $# ifTrue: [in next].
	[in atEnd not and:[in peek isAlphaNumeric or:['-_' includes: in peek]]] 
		whileTrue: [out nextPut: in next].
	^self new value: out contents.! !

!CSSIDSelector methodsFor: 'as yet unclassified' stamp: 'tb 12/20/2005 21:23'!
initialize

	super initialize.
	self attribute: 'id'! !

!CSSIDSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/3/2006 05:13'!
printOn:aStream

	aStream nextPut: $#.
	aStream nextPutAll: self value.! !

CSSAttributeSelector subclass: #CSSAttributeExistsSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAttributeExistsSelector methodsFor: 'as yet unclassified' stamp: 'tb 12/21/2005 08:24'!
matches: aNode

	^(self privateAttributeKeyInNode: aNode) notNil! !

CSSAttributeSelector subclass: #CSSAttributeHasPrefixSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAttributeHasPrefixSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 21:34'!
isCSS3

	^true! !

!CSSAttributeHasPrefixSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 21:03'!
matches: aNode

	| attr |

	attr := self privateAttributeKeyInNode: aNode.

	^attr isNil 
		ifTrue: [false]
		ifFalse: 
		[
			(aNode attributes at: attr) beginsWith: self value unquoted
		]! !

CSSAttributeSelector subclass: #CSSAttributeHasSubstringSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAttributeHasSubstringSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 21:34'!
isCSS3

	^true! !

!CSSAttributeHasSubstringSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 21:05'!
matches: aNode

	| attr |

	attr := self privateAttributeKeyInNode: aNode.

	^attr isNil 
		ifTrue: [false]
		ifFalse: 
		[
			((aNode attributes at: attr) findString: self value unquoted) > 0
		]! !

CSSAttributeSelector subclass: #CSSAttributeHasSuffixSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAttributeHasSuffixSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 21:33'!
isCSS3

	^true! !

!CSSAttributeHasSuffixSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 21:04'!
matches: aNode

	| attr |

	attr := self privateAttributeKeyInNode: aNode.

	^attr isNil 
		ifTrue: [false]
		ifFalse: 
		[
			(aNode attributes at: attr) endsWith: self value unquoted
		]! !

CSSAttributeSelector subclass: #CSSAttributeIncludesSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAttributeIncludesSelector methodsFor: 'as yet unclassified' stamp: 'tb 12/21/2005 08:30'!
matches: aNode

	| attr |

	attr := self privateAttributeKeyInNode: aNode.

	^attr isNil 
		ifTrue: [false]
		ifFalse: 
		[
			(((aNode attributes at: attr) findTokens: ' ') 
				collect: [:ea | ea unquoted])
					includes: self value unquoted
		]! !

CSSAttributeIncludesSelector subclass: #CSSClassSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSClassSelector class methodsFor: 'as yet unclassified' stamp: 'tb 1/3/2006 17:54'!
readFrom: in

	| out | 
	
	out := WriteStream on: String new.
	in peek = $. ifTrue: [in next].
	[in atEnd not and: [in peek isAlphaNumeric or:['-_' includes: in peek]]] 
		whileTrue: [out nextPut: in next].
	^self new value: out contents unquoted.! !

!CSSClassSelector methodsFor: 'as yet unclassified' stamp: 'tb 12/20/2005 21:22'!
initialize

	super initialize.
	self attribute: 'class'! !

!CSSClassSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/3/2006 17:54'!
matches: aNode

	^aNode classes includes: self value! !

!CSSClassSelector methodsFor: 'as yet unclassified' stamp: 'tb 12/21/2005 14:16'!
printOn:aStream

	aStream nextPut: $..
	aStream nextPutAll: self value.! !

!CSSAttributeSelector class methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 21:08'!
readFrom: in

	| out |
	in peek = $[ ifFalse: [self halt].
	in next.
	out := WriteStream on: String new.
	[in atEnd not and: [in peek isLetter or: [in peek isDigit or:['-_' includes: in peek]]]] 
		whileTrue:[out nextPut: in next].

	in skipSeparators.
	in peek = $] ifTrue:
	[
		in next.
		^CSSAttributeExistsSelector new attribute: out contents
	]
	ifFalse:
	[
		| sel val |
		sel := nil.
		in peek = $= ifTrue: [sel := CSSAttributeEqualsSelector new attribute: out contents]
		ifFalse: 
		[
		 	(in peek = $~ and: [in next. in peek = $=]) 
			ifTrue: [sel := CSSAttributeIncludesSelector new attribute: out contents]
			ifFalse: 
			[
				(in peek = $| and: [in next. in peek = $=]) 
				ifTrue:[sel := CSSAttributeBeginsWithSelector new attribute: out contents]
				ifFalse:
				[
					(in peek = $^ and: [in next. in peek = $=]) ifTrue:
					[sel := CSSAttributeHasPrefixSelector new attribute: out contents]
					ifFalse:
					[
						(in peek = $$ and: [in next. in peek = $=]) ifTrue:
						[
							sel := CSSAttributeHasSuffixSelector new attribute: out contents
						]
						ifFalse:
						[
							(in peek = $* and: [in next. in peek = $=]) ifTrue:
							[
								sel := CSSAttributeHasSubstringSelector new 
									attribute: out contents
							]
							ifFalse:
							[
								in upTo: $].
								^CSSInvalidAttributeSelector new attribute: out contents.
							]
						]
					]
				]
			]
		].
		in next. "Skip ="
		in skipSeparators.
		in peek = $" 
			ifTrue:[in next. val := in upTo: $"]
			ifFalse:
			[
				in peek = $' 
					ifTrue:[in next. val := in upTo: $']
					ifFalse:
					[
						out := WriteStream on: String new.
						[in peek isLetter or: [in peek isDigit or: ['-.' includes: in peek]]]
						whileTrue: [out nextPut: in next].
						val := out contents.
					]
			].
		in next. "Consume the last ]"
		^sel value: val.
	]
	! !

!CSSAttributeSelector methodsFor: 'comparing' stamp: 'tb 1/3/2006 04:20'!
= anObject

	^anObject class == self class and: 
		[anObject attribute = self attribute and: 
			[anObject value = self value]
		]! !

!CSSAttributeSelector methodsFor: 'accessing' stamp: 'tb 12/20/2005 21:21'!
attribute
	"Answer the value of attribute"

	^ attribute! !

!CSSAttributeSelector methodsFor: 'accessing' stamp: 'tb 7/12/2007 11:17'!
attribute: anObject
	"Set the value of attribute"

	attribute := anObject asLowercase asSymbol! !

!CSSAttributeSelector methodsFor: 'comparing' stamp: 'tb 1/3/2006 04:20'!
hash

	^self class hash + self attribute hash + self value hash! !

!CSSAttributeSelector methodsFor: 'testing' stamp: 'tb 1/3/2006 04:55'!
isAttributeSelector

	^true! !

!CSSAttributeSelector methodsFor: 'accessing' stamp: 'tb 7/12/2007 11:19'!
privateAttributeKeyInNode: aNode

	^(aNode dtd isXML 
		ifTrue: 
		[
			(aNode attributes includesKey: self attribute) 
				ifFalse: [nil] 
				ifTrue:[self attribute]
		] 
		ifFalse: 
		[
			(aNode attributes includesKey: self attribute asLowercase) 
				ifFalse: [nil] 
				ifTrue:[self attribute]
		]).
! !

!CSSAttributeSelector methodsFor: 'accessing' stamp: 'tb 12/21/2005 12:06'!
value

	^value! !

!CSSAttributeSelector methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
value: anObject
	"Set the value of value"

	value := anObject! !

CSSAttributeSelector subclass: #CSSInvalidAttributeSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSInvalidAttributeSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 20:35'!
isValid

	^false! !

CSSSelector subclass: #CSSInvalidSelector
	instanceVariableNames: 'selectorString'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSInvalidSelector methodsFor: 'accessing' stamp: 'tb 12/30/2005 19:19'!
isValid

	^false! !

!CSSInvalidSelector methodsFor: 'printing' stamp: 'tb 12/30/2005 19:34'!
printOn: aStream

	aStream nextPutAll: 'INVALID: '; nextPutAll: self selectorString! !

!CSSInvalidSelector methodsFor: 'accessing' stamp: 'tb 12/30/2005 19:30'!
selectorString
	"Answer the value of selectorString"

	^ selectorString! !

!CSSInvalidSelector methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:15'!
selectorString: anObject
	"Set the value of selectorString"

	selectorString := anObject! !

CSSSelector subclass: #CSSPseudoClassSelector
	instanceVariableNames: 'pseudoClassName arguments'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSActiveSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSCheckedSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSDisabledSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSEnabledSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSFirstChildSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSFirstChildSelector class methodsFor: 'accessing' stamp: 'tb 12/22/2005 22:23'!
operator

	^'>'! !

CSSPseudoClassSelector subclass: #CSSFirstOfTypeSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSFirstSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSFocusSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSHoverSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSIndeterminateSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSInvalidPseudoClassSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSInvalidPseudoClassSelector methodsFor: 'testing' stamp: 'tb 1/18/2006 20:55'!
isValid

	^false! !

CSSPseudoClassSelector subclass: #CSSLangSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSLangSelector methodsFor: 'accessing' stamp: 'tb 1/18/2006 21:09'!
language
	"Answer the value of language"

	^ self arguments! !

CSSPseudoClassSelector subclass: #CSSLastChildSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSLastOfTypeSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSLinkSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSNotSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSNthChildSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSNthLastChildSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSNthLastOfTypeSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSNthOfTypeSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSPseudoClassSelector class methodsFor: 'as yet unclassified' stamp: 'tb 1/18/2006 21:08'!
readFrom: in

	| cls obj out |
	in next = $: ifFalse: [self halt].
	out := WriteStream on: String new.
	"out nextPutAll: 'CSS'."
	out nextPut: in next asUppercase.
	[in peek notNil and: [in peek isLetter or: [in peek = $-]]] whileTrue: 
	[
		in peek = $- 
			ifTrue: [in next. out nextPut: in next asUppercase] 
			ifFalse: [out nextPut: in next]
	].
	"out nextPutAll: 'Selector'."

	cls := (Smalltalk at: ('CSS',(out contents),'Selector') asSymbol ifAbsent: [self]).
	cls = self ifTrue: [obj := cls new pseudoClassName: out contents] ifFalse: [obj := cls new].
	in skipSeparators.
	in peek = $( ifTrue:[ in next. obj arguments: (in upTo: $) )].
	^obj! !

!CSSPseudoClassSelector methodsFor: 'accessing' stamp: 'tb 1/18/2006 21:08'!
arguments
	"Answer the value of arguments"

	^ arguments! !

!CSSPseudoClassSelector methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:18'!
arguments: anObject
	"Set the value of arguments"

	arguments := anObject! !

!CSSPseudoClassSelector methodsFor: 'as yet unclassified' stamp: 'tb 12/21/2005 15:14'!
matches: aNode

	^true! !

!CSSPseudoClassSelector methodsFor: 'as yet unclassified' stamp: 'tb 1/18/2006 23:33'!
printOn: aStream

	| s |

	s := self pseudoClassName asString.
	aStream nextPut: $: ; nextPutAll: s asLowercase! !

!CSSPseudoClassSelector methodsFor: 'accessing' stamp: 'tb 1/18/2006 21:06'!
pseudoClassName
	"Answer the value of pseudoClassName"

	^ pseudoClassName isNil 
		ifFalse: [pseudoClassName] 
		ifTrue: 
		[
			| nm in out |
			nm := self class name asString.
			nm := nm last: (nm size - ('CSS' size)).
			nm := nm first: (nm size - ('Selector' size)).
			out := WriteStream on: (String new: nm size *2).
			in := ReadStream on: nm.
			out nextPut: in next asLowercase.
			[in atEnd] whileFalse:
			[
				in peek isUppercase 
					ifTrue: [out nextPut: $-. out nextPut: in next asLowercase]
					ifFalse:[out nextPut: in next].
			].
			out contents
		]! !

!CSSPseudoClassSelector methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:18'!
pseudoClassName: anObject
	"Set the value of pseudoClassName"

	pseudoClassName := anObject! !

CSSPseudoClassSelector subclass: #CSSRootSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSTargetSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSPseudoClassSelector subclass: #CSSVisitedSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSSelector subclass: #CSSPseudoElementSelector
	instanceVariableNames: 'pseudoElementName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSSelector subclass: #CSSRelationshipSelector
	instanceVariableNames: 'selectors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSRelationshipSelector subclass: #CSSAdjacentSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAdjacentSelector class methodsFor: 'accessing' stamp: 'tb 12/22/2005 22:23'!
operator

	^'+'! !

!CSSAdjacentSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 12:44'!
isAdjacentSelector

	^true! !

!CSSAdjacentSelector methodsFor: 'testing' stamp: 'tb 12/22/2005 22:38'!
matches: aNode

	| node |
	node := aNode.
	selectors reverseDo: 
	[:ea | 
		(ea matches: node) 
			ifFalse: [^false] 
			ifTrue:
			[
				selectors first == ea ifFalse:
				[
					node := node parent children before: node ifAbsent:[^false]
				]
			]
	].
	^true! !

CSSRelationshipSelector subclass: #CSSChildSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSChildSelector class methodsFor: 'as yet unclassified' stamp: 'tb 1/13/2006 17:42'!
operator

	^'>'! !

!CSSChildSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 12:44'!
isChildSelector

	^true! !

!CSSChildSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 00:12'!
matches: aNode

	| node |

	node := aNode.
	selectors reverseDo:[:ea | (ea matches: node) ifFalse:[^false]. node := node parent].
	^true! !

CSSRelationshipSelector subclass: #CSSDescendantSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSDescendantSelector class methodsFor: 'accessing' stamp: 'tb 12/22/2005 22:22'!
operator

	^' '! !

!CSSDescendantSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 12:45'!
isDescendantSelector

	^true! !

!CSSDescendantSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 00:50'!
matches: aNode

	| node |

	(selectors last matches: aNode) ifFalse: [^false].

	node := aNode parent.
	selectors size -1 to: 1 by: -1 do:
	[:idx |
		| sel | 
		node isNil ifTrue:[^false].
		sel := selectors at: idx.
		[node isNil or: [sel matches: node]] whileFalse:[node := node parent]. 
	].
	^node notNil! !

!CSSDescendantSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 15:26'!
printOn: aStream

	self selectors do: [:ea | ea printOn: aStream. aStream nextPut: $ ]! !

CSSRelationshipSelector subclass: #CSSOrSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSOrSelector class methodsFor: 'accessing' stamp: 'tb 1/18/2006 22:51'!
operator

	^''! !

!CSSOrSelector methodsFor: 'comparing' stamp: 'tb 1/3/2006 04:31'!
= anObject

	^self class = anObject class and: [self selectors asSet = anObject selectors asSet]! !

!CSSOrSelector methodsFor: 'comparing' stamp: 'tb 1/3/2006 04:32'!
hash

	^(self class hash) + (self selectors asSet hash)! !

!CSSOrSelector methodsFor: 'accessing' stamp: 'tb 1/3/2006 03:02'!
keySelectors

	| list |

	list := OrderedCollection new.
	self selectors do: [:ea | list addAll: ea keySelectors].
	^list! !

!CSSOrSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 13:12'!
matches: aNode

	^(selectors detect:[:ea | ea matches: aNode] ifNone:[nil]) notNil! !

!CSSOrSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 15:25'!
printOn: aStream

	self selectors do: [:ea | ea printOn: aStream. self selectors last == ea ifFalse: [aStream nextPutAll: ', ']]! !

!CSSRelationshipSelector class methodsFor: 'accessing' stamp: 'tb 1/18/2006 22:16'!
operator

	self subclassResponsibility! !

!CSSRelationshipSelector class methodsFor: 'accessing' stamp: 'tb 1/18/2006 22:48'!
operators

	| out |
	out := WriteStream on: String new.
	self subclasses do:[:ea | out nextPutAll: ea operator].
	^out contents! !

!CSSRelationshipSelector class methodsFor: 'instance creation' stamp: 'tb 1/18/2006 22:18'!
selectorForOperator: aString

	^(self subclasses detect:[:ea | ea operator = aString] ifNone: [CSSDescendantSelector]) new! !

!CSSRelationshipSelector methodsFor: 'comparing' stamp: 'tb 1/3/2006 04:29'!
= anObject

	self class = anObject class and: [self selectors = anObject selectors]! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 12/21/2005 12:39'!
addSelector: aSelector

	self selectors: (self selectors copyWith: aSelector)! !

!CSSRelationshipSelector methodsFor: 'comparing' stamp: 'tb 1/3/2006 04:30'!
hash

	^(self class hash) + (selectors hash)! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 1/14/2006 21:36'!
isCSS3

	^(self selectors detect:[:ea | ea isCSS3] ifNone: [nil]) notNil! !

!CSSRelationshipSelector methodsFor: 'testing' stamp: 'tb 1/18/2006 22:14'!
isRelationshipSelector

	^true! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 1/14/2006 20:40'!
isValid

	^super isValid and: [(self selectors detect:[:ea | ea isValid not] ifNone: [nil]) isNil]! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 1/3/2006 03:03'!
keySelectors

	^self selectors last keySelectors! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 12/22/2005 22:22'!
operator

	^self class operator! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 12/22/2005 22:26'!
printOn: aStream

	selectors do:
	[:ea | 
		ea printOn: aStream. 
		ea == selectors last 
			ifFalse: [aStream nextPutAll: self operator]
	]! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 1/18/2006 22:41'!
removeLastSelector

	| last |
	last := selectors last.
	self selectors: (selectors first: (selectors size-1)).
	^last! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 12/21/2005 15:16'!
selectors
	"Answer the value of selectors"

	^ selectors isNil ifTrue:[#()] ifFalse: [selectors]! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
selectors: anObject
	"Set the value of selectors"

	selectors := anObject! !

!CSSRelationshipSelector methodsFor: 'accessing' stamp: 'tb 12/30/2005 19:53'!
selectorsMatching: aBlock

	| col |
	col := OrderedCollection new.

	(aBlock value: self) ifTrue: [col add: self].
	self selectors do: [:ea | col addAll: (ea selectorsMatching: aBlock)].
	^col asArray! !

CSSRelationshipSelector subclass: #CSSSiblingSelector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSSiblingSelector class methodsFor: 'accessing' stamp: 'tb 1/18/2006 21:49'!
operator

	^'~'! !

!CSSSiblingSelector methodsFor: 'testing' stamp: 'tb 1/18/2006 21:52'!
isAdjacentSelector

	^false! !

!CSSSiblingSelector methodsFor: 'testing' stamp: 'tb 1/18/2006 22:04'!
matches: aNode

	| predecessors sels |

	(selectors last matches: aNode) ifTrue:
	[
		predecessors := aNode parent children.
		predecessors := predecessors first: (predecessors indexOf: aNode)-1.
		sels := selectors first: (selectors size -1).

		[predecessors isEmpty not] whileTrue:
		[
			(sels last matches: (predecessors last)) ifTrue:
			[
				sels size = 1 ifTrue: [^true] ifFalse: [sels := sels first: (sels size -1)]
			].
			predecessors := predecessors first: (predecessors size-1).
		]		
	].
	^false! !

!CSSSelector class methodsFor: 'as yet unclassified' stamp: 'tb 4/8/2006 20:45'!
fromString: aString

	(aString includes: $,) 
	ifTrue:
	[
		| in list |
		in := ReadStream on: (aString, ' ').
		list := OrderedCollection new.
		[in atEnd] whileFalse:
		[
			| sel |
			sel := CSSSelector oneSelectorFromString: (in upTo: $,).
			sel isNil ifTrue: [^CSSInvalidSelector new selectorString: aString].
			list add: sel.
		].
			
		^CSSOrSelector new selectors: list asArray.
	]
	ifFalse:
	[
		| sel |
		sel := CSSSelector oneSelectorFromString: aString.
		^sel isNil ifTrue: [CSSInvalidSelector new selectorString: aString] ifFalse: [sel]
	]! !

!CSSSelector class methodsFor: 'as yet unclassified' stamp: 'tb 4/8/2006 20:45'!
oneSelectorFromString: aString

	| in sel lastOperator pos operators |
	operators := CSSRelationshipSelector operators.
	in := ReadStream on: aString.
	sel := nil.
	lastOperator := ' '.
	in skipSeparators.
	pos := -5.
	[in atEnd] whileFalse:
	[
		| c |
		in position = pos ifTrue: [^CSSInvalidSelector new selectorString: aString]. "Infinite loop detection"
		pos := in position.
		c := in peek.
		('.[*#:' includes: c) 
			ifTrue: 
			[
				sel isNil 
					ifTrue: [sel := (CSSUniversalSelector readFrom: in)]
					ifFalse: 
					[
						sel isRelationshipSelector ifFalse:
						[
							sel := (CSSRelationshipSelector selectorForOperator: lastOperator) 
								addSelector: sel.
							sel addSelector: (CSSUniversalSelector readFrom: in)
						]
						ifTrue:
						[ 
							lastOperator = sel operator ifFalse:
							[
								| last comb | 		
								last := sel removeLastSelector.
								comb := CSSRelationshipSelector selectorForOperator: lastOperator.
								comb addSelector: last.
								sel selectors isEmpty 
									ifFalse: [sel addSelector: comb] 
									ifTrue: [sel := comb]. 
								comb addSelector: (self oneSelectorFromString: in upToEnd).
							]
							ifTrue:
							[
								sel addSelector: (CSSUniversalSelector readFrom: in).
							]
						]
					]
			]
			ifFalse:
			[
				((operators includes:c) and:[sel notNil]) ifTrue: 
				[
					lastOperator := in next asString.
					sel isRelationshipSelector ifFalse:
					[
						sel := (CSSRelationshipSelector selectorForOperator: lastOperator)
							addSelector: sel.
					]
					ifTrue:
					[
						sel operator = lastOperator ifFalse:
						[
							| last comb | 		
							last := sel removeLastSelector.
							comb := CSSRelationshipSelector selectorForOperator: lastOperator.
							comb addSelector: last.
							sel selectors isEmpty 
								ifFalse: [sel addSelector: comb] 
								ifTrue: [sel := comb]. 
							comb addSelector: (self oneSelectorFromString: in upToEnd)
						]
					]
				]
				ifFalse:
				[
					c isLetter ifTrue:
					[
						sel isNil ifTrue: 
						[
							sel := (CSSTypeSelector readFrom: in).
						]
						ifFalse:
						[
							sel isRelationshipSelector ifFalse:
							[
								sel := (CSSRelationshipSelector selectorForOperator: lastOperator) 
									addSelector: sel.
								sel addSelector: (CSSTypeSelector readFrom: in)
							]
							ifTrue:
							[ 
								lastOperator = sel operator ifFalse:
								[
									| last comb | 		
									last := sel removeLastSelector.
									comb := CSSTypeSelector selectorForOperator: lastOperator.
									comb addSelector: last.
									sel selectors isEmpty 
										ifFalse: [sel addSelector: comb] 
										ifTrue: [sel := comb]. 
									comb addSelector: (self oneSelectorFromString: in upToEnd).
								]
								ifTrue:
								[
									sel addSelector: (CSSTypeSelector readFrom: in).
								]
							]
						]
					]
					ifFalse: 
					[
						(c = $/) ifTrue:
						[
							in next. 
							in peek = $* ifTrue:
							[
								in back.
								CSSComment new parseContents: in.
							]
							ifFalse:
							[
								in back.
							]
						]
					].
				] 
		].
	].
	^sel! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 12:40'!
isAdjacentSelector

	^false! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 1/3/2006 04:55'!
isAttributeSelector

	^false! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 1/14/2006 21:34'!
isCSS3

	^false! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 12:40'!
isChildSelector

	^false! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 12:40'!
isDescendantSelector

	^false! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 1/18/2006 22:14'!
isRelationshipSelector

	^false! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 1/3/2006 02:47'!
isSimple

	^false! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 12/30/2005 19:46'!
isTypeSelector

	^false! !

!CSSSelector methodsFor: 'accessing' stamp: 'tb 12/30/2005 19:09'!
isValid
	"Answer the value of isValid"

	^ isValid isNil or: [isValid]! !

!CSSSelector methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:21'!
isValid: anObject
	"Set the value of isValid"

	isValid := anObject! !

!CSSSelector methodsFor: 'accessing' stamp: 'tb 1/3/2006 02:58'!
keySelectors

	^#()! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 12/21/2005 00:08'!
matches: aNode

	self subclassResponsibility! !

!CSSSelector methodsFor: 'accessing' stamp: 'tb 12/21/2005 14:21'!
printOn: aStream

	self subclassResponsibility! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 12/30/2005 19:50'!
selectorsMatching: aBlock

	^(aBlock value: self) ifTrue: [Array with: self] ifFalse: [#()]! !

!CSSSelector methodsFor: 'testing' stamp: 'tb 12/30/2005 21:26'!
typeSelectors

	^self selectorsMatching: [:ea | ea isTypeSelector]! !

CSSSelector subclass: #CSSUniversalSelector
	instanceVariableNames: 'selectors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSUniversalSelector subclass: #CSSTypeSelector
	instanceVariableNames: 'element'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSTypeSelector class methodsFor: 'as yet unclassified' stamp: 'tb 12/22/2005 18:53'!
readFrom: in

	| out |
	out := WriteStream on: String new.
	(in atEnd not and: [in peek isLetter]) ifTrue: 
	[
		out nextPut: in next.
		[in atEnd not and: [in peek isLetter or: [in peek isDigit or:['-_' includes: in peek]]]] 
			whileTrue:[out nextPut: in next].
	]
	ifFalse: [self halt].
	^self new element: out contents; readAttributeSelectorsFrom: in! !

!CSSTypeSelector methodsFor: 'accessing' stamp: 'tb 12/20/2005 18:03'!
element
	"Answer the value of element"

	^ element! !

!CSSTypeSelector methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:24'!
element: anObject
	"Set the value of element"

	element := anObject! !

!CSSTypeSelector methodsFor: 'testing' stamp: 'tb 12/30/2005 19:56'!
isTypeSelector

	^true! !

!CSSTypeSelector methodsFor: 'accessing' stamp: 'tb 1/18/2006 23:46'!
matches: aNode

	aNode isCSS ifTrue: [^false].
	^(aNode dtd isXML 
		ifTrue:[element = aNode tag] 
		ifFalse: [element asLowercase = aNode tag]) and: [super matches: aNode]! !

!CSSTypeSelector methodsFor: 'accessing' stamp: 'tb 12/21/2005 14:39'!
printOn: aStream

	aStream nextPutAll: self element.
	self selectors do:[:ea | ea printOn: aStream]! !

!CSSUniversalSelector class methodsFor: 'as yet unclassified' stamp: 'tb 1/13/2006 18:28'!
readFrom: in

	in peek = $* ifTrue: [in next].
	^self new readAttributeSelectorsFrom: in! !

!CSSUniversalSelector methodsFor: 'comparing' stamp: 'tb 1/3/2006 04:24'!
= anObject

	^self class = anObject class and: [self selectors = anObject selectors and:[self element = anObject element]]! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 12/20/2005 18:04'!
addSelector: aSelector

	self selectors: (self selectors copyWith: aSelector) ! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 1/3/2006 05:18'!
classSelectors

	^self selectors select:[:ea | ea isAttributeSelector and: [ea attribute asLowercase = 'class']]! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 1/3/2006 02:50'!
element

	^nil! !

!CSSUniversalSelector methodsFor: 'comparing' stamp: 'tb 1/3/2006 04:24'!
hash

	^(self class hash) + (self selectors hash) + (self element hash)! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 1/3/2006 04:56'!
idSelector

	^self selectors detect:[:ea | ea isAttributeSelector and: [ea attribute asLowercase = 'id']] ifNone: [nil]! !

!CSSUniversalSelector methodsFor: 'testing' stamp: 'tb 1/14/2006 21:35'!
isCSS3

	^(self selectors detect:[:ea | ea isCSS3] ifNone: [nil]) notNil! !

!CSSUniversalSelector methodsFor: 'testing' stamp: 'tb 1/3/2006 02:47'!
isSimple

	^true! !

!CSSUniversalSelector methodsFor: 'testing' stamp: 'tb 1/14/2006 20:40'!
isValid

	^super isValid and: [(self selectors detect:[:ea | ea isValid not] ifNone: [nil]) isNil]! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 1/3/2006 03:03'!
keySelectors

	^Array with: self! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 12/21/2005 08:13'!
matches: aNode

	^(self selectors detect:[:ea | (ea matches: aNode) not] ifNone:[nil]) isNil! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 12/21/2005 14:39'!
printOn: aStream

	aStream nextPut: $*.
	self selectors do: [:ea | ea printOn: aStream]! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 1/3/2006 03:50'!
readAttributeSelectorsFrom: in

	| c |

	in skipSeparators.
	c := in peek.
	[c isNil or: [c isSeparator]] whileFalse:
	[
		c = $. ifTrue: [self addSelector: (CSSClassSelector readFrom: in)] ifFalse:[
		c = $# ifTrue: [self addSelector: (CSSIDSelector readFrom: in)] ifFalse:[
		c = $[ ifTrue: [self addSelector: (CSSAttributeSelector readFrom: in)] ifFalse:[
		c = $: ifTrue: [self addSelector: (CSSPseudoClassSelector readFrom: in)] ifFalse:[
		^self]]]].
		
		c := in peek.
	]! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'tb 12/20/2005 18:03'!
selectors
	"Answer the value of selectors"

	^ selectors isNil ifTrue: [#()] ifFalse: [selectors]! !

!CSSUniversalSelector methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
selectors: anObject
	"Set the value of selectors"

	selectors := anObject! !

Object subclass: #CSSValue
	instanceVariableNames: 'valueString'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSValue subclass: #CSSAbsoluteSizeValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAbsoluteSizeValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:37'!
isValid

	^#('xx-small' 'x-small' 'small' 'medium' 'large' 'x-large' 'xx-large') includes: self valueString trimBoth! !

CSSValue subclass: #CSSAttrXValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAttrXValue class methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:38'!
readFrom: in

	|  pos |
	in skipSeparators.
	pos := in position.
	(in match: 'attr(') ifFalse: [in position: pos. ^nil].
	^self new valueString: (in upTo: $)) trimBoth
	
	! !

!CSSAttrXValue methodsFor: 'as yet unclassified' stamp: 'tb 12/29/2005 01:03'!
isValid

	^(self valueString beginsWith: 'attr(') and: [self valueString endsWith: ')']! !

CSSValue subclass: #CSSBorderStyleValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSBorderStyleValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

^#( 'none'
	'hidden'
	'dotted'
	'dashed'
	'solid'
	'double'
	'groove'
	'ridge'
	'inset'
	'outset') includes: self valueString trimBoth
! !

CSSValue subclass: #CSSBorderWidthValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSBorderWidthValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

	| val |
	val := self valueString trimBoth.
	^(#( 'thin' 'medium' 'thick' ) includes: val) or: [(CSSLengthValue fromString: val) isValid]! !

CSSValue subclass: #CSSColorValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSColorValue class methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 02:54'!
readFrom: in

	| out |
	out := WriteStream on: String new.
	in skipSeparators.
	(in peek = $#) ifFalse: 
	[
		[in atEnd not and: [in peek isLetter]] whileTrue: [out nextPut: in next]
	]
	ifTrue: 
	[
		out nextPut: in next.
		[in atEnd not and: [in peek isAlphaNumeric]] whileTrue: [out nextPut: in next]
	].
	^self new valueString: out contents! !

!CSSColorValue methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 02:33'!
isValid

	| s |

	s := valueString asLowercase.

	(#(	'black' 
		'silver' 
		'gray' 
		'white' 
		'maroon' 
		'red' 
		'purple' 
		'fuchsia' 
		'green' 
		'lime' 
		'olive' 
		'yellow' 
		'navy' 
		'blue' 
		'teal' 
		'aqua') includes: s) ifTrue:[^true].

		((s beginsWith: 'rgb(') and: [s endsWith: ')']) ifTrue:
		[ 
			| rgb |
			rgb := (s findTokens: '(), ') copyWithoutFirst.
			(rgb size = 3 and:
			[
				((rgb first endsWith: '%') ifTrue: 
				[(rgb detect:
					[:ea | 
						((ea endsWith: '%') and: 
						[
							((ea first: (ea size-1)) asNumber >=0) 
								and: [(ea first:(ea size-1)) asNumber <=100]
						]) not
					] ifNone:[nil]) isNil ifTrue: [^true]
				]
				ifFalse:
				[
					(rgb detect:
					[:ea |
						(ea asNumber >= 0 and: [ea asNumber <= 255]) not
					] ifNone:[nil]) isNil ifTrue: [^true]
				])
			])
		].
		((s beginsWith: '#') and: 
			[(s size = 7 or: [s size = 4]) and: 
				[(s copyWithoutFirst detect:
					[:ea| ('0123456789abcdef' includes: ea) not] ifNone:[nil]) isNil]
			]) ifTrue: [^true].

	^false
	! !

CSSValue subclass: #CSSCounterValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSCounterValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

	| s |
	s := self valueString trimBoth.
	((s beginsWith: 'counter(') and: [s endsWith: ')' ]) ifTrue:
	[
		| args |
		args := (s findTokens: '(,) ') copyWithoutFirst.
		^args size = 1 or:[args size = 2 and: [self isValidStyle: args last]]
	]
	ifFalse:
	[
		((s beginsWith: 'counters(') and: [s endsWith: ')']) ifTrue:
		[
			| args |
			args := (s findTokens: '(,) ') copyWithoutFirst.
			^args size = 2 or:[args size = 3 and: [self isValidStyle: args last]]
			
		]
	].
	^false! !

!CSSCounterValue methodsFor: 'as yet unclassified' stamp: 'tb 12/29/2005 17:13'!
isValidStyle: aString

	^(CSSPropertyDefinition named: 'list-style-type') valueType isValidValue: aString! !

CSSValue subclass: #CSSEnumeratedValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSEnumeratedValue class methodsFor: 'as yet unclassified' stamp: 'tb 12/30/2005 12:53'!
readFrom: in

	| out set |
	set := (Character separators, $;).
	out := WriteStream on: String new.
	[set includes: in peek] whileTrue: [out nextPut: in next].
	^self new valueString: out contents. 
	 

	! !

CSSValue subclass: #CSSGenericFamilyValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSGenericFamilyValue class methodsFor: 'instance creation' stamp: 'tb 1/6/2006 11:49'!
readFrom: in

	| out pos |
	
	pos := in position.
	out := WriteStream on: String new.
	in skipSeparators.
	(in peek = $') ifTrue: 
	[
		in next.
		[in peek = $'] whileFalse: 
		[
			in peek = $\ ifTrue: [in next].
			out nextPut: in next
		]
	] 
	ifFalse: "better be a double quote"
	[
		in peek = $" ifTrue:
		[
			in next.
			[in peek = $"] whileFalse: 
			[
				in peek = $\ ifTrue: [in next].
				out nextPut: in next
			]
		]
		ifFalse:
		[
			| seps |
			seps := (Character separators, ',;/').
			[in atEnd or: [seps includes: in peek]]whileFalse: [out nextPut: in next]
		]
	].

	^out contents size > 0 
		ifTrue: [self new valueString: out contents] 
		ifFalse: [in position: pos. nil]! !

!CSSGenericFamilyValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

	^#( 'serif' 'sans-serif' 'cursive' 'fantasy' 'monospace' ) includes: self valueString trimBoth! !

CSSValue subclass: #CSSGenericVoiceValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSGenericVoiceValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

	^#( 'male' 'female' 'child' ) includes: self valueString trimBoth! !

CSSValue subclass: #CSSIdentifierValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSIdentifierValue methodsFor: 'testing' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

	| s |
	s := self valueString trimBoth.
	s first isLetter ifFalse:[^false].
	^(s detect:[:ea | (ea isAlphaNumeric or: ['_-' includes: ea]) not] ifNone: [nil]) isNil! !

CSSValue subclass: #CSSIntegerValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSIntegerValue class methodsFor: 'as yet unclassified' stamp: 'tb 12/30/2005 23:49'!
readFrom: in

	| out |
	out := WriteStream on: String new.
	('-+' includes: in peek) ifTrue: [out nextPut: in next].
	[in atEnd not and: [in peek isDigit]] whileTrue: [out nextPut: in next].
	^self new valueString: out contents ! !

!CSSIntegerValue methodsFor: 'as yet unclassified' stamp: 'tb 12/14/2005 23:38'!
isValid

	| str |

	str := ('+-' includes: valueString first)
		ifTrue: [valueString copyWithoutFirst]
		ifFalse:[valueString].

	^(str detect:[:ea | ea isDigit not] ifNone:[nil]) isNil! !

!CSSIntegerValue methodsFor: 'as yet unclassified' stamp: 'tb 12/14/2005 23:55'!
value

	^self isValid ifTrue: [valueString asNumber] ifFalse: [nil]! !

CSSIntegerValue subclass: #CSSNumberValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSNumberValue class methodsFor: 'instance creation' stamp: 'tb 12/31/2005 02:17'!
readFrom: in

	| out |
	out := WriteStream on: String new.
	('-+' includes: in peek) ifTrue: [out nextPut: in next].
	[in atEnd not and: [in peek isDigit]] whileTrue: [out nextPut: in next].
	(in peek = $.) ifTrue: 
	[
		out size = 0 ifTrue: [out nextPut: $0].
		out nextPut: in next.
		[in atEnd not and: [in peek isDigit]] whileTrue: [out nextPut: in next].
	].
	^(in peek isNil or: [(Character separators, ';, /') includes: in peek]) 
		ifFalse: [nil]
		ifTrue:[self new valueString: out contents] ! !

!CSSNumberValue methodsFor: 'as yet unclassified' stamp: 'tb 12/14/2005 23:48'!
isValid

	^([valueString asNumber] on: Exception do:[:ex | ex return: nil]) notNil! !

CSSNumberValue subclass: #CSSUnitizedNumberValue
	instanceVariableNames: 'units'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSUnitizedNumberValue subclass: #CSSAngleValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAngleValue methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:21'!
allowableUnits

	^#('deg' 'grad' 'rad')! !

CSSUnitizedNumberValue subclass: #CSSFrequencyValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSFrequencyValue methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:24'!
allowableUnits

	^#( 'hz' 'khz')! !

!CSSFrequencyValue methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:23'!
isValid

	^super isValid and: [valueString asNumber >= 0]! !

CSSUnitizedNumberValue subclass: #CSSLengthValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSLengthValue methodsFor: 'accessing' stamp: 'tb 12/15/2005 01:35'!
allowableUnits

	^#('em' 'ex' 'px' 'in' 'cm' 'mm' 'pt' 'pc') ! !

!CSSLengthValue methodsFor: 'accessing' stamp: 'tb 12/15/2005 00:06'!
isRelative

	^#( 'em' 'ex' 'px') includes: self units! !

!CSSLengthValue methodsFor: 'accessing' stamp: 'tb 12/15/2005 01:20'!
isValid

	^super isValid and:[units notNil or: [valueString asNumber = 0]]! !

CSSLengthValue subclass: #CSSPaddingWidthValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSPaddingWidthValue methodsFor: 'accessing' stamp: 'tb 12/28/2005 23:24'!
allowableUnits

	^super allowableUnits copyWith: '%'! !

CSSUnitizedNumberValue subclass: #CSSPercentageValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSPercentageValue methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:37'!
allowableUnits

	^#( '%' )! !

!CSSPercentageValue methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:37'!
isValid

	^super isValid and: [valueString asNumber >= 0]! !

CSSUnitizedNumberValue subclass: #CSSTimeValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSTimeValue methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:22'!
allowableUnits

	^#( 's' 'ms')! !

!CSSTimeValue methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:22'!
isValid

	^super isValid and: [valueString asNumber >= 0]! !

!CSSUnitizedNumberValue class methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:36'!
allowableUnits

	^self new allowableUnits! !

!CSSUnitizedNumberValue class methodsFor: 'as yet unclassified' stamp: 'tb 1/5/2006 00:06'!
readFrom: in

	| out |
	out := WriteStream on: String new.
	in skipSeparators.
	('-+' includes: in peek) ifTrue: [out nextPut: in next].
	[in atEnd not and: [in peek isDigit]] whileTrue: [out nextPut: in next].
	(in peek = $.) ifTrue: 
	[
		out size = 0 ifTrue: [out nextPut: $0].
		out nextPut: in next.
		[in atEnd not and: [in peek isDigit]] whileTrue: [out nextPut: in next].
	].
	[in atEnd or: [in peek isSeparator or:[';/,' includes: in peek]]] 
		whileFalse: [out nextPut: in next].
	^out contents size > 0 ifTrue: [self new valueString: out contents] ifFalse: [nil]! !

!CSSUnitizedNumberValue methodsFor: 'accessing' stamp: 'tb 12/15/2005 01:18'!
allowableUnits

	self subclassResponsibility! !

!CSSUnitizedNumberValue methodsFor: 'accessing' stamp: 'tb 12/15/2005 01:15'!
units
	"Answer the value of units"

	^ units! !

!CSSUnitizedNumberValue methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:24'!
units: anObject
	"Set the value of units"

	units := anObject! !

!CSSUnitizedNumberValue methodsFor: 'accessing' stamp: 'tb 12/15/2005 01:17'!
valueString: aString

	| lc |

	lc := aString asLowercase.
	units := (self allowableUnits
			detect: [:ea | lc endsWith: ea] ifNone: [nil]).
	valueString := units notNil 
		ifTrue: [aString first: (aString size - units size)] 
		ifFalse: [aString].
! !

CSSValue subclass: #CSSMarginWidthValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSMarginWidthValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

	| s | 
	s := self valueString trimBoth.
	s = 'auto' ifTrue: [^true].
	(CSSLengthValue fromString: s) isValid ifTrue: [^true].
	(CSSPercentageValue fromString: s) isValid ifTrue: [^true].
	^false! !

CSSValue subclass: #CSSRelativeSizeValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSRelativeSizeValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

	^#( 'smaller' 'larger' ) includes: self valueString trimBoth! !

CSSValue subclass: #CSSShapeValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSShapeValue methodsFor: 'testing' stamp: 'SebastianSastre 1/15/2012 16:38'!
isValid

	| s items |

	s := valueString trimBoth asLowercase.

	((s beginsWith: 'rect(') and: [s endsWith: ')']) ifTrue:
	[
		s := s last: (s size) - ('rect(' size).
		s := s first: s size -1.
		items := s findTokens: ' '.
		^items size = 4 and: [(items detect:
			[:ea | (ea = 'auto' or: [(CSSLengthValue fromString: ea) isValid]) not] ifNone: [nil]) isNil]	
	].
	^false! !

CSSValue subclass: #CSSSpecificVoiceValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSSpecificVoiceValue methodsFor: 'as yet unclassified' stamp: 'tb 12/29/2005 00:20'!
isValid

	^true! !

CSSValue subclass: #CSSStringValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSStringValue subclass: #CSSFamilyNameValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSFamilyNameValue class methodsFor: 'instance creation' stamp: 'tb 1/6/2006 11:49'!
readFrom: in

	| out pos |

	pos := in position.
	out := WriteStream on: String new.
	in skipSeparators.
	(in peek = $') ifTrue: 
	[
		in next.
		[in peek = $'] whileFalse: 
		[
			in peek = $\ ifTrue: [in next].
			out nextPut: in next
		]
	] 
	ifFalse: "better be a double quote"
	[
		in peek = $" ifTrue:
		[
			in next.
			[in peek = $"] whileFalse: 
			[
				in peek = $\ ifTrue: [in next].
				out nextPut: in next
			].
			in next. "Clear the trailing quote"
		]
		ifFalse:
		[
			| seps |
			seps := (Character separators, ',;/').
			[in atEnd or: [seps includes: in peek]]whileFalse: [out nextPut: in next]
		]
	].
	(CSSGenericFamilyValue new valueString: out contents asLowercase) isValid 
		ifTrue: [in position: pos. ^nil].

	^out contents size > 0 
		ifTrue: [self new valueString: out contents] 
		ifFalse: [in position: pos. nil]! !

!CSSStringValue class methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 17:26'!
readFrom: in

	| out |
	out := WriteStream on: String new.
	(in peek = $') ifTrue: 
	[
		in next.
		[in peek = $'] whileFalse: 
		[
			in peek = $\ ifTrue: [in next].
			out nextPut: in next
		].
		in next.
	] 
	ifFalse: "better be a double quote"
	[
		in peek = $" ifTrue:
		[
			in next.
			[in peek = $"] whileFalse: 
			[
				in peek = $\ ifTrue: [in next].
				out nextPut: in next
			].
			in next. "Read last quote"
		]
	].
	^out contents size > 0 ifTrue: [self new valueString: out contents] ifFalse: [nil]! !

!CSSStringValue methodsFor: 'testing' stamp: 'SebastianSastre 1/15/2012 16:39'!
isValid

	| s |

	s := valueString trimBoth.
	^s size > 0! !

!CSSStringValue methodsFor: 'accessing' stamp: 'tb 12/29/2005 00:09'!
value

	^self valueString unquoted! !

CSSValue subclass: #CSSUrlValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSUrlValue subclass: #CSSUriValue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSUrlValue class methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 19:28'!
readFrom: aStream

	| s pos |
	aStream skipSeparators.
	pos := aStream position.
	(aStream match: 'url(' ) ifTrue: [s := aStream upTo: $)]
	ifFalse:
	[
		aStream position: pos.
		aStream peek = $' ifTrue: [aStream next. s:= aStream upTo: $']
		ifFalse: 
		[
			aStream peek = $" ifTrue: [aStream next. s:= aStream upTo: $"]
			ifFalse:
			[
				^nil 
			]
		]
	].
	^self new valueString: s unquoted. 
	! !

!CSSUrlValue methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 19:32'!
isValid

	^true.! !

!CSSUrlValue methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:39'!
value

	| s |

	s := valueString trimBoth.

	^self isValid ifTrue:
	[
		(s beginsWith: 'url(') ifTrue:
		[
			s := s last: s size - 4.
			s := s first: s size - 1.
			s := s trimBoth unquoted.
		]
		ifFalse: [ s unquoted ]
	]
	ifFalse: ['']! !

!CSSValue class methodsFor: 'as yet unclassified' stamp: 'tb 12/15/2005 01:57'!
forType: aType

	^self allSubclasses detect: [:ea | ea type = aType] ifNone: [nil].! !

!CSSValue class methodsFor: 'as yet unclassified' stamp: 'tb 12/30/2005 12:23'!
fromString: aString
	
	^self readFrom: (ReadStream on: aString)
! !

!CSSValue class methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:40'!
readFrom: in

	| out seps |
	out := WriteStream on: String new.
	seps := (Character separators, ';,').
	in skipSeparators.
	[in atEnd or:[seps includes: in peek]] whileFalse: [out nextPut: in next].
	out contents trimBoth size < 1 ifTrue: [^nil].
	^self new valueString: out contents! !

!CSSValue class methodsFor: 'as yet unclassified' stamp: 'tb 12/28/2005 22:15'!
type

	| s in out |
	s := self name.
	s := s last: s size - 'CSS' size.
	s := s first: s size - 'Value' size.
	out := WriteStream on: String new.
	in := ReadStream on: s.
	out nextPut: $<; nextPut: in next asLowercase.
	[in atEnd] whileFalse: 
	[
		in peek isUppercase ifTrue: [out nextPut: $-]. 
		out nextPut: in next asLowercase
	].
	out nextPut: $>.
	^out contents ! !

!CSSValue methodsFor: 'accessing' stamp: 'tb 1/5/2006 08:24'!
isProperty

	^false! !

!CSSValue methodsFor: 'accessing' stamp: 'tb 12/14/2005 23:24'!
isValid

	self subclassResponsibility! !

!CSSValue methodsFor: 'accessing' stamp: 'tb 12/14/2005 23:54'!
value

	self subclassResponsibility! !

!CSSValue methodsFor: 'accessing' stamp: 'tb 12/14/2005 23:23'!
valueString
	"Answer the value of valueString"

	^ valueString! !

!CSSValue methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:25'!
valueString: anObject
	"Set the value of valueString"

	valueString := anObject! !

Object subclass: #CSSValueType
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueType class methodsFor: 'instance creation' stamp: 'tb 12/27/2005 10:44'!
fromString: aString

	| in | 

	in := ReadStream on: aString.

	^self readFrom: in.
	! !

!CSSValueType class methodsFor: 'instance creation' stamp: 'tb 1/3/2006 09:39'!
readFrom: aStream

	| in val pos operator |

	in := aStream. "in is easier to type"
	val := nil.
	operator := nil.
	pos := -100.	
	[in atEnd] whileFalse:
	[
		in skipSeparators.
		in position = pos ifTrue: [self halt].
		pos := in position.
		(in peek isAlphaNumeric or: [',/' includes: in peek]) ifTrue:
		[
			val isNil 
				ifTrue: [val := CSSValueTypeLiteral readFrom: aStream]
				ifFalse: 
				[
					((val respondsTo: #addType:) and: [val operator = operator])ifFalse: 
					[
						val := CSSValueTypeAND new addType: val.
						operator := val operator.
					].
					val addType: (CSSValueTypeLiteral readFrom: aStream)
				]
		]
		ifFalse:
		[
			in peek = $| ifTrue:
			[
				in next.
				in peek = $| ifTrue:
				[
					in next.
					operator isNil ifTrue:
					[
						val := CSSValueTypeANDOR new addType: val.
						operator := val operator.	
					]
				]
				ifFalse:
				[
					operator isNil ifTrue:
					[
						val := CSSValueTypeOR new addType: val.
						operator := val operator.
					]
				]
			]
			ifFalse:
			[
				in peek = $[ ifTrue:
				[
					in next.
					val isNil 
						ifTrue: [val := self readFrom: in]
						ifFalse: 
						[
							(val respondsTo: #addType:) ifFalse: 
							[
								val := CSSValueTypeAND new addType: val.
								operator = val operator.
							].
							val addType: (self readFrom: in)
						]					
				]
				ifFalse:
				[
					in peek = $] ifTrue:
					[
						in next.
						(CSSValueTypeConstraint allTypes includes: in peek) ifTrue:
						[
							val := (CSSValueTypeConstraint readFrom: in) type: val.
						].
						^val
					]
					ifFalse:
					[
						in peek = $< ifTrue:
						[
							val isNil 
								ifTrue: [val := CSSValueTypeValue readFrom: in]
								ifFalse: 
								[
									(val respondsTo: #addType:) ifFalse: 
									[
										val := CSSValueTypeAND new addType: val.
										operator := val operator.
									].
									val addType: (CSSValueTypeValue readFrom: in)
								]
						]
						ifFalse:
						[
							in peek = $' ifTrue:
							[
								val isNil 
									ifTrue: [val := CSSValueTypeProperty readFrom: in]
									ifFalse: 
									[
										(val respondsTo: #addType:) ifFalse: 
										[
											val := CSSValueTypeAND new addType: val.
											operator := val operator.
										].
										val addType: (CSSValueTypeProperty readFrom: in)
									]				
							]
						]
					]		
				]
			]
		]
	].
	^val! !

!CSSValueType methodsFor: 'testing' stamp: 'tb 12/30/2005 22:55'!
addValues: aCollection fromStream: in

	self subclassResponsibility! !

!CSSValueType methodsFor: 'testing' stamp: 'tb 12/27/2005 15:49'!
isTypeAND


	^false! !

!CSSValueType methodsFor: 'testing' stamp: 'tb 12/27/2005 15:26'!
isTypeANDOR


	^false! !

!CSSValueType methodsFor: 'testing' stamp: 'tb 12/27/2005 15:26'!
isTypeOR


	^false! !

!CSSValueType methodsFor: 'testing' stamp: 'tb 12/27/2005 10:20'!
isTypeOneOf


	^false! !

!CSSValueType methodsFor: 'testing' stamp: 'tb 12/27/2005 10:20'!
isTypeOneOrMoreOf


	^false! !

!CSSValueType methodsFor: 'testing' stamp: 'tb 12/27/2005 10:49'!
isValidValue: anObject

	self subclassResponsibility! !

CSSValueType subclass: #CSSValueTypeConstraint
	instanceVariableNames: 'type'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeConstraint class methodsFor: 'accessing' stamp: 'tb 12/27/2005 15:44'!
allTypes

	^self subclasses collect:[:ea | ea type]! !

!CSSValueTypeConstraint class methodsFor: 'instance creation' stamp: 'tb 12/27/2005 15:36'!
forType: aCharacter

	^self subclasses detect: [:ea | ea type = aCharacter] 
		ifNone: [self error: ('Type not found ', (aCharacter asString))].
	! !

!CSSValueTypeConstraint class methodsFor: 'accessing' stamp: 'tb 12/27/2005 20:16'!
isSimple

	^true! !

!CSSValueTypeConstraint class methodsFor: 'instance creation' stamp: 'tb 12/27/2005 21:43'!
readFrom: in

	| cls |

	cls := self forType: in peek.
	^cls isSimple 
		ifTrue: [in next. cls new]
		ifFalse: [cls readFrom: in]! !

!CSSValueTypeConstraint methodsFor: 'printing' stamp: 'tb 12/27/2005 16:05'!
printOn: aStream

	self type printOn: aStream.
	aStream nextPut: self class type! !

!CSSValueTypeConstraint methodsFor: 'accessing' stamp: 'tb 12/27/2005 15:21'!
type
	"Answer the value of type"

	^ type! !

!CSSValueTypeConstraint methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:18'!
type: anObject
	"Set the value of type"

	type := anObject! !

CSSValueTypeConstraint subclass: #CSSValueTypeNToMOf
	instanceVariableNames: 'n m'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeNToMOf class methodsFor: 'accessing' stamp: 'tb 12/27/2005 20:16'!
isSimple

	^false! !

!CSSValueTypeNToMOf class methodsFor: 'instance creation' stamp: 'tb 12/27/2005 20:18'!
readFrom: in

	in peek = ${ ifTrue:
	[
		| s n m |
		s := (in upTo: $}) findTokens: ', {}'.
		n := s first asNumber.
		m := s last asNumber.
		^((CSSValueTypeNToMOf new n: n) m: m).
	].
	^nil! !

!CSSValueTypeNToMOf class methodsFor: 'accessing' stamp: 'tb 12/27/2005 15:58'!
type

	^${! !

!CSSValueTypeNToMOf methodsFor: 'accessing' stamp: 'tb 12/31/2005 02:47'!
addValues: aCollection fromStream: in

	| pos |
	"At least n"
	n timesRepeat: 
	[
		pos := in position. 
		(self type addValues: aCollection fromStream: in) ifFalse: [in position: pos. ^false]
	].

	"Fulfilled the minimum - read up to m elements - result is success"
	m-n timesRepeat:
	[
		pos := in position.
		(self type addValues: aCollection fromStream: in) ifFalse: [in position: pos. ^true]
	].
	^true! !

!CSSValueTypeNToMOf methodsFor: 'accessing' stamp: 'tb 12/27/2005 14:35'!
m
	"Answer the value of m"

	^ m! !

!CSSValueTypeNToMOf methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:18'!
m: anObject
	"Set the value of m"

	m := anObject! !

!CSSValueTypeNToMOf methodsFor: 'accessing' stamp: 'tb 12/27/2005 14:35'!
n
	"Answer the value of n"

	^ n! !

!CSSValueTypeNToMOf methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:18'!
n: anObject
	"Set the value of n"

	n := anObject! !

!CSSValueTypeNToMOf methodsFor: 'printing' stamp: 'tb 12/27/2005 14:47'!
printOn: aStream

	type printOn: aStream.
	aStream nextPut: ${; 
		nextPutAll: self n asString; 
		nextPut: $,; 
		nextPutAll: self m asString; 
		nextPut: $}.! !

CSSValueTypeConstraint subclass: #CSSValueTypeOneOrMoreOf
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeOneOrMoreOf class methodsFor: 'accessing' stamp: 'tb 12/27/2005 15:37'!
type

	^$+! !

!CSSValueTypeOneOrMoreOf methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 01:30'!
addValues: aCollection fromStream: in

	| one pos |
	one := false.
	[
		pos := in position.
		type addValues: aCollection fromStream: in
	] whileTrue: [one := true].
	in position: pos.
	^one! !

CSSValueTypeConstraint subclass: #CSSValueTypeZeroOrMoreOf
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeZeroOrMoreOf class methodsFor: 'accessing' stamp: 'tb 12/27/2005 15:37'!
type

	^$*! !

!CSSValueTypeZeroOrMoreOf methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 01:28'!
addValues: aCollection fromStream: in

	| pos |
	[
		pos := in position.
		type addValues: aCollection fromStream: in
	] whileTrue:[].
	in position: pos.
	^true! !

CSSValueTypeConstraint subclass: #CSSValueTypeZeroOrOneOf
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeZeroOrOneOf class methodsFor: 'accessing' stamp: 'tb 12/27/2005 15:36'!
type

	^$?! !

!CSSValueTypeZeroOrOneOf methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 01:29'!
addValues: aCollection fromStream: in

	| pos |
	pos := in position.
	(type addValues: aCollection fromStream: in) ifFalse:[in position: pos].
	^true! !

CSSValueType subclass: #CSSValueTypeLiteral
	instanceVariableNames: 'literal'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeLiteral class methodsFor: 'instance creation' stamp: 'tb 12/29/2005 00:30'!
readFrom: in

	| out val |

	out := WriteStream on: String new.
	in peek isAlphaNumeric 
		ifTrue: 
		[ 
			out nextPut: in next.
			[in atEnd not and:[in peek isAlphaNumeric or: ['_-' includes: in peek]]] 
			whileTrue: [out nextPut: in next].
		] 
		ifFalse: 
		[
			('/,' includes: in peek) ifTrue: [out nextPut: in next]. 
		].
	val := self new literal: out contents.
	(CSSValueTypeConstraint allTypes includes: in peek) ifTrue:
	[
		val := (CSSValueTypeConstraint readFrom: in) type: val.
	].
	^val! !

!CSSValueTypeLiteral methodsFor: 'testing' stamp: 'tb 1/14/2006 00:28'!
addValues: aCollection fromStream: in

	| pos s out |

	pos := in position.
	in skipSeparators.
	out := WriteStream on: String new.
	[in atEnd not and: [literal includes: in peek]] 
		whileTrue: [out nextPut: in next].
	s := out contents.
	^(s asLowercase = literal) 
		ifFalse: [in position: pos. false] 
		ifTrue: [aCollection add: (CSSIdentifierValue new valueString: self literal). true]! !

!CSSValueTypeLiteral methodsFor: 'testing' stamp: 'tb 12/27/2005 10:52'!
isValueValid: anObject

	^anObject isString and: [self literal = anObject]! !

!CSSValueTypeLiteral methodsFor: 'accessing' stamp: 'tb 12/27/2005 09:53'!
literal
	"Answer the value of literal"

	^ literal! !

!CSSValueTypeLiteral methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:18'!
literal: anObject
	"Set the value of literal"

	literal := anObject! !

!CSSValueTypeLiteral methodsFor: 'printing' stamp: 'tb 12/27/2005 10:48'!
printOn: aStream

	aStream nextPutAll: self literal! !

CSSValueType subclass: #CSSValueTypeProperty
	instanceVariableNames: 'propertyName valueType'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeProperty class methodsFor: 'instance creation' stamp: 'tb 12/27/2005 20:13'!
readFrom: in

	| s val | 
	in skipSeparators.
	in peek = $' ifFalse: [self error: 'CSSValueTypeProperty must begin with "''"'].
	in next.
	s := in upTo: $'.
	val := self new propertyName: s.
	(CSSValueTypeConstraint allTypes includes: in peek) ifTrue:
	[
		val := (CSSValueTypeConstraint readFrom: in) type: val.
	].
	^val! !

!CSSValueTypeProperty methodsFor: 'accessing' stamp: 'tb 1/5/2006 21:23'!
addValues: aCollection fromStream: in

	| returnValue size |
	size := aCollection size.
	returnValue := self valueType addValues: aCollection fromStream: in.
	returnValue ifTrue:
	[
		| vals newSize |
		
		newSize := aCollection size.
		vals := aCollection last: newSize - size.
		[aCollection size > size ] whileTrue: [aCollection removeLast].
		aCollection add: 
			(((CSSProperty propertyNamed: self propertyName)
				isDerived: true) propertyValue: vals).
	].
	^returnValue! !

!CSSValueTypeProperty methodsFor: 'accessing' stamp: 'tb 12/29/2005 00:58'!
isValueValid: anObject

	^self valueType isValueValid: anObject! !

!CSSValueTypeProperty methodsFor: 'printing' stamp: 'tb 12/29/2005 00:59'!
printOn: aStream

	"self propertyName printOn: aStream"
	self valueType printOn: aStream! !

!CSSValueTypeProperty methodsFor: 'accessing' stamp: 'tb 12/27/2005 10:33'!
propertyName
	"Answer the value of propertyName"

	^ propertyName! !

!CSSValueTypeProperty methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:18'!
propertyName: anObject
	"Set the value of propertyName"

	propertyName := anObject! !

!CSSValueTypeProperty methodsFor: 'accessing' stamp: 'tb 12/29/2005 17:11'!
valueType

	valueType isNil 
		ifTrue: [valueType := (CSSPropertyDefinition named: self propertyName) valueType].
	^valueType! !

CSSValueType subclass: #CSSValueTypeRelation
	instanceVariableNames: 'types'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSValueTypeRelation subclass: #CSSValueTypeAND
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeAND methodsFor: 'testing' stamp: 'tb 1/6/2006 11:29'!
addValues: aCollection fromStream: in

	| pos sz |
	pos := in position.
	sz := aCollection size.
	self types do: 	
	[:ea |
		(ea addValues: aCollection fromStream: in) ifFalse: 
		[
			in position: pos.
			[aCollection size > sz] whileTrue: [aCollection removeLast].
			^false
		]
	].
	^true! !

!CSSValueTypeAND methodsFor: 'testing' stamp: 'tb 12/27/2005 15:48'!
isTypeAND

	^true! !

!CSSValueTypeAND methodsFor: 'testing' stamp: 'tb 12/28/2005 22:57'!
operator

	^' '! !

!CSSValueTypeAND methodsFor: 'printing' stamp: 'tb 12/27/2005 16:04'!
printOn: aStream

	aStream nextPut: $[.
	self types do: [:ea | ea printOn: aStream.  ea == types last ifFalse: [ aStream nextPutAll: ' ']].
	aStream nextPut: $]! !

CSSValueTypeRelation subclass: #CSSValueTypeANDOR
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeANDOR methodsFor: 'testing' stamp: 'tb 1/5/2006 00:07'!
addValues: aCollection fromStream: in

	| oneWorked gotOne |

	oneWorked := false.
	gotOne := false.
	[
		self types do:
		[:ea |
			(ea addValues: aCollection fromStream: in) 
				ifTrue: [gotOne := oneWorked := true]
		].
	(gotOne and: [in atEnd not])
	] whileTrue: [ gotOne := false ].

	^oneWorked! !

!CSSValueTypeANDOR methodsFor: 'testing' stamp: 'tb 12/27/2005 15:25'!
isTypeANDOR

	^true! !

!CSSValueTypeANDOR methodsFor: 'testing' stamp: 'tb 12/28/2005 22:58'!
operator

	^'||'! !

!CSSValueTypeANDOR methodsFor: 'printing' stamp: 'tb 12/27/2005 14:44'!
printOn: aStream

	aStream nextPut: $[.
	self types do: [:ea | ea printOn: aStream.  ea == types last ifFalse: [ aStream nextPutAll: ' || ']].
	aStream nextPut: $]! !

CSSValueTypeRelation subclass: #CSSValueTypeOR
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeOR methodsFor: 'testing' stamp: 'tb 12/31/2005 01:50'!
addValues: aCollection fromStream: in

	| pos |
	pos := in position.
	self types do:
	[:ea |
		(ea addValues: aCollection fromStream: in) ifTrue: [^true].
		in position: pos.
	].
	^false! !

!CSSValueTypeOR methodsFor: 'testing' stamp: 'tb 12/27/2005 15:26'!
isTypeOR

	^true! !

!CSSValueTypeOR methodsFor: 'testing' stamp: 'tb 12/27/2005 10:51'!
isValueValid: anObject

	^(self types detect: [:ea | ea isValueValid: anObject] ifNone: [nil]) notNil! !

!CSSValueTypeOR methodsFor: 'testing' stamp: 'tb 12/28/2005 16:13'!
operator

	^'|'! !

!CSSValueTypeOR methodsFor: 'printing' stamp: 'tb 12/27/2005 14:44'!
printOn: aStream

	aStream nextPut: $[.
	self types do: [:ea | ea printOn: aStream.  ea == types last ifFalse: [ aStream nextPutAll: ' | ']].
	aStream nextPut: $]! !

!CSSValueTypeRelation methodsFor: 'accessing' stamp: 'tb 12/27/2005 09:56'!
addType: anObject

	self types: (self types copyWith: anObject)! !

!CSSValueTypeRelation methodsFor: 'accessing' stamp: 'tb 12/28/2005 16:12'!
operator

	self subclassResponsibility! !

!CSSValueTypeRelation methodsFor: 'accessing' stamp: 'tb 12/27/2005 09:55'!
types
	"Answer the value of types"

	^ types isNil ifTrue: [#()] ifFalse: [types]! !

!CSSValueTypeRelation methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:17'!
types: anObject
	"Set the value of types"

	types := anObject! !

CSSValueType subclass: #CSSValueTypeValue
	instanceVariableNames: 'valueClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSValueTypeValue class methodsFor: 'instance creation' stamp: 'tb 12/28/2005 23:45'!
readFrom: in

	| str vt val |

	in peek = $< ifFalse: [self error: 'CSSValueType must begin with ''<'''].

	str := (in upTo: $>),'>'.
	vt := CSSValue forType: str.
	val := vt notNil 
		ifTrue: [self new valueClass: vt] 
		ifFalse: [self error: ('Unknown CSS value type ',str)].
	in skipSeparators.
	(CSSValueTypeConstraint allTypes includes: in peek) ifTrue:
	[
		val := (CSSValueTypeConstraint readFrom: in) type: val.
	].
	^val! !

!CSSValueTypeValue methodsFor: 'testing' stamp: 'tb 12/30/2005 22:58'!
addValues: aCollection fromStream: in

	| pos val |
	pos := in position.
	val := self valueClass readFrom: in.
	^(val notNil and: [val isValid]) 
		ifTrue: [aCollection add: val. true] 
		ifFalse: [in position: pos. false].
	! !

!CSSValueTypeValue methodsFor: 'testing' stamp: 'tb 12/27/2005 14:07'!
isValueValid: anObject

	^(self valueClass new valueString: anObject asString) isValid! !

!CSSValueTypeValue methodsFor: 'printing' stamp: 'tb 12/27/2005 14:10'!
printOn: aStream

	aStream nextPutAll: self valueClass type! !

!CSSValueTypeValue methodsFor: 'accessing' stamp: 'tb 12/27/2005 09:58'!
valueClass
	"Answer the value of valueClass"

	^ valueClass! !

!CSSValueTypeValue methodsFor: 'accessing' stamp: 'sebastian.sastre 9/19/2010 11:17'!
valueClass: anObject
	"Set the value of valueClass"

	valueClass := anObject! !

Object subclass: #HtmlCache
	instanceVariableNames: 'urlToDocument url'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlCache methodsFor: 'accessing' stamp: 'tb 1/17/2008 14:49'!
canonicalize: aStringOrUrl

	| href |

	href := aStringOrUrl asString.
	^((href beginsWith: 'http:') or: [(href beginsWith: 'https:') or:[(href beginsWith: 'ftp:') or: [href beginsWith: 'file:']]]) 
	ifFalse:
	[
		(Url combine: url asString withRelative: href).
	]
	ifTrue: [href].! !

!HtmlCache methodsFor: 'accessing' stamp: 'tb 1/10/2006 10:58'!
contentsForUrl: anUrl

	^(self documentForUrl: anUrl) contents! !

!HtmlCache methodsFor: 'accessing' stamp: 'tb 1/15/2006 19:46'!
documentForUrl: anUrl

	^(self queueUrl: anUrl) ! !

!HtmlCache methodsFor: 'initialize-release' stamp: 'tb 1/10/2006 19:35'!
initialize

	super initialize.
	urlToDocument := Dictionary new.! !

!HtmlCache methodsFor: 'accessing' stamp: 'tb 1/17/2008 14:55'!
queueUrl: anUrl

	| absoluteUrl |
	absoluteUrl := self canonicalize: anUrl.
	^urlToDocument at: absoluteUrl ifAbsentPut: 
	[
		"| request loader |
		loader := HTTPLoader default.
		request := HTTPLoader httpRequestClass for: absoluteUrl in: loader.
		loader addRequest: request.
		request"
		[absoluteUrl asUrl retrieveContents]
		on: Error
		do: ['']
	]	! !

!HtmlCache methodsFor: 'accessing' stamp: 'tb 1/10/2006 10:44'!
url
	"Answer the value of url"

	^ url! !

!HtmlCache methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:51'!
url: anObject
	"Set the value of url"

	url := anObject! !

Object subclass: #HtmlDTD
	instanceVariableNames: 'url document errors dom content comments elements attLists entities'
	classVariableNames: 'Cache'
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTD commentStamp: 'tb 11/30/2005 00:38' prior: 0!
I represent the knowledge in a DTD document regarding valid attributes and tag combinations.

After working with this format for awhile, I can honestly say the inventors of DVD have been taking way too much crack.!

HtmlDTD subclass: #Html401Frameset
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!Html401Frameset class methodsFor: 'as yet unclassified' stamp: 'tb 12/19/2005 21:53'!
defaultUrl

	^'http://www.w3.org/TR/html4/frameset.dtd' asUrl! !

!Html401Frameset class methodsFor: 'as yet unclassified' stamp: 'tb 12/19/2005 21:52'!
docType

	^'<!!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Frameset//EN" "http://www.w3.org/TR/html4/frameset.dtd">'! !

!Html401Frameset methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 12:43'!
dtd
	^ '<!!--
    This is HTML 4.01 Strict DTD, which excludes the presentation 
    attributes and elements that W3C expects to phase out as 
    support for style sheets matures. Authors should use the Strict
    DTD when possible, but may use the Transitional DTD when support
    for presentation attribute and elements is required.
    
    HTML 4 includes mechanisms for style sheets, scripting,
    embedding objects, improved support for right to left and mixed
    direction text, and enhancements to forms for improved
    accessibility for people with disabilities.

          Draft: $Date: 1999/12/24 23:37:48 $

          Authors:
              Dave Raggett <dsr@w3.org>
              Arnaud Le Hors <lehors@w3.org>
              Ian Jacobs <ij@w3.org>

    Further information about HTML 4.01 is available at:

        http://www.w3.org/TR/1999/REC-html401-19991224


    The HTML 4.01 specification includes additional
    syntactic constraints that cannot be expressed within
    the DTDs.

-->
<!!--
    Typical usage:

    <!!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
            "http://www.w3.org/TR/html4/strict.dtd">
    <html>
    <head>
    ...
    </head>
    <body>
    ...
    </body>
    </html>

    The URI used as a system identifier with the public identifier allows
    the user agent to download the DTD and entity sets as needed.

    The FPI for the Transitional HTML 4.01 DTD is:

        "-//W3C//DTD HTML 4.01 Transitional//EN"

    This version of the transitional DTD is:

        http://www.w3.org/TR/1999/REC-html401-19991224/loose.dtd

    If you are writing a document that includes frames, use 
    the following FPI:

        "-//W3C//DTD HTML 4.01 Frameset//EN"

    This version of the frameset DTD is:

        http://www.w3.org/TR/1999/REC-html401-19991224/frameset.dtd

    Use the following (relative) URIs to refer to 
    the DTDs and entity definitions of this specification:

    "strict.dtd"
    "loose.dtd"
    "frameset.dtd"
    "HTMLlat1.ent"
    "HTMLsymbol.ent"
    "HTMLspecial.ent"

-->

<!!--================== Imported Names ====================================-->
<!!-- Feature Switch for frameset documents -->
<!!ENTITY % HTML.Frameset "INCLUDE">

<!!ENTITY % ContentType "CDATA"
    -- media type, as per [RFC2045]
    -->

<!!ENTITY % ContentTypes "CDATA"
    -- comma-separated list of media types, as per [RFC2045]
    -->

<!!ENTITY % Charset "CDATA"
    -- a character encoding, as per [RFC2045]
    -->

<!!ENTITY % Charsets "CDATA"
    -- a space-separated list of character encodings, as per [RFC2045]
    -->

<!!ENTITY % LanguageCode "NAME"
    -- a language code, as per [RFC1766]
    -->

<!!ENTITY % Character "CDATA"
    -- a single character from [ISO10646] 
    -->

<!!ENTITY % LinkTypes "CDATA"
    -- space-separated list of link types
    -->

<!!ENTITY % MediaDesc "CDATA"
    -- single or comma-separated list of media descriptors
    -->

<!!ENTITY % URI "CDATA"
    -- a Uniform Resource Identifier,
       see [URI]
    -->

<!!ENTITY % Datetime "CDATA" -- date and time information. ISO date format -->


<!!ENTITY % Script "CDATA" -- script expression -->

<!!ENTITY % StyleSheet "CDATA" -- style sheet data -->



<!!ENTITY % Text "CDATA">


<!!-- Parameter Entities -->

<!!ENTITY % head.misc "SCRIPT|STYLE|META|LINK|OBJECT" -- repeatable head elements -->

<!!ENTITY % heading "H1|H2|H3|H4|H5|H6">

<!!ENTITY % list "UL | OL">

<!!ENTITY % preformatted "PRE">


<!!--================ Character mnemonic entities =========================-->

<!!ENTITY % HTMLlat1 PUBLIC
   "-//W3C//ENTITIES Latin1//EN//HTML"
   "HTMLlat1.ent">
%HTMLlat1;

<!!ENTITY % HTMLsymbol PUBLIC
   "-//W3C//ENTITIES Symbols//EN//HTML"
   "HTMLsymbol.ent">
%HTMLsymbol;

<!!ENTITY % HTMLspecial PUBLIC
   "-//W3C//ENTITIES Special//EN//HTML"
   "HTMLspecial.ent">
%HTMLspecial;
<!!--=================== Generic Attributes ===============================-->

<!!ENTITY % coreattrs
 "id          ID             #IMPLIED  -- document-wide unique id --
  class       CDATA          #IMPLIED  -- space-separated list of classes --
  style       %StyleSheet;   #IMPLIED  -- associated style info --
  title       %Text;         #IMPLIED  -- advisory title --"
  >

<!!ENTITY % i18n
 "lang        %LanguageCode; #IMPLIED  -- language code --
  dir         (ltr|rtl)      #IMPLIED  -- direction for weak/neutral text --"
  >

<!!ENTITY % events
 "onclick     %Script;       #IMPLIED  -- a pointer button was clicked --
  ondblclick  %Script;       #IMPLIED  -- a pointer button was double clicked--
  onmousedown %Script;       #IMPLIED  -- a pointer button was pressed down --
  onmouseup   %Script;       #IMPLIED  -- a pointer button was released --
  onmouseover %Script;       #IMPLIED  -- a pointer was moved onto --
  onmousemove %Script;       #IMPLIED  -- a pointer was moved within --
  onmouseout  %Script;       #IMPLIED  -- a pointer was moved away --
  onkeypress  %Script;       #IMPLIED  -- a key was pressed and released --
  onkeydown   %Script;       #IMPLIED  -- a key was pressed down --
  onkeyup     %Script;       #IMPLIED  -- a key was released --"
  >

<!!-- Reserved Feature Switch -->
<!!ENTITY % HTML.Reserved "IGNORE">

<!!-- The following attributes are reserved for possible future use -->
<!![ %HTML.Reserved; [
<!!ENTITY % reserved
 "datasrc     %URI;          #IMPLIED  -- a single or tabular Data Source --
  datafld     CDATA          #IMPLIED  -- the property or column name --
  dataformatas (plaintext|html) plaintext -- text or html --"
  >
]]>

<!!ENTITY % reserved "">

<!!ENTITY % attrs "%coreattrs; %i18n; %events;">


<!!--=================== Text Markup ======================================-->

<!!ENTITY % fontstyle
 "TT | I | B | BIG | SMALL">

<!!ENTITY % phrase "EM | STRONG | DFN | CODE |
                   SAMP | KBD | VAR | CITE | ABBR | ACRONYM" >

<!!ENTITY % special
   "A | IMG | OBJECT | BR | SCRIPT | MAP | Q | SUB | SUP | SPAN | BDO">

<!!ENTITY % formctrl "INPUT | SELECT | TEXTAREA | LABEL | BUTTON">

<!!-- %inline; covers inline or "text-level" elements -->
<!!ENTITY % inline "#PCDATA | %fontstyle; | %phrase; | %special; | %formctrl;">

<!!ELEMENT (%fontstyle;|%phrase;) - - (%inline;)*>
<!!ATTLIST (%fontstyle;|%phrase;)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT (SUB|SUP) - - (%inline;)*    -- subscript, superscript -->
<!!ATTLIST (SUB|SUP)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT SPAN - - (%inline;)*         -- generic language/style container -->
<!!ATTLIST SPAN
  %attrs;                              -- %coreattrs, %i18n, %events --
  %reserved;			       -- reserved for possible future use --
  >

<!!ELEMENT BDO - - (%inline;)*          -- I18N BiDi over-ride -->
<!!ATTLIST BDO
  %coreattrs;                          -- id, class, style, title --
  lang        %LanguageCode; #IMPLIED  -- language code --
  dir         (ltr|rtl)      #REQUIRED -- directionality --
  >


<!!ELEMENT BR - O EMPTY                 -- forced line break -->
<!!ATTLIST BR
  %coreattrs;                          -- id, class, style, title --
  >

<!!--================== HTML content models ===============================-->

<!!--
    HTML has two basic content models:

        %inline;     character level elements and text strings
        %block;      block-like elements e.g. paragraphs and lists
-->

<!!ENTITY % block
     "P | %heading; | %list; | %preformatted; | DL | DIV | NOSCRIPT |
      BLOCKQUOTE | FORM | HR | TABLE | FIELDSET | ADDRESS">

<!!ENTITY % flow "%block; | %inline;">

<!!--=================== Document Body ====================================-->

<!!ELEMENT BODY O O (%block;|SCRIPT)+ +(INS|DEL) -- document body -->
<!!ATTLIST BODY
  %attrs;                              -- %coreattrs, %i18n, %events --
  onload          %Script;   #IMPLIED  -- the document has been loaded --
  onunload        %Script;   #IMPLIED  -- the document has been removed --
  >

<!!ELEMENT ADDRESS - - (%inline;)* -- information on author -->
<!!ATTLIST ADDRESS
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT DIV - - (%flow;)*            -- generic language/style container -->
<!!ATTLIST DIV
  %attrs;                              -- %coreattrs, %i18n, %events --
  %reserved;                           -- reserved for possible future use --
  >


<!!--================== The Anchor Element ================================-->

<!!ENTITY % Shape "(rect|circle|poly|default)">
<!!ENTITY % Coords "CDATA" -- comma-separated list of lengths -->

<!!ELEMENT A - - (%inline;)* -(A)       -- anchor -->
<!!ATTLIST A
  %attrs;                              -- %coreattrs, %i18n, %events --
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  type        %ContentType;  #IMPLIED  -- advisory content type --
  name        CDATA          #IMPLIED  -- named link end --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  hreflang    %LanguageCode; #IMPLIED  -- language code --
  rel         %LinkTypes;    #IMPLIED  -- forward link types --
  rev         %LinkTypes;    #IMPLIED  -- reverse link types --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  shape       %Shape;        rect      -- for use with client-side image maps --
  coords      %Coords;       #IMPLIED  -- for use with client-side image maps --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!--================== Client-side image maps ============================-->

<!!-- These can be placed in the same document or grouped in a
     separate document although this isn''t yet widely supported -->

<!!ELEMENT MAP - - ((%block;) | AREA)+ -- client-side image map -->
<!!ATTLIST MAP
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #REQUIRED -- for reference by usemap --
  >

<!!ELEMENT AREA - O EMPTY               -- client-side image map area -->
<!!ATTLIST AREA
  %attrs;                              -- %coreattrs, %i18n, %events --
  shape       %Shape;        rect      -- controls interpretation of coords --
  coords      %Coords;       #IMPLIED  -- comma-separated list of lengths --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  nohref      (nohref)       #IMPLIED  -- this region has no action --
  alt         %Text;         #REQUIRED -- short description --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!--================== The LINK Element ==================================-->

<!!--
  Relationship values can be used in principle:

   a) for document specific toolbars/menus when used
      with the LINK element in document head e.g.
        start, contents, previous, next, index, end, help
   b) to link to a separate style sheet (rel=stylesheet)
   c) to make a link to a script (rel=script)
   d) by stylesheets to control how collections of
      html nodes are rendered into printed documents
   e) to make a link to a printable version of this document
      e.g. a postscript or pdf version (rel=alternate media=print)
-->

<!!ELEMENT LINK - O EMPTY               -- a media-independent link -->
<!!ATTLIST LINK
  %attrs;                              -- %coreattrs, %i18n, %events --
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  hreflang    %LanguageCode; #IMPLIED  -- language code --
  type        %ContentType;  #IMPLIED  -- advisory content type --
  rel         %LinkTypes;    #IMPLIED  -- forward link types --
  rev         %LinkTypes;    #IMPLIED  -- reverse link types --
  media       %MediaDesc;    #IMPLIED  -- for rendering on these media --
  >


<!!--=================== Images ===========================================-->

<!!-- Length defined in strict DTD for cellpadding/cellspacing -->
<!!ENTITY % Length "CDATA" -- nn for pixels or nn% for percentage length -->
<!!ENTITY % MultiLength "CDATA" -- pixel, percentage, or relative -->

<!![ %HTML.Frameset; [
<!!ENTITY % MultiLengths "CDATA" -- comma-separated list of MultiLength -->
]]>

<!!ENTITY % Pixels "CDATA" -- integer representing length in pixels -->


<!!-- To avoid problems with text-only UAs as well as 
   to make image content understandable and navigable 
   to users of non-visual UAs, you need to provide
   a description with ALT, and avoid server-side image maps -->
<!!ELEMENT IMG - O EMPTY                -- Embedded image -->
<!!ATTLIST IMG
  %attrs;                              -- %coreattrs, %i18n, %events --
  src         %URI;          #REQUIRED -- URI of image to embed --
  alt         %Text;         #REQUIRED -- short description --
  longdesc    %URI;          #IMPLIED  -- link to long description
                                          (complements alt) --
  name        CDATA          #IMPLIED  -- name of image for scripting --
  height      %Length;       #IMPLIED  -- override height --
  width       %Length;       #IMPLIED  -- override width --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  ismap       (ismap)        #IMPLIED  -- use server-side image map --
  >

<!!-- USEMAP points to a MAP element which may be in this document
  or an external document, although the latter is not widely supported -->

<!!--==================== OBJECT ======================================-->
<!!--
  OBJECT is used to embed objects as part of HTML pages 
  PARAM elements should precede other content. SGML mixed content
  model technicality precludes specifying this formally ...
-->

<!!ELEMENT OBJECT - - (PARAM | %flow;)*
 -- generic embedded object -->
<!!ATTLIST OBJECT
  %attrs;                              -- %coreattrs, %i18n, %events --
  declare     (declare)      #IMPLIED  -- declare but don''t instantiate flag --
  classid     %URI;          #IMPLIED  -- identifies an implementation --
  codebase    %URI;          #IMPLIED  -- base URI for classid, data, archive--
  data        %URI;          #IMPLIED  -- reference to object''s data --
  type        %ContentType;  #IMPLIED  -- content type for data --
  codetype    %ContentType;  #IMPLIED  -- content type for code --
  archive     CDATA          #IMPLIED  -- space-separated list of URIs --
  standby     %Text;         #IMPLIED  -- message to show while loading --
  height      %Length;       #IMPLIED  -- override height --
  width       %Length;       #IMPLIED  -- override width --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  name        CDATA          #IMPLIED  -- submit as part of form --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT PARAM - O EMPTY              -- named property value -->
<!!ATTLIST PARAM
  id          ID             #IMPLIED  -- document-wide unique id --
  name        CDATA          #REQUIRED -- property name --
  value       CDATA          #IMPLIED  -- property value --
  valuetype   (DATA|REF|OBJECT) DATA   -- How to interpret value --
  type        %ContentType;  #IMPLIED  -- content type for value
                                          when valuetype=ref --
  >


<!!--=================== Horizontal Rule ==================================-->

<!!ELEMENT HR - O EMPTY -- horizontal rule -->
<!!ATTLIST HR
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--=================== Paragraphs =======================================-->

<!!ELEMENT P - O (%inline;)*            -- paragraph -->
<!!ATTLIST P
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--=================== Headings =========================================-->

<!!--
  There are six levels of headings from H1 (the most important)
  to H6 (the least important).
-->

<!!ELEMENT (%heading;)  - - (%inline;)* -- heading -->
<!!ATTLIST (%heading;)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--=================== Preformatted Text ================================-->

<!!-- excludes markup for images and changes in font size -->
<!!ENTITY % pre.exclusion "IMG|OBJECT|BIG|SMALL|SUB|SUP">

<!!ELEMENT PRE - - (%inline;)* -(%pre.exclusion;) -- preformatted text -->
<!!ATTLIST PRE
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--===================== Inline Quotes ==================================-->

<!!ELEMENT Q - - (%inline;)*            -- short inline quotation -->
<!!ATTLIST Q
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- URI for source document or msg --
  >

<!!--=================== Block-like Quotes ================================-->

<!!ELEMENT BLOCKQUOTE - - (%block;|SCRIPT)+ -- long quotation -->
<!!ATTLIST BLOCKQUOTE
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- URI for source document or msg --
  >

<!!--=================== Inserted/Deleted Text ============================-->


<!!-- INS/DEL are handled by inclusion on BODY -->
<!!ELEMENT (INS|DEL) - - (%flow;)*      -- inserted text, deleted text -->
<!!ATTLIST (INS|DEL)
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- info on reason for change --
  datetime    %Datetime;     #IMPLIED  -- date and time of change --
  >

<!!--=================== Lists ============================================-->

<!!-- definition lists - DT for term, DD for its definition -->

<!!ELEMENT DL - - (DT|DD)+              -- definition list -->
<!!ATTLIST DL
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT DT - O (%inline;)*           -- definition term -->
<!!ELEMENT DD - O (%flow;)*             -- definition description -->
<!!ATTLIST (DT|DD)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >


<!!ELEMENT OL - - (LI)+                 -- ordered list -->
<!!ATTLIST OL
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!-- Unordered Lists (UL) bullet styles -->
<!!ELEMENT UL - - (LI)+                 -- unordered list -->
<!!ATTLIST UL
  %attrs;                              -- %coreattrs, %i18n, %events --
  >



<!!ELEMENT LI - O (%flow;)*             -- list item -->
<!!ATTLIST LI
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--================ Forms ===============================================-->
<!!ELEMENT FORM - - (%block;|SCRIPT)+ -(FORM) -- interactive form -->
<!!ATTLIST FORM
  %attrs;                              -- %coreattrs, %i18n, %events --
  action      %URI;          #REQUIRED -- server-side form handler --
  method      (get|post)     GET       -- HTTP method used to submit the form--
  enctype     %ContentType;  "application/x-www-form-urlencoded"
  accept      %ContentTypes; #IMPLIED  -- list of MIME types for file upload --
  name        CDATA          #IMPLIED  -- name of form for scripting --
  onsubmit    %Script;       #IMPLIED  -- the form was submitted --
  onreset     %Script;       #IMPLIED  -- the form was reset --
  accept-charset %Charsets;  #IMPLIED  -- list of supported charsets --
  >

<!!-- Each label must not contain more than ONE field -->
<!!ELEMENT LABEL - - (%inline;)* -(LABEL) -- form field label text -->
<!!ATTLIST LABEL
  %attrs;                              -- %coreattrs, %i18n, %events --
  for         IDREF          #IMPLIED  -- matches field ID value --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!ENTITY % InputType
  "(text | password | checkbox |
    radio | submit | reset |
    file | hidden | image | button)"
   >

<!!-- attribute name required for all but submit and reset -->
<!!ELEMENT INPUT - O EMPTY              -- form control -->
<!!ATTLIST INPUT
  %attrs;                              -- %coreattrs, %i18n, %events --
  type        %InputType;    TEXT      -- what kind of widget is needed --
  name        CDATA          #IMPLIED  -- submit as part of form --
  value       CDATA          #IMPLIED  -- Specify for radio buttons and checkboxes --
  checked     (checked)      #IMPLIED  -- for radio buttons and check boxes --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  readonly    (readonly)     #IMPLIED  -- for text and passwd --
  size        CDATA          #IMPLIED  -- specific to each type of field --
  maxlength   NUMBER         #IMPLIED  -- max chars for text fields --
  src         %URI;          #IMPLIED  -- for fields with images --
  alt         CDATA          #IMPLIED  -- short description --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  ismap       (ismap)        #IMPLIED  -- use server-side image map --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onselect    %Script;       #IMPLIED  -- some text was selected --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  accept      %ContentTypes; #IMPLIED  -- list of MIME types for file upload --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT SELECT - - (OPTGROUP|OPTION)+ -- option selector -->
<!!ATTLIST SELECT
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED  -- field name --
  size        NUMBER         #IMPLIED  -- rows visible --
  multiple    (multiple)     #IMPLIED  -- default is single selection --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT OPTGROUP - - (OPTION)+ -- option group -->
<!!ATTLIST OPTGROUP
  %attrs;                              -- %coreattrs, %i18n, %events --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  label       %Text;         #REQUIRED -- for use in hierarchical menus --
  >

<!!ELEMENT OPTION - O (#PCDATA)         -- selectable choice -->
<!!ATTLIST OPTION
  %attrs;                              -- %coreattrs, %i18n, %events --
  selected    (selected)     #IMPLIED
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  label       %Text;         #IMPLIED  -- for use in hierarchical menus --
  value       CDATA          #IMPLIED  -- defaults to element content --
  >

<!!ELEMENT TEXTAREA - - (#PCDATA)       -- multi-line text field -->
<!!ATTLIST TEXTAREA
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED
  rows        NUMBER         #REQUIRED
  cols        NUMBER         #REQUIRED
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  readonly    (readonly)     #IMPLIED
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onselect    %Script;       #IMPLIED  -- some text was selected --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  %reserved;                           -- reserved for possible future use --
  >

<!!--
  #PCDATA is to solve the mixed content problem,
  per specification only whitespace is allowed there!!
 -->
<!!ELEMENT FIELDSET - - (#PCDATA,LEGEND,(%flow;)*) -- form control group -->
<!!ATTLIST FIELDSET
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT LEGEND - - (%inline;)*       -- fieldset legend -->

<!!ATTLIST LEGEND
  %attrs;                              -- %coreattrs, %i18n, %events --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  >

<!!ELEMENT BUTTON - -
     (%flow;)* -(A|%formctrl;|FORM|FIELDSET)
     -- push button -->
<!!ATTLIST BUTTON
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED
  value       CDATA          #IMPLIED  -- sent to server when submitted --
  type        (button|submit|reset) submit -- for use as form button --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  %reserved;                           -- reserved for possible future use --
  >

<!!--======================= Tables =======================================-->

<!!-- IETF HTML table standard, see [RFC1942] -->

<!!--
 The BORDER attribute sets the thickness of the frame around the
 table. The default units are screen pixels.

 The FRAME attribute specifies which parts of the frame around
 the table should be rendered. The values are not the same as
 CALS to avoid a name clash with the VALIGN attribute.

 The value "border" is included for backwards compatibility with
 <TABLE BORDER> which yields frame=border and border=implied
 For <TABLE BORDER=1> you get border=1 and frame=implied. In this
 case, it is appropriate to treat this as frame=border for backwards
 compatibility with deployed browsers.
-->
<!!ENTITY % TFrame "(void|above|below|hsides|lhs|rhs|vsides|box|border)">

<!!--
 The RULES attribute defines which rules to draw between cells:

 If RULES is absent then assume:
     "none" if BORDER is absent or BORDER=0 otherwise "all"
-->

<!!ENTITY % TRules "(none | groups | rows | cols | all)">
  
<!!-- horizontal placement of table relative to document -->
<!!ENTITY % TAlign "(left|center|right)">

<!!-- horizontal alignment attributes for cell contents -->
<!!ENTITY % cellhalign
  "align      (left|center|right|justify|char) #IMPLIED
   char       %Character;    #IMPLIED  -- alignment char, e.g. char='':'' --
   charoff    %Length;       #IMPLIED  -- offset for alignment char --"
  >

<!!-- vertical alignment attributes for cell contents -->
<!!ENTITY % cellvalign
  "valign     (top|middle|bottom|baseline) #IMPLIED"
  >

<!!ELEMENT TABLE - -
     (CAPTION?, (COL*|COLGROUP*), THEAD?, TFOOT?, (TBODY|TR+))>
<!!ELEMENT CAPTION  - - (%inline;)*     -- table caption -->
<!!ELEMENT THEAD    - O (TR)+           -- table header -->
<!!ELEMENT TFOOT    - O (TR)+           -- table footer -->
<!!ELEMENT TBODY    O O (TR)+           -- table body -->
<!!ELEMENT COLGROUP - O (COL)*          -- table column group -->
<!!ELEMENT COL      - O EMPTY           -- table column -->
<!!ELEMENT TR       - O (TH|TD)+        -- table row -->
<!!ELEMENT (TH|TD)  - O (%flow;)*       -- table header cell, table data cell-->

<!!ATTLIST TABLE                        -- table element --
  %attrs;                              -- %coreattrs, %i18n, %events --
  summary     %Text;         #IMPLIED  -- purpose/structure for speech output--
  width       %Length;       #IMPLIED  -- table width --
  border      %Pixels;       #IMPLIED  -- controls frame width around table --
  frame       %TFrame;       #IMPLIED  -- which parts of frame to render --
  rules       %TRules;       #IMPLIED  -- rulings between rows and cols --
  cellspacing %Length;       #IMPLIED  -- spacing between cells --
  cellpadding %Length;       #IMPLIED  -- spacing within cells --
  %reserved;                           -- reserved for possible future use --
  datapagesize CDATA         #IMPLIED  -- reserved for possible future use --
  >


<!!ATTLIST CAPTION
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--
COLGROUP groups a set of COL elements. It allows you to group
several semantically related columns together.
-->
<!!ATTLIST COLGROUP
  %attrs;                              -- %coreattrs, %i18n, %events --
  span        NUMBER         1         -- default number of columns in group --
  width       %MultiLength;  #IMPLIED  -- default width for enclosed COLs --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!--
 COL elements define the alignment properties for cells in
 one or more columns.

 The WIDTH attribute specifies the width of the columns, e.g.

     width=64        width in screen pixels
     width=0.5*      relative width of 0.5

 The SPAN attribute causes the attributes of one
 COL element to apply to more than one column.
-->
<!!ATTLIST COL                          -- column groups and properties --
  %attrs;                              -- %coreattrs, %i18n, %events --
  span        NUMBER         1         -- COL attributes affect N columns --
  width       %MultiLength;  #IMPLIED  -- column width specification --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!--
    Use THEAD to duplicate headers when breaking table
    across page boundaries, or for static headers when
    TBODY sections are rendered in scrolling panel.

    Use TFOOT to duplicate footers when breaking table
    across page boundaries, or for static footers when
    TBODY sections are rendered in scrolling panel.

    Use multiple TBODY sections when rules are needed
    between groups of table rows.
-->
<!!ATTLIST (THEAD|TBODY|TFOOT)          -- table section --
  %attrs;                              -- %coreattrs, %i18n, %events --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!ATTLIST TR                           -- table row --
  %attrs;                              -- %coreattrs, %i18n, %events --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >


<!!-- Scope is simpler than headers attribute for common tables -->
<!!ENTITY % Scope "(row|col|rowgroup|colgroup)">

<!!-- TH is for headers, TD for data, but for cells acting as both use TD -->
<!!ATTLIST (TH|TD)                      -- header or data cell --
  %attrs;                              -- %coreattrs, %i18n, %events --
  abbr        %Text;         #IMPLIED  -- abbreviation for header cell --
  axis        CDATA          #IMPLIED  -- comma-separated list of related headers--
  headers     IDREFS         #IMPLIED  -- list of id''s for header cells --
  scope       %Scope;        #IMPLIED  -- scope covered by header cells --
  rowspan     NUMBER         1         -- number of rows spanned by cell --
  colspan     NUMBER         1         -- number of cols spanned by cell --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >


<!!--================ Document Head =======================================-->
<!!-- %head.misc; defined earlier on as "SCRIPT|STYLE|META|LINK|OBJECT" -->
<!!ENTITY % head.content "TITLE & BASE?">

<!!ELEMENT HEAD O O (%head.content;) +(%head.misc;) -- document head -->
<!!ATTLIST HEAD
  %i18n;                               -- lang, dir --
  profile     %URI;          #IMPLIED  -- named dictionary of meta info --
  >

<!!-- The TITLE element is not considered part of the flow of text.
       It should be displayed, for example as the page header or
       window title. Exactly one title is required per document.
    -->
<!!ELEMENT TITLE - - (#PCDATA) -(%head.misc;) -- document title -->
<!!ATTLIST TITLE %i18n>


<!!ELEMENT BASE - O EMPTY               -- document base URI -->
<!!ATTLIST BASE
  href        %URI;          #REQUIRED -- URI that acts as base URI --
  >

<!!ELEMENT META - O EMPTY               -- generic metainformation -->
<!!ATTLIST META
  %i18n;                               -- lang, dir, for use with content --
  http-equiv  NAME           #IMPLIED  -- HTTP response header name  --
  name        NAME           #IMPLIED  -- metainformation name --
  content     CDATA          #REQUIRED -- associated information --
  scheme      CDATA          #IMPLIED  -- select form of content --
  >

<!!ELEMENT STYLE - - %StyleSheet        -- style info -->
<!!ATTLIST STYLE
  %i18n;                               -- lang, dir, for use with title --
  type        %ContentType;  #REQUIRED -- content type of style language --
  media       %MediaDesc;    #IMPLIED  -- designed for use with these media --
  title       %Text;         #IMPLIED  -- advisory title --
  >

<!!ELEMENT SCRIPT - - %Script;          -- script statements -->
<!!ATTLIST SCRIPT
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  type        %ContentType;  #REQUIRED -- content type of script language --
  src         %URI;          #IMPLIED  -- URI for an external script --
  defer       (defer)        #IMPLIED  -- UA may defer execution of script --
  event       CDATA          #IMPLIED  -- reserved for possible future use --
  for         %URI;          #IMPLIED  -- reserved for possible future use --
  >

<!!ELEMENT NOSCRIPT - - (%block;)+
  -- alternate content container for non script-based rendering -->
<!!ATTLIST NOSCRIPT
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--================ Document Structure ==================================-->
<!!ENTITY % html.content "HEAD, FRAMESET">

<!!ELEMENT HTML O O (%html.content;)    -- document root element -->
<!!ATTLIST HTML
  %i18n;                               -- lang, dir --
  >
'! !

!Html401Frameset methodsFor: 'as yet unclassified' stamp: 'tb 12/19/2005 21:50'!
initialize

	super initialize.
	self parse.! !

HtmlDTD subclass: #Html401Strict
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!Html401Strict class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:54'!
defaultUrl

	^'http://www.w3.org/TR/REC-html40/strict.dtd' asUrl! !

!Html401Strict class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:32'!
docType

	^'<!!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN" "http://www.w3.org/TR/html4/strict.dtd">'! !

!Html401Strict methodsFor: 'resources' stamp: 'tb 1/14/2006 12:41'!
dtd
	^ '<!!--
    This is HTML 4.01 Strict DTD, which excludes the presentation 
    attributes and elements that W3C expects to phase out as 
    support for style sheets matures. Authors should use the Strict
    DTD when possible, but may use the Transitional DTD when support
    for presentation attribute and elements is required.
    
    HTML 4 includes mechanisms for style sheets, scripting,
    embedding objects, improved support for right to left and mixed
    direction text, and enhancements to forms for improved
    accessibility for people with disabilities.

          Draft: $Date: 1999/12/24 23:37:48 $

          Authors:
              Dave Raggett <dsr@w3.org>
              Arnaud Le Hors <lehors@w3.org>
              Ian Jacobs <ij@w3.org>

    Further information about HTML 4.01 is available at:

        http://www.w3.org/TR/1999/REC-html401-19991224


    The HTML 4.01 specification includes additional
    syntactic constraints that cannot be expressed within
    the DTDs.

-->
<!!--
    Typical usage:

    <!!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
            "http://www.w3.org/TR/html4/strict.dtd">
    <html>
    <head>
    ...
    </head>
    <body>
    ...
    </body>
    </html>

    The URI used as a system identifier with the public identifier allows
    the user agent to download the DTD and entity sets as needed.

    The FPI for the Transitional HTML 4.01 DTD is:

        "-//W3C//DTD HTML 4.01 Transitional//EN"

    This version of the transitional DTD is:

        http://www.w3.org/TR/1999/REC-html401-19991224/loose.dtd

    If you are writing a document that includes frames, use 
    the following FPI:

        "-//W3C//DTD HTML 4.01 Frameset//EN"

    This version of the frameset DTD is:

        http://www.w3.org/TR/1999/REC-html401-19991224/frameset.dtd

    Use the following (relative) URIs to refer to 
    the DTDs and entity definitions of this specification:

    "strict.dtd"
    "loose.dtd"
    "frameset.dtd"
    "HTMLlat1.ent"
    "HTMLsymbol.ent"
    "HTMLspecial.ent"

-->

<!!--================== Imported Names ====================================-->
<!!-- Feature Switch for frameset documents -->
<!!ENTITY % HTML.Frameset "IGNORE">

<!!ENTITY % ContentType "CDATA"
    -- media type, as per [RFC2045]
    -->

<!!ENTITY % ContentTypes "CDATA"
    -- comma-separated list of media types, as per [RFC2045]
    -->

<!!ENTITY % Charset "CDATA"
    -- a character encoding, as per [RFC2045]
    -->

<!!ENTITY % Charsets "CDATA"
    -- a space-separated list of character encodings, as per [RFC2045]
    -->

<!!ENTITY % LanguageCode "NAME"
    -- a language code, as per [RFC1766]
    -->

<!!ENTITY % Character "CDATA"
    -- a single character from [ISO10646] 
    -->

<!!ENTITY % LinkTypes "CDATA"
    -- space-separated list of link types
    -->

<!!ENTITY % MediaDesc "CDATA"
    -- single or comma-separated list of media descriptors
    -->

<!!ENTITY % URI "CDATA"
    -- a Uniform Resource Identifier,
       see [URI]
    -->

<!!ENTITY % Datetime "CDATA" -- date and time information. ISO date format -->


<!!ENTITY % Script "CDATA" -- script expression -->

<!!ENTITY % StyleSheet "CDATA" -- style sheet data -->



<!!ENTITY % Text "CDATA">


<!!-- Parameter Entities -->

<!!ENTITY % head.misc "SCRIPT|STYLE|META|LINK|OBJECT" -- repeatable head elements -->

<!!ENTITY % heading "H1|H2|H3|H4|H5|H6">

<!!ENTITY % list "UL | OL">

<!!ENTITY % preformatted "PRE">


<!!--================ Character mnemonic entities =========================-->

<!!ENTITY % HTMLlat1 PUBLIC
   "-//W3C//ENTITIES Latin1//EN//HTML"
   "HTMLlat1.ent">
%HTMLlat1;

<!!ENTITY % HTMLsymbol PUBLIC
   "-//W3C//ENTITIES Symbols//EN//HTML"
   "HTMLsymbol.ent">
%HTMLsymbol;

<!!ENTITY % HTMLspecial PUBLIC
   "-//W3C//ENTITIES Special//EN//HTML"
   "HTMLspecial.ent">
%HTMLspecial;
<!!--=================== Generic Attributes ===============================-->

<!!ENTITY % coreattrs
 "id          ID             #IMPLIED  -- document-wide unique id --
  class       CDATA          #IMPLIED  -- space-separated list of classes --
  style       %StyleSheet;   #IMPLIED  -- associated style info --
  title       %Text;         #IMPLIED  -- advisory title --"
  >

<!!ENTITY % i18n
 "lang        %LanguageCode; #IMPLIED  -- language code --
  dir         (ltr|rtl)      #IMPLIED  -- direction for weak/neutral text --"
  >

<!!ENTITY % events
 "onclick     %Script;       #IMPLIED  -- a pointer button was clicked --
  ondblclick  %Script;       #IMPLIED  -- a pointer button was double clicked--
  onmousedown %Script;       #IMPLIED  -- a pointer button was pressed down --
  onmouseup   %Script;       #IMPLIED  -- a pointer button was released --
  onmouseover %Script;       #IMPLIED  -- a pointer was moved onto --
  onmousemove %Script;       #IMPLIED  -- a pointer was moved within --
  onmouseout  %Script;       #IMPLIED  -- a pointer was moved away --
  onkeypress  %Script;       #IMPLIED  -- a key was pressed and released --
  onkeydown   %Script;       #IMPLIED  -- a key was pressed down --
  onkeyup     %Script;       #IMPLIED  -- a key was released --"
  >

<!!-- Reserved Feature Switch -->
<!!ENTITY % HTML.Reserved "IGNORE">

<!!-- The following attributes are reserved for possible future use -->
<!![ %HTML.Reserved; [
<!!ENTITY % reserved
 "datasrc     %URI;          #IMPLIED  -- a single or tabular Data Source --
  datafld     CDATA          #IMPLIED  -- the property or column name --
  dataformatas (plaintext|html) plaintext -- text or html --"
  >
]]>

<!!ENTITY % reserved "">

<!!ENTITY % attrs "%coreattrs; %i18n; %events;">


<!!--=================== Text Markup ======================================-->

<!!ENTITY % fontstyle
 "TT | I | B | BIG | SMALL">

<!!ENTITY % phrase "EM | STRONG | DFN | CODE |
                   SAMP | KBD | VAR | CITE | ABBR | ACRONYM" >

<!!ENTITY % special
   "A | IMG | OBJECT | BR | SCRIPT | MAP | Q | SUB | SUP | SPAN | BDO">

<!!ENTITY % formctrl "INPUT | SELECT | TEXTAREA | LABEL | BUTTON">

<!!-- %inline; covers inline or "text-level" elements -->
<!!ENTITY % inline "#PCDATA | %fontstyle; | %phrase; | %special; | %formctrl;">

<!!ELEMENT (%fontstyle;|%phrase;) - - (%inline;)*>
<!!ATTLIST (%fontstyle;|%phrase;)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT (SUB|SUP) - - (%inline;)*    -- subscript, superscript -->
<!!ATTLIST (SUB|SUP)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT SPAN - - (%inline;)*         -- generic language/style container -->
<!!ATTLIST SPAN
  %attrs;                              -- %coreattrs, %i18n, %events --
  %reserved;			       -- reserved for possible future use --
  >

<!!ELEMENT BDO - - (%inline;)*          -- I18N BiDi over-ride -->
<!!ATTLIST BDO
  %coreattrs;                          -- id, class, style, title --
  lang        %LanguageCode; #IMPLIED  -- language code --
  dir         (ltr|rtl)      #REQUIRED -- directionality --
  >


<!!ELEMENT BR - O EMPTY                 -- forced line break -->
<!!ATTLIST BR
  %coreattrs;                          -- id, class, style, title --
  >

<!!--================== HTML content models ===============================-->

<!!--
    HTML has two basic content models:

        %inline;     character level elements and text strings
        %block;      block-like elements e.g. paragraphs and lists
-->

<!!ENTITY % block
     "P | %heading; | %list; | %preformatted; | DL | DIV | NOSCRIPT |
      BLOCKQUOTE | FORM | HR | TABLE | FIELDSET | ADDRESS">

<!!ENTITY % flow "%block; | %inline;">

<!!--=================== Document Body ====================================-->

<!!ELEMENT BODY O O (%block;|SCRIPT)+ +(INS|DEL) -- document body -->
<!!ATTLIST BODY
  %attrs;                              -- %coreattrs, %i18n, %events --
  onload          %Script;   #IMPLIED  -- the document has been loaded --
  onunload        %Script;   #IMPLIED  -- the document has been removed --
  >

<!!ELEMENT ADDRESS - - (%inline;)* -- information on author -->
<!!ATTLIST ADDRESS
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT DIV - - (%flow;)*            -- generic language/style container -->
<!!ATTLIST DIV
  %attrs;                              -- %coreattrs, %i18n, %events --
  %reserved;                           -- reserved for possible future use --
  >


<!!--================== The Anchor Element ================================-->

<!!ENTITY % Shape "(rect|circle|poly|default)">
<!!ENTITY % Coords "CDATA" -- comma-separated list of lengths -->

<!!ELEMENT A - - (%inline;)* -(A)       -- anchor -->
<!!ATTLIST A
  %attrs;                              -- %coreattrs, %i18n, %events --
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  type        %ContentType;  #IMPLIED  -- advisory content type --
  name        CDATA          #IMPLIED  -- named link end --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  hreflang    %LanguageCode; #IMPLIED  -- language code --
  rel         %LinkTypes;    #IMPLIED  -- forward link types --
  rev         %LinkTypes;    #IMPLIED  -- reverse link types --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  shape       %Shape;        rect      -- for use with client-side image maps --
  coords      %Coords;       #IMPLIED  -- for use with client-side image maps --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!--================== Client-side image maps ============================-->

<!!-- These can be placed in the same document or grouped in a
     separate document although this isn''t yet widely supported -->

<!!ELEMENT MAP - - ((%block;) | AREA)+ -- client-side image map -->
<!!ATTLIST MAP
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #REQUIRED -- for reference by usemap --
  >

<!!ELEMENT AREA - O EMPTY               -- client-side image map area -->
<!!ATTLIST AREA
  %attrs;                              -- %coreattrs, %i18n, %events --
  shape       %Shape;        rect      -- controls interpretation of coords --
  coords      %Coords;       #IMPLIED  -- comma-separated list of lengths --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  nohref      (nohref)       #IMPLIED  -- this region has no action --
  alt         %Text;         #REQUIRED -- short description --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!--================== The LINK Element ==================================-->

<!!--
  Relationship values can be used in principle:

   a) for document specific toolbars/menus when used
      with the LINK element in document head e.g.
        start, contents, previous, next, index, end, help
   b) to link to a separate style sheet (rel=stylesheet)
   c) to make a link to a script (rel=script)
   d) by stylesheets to control how collections of
      html nodes are rendered into printed documents
   e) to make a link to a printable version of this document
      e.g. a postscript or pdf version (rel=alternate media=print)
-->

<!!ELEMENT LINK - O EMPTY               -- a media-independent link -->
<!!ATTLIST LINK
  %attrs;                              -- %coreattrs, %i18n, %events --
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  hreflang    %LanguageCode; #IMPLIED  -- language code --
  type        %ContentType;  #IMPLIED  -- advisory content type --
  rel         %LinkTypes;    #IMPLIED  -- forward link types --
  rev         %LinkTypes;    #IMPLIED  -- reverse link types --
  media       %MediaDesc;    #IMPLIED  -- for rendering on these media --
  >

<!!--=================== Images ===========================================-->

<!!-- Length defined in strict DTD for cellpadding/cellspacing -->
<!!ENTITY % Length "CDATA" -- nn for pixels or nn% for percentage length -->
<!!ENTITY % MultiLength "CDATA" -- pixel, percentage, or relative -->

<!![ %HTML.Frameset; [
<!!ENTITY % MultiLengths "CDATA" -- comma-separated list of MultiLength -->
]]>

<!!ENTITY % Pixels "CDATA" -- integer representing length in pixels -->


<!!-- To avoid problems with text-only UAs as well as 
   to make image content understandable and navigable 
   to users of non-visual UAs, you need to provide
   a description with ALT, and avoid server-side image maps -->
<!!ELEMENT IMG - O EMPTY                -- Embedded image -->
<!!ATTLIST IMG
  %attrs;                              -- %coreattrs, %i18n, %events --
  src         %URI;          #REQUIRED -- URI of image to embed --
  alt         %Text;         #REQUIRED -- short description --
  longdesc    %URI;          #IMPLIED  -- link to long description
                                          (complements alt) --
  name        CDATA          #IMPLIED  -- name of image for scripting --
  height      %Length;       #IMPLIED  -- override height --
  width       %Length;       #IMPLIED  -- override width --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  ismap       (ismap)        #IMPLIED  -- use server-side image map --
  >

<!!-- USEMAP points to a MAP element which may be in this document
  or an external document, although the latter is not widely supported -->

<!!--==================== OBJECT ======================================-->
<!!--
  OBJECT is used to embed objects as part of HTML pages 
  PARAM elements should precede other content. SGML mixed content
  model technicality precludes specifying this formally ...
-->

<!!ELEMENT OBJECT - - (PARAM | %flow;)*
 -- generic embedded object -->
<!!ATTLIST OBJECT
  %attrs;                              -- %coreattrs, %i18n, %events --
  declare     (declare)      #IMPLIED  -- declare but don''t instantiate flag --
  classid     %URI;          #IMPLIED  -- identifies an implementation --
  codebase    %URI;          #IMPLIED  -- base URI for classid, data, archive--
  data        %URI;          #IMPLIED  -- reference to object''s data --
  type        %ContentType;  #IMPLIED  -- content type for data --
  codetype    %ContentType;  #IMPLIED  -- content type for code --
  archive     CDATA          #IMPLIED  -- space-separated list of URIs --
  standby     %Text;         #IMPLIED  -- message to show while loading --
  height      %Length;       #IMPLIED  -- override height --
  width       %Length;       #IMPLIED  -- override width --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  name        CDATA          #IMPLIED  -- submit as part of form --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT PARAM - O EMPTY              -- named property value -->
<!!ATTLIST PARAM
  id          ID             #IMPLIED  -- document-wide unique id --
  name        CDATA          #REQUIRED -- property name --
  value       CDATA          #IMPLIED  -- property value --
  valuetype   (DATA|REF|OBJECT) DATA   -- How to interpret value --
  type        %ContentType;  #IMPLIED  -- content type for value
                                          when valuetype=ref --
  >


<!!--=================== Horizontal Rule ==================================-->

<!!ELEMENT HR - O EMPTY -- horizontal rule -->
<!!ATTLIST HR
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--=================== Paragraphs =======================================-->

<!!ELEMENT P - O (%inline;)*            -- paragraph -->
<!!ATTLIST P
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--=================== Headings =========================================-->

<!!--
  There are six levels of headings from H1 (the most important)
  to H6 (the least important).
-->

<!!ELEMENT (%heading;)  - - (%inline;)* -- heading -->
<!!ATTLIST (%heading;)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--=================== Preformatted Text ================================-->

<!!-- excludes markup for images and changes in font size -->
<!!ENTITY % pre.exclusion "IMG|OBJECT|BIG|SMALL|SUB|SUP">

<!!ELEMENT PRE - - (%inline;)* -(%pre.exclusion;) -- preformatted text -->
<!!ATTLIST PRE
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--===================== Inline Quotes ==================================-->

<!!ELEMENT Q - - (%inline;)*            -- short inline quotation -->
<!!ATTLIST Q
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- URI for source document or msg --
  >

<!!--=================== Block-like Quotes ================================-->

<!!ELEMENT BLOCKQUOTE - - (%block;|SCRIPT)+ -- long quotation -->
<!!ATTLIST BLOCKQUOTE
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- URI for source document or msg --
  >

<!!--=================== Inserted/Deleted Text ============================-->


<!!-- INS/DEL are handled by inclusion on BODY -->
<!!ELEMENT (INS|DEL) - - (%flow;)*      -- inserted text, deleted text -->
<!!ATTLIST (INS|DEL)
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- info on reason for change --
  datetime    %Datetime;     #IMPLIED  -- date and time of change --
  >

<!!--=================== Lists ============================================-->

<!!-- definition lists - DT for term, DD for its definition -->

<!!ELEMENT DL - - (DT|DD)+              -- definition list -->
<!!ATTLIST DL
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT DT - O (%inline;)*           -- definition term -->
<!!ELEMENT DD - O (%flow;)*             -- definition description -->
<!!ATTLIST (DT|DD)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >


<!!ELEMENT OL - - (LI)+                 -- ordered list -->
<!!ATTLIST OL
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!-- Unordered Lists (UL) bullet styles -->
<!!ELEMENT UL - - (LI)+                 -- unordered list -->
<!!ATTLIST UL
  %attrs;                              -- %coreattrs, %i18n, %events --
  >



<!!ELEMENT LI - O (%flow;)*             -- list item -->
<!!ATTLIST LI
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--================ Forms ===============================================-->
<!!ELEMENT FORM - - (%block;|SCRIPT)+ -(FORM) -- interactive form -->
<!!ATTLIST FORM
  %attrs;                              -- %coreattrs, %i18n, %events --
  action      %URI;          #REQUIRED -- server-side form handler --
  method      (get|post)     GET       -- HTTP method used to submit the form--
  enctype     %ContentType;  "application/x-www-form-urlencoded"
  accept      %ContentTypes; #IMPLIED  -- list of MIME types for file upload --
  name        CDATA          #IMPLIED  -- name of form for scripting --
  onsubmit    %Script;       #IMPLIED  -- the form was submitted --
  onreset     %Script;       #IMPLIED  -- the form was reset --
  accept-charset %Charsets;  #IMPLIED  -- list of supported charsets --
  >

<!!-- Each label must not contain more than ONE field -->
<!!ELEMENT LABEL - - (%inline;)* -(LABEL) -- form field label text -->
<!!ATTLIST LABEL
  %attrs;                              -- %coreattrs, %i18n, %events --
  for         IDREF          #IMPLIED  -- matches field ID value --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!ENTITY % InputType
  "(text | password | checkbox |
    radio | submit | reset |
    file | hidden | image | button)"
   >

<!!-- attribute name required for all but submit and reset -->
<!!ELEMENT INPUT - O EMPTY              -- form control -->
<!!ATTLIST INPUT
  %attrs;                              -- %coreattrs, %i18n, %events --
  type        %InputType;    TEXT      -- what kind of widget is needed --
  name        CDATA          #IMPLIED  -- submit as part of form --
  value       CDATA          #IMPLIED  -- Specify for radio buttons and checkboxes --
  checked     (checked)      #IMPLIED  -- for radio buttons and check boxes --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  readonly    (readonly)     #IMPLIED  -- for text and passwd --
  size        CDATA          #IMPLIED  -- specific to each type of field --
  maxlength   NUMBER         #IMPLIED  -- max chars for text fields --
  src         %URI;          #IMPLIED  -- for fields with images --
  alt         CDATA          #IMPLIED  -- short description --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  ismap       (ismap)        #IMPLIED  -- use server-side image map --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onselect    %Script;       #IMPLIED  -- some text was selected --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  accept      %ContentTypes; #IMPLIED  -- list of MIME types for file upload --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT SELECT - - (OPTGROUP|OPTION)+ -- option selector -->
<!!ATTLIST SELECT
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED  -- field name --
  size        NUMBER         #IMPLIED  -- rows visible --
  multiple    (multiple)     #IMPLIED  -- default is single selection --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT OPTGROUP - - (OPTION)+ -- option group -->
<!!ATTLIST OPTGROUP
  %attrs;                              -- %coreattrs, %i18n, %events --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  label       %Text;         #REQUIRED -- for use in hierarchical menus --
  >

<!!ELEMENT OPTION - O (#PCDATA)         -- selectable choice -->
<!!ATTLIST OPTION
  %attrs;                              -- %coreattrs, %i18n, %events --
  selected    (selected)     #IMPLIED
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  label       %Text;         #IMPLIED  -- for use in hierarchical menus --
  value       CDATA          #IMPLIED  -- defaults to element content --
  >

<!!ELEMENT TEXTAREA - - (#PCDATA)       -- multi-line text field -->
<!!ATTLIST TEXTAREA
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED
  rows        NUMBER         #REQUIRED
  cols        NUMBER         #REQUIRED
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  readonly    (readonly)     #IMPLIED
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onselect    %Script;       #IMPLIED  -- some text was selected --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  %reserved;                           -- reserved for possible future use --
  >

<!!--
  #PCDATA is to solve the mixed content problem,
  per specification only whitespace is allowed there!!
 -->
<!!ELEMENT FIELDSET - - (#PCDATA,LEGEND,(%flow;)*) -- form control group -->
<!!ATTLIST FIELDSET
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT LEGEND - - (%inline;)*       -- fieldset legend -->

<!!ATTLIST LEGEND
  %attrs;                              -- %coreattrs, %i18n, %events --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  >

<!!ELEMENT BUTTON - -
     (%flow;)* -(A|%formctrl;|FORM|FIELDSET)
     -- push button -->
<!!ATTLIST BUTTON
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED
  value       CDATA          #IMPLIED  -- sent to server when submitted --
  type        (button|submit|reset) submit -- for use as form button --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  %reserved;                           -- reserved for possible future use --
  >

<!!--======================= Tables =======================================-->

<!!-- IETF HTML table standard, see [RFC1942] -->

<!!--
 The BORDER attribute sets the thickness of the frame around the
 table. The default units are screen pixels.

 The FRAME attribute specifies which parts of the frame around
 the table should be rendered. The values are not the same as
 CALS to avoid a name clash with the VALIGN attribute.

 The value "border" is included for backwards compatibility with
 <TABLE BORDER> which yields frame=border and border=implied
 For <TABLE BORDER=1> you get border=1 and frame=implied. In this
 case, it is appropriate to treat this as frame=border for backwards
 compatibility with deployed browsers.
-->
<!!ENTITY % TFrame "(void|above|below|hsides|lhs|rhs|vsides|box|border)">

<!!--
 The RULES attribute defines which rules to draw between cells:

 If RULES is absent then assume:
     "none" if BORDER is absent or BORDER=0 otherwise "all"
-->

<!!ENTITY % TRules "(none | groups | rows | cols | all)">
  
<!!-- horizontal placement of table relative to document -->
<!!ENTITY % TAlign "(left|center|right)">

<!!-- horizontal alignment attributes for cell contents -->
<!!ENTITY % cellhalign
  "align      (left|center|right|justify|char) #IMPLIED
   char       %Character;    #IMPLIED  -- alignment char, e.g. char='':'' --
   charoff    %Length;       #IMPLIED  -- offset for alignment char --"
  >

<!!-- vertical alignment attributes for cell contents -->
<!!ENTITY % cellvalign
  "valign     (top|middle|bottom|baseline) #IMPLIED"
  >

<!!ELEMENT TABLE - -
     (CAPTION?, (COL*|COLGROUP*), THEAD?, TFOOT?, (TBODY|TR+))>
<!!ELEMENT CAPTION  - - (%inline;)*     -- table caption -->
<!!ELEMENT THEAD    - O (TR)+           -- table header -->
<!!ELEMENT TFOOT    - O (TR)+           -- table footer -->
<!!ELEMENT TBODY    O O (TR)+           -- table body -->
<!!ELEMENT COLGROUP - O (COL)*          -- table column group -->
<!!ELEMENT COL      - O EMPTY           -- table column -->
<!!ELEMENT TR       - O (TH|TD)+        -- table row -->
<!!ELEMENT (TH|TD)  - O (%flow;)*       -- table header cell, table data cell-->

<!!ATTLIST TABLE                        -- table element --
  %attrs;                              -- %coreattrs, %i18n, %events --
  summary     %Text;         #IMPLIED  -- purpose/structure for speech output--
  width       %Length;       #IMPLIED  -- table width --
  border      %Pixels;       #IMPLIED  -- controls frame width around table --
  frame       %TFrame;       #IMPLIED  -- which parts of frame to render --
  rules       %TRules;       #IMPLIED  -- rulings between rows and cols --
  cellspacing %Length;       #IMPLIED  -- spacing between cells --
  cellpadding %Length;       #IMPLIED  -- spacing within cells --
  %reserved;                           -- reserved for possible future use --
  datapagesize CDATA         #IMPLIED  -- reserved for possible future use --
  >


<!!ATTLIST CAPTION
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--
COLGROUP groups a set of COL elements. It allows you to group
several semantically related columns together.
-->
<!!ATTLIST COLGROUP
  %attrs;                              -- %coreattrs, %i18n, %events --
  span        NUMBER         1         -- default number of columns in group --
  width       %MultiLength;  #IMPLIED  -- default width for enclosed COLs --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!--
 COL elements define the alignment properties for cells in
 one or more columns.

 The WIDTH attribute specifies the width of the columns, e.g.

     width=64        width in screen pixels
     width=0.5*      relative width of 0.5

 The SPAN attribute causes the attributes of one
 COL element to apply to more than one column.
-->
<!!ATTLIST COL                          -- column groups and properties --
  %attrs;                              -- %coreattrs, %i18n, %events --
  span        NUMBER         1         -- COL attributes affect N columns --
  width       %MultiLength;  #IMPLIED  -- column width specification --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!--
    Use THEAD to duplicate headers when breaking table
    across page boundaries, or for static headers when
    TBODY sections are rendered in scrolling panel.

    Use TFOOT to duplicate footers when breaking table
    across page boundaries, or for static footers when
    TBODY sections are rendered in scrolling panel.

    Use multiple TBODY sections when rules are needed
    between groups of table rows.
-->
<!!ATTLIST (THEAD|TBODY|TFOOT)          -- table section --
  %attrs;                              -- %coreattrs, %i18n, %events --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!ATTLIST TR                           -- table row --
  %attrs;                              -- %coreattrs, %i18n, %events --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >


<!!-- Scope is simpler than headers attribute for common tables -->
<!!ENTITY % Scope "(row|col|rowgroup|colgroup)">

<!!-- TH is for headers, TD for data, but for cells acting as both use TD -->
<!!ATTLIST (TH|TD)                      -- header or data cell --
  %attrs;                              -- %coreattrs, %i18n, %events --
  abbr        %Text;         #IMPLIED  -- abbreviation for header cell --
  axis        CDATA          #IMPLIED  -- comma-separated list of related headers--
  headers     IDREFS         #IMPLIED  -- list of id''s for header cells --
  scope       %Scope;        #IMPLIED  -- scope covered by header cells --
  rowspan     NUMBER         1         -- number of rows spanned by cell --
  colspan     NUMBER         1         -- number of cols spanned by cell --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >


<!!--================ Document Head =======================================-->
<!!-- %head.misc; defined earlier on as "SCRIPT|STYLE|META|LINK|OBJECT" -->
<!!ENTITY % head.content "TITLE & BASE?">

<!!ELEMENT HEAD O O (%head.content;) +(%head.misc;) -- document head -->
<!!ATTLIST HEAD
  %i18n;                               -- lang, dir --
  profile     %URI;          #IMPLIED  -- named dictionary of meta info --
  >

<!!-- The TITLE element is not considered part of the flow of text.
       It should be displayed, for example as the page header or
       window title. Exactly one title is required per document.
    -->
<!!ELEMENT TITLE - - (#PCDATA) -(%head.misc;) -- document title -->
<!!ATTLIST TITLE %i18n>


<!!ELEMENT BASE - O EMPTY               -- document base URI -->
<!!ATTLIST BASE
  href        %URI;          #REQUIRED -- URI that acts as base URI --
  >

<!!ELEMENT META - O EMPTY               -- generic metainformation -->
<!!ATTLIST META
  %i18n;                               -- lang, dir, for use with content --
  http-equiv  NAME           #IMPLIED  -- HTTP response header name  --
  name        NAME           #IMPLIED  -- metainformation name --
  content     CDATA          #REQUIRED -- associated information --
  scheme      CDATA          #IMPLIED  -- select form of content --
  >

<!!ELEMENT STYLE - - %StyleSheet        -- style info -->
<!!ATTLIST STYLE
  %i18n;                               -- lang, dir, for use with title --
  type        %ContentType;  #REQUIRED -- content type of style language --
  media       %MediaDesc;    #IMPLIED  -- designed for use with these media --
  title       %Text;         #IMPLIED  -- advisory title --
  >

<!!ELEMENT SCRIPT - - %Script;          -- script statements -->
<!!ATTLIST SCRIPT
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  type        %ContentType;  #REQUIRED -- content type of script language --
  src         %URI;          #IMPLIED  -- URI for an external script --
  defer       (defer)        #IMPLIED  -- UA may defer execution of script --
  event       CDATA          #IMPLIED  -- reserved for possible future use --
  for         %URI;          #IMPLIED  -- reserved for possible future use --
  >

<!!ELEMENT NOSCRIPT - - (%block;)+
  -- alternate content container for non script-based rendering -->
<!!ATTLIST NOSCRIPT
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--================ Document Structure ==================================-->
<!!ENTITY % html.content "HEAD, BODY">

<!!ELEMENT HTML O O (%html.content;)    -- document root element -->
<!!ATTLIST HTML
  %i18n;                               -- lang, dir --
  >
'! !

!Html401Strict methodsFor: 'initialize-release' stamp: 'tb 11/30/2005 22:56'!
initialize

	super initialize.
	self parse.! !

HtmlDTD subclass: #Html401Transitional
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!Html401Transitional class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:54'!
defaultUrl

	^'http://www.w3.org/TR/html4/loose.dtd' asUrl! !

!Html401Transitional class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:35'!
docType

	^'<!!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN" "http://www.w3.org/TR/html4/loose.dtd">'! !

!Html401Transitional methodsFor: 'resources' stamp: 'tb 1/14/2006 12:39'!
dtd
	^ '<!!--
    This is the HTML 4.01 Transitional DTD, which includes
    presentation attributes and elements that W3C expects to phase out
    as support for style sheets matures. Authors should use the Strict
    DTD when possible, but may use the Transitional DTD when support
    for presentation attribute and elements is required.

    HTML 4 includes mechanisms for style sheets, scripting,
    embedding objects, improved support for right to left and mixed
    direction text, and enhancements to forms for improved
    accessibility for people with disabilities.

          Draft: $Date: 1999/12/24 23:37:48 $

          Authors:
              Dave Raggett <dsr@w3.org>
              Arnaud Le Hors <lehors@w3.org>
              Ian Jacobs <ij@w3.org>

    Further information about HTML 4.01 is available at:

        http://www.w3.org/TR/1999/REC-html401-19991224


    The HTML 4.01 specification includes additional
    syntactic constraints that cannot be expressed within
    the DTDs.

-->
<!!ENTITY % HTML.Version "-//W3C//DTD HTML 4.01 Transitional//EN"
>  <!!-- Typical usage:

    <!!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
            "http://www.w3.org/TR/html4/loose.dtd">
    <html>
    <head>
    ...
    </head>
    <body>
    ...
    </body>
    </html>

    The URI used as a system identifier with the public identifier allows
    the user agent to download the DTD and entity sets as needed.

    The FPI for the Strict HTML 4.01 DTD is:

        "-//W3C//DTD HTML 4.01//EN"

    This version of the strict DTD is:

        http://www.w3.org/TR/1999/REC-html401-19991224/strict.dtd

    Authors should use the Strict DTD unless they need the
    presentation control for user agents that don''t (adequately)
    support style sheets.

    If you are writing a document that includes frames, use 
    the following FPI:

        "-//W3C//DTD HTML 4.01 Frameset//EN"

    This version of the frameset DTD is:

        http://www.w3.org/TR/1999/REC-html401-19991224/frameset.dtd

    Use the following (relative) URIs to refer to 
    the DTDs and entity definitions of this specification:

    "strict.dtd"
    "loose.dtd"
    "frameset.dtd"
    "HTMLlat1.ent"
    "HTMLsymbol.ent"
    "HTMLspecial.ent"

-->

<!!--================== Imported Names ====================================-->
<!!-- Feature Switch for frameset documents -->
<!!ENTITY % HTML.Frameset "IGNORE">

<!!ENTITY % ContentType "CDATA"
    -- media type, as per [RFC2045]
    -->

<!!ENTITY % ContentTypes "CDATA"
    -- comma-separated list of media types, as per [RFC2045]
    -->

<!!ENTITY % Charset "CDATA"
    -- a character encoding, as per [RFC2045]
    -->

<!!ENTITY % Charsets "CDATA"
    -- a space-separated list of character encodings, as per [RFC2045]
    -->

<!!ENTITY % LanguageCode "NAME"
    -- a language code, as per [RFC1766]
    -->

<!!ENTITY % Character "CDATA"
    -- a single character from [ISO10646] 
    -->

<!!ENTITY % LinkTypes "CDATA"
    -- space-separated list of link types
    -->

<!!ENTITY % MediaDesc "CDATA"
    -- single or comma-separated list of media descriptors
    -->

<!!ENTITY % URI "CDATA"
    -- a Uniform Resource Identifier,
       see [URI]
    -->

<!!ENTITY % Datetime "CDATA" -- date and time information. ISO date format -->


<!!ENTITY % Script "CDATA" -- script expression -->

<!!ENTITY % StyleSheet "CDATA" -- style sheet data -->

<!!ENTITY % FrameTarget "CDATA" -- render in this frame -->


<!!ENTITY % Text "CDATA">


<!!-- Parameter Entities -->

<!!ENTITY % head.misc "SCRIPT|STYLE|META|LINK|OBJECT" -- repeatable head elements -->

<!!ENTITY % heading "H1|H2|H3|H4|H5|H6">

<!!ENTITY % list "UL | OL |  DIR | MENU">

<!!ENTITY % preformatted "PRE">

<!!ENTITY % Color "CDATA" -- a color using sRGB: #RRGGBB as Hex values -->

<!!-- There are also 16 widely known color names with their sRGB values:

    Black  = #000000    Green  = #008000
    Silver = #C0C0C0    Lime   = #00FF00
    Gray   = #808080    Olive  = #808000
    White  = #FFFFFF    Yellow = #FFFF00
    Maroon = #800000    Navy   = #000080
    Red    = #FF0000    Blue   = #0000FF
    Purple = #800080    Teal   = #008080
    Fuchsia= #FF00FF    Aqua   = #00FFFF
 -->

<!!ENTITY % bodycolors "
  bgcolor     %Color;        #IMPLIED  -- document background color --
  text        %Color;        #IMPLIED  -- document text color --
  link        %Color;        #IMPLIED  -- color of links --
  vlink       %Color;        #IMPLIED  -- color of visited links --
  alink       %Color;        #IMPLIED  -- color of selected links --
  ">

<!!--================ Character mnemonic entities =========================-->

<!!ENTITY % HTMLlat1 PUBLIC
   "-//W3C//ENTITIES Latin1//EN//HTML"
   "HTMLlat1.ent">
%HTMLlat1;

<!!ENTITY % HTMLsymbol PUBLIC
   "-//W3C//ENTITIES Symbols//EN//HTML"
   "HTMLsymbol.ent">
%HTMLsymbol;

<!!ENTITY % HTMLspecial PUBLIC
   "-//W3C//ENTITIES Special//EN//HTML"
   "HTMLspecial.ent">
%HTMLspecial;
<!!--=================== Generic Attributes ===============================-->

<!!ENTITY % coreattrs
 "id          ID             #IMPLIED  -- document-wide unique id --
  class       CDATA          #IMPLIED  -- space-separated list of classes --
  style       %StyleSheet;   #IMPLIED  -- associated style info --
  title       %Text;         #IMPLIED  -- advisory title --"
  >

<!!ENTITY % i18n
 "lang        %LanguageCode; #IMPLIED  -- language code --
  dir         (ltr|rtl)      #IMPLIED  -- direction for weak/neutral text --"
  >

<!!ENTITY % events
 "onclick     %Script;       #IMPLIED  -- a pointer button was clicked --
  ondblclick  %Script;       #IMPLIED  -- a pointer button was double clicked--
  onmousedown %Script;       #IMPLIED  -- a pointer button was pressed down --
  onmouseup   %Script;       #IMPLIED  -- a pointer button was released --
  onmouseover %Script;       #IMPLIED  -- a pointer was moved onto --
  onmousemove %Script;       #IMPLIED  -- a pointer was moved within --
  onmouseout  %Script;       #IMPLIED  -- a pointer was moved away --
  onkeypress  %Script;       #IMPLIED  -- a key was pressed and released --
  onkeydown   %Script;       #IMPLIED  -- a key was pressed down --
  onkeyup     %Script;       #IMPLIED  -- a key was released --"
  >

<!!-- Reserved Feature Switch -->
<!!ENTITY % HTML.Reserved "IGNORE">

<!!-- The following attributes are reserved for possible future use -->
<!![ %HTML.Reserved; [
<!!ENTITY % reserved
 "datasrc     %URI;          #IMPLIED  -- a single or tabular Data Source --
  datafld     CDATA          #IMPLIED  -- the property or column name --
  dataformatas (plaintext|html) plaintext -- text or html --"
  >
]]>

<!!ENTITY % reserved "">

<!!ENTITY % attrs "%coreattrs; %i18n; %events;">

<!!ENTITY % align "align (left|center|right|justify)  #IMPLIED"
                   -- default is left for ltr paragraphs, right for rtl --
  >

<!!--=================== Text Markup ======================================-->

<!!ENTITY % fontstyle
 "TT | I | B | U | S | STRIKE | BIG | SMALL">

<!!ENTITY % phrase "EM | STRONG | DFN | CODE |
                   SAMP | KBD | VAR | CITE | ABBR | ACRONYM" >

<!!ENTITY % special
   "A | IMG | APPLET | OBJECT | FONT | BASEFONT | BR | SCRIPT |
    MAP | Q | SUB | SUP | SPAN | BDO | IFRAME">

<!!ENTITY % formctrl "INPUT | SELECT | TEXTAREA | LABEL | BUTTON">

<!!-- %inline; covers inline or "text-level" elements -->
<!!ENTITY % inline "#PCDATA | %fontstyle; | %phrase; | %special; | %formctrl;">

<!!ELEMENT (%fontstyle;|%phrase;) - - (%inline;)*>
<!!ATTLIST (%fontstyle;|%phrase;)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT (SUB|SUP) - - (%inline;)*    -- subscript, superscript -->
<!!ATTLIST (SUB|SUP)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT SPAN - - (%inline;)*         -- generic language/style container -->
<!!ATTLIST SPAN
  %attrs;                              -- %coreattrs, %i18n, %events --
  %reserved;			       -- reserved for possible future use --
  >

<!!ELEMENT BDO - - (%inline;)*          -- I18N BiDi over-ride -->
<!!ATTLIST BDO
  %coreattrs;                          -- id, class, style, title --
  lang        %LanguageCode; #IMPLIED  -- language code --
  dir         (ltr|rtl)      #REQUIRED -- directionality --
  >

<!!ELEMENT BASEFONT - O EMPTY           -- base font size -->
<!!ATTLIST BASEFONT
  id          ID             #IMPLIED  -- document-wide unique id --
  size        CDATA          #REQUIRED -- base font size for FONT elements --
  color       %Color;        #IMPLIED  -- text color --
  face        CDATA          #IMPLIED  -- comma-separated list of font names --
  >

<!!ELEMENT FONT - - (%inline;)*         -- local change to font -->
<!!ATTLIST FONT
  %coreattrs;                          -- id, class, style, title --
  %i18n;		               -- lang, dir --
  size        CDATA          #IMPLIED  -- [+|-]nn e.g. size="+1", size="4" --
  color       %Color;        #IMPLIED  -- text color --
  face        CDATA          #IMPLIED  -- comma-separated list of font names --
  >

<!!ELEMENT BR - O EMPTY                 -- forced line break -->
<!!ATTLIST BR
  %coreattrs;                          -- id, class, style, title --
  clear       (left|all|right|none) none -- control of text flow --
  >

<!!--================== HTML content models ===============================-->

<!!--
    HTML has two basic content models:

        %inline;     character level elements and text strings
        %block;      block-like elements e.g. paragraphs and lists
-->

<!!ENTITY % block
     "P | %heading; | %list; | %preformatted; | DL | DIV | CENTER |
      NOSCRIPT | NOFRAMES | BLOCKQUOTE | FORM | ISINDEX | HR |
      TABLE | FIELDSET | ADDRESS">

<!!ENTITY % flow "%block; | %inline;">

<!!--=================== Document Body ====================================-->

<!!ELEMENT BODY O O (%flow;)* +(INS|DEL) -- document body -->
<!!ATTLIST BODY
  %attrs;                              -- %coreattrs, %i18n, %events --
  onload          %Script;   #IMPLIED  -- the document has been loaded --
  onunload        %Script;   #IMPLIED  -- the document has been removed --
  background      %URI;      #IMPLIED  -- texture tile for document
                                          background --
  %bodycolors;                         -- bgcolor, text, link, vlink, alink --
  >

<!!ELEMENT ADDRESS - - ((%inline;)|P)*  -- information on author -->
<!!ATTLIST ADDRESS
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT DIV - - (%flow;)*            -- generic language/style container -->
<!!ATTLIST DIV
  %attrs;                              -- %coreattrs, %i18n, %events --
  %align;                              -- align, text alignment --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT CENTER - - (%flow;)*         -- shorthand for DIV align=center -->
<!!ATTLIST CENTER
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--================== The Anchor Element ================================-->

<!!ENTITY % Shape "(rect|circle|poly|default)">
<!!ENTITY % Coords "CDATA" -- comma-separated list of lengths -->

<!!ELEMENT A - - (%inline;)* -(A)       -- anchor -->
<!!ATTLIST A
  %attrs;                              -- %coreattrs, %i18n, %events --
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  type        %ContentType;  #IMPLIED  -- advisory content type --
  name        CDATA          #IMPLIED  -- named link end --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  hreflang    %LanguageCode; #IMPLIED  -- language code --
  target      %FrameTarget;  #IMPLIED  -- render in this frame --
  rel         %LinkTypes;    #IMPLIED  -- forward link types --
  rev         %LinkTypes;    #IMPLIED  -- reverse link types --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  shape       %Shape;        rect      -- for use with client-side image maps --
  coords      %Coords;       #IMPLIED  -- for use with client-side image maps --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!--================== Client-side image maps ============================-->

<!!-- These can be placed in the same document or grouped in a
     separate document although this isn''t yet widely supported -->

<!!ELEMENT MAP - - ((%block;) | AREA)+ -- client-side image map -->
<!!ATTLIST MAP
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #REQUIRED -- for reference by usemap --
  >

<!!ELEMENT AREA - O EMPTY               -- client-side image map area -->
<!!ATTLIST AREA
  %attrs;                              -- %coreattrs, %i18n, %events --
  shape       %Shape;        rect      -- controls interpretation of coords --
  coords      %Coords;       #IMPLIED  -- comma-separated list of lengths --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  target      %FrameTarget;  #IMPLIED  -- render in this frame --
  nohref      (nohref)       #IMPLIED  -- this region has no action --
  alt         %Text;         #REQUIRED -- short description --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!--================== The LINK Element ==================================-->

<!!--
  Relationship values can be used in principle:

   a) for document specific toolbars/menus when used
      with the LINK element in document head e.g.
        start, contents, previous, next, index, end, help
   b) to link to a separate style sheet (rel=stylesheet)
   c) to make a link to a script (rel=script)
   d) by stylesheets to control how collections of
      html nodes are rendered into printed documents
   e) to make a link to a printable version of this document
      e.g. a postscript or pdf version (rel=alternate media=print)
-->

<!!ELEMENT LINK - O EMPTY               -- a media-independent link -->
<!!ATTLIST LINK
  %attrs;                              -- %coreattrs, %i18n, %events --
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  href        %URI;          #IMPLIED  -- URI for linked resource --
  hreflang    %LanguageCode; #IMPLIED  -- language code --
  type        %ContentType;  #IMPLIED  -- advisory content type --
  rel         %LinkTypes;    #IMPLIED  -- forward link types --
  rev         %LinkTypes;    #IMPLIED  -- reverse link types --
  media       %MediaDesc;    #IMPLIED  -- for rendering on these media --
  target      %FrameTarget;  #IMPLIED  -- render in this frame --
  >

<!!--=================== Images ===========================================-->

<!!-- Length defined in strict DTD for cellpadding/cellspacing -->
<!!ENTITY % Length "CDATA" -- nn for pixels or nn% for percentage length -->
<!!ENTITY % MultiLength "CDATA" -- pixel, percentage, or relative -->

<!![ %HTML.Frameset; [
<!!ENTITY % MultiLengths "CDATA" -- comma-separated list of MultiLength -->
]]>

<!!ENTITY % Pixels "CDATA" -- integer representing length in pixels -->

<!!ENTITY % IAlign "(top|middle|bottom|left|right)" -- center? -->

<!!-- To avoid problems with text-only UAs as well as 
   to make image content understandable and navigable 
   to users of non-visual UAs, you need to provide
   a description with ALT, and avoid server-side image maps -->
<!!ELEMENT IMG - O EMPTY                -- Embedded image -->
<!!ATTLIST IMG
  %attrs;                              -- %coreattrs, %i18n, %events --
  src         %URI;          #REQUIRED -- URI of image to embed --
  alt         %Text;         #REQUIRED -- short description --
  longdesc    %URI;          #IMPLIED  -- link to long description
                                          (complements alt) --
  name        CDATA          #IMPLIED  -- name of image for scripting --
  height      %Length;       #IMPLIED  -- override height --
  width       %Length;       #IMPLIED  -- override width --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  ismap       (ismap)        #IMPLIED  -- use server-side image map --
  align       %IAlign;       #IMPLIED  -- vertical or horizontal alignment --
  border      %Pixels;       #IMPLIED  -- link border width --
  hspace      %Pixels;       #IMPLIED  -- horizontal gutter --
  vspace      %Pixels;       #IMPLIED  -- vertical gutter --
  >

<!!-- USEMAP points to a MAP element which may be in this document
  or an external document, although the latter is not widely supported -->

<!!--==================== OBJECT ======================================-->
<!!--
  OBJECT is used to embed objects as part of HTML pages 
  PARAM elements should precede other content. SGML mixed content
  model technicality precludes specifying this formally ...
-->

<!!ELEMENT OBJECT - - (PARAM | %flow;)*
 -- generic embedded object -->
<!!ATTLIST OBJECT
  %attrs;                              -- %coreattrs, %i18n, %events --
  declare     (declare)      #IMPLIED  -- declare but don''t instantiate flag --
  classid     %URI;          #IMPLIED  -- identifies an implementation --
  codebase    %URI;          #IMPLIED  -- base URI for classid, data, archive--
  data        %URI;          #IMPLIED  -- reference to object''s data --
  type        %ContentType;  #IMPLIED  -- content type for data --
  codetype    %ContentType;  #IMPLIED  -- content type for code --
  archive     CDATA          #IMPLIED  -- space-separated list of URIs --
  standby     %Text;         #IMPLIED  -- message to show while loading --
  height      %Length;       #IMPLIED  -- override height --
  width       %Length;       #IMPLIED  -- override width --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  name        CDATA          #IMPLIED  -- submit as part of form --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  align       %IAlign;       #IMPLIED  -- vertical or horizontal alignment --
  border      %Pixels;       #IMPLIED  -- link border width --
  hspace      %Pixels;       #IMPLIED  -- horizontal gutter --
  vspace      %Pixels;       #IMPLIED  -- vertical gutter --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT PARAM - O EMPTY              -- named property value -->
<!!ATTLIST PARAM
  id          ID             #IMPLIED  -- document-wide unique id --
  name        CDATA          #REQUIRED -- property name --
  value       CDATA          #IMPLIED  -- property value --
  valuetype   (DATA|REF|OBJECT) DATA   -- How to interpret value --
  type        %ContentType;  #IMPLIED  -- content type for value
                                          when valuetype=ref --
  >

<!!--=================== Java APPLET ==================================-->
<!!--
  One of code or object attributes must be present.
  Place PARAM elements before other content.
-->
<!!ELEMENT APPLET - - (PARAM | %flow;)* -- Java applet -->
<!!ATTLIST APPLET
  %coreattrs;                          -- id, class, style, title --
  codebase    %URI;          #IMPLIED  -- optional base URI for applet --
  archive     CDATA          #IMPLIED  -- comma-separated archive list --
  code        CDATA          #IMPLIED  -- applet class file --
  object      CDATA          #IMPLIED  -- serialized applet file --
  alt         %Text;         #IMPLIED  -- short description --
  name        CDATA          #IMPLIED  -- allows applets to find each other --
  width       %Length;       #REQUIRED -- initial width --
  height      %Length;       #REQUIRED -- initial height --
  align       %IAlign;       #IMPLIED  -- vertical or horizontal alignment --
  hspace      %Pixels;       #IMPLIED  -- horizontal gutter --
  vspace      %Pixels;       #IMPLIED  -- vertical gutter --
  >

<!!--=================== Horizontal Rule ==================================-->

<!!ELEMENT HR - O EMPTY -- horizontal rule -->
<!!ATTLIST HR
  %attrs;                              -- %coreattrs, %i18n, %events --
  align       (left|center|right) #IMPLIED
  noshade     (noshade)      #IMPLIED
  size        %Pixels;       #IMPLIED
  width       %Length;       #IMPLIED
  >

<!!--=================== Paragraphs =======================================-->

<!!ELEMENT P - O (%inline;)*            -- paragraph -->
<!!ATTLIST P
  %attrs;                              -- %coreattrs, %i18n, %events --
  %align;                              -- align, text alignment --
  >

<!!--=================== Headings =========================================-->

<!!--
  There are six levels of headings from H1 (the most important)
  to H6 (the least important).
-->

<!!ELEMENT (%heading;)  - - (%inline;)* -- heading -->
<!!ATTLIST (%heading;)
  %attrs;                              -- %coreattrs, %i18n, %events --
  %align;                              -- align, text alignment --
  >

<!!--=================== Preformatted Text ================================-->

<!!-- excludes markup for images and changes in font size -->
<!!ENTITY % pre.exclusion "IMG|OBJECT|APPLET|BIG|SMALL|SUB|SUP|FONT|BASEFONT">

<!!ELEMENT PRE - - (%inline;)* -(%pre.exclusion;) -- preformatted text -->
<!!ATTLIST PRE
  %attrs;                              -- %coreattrs, %i18n, %events --
  width       NUMBER         #IMPLIED
  >

<!!--===================== Inline Quotes ==================================-->

<!!ELEMENT Q - - (%inline;)*            -- short inline quotation -->
<!!ATTLIST Q
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- URI for source document or msg --
  >

<!!--=================== Block-like Quotes ================================-->

<!!ELEMENT BLOCKQUOTE - - (%flow;)*     -- long quotation -->
<!!ATTLIST BLOCKQUOTE
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- URI for source document or msg --
  >

<!!--=================== Inserted/Deleted Text ============================-->


<!!-- INS/DEL are handled by inclusion on BODY -->
<!!ELEMENT (INS|DEL) - - (%flow;)*      -- inserted text, deleted text -->
<!!ATTLIST (INS|DEL)
  %attrs;                              -- %coreattrs, %i18n, %events --
  cite        %URI;          #IMPLIED  -- info on reason for change --
  datetime    %Datetime;     #IMPLIED  -- date and time of change --
  >

<!!--=================== Lists ============================================-->

<!!-- definition lists - DT for term, DD for its definition -->

<!!ELEMENT DL - - (DT|DD)+              -- definition list -->
<!!ATTLIST DL
  %attrs;                              -- %coreattrs, %i18n, %events --
  compact     (compact)      #IMPLIED  -- reduced interitem spacing --
  >

<!!ELEMENT DT - O (%inline;)*           -- definition term -->
<!!ELEMENT DD - O (%flow;)*             -- definition description -->
<!!ATTLIST (DT|DD)
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!-- Ordered lists (OL) Numbering style

    1   arablic numbers     1, 2, 3, ...
    a   lower alpha         a, b, c, ...
    A   upper alpha         A, B, C, ...
    i   lower roman         i, ii, iii, ...
    I   upper roman         I, II, III, ...

    The style is applied to the sequence number which by default
    is reset to 1 for the first list item in an ordered list.

    This can''t be expressed directly in SGML due to case folding.
-->

<!!ENTITY % OLStyle "CDATA"      -- constrained to: "(1|a|A|i|I)" -->

<!!ELEMENT OL - - (LI)+                 -- ordered list -->
<!!ATTLIST OL
  %attrs;                              -- %coreattrs, %i18n, %events --
  type        %OLStyle;      #IMPLIED  -- numbering style --
  compact     (compact)      #IMPLIED  -- reduced interitem spacing --
  start       NUMBER         #IMPLIED  -- starting sequence number --
  >

<!!-- Unordered Lists (UL) bullet styles -->
<!!ENTITY % ULStyle "(disc|square|circle)">

<!!ELEMENT UL - - (LI)+                 -- unordered list -->
<!!ATTLIST UL
  %attrs;                              -- %coreattrs, %i18n, %events --
  type        %ULStyle;      #IMPLIED  -- bullet style --
  compact     (compact)      #IMPLIED  -- reduced interitem spacing --
  >

<!!ELEMENT (DIR|MENU) - - (LI)+ -(%block;) -- directory list, menu list -->
<!!ATTLIST DIR
  %attrs;                              -- %coreattrs, %i18n, %events --
  compact     (compact)      #IMPLIED -- reduced interitem spacing --
  >
<!!ATTLIST MENU
  %attrs;                              -- %coreattrs, %i18n, %events --
  compact     (compact)      #IMPLIED -- reduced interitem spacing --
  >

<!!ENTITY % LIStyle "CDATA" -- constrained to: "(%ULStyle;|%OLStyle;)" -->

<!!ELEMENT LI - O (%flow;)*             -- list item -->
<!!ATTLIST LI
  %attrs;                              -- %coreattrs, %i18n, %events --
  type        %LIStyle;      #IMPLIED  -- list item style --
  value       NUMBER         #IMPLIED  -- reset sequence number --
  >

<!!--================ Forms ===============================================-->
<!!ELEMENT FORM - - (%flow;)* -(FORM)   -- interactive form -->
<!!ATTLIST FORM
  %attrs;                              -- %coreattrs, %i18n, %events --
  action      %URI;          #REQUIRED -- server-side form handler --
  method      (get|post)     GET       -- HTTP method used to submit the form--
  enctype     %ContentType;  "application/x-www-form-urlencoded"
  accept      %ContentTypes; #IMPLIED  -- list of MIME types for file upload --
  name        CDATA          #IMPLIED  -- name of form for scripting --
  onsubmit    %Script;       #IMPLIED  -- the form was submitted --
  onreset     %Script;       #IMPLIED  -- the form was reset --
  target      %FrameTarget;  #IMPLIED  -- render in this frame --
  accept-charset %Charsets;  #IMPLIED  -- list of supported charsets --
  >

<!!-- Each label must not contain more than ONE field -->
<!!ELEMENT LABEL - - (%inline;)* -(LABEL) -- form field label text -->
<!!ATTLIST LABEL
  %attrs;                              -- %coreattrs, %i18n, %events --
  for         IDREF          #IMPLIED  -- matches field ID value --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  >

<!!ENTITY % InputType
  "(text | password | checkbox |
    radio | submit | reset |
    file | hidden | image | button)"
   >

<!!-- attribute name required for all but submit and reset -->
<!!ELEMENT INPUT - O EMPTY              -- form control -->
<!!ATTLIST INPUT
  %attrs;                              -- %coreattrs, %i18n, %events --
  type        %InputType;    TEXT      -- what kind of widget is needed --
  name        CDATA          #IMPLIED  -- submit as part of form --
  value       CDATA          #IMPLIED  -- Specify for radio buttons and checkboxes --
  checked     (checked)      #IMPLIED  -- for radio buttons and check boxes --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  readonly    (readonly)     #IMPLIED  -- for text and passwd --
  size        CDATA          #IMPLIED  -- specific to each type of field --
  maxlength   NUMBER         #IMPLIED  -- max chars for text fields --
  src         %URI;          #IMPLIED  -- for fields with images --
  alt         CDATA          #IMPLIED  -- short description --
  usemap      %URI;          #IMPLIED  -- use client-side image map --
  ismap       (ismap)        #IMPLIED  -- use server-side image map --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onselect    %Script;       #IMPLIED  -- some text was selected --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  accept      %ContentTypes; #IMPLIED  -- list of MIME types for file upload --
  align       %IAlign;       #IMPLIED  -- vertical or horizontal alignment --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT SELECT - - (OPTGROUP|OPTION)+ -- option selector -->
<!!ATTLIST SELECT
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED  -- field name --
  size        NUMBER         #IMPLIED  -- rows visible --
  multiple    (multiple)     #IMPLIED  -- default is single selection --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  %reserved;                           -- reserved for possible future use --
  >

<!!ELEMENT OPTGROUP - - (OPTION)+ -- option group -->
<!!ATTLIST OPTGROUP
  %attrs;                              -- %coreattrs, %i18n, %events --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  label       %Text;         #REQUIRED -- for use in hierarchical menus --
  >

<!!ELEMENT OPTION - O (#PCDATA)         -- selectable choice -->
<!!ATTLIST OPTION
  %attrs;                              -- %coreattrs, %i18n, %events --
  selected    (selected)     #IMPLIED
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  label       %Text;         #IMPLIED  -- for use in hierarchical menus --
  value       CDATA          #IMPLIED  -- defaults to element content --
  >

<!!ELEMENT TEXTAREA - - (#PCDATA)       -- multi-line text field -->
<!!ATTLIST TEXTAREA
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED
  rows        NUMBER         #REQUIRED
  cols        NUMBER         #REQUIRED
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  readonly    (readonly)     #IMPLIED
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  onselect    %Script;       #IMPLIED  -- some text was selected --
  onchange    %Script;       #IMPLIED  -- the element value was changed --
  %reserved;                           -- reserved for possible future use --
  >

<!!--
  #PCDATA is to solve the mixed content problem,
  per specification only whitespace is allowed there!!
 -->
<!!ELEMENT FIELDSET - - (#PCDATA,LEGEND,(%flow;)*) -- form control group -->
<!!ATTLIST FIELDSET
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!ELEMENT LEGEND - - (%inline;)*       -- fieldset legend -->
<!!ENTITY % LAlign "(top|bottom|left|right)">

<!!ATTLIST LEGEND
  %attrs;                              -- %coreattrs, %i18n, %events --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  align       %LAlign;       #IMPLIED  -- relative to fieldset --
  >

<!!ELEMENT BUTTON - -
     (%flow;)* -(A|%formctrl;|FORM|ISINDEX|FIELDSET|IFRAME)
     -- push button -->
<!!ATTLIST BUTTON
  %attrs;                              -- %coreattrs, %i18n, %events --
  name        CDATA          #IMPLIED
  value       CDATA          #IMPLIED  -- sent to server when submitted --
  type        (button|submit|reset) submit -- for use as form button --
  disabled    (disabled)     #IMPLIED  -- unavailable in this context --
  tabindex    NUMBER         #IMPLIED  -- position in tabbing order --
  accesskey   %Character;    #IMPLIED  -- accessibility key character --
  onfocus     %Script;       #IMPLIED  -- the element got the focus --
  onblur      %Script;       #IMPLIED  -- the element lost the focus --
  %reserved;                           -- reserved for possible future use --
  >

<!!--======================= Tables =======================================-->

<!!-- IETF HTML table standard, see [RFC1942] -->

<!!--
 The BORDER attribute sets the thickness of the frame around the
 table. The default units are screen pixels.

 The FRAME attribute specifies which parts of the frame around
 the table should be rendered. The values are not the same as
 CALS to avoid a name clash with the VALIGN attribute.

 The value "border" is included for backwards compatibility with
 <TABLE BORDER> which yields frame=border and border=implied
 For <TABLE BORDER=1> you get border=1 and frame=implied. In this
 case, it is appropriate to treat this as frame=border for backwards
 compatibility with deployed browsers.
-->
<!!ENTITY % TFrame "(void|above|below|hsides|lhs|rhs|vsides|box|border)">

<!!--
 The RULES attribute defines which rules to draw between cells:

 If RULES is absent then assume:
     "none" if BORDER is absent or BORDER=0 otherwise "all"
-->

<!!ENTITY % TRules "(none | groups | rows | cols | all)">
  
<!!-- horizontal placement of table relative to document -->
<!!ENTITY % TAlign "(left|center|right)">

<!!-- horizontal alignment attributes for cell contents -->
<!!ENTITY % cellhalign
  "align      (left|center|right|justify|char) #IMPLIED
   char       %Character;    #IMPLIED  -- alignment char, e.g. char='':'' --
   charoff    %Length;       #IMPLIED  -- offset for alignment char --"
  >

<!!-- vertical alignment attributes for cell contents -->
<!!ENTITY % cellvalign
  "valign     (top|middle|bottom|baseline) #IMPLIED"
  >

<!!ELEMENT TABLE - -
     (CAPTION?, (COL*|COLGROUP*), THEAD?, TFOOT?, (TBODY|TR+))>
<!!ELEMENT CAPTION  - - (%inline;)*     -- table caption -->
<!!ELEMENT THEAD    - O (TR)+           -- table header -->
<!!ELEMENT TFOOT    - O (TR)+           -- table footer -->
<!!ELEMENT TBODY    O O (TR)+           -- table body -->
<!!ELEMENT COLGROUP - O (COL)*          -- table column group -->
<!!ELEMENT COL      - O EMPTY           -- table column -->
<!!ELEMENT TR       - O (TH|TD)+        -- table row -->
<!!ELEMENT (TH|TD)  - O (%flow;)*       -- table header cell, table data cell-->

<!!ATTLIST TABLE                        -- table element --
  %attrs;                              -- %coreattrs, %i18n, %events --
  summary     %Text;         #IMPLIED  -- purpose/structure for speech output--
  width       %Length;       #IMPLIED  -- table width --
  border      %Pixels;       #IMPLIED  -- controls frame width around table --
  frame       %TFrame;       #IMPLIED  -- which parts of frame to render --
  rules       %TRules;       #IMPLIED  -- rulings between rows and cols --
  cellspacing %Length;       #IMPLIED  -- spacing between cells --
  cellpadding %Length;       #IMPLIED  -- spacing within cells --
  align       %TAlign;       #IMPLIED  -- table position relative to window --
  bgcolor     %Color;        #IMPLIED  -- background color for cells --
  %reserved;                           -- reserved for possible future use --
  datapagesize CDATA         #IMPLIED  -- reserved for possible future use --
  >

<!!ENTITY % CAlign "(top|bottom|left|right)">

<!!ATTLIST CAPTION
  %attrs;                              -- %coreattrs, %i18n, %events --
  align       %CAlign;       #IMPLIED  -- relative to table --
  >

<!!--
COLGROUP groups a set of COL elements. It allows you to group
several semantically related columns together.
-->
<!!ATTLIST COLGROUP
  %attrs;                              -- %coreattrs, %i18n, %events --
  span        NUMBER         1         -- default number of columns in group --
  width       %MultiLength;  #IMPLIED  -- default width for enclosed COLs --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!--
 COL elements define the alignment properties for cells in
 one or more columns.

 The WIDTH attribute specifies the width of the columns, e.g.

     width=64        width in screen pixels
     width=0.5*      relative width of 0.5

 The SPAN attribute causes the attributes of one
 COL element to apply to more than one column.
-->
<!!ATTLIST COL                          -- column groups and properties --
  %attrs;                              -- %coreattrs, %i18n, %events --
  span        NUMBER         1         -- COL attributes affect N columns --
  width       %MultiLength;  #IMPLIED  -- column width specification --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!--
    Use THEAD to duplicate headers when breaking table
    across page boundaries, or for static headers when
    TBODY sections are rendered in scrolling panel.

    Use TFOOT to duplicate footers when breaking table
    across page boundaries, or for static footers when
    TBODY sections are rendered in scrolling panel.

    Use multiple TBODY sections when rules are needed
    between groups of table rows.
-->
<!!ATTLIST (THEAD|TBODY|TFOOT)          -- table section --
  %attrs;                              -- %coreattrs, %i18n, %events --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  >

<!!ATTLIST TR                           -- table row --
  %attrs;                              -- %coreattrs, %i18n, %events --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  bgcolor     %Color;        #IMPLIED  -- background color for row --
  >


<!!-- Scope is simpler than headers attribute for common tables -->
<!!ENTITY % Scope "(row|col|rowgroup|colgroup)">

<!!-- TH is for headers, TD for data, but for cells acting as both use TD -->
<!!ATTLIST (TH|TD)                      -- header or data cell --
  %attrs;                              -- %coreattrs, %i18n, %events --
  abbr        %Text;         #IMPLIED  -- abbreviation for header cell --
  axis        CDATA          #IMPLIED  -- comma-separated list of related headers--
  headers     IDREFS         #IMPLIED  -- list of id''s for header cells --
  scope       %Scope;        #IMPLIED  -- scope covered by header cells --
  rowspan     NUMBER         1         -- number of rows spanned by cell --
  colspan     NUMBER         1         -- number of cols spanned by cell --
  %cellhalign;                         -- horizontal alignment in cells --
  %cellvalign;                         -- vertical alignment in cells --
  nowrap      (nowrap)       #IMPLIED  -- suppress word wrap --
  bgcolor     %Color;        #IMPLIED  -- cell background color --
  width       %Length;       #IMPLIED  -- width for cell --
  height      %Length;       #IMPLIED  -- height for cell --
  >

<!!--================== Document Frames ===================================-->

<!!--
  The content model for HTML documents depends on whether the HEAD is
  followed by a FRAMESET or BODY element. The widespread omission of
  the BODY start tag makes it impractical to define the content model
  without the use of a marked section.
-->

<!![ %HTML.Frameset; [
<!!ELEMENT FRAMESET - - ((FRAMESET|FRAME)+ & NOFRAMES?) -- window subdivision-->
<!!ATTLIST FRAMESET
  %coreattrs;                          -- id, class, style, title --
  rows        %MultiLengths; #IMPLIED  -- list of lengths,
                                          default: 100% (1 row) --
  cols        %MultiLengths; #IMPLIED  -- list of lengths,
                                          default: 100% (1 col) --
  onload      %Script;       #IMPLIED  -- all the frames have been loaded  -- 
  onunload    %Script;       #IMPLIED  -- all the frames have been removed -- 
  >
]]>

<!![ %HTML.Frameset; [
<!!-- reserved frame names start with "_" otherwise starts with letter -->
<!!ELEMENT FRAME - O EMPTY              -- subwindow -->
<!!ATTLIST FRAME
  %coreattrs;                          -- id, class, style, title --
  longdesc    %URI;          #IMPLIED  -- link to long description
                                          (complements title) --
  name        CDATA          #IMPLIED  -- name of frame for targetting --
  src         %URI;          #IMPLIED  -- source of frame content --
  frameborder (1|0)          1         -- request frame borders? --
  marginwidth %Pixels;       #IMPLIED  -- margin widths in pixels --
  marginheight %Pixels;      #IMPLIED  -- margin height in pixels --
  noresize    (noresize)     #IMPLIED  -- allow users to resize frames? --
  scrolling   (yes|no|auto)  auto      -- scrollbar or none --
  >
]]>

<!!ELEMENT IFRAME - - (%flow;)*         -- inline subwindow -->
<!!ATTLIST IFRAME
  %coreattrs;                          -- id, class, style, title --
  longdesc    %URI;          #IMPLIED  -- link to long description
                                          (complements title) --
  name        CDATA          #IMPLIED  -- name of frame for targetting --
  src         %URI;          #IMPLIED  -- source of frame content --
  frameborder (1|0)          1         -- request frame borders? --
  marginwidth %Pixels;       #IMPLIED  -- margin widths in pixels --
  marginheight %Pixels;      #IMPLIED  -- margin height in pixels --
  scrolling   (yes|no|auto)  auto      -- scrollbar or none --
  align       %IAlign;       #IMPLIED  -- vertical or horizontal alignment --
  height      %Length;       #IMPLIED  -- frame height --
  width       %Length;       #IMPLIED  -- frame width --
  >

<!![ %HTML.Frameset; [
<!!ENTITY % noframes.content "(BODY) -(NOFRAMES)">
]]>

<!!ENTITY % noframes.content "(%flow;)*">

<!!ELEMENT NOFRAMES - - %noframes.content;
 -- alternate content container for non frame-based rendering -->
<!!ATTLIST NOFRAMES
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--================ Document Head =======================================-->
<!!-- %head.misc; defined earlier on as "SCRIPT|STYLE|META|LINK|OBJECT" -->
<!!ENTITY % head.content "TITLE & ISINDEX? & BASE?">

<!!ELEMENT HEAD O O (%head.content;) +(%head.misc;) -- document head -->
<!!ATTLIST HEAD
  %i18n;                               -- lang, dir --
  profile     %URI;          #IMPLIED  -- named dictionary of meta info --
  >

<!!-- The TITLE element is not considered part of the flow of text.
       It should be displayed, for example as the page header or
       window title. Exactly one title is required per document.
    -->
<!!ELEMENT TITLE - - (#PCDATA) -(%head.misc;) -- document title -->
<!!ATTLIST TITLE %i18n>

<!!ELEMENT ISINDEX - O EMPTY            -- single line prompt -->
<!!ATTLIST ISINDEX
  %coreattrs;                          -- id, class, style, title --
  %i18n;                               -- lang, dir --
  prompt      %Text;         #IMPLIED  -- prompt message -->

<!!ELEMENT BASE - O EMPTY               -- document base URI -->
<!!ATTLIST BASE
  href        %URI;          #IMPLIED  -- URI that acts as base URI --
  target      %FrameTarget;  #IMPLIED  -- render in this frame --
  >

<!!ELEMENT META - O EMPTY               -- generic metainformation -->
<!!ATTLIST META
  %i18n;                               -- lang, dir, for use with content --
  http-equiv  NAME           #IMPLIED  -- HTTP response header name  --
  name        NAME           #IMPLIED  -- metainformation name --
  content     CDATA          #REQUIRED -- associated information --
  scheme      CDATA          #IMPLIED  -- select form of content --
  >

<!!ELEMENT STYLE - - %StyleSheet        -- style info -->
<!!ATTLIST STYLE
  %i18n;                               -- lang, dir, for use with title --
  type        %ContentType;  #REQUIRED -- content type of style language --
  media       %MediaDesc;    #IMPLIED  -- designed for use with these media --
  title       %Text;         #IMPLIED  -- advisory title --
  >

<!!ELEMENT SCRIPT - - %Script;          -- script statements -->
<!!ATTLIST SCRIPT
  charset     %Charset;      #IMPLIED  -- char encoding of linked resource --
  type        %ContentType;  #REQUIRED -- content type of script language --
  language    CDATA          #IMPLIED  -- predefined script language name --
  src         %URI;          #IMPLIED  -- URI for an external script --
  defer       (defer)        #IMPLIED  -- UA may defer execution of script --
  event       CDATA          #IMPLIED  -- reserved for possible future use --
  for         %URI;          #IMPLIED  -- reserved for possible future use --
  >

<!!ELEMENT NOSCRIPT - - (%flow;)*
  -- alternate content container for non script-based rendering -->
<!!ATTLIST NOSCRIPT
  %attrs;                              -- %coreattrs, %i18n, %events --
  >

<!!--================ Document Structure ==================================-->
<!!ENTITY % version "version CDATA #FIXED ''%HTML.Version;''">

<!![ %HTML.Frameset; [
<!!ENTITY % html.content "HEAD, FRAMESET">
]]>

<!!ENTITY % html.content "HEAD, BODY">

<!!ELEMENT HTML O O (%html.content;)    -- document root element -->
<!!ATTLIST HTML
  %i18n;                               -- lang, dir --
  %version;
  >
'! !

!Html401Transitional methodsFor: 'initialize-release' stamp: 'tb 12/1/2005 09:34'!
initialize

	super initialize.
	self parse.! !

!HtmlDTD class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 13:04'!
default

	^self forDocType: self docType! !

!HtmlDTD class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:54'!
defaultUrl

	self subclassResponsibility! !

!HtmlDTD class methodsFor: 'as yet unclassified' stamp: 'tb 12/3/2005 15:05'!
docType

	^'abstract'! !

!HtmlDTD class methodsFor: 'as yet unclassified' stamp: 'tb 12/3/2005 15:05'!
forDocType: aDocType

	| dt delims |

	delims := (Character separators copyWith: $<) copyWith: $>.

	dt := aDocType copyWithoutAll: delims.

	Cache isNil ifTrue: [Cache := Dictionary new].

	^Cache at: dt ifAbsentPut: 
	[ 
		(self allSubclasses
			detect: [:ea | (ea docType copyWithoutAll: delims) = dt] 
			ifNone: [Html401Strict]) new.
	]! !

!HtmlDTD class methodsFor: 'as yet unclassified' stamp: 'tb 11/29/2005 20:22'!
on: aString

	^self new content: aString! !

!HtmlDTD class methodsFor: 'as yet unclassified' stamp: 'tb 11/29/2005 20:22'!
onUrl: aStringOrUrl

	^self new url: aStringOrUrl asUrl ! !

!HtmlDTD class methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 09:41'!
resetCache

	"HtmlDTD resetCache"
	Cache := nil! !

!HtmlDTD class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 10:16'!
updateFromWeb

	self compileSilently: ('dtd',(Character cr asString), '^',(self defaultUrl retrieveContents contents printString)) classified: 'resources'  ! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 21:09'!
addToAttLists: anAttList

	self attLists: (self attLists copyWith: anAttList)! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 20:56'!
addToComments: aComment

	self comments: (self comments copyWith: aComment)! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 20:59'!
addToElements: anElement

	self elements: (self elements copyWith: anElement)! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 21:08'!
addToEntities: anEntity

	self entities: (self entities copyWith: anEntity)! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 21:20'!
attLists
	"Answer the value of attLists"

	^ attLists isNil 
		ifTrue:[#()]
		ifFalse:[attLists]! !

!HtmlDTD methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
attLists: anObject
	"Set the value of attLists"

	attLists := anObject! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 12/5/2005 12:13'!
attrListForTag: aTag

	^attLists detect:[:ea | ea tagMatches: aTag] ifNone:[nil]! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 21:20'!
comments

	^comments isNil 
		ifTrue: [#()]
		ifFalse: [comments]! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 20:56'!
comments: anArray

	comments := anArray! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 12/1/2005 09:46'!
content
	"Answer the value of content"

	^ content isNil 
		ifFalse: [content]
		ifTrue: 
		[
			(self respondsTo: #dtd) 
				ifTrue: [content := self dtd]
				ifFalse: [content := self document content]
		]
	! !

!HtmlDTD methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:26'!
content: anObject
	"Set the value of content"

	content := anObject! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 12/1/2005 09:55'!
defaultUrl

	^self class defaultUrl! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 20:22'!
document
	"Answer the value of document"

	^ document isNil 
		ifFalse: [document]
		ifTrue: [url isNil ifFalse: [document := url asUrl retrieveContents] ifTrue: [nil]] ! !

!HtmlDTD methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:26'!
document: anObject
	"Set the value of document"

	document := anObject! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 20:22'!
dom
	"Answer the value of dom"

	^ dom! !

!HtmlDTD methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:27'!
dom: anObject
	"Set the value of dom"

	dom := anObject! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 12/5/2005 12:12'!
elementForTag: aTag

	^elements detect:[:ea | ea tagMatches: aTag] ifNone:[nil]! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 21:21'!
elements
	"Answer the value of elements"

	^ elements isNil 
		ifTrue:[#()]
		ifFalse:[elements]! !

!HtmlDTD methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
elements: anObject
	"Set the value of elements"

	elements := anObject! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 21:22'!
entities
	"Answer the value of entities"

	^ entities isNil 
		ifTrue:[#()]
		ifFalse: [entities]! !

!HtmlDTD methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
entities: anObject
	"Set the value of entities"

	entities := anObject! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 20:22'!
errors
	"Answer the value of errors"

	^ errors! !

!HtmlDTD methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:28'!
errors: anObject
	"Set the value of errors"

	errors := anObject! !

!HtmlDTD methodsFor: 'private' stamp: 'tb 11/30/2005 18:07'!
expandAllTokens

	self expandElementTokens.
	self expandAttListTokens.! !

!HtmlDTD methodsFor: 'private' stamp: 'tb 11/30/2005 18:06'!
expandAttListTokens

	| expanded changed |

	changed := true.
	expanded := attLists.

	[changed] whileTrue:
	[
		changed := false.
		expanded := expanded collect:
		[:item |
			item collect: 
			[:ea |
				| s |
				s := self expandTokensInLine: ea.
				s = ea ifFalse: [changed := true].
				s
			]
		]
	].
	^attLists := expanded! !

!HtmlDTD methodsFor: 'private' stamp: 'tb 11/30/2005 18:07'!
expandElementTokens

	| expanded changed |

	changed := true.
	expanded := elements.

	[changed] whileTrue:
	[
		changed := false.
		expanded := expanded collect:
		[:ea |
			| s |
			s := self expandTokensInLine: ea.
			s = ea ifFalse: [changed := true].
			s
		]
	].
	^elements := expanded! !

!HtmlDTD methodsFor: 'private' stamp: 'tb 11/30/2005 17:20'!
expandTokensInLine: aString

	| in out s |

	in := ReadStream on: aString.
	out := WriteStream on: (String new: 20).
	[in atEnd] whileFalse:
	[
		s := in upTo: $%.
		out nextPutAll: s.
		in atEnd 
			ifFalse: 
			[ 
				out nextPutAll: 
					(((self tokenDefinition: 
						(in upTo: $;)) findTokens: ' ')joinOn: ' ') 
			]
	].
	^out contents! !

!HtmlDTD methodsFor: 'initialize-release' stamp: 'tb 11/29/2005 20:22'!
initialize

	super initialize.
! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 12/5/2005 12:06'!
isXML

	^false! !

!HtmlDTD methodsFor: 'private' stamp: 'tb 11/30/2005 21:46'!
objectifyAttLists

	attLists := attLists collect: [:ea | HtmlDTDAttrList fromParsedArray: ea]! !

!HtmlDTD methodsFor: 'private' stamp: 'tb 12/6/2005 09:55'!
objectifyElements

	elements := elements collect: [:ea | HtmlDTDElement fromParsedString: ea].
	elements do: [:ea | ea cacheAllowedElementsForDTD: self].! !

!HtmlDTD methodsFor: 'actions' stamp: 'SebastianSastre 1/15/2012 16:34'!
parse

	| in s |
	
	self reset.
	in := ReadStream on: self content.

	[in atEnd] whileFalse:
	[
		in skipSeparators.
		in upTo: $<.
		s := in upTo: $ .
		(s beginsWith: '!!--') 
		ifTrue: [self readComment: in]
		ifFalse:
		[
			(s beginsWith: '!!ELEMENT') 
			ifTrue: [self readElement: in]
			ifFalse:
			[
				(s beginsWith: '!!ATTLIST')
				ifTrue: [self readAttList: in]
				ifFalse:
				[
					(s beginsWith: '!!ENTITY')
					ifTrue:[self readEntity: in]
					ifFalse:
					[
						(s beginsWith: '!![')
						ifTrue:[self readConditionalElement: in]
						ifFalse:[s trimBoth size > 0 ifTrue: [self halt]]
					] 
				] 
			]
		]
	].

	self expandAllTokens.
	self objectifyAttLists.
	self objectifyElements.
	
	! !

!HtmlDTD methodsFor: 'actions' stamp: 'SebastianSastre 1/15/2012 16:34'!
readAttList: aStream

	| s list |

	list := OrderedCollection new.
	s := (aStream upTo: $>) trimBoth findTokens: Character separators.

	list add: s first.
	s := s copyWithoutFirst joinOn: ' '.
	list add: (self stripCommentsFrom: s).
	self addToAttLists: list asArray.! !

!HtmlDTD methodsFor: 'actions' stamp: 'tb 11/29/2005 21:17'!
readComment: aStream

	| s |

	s := aStream upTo: $>.
	[s endsWith: '--'] whileFalse: [s:= (s, '>',(aStream upTo: $>))].
	self addToComments: s.! !

!HtmlDTD methodsFor: 'actions' stamp: 'tb 11/29/2005 21:31'!
readConditionalElement: aStream

	| s |

	s := aStream upTo: $>.
	[s endsWith: ']]'] whileFalse: [s:= (s, '>',(aStream upTo: $>))].
! !

!HtmlDTD methodsFor: 'actions' stamp: 'SebastianSastre 1/15/2012 16:34'!
readElement: aStream

	| s |

	s := aStream upTo: $>.
	self addToElements: (self stripCommentsFrom: s) trimBoth.! !

!HtmlDTD methodsFor: 'actions' stamp: 'SebastianSastre 1/15/2012 16:33'!
readEntity: aStream

	| s in out |

	out := WriteStream on: (String new: 100).

	s := (self stripCommentsFrom: (aStream upTo: $>)) trimBoth.

	in := ReadStream on: s.
	in upTo: $%.
	out nextPutAll: '% '.

	in skipSeparators.
	[in peek isSeparator]whileFalse:[out nextPut: in next].

	in skipSeparators.
	out nextPut: $ .
	out nextPutAll: in upToEnd.
	self addToEntities: out contents trimBoth.
	! !

!HtmlDTD methodsFor: 'actions' stamp: 'tb 11/29/2005 21:22'!
reset

	entities := attributes := comments := elements := attLists := nil! !

!HtmlDTD methodsFor: 'private' stamp: 'tb 12/2/2005 17:08'!
stripCommentsFrom: aString

	| in out |

	in := ReadStream on: aString.
	out := WriteStream on: (String new: aString size).

	[in atEnd] whileFalse:
	[
		out nextPutAll: (in upToAll: '--').
		in upToAll: '--'.
	].
	^out contents! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 11/30/2005 23:16'!
tag: aTag allowsAttribute: anAttribute

	| a |
	a := (self attrListForTag: aTag).
	^a isNil 
		ifFalse: [(a attributeNamed: anAttribute) notNil] 
		ifTrue:[false]! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 12/1/2005 14:27'!
tag: aTag attribute: anAttribute hasValidValue: aValue

	| a adef |
	a := (self attrListForTag: aTag).
	^a isNil 
		ifTrue:[false]
		ifFalse: 
		[
			adef := a attributeNamed: anAttribute.
			adef isNil 
				ifTrue:[false]
				ifFalse:
				[
					adef type isString 
						ifTrue: [true]
						ifFalse: [adef type allowsItem: aValue]
				]
		] 
! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 12/6/2005 15:52'!
tag: aTag canContainTag: anotherTag

	| element |
		element := self elementForTag: aTag asUppercase.
		^element notNil 
			ifTrue: [element allowsTag: anotherTag asUppercase]
			ifFalse:[false]
		! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 12/2/2005 10:36'!
tagCloseIsOptional: aTag

	| element |

	element := self elementForTag: aTag.
	^element isNil 
		ifTrue:[false] 
		ifFalse:[element endIsOptional]! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 12/2/2005 12:22'!
tagIsEmpty: aTag

	| e |

	e := elements detect: [:ea | ea tagMatches: aTag] ifNone: [nil].
	^e notNil ifTrue: [e isEmpty] ifFalse: [false]! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 11/29/2005 23:34'!
tagIsOptional: aTag

	| decl |
	decl := (self elements detect: [:ea | ea beginsWith: (aTag asUppercase)]).
	^((decl splitOn: ' ') at: 2) = 'O'.! !

!HtmlDTD methodsFor: 'testing' stamp: 'tb 12/2/2005 09:50'!
tagIsReal: aTag

	^(self elementForTag: aTag) notNil! !

!HtmlDTD methodsFor: 'private' stamp: 'tb 12/2/2005 12:48'!
tokenDefinition: aToken

	| def fmt s in |

	fmt := (aToken copyWithout: $%)	copyWithout: $;.
	fmt := ('% ',fmt, ' ').
	def := entities detect: [:ea | ea beginsWith: fmt] ifNone: ['"', aToken, '"'].
	in := ReadStream on: def.
	in upTo: $".
	s := in upTo: $".
	^s	! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/30/2005 23:04'!
url
	"Answer the value of url"

	^ url isNil ifTrue: [url := self defaultUrl]! !

!HtmlDTD methodsFor: 'accessing' stamp: 'tb 11/29/2005 20:22'!
url: anObject
	"Set the value of url"

	url := anObject.
! !

!HtmlDTD methodsFor: 'actions' stamp: 'tb 1/3/2006 05:42'!
validateAttributesOfNode: aNode

	aNode attributes copy keysAndValuesDo:
	[:k :v || attList lck |
		attList := self attrListForTag: aNode dtdTag.
		lck := k asLowercase.
		(attList includesAttribute: lck) ifFalse:
		[
			aNode addError: (HtmlAttributeUnknownError new attribute: k).
		]
		ifTrue:
		[
			(k = lck) 
			ifFalse: 
			[
				aNode addError: (HtmlAttributeNotLowerCaseError new attribute: k).
				aNode attributes at: lck put: v.
				aNode attributes removeKey: k.
			].
			v isNil ifTrue:
			[
				(self isXML or: [((attList attributeNamed: lck) allowsValue: lck) not]) ifTrue:
				[
					aNode addError: (HtmlAttributeValueMissingError new attribute: k).
				]
			]
			ifFalse:
			[
				((attList attributeNamed: lck) allowsValue: v) ifFalse:
				[
					aNode addError: (HtmlAttributeValueNotValidError new attribute: k).
				]
			]
		]					
	]
! !

!HtmlDTD methodsFor: 'actions' stamp: 'tb 12/16/2005 23:29'!
validateChildrenOfNode: aNode

	| element children stream |

	(aNode isComment or: [aNode isCDATA]) ifTrue: [^self].
	element := self elementForTag: aNode dtdTag.
	children := aNode children reject: [:ea | ea isComment].	

	stream := ReadStream on: children.
	(element contentRule validateChildren: stream ofNode: aNode)  
		ifTrue:[]	
		ifFalse:
		[
			stream atEnd ifFalse: 
			[
				(self tag: aNode dtdTag canContainTag: stream peek dtdTag) ifFalse:
				[
					stream peek addError: (HtmlTagImproperlyNestedError new)
				]
			].
		]
! !

HtmlDTD subclass: #XHtmlDTD
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

XHtmlDTD subclass: #XHtml1Strict
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!XHtml1Strict class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:53'!
defaultUrl

	^'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd' asUrl! !

!XHtml1Strict class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:52'!
docType

	^'<!!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'! !

!XHtml1Strict methodsFor: 'resources' stamp: 'tb 12/3/2005 14:45'!
dtd
	^ '<!!--
   Extensible HTML version 1.0 Strict DTD

   This is the same as HTML 4 Strict except for
   changes due to the differences between XML and SGML.

   Namespace = http://www.w3.org/1999/xhtml

   For further information, see: http://www.w3.org/TR/xhtml1

   Copyright (c) 1998-2002 W3C (MIT, INRIA, Keio),
   All Rights Reserved. 

   This DTD module is identified by the PUBLIC and SYSTEM identifiers:

   PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
   SYSTEM "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd"

   $Revision: 1.1 $
   $Date: 2002/08/01 13:56:03 $

-->

<!!--================ Character mnemonic entities =========================-->

<!!ENTITY % HTMLlat1 PUBLIC
   "-//W3C//ENTITIES Latin 1 for XHTML//EN"
   "xhtml-lat1.ent">
%HTMLlat1;

<!!ENTITY % HTMLsymbol PUBLIC
   "-//W3C//ENTITIES Symbols for XHTML//EN"
   "xhtml-symbol.ent">
%HTMLsymbol;

<!!ENTITY % HTMLspecial PUBLIC
   "-//W3C//ENTITIES Special for XHTML//EN"
   "xhtml-special.ent">
%HTMLspecial;

<!!--================== Imported Names ====================================-->

<!!ENTITY % ContentType "CDATA">
    <!!-- media type, as per [RFC2045] -->

<!!ENTITY % ContentTypes "CDATA">
    <!!-- comma-separated list of media types, as per [RFC2045] -->

<!!ENTITY % Charset "CDATA">
    <!!-- a character encoding, as per [RFC2045] -->

<!!ENTITY % Charsets "CDATA">
    <!!-- a space separated list of character encodings, as per [RFC2045] -->

<!!ENTITY % LanguageCode "NMTOKEN">
    <!!-- a language code, as per [RFC3066] -->

<!!ENTITY % Character "CDATA">
    <!!-- a single character, as per section 2.2 of [XML] -->

<!!ENTITY % Number "CDATA">
    <!!-- one or more digits -->

<!!ENTITY % LinkTypes "CDATA">
    <!!-- space-separated list of link types -->

<!!ENTITY % MediaDesc "CDATA">
    <!!-- single or comma-separated list of media descriptors -->

<!!ENTITY % URI "CDATA">
    <!!-- a Uniform Resource Identifier, see [RFC2396] -->

<!!ENTITY % UriList "CDATA">
    <!!-- a space separated list of Uniform Resource Identifiers -->

<!!ENTITY % Datetime "CDATA">
    <!!-- date and time information. ISO date format -->

<!!ENTITY % Script "CDATA">
    <!!-- script expression -->

<!!ENTITY % StyleSheet "CDATA">
    <!!-- style sheet data -->

<!!ENTITY % Text "CDATA">
    <!!-- used for titles etc. -->

<!!ENTITY % Length "CDATA">
    <!!-- nn for pixels or nn% for percentage length -->

<!!ENTITY % MultiLength "CDATA">
    <!!-- pixel, percentage, or relative -->

<!!ENTITY % Pixels "CDATA">
    <!!-- integer representing length in pixels -->

<!!-- these are used for image maps -->

<!!ENTITY % Shape "(rect|circle|poly|default)">

<!!ENTITY % Coords "CDATA">
    <!!-- comma separated list of lengths -->

<!!--=================== Generic Attributes ===============================-->

<!!-- core attributes common to most elements
  id       document-wide unique id
  class    space separated list of classes
  style    associated style info
  title    advisory title/amplification
-->
<!!ENTITY % coreattrs
 "id          ID             #IMPLIED
  class       CDATA          #IMPLIED
  style       %StyleSheet;   #IMPLIED
  title       %Text;         #IMPLIED"
  >

<!!-- internationalization attributes
  lang        language code (backwards compatible)
  xml:lang    language code (as per XML 1.0 spec)
  dir         direction for weak/neutral text
-->
<!!ENTITY % i18n
 "lang        %LanguageCode; #IMPLIED
  xml:lang    %LanguageCode; #IMPLIED
  dir         (ltr|rtl)      #IMPLIED"
  >

<!!-- attributes for common UI events
  onclick     a pointer button was clicked
  ondblclick  a pointer button was double clicked
  onmousedown a pointer button was pressed down
  onmouseup   a pointer button was released
  onmousemove a pointer was moved onto the element
  onmouseout  a pointer was moved away from the element
  onkeypress  a key was pressed and released
  onkeydown   a key was pressed down
  onkeyup     a key was released
-->
<!!ENTITY % events
 "onclick     %Script;       #IMPLIED
  ondblclick  %Script;       #IMPLIED
  onmousedown %Script;       #IMPLIED
  onmouseup   %Script;       #IMPLIED
  onmouseover %Script;       #IMPLIED
  onmousemove %Script;       #IMPLIED
  onmouseout  %Script;       #IMPLIED
  onkeypress  %Script;       #IMPLIED
  onkeydown   %Script;       #IMPLIED
  onkeyup     %Script;       #IMPLIED"
  >

<!!-- attributes for elements that can get the focus
  accesskey   accessibility key character
  tabindex    position in tabbing order
  onfocus     the element got the focus
  onblur      the element lost the focus
-->
<!!ENTITY % focus
 "accesskey   %Character;    #IMPLIED
  tabindex    %Number;       #IMPLIED
  onfocus     %Script;       #IMPLIED
  onblur      %Script;       #IMPLIED"
  >

<!!ENTITY % attrs "%coreattrs; %i18n; %events;">

<!!--=================== Text Elements ====================================-->

<!!ENTITY % special.pre
   "br | span | bdo | map">


<!!ENTITY % special
   "%special.pre; | object | img ">

<!!ENTITY % fontstyle "tt | i | b | big | small ">

<!!ENTITY % phrase "em | strong | dfn | code | q |
                   samp | kbd | var | cite | abbr | acronym | sub | sup ">

<!!ENTITY % inline.forms "input | select | textarea | label | button">

<!!-- these can occur at block or inline level -->
<!!ENTITY % misc.inline "ins | del | script">

<!!-- these can only occur at block level -->
<!!ENTITY % misc "noscript | %misc.inline;">

<!!ENTITY % inline "a | %special; | %fontstyle; | %phrase; | %inline.forms;">

<!!-- %Inline; covers inline or "text-level" elements -->
<!!ENTITY % Inline "(#PCDATA | %inline; | %misc.inline;)*">

<!!--================== Block level elements ==============================-->

<!!ENTITY % heading "h1|h2|h3|h4|h5|h6">
<!!ENTITY % lists "ul | ol | dl">
<!!ENTITY % blocktext "pre | hr | blockquote | address">

<!!ENTITY % block
     "p | %heading; | div | %lists; | %blocktext; | fieldset | table">

<!!ENTITY % Block "(%block; | form | %misc;)*">

<!!-- %Flow; mixes block and inline and is used for list items etc. -->
<!!ENTITY % Flow "(#PCDATA | %block; | form | %inline; | %misc;)*">

<!!--================== Content models for exclusions =====================-->

<!!-- a elements use %Inline; excluding a -->

<!!ENTITY % a.content
   "(#PCDATA | %special; | %fontstyle; | %phrase; | %inline.forms; | %misc.inline;)*">

<!!-- pre uses %Inline excluding big, small, sup or sup -->

<!!ENTITY % pre.content
   "(#PCDATA | a | %fontstyle; | %phrase; | %special.pre; | %misc.inline;
      | %inline.forms;)*">

<!!-- form uses %Block; excluding form -->

<!!ENTITY % form.content "(%block; | %misc;)*">

<!!-- button uses %Flow; but excludes a, form and form controls -->

<!!ENTITY % button.content
   "(#PCDATA | p | %heading; | div | %lists; | %blocktext; |
    table | %special; | %fontstyle; | %phrase; | %misc;)*">

<!!--================ Document Structure ==================================-->

<!!-- the namespace URI designates the document profile -->

<!!ELEMENT html (head, body)>
<!!ATTLIST html
  %i18n;
  id          ID             #IMPLIED
  xmlns       %URI;          #FIXED ''http://www.w3.org/1999/xhtml''
  >

<!!--================ Document Head =======================================-->

<!!ENTITY % head.misc "(script|style|meta|link|object)*">

<!!-- content model is %head.misc; combined with a single
     title and an optional base element in any order -->

<!!ELEMENT head (%head.misc;,
     ((title, %head.misc;, (base, %head.misc;)?) |
      (base, %head.misc;, (title, %head.misc;))))>

<!!ATTLIST head
  %i18n;
  id          ID             #IMPLIED
  profile     %URI;          #IMPLIED
  >

<!!-- The title element is not considered part of the flow of text.
       It should be displayed, for example as the page header or
       window title. Exactly one title is required per document.
    -->
<!!ELEMENT title (#PCDATA)>
<!!ATTLIST title 
  %i18n;
  id          ID             #IMPLIED
  >

<!!-- document base URI -->

<!!ELEMENT base EMPTY>
<!!ATTLIST base
  href        %URI;          #REQUIRED
  id          ID             #IMPLIED
  >

<!!-- generic metainformation -->
<!!ELEMENT meta EMPTY>
<!!ATTLIST meta
  %i18n;
  id          ID             #IMPLIED
  http-equiv  CDATA          #IMPLIED
  name        CDATA          #IMPLIED
  content     CDATA          #REQUIRED
  scheme      CDATA          #IMPLIED
  >

<!!--
  Relationship values can be used in principle:

   a) for document specific toolbars/menus when used
      with the link element in document head e.g.
        start, contents, previous, next, index, end, help
   b) to link to a separate style sheet (rel="stylesheet")
   c) to make a link to a script (rel="script")
   d) by stylesheets to control how collections of
      html nodes are rendered into printed documents
   e) to make a link to a printable version of this document
      e.g. a PostScript or PDF version (rel="alternate" media="print")
-->

<!!ELEMENT link EMPTY>
<!!ATTLIST link
  %attrs;
  charset     %Charset;      #IMPLIED
  href        %URI;          #IMPLIED
  hreflang    %LanguageCode; #IMPLIED
  type        %ContentType;  #IMPLIED
  rel         %LinkTypes;    #IMPLIED
  rev         %LinkTypes;    #IMPLIED
  media       %MediaDesc;    #IMPLIED
  >

<!!-- style info, which may include CDATA sections -->
<!!ELEMENT style (#PCDATA)>
<!!ATTLIST style
  %i18n;
  id          ID             #IMPLIED
  type        %ContentType;  #REQUIRED
  media       %MediaDesc;    #IMPLIED
  title       %Text;         #IMPLIED
  xml:space   (preserve)     #FIXED ''preserve''
  >

<!!-- script statements, which may include CDATA sections -->
<!!ELEMENT script (#PCDATA)>
<!!ATTLIST script
  id          ID             #IMPLIED
  charset     %Charset;      #IMPLIED
  type        %ContentType;  #REQUIRED
  src         %URI;          #IMPLIED
  defer       (defer)        #IMPLIED
  xml:space   (preserve)     #FIXED ''preserve''
  >

<!!-- alternate content container for non script-based rendering -->

<!!ELEMENT noscript %Block;>
<!!ATTLIST noscript
  %attrs;
  >

<!!--=================== Document Body ====================================-->

<!!ELEMENT body %Block;>
<!!ATTLIST body
  %attrs;
  onload          %Script;   #IMPLIED
  onunload        %Script;   #IMPLIED
  >

<!!ELEMENT div %Flow;>  <!!-- generic language/style container -->
<!!ATTLIST div
  %attrs;
  >

<!!--=================== Paragraphs =======================================-->

<!!ELEMENT p %Inline;>
<!!ATTLIST p
  %attrs;
  >

<!!--=================== Headings =========================================-->

<!!--
  There are six levels of headings from h1 (the most important)
  to h6 (the least important).
-->

<!!ELEMENT h1  %Inline;>
<!!ATTLIST h1
   %attrs;
   >

<!!ELEMENT h2 %Inline;>
<!!ATTLIST h2
   %attrs;
   >

<!!ELEMENT h3 %Inline;>
<!!ATTLIST h3
   %attrs;
   >

<!!ELEMENT h4 %Inline;>
<!!ATTLIST h4
   %attrs;
   >

<!!ELEMENT h5 %Inline;>
<!!ATTLIST h5
   %attrs;
   >

<!!ELEMENT h6 %Inline;>
<!!ATTLIST h6
   %attrs;
   >

<!!--=================== Lists ============================================-->

<!!-- Unordered list -->

<!!ELEMENT ul (li)+>
<!!ATTLIST ul
  %attrs;
  >

<!!-- Ordered (numbered) list -->

<!!ELEMENT ol (li)+>
<!!ATTLIST ol
  %attrs;
  >

<!!-- list item -->

<!!ELEMENT li %Flow;>
<!!ATTLIST li
  %attrs;
  >

<!!-- definition lists - dt for term, dd for its definition -->

<!!ELEMENT dl (dt|dd)+>
<!!ATTLIST dl
  %attrs;
  >

<!!ELEMENT dt %Inline;>
<!!ATTLIST dt
  %attrs;
  >

<!!ELEMENT dd %Flow;>
<!!ATTLIST dd
  %attrs;
  >

<!!--=================== Address ==========================================-->

<!!-- information on author -->

<!!ELEMENT address %Inline;>
<!!ATTLIST address
  %attrs;
  >

<!!--=================== Horizontal Rule ==================================-->

<!!ELEMENT hr EMPTY>
<!!ATTLIST hr
  %attrs;
  >

<!!--=================== Preformatted Text ================================-->

<!!-- content is %Inline; excluding "img|object|big|small|sub|sup" -->

<!!ELEMENT pre %pre.content;>
<!!ATTLIST pre
  %attrs;
  xml:space (preserve) #FIXED ''preserve''
  >

<!!--=================== Block-like Quotes ================================-->

<!!ELEMENT blockquote %Block;>
<!!ATTLIST blockquote
  %attrs;
  cite        %URI;          #IMPLIED
  >

<!!--=================== Inserted/Deleted Text ============================-->

<!!--
  ins/del are allowed in block and inline content, but its
  inappropriate to include block content within an ins element
  occurring in inline content.
-->
<!!ELEMENT ins %Flow;>
<!!ATTLIST ins
  %attrs;
  cite        %URI;          #IMPLIED
  datetime    %Datetime;     #IMPLIED
  >

<!!ELEMENT del %Flow;>
<!!ATTLIST del
  %attrs;
  cite        %URI;          #IMPLIED
  datetime    %Datetime;     #IMPLIED
  >

<!!--================== The Anchor Element ================================-->

<!!-- content is %Inline; except that anchors shouldn''t be nested -->

<!!ELEMENT a %a.content;>
<!!ATTLIST a
  %attrs;
  %focus;
  charset     %Charset;      #IMPLIED
  type        %ContentType;  #IMPLIED
  name        NMTOKEN        #IMPLIED
  href        %URI;          #IMPLIED
  hreflang    %LanguageCode; #IMPLIED
  rel         %LinkTypes;    #IMPLIED
  rev         %LinkTypes;    #IMPLIED
  shape       %Shape;        "rect"
  coords      %Coords;       #IMPLIED
  >

<!!--===================== Inline Elements ================================-->

<!!ELEMENT span %Inline;> <!!-- generic language/style container -->
<!!ATTLIST span
  %attrs;
  >

<!!ELEMENT bdo %Inline;>  <!!-- I18N BiDi over-ride -->
<!!ATTLIST bdo
  %coreattrs;
  %events;
  lang        %LanguageCode; #IMPLIED
  xml:lang    %LanguageCode; #IMPLIED
  dir         (ltr|rtl)      #REQUIRED
  >

<!!ELEMENT br EMPTY>   <!!-- forced line break -->
<!!ATTLIST br
  %coreattrs;
  >

<!!ELEMENT em %Inline;>   <!!-- emphasis -->
<!!ATTLIST em %attrs;>

<!!ELEMENT strong %Inline;>   <!!-- strong emphasis -->
<!!ATTLIST strong %attrs;>

<!!ELEMENT dfn %Inline;>   <!!-- definitional -->
<!!ATTLIST dfn %attrs;>

<!!ELEMENT code %Inline;>   <!!-- program code -->
<!!ATTLIST code %attrs;>

<!!ELEMENT samp %Inline;>   <!!-- sample -->
<!!ATTLIST samp %attrs;>

<!!ELEMENT kbd %Inline;>  <!!-- something user would type -->
<!!ATTLIST kbd %attrs;>

<!!ELEMENT var %Inline;>   <!!-- variable -->
<!!ATTLIST var %attrs;>

<!!ELEMENT cite %Inline;>   <!!-- citation -->
<!!ATTLIST cite %attrs;>

<!!ELEMENT abbr %Inline;>   <!!-- abbreviation -->
<!!ATTLIST abbr %attrs;>

<!!ELEMENT acronym %Inline;>   <!!-- acronym -->
<!!ATTLIST acronym %attrs;>

<!!ELEMENT q %Inline;>   <!!-- inlined quote -->
<!!ATTLIST q
  %attrs;
  cite        %URI;          #IMPLIED
  >

<!!ELEMENT sub %Inline;> <!!-- subscript -->
<!!ATTLIST sub %attrs;>

<!!ELEMENT sup %Inline;> <!!-- superscript -->
<!!ATTLIST sup %attrs;>

<!!ELEMENT tt %Inline;>   <!!-- fixed pitch font -->
<!!ATTLIST tt %attrs;>

<!!ELEMENT i %Inline;>   <!!-- italic font -->
<!!ATTLIST i %attrs;>

<!!ELEMENT b %Inline;>   <!!-- bold font -->
<!!ATTLIST b %attrs;>

<!!ELEMENT big %Inline;>   <!!-- bigger font -->
<!!ATTLIST big %attrs;>

<!!ELEMENT small %Inline;>   <!!-- smaller font -->
<!!ATTLIST small %attrs;>

<!!--==================== Object ======================================-->
<!!--
  object is used to embed objects as part of HTML pages.
  param elements should precede other content. Parameters
  can also be expressed as attribute/value pairs on the
  object element itself when brevity is desired.
-->

<!!ELEMENT object (#PCDATA | param | %block; | form | %inline; | %misc;)*>
<!!ATTLIST object
  %attrs;
  declare     (declare)      #IMPLIED
  classid     %URI;          #IMPLIED
  codebase    %URI;          #IMPLIED
  data        %URI;          #IMPLIED
  type        %ContentType;  #IMPLIED
  codetype    %ContentType;  #IMPLIED
  archive     %UriList;      #IMPLIED
  standby     %Text;         #IMPLIED
  height      %Length;       #IMPLIED
  width       %Length;       #IMPLIED
  usemap      %URI;          #IMPLIED
  name        NMTOKEN        #IMPLIED
  tabindex    %Number;       #IMPLIED
  >

<!!--
  param is used to supply a named property value.
  In XML it would seem natural to follow RDF and support an
  abbreviated syntax where the param elements are replaced
  by attribute value pairs on the object start tag.
-->
<!!ELEMENT param EMPTY>
<!!ATTLIST param
  id          ID             #IMPLIED
  name        CDATA          #IMPLIED
  value       CDATA          #IMPLIED
  valuetype   (data|ref|object) "data"
  type        %ContentType;  #IMPLIED
  >

<!!--=================== Images ===========================================-->

<!!--
   To avoid accessibility problems for people who aren''t
   able to see the image, you should provide a text
   description using the alt and longdesc attributes.
   In addition, avoid the use of server-side image maps.
   Note that in this DTD there is no name attribute. That
   is only available in the transitional and frameset DTD.
-->

<!!ELEMENT img EMPTY>
<!!ATTLIST img
  %attrs;
  src         %URI;          #REQUIRED
  alt         %Text;         #REQUIRED
  longdesc    %URI;          #IMPLIED
  height      %Length;       #IMPLIED
  width       %Length;       #IMPLIED
  usemap      %URI;          #IMPLIED
  ismap       (ismap)        #IMPLIED
  >

<!!-- usemap points to a map element which may be in this document
  or an external document, although the latter is not widely supported -->

<!!--================== Client-side image maps ============================-->

<!!-- These can be placed in the same document or grouped in a
     separate document although this isn''t yet widely supported -->

<!!ELEMENT map ((%block; | form | %misc;)+ | area+)>
<!!ATTLIST map
  %i18n;
  %events;
  id          ID             #REQUIRED
  class       CDATA          #IMPLIED
  style       %StyleSheet;   #IMPLIED
  title       %Text;         #IMPLIED
  name        NMTOKEN        #IMPLIED
  >

<!!ELEMENT area EMPTY>
<!!ATTLIST area
  %attrs;
  %focus;
  shape       %Shape;        "rect"
  coords      %Coords;       #IMPLIED
  href        %URI;          #IMPLIED
  nohref      (nohref)       #IMPLIED
  alt         %Text;         #REQUIRED
  >

<!!--================ Forms ===============================================-->
<!!ELEMENT form %form.content;>   <!!-- forms shouldn''t be nested -->

<!!ATTLIST form
  %attrs;
  action      %URI;          #REQUIRED
  method      (get|post)     "get"
  enctype     %ContentType;  "application/x-www-form-urlencoded"
  onsubmit    %Script;       #IMPLIED
  onreset     %Script;       #IMPLIED
  accept      %ContentTypes; #IMPLIED
  accept-charset %Charsets;  #IMPLIED
  >

<!!--
  Each label must not contain more than ONE field
  Label elements shouldn''t be nested.
-->
<!!ELEMENT label %Inline;>
<!!ATTLIST label
  %attrs;
  for         IDREF          #IMPLIED
  accesskey   %Character;    #IMPLIED
  onfocus     %Script;       #IMPLIED
  onblur      %Script;       #IMPLIED
  >

<!!ENTITY % InputType
  "(text | password | checkbox |
    radio | submit | reset |
    file | hidden | image | button)"
   >

<!!-- the name attribute is required for all but submit & reset -->

<!!ELEMENT input EMPTY>     <!!-- form control -->
<!!ATTLIST input
  %attrs;
  %focus;
  type        %InputType;    "text"
  name        CDATA          #IMPLIED
  value       CDATA          #IMPLIED
  checked     (checked)      #IMPLIED
  disabled    (disabled)     #IMPLIED
  readonly    (readonly)     #IMPLIED
  size        CDATA          #IMPLIED
  maxlength   %Number;       #IMPLIED
  src         %URI;          #IMPLIED
  alt         CDATA          #IMPLIED
  usemap      %URI;          #IMPLIED
  onselect    %Script;       #IMPLIED
  onchange    %Script;       #IMPLIED
  accept      %ContentTypes; #IMPLIED
  >

<!!ELEMENT select (optgroup|option)+>  <!!-- option selector -->
<!!ATTLIST select
  %attrs;
  name        CDATA          #IMPLIED
  size        %Number;       #IMPLIED
  multiple    (multiple)     #IMPLIED
  disabled    (disabled)     #IMPLIED
  tabindex    %Number;       #IMPLIED
  onfocus     %Script;       #IMPLIED
  onblur      %Script;       #IMPLIED
  onchange    %Script;       #IMPLIED
  >

<!!ELEMENT optgroup (option)+>   <!!-- option group -->
<!!ATTLIST optgroup
  %attrs;
  disabled    (disabled)     #IMPLIED
  label       %Text;         #REQUIRED
  >

<!!ELEMENT option (#PCDATA)>     <!!-- selectable choice -->
<!!ATTLIST option
  %attrs;
  selected    (selected)     #IMPLIED
  disabled    (disabled)     #IMPLIED
  label       %Text;         #IMPLIED
  value       CDATA          #IMPLIED
  >

<!!ELEMENT textarea (#PCDATA)>     <!!-- multi-line text field -->
<!!ATTLIST textarea
  %attrs;
  %focus;
  name        CDATA          #IMPLIED
  rows        %Number;       #REQUIRED
  cols        %Number;       #REQUIRED
  disabled    (disabled)     #IMPLIED
  readonly    (readonly)     #IMPLIED
  onselect    %Script;       #IMPLIED
  onchange    %Script;       #IMPLIED
  >

<!!--
  The fieldset element is used to group form fields.
  Only one legend element should occur in the content
  and if present should only be preceded by whitespace.
-->
<!!ELEMENT fieldset (#PCDATA | legend | %block; | form | %inline; | %misc;)*>
<!!ATTLIST fieldset
  %attrs;
  >

<!!ELEMENT legend %Inline;>     <!!-- fieldset label -->
<!!ATTLIST legend
  %attrs;
  accesskey   %Character;    #IMPLIED
  >

<!!--
 Content is %Flow; excluding a, form and form controls
--> 
<!!ELEMENT button %button.content;>  <!!-- push button -->
<!!ATTLIST button
  %attrs;
  %focus;
  name        CDATA          #IMPLIED
  value       CDATA          #IMPLIED
  type        (button|submit|reset) "submit"
  disabled    (disabled)     #IMPLIED
  >

<!!--======================= Tables =======================================-->

<!!-- Derived from IETF HTML table standard, see [RFC1942] -->

<!!--
 The border attribute sets the thickness of the frame around the
 table. The default units are screen pixels.

 The frame attribute specifies which parts of the frame around
 the table should be rendered. The values are not the same as
 CALS to avoid a name clash with the valign attribute.
-->
<!!ENTITY % TFrame "(void|above|below|hsides|lhs|rhs|vsides|box|border)">

<!!--
 The rules attribute defines which rules to draw between cells:

 If rules is absent then assume:
     "none" if border is absent or border="0" otherwise "all"
-->

<!!ENTITY % TRules "(none | groups | rows | cols | all)">
  
<!!-- horizontal alignment attributes for cell contents

  char        alignment char, e.g. char='':''
  charoff     offset for alignment char
-->
<!!ENTITY % cellhalign
  "align      (left|center|right|justify|char) #IMPLIED
   char       %Character;    #IMPLIED
   charoff    %Length;       #IMPLIED"
  >

<!!-- vertical alignment attributes for cell contents -->
<!!ENTITY % cellvalign
  "valign     (top|middle|bottom|baseline) #IMPLIED"
  >

<!!ELEMENT table
     (caption?, (col*|colgroup*), thead?, tfoot?, (tbody+|tr+))>
<!!ELEMENT caption  %Inline;>
<!!ELEMENT thead    (tr)+>
<!!ELEMENT tfoot    (tr)+>
<!!ELEMENT tbody    (tr)+>
<!!ELEMENT colgroup (col)*>
<!!ELEMENT col      EMPTY>
<!!ELEMENT tr       (th|td)+>
<!!ELEMENT th       %Flow;>
<!!ELEMENT td       %Flow;>

<!!ATTLIST table
  %attrs;
  summary     %Text;         #IMPLIED
  width       %Length;       #IMPLIED
  border      %Pixels;       #IMPLIED
  frame       %TFrame;       #IMPLIED
  rules       %TRules;       #IMPLIED
  cellspacing %Length;       #IMPLIED
  cellpadding %Length;       #IMPLIED
  >

<!!ATTLIST caption
  %attrs;
  >

<!!--
colgroup groups a set of col elements. It allows you to group
several semantically related columns together.
-->
<!!ATTLIST colgroup
  %attrs;
  span        %Number;       "1"
  width       %MultiLength;  #IMPLIED
  %cellhalign;
  %cellvalign;
  >

<!!--
 col elements define the alignment properties for cells in
 one or more columns.

 The width attribute specifies the width of the columns, e.g.

     width=64        width in screen pixels
     width=0.5*      relative width of 0.5

 The span attribute causes the attributes of one
 col element to apply to more than one column.
-->
<!!ATTLIST col
  %attrs;
  span        %Number;       "1"
  width       %MultiLength;  #IMPLIED
  %cellhalign;
  %cellvalign;
  >

<!!--
    Use thead to duplicate headers when breaking table
    across page boundaries, or for static headers when
    tbody sections are rendered in scrolling panel.

    Use tfoot to duplicate footers when breaking table
    across page boundaries, or for static footers when
    tbody sections are rendered in scrolling panel.

    Use multiple tbody sections when rules are needed
    between groups of table rows.
-->
<!!ATTLIST thead
  %attrs;
  %cellhalign;
  %cellvalign;
  >

<!!ATTLIST tfoot
  %attrs;
  %cellhalign;
  %cellvalign;
  >

<!!ATTLIST tbody
  %attrs;
  %cellhalign;
  %cellvalign;
  >

<!!ATTLIST tr
  %attrs;
  %cellhalign;
  %cellvalign;
  >


<!!-- Scope is simpler than headers attribute for common tables -->
<!!ENTITY % Scope "(row|col|rowgroup|colgroup)">

<!!-- th is for headers, td for data and for cells acting as both -->

<!!ATTLIST th
  %attrs;
  abbr        %Text;         #IMPLIED
  axis        CDATA          #IMPLIED
  headers     IDREFS         #IMPLIED
  scope       %Scope;        #IMPLIED
  rowspan     %Number;       "1"
  colspan     %Number;       "1"
  %cellhalign;
  %cellvalign;
  >

<!!ATTLIST td
  %attrs;
  abbr        %Text;         #IMPLIED
  axis        CDATA          #IMPLIED
  headers     IDREFS         #IMPLIED
  scope       %Scope;        #IMPLIED
  rowspan     %Number;       "1"
  colspan     %Number;       "1"
  %cellhalign;
  %cellvalign;
  >


'! !

!XHtml1Strict methodsFor: 'initialize-release' stamp: 'tb 12/1/2005 09:50'!
initialize

	super initialize.
	self parse.! !

XHtmlDTD subclass: #XHtml1Transitional
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!XHtml1Transitional class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:51'!
defaultUrl

	^'http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd' asUrl! !

!XHtml1Transitional class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:48'!
docType

	^'<!!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'! !

!XHtml1Transitional methodsFor: 'resources' stamp: ''!
dtd
	^ '<!!--
   Extensible HTML version 1.0 Transitional DTD

   This is the same as HTML 4 Transitional except for
   changes due to the differences between XML and SGML.

   Namespace = http://www.w3.org/1999/xhtml

   For further information, see: http://www.w3.org/TR/xhtml1

   Copyright (c) 1998-2002 W3C (MIT, INRIA, Keio),
   All Rights Reserved. 

   This DTD module is identified by the PUBLIC and SYSTEM identifiers:

   PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
   SYSTEM "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd"

   $Revision: 1.2 $
   $Date: 2002/08/01 18:37:55 $

-->

<!!--================ Character mnemonic entities =========================-->

<!!ENTITY % HTMLlat1 PUBLIC
   "-//W3C//ENTITIES Latin 1 for XHTML//EN"
   "xhtml-lat1.ent">
%HTMLlat1;

<!!ENTITY % HTMLsymbol PUBLIC
   "-//W3C//ENTITIES Symbols for XHTML//EN"
   "xhtml-symbol.ent">
%HTMLsymbol;

<!!ENTITY % HTMLspecial PUBLIC
   "-//W3C//ENTITIES Special for XHTML//EN"
   "xhtml-special.ent">
%HTMLspecial;

<!!--================== Imported Names ====================================-->

<!!ENTITY % ContentType "CDATA">
    <!!-- media type, as per [RFC2045] -->

<!!ENTITY % ContentTypes "CDATA">
    <!!-- comma-separated list of media types, as per [RFC2045] -->

<!!ENTITY % Charset "CDATA">
    <!!-- a character encoding, as per [RFC2045] -->

<!!ENTITY % Charsets "CDATA">
    <!!-- a space separated list of character encodings, as per [RFC2045] -->

<!!ENTITY % LanguageCode "NMTOKEN">
    <!!-- a language code, as per [RFC3066] -->

<!!ENTITY % Character "CDATA">
    <!!-- a single character, as per section 2.2 of [XML] -->

<!!ENTITY % Number "CDATA">
    <!!-- one or more digits -->

<!!ENTITY % LinkTypes "CDATA">
    <!!-- space-separated list of link types -->

<!!ENTITY % MediaDesc "CDATA">
    <!!-- single or comma-separated list of media descriptors -->

<!!ENTITY % URI "CDATA">
    <!!-- a Uniform Resource Identifier, see [RFC2396] -->

<!!ENTITY % UriList "CDATA">
    <!!-- a space separated list of Uniform Resource Identifiers -->

<!!ENTITY % Datetime "CDATA">
    <!!-- date and time information. ISO date format -->

<!!ENTITY % Script "CDATA">
    <!!-- script expression -->

<!!ENTITY % StyleSheet "CDATA">
    <!!-- style sheet data -->

<!!ENTITY % Text "CDATA">
    <!!-- used for titles etc. -->

<!!ENTITY % FrameTarget "NMTOKEN">
    <!!-- render in this frame -->

<!!ENTITY % Length "CDATA">
    <!!-- nn for pixels or nn% for percentage length -->

<!!ENTITY % MultiLength "CDATA">
    <!!-- pixel, percentage, or relative -->

<!!ENTITY % Pixels "CDATA">
    <!!-- integer representing length in pixels -->

<!!-- these are used for image maps -->

<!!ENTITY % Shape "(rect|circle|poly|default)">

<!!ENTITY % Coords "CDATA">
    <!!-- comma separated list of lengths -->

<!!-- used for object, applet, img, input and iframe -->
<!!ENTITY % ImgAlign "(top|middle|bottom|left|right)">

<!!-- a color using sRGB: #RRGGBB as Hex values -->
<!!ENTITY % Color "CDATA">

<!!-- There are also 16 widely known color names with their sRGB values:

    Black  = #000000    Green  = #008000
    Silver = #C0C0C0    Lime   = #00FF00
    Gray   = #808080    Olive  = #808000
    White  = #FFFFFF    Yellow = #FFFF00
    Maroon = #800000    Navy   = #000080
    Red    = #FF0000    Blue   = #0000FF
    Purple = #800080    Teal   = #008080
    Fuchsia= #FF00FF    Aqua   = #00FFFF
-->

<!!--=================== Generic Attributes ===============================-->

<!!-- core attributes common to most elements
  id       document-wide unique id
  class    space separated list of classes
  style    associated style info
  title    advisory title/amplification
-->
<!!ENTITY % coreattrs
 "id          ID             #IMPLIED
  class       CDATA          #IMPLIED
  style       %StyleSheet;   #IMPLIED
  title       %Text;         #IMPLIED"
  >

<!!-- internationalization attributes
  lang        language code (backwards compatible)
  xml:lang    language code (as per XML 1.0 spec)
  dir         direction for weak/neutral text
-->
<!!ENTITY % i18n
 "lang        %LanguageCode; #IMPLIED
  xml:lang    %LanguageCode; #IMPLIED
  dir         (ltr|rtl)      #IMPLIED"
  >

<!!-- attributes for common UI events
  onclick     a pointer button was clicked
  ondblclick  a pointer button was double clicked
  onmousedown a pointer button was pressed down
  onmouseup   a pointer button was released
  onmousemove a pointer was moved onto the element
  onmouseout  a pointer was moved away from the element
  onkeypress  a key was pressed and released
  onkeydown   a key was pressed down
  onkeyup     a key was released
-->
<!!ENTITY % events
 "onclick     %Script;       #IMPLIED
  ondblclick  %Script;       #IMPLIED
  onmousedown %Script;       #IMPLIED
  onmouseup   %Script;       #IMPLIED
  onmouseover %Script;       #IMPLIED
  onmousemove %Script;       #IMPLIED
  onmouseout  %Script;       #IMPLIED
  onkeypress  %Script;       #IMPLIED
  onkeydown   %Script;       #IMPLIED
  onkeyup     %Script;       #IMPLIED"
  >

<!!-- attributes for elements that can get the focus
  accesskey   accessibility key character
  tabindex    position in tabbing order
  onfocus     the element got the focus
  onblur      the element lost the focus
-->
<!!ENTITY % focus
 "accesskey   %Character;    #IMPLIED
  tabindex    %Number;       #IMPLIED
  onfocus     %Script;       #IMPLIED
  onblur      %Script;       #IMPLIED"
  >

<!!ENTITY % attrs "%coreattrs; %i18n; %events;">

<!!-- text alignment for p, div, h1-h6. The default is
     align="left" for ltr headings, "right" for rtl -->

<!!ENTITY % TextAlign "align (left|center|right|justify) #IMPLIED">

<!!--=================== Text Elements ====================================-->

<!!ENTITY % special.extra
   "object | applet | img | map | iframe">
	
<!!ENTITY % special.basic
	"br | span | bdo">

<!!ENTITY % special
   "%special.basic; | %special.extra;">

<!!ENTITY % fontstyle.extra "big | small | font | basefont">

<!!ENTITY % fontstyle.basic "tt | i | b | u
                      | s | strike ">

<!!ENTITY % fontstyle "%fontstyle.basic; | %fontstyle.extra;">

<!!ENTITY % phrase.extra "sub | sup">
<!!ENTITY % phrase.basic "em | strong | dfn | code | q |
                   samp | kbd | var | cite | abbr | acronym">

<!!ENTITY % phrase "%phrase.basic; | %phrase.extra;">

<!!ENTITY % inline.forms "input | select | textarea | label | button">

<!!-- these can occur at block or inline level -->
<!!ENTITY % misc.inline "ins | del | script">

<!!-- these can only occur at block level -->
<!!ENTITY % misc "noscript | %misc.inline;">

<!!ENTITY % inline "a | %special; | %fontstyle; | %phrase; | %inline.forms;">

<!!-- %Inline; covers inline or "text-level" elements -->
<!!ENTITY % Inline "(#PCDATA | %inline; | %misc.inline;)*">

<!!--================== Block level elements ==============================-->

<!!ENTITY % heading "h1|h2|h3|h4|h5|h6">
<!!ENTITY % lists "ul | ol | dl | menu | dir">
<!!ENTITY % blocktext "pre | hr | blockquote | address | center | noframes">

<!!ENTITY % block
    "p | %heading; | div | %lists; | %blocktext; | isindex |fieldset | table">

<!!-- %Flow; mixes block and inline and is used for list items etc. -->
<!!ENTITY % Flow "(#PCDATA | %block; | form | %inline; | %misc;)*">

<!!--================== Content models for exclusions =====================-->

<!!-- a elements use %Inline; excluding a -->

<!!ENTITY % a.content
   "(#PCDATA | %special; | %fontstyle; | %phrase; | %inline.forms; | %misc.inline;)*">

<!!-- pre uses %Inline excluding img, object, applet, big, small,
     font, or basefont -->

<!!ENTITY % pre.content
   "(#PCDATA | a | %special.basic; | %fontstyle.basic; | %phrase.basic; |
	   %inline.forms; | %misc.inline;)*">

<!!-- form uses %Flow; excluding form -->

<!!ENTITY % form.content "(#PCDATA | %block; | %inline; | %misc;)*">

<!!-- button uses %Flow; but excludes a, form, form controls, iframe -->

<!!ENTITY % button.content
   "(#PCDATA | p | %heading; | div | %lists; | %blocktext; |
      table | br | span | bdo | object | applet | img | map |
      %fontstyle; | %phrase; | %misc;)*">

<!!--================ Document Structure ==================================-->

<!!-- the namespace URI designates the document profile -->

<!!ELEMENT html (head, body)>
<!!ATTLIST html
  %i18n;
  id          ID             #IMPLIED
  xmlns       %URI;          #FIXED ''http://www.w3.org/1999/xhtml''
  >

<!!--================ Document Head =======================================-->

<!!ENTITY % head.misc "(script|style|meta|link|object|isindex)*">

<!!-- content model is %head.misc; combined with a single
     title and an optional base element in any order -->

<!!ELEMENT head (%head.misc;,
     ((title, %head.misc;, (base, %head.misc;)?) |
      (base, %head.misc;, (title, %head.misc;))))>

<!!ATTLIST head
  %i18n;
  id          ID             #IMPLIED
  profile     %URI;          #IMPLIED
  >

<!!-- The title element is not considered part of the flow of text.
       It should be displayed, for example as the page header or
       window title. Exactly one title is required per document.
    -->
<!!ELEMENT title (#PCDATA)>
<!!ATTLIST title 
  %i18n;
  id          ID             #IMPLIED
  >

<!!-- document base URI -->

<!!ELEMENT base EMPTY>
<!!ATTLIST base
  id          ID             #IMPLIED
  href        %URI;          #IMPLIED
  target      %FrameTarget;  #IMPLIED
  >

<!!-- generic metainformation -->
<!!ELEMENT meta EMPTY>
<!!ATTLIST meta
  %i18n;
  id          ID             #IMPLIED
  http-equiv  CDATA          #IMPLIED
  name        CDATA          #IMPLIED
  content     CDATA          #REQUIRED
  scheme      CDATA          #IMPLIED
  >

<!!--
  Relationship values can be used in principle:

   a) for document specific toolbars/menus when used
      with the link element in document head e.g.
        start, contents, previous, next, index, end, help
   b) to link to a separate style sheet (rel="stylesheet")
   c) to make a link to a script (rel="script")
   d) by stylesheets to control how collections of
      html nodes are rendered into printed documents
   e) to make a link to a printable version of this document
      e.g. a PostScript or PDF version (rel="alternate" media="print")
-->

<!!ELEMENT link EMPTY>
<!!ATTLIST link
  %attrs;
  charset     %Charset;      #IMPLIED
  href        %URI;          #IMPLIED
  hreflang    %LanguageCode; #IMPLIED
  type        %ContentType;  #IMPLIED
  rel         %LinkTypes;    #IMPLIED
  rev         %LinkTypes;    #IMPLIED
  media       %MediaDesc;    #IMPLIED
  target      %FrameTarget;  #IMPLIED
  >

<!!-- style info, which may include CDATA sections -->
<!!ELEMENT style (#PCDATA)>
<!!ATTLIST style
  %i18n;
  id          ID             #IMPLIED
  type        %ContentType;  #REQUIRED
  media       %MediaDesc;    #IMPLIED
  title       %Text;         #IMPLIED
  xml:space   (preserve)     #FIXED ''preserve''
  >

<!!-- script statements, which may include CDATA sections -->
<!!ELEMENT script (#PCDATA)>
<!!ATTLIST script
  id          ID             #IMPLIED
  charset     %Charset;      #IMPLIED
  type        %ContentType;  #REQUIRED
  language    CDATA          #IMPLIED
  src         %URI;          #IMPLIED
  defer       (defer)        #IMPLIED
  xml:space   (preserve)     #FIXED ''preserve''
  >

<!!-- alternate content container for non script-based rendering -->

<!!ELEMENT noscript %Flow;>
<!!ATTLIST noscript
  %attrs;
  >

<!!--======================= Frames =======================================-->

<!!-- inline subwindow -->

<!!ELEMENT iframe %Flow;>
<!!ATTLIST iframe
  %coreattrs;
  longdesc    %URI;          #IMPLIED
  name        NMTOKEN        #IMPLIED
  src         %URI;          #IMPLIED
  frameborder (1|0)          "1"
  marginwidth %Pixels;       #IMPLIED
  marginheight %Pixels;      #IMPLIED
  scrolling   (yes|no|auto)  "auto"
  align       %ImgAlign;     #IMPLIED
  height      %Length;       #IMPLIED
  width       %Length;       #IMPLIED
  >

<!!-- alternate content container for non frame-based rendering -->

<!!ELEMENT noframes %Flow;>
<!!ATTLIST noframes
  %attrs;
  >

<!!--=================== Document Body ====================================-->

<!!ELEMENT body %Flow;>
<!!ATTLIST body
  %attrs;
  onload      %Script;       #IMPLIED
  onunload    %Script;       #IMPLIED
  background  %URI;          #IMPLIED
  bgcolor     %Color;        #IMPLIED
  text        %Color;        #IMPLIED
  link        %Color;        #IMPLIED
  vlink       %Color;        #IMPLIED
  alink       %Color;        #IMPLIED
  >

<!!ELEMENT div %Flow;>  <!!-- generic language/style container -->
<!!ATTLIST div
  %attrs;
  %TextAlign;
  >

<!!--=================== Paragraphs =======================================-->

<!!ELEMENT p %Inline;>
<!!ATTLIST p
  %attrs;
  %TextAlign;
  >

<!!--=================== Headings =========================================-->

<!!--
  There are six levels of headings from h1 (the most important)
  to h6 (the least important).
-->

<!!ELEMENT h1  %Inline;>
<!!ATTLIST h1
  %attrs;
  %TextAlign;
  >

<!!ELEMENT h2 %Inline;>
<!!ATTLIST h2
  %attrs;
  %TextAlign;
  >

<!!ELEMENT h3 %Inline;>
<!!ATTLIST h3
  %attrs;
  %TextAlign;
  >

<!!ELEMENT h4 %Inline;>
<!!ATTLIST h4
  %attrs;
  %TextAlign;
  >

<!!ELEMENT h5 %Inline;>
<!!ATTLIST h5
  %attrs;
  %TextAlign;
  >

<!!ELEMENT h6 %Inline;>
<!!ATTLIST h6
  %attrs;
  %TextAlign;
  >

<!!--=================== Lists ============================================-->

<!!-- Unordered list bullet styles -->

<!!ENTITY % ULStyle "(disc|square|circle)">

<!!-- Unordered list -->

<!!ELEMENT ul (li)+>
<!!ATTLIST ul
  %attrs;
  type        %ULStyle;     #IMPLIED
  compact     (compact)     #IMPLIED
  >

<!!-- Ordered list numbering style

    1   arabic numbers      1, 2, 3, ...
    a   lower alpha         a, b, c, ...
    A   upper alpha         A, B, C, ...
    i   lower roman         i, ii, iii, ...
    I   upper roman         I, II, III, ...

    The style is applied to the sequence number which by default
    is reset to 1 for the first list item in an ordered list.
-->
<!!ENTITY % OLStyle "CDATA">

<!!-- Ordered (numbered) list -->

<!!ELEMENT ol (li)+>
<!!ATTLIST ol
  %attrs;
  type        %OLStyle;      #IMPLIED
  compact     (compact)      #IMPLIED
  start       %Number;       #IMPLIED
  >

<!!-- single column list (DEPRECATED) --> 
<!!ELEMENT menu (li)+>
<!!ATTLIST menu
  %attrs;
  compact     (compact)     #IMPLIED
  >

<!!-- multiple column list (DEPRECATED) --> 
<!!ELEMENT dir (li)+>
<!!ATTLIST dir
  %attrs;
  compact     (compact)     #IMPLIED
  >

<!!-- LIStyle is constrained to: "(%ULStyle;|%OLStyle;)" -->
<!!ENTITY % LIStyle "CDATA">

<!!-- list item -->

<!!ELEMENT li %Flow;>
<!!ATTLIST li
  %attrs;
  type        %LIStyle;      #IMPLIED
  value       %Number;       #IMPLIED
  >

<!!-- definition lists - dt for term, dd for its definition -->

<!!ELEMENT dl (dt|dd)+>
<!!ATTLIST dl
  %attrs;
  compact     (compact)      #IMPLIED
  >

<!!ELEMENT dt %Inline;>
<!!ATTLIST dt
  %attrs;
  >

<!!ELEMENT dd %Flow;>
<!!ATTLIST dd
  %attrs;
  >

<!!--=================== Address ==========================================-->

<!!-- information on author -->

<!!ELEMENT address (#PCDATA | %inline; | %misc.inline; | p)*>
<!!ATTLIST address
  %attrs;
  >

<!!--=================== Horizontal Rule ==================================-->

<!!ELEMENT hr EMPTY>
<!!ATTLIST hr
  %attrs;
  align       (left|center|right) #IMPLIED
  noshade     (noshade)      #IMPLIED
  size        %Pixels;       #IMPLIED
  width       %Length;       #IMPLIED
  >

<!!--=================== Preformatted Text ================================-->

<!!-- content is %Inline; excluding 
        "img|object|applet|big|small|sub|sup|font|basefont" -->

<!!ELEMENT pre %pre.content;>
<!!ATTLIST pre
  %attrs;
  width       %Number;      #IMPLIED
  xml:space   (preserve)    #FIXED ''preserve''
  >

<!!--=================== Block-like Quotes ================================-->

<!!ELEMENT blockquote %Flow;>
<!!ATTLIST blockquote
  %attrs;
  cite        %URI;          #IMPLIED
  >

<!!--=================== Text alignment ===================================-->

<!!-- center content -->
<!!ELEMENT center %Flow;>
<!!ATTLIST center
  %attrs;
  >

<!!--=================== Inserted/Deleted Text ============================-->

<!!--
  ins/del are allowed in block and inline content, but its
  inappropriate to include block content within an ins element
  occurring in inline content.
-->
<!!ELEMENT ins %Flow;>
<!!ATTLIST ins
  %attrs;
  cite        %URI;          #IMPLIED
  datetime    %Datetime;     #IMPLIED
  >

<!!ELEMENT del %Flow;>
<!!ATTLIST del
  %attrs;
  cite        %URI;          #IMPLIED
  datetime    %Datetime;     #IMPLIED
  >

<!!--================== The Anchor Element ================================-->

<!!-- content is %Inline; except that anchors shouldn''t be nested -->

<!!ELEMENT a %a.content;>
<!!ATTLIST a
  %attrs;
  %focus;
  charset     %Charset;      #IMPLIED
  type        %ContentType;  #IMPLIED
  name        NMTOKEN        #IMPLIED
  href        %URI;          #IMPLIED
  hreflang    %LanguageCode; #IMPLIED
  rel         %LinkTypes;    #IMPLIED
  rev         %LinkTypes;    #IMPLIED
  shape       %Shape;        "rect"
  coords      %Coords;       #IMPLIED
  target      %FrameTarget;  #IMPLIED
  >

<!!--===================== Inline Elements ================================-->

<!!ELEMENT span %Inline;> <!!-- generic language/style container -->
<!!ATTLIST span
  %attrs;
  >

<!!ELEMENT bdo %Inline;>  <!!-- I18N BiDi over-ride -->
<!!ATTLIST bdo
  %coreattrs;
  %events;
  lang        %LanguageCode; #IMPLIED
  xml:lang    %LanguageCode; #IMPLIED
  dir         (ltr|rtl)      #REQUIRED
  >

<!!ELEMENT br EMPTY>   <!!-- forced line break -->
<!!ATTLIST br
  %coreattrs;
  clear       (left|all|right|none) "none"
  >

<!!ELEMENT em %Inline;>   <!!-- emphasis -->
<!!ATTLIST em %attrs;>

<!!ELEMENT strong %Inline;>   <!!-- strong emphasis -->
<!!ATTLIST strong %attrs;>

<!!ELEMENT dfn %Inline;>   <!!-- definitional -->
<!!ATTLIST dfn %attrs;>

<!!ELEMENT code %Inline;>   <!!-- program code -->
<!!ATTLIST code %attrs;>

<!!ELEMENT samp %Inline;>   <!!-- sample -->
<!!ATTLIST samp %attrs;>

<!!ELEMENT kbd %Inline;>  <!!-- something user would type -->
<!!ATTLIST kbd %attrs;>

<!!ELEMENT var %Inline;>   <!!-- variable -->
<!!ATTLIST var %attrs;>

<!!ELEMENT cite %Inline;>   <!!-- citation -->
<!!ATTLIST cite %attrs;>

<!!ELEMENT abbr %Inline;>   <!!-- abbreviation -->
<!!ATTLIST abbr %attrs;>

<!!ELEMENT acronym %Inline;>   <!!-- acronym -->
<!!ATTLIST acronym %attrs;>

<!!ELEMENT q %Inline;>   <!!-- inlined quote -->
<!!ATTLIST q
  %attrs;
  cite        %URI;          #IMPLIED
  >

<!!ELEMENT sub %Inline;> <!!-- subscript -->
<!!ATTLIST sub %attrs;>

<!!ELEMENT sup %Inline;> <!!-- superscript -->
<!!ATTLIST sup %attrs;>

<!!ELEMENT tt %Inline;>   <!!-- fixed pitch font -->
<!!ATTLIST tt %attrs;>

<!!ELEMENT i %Inline;>   <!!-- italic font -->
<!!ATTLIST i %attrs;>

<!!ELEMENT b %Inline;>   <!!-- bold font -->
<!!ATTLIST b %attrs;>

<!!ELEMENT big %Inline;>   <!!-- bigger font -->
<!!ATTLIST big %attrs;>

<!!ELEMENT small %Inline;>   <!!-- smaller font -->
<!!ATTLIST small %attrs;>

<!!ELEMENT u %Inline;>   <!!-- underline -->
<!!ATTLIST u %attrs;>

<!!ELEMENT s %Inline;>   <!!-- strike-through -->
<!!ATTLIST s %attrs;>

<!!ELEMENT strike %Inline;>   <!!-- strike-through -->
<!!ATTLIST strike %attrs;>

<!!ELEMENT basefont EMPTY>  <!!-- base font size -->
<!!ATTLIST basefont
  id          ID             #IMPLIED
  size        CDATA          #REQUIRED
  color       %Color;        #IMPLIED
  face        CDATA          #IMPLIED
  >

<!!ELEMENT font %Inline;> <!!-- local change to font -->
<!!ATTLIST font
  %coreattrs;
  %i18n;
  size        CDATA          #IMPLIED
  color       %Color;        #IMPLIED
  face        CDATA          #IMPLIED
  >

<!!--==================== Object ======================================-->
<!!--
  object is used to embed objects as part of HTML pages.
  param elements should precede other content. Parameters
  can also be expressed as attribute/value pairs on the
  object element itself when brevity is desired.
-->

<!!ELEMENT object (#PCDATA | param | %block; | form | %inline; | %misc;)*>
<!!ATTLIST object
  %attrs;
  declare     (declare)      #IMPLIED
  classid     %URI;          #IMPLIED
  codebase    %URI;          #IMPLIED
  data        %URI;          #IMPLIED
  type        %ContentType;  #IMPLIED
  codetype    %ContentType;  #IMPLIED
  archive     %UriList;      #IMPLIED
  standby     %Text;         #IMPLIED
  height      %Length;       #IMPLIED
  width       %Length;       #IMPLIED
  usemap      %URI;          #IMPLIED
  name        NMTOKEN        #IMPLIED
  tabindex    %Number;       #IMPLIED
  align       %ImgAlign;     #IMPLIED
  border      %Pixels;       #IMPLIED
  hspace      %Pixels;       #IMPLIED
  vspace      %Pixels;       #IMPLIED
  >

<!!--
  param is used to supply a named property value.
  In XML it would seem natural to follow RDF and support an
  abbreviated syntax where the param elements are replaced
  by attribute value pairs on the object start tag.
-->
<!!ELEMENT param EMPTY>
<!!ATTLIST param
  id          ID             #IMPLIED
  name        CDATA          #REQUIRED
  value       CDATA          #IMPLIED
  valuetype   (data|ref|object) "data"
  type        %ContentType;  #IMPLIED
  >

<!!--=================== Java applet ==================================-->
<!!--
  One of code or object attributes must be present.
  Place param elements before other content.
-->
<!!ELEMENT applet (#PCDATA | param | %block; | form | %inline; | %misc;)*>
<!!ATTLIST applet
  %coreattrs;
  codebase    %URI;          #IMPLIED
  archive     CDATA          #IMPLIED
  code        CDATA          #IMPLIED
  object      CDATA          #IMPLIED
  alt         %Text;         #IMPLIED
  name        NMTOKEN        #IMPLIED
  width       %Length;       #REQUIRED
  height      %Length;       #REQUIRED
  align       %ImgAlign;     #IMPLIED
  hspace      %Pixels;       #IMPLIED
  vspace      %Pixels;       #IMPLIED
  >

<!!--=================== Images ===========================================-->

<!!--
   To avoid accessibility problems for people who aren''t
   able to see the image, you should provide a text
   description using the alt and longdesc attributes.
   In addition, avoid the use of server-side image maps.
-->

<!!ELEMENT img EMPTY>
<!!ATTLIST img
  %attrs;
  src         %URI;          #REQUIRED
  alt         %Text;         #REQUIRED
  name        NMTOKEN        #IMPLIED
  longdesc    %URI;          #IMPLIED
  height      %Length;       #IMPLIED
  width       %Length;       #IMPLIED
  usemap      %URI;          #IMPLIED
  ismap       (ismap)        #IMPLIED
  align       %ImgAlign;     #IMPLIED
  border      %Length;       #IMPLIED
  hspace      %Pixels;       #IMPLIED
  vspace      %Pixels;       #IMPLIED
  >

<!!-- usemap points to a map element which may be in this document
  or an external document, although the latter is not widely supported -->

<!!--================== Client-side image maps ============================-->

<!!-- These can be placed in the same document or grouped in a
     separate document although this isn''t yet widely supported -->

<!!ELEMENT map ((%block; | form | %misc;)+ | area+)>
<!!ATTLIST map
  %i18n;
  %events;
  id          ID             #REQUIRED
  class       CDATA          #IMPLIED
  style       %StyleSheet;   #IMPLIED
  title       %Text;         #IMPLIED
  name        CDATA          #IMPLIED
  >

<!!ELEMENT area EMPTY>
<!!ATTLIST area
  %attrs;
  %focus;
  shape       %Shape;        "rect"
  coords      %Coords;       #IMPLIED
  href        %URI;          #IMPLIED
  nohref      (nohref)       #IMPLIED
  alt         %Text;         #REQUIRED
  target      %FrameTarget;  #IMPLIED
  >

<!!--================ Forms ===============================================-->

<!!ELEMENT form %form.content;>   <!!-- forms shouldn''t be nested -->

<!!ATTLIST form
  %attrs;
  action      %URI;          #REQUIRED
  method      (get|post)     "get"
  name        NMTOKEN        #IMPLIED
  enctype     %ContentType;  "application/x-www-form-urlencoded"
  onsubmit    %Script;       #IMPLIED
  onreset     %Script;       #IMPLIED
  accept      %ContentTypes; #IMPLIED
  accept-charset %Charsets;  #IMPLIED
  target      %FrameTarget;  #IMPLIED
  >

<!!--
  Each label must not contain more than ONE field
  Label elements shouldn''t be nested.
-->
<!!ELEMENT label %Inline;>
<!!ATTLIST label
  %attrs;
  for         IDREF          #IMPLIED
  accesskey   %Character;    #IMPLIED
  onfocus     %Script;       #IMPLIED
  onblur      %Script;       #IMPLIED
  >

<!!ENTITY % InputType
  "(text | password | checkbox |
    radio | submit | reset |
    file | hidden | image | button)"
   >

<!!-- the name attribute is required for all but submit & reset -->

<!!ELEMENT input EMPTY>     <!!-- form control -->
<!!ATTLIST input
  %attrs;
  %focus;
  type        %InputType;    "text"
  name        CDATA          #IMPLIED
  value       CDATA          #IMPLIED
  checked     (checked)      #IMPLIED
  disabled    (disabled)     #IMPLIED
  readonly    (readonly)     #IMPLIED
  size        CDATA          #IMPLIED
  maxlength   %Number;       #IMPLIED
  src         %URI;          #IMPLIED
  alt         CDATA          #IMPLIED
  usemap      %URI;          #IMPLIED
  onselect    %Script;       #IMPLIED
  onchange    %Script;       #IMPLIED
  accept      %ContentTypes; #IMPLIED
  align       %ImgAlign;     #IMPLIED
  >

<!!ELEMENT select (optgroup|option)+>  <!!-- option selector -->
<!!ATTLIST select
  %attrs;
  name        CDATA          #IMPLIED
  size        %Number;       #IMPLIED
  multiple    (multiple)     #IMPLIED
  disabled    (disabled)     #IMPLIED
  tabindex    %Number;       #IMPLIED
  onfocus     %Script;       #IMPLIED
  onblur      %Script;       #IMPLIED
  onchange    %Script;       #IMPLIED
  >

<!!ELEMENT optgroup (option)+>   <!!-- option group -->
<!!ATTLIST optgroup
  %attrs;
  disabled    (disabled)     #IMPLIED
  label       %Text;         #REQUIRED
  >

<!!ELEMENT option (#PCDATA)>     <!!-- selectable choice -->
<!!ATTLIST option
  %attrs;
  selected    (selected)     #IMPLIED
  disabled    (disabled)     #IMPLIED
  label       %Text;         #IMPLIED
  value       CDATA          #IMPLIED
  >

<!!ELEMENT textarea (#PCDATA)>     <!!-- multi-line text field -->
<!!ATTLIST textarea
  %attrs;
  %focus;
  name        CDATA          #IMPLIED
  rows        %Number;       #REQUIRED
  cols        %Number;       #REQUIRED
  disabled    (disabled)     #IMPLIED
  readonly    (readonly)     #IMPLIED
  onselect    %Script;       #IMPLIED
  onchange    %Script;       #IMPLIED
  >

<!!--
  The fieldset element is used to group form fields.
  Only one legend element should occur in the content
  and if present should only be preceded by whitespace.
-->
<!!ELEMENT fieldset (#PCDATA | legend | %block; | form | %inline; | %misc;)*>
<!!ATTLIST fieldset
  %attrs;
  >

<!!ENTITY % LAlign "(top|bottom|left|right)">

<!!ELEMENT legend %Inline;>     <!!-- fieldset label -->
<!!ATTLIST legend
  %attrs;
  accesskey   %Character;    #IMPLIED
  align       %LAlign;       #IMPLIED
  >

<!!--
 Content is %Flow; excluding a, form, form controls, iframe
--> 
<!!ELEMENT button %button.content;>  <!!-- push button -->
<!!ATTLIST button
  %attrs;
  %focus;
  name        CDATA          #IMPLIED
  value       CDATA          #IMPLIED
  type        (button|submit|reset) "submit"
  disabled    (disabled)     #IMPLIED
  >

<!!-- single-line text input control (DEPRECATED) -->
<!!ELEMENT isindex EMPTY>
<!!ATTLIST isindex
  %coreattrs;
  %i18n;
  prompt      %Text;         #IMPLIED
  >

<!!--======================= Tables =======================================-->

<!!-- Derived from IETF HTML table standard, see [RFC1942] -->

<!!--
 The border attribute sets the thickness of the frame around the
 table. The default units are screen pixels.

 The frame attribute specifies which parts of the frame around
 the table should be rendered. The values are not the same as
 CALS to avoid a name clash with the valign attribute.
-->
<!!ENTITY % TFrame "(void|above|below|hsides|lhs|rhs|vsides|box|border)">

<!!--
 The rules attribute defines which rules to draw between cells:

 If rules is absent then assume:
     "none" if border is absent or border="0" otherwise "all"
-->

<!!ENTITY % TRules "(none | groups | rows | cols | all)">
  
<!!-- horizontal placement of table relative to document -->
<!!ENTITY % TAlign "(left|center|right)">

<!!-- horizontal alignment attributes for cell contents

  char        alignment char, e.g. char='':''
  charoff     offset for alignment char
-->
<!!ENTITY % cellhalign
  "align      (left|center|right|justify|char) #IMPLIED
   char       %Character;    #IMPLIED
   charoff    %Length;       #IMPLIED"
  >

<!!-- vertical alignment attributes for cell contents -->
<!!ENTITY % cellvalign
  "valign     (top|middle|bottom|baseline) #IMPLIED"
  >

<!!ELEMENT table
     (caption?, (col*|colgroup*), thead?, tfoot?, (tbody+|tr+))>
<!!ELEMENT caption  %Inline;>
<!!ELEMENT thead    (tr)+>
<!!ELEMENT tfoot    (tr)+>
<!!ELEMENT tbody    (tr)+>
<!!ELEMENT colgroup (col)*>
<!!ELEMENT col      EMPTY>
<!!ELEMENT tr       (th|td)+>
<!!ELEMENT th       %Flow;>
<!!ELEMENT td       %Flow;>

<!!ATTLIST table
  %attrs;
  summary     %Text;         #IMPLIED
  width       %Length;       #IMPLIED
  border      %Pixels;       #IMPLIED
  frame       %TFrame;       #IMPLIED
  rules       %TRules;       #IMPLIED
  cellspacing %Length;       #IMPLIED
  cellpadding %Length;       #IMPLIED
  align       %TAlign;       #IMPLIED
  bgcolor     %Color;        #IMPLIED
  >

<!!ENTITY % CAlign "(top|bottom|left|right)">

<!!ATTLIST caption
  %attrs;
  align       %CAlign;       #IMPLIED
  >

<!!--
colgroup groups a set of col elements. It allows you to group
several semantically related columns together.
-->
<!!ATTLIST colgroup
  %attrs;
  span        %Number;       "1"
  width       %MultiLength;  #IMPLIED
  %cellhalign;
  %cellvalign;
  >

<!!--
 col elements define the alignment properties for cells in
 one or more columns.

 The width attribute specifies the width of the columns, e.g.

     width=64        width in screen pixels
     width=0.5*      relative width of 0.5

 The span attribute causes the attributes of one
 col element to apply to more than one column.
-->
<!!ATTLIST col
  %attrs;
  span        %Number;       "1"
  width       %MultiLength;  #IMPLIED
  %cellhalign;
  %cellvalign;
  >

<!!--
    Use thead to duplicate headers when breaking table
    across page boundaries, or for static headers when
    tbody sections are rendered in scrolling panel.

    Use tfoot to duplicate footers when breaking table
    across page boundaries, or for static footers when
    tbody sections are rendered in scrolling panel.

    Use multiple tbody sections when rules are needed
    between groups of table rows.
-->
<!!ATTLIST thead
  %attrs;
  %cellhalign;
  %cellvalign;
  >

<!!ATTLIST tfoot
  %attrs;
  %cellhalign;
  %cellvalign;
  >

<!!ATTLIST tbody
  %attrs;
  %cellhalign;
  %cellvalign;
  >

<!!ATTLIST tr
  %attrs;
  %cellhalign;
  %cellvalign;
  bgcolor     %Color;        #IMPLIED
  >

<!!-- Scope is simpler than headers attribute for common tables -->
<!!ENTITY % Scope "(row|col|rowgroup|colgroup)">

<!!-- th is for headers, td for data and for cells acting as both -->

<!!ATTLIST th
  %attrs;
  abbr        %Text;         #IMPLIED
  axis        CDATA          #IMPLIED
  headers     IDREFS         #IMPLIED
  scope       %Scope;        #IMPLIED
  rowspan     %Number;       "1"
  colspan     %Number;       "1"
  %cellhalign;
  %cellvalign;
  nowrap      (nowrap)       #IMPLIED
  bgcolor     %Color;        #IMPLIED
  width       %Length;       #IMPLIED
  height      %Length;       #IMPLIED
  >

<!!ATTLIST td
  %attrs;
  abbr        %Text;         #IMPLIED
  axis        CDATA          #IMPLIED
  headers     IDREFS         #IMPLIED
  scope       %Scope;        #IMPLIED
  rowspan     %Number;       "1"
  colspan     %Number;       "1"
  %cellhalign;
  %cellvalign;
  nowrap      (nowrap)       #IMPLIED
  bgcolor     %Color;        #IMPLIED
  width       %Length;       #IMPLIED
  height      %Length;       #IMPLIED
  >

'! !

!XHtml1Transitional methodsFor: 'initialize-release' stamp: 'tb 12/1/2005 09:41'!
initialize

	super initialize.
	self parse.! !

!XHtmlDTD methodsFor: 'private' stamp: 'tb 12/5/2005 12:06'!
isXML

	^true! !

!XHtmlDTD methodsFor: 'private' stamp: 'tb 12/3/2005 15:00'!
objectifyElements

	elements := elements collect: [:ea | XHtmlDTDElement fromParsedString: ea]! !

!XHtmlDTD methodsFor: 'private' stamp: 'tb 12/6/2005 15:52'!
tag: aTag canContainTag: anotherTag

	| element |
		element := self elementForTag: aTag asLowercase.
		^element notNil 
			ifTrue: [element allowsTag: anotherTag asLowercase]
			ifFalse:[false]
		! !

Object subclass: #HtmlDTDConstruct
	instanceVariableNames: 'tag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

HtmlDTDConstruct subclass: #HtmlDTDAttrList
	instanceVariableNames: 'attributes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDAttrList class methodsFor: 'as yet unclassified' stamp: 'tb 12/3/2005 14:30'!
fromParsedArray: anArray

	| tag attrs list |

	tag := anArray first.
	attrs := anArray copyWithoutFirst.
	list := OrderedCollection new.

	attrs do: 
	[:ea |

		| in attribute type required |

		in := ReadStream on: ea.

		[in atEnd] whileFalse: 
		[
			attribute := self readExpression: in.
			type := self readExpression: in.
			required := self readExpression: in.
			required = '#FIXED' ifTrue: [in upTo: $'. type := ('(',(in upTo: $'),')')].
			in skipSeparators.
			list add: (HtmlDTDAttribute new attribute: attribute;
						type: type; isRequired: required).
		]
	].

	^self new tag: tag; attributes: list asArray
		! !

!HtmlDTDAttrList methodsFor: 'accessing' stamp: 'tb 11/30/2005 19:31'!
addToAttributes: anAttribute

	self attributes: (self attributes copyWith: anAttribute)! !

!HtmlDTDAttrList methodsFor: 'accessing' stamp: 'tb 11/30/2005 23:12'!
attributeNamed: aString

	| nm |
	nm := aString asLowercase.
	^self attributes detect:[:ea | ea attribute = nm] ifNone: [nil]! !

!HtmlDTDAttrList methodsFor: 'accessing' stamp: 'tb 11/30/2005 19:32'!
attributes

	^attributes isNil ifTrue:[attributes := #()] ifFalse: [attributes]! !

!HtmlDTDAttrList methodsFor: 'accessing' stamp: 'tb 11/30/2005 19:52'!
attributes: anArray

	attributes := anArray! !

!HtmlDTDAttrList methodsFor: 'testing' stamp: 'tb 12/1/2005 00:57'!
includesAttribute: anAttribute

	^(self attributeNamed: anAttribute) notNil! !

!HtmlDTDAttrList methodsFor: 'as yet unclassified' stamp: 'tb 11/30/2005 22:58'!
printOn: aStream

	aStream nextPutAll: '@@'; nextPut: $(. 
	tag expression printOn: aStream.
	aStream space.
	self attributes do:
	[:ea | 
		ea printOn: aStream.
		aStream space.
	].
	aStream nextPut: $).! !

!HtmlDTDAttrList methodsFor: 'accessing' stamp: 'tb 12/1/2005 01:01'!
requiredAttributes

	^self attributes select:[:ea | ea isRequired]! !

HtmlDTDConstruct subclass: #HtmlDTDAttribute
	instanceVariableNames: 'attribute type isRequired'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDAttribute methodsFor: 'accessing' stamp: 'tb 12/2/2005 10:39'!
allowsValue: aValue
	
	self type isNil ifTrue:[^true].
	^self type isEnumeration
			ifFalse: [true]
			ifTrue: [self type allowsItem: aValue]! !

!HtmlDTDAttribute methodsFor: 'accessing' stamp: 'tb 11/30/2005 19:27'!
attribute
	"Answer the value of attribute"

	^ attribute! !

!HtmlDTDAttribute methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
attribute: anObject
	"Set the value of attribute"

	attribute := anObject! !

!HtmlDTDAttribute methodsFor: 'accessing' stamp: 'tb 11/30/2005 19:29'!
isRequired
	"Answer the value of isRequired"

	^ isRequired = '#REQUIRED'! !

!HtmlDTDAttribute methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
isRequired: anObject
	"Set the value of isRequired"

	isRequired := anObject! !

!HtmlDTDAttribute methodsFor: 'accessing' stamp: 'tb 11/30/2005 21:31'!
printOn: aStream

	aStream nextPut: $@; nextPut:$(.
	attribute printOn: aStream.
	aStream space.
	type printOn: aStream.
	aStream space.
	isRequired printOn: aStream.
	aStream nextPut:$).! !

!HtmlDTDAttribute methodsFor: 'accessing' stamp: 'tb 11/30/2005 19:27'!
type
	"Answer the value of type"

	^ type! !

!HtmlDTDAttribute methodsFor: 'accessing' stamp: 'tb 12/2/2005 13:53'!
type: anObject
	"Set the value of type"

	type := HtmlDTDMatchExpression on: anObject! !

!HtmlDTDConstruct class methodsFor: 'as yet unclassified' stamp: 'tb 11/30/2005 21:50'!
readExpression: in

	| s |

	in skipSeparators.
	^in peek = $( ifTrue:
	[
		s := in upTo: $).
		(s,')',(in upTo: $ ))
	]
	ifFalse:
	[
		in upTo: $ .
	]! !

!HtmlDTDConstruct methodsFor: 'accessing' stamp: 'tb 11/30/2005 21:55'!
tag
	"Answer the value of tag"

	^ tag! !

!HtmlDTDConstruct methodsFor: 'accessing' stamp: 'SebastianSastre 1/15/2012 16:34'!
tag: anObject
	"Set the value of tag"

	tag := HtmlDTDMatchExpression on: anObject trimBoth! !

!HtmlDTDConstruct methodsFor: 'accessing' stamp: 'tb 12/1/2005 14:28'!
tagMatches: anObject

	^tag allowsItem: anObject! !

HtmlDTDConstruct subclass: #HtmlDTDElement
	instanceVariableNames: 'dtd startIsOptional endIsOptional contentRule allowedElements'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDElement class methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:34'!
fromParsedString: aString

	| in tag start end contents |

	in := ReadStream on: aString.

	tag := self readExpression: in.
	start := self readExpression: in.
	end := self readExpression: in.
	contents := in upToEnd trimBoth.

	^self new tag: tag;
		startIsOptional: start;
		endIsOptional: end;
		contentRule: contents		! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'tb 12/8/2005 15:25'!
allowsTag: aTag

	^contentRule allowsItem: aTag! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'tb 12/6/2005 15:27'!
cacheAllowedElementsForDTD: aDTD

	| |

	allowedElements isNil ifTrue:
	[
		allowedElements := Set new.

		allowedElements addAll: self contentRule allowedElements.

		allowedElements copy do: 
		[:ea | 
			ea isLiteral ifTrue: 
			[
				| el |
				el := aDTD elementForTag: ea expression.
				el startIsOptional ifTrue:
					[allowedElements addAll: (el cacheAllowedElementsForDTD: aDTD)]
			]
		]
	].

	^allowedElements! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'tb 11/30/2005 22:06'!
contentRule
	"Answer the value of contentRule"

	^ contentRule! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'tb 12/2/2005 08:53'!
contentRule: anObject
	"Set the value of contentRule"

	contentRule := HtmlDTDMatchExpression on: anObject! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'tb 11/30/2005 22:25'!
endIsOptional
	"Answer the value of endIsOptional"

	^ endIsOptional first = $O! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
endIsOptional: anObject
	"Set the value of endIsOptional"

	endIsOptional := anObject! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'tb 12/1/2005 14:29'!
isEmpty

	^contentRule isEmpty! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'tb 11/30/2005 22:29'!
printOn: aStream

	aStream nextPutAll: 'E('.
	tag printOn: aStream.
	aStream space; 
	nextPutAll: startIsOptional; space;
	nextPutAll: endIsOptional; space.
	contentRule printOn: aStream.
	aStream nextPut: $)! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'tb 11/30/2005 22:02'!
startIsOptional
	"Answer the value of startIsOptional"

	^ startIsOptional first = $O! !

!HtmlDTDElement methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
startIsOptional: anObject
	"Set the value of startIsOptional"

	startIsOptional := anObject! !

HtmlDTDElement subclass: #XHtmlDTDElement
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!XHtmlDTDElement class methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:42'!
fromParsedString: aString

	| in tag start end contents |

	in := ReadStream on: aString.

	tag := self readExpression: in.
	start := '-'.
	end := '-'.
	contents := in upToEnd trimBoth.

	^self new tag: tag;
		startIsOptional: start;
		endIsOptional: end;
		contentRule: contents		! !

Object subclass: #HtmlDTDMatchExpression
	instanceVariableNames: 'expression'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchExpression commentStamp: '<historical>' prior: 0!
The expression rules are: 
( ... )
    Delimits a group.
A
    A must occur, one time only.
A+
    A must occur one or more times.
A?
    A must occur zero or one time.
A*
    A may occur zero or more times.
+(A)
    A may occur.
-(A)
    A must not occur.
A | B
    Either A or B must occur, but not both.
A , B
    Both A and B must occur, in that order.
A & B
    Both A and B must occur, in any order. !

HtmlDTDMatchExpression subclass: #HtmlDTDMatchAllow
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchAllow class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 22:13'!
on: anObject

	| in | 

	in := anObject isStream 
		ifTrue:[anObject] 
		ifFalse: [ReadStream on: anObject].

	in next = $+ ifFalse: [self error: 'HtmlDTDMatchDeny called with invalid stream state'].

	^self new expression: (HtmlDTDMatchExpression on: in).! !

!HtmlDTDMatchAllow methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 15:25'!
allowedElements

	^expression allowedElements! !

!HtmlDTDMatchAllow methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 14:07'!
allowsItem: anItem

	^self expression allowsItem: anItem! !

!HtmlDTDMatchAllow methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 21:22'!
printOn: aStream

	aStream nextPut: $+.
	expression printOn: aStream.

! !

!HtmlDTDMatchAllow methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 16:44'!
validateChildren: aStream ofNode: aNode

	^expression validateChildren: aStream ofNode: aNode! !

HtmlDTDMatchExpression subclass: #HtmlDTDMatchCDATA
	instanceVariableNames: ''
	classVariableNames: 'CData'
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchCDATA class methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 14:53'!
cdata

	^CData isNil ifTrue: [CData := self new] ifFalse: [CData]! !

!HtmlDTDMatchCDATA methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 15:09'!
allowedElements

	^#()! !

!HtmlDTDMatchCDATA methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 17:35'!
allowsItem: aTag

	^(aTag asUppercase findString: 'CDATA') > 0! !

!HtmlDTDMatchCDATA methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 15:07'!
initialize

	super initialize.
	expression := '#CDATA'! !

!HtmlDTDMatchCDATA methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 17:39'!
printOn: aStream

	aStream nextPutAll: expression! !

!HtmlDTDMatchCDATA methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 17:37'!
validateChildren: aStream ofNode: aNode

	^(aStream atEnd not and: [self allowsItem: aStream peek dtdTag]) 
		ifTrue: [aStream next. true] 
		ifFalse: [false]
! !

HtmlDTDMatchExpression subclass: #HtmlDTDMatchConstraint
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchConstraint methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 10:07'!
allowedElements

	^expression allowedElements! !

!HtmlDTDMatchConstraint methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 10:32'!
allowsItem: anItem

	^expression allowsItem: anItem! !

HtmlDTDMatchConstraint subclass: #HtmlDTDMatchOneOrMore
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchOneOrMore methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 19:39'!
printOn: aStream

	expression printOn: aStream.
	aStream nextPut: $+

	! !

!HtmlDTDMatchOneOrMore methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 12:18'!
validateChildren: aStream ofNode: aNode

	| matchedOne |
	matchedOne := false.
	[expression validateChildren: aStream ofNode: aNode] whileTrue:[matchedOne := true].
	^matchedOne! !

HtmlDTDMatchConstraint subclass: #HtmlDTDMatchZeroOrMore
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchZeroOrMore methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 21:20'!
printOn: aStream

	expression printOn: aStream.
	aStream nextPut: $*

	! !

!HtmlDTDMatchZeroOrMore methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 12:17'!
validateChildren: aStream ofNode: aNode

	[expression validateChildren: aStream ofNode: aNode] whileTrue:[].
	^true! !

HtmlDTDMatchConstraint subclass: #HtmlDTDMatchZeroOrOne
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchZeroOrOne methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 21:20'!
printOn: aStream

	expression printOn: aStream.
	aStream nextPut: $?

	! !

!HtmlDTDMatchZeroOrOne methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 12:19'!
validateChildren: aStream ofNode: aNode

	expression validateChildren: aStream ofNode: aNode.
	^true! !

HtmlDTDMatchExpression subclass: #HtmlDTDMatchDeny
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchDeny class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 22:12'!
on: anObject

	| in | 

	in := anObject isStream 
		ifTrue:[anObject] 
		ifFalse: [ReadStream on: anObject].

	in next = $- ifFalse: [self error: 'HtmlDTDMatchDeny called with invalid stream state'].

	^self new expression: (HtmlDTDMatchExpression on: in).! !

!HtmlDTDMatchDeny methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 10:02'!
allowedElements

	^#()! !

!HtmlDTDMatchDeny methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 14:06'!
allowsItem: anItem

	^(self expression allowsItem: anItem) not! !

!HtmlDTDMatchDeny methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 21:21'!
printOn: aStream

	aStream nextPut: $-.
	expression printOn: aStream.

	! !

!HtmlDTDMatchDeny methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 16:45'!
validateChildren: aStream ofNode: aNode

	^(expression validateChildren: aStream ofNode: aNode) not! !

HtmlDTDMatchExpression subclass: #HtmlDTDMatchEmpty
	instanceVariableNames: ''
	classVariableNames: 'Empty'
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchEmpty class methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 14:50'!
empty

	^Empty isNil ifTrue: [Empty := self new] ifFalse: [Empty]! !

!HtmlDTDMatchEmpty methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 10:02'!
allowedElements

	^#()! !

!HtmlDTDMatchEmpty methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 14:25'!
allowsItem: aTag

	^false! !

!HtmlDTDMatchEmpty methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 15:06'!
initialize

	super initialize.
	expression := 'EMPTY'! !

!HtmlDTDMatchEmpty methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 10:41'!
isEmpty

	^true! !

!HtmlDTDMatchEmpty methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 14:26'!
printOn: aStream

	aStream nextPutAll: 'EMPTY'! !

!HtmlDTDMatchEmpty methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 11:51'!
validateChildren: aStream ofNode: aNode

	^(aStream atEnd) 
		ifTrue: [true] 
		ifFalse:
		[
			aNode addError: (HtmlTagMustBeEmptyError new).
			false
		]
! !

!HtmlDTDMatchExpression class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 22:15'!
on: anObject

" 
( ... )
    Delimits a group.
A
    A must occur, one time only.
A+
    A must occur one or more times.
A?
    A must occur zero or one time.
A*
    A may occur zero or more times.
+(A)
    A may occur.
-(A)
    A must not occur.
A | B
    Either A or B must occur, but not both.
A , B
    Both A and B must occur, in that order.
A & B
    Both A and B must occur, in any order. "

	| in seq | 

	in := anObject isStream 
		ifTrue:[anObject] 
		ifFalse: [ReadStream on: anObject].

	seq := OrderedCollection new.	


	[in atEnd] whileFalse:
	[
		in skipSeparators.
		(in peek = $-) 
		ifTrue: [seq add:(HtmlDTDMatchDeny on: in)]
		ifFalse:
		[
			(in peek = $+) 
			ifTrue: [seq add:(HtmlDTDMatchAllow on: in)]
			ifFalse: 
			[
				(in peek = $( ) 
				ifTrue: 
				[
					seq add: 
					(self readConstraintFor:(HtmlDTDMatchGroup on: in) from: in)
				]
				ifFalse: 
				[
					('|),&' includes: in peek) ifTrue: [^seq last]
					ifFalse:
					[
						seq add: 
						(self readConstraintFor:(HtmlDTDMatchLiteral on: in) from: in)
					]
				]
			]	
		]
	].
	^seq size > 1 
		ifTrue:[self new expression: seq] 
		ifFalse:[seq last]
	
! !

!HtmlDTDMatchExpression class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 21:47'!
readConstraintFor: anExpn from: aStream

	| c |

	^('+*?' includes: aStream peek) ifFalse: [anExpn]
	ifTrue:
	[
		c := aStream next.

		c = $+ ifTrue:[HtmlDTDMatchOneOrMore new expression: anExpn] ifFalse: [
		c = $? ifTrue:[HtmlDTDMatchZeroOrOne new expression: anExpn] ifFalse: [
		c = $* ifTrue:[HtmlDTDMatchZeroOrMore new expression: anExpn]]]
	]	! !

!HtmlDTDMatchExpression methodsFor: 'accessing' stamp: 'tb 12/6/2005 10:01'!
allowedElements

	| collection |

	collection := OrderedCollection new.

	expression do: [:ea | collection addAll: ea allowedElements].

	^collection asArray! !

!HtmlDTDMatchExpression methodsFor: 'matching' stamp: 'tb 12/2/2005 10:13'!
allowsItem: anItem

	^(expression detect: [:ea | ea allowsItem: anItem] ifNone: [nil]) notNil! !

!HtmlDTDMatchExpression methodsFor: 'accessing' stamp: 'tb 11/30/2005 19:18'!
expression
	"Answer the value of expression"

	^ expression! !

!HtmlDTDMatchExpression methodsFor: 'accessing' stamp: 'tb 1/13/2006 20:28'!
expression: anObject
	
	expression := anObject! !

!HtmlDTDMatchExpression methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 14:29'!
isEmpty

	^false! !

!HtmlDTDMatchExpression methodsFor: 'matching' stamp: 'tb 12/2/2005 08:55'!
isEnumeration

	^false! !

!HtmlDTDMatchExpression methodsFor: 'matching' stamp: 'tb 12/6/2005 10:13'!
isLiteral

	^false! !

!HtmlDTDMatchExpression methodsFor: 'matching' stamp: 'tb 12/1/2005 23:56'!
matchTagList: aList

	| in |

	in := ReadStream on: aList.
	self match: in.
	^in upToEnd! !

!HtmlDTDMatchExpression methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 22:16'!
printOn: aStream

	expression do: [:ea | ea printOn: aStream. aStream nextPut: $ ]! !

!HtmlDTDMatchExpression methodsFor: 'matching' stamp: 'tb 12/5/2005 16:54'!
validateChildren: aStream ofNode: aNode

	expression do:
	[:ea | 
		
		(ea validateChildren: aStream ofNode: aNode) ifFalse: [^false]
	].
	^true
	! !

HtmlDTDMatchExpression subclass: #HtmlDTDMatchGroup
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

HtmlDTDMatchGroup subclass: #HtmlDTDMatchAND
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchAND methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 17:14'!
separator

	^'&'! !

!HtmlDTDMatchAND methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 12:35'!
validateChildren: aStream ofNode: aNode

	| position matched subexpressions |

	position := aStream position.
	subexpressions := expression.

	[subexpressions isEmpty] whileFalse:
	[
		matched := subexpressions detect:
		[:ea | 
			aStream position: position.		
			(ea validateChildren: aStream ofNode: aNode)
		] ifNone: [nil].
		matched notNil 
			ifFalse: [^false]
			ifTrue: 
			[
				position := aStream position.
				subexpressions := subexpressions copyWithout: matched.
			]
	].
	^true! !

!HtmlDTDMatchGroup class methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 21:09'!
on: aStringOrStream

	| seq in seps | 
	
	in := aStringOrStream isStream
		ifFalse: [ReadStream on: aStringOrStream] 
		ifTrue: [aStringOrStream].

	in skipSeparators.
	seq := OrderedCollection new.
	seps := #().
	(in peek = $() ifFalse: [self error: 'Groups start with $('].

	in next; skipSeparators.

	[in peek = $)] whileFalse:
	[
		in skipSeparators.
		seq add: (HtmlDTDMatchExpression on: in).
		in skipSeparators.
		('|&,' includes: in peek) ifTrue: [seps:= seps copyWith: in next. in skipSeparators]
			ifFalse: [in peek = $) ifFalse: [self error: 'Huh?']]
	].
	in next.
	
	^(seps isEmpty) ifTrue: [seq last]
		ifFalse: [(seps first = $|) ifTrue: [HtmlDTDMatchOR new expression: seq]
		ifFalse:[(seps first = $&) ifTrue: [HtmlDTDMatchAND new expression: seq]
			ifFalse:[(seps first = $,) ifTrue: [HtmlDTDMatchSequence new expression: seq]
				ifFalse:[self error: 'Invalid group separator']]]]
	! !

!HtmlDTDMatchGroup methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 10:07'!
allowedElements

	| collection |

	collection := OrderedCollection new.

	expression do: [:ea | collection addAll: ea allowedElements].

	^collection asArray! !

!HtmlDTDMatchGroup methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 16:01'!
allowsItem: anItem

	^(expression detect:[:ea| ea allowsItem: anItem] ifNone: [nil]) notNil! !

!HtmlDTDMatchGroup methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 17:10'!
expression: aCollection

	super expression: aCollection asArray! !

!HtmlDTDMatchGroup methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 22:00'!
printOn: aStream

	aStream nextPut: $(.
	aStream nextPutAll: (expression joinOn: (' ',(self separator asString),' ')).
	aStream nextPut: $)
! !

!HtmlDTDMatchGroup methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 17:11'!
separator

	self subclassResponsibility ! !

HtmlDTDMatchGroup subclass: #HtmlDTDMatchOR
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchOR methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 08:55'!
isEnumeration

	^true! !

!HtmlDTDMatchOR methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 17:14'!
separator

	^'|'! !

!HtmlDTDMatchOR methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 12:27'!
validateChildren: aStream ofNode: aNode

	| position matched |

	position := aStream position.

	matched := expression detect:
	[:ea | 
		aStream position: position.		
		(ea validateChildren: aStream ofNode: aNode)
	] ifNone: [nil].
	^matched notNil! !

HtmlDTDMatchGroup subclass: #HtmlDTDMatchSequence
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchSequence methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 17:14'!
separator

	^','! !

!HtmlDTDMatchSequence methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 11:35'!
validateChildren: aStream ofNode: aNode

	expression do:
	[:ea | 
		
		(ea validateChildren: aStream ofNode: aNode) ifFalse: [^false]
	].
	^true! !

HtmlDTDMatchExpression subclass: #HtmlDTDMatchLiteral
	instanceVariableNames: ''
	classVariableNames: 'LiteralsByTagAsSymbol'
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlDTDMatchLiteral class methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 15:00'!
expression: aTag

	LiteralsByTagAsSymbol isNil ifTrue: [LiteralsByTagAsSymbol := WeakValueDictionary new].

	^LiteralsByTagAsSymbol at: aTag asSymbol ifAbsentPut: [self new expression: aTag].	! !

!HtmlDTDMatchLiteral class methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 14:52'!
on: aStringOrStream

	| in out delims s |

	in := aStringOrStream isStream 
		ifFalse: [ReadStream on: aStringOrStream]
		ifTrue:[aStringOrStream].

	delims := ' ()|*?+-,&'.
	out := WriteStream on: String new.

	in skipSeparators.
	[in atEnd or: [delims includes: in peek]] whileFalse: [out nextPut: in next].
	s := out contents.
	^s = 'EMPTY' 
		ifTrue: [HtmlDTDMatchEmpty empty]	
		ifFalse: 
		[
			(s asUppercase findString: 'CDATA') > 0 
			ifTrue:[HtmlDTDMatchCDATA cdata]
			ifFalse:[HtmlDTDMatchLiteral new expression: s]
		]! !

!HtmlDTDMatchLiteral methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 15:09'!
allowedElements

	^ Array with: self! !

!HtmlDTDMatchLiteral methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 14:25'!
allowsItem: aTag

	^expression = aTag! !

!HtmlDTDMatchLiteral methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 23:22'!
consumeStream: in

	expression = in peek ifTrue:[in next]! !

!HtmlDTDMatchLiteral methodsFor: 'as yet unclassified' stamp: 'tb 12/6/2005 10:13'!
isLiteral

	^true! !

!HtmlDTDMatchLiteral methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 17:28'!
printOn: aStream

	aStream nextPut: $"; nextPutAll: expression; nextPut:$".
	! !

!HtmlDTDMatchLiteral methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 12:48'!
validateChildren: aStream ofNode: aNode

	^(aStream atEnd not and: [expression = aStream peek dtdTag]) 
		ifTrue: [aStream next. true] 
		ifFalse: [false]
! !

Object subclass: #HtmlError
	instanceVariableNames: 'parent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

HtmlError subclass: #CSSError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

!CSSError methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 00:07'!
isCSSError

	^true! !

!CSSError methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 00:15'!
printOn: aStream

	super printOn: aStream.
! !

!CSSError methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 23:18'!
title

	| tt |
	tt := self class name. 
	^(tt last: tt size - 3) asCapitalizedPhrase.
	! !

CSSError subclass: #CSSImportAfterStyleRuleError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

CSSError subclass: #CSSPropertyError
	instanceVariableNames: 'propertyName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

CSSPropertyError subclass: #CSSColorAndBackgroundColorNotSet
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

!CSSPropertyError methodsFor: 'accessing' stamp: 'tb 12/31/2005 00:11'!
printOn: aStream

	super printOn: aStream.
	aStream nextPut: $ ; nextPutAll: self propertyName; nextPut: $ .
	self parent printOn: aStream.! !

!CSSPropertyError methodsFor: 'accessing' stamp: 'tb 12/30/2005 21:53'!
propertyName
	"Answer the value of propertyName"

	^ propertyName! !

!CSSPropertyError methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:07'!
propertyName: anObject
	"Set the value of propertyName"

	propertyName := anObject! !

CSSPropertyError subclass: #CSSPropertyNameNotLowercaseError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

CSSPropertyError subclass: #CSSPropertyUnknownError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

CSSPropertyError subclass: #CSSPropertyValueNotValidError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

CSSError subclass: #CSSSelectorError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

CSSSelectorError subclass: #CSSSelectorElementNotLowercaseError
	instanceVariableNames: 'element'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

!CSSSelectorElementNotLowercaseError methodsFor: 'accessing' stamp: 'tb 12/30/2005 21:29'!
element
	"Answer the value of element"

	^ element! !

!CSSSelectorElementNotLowercaseError methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:38'!
element: anObject
	"Set the value of element"

	element := anObject! !

CSSSelectorError subclass: #CSSSelectorElementNotValidError
	instanceVariableNames: 'element'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

!CSSSelectorElementNotValidError methodsFor: 'accessing' stamp: 'tb 12/30/2005 21:45'!
element
	"Answer the value of element"

	^ element! !

!CSSSelectorElementNotValidError methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:38'!
element: anObject
	"Set the value of element"

	element := anObject! !

!CSSSelectorElementNotValidError methodsFor: 'accessing' stamp: 'tb 12/31/2005 01:40'!
printOn: aStream

	self class printOn: aStream.
	aStream nextPut: $ ; nextPutAll: element; nextPut: $ .
	self parent printOn: aStream.! !

!CSSSelectorError methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 00:14'!
printOn: aStream

	super printOn: aStream.
	self parent selector printOn: aStream.! !

CSSSelectorError subclass: #CSSSelectorNotValidError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

CSSError subclass: #CSSUnmatchedRuleError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

!CSSUnmatchedRuleError methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 00:16'!
printOn: aStream

	super printOn: aStream.
	self parent printOn: aStream.! !

CSSError subclass: #CSSUnterminatedPropertyError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

CSSError subclass: #CSSVersion3ConstructFoundError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS-Errors'!

!CSSVersion3ConstructFoundError methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 21:42'!
printOn: aStream

	super printOn: aStream.
	self parent printOn: aStream.! !

HtmlError subclass: #HtmlAttributeError
	instanceVariableNames: 'attribute'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlAttributeError methodsFor: 'accessing' stamp: 'tb 12/4/2005 21:00'!
attribute
	"Answer the value of attribute"

	^ attribute! !

!HtmlAttributeError methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:55'!
attribute: anObject
	"Set the value of attribute"

	attribute := anObject! !

!HtmlAttributeError methodsFor: 'testing' stamp: 'tb 12/5/2005 00:04'!
isAttributeError

	^true! !

!HtmlAttributeError methodsFor: 'accessing' stamp: 'tb 12/5/2005 00:30'!
position

	^self parent start + (self parent prefix findString: self attribute).! !

!HtmlAttributeError methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 00:34'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' Attribute: '.
	self attribute printOn: aStream.
	aStream space.
	self parent prefix printOn: aStream.
	aStream nextPut: $..! !

HtmlAttributeError subclass: #HtmlAttributeMultiplySpecifiedError
	instanceVariableNames: 'value1 value2'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlAttributeMultiplySpecifiedError methodsFor: 'accessing' stamp: 'tb 12/8/2005 23:03'!
printOn: aStream

	self attribute printOn: aStream.
	aStream nextPutAll: ' is specified multiple times in '.
	self parent prefix printOn: aStream.
	aStream nextPutAll: '.'! !

!HtmlAttributeMultiplySpecifiedError methodsFor: 'accessing' stamp: 'tb 12/4/2005 21:08'!
value: v1 value: v2

	value1 := v1.
	value2 := v2! !

HtmlAttributeError subclass: #HtmlAttributeNotLowerCaseError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlAttributeNotLowerCaseError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:01'!
isCaseError! !

!HtmlAttributeNotLowerCaseError methodsFor: 'testing' stamp: 'tb 12/8/2005 22:44'!
printOn: aStream

	| |
	self attribute printOn: aStream.
	aStream nextPutAll: ' should be '.
	self attribute asLowercase printOn: aStream.
	aStream nextPutAll: ' in '.
	self parent prefix printOn: aStream" Valid attributes are: '.
	attrList := self parent dtd attrListForTag: self parent dtdTag.
	(attrList attributes collect:[:ea | ea attribute]) printOn: aStream."! !

HtmlAttributeError subclass: #HtmlAttributeUnknownError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlAttributeUnknownError methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 22:49'!
printOn: aStream

	| attrList |

	self attribute printOn: aStream.
	aStream nextPutAll: ' in: '.
	self parent prefix printOn: aStream.
	aStream nextPutAll: ' should be one of :'.
	attrList := self parent dtd attrListForTag: self parent dtdTag.
	(attrList attributes collect:[:ea | ea attribute]) printOn: aStream.	! !

HtmlAttributeError subclass: #HtmlAttributeValueMissingError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlAttributeValueMissingError methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 23:02'!
printOn: aStream

	aStream nextPutAll: 'Value missing for: '.
	self attribute printOn: aStream.
	aStream nextPutAll: ' in '.
	self parent prefix printOn: aStream.
	aStream nextPutAll: '.'.! !

HtmlAttributeError subclass: #HtmlAttributeValueNotQuotedError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlAttributeValueNotQuotedError methodsFor: 'testing' stamp: 'tb 12/5/2005 00:07'!
isSyntaxError

	^true! !

!HtmlAttributeValueNotQuotedError methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 22:59'!
printOn: aStream

	aStream nextPutAll: 'Value of: '.
	self attribute printOn: aStream.
	aStream nextPutAll: ' in '.
	self parent prefix printOn: aStream.
	aStream nextPutAll: ' should be quoted.'.! !

HtmlAttributeError subclass: #HtmlAttributeValueNotValidError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlAttributeValueNotValidError methodsFor: 'accessing' stamp: 'tb 12/8/2005 22:55'!
printOn: aStream

	| attrList attr |

	"super printOn: aStream."
	self attribute printOn: aStream.
	aStream nextPutAll: ' in '.
	self parent prefix printOn: aStream.
	aStream nextPutAll: ' must be one of: '.
	attrList := self parent dtd attrListForTag: self parent dtdTag.
	attr := attrList attributeNamed: self attribute.
	attr type asString printOn: aStream.	! !

!HtmlAttributeValueNotValidError methodsFor: 'accessing' stamp: 'tb 12/4/2005 21:39'!
validTypeExpression

	| p attList attr |

	p := self parent.
	attList := p dtd attrListForTag: p dtdTag.
	attr := attList attributeNamed: self attribute.
	^ attr type expression asString.! !

!HtmlError methodsFor: 'testing' stamp: 'tb 12/4/2005 21:58'!
isAttributeError

	^false! !

!HtmlError methodsFor: 'testing' stamp: 'tb 12/31/2005 00:08'!
isCSSError

	^false! !

!HtmlError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:01'!
isCaseError

	^false! !

!HtmlError methodsFor: 'testing' stamp: 'tb 12/4/2005 23:31'!
isFatal

	^false! !

!HtmlError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:02'!
isStructuralError

	^false! !

!HtmlError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:00'!
isSyntaxError

	^false! !

!HtmlError methodsFor: 'testing' stamp: 'tb 12/4/2005 21:58'!
isTagError

	^false! !

!HtmlError methodsFor: 'accessing' stamp: 'tb 12/4/2005 22:10'!
parent
	"Answer the value of parent"

	^ parent isNil ifTrue: [nil] ifFalse: [parent first].! !

!HtmlError methodsFor: 'accessing' stamp: 'tb 12/4/2005 22:10'!
parent: anObject
	"Set the value of parent"

	parent := WeakArray with: anObject! !

!HtmlError methodsFor: 'accessing' stamp: 'tb 12/5/2005 00:29'!
position

	^self parent start! !

!HtmlError methodsFor: 'printing' stamp: 'tb 12/4/2005 23:59'!
printOn: aStream

	self class printOn: aStream. 
	aStream nextPutAll: ': Character: '.
	self position printOn: aStream.
	aStream space.! !

!HtmlError methodsFor: 'accessing' stamp: 'tb 12/5/2005 00:50'!
tag

	^self parent tag.! !

!HtmlError methodsFor: 'accessing' stamp: 'tb 12/31/2005 23:17'!
title

	| tt |
	tt := self class name. 
	^(tt last: tt size - 4) asCapitalizedPhrase.
	! !

!HtmlError methodsFor: 'testing' stamp: 'tb 12/5/2005 01:09'!
uniqueId

	^self parent uniqueId! !

HtmlError subclass: #HtmlFatalError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlFatalError methodsFor: 'as yet unclassified' stamp: 'tb 12/4/2005 23:31'!
isFatal

	^true! !

!HtmlFatalError methodsFor: 'as yet unclassified' stamp: 'tb 12/4/2005 22:59'!
printOn: aStream

	super printOn: aStream.
	'ERROR!!  This page is much too broken to parse.' printOn: aStream.! !

HtmlError subclass: #HtmlMissingDocTypeError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlMissingDocTypeError methodsFor: 'as yet unclassified' stamp: 'tb 1/15/2006 20:09'!
isStructuralError

	^true! !

HtmlError subclass: #HtmlSyntaxError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlSyntaxError methodsFor: 'testing' stamp: 'tb 12/4/2005 21:59'!
isSyntaxError

	^true! !

HtmlSyntaxError subclass: #HtmlUnclosedCommentError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlUnclosedCommentError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:02'!
isStructuralError

	^true! !

!HtmlUnclosedCommentError methodsFor: 'testing' stamp: 'tb 12/8/2005 23:08'!
printOn: aStream

	aStream nextPutAll: 'Comment section not properly terminated.'! !

HtmlSyntaxError subclass: #HtmlUnclosedProcessingDirectiveError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlUnclosedProcessingDirectiveError methodsFor: 'testing' stamp: 'tb 12/11/2005 20:26'!
isStructuralError

	^true! !

!HtmlUnclosedProcessingDirectiveError methodsFor: 'testing' stamp: 'tb 12/11/2005 20:27'!
printOn: aStream

	aStream nextPutAll: 'Processing directive not properly terminated.'! !

HtmlSyntaxError subclass: #HtmlUnclosedXmlCDataSectionError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlUnclosedXmlCDataSectionError methodsFor: 'testing' stamp: 'tb 3/12/2006 14:09'!
isStructuralError

	^true! !

!HtmlUnclosedXmlCDataSectionError methodsFor: 'testing' stamp: 'tb 3/12/2006 14:09'!
printOn: aStream

	aStream nextPutAll: 'CDATA section not properly terminated.'! !

HtmlSyntaxError subclass: #HtmlUncommentedScriptError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlUncommentedScriptError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:07'!
isStructuralError

	^true! !

!HtmlUncommentedScriptError methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 23:07'!
printOn: aStream

	'Script should be surrounded by html comments: <!!-- script -->' printOn: aStream.! !

HtmlSyntaxError subclass: #HtmlUnescapedSpecialCharacterError
	instanceVariableNames: 'position'
	classVariableNames: 'CharactersToEntities SpecialCharacters'
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlUnescapedSpecialCharacterError class methodsFor: 'as yet unclassified' stamp: 'tb 12/11/2005 21:58'!
initialize

	"HtmlUnescapedSpecialCharacterError initialize"

	CharactersToEntities := Dictionary new. 

	CharactersToEntities 
		at: $< put: '&lt;';
		at: $> put: '&gt;';
		at: $& put: '&amp;'.

	SpecialCharacters := CharactersToEntities keys! !

!HtmlUnescapedSpecialCharacterError class methodsFor: 'as yet unclassified' stamp: 'tb 12/10/2005 22:51'!
specialCharacters

	^SpecialCharacters! !

!HtmlUnescapedSpecialCharacterError methodsFor: 'accessing' stamp: 'tb 12/10/2005 22:57'!
position

	^position! !

!HtmlUnescapedSpecialCharacterError methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:34'!
position: anObject
	"Set the value of position"

	position := anObject! !

!HtmlUnescapedSpecialCharacterError methodsFor: 'as yet unclassified' stamp: 'tb 12/10/2005 22:52'!
printOn: aStream

	| c |

	c := self parent document content at: self position.

	aStream nextPutAll: 'Character: '.
	c asString printOn: aStream.
	aStream nextPutAll: ' should be '.
	aStream nextPutAll: (CharactersToEntities at: c ifAbsent: ['escaped.']).! !

HtmlError subclass: #HtmlTagError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

HtmlTagError subclass: #HtmlHeadBodyNestingError
	instanceVariableNames: 'nesting'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlHeadBodyNestingError methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 02:41'!
isStructuralError

	^true! !

!HtmlHeadBodyNestingError methodsFor: 'accessing' stamp: 'tb 12/4/2005 23:06'!
nesting
	"Answer the value of nesting"

	^ nesting! !

!HtmlHeadBodyNestingError methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:31'!
nesting: anObject
	"Set the value of nesting"

	nesting := anObject! !

!HtmlHeadBodyNestingError methodsFor: 'accessing' stamp: 'tb 12/5/2005 02:43'!
printOn: aStream

	super printOn: aStream.
	aStream nextPut: $ .
	('Actual tag nesting order found in document is: ', self nesting asString) printOn: aStream.! !

!HtmlTagError methodsFor: 'testing' stamp: 'tb 12/4/2005 21:58'!
isTagError

	^true! !

!HtmlTagError methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 00:48'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: (' Tag: ', self tag)! !

HtmlTagError subclass: #HtmlTagUnknownError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlTagUnknownError methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 23:07'!
printOn: aStream

	self parent tag printOn: aStream.
	aStream nextPutAll: ' is not valid within this doctype.  It can probably be replaced with a span and CSS style.'

	! !

HtmlError subclass: #HtmlTagNotLowerCaseError
	instanceVariableNames: 'tag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlTagNotLowerCaseError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:01'!
isCaseError

	^true! !

!HtmlTagNotLowerCaseError methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 22:35'!
printOn: aStream

	self parent rawTag printOn: aStream.
	aStream nextPutAll: ' should probably be '.
	self parent tag printOn: aStream.! !

!HtmlTagNotLowerCaseError methodsFor: 'accessing' stamp: 'tb 12/4/2005 21:22'!
tag
	"Answer the value of tag"

	^ tag! !

!HtmlTagNotLowerCaseError methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:34'!
tag: anObject
	"Set the value of tag"

	tag := anObject! !

HtmlError subclass: #HtmlTagStructureError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

HtmlTagStructureError subclass: #HtmlTagImproperlyNestedError
	instanceVariableNames: 'wrongTag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlTagImproperlyNestedError methodsFor: 'as yet unclassified' stamp: 'tb 12/16/2005 23:31'!
printOn: aStream

	| rule |

	self parent tag printOn: aStream.
	aStream nextPutAll: ' found within tag: '.
	self parent parent tag printOn: aStream.

	rule := (self parent dtd elementForTag: self parent parent dtdTag).
	rule notNil ifTrue:
	[
		aStream nextPutAll: '. Tag Rule: '.
		rule contentRule printOn: aStream.
	]
	! !

!HtmlTagImproperlyNestedError methodsFor: 'accessing' stamp: 'tb 12/5/2005 11:46'!
wrongTag
	"Answer the value of wrongTag"

	^ wrongTag! !

!HtmlTagImproperlyNestedError methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:33'!
wrongTag: anObject
	"Set the value of wrongTag"

	wrongTag := anObject! !

HtmlTagStructureError subclass: #HtmlTagMissingEndError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlTagMissingEndError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:03'!
isStructuralError

	^true! !

!HtmlTagMissingEndError methodsFor: 'testing' stamp: 'tb 12/8/2005 19:14'!
printOn: aStream

	aStream nextPutAll: 'No closing tag found for tag: '.
	self parent tag printOn: aStream.
	! !

!HtmlTagMissingEndError methodsFor: 'testing' stamp: 'tb 12/5/2005 01:35'!
uniqueId

	^self parent uniqueId, '-end-missing'! !

HtmlTagStructureError subclass: #HtmlTagMissingStartError
	instanceVariableNames: 'position tag'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlTagMissingStartError methodsFor: 'testing' stamp: 'tb 12/4/2005 22:04'!
isStructuralError

	^true! !

!HtmlTagMissingStartError methodsFor: 'accessing' stamp: 'tb 12/4/2005 21:14'!
position
	"Answer the value of position"

	^ position! !

!HtmlTagMissingStartError methodsFor: 'accessing' stamp: 'tb 10/30/2008 16:31'!
position: anObject
	"Set the value of position"

	position := anObject.
	"self halt."

! !

!HtmlTagMissingStartError methodsFor: 'accessing' stamp: 'tb 12/8/2005 19:17'!
printOn: aStream

	aStream nextPutAll: 'Closing tag: '.
	self tag printOn: aStream.
	aStream nextPutAll: ' with no start found'! !

!HtmlTagMissingStartError methodsFor: 'accessing' stamp: 'tb 12/4/2005 21:14'!
tag
	"Answer the value of tag"

	^ tag! !

!HtmlTagMissingStartError methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:55'!
tag: anObject
	"Set the value of tag"

	tag := anObject! !

!HtmlTagMissingStartError methodsFor: 'accessing' stamp: 'tb 12/5/2005 01:26'!
uniqueId

	^super uniqueId, '-end'! !

HtmlTagStructureError subclass: #HtmlTagMustBeEmptyError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM-Errors'!

!HtmlTagMustBeEmptyError methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 23:06'!
printOn: aStream

	self parent tag printOn: aStream.
	aStream nextPutAll: ' must be empty.'! !

!HtmlTagStructureError methodsFor: 'testing' stamp: 'tb 12/5/2005 17:26'!
isStructuralError

	^true! !

Object subclass: #HtmlPListSerializer
	instanceVariableNames: 'stream arrayItemsPerLine dictionaryItemsPerLine indentLevel'
	classVariableNames: 'ReadTranslations WriteTranslations'
	poolDictionaries: ''
	category: 'HTML-PList'!

!HtmlPListSerializer class methodsFor: 'class initialization' stamp: 'tb 8/1/2006 20:36'!
initialize
	"comment stating purpose of message"
	| dict |

	ReadTranslations notNil ifTrue: [^self].

	dict := Dictionary new.
			dict at: $" put: $".
			dict at: $' put: $'.
			dict at: $\ put: $\.
			dict at: $n put: (Character cr).
			dict at: $r put: (Character cr).
			dict at: $t put: (Character tab).
	ReadTranslations := dict.

	dict := Dictionary new.
			dict at: Character tab put: '\t'.
			dict at: Character cr put: '\n'.
			dict at: $" put: '\"'.
			dict at: $' put: ('\', ($' asString)).
			dict at: $\ put: '\\'.
	WriteTranslations := dict
! !

!HtmlPListSerializer class methodsFor: 'instance creation' stamp: 'tb 8/1/2006 21:55'!
on: aStream 
	"Answer an instance of me initialized on a stream."

	self initialize.
	^self basicNew on: aStream! !

!HtmlPListSerializer methodsFor: 'initialize' stamp: 'tb 5/12/2006 10:15'!
close

	stream close.
	stream := nil.! !

!HtmlPListSerializer methodsFor: 'reading' stamp: 'tb 5/12/2006 10:15'!
next
	^self read! !

!HtmlPListSerializer methodsFor: 'reading' stamp: 'tb 5/12/2006 10:15'!
nextPut: anObject

	self write: anObject! !

!HtmlPListSerializer methodsFor: 'initialize' stamp: 'tb 5/12/2006 10:15'!
on: aStream
	"Initialize myself with a stream"

	stream := aStream.
	(stream respondsTo: #ascii) ifTrue: [stream ascii].
	arrayItemsPerLine := 5.
	dictionaryItemsPerLine := 3.
	indentLevel := 0! !

!HtmlPListSerializer methodsFor: 'reading' stamp: 'tb 5/12/2006 10:15'!
read
	"read a complete plist object assembly from the stream and return it"

	| c |

	self skipWhitespace.
	c := stream peek.
		
	 ^(c = ${)  ifTrue: [ self readDictionary ]
		ifFalse: 
		[
			( c = $( ) ifTrue: [ self readArray ]
			ifFalse: 
			[
				( c = $" ) ifTrue: [ self readQuotedString ] 
				ifFalse: [(self readString)]
			]
		]! !

!HtmlPListSerializer methodsFor: 'reading' stamp: 'tb 5/12/2006 10:15'!
readArray
	"read an array"

	| item list |

	self skipWhitespace.

	( stream peek = $( ) 
		ifTrue:
		[
			stream next.
			self skipWhitespace.
			list := OrderedCollection new.
			[stream peek = $) ] whileFalse:
			[
				item := self read.
				list add: item.
				self skipWhitespace.
				(stream peek = $,) ifTrue: [stream next. self skipWhitespace].
			].
			stream next. "consume the trailing paren"
		]
		ifFalse:
		[ 
			"error of some sort here"
		].
		^list! !

!HtmlPListSerializer methodsFor: 'reading' stamp: 'tb 5/12/2006 10:15'!
readDictionary
	"read a dictionary - assumes leading { has already been read"

	| c key value dict |

	self skipWhitespace.

	c := stream next.
	( c = ${ ) 
		ifTrue:
		[
			dict := Dictionary new.
			self skipWhitespace.
			[stream peek = $} ] whileFalse:
			[
				key := self read.
				self skipWhitespace.
				c := stream next.
				( c = $=) ifFalse: [ self error: 'expected = here' ].
				value := self read.
				self skipWhitespace.
				c := stream next.
				( c = $;) ifFalse: [ self error: 'expected ; here' ].
				self skipWhitespace.
				dict at: key put: value.
			].
			stream next.
		]
		ifFalse:
		[ 
			self error: 'not a dictionary'
		].
		^dict! !

!HtmlPListSerializer methodsFor: 'reading' stamp: 'sebastian.sastre 9/19/2010 11:16'!
readQuotedString
	"read a complete plist object assembly from the stream and return it"
	| out c |
	out := WriteStream on: String new.
	self skipWhitespace.
	stream peek = $"
		ifTrue: [stream next.
			[stream peek = $" & stream peek notNil]
				whileFalse: [c := stream next.
					c = $\
						ifTrue: [c := stream next.
							out
								nextPut: (ReadTranslations at: c)]
						ifFalse: [out nextPut: c]].
			stream next].
	^ out contents! !

!HtmlPListSerializer methodsFor: 'reading' stamp: 'tb 5/12/2006 10:15'!
readString
	"read a complete plist object assembly from the stream and return it"

	| out c |

	out := WriteStream on: String new.

	self skipWhitespace.
	[c := stream peek.
		c isAlphaNumeric | ('-_.' includes: c)] whileTrue: [ out nextPut: stream next ].
		
	^out contents.! !

!HtmlPListSerializer methodsFor: 'reading' stamp: 'tb 5/12/2006 10:15'!
skipWhitespace
	"read a dictionary - assumes leading { has already been read"

	[
		| c |
		c := stream peek.
		(c notNil) and: [c isSeparator]
	] whileTrue: [ stream next].
! !

!HtmlPListSerializer methodsFor: 'writing' stamp: 'tb 5/12/2006 10:15'!
write: object
	"write a serialized version of object - either an ordered collection, a dictionary, or something we can convert to a string"

	object isString 
		ifFalse: [ self writeCollection: object]
		ifTrue: [ self writeString: object asString] ! !

!HtmlPListSerializer methodsFor: 'writing' stamp: 'tb 5/12/2006 10:15'!
writeArray: aCollection
	"write a serialized version of the ordered collection"

	| itemsPerLine | 

	itemsPerLine := 0.
	stream nextPutAll: '( '.
	indentLevel := indentLevel + 1.
	aCollection do: 
		[:x | 
			self write: x. 
			stream nextPutAll: ', '. 
			itemsPerLine := itemsPerLine + 1.
			itemsPerLine = arrayItemsPerLine ifTrue:[stream cr. itemsPerLine := 0] 
		].
	indentLevel := indentLevel - 1.
	aCollection isEmpty ifFalse: [ stream position: (stream position-2) ]. "Trim last comma space"
	stream nextPutAll: ')'.! !

!HtmlPListSerializer methodsFor: 'writing' stamp: 'tb 5/12/2006 10:15'!
writeCollection: aCollection
	"write a serialized version of the collection"

	aCollection isSequenceable 
		ifTrue: [ self writeArray: aCollection]
		ifFalse: [ self writeDictionary: aCollection] ! !

!HtmlPListSerializer methodsFor: 'writing' stamp: 'sebastian.sastre 9/6/2010 17:07'!
writeDictionary: aCollection
	"write a serialized version of the ordered collection"

	stream nextPutAll: '{ '; nextPut: Character lf.
	indentLevel := indentLevel + 1.
	aCollection keysAndValuesDo: [:key :value | 
		0 to: indentLevel do: [:ea | stream nextPut: $ ].
		self write: key. 
		stream nextPutAll: ' = '. 
		self write: value.
		stream nextPutAll: '; '; nextPut: Character lf].
	indentLevel := -1.
	stream nextPutAll: '}'.! !

!HtmlPListSerializer methodsFor: 'writing' stamp: 'tb 5/12/2006 10:15'!
writeIndent
	"write the indent"

	0 to: indentLevel do: [:x | stream nextPutAll: '    '] ! !

!HtmlPListSerializer methodsFor: 'writing' stamp: 'tb 5/12/2006 10:15'!
writeString: aString
	"write a string to the stream - quoting and escaping if necessary"

	(aString detect: [:c | ((c isAlphaNumeric) | ('_-.' includes: c)) not] ifNone: [ nil ])
	ifNil: [ stream nextPutAll: aString ]
	ifNotNil: [ stream nextPut: $". 
		aString do: [:c | stream nextPutAll: (WriteTranslations at: c ifAbsent: [ c asString] )]. stream nextPutAll: '" ']! !

Object subclass: #HtmlParseNode
	instanceVariableNames: 'children parent errors start end'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

HtmlParseNode subclass: #CSSElement
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSElement commentStamp: '<historical>' prior: 0!
I am a base class for CSS parse nodes.!

CSSElement subclass: #CSSComment
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSComment methodsFor: 'as yet unclassified' stamp: 'tb 1/6/2006 20:06'!
isAtImportRule

	^false! !

!CSSComment methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 15:42'!
parseContents: aStream

	aStream skip: 2.
	[aStream atEnd] whileFalse:
	[
		(aStream peek = $* and: [aStream next. aStream peek = $/]) ifTrue:
		[
			aStream next.
			^self end: aStream position.
		]
		ifFalse:
		[
			aStream next = $\ ifTrue: [aStream next].
		]
	].
	self end: aStream position.! !

!CSSComment methodsFor: 'as yet unclassified' stamp: 'tb 12/19/2005 23:00'!
tag

	^'comment'! !

!CSSElement methodsFor: 'as yet unclassified' stamp: 'tb 1/6/2006 20:06'!
isAtImportRule

	^false! !

!CSSElement methodsFor: 'as yet unclassified' stamp: 'tb 12/19/2005 23:01'!
isCSS

	^true! !

!CSSElement methodsFor: 'as yet unclassified' stamp: 'tb 1/6/2006 13:03'!
isProperty

	^false! !

!CSSElement methodsFor: 'as yet unclassified' stamp: 'tb 12/31/2005 23:05'!
isStyleRule

	^false! !

!CSSElement methodsFor: 'as yet unclassified' stamp: 'tb 12/21/2005 16:13'!
tag
	^'css'! !

!CSSElement methodsFor: 'as yet unclassified' stamp: 'tb 12/20/2005 13:43'!
validateAgainstDTD: aDTD! !

CSSElement subclass: #CSSProperty
	instanceVariableNames: 'propertyName propertyValue propertyString definition important isDerived'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSProperty class methodsFor: 'instance creation' stamp: 'tb 12/18/2005 01:07'!
propertyNamed: aString

	^self new propertyName: aString! !

!CSSProperty methodsFor: 'accessing' stamp: 'SebastianSastre 1/15/2012 16:38'!
definition

	definition isNil ifTrue: 
	[
		definition := CSSPropertyDefinition named: 
			self propertyName asLowercase trimBoth
	].
	^definition! !

!CSSProperty methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:16'!
definition: anObject
	"Set the value of definition"

	definition := anObject! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 1/4/2006 21:06'!
important
	"Answer the value of important"

	^ important! !

!CSSProperty methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:57'!
important: anObject
	"Set the value of important"

	important := anObject! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 1/5/2006 08:29'!
isDerived
	"Answer the value of isDerived"

	isDerived isNil ifTrue: [isDerived := false].
	^ isDerived! !

!CSSProperty methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:16'!
isDerived: anObject
	"Set the value of isDerived"

	isDerived := anObject! !

!CSSProperty methodsFor: 'testing' stamp: 'tb 1/5/2006 08:25'!
isProperty

	^true! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 1/4/2006 21:08'!
parseContents: aStream

	| s pos in |
	aStream skipSeparators.
	pos := aStream position.
	s := aStream upTo: $;.
	(s includes: $}) ifTrue: 
	[
		aStream position: pos. 
		s := aStream upTo: $}.  
		self addError: CSSUnterminatedPropertyError new.
	].
	in := ReadStream on: s.
	self propertyString: (in upTo: $!!).
	self important: (in match: 'important').! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 1/5/2006 21:21'!
prefix

	^self isDerived ifTrue: 
	[
		| out |
		out := WriteStream on: String new.
		out nextPutAll: '<<';
			nextPutAll: self propertyName;
			nextPutAll: '>>';
			nextPut: $:;
			nextPut: $ .
		self propertyValue do:[:ea | out nextPut: $ ; nextPutAll: ea valueString].
		out nextPut: $;.
		out contents
	]
	ifFalse:
	[
		super prefix
	]! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 12/20/2005 11:20'!
printOn: aStream

	aStream nextPutAll: self prefix! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 12/17/2005 22:05'!
propertyName
	"Answer the value of propertyName"

	^ propertyName! !

!CSSProperty methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
propertyName: anObject
	"Set the value of propertyName"

	propertyName := anObject! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 12/30/2005 01:44'!
propertyString
	"Answer the value of propertyString"

	^ propertyString! !

!CSSProperty methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:57'!
propertyString: anObject
	"Set the value of propertyString"

	propertyString := anObject! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 12/17/2005 23:58'!
propertyValue
	"Answer the value of propertyValue"

	^ propertyValue! !

!CSSProperty methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:17'!
propertyValue: anObject
	"Set the value of propertyValue"

	propertyValue := anObject! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 12/17/2005 23:58'!
tag

	^propertyName! !

!CSSProperty methodsFor: 'accessing' stamp: 'tb 1/6/2006 11:49'!
valueIsValid

	"self propertyName = 'font-family' ifTrue: [self halt]."
	^self definition isNil ifTrue: [false]
	ifFalse: 
	[
		propertyValue isNil ifTrue:
		[
			| list in |
			in := ReadStream on: self propertyString.
			list := OrderedCollection new.
			((self definition valueType addValues: list fromStream: in) and: 
				[in skipSeparators atEnd]) 
			ifTrue: 
			[
				self propertyValue: list.
				list do: [:ea | (ea isString not and: [ea isProperty]) ifTrue: [self parent addProperty: (ea start: self start; end: self end)]]. 
				true
			]
			ifFalse: [false]
		]
		ifFalse: [true]
	].
	! !

CSSElement subclass: #CSSRule
	instanceVariableNames: 'media'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSRule subclass: #CSSAtRule
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

CSSAtRule subclass: #CSSAtImportRule
	instanceVariableNames: 'urlValue stylesheet'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAtImportRule methodsFor: 'testing' stamp: 'tb 12/30/2005 20:05'!
isAtImportRule

	^true! !

!CSSAtImportRule methodsFor: 'as yet unclassified' stamp: 'tb 1/3/2006 17:02'!
parseContents: aStream

	| s m |

	urlValue := CSSUriValue readFrom: aStream.	
	aStream skipSeparators.
	s := aStream upTo: $;.
	self end: aStream position.
	m := s findTokens: ', '.
	m size > 0 ifTrue: [self media: m].	
! !

!CSSAtImportRule methodsFor: 'accessing' stamp: 'tb 12/31/2005 23:00'!
stylesheet
	"Answer the value of styleSheet"

	stylesheet isNil ifTrue: [ stylesheet := CSSStyleSheet fromUrl: urlValue value parent: self].
	^stylesheet! !

!CSSAtImportRule methodsFor: 'accessing' stamp: 'tb 12/30/2005 14:36'!
stylesheet: anObject
	"Set the value of styleSheet"

	stylesheet := anObject! !

!CSSAtImportRule methodsFor: 'accessing' stamp: 'tb 12/30/2005 14:26'!
urlValue
	"Answer the value of urlValue"

	^ urlValue! !

!CSSAtImportRule methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:15'!
urlValue: anObject
	"Set the value of urlValue"

	urlValue := anObject! !

CSSAtRule subclass: #CSSAtPageRule
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSAtRule class methodsFor: 'instance creation' stamp: 'tb 1/6/2006 09:46'!
named: aString

	| out cls |

	out := WriteStream on: String new.
	out nextPutAll: 'CSSAt'.
	out nextPut: aString first asUppercase.
	out nextPutAll: (aString last: aString size -1).
	out nextPutAll: 'Rule'.

	cls := Smalltalk at: out contents asSymbol ifAbsent:[nil].

	^cls new.! !

!CSSAtRule methodsFor: 'as yet unclassified' stamp: 'tb 1/3/2006 17:03'!
parseContents: aStream

	| s |

	s := aStream upTo: $;.
	(s includes: ${) ifTrue: [aStream position: self start. aStream upTo: $}].
	self end: aStream position.
	! !

!CSSRule methodsFor: 'testing' stamp: 'tb 12/30/2005 20:05'!
isAtImportRule

	^false! !

!CSSRule methodsFor: 'testing' stamp: 'tb 12/30/2005 20:05'!
isStyleRule

	^false! !

!CSSRule methodsFor: 'accessing' stamp: 'tb 12/20/2005 00:28'!
media
	"Answer the value of media"

	^ media! !

!CSSRule methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:19'!
media: anObject
	"Set the value of media"

	media := anObject! !

CSSRule subclass: #CSSStyleRule
	instanceVariableNames: 'selector properties matchCount keySelector'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSStyleRule commentStamp: 'tb 12/12/2005 09:05' prior: 0!
I am a parsed CSS rule with selector and dictionary of attributes.

<Foo>
    Value of type Foo. Common types are discussed on the Units page.
Foo
    A keyword that must appear literally (though without case sensitivity). Commas and slashes must also appear literally.
A B C
    A must occur, then B, then C, in that order.
A | B
    A or B must occur.
A || B
    A or B or both must occur, in any order.
[ Foo ]
    Brackets are used to group items together.
Foo*
    Foo is repeated zero or more times.
Foo+
    Foo is repeated one or more times.
Foo?
    Foo is optional.
Foo{A,B}
    Foo must occur at least A times and at most B times. !

CSSStyleRule subclass: #CSSNodeAttributeStyleRule
	instanceVariableNames: 'content'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSNodeAttributeStyleRule methodsFor: 'as yet unclassified' stamp: 'tb 7/12/2007 11:53'!
content

	^content! !

!CSSNodeAttributeStyleRule methodsFor: 'as yet unclassified' stamp: 'tb 7/12/2007 12:09'!
content: aString

	content := '[style="',aString,'"]{',aString,'}'.
	self start: 1.
	self end: content size.
	self parseContents: (ReadStream on: content)! !

!CSSNodeAttributeStyleRule methodsFor: 'as yet unclassified' stamp: 'tb 7/12/2007 12:16'!
printOn: aStream

	aStream nextPutAll: self content! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'tb 12/15/2005 02:05'!
allRules

	| dict |

	dict := Dictionary new.

	dict addAll: (self structuredRulesFrom: self boxPropertyRules).
	dict addAll: (self structuredRulesFrom: self classificationRules).
	dict addAll: (self structuredRulesFrom: self colorAndBackgroundRules).
	dict addAll: (self structuredRulesFrom: self fontRules).
	dict addAll: (self structuredRulesFrom: self textPropertyRules).
	^dict! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'tb 12/12/2005 10:54'!
boxPropertyRules

	^'Top Margin
Syntax: 	margin-top: <value>
Possible Values: 	<length> | <percentage> | auto
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The margin-top property sets the top margin of an element by specifying a length or a percentage. Percentage values refer to the parent element''s width. Negative margins are permitted.

For example, the following rule would eliminate the top margin of a document:

BODY { margin-top: 0 }

Note that adjoining vertical margins are collapsed to use the maximum of the margin values.
Right Margin
Syntax: 	margin-right: <value>
Possible Values: 	<length> | <percentage> | auto
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The margin-right property sets the right margin of an element by specifying a length or a percentage. Percentage values refer to the parent element''s width. Negative margins are permitted.

Example:

P.narrow { margin-right: 50% }

Note that adjoining horizontal margins are not collapsed.
Bottom Margin
Syntax: 	margin-bottom: <value>
Possible Values: 	<length> | <percentage> | auto
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The margin-bottom property sets the bottom margin of an element by specifying a length or a percentage. Percentage values refer to the parent element''s width. Negative margins are permitted.

Example:

DT { margin-bottom: 3em }

Note that adjoining vertical margins are collapsed to use the maximum of the margin values.
Left Margin
Syntax: 	margin-left: <value>
Possible Values: 	<length> | <percentage> | auto
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The margin-left property sets the left margin of an element by specifying a length or a percentage. Percentage values refer to the parent element''s width. Negative margins are permitted.

Example:

ADDRESS { margin-left: 50% }

Note that adjoining horizontal margins are not collapsed.
Margin
Syntax: 	margin: <value>
Possible Values: 	[ <length> | <percentage> | auto ]{1,4}
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	No

The margin property sets the margins of an element by specifying between one and four values, where each value is a length, a percentage, or auto. Percentage values refer to the parent element''s width. Negative margins are permitted.

If four values are given, they apply to top, right, bottom, and left margin, respectively. If one value is given, it applies to all sides. If two or three values are given, the missing values are taken from the opposite side.

Examples of margin declarations include:

BODY { margin: 5em }             /* all margins 5em */
P    { margin: 2em 4em }         /* top and bottom margins 2em,
                                    left and right margins 4em */
DIV  { margin: 1em 2em 3em 4em } /* top margin 1em,
                                    right margin 2em,
                                    bottom margin 3em,
                                    left margin 4em */

Note that adjoining vertical margins are collapsed to use the maximum of the margin values. Horizontal margins are not collapsed.

Using the margin property allows one to set all margins; alternatively, the properties margin-top, margin-bottom, margin-left, and margin-right may be used.
Top Padding
Syntax: 	padding-top: <value>
Possible Values: 	<length> | <percentage>
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The padding-top property describes how much space to put between the top border and the content of the selector. The value is either a length or a percentage. Percentage values refer to the parent element''s width. Negative values are not permitted.
Right Padding
Syntax: 	padding-right: <value>
Possible Values: 	<length> | <percentage>
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The padding-right property describes how much space to put between the right border and the content of the selector. The value is either a length or a percentage. Percentage values refer to the parent element''s width. Negative values are not permitted.
Bottom Padding
Syntax: 	padding-bottom: <value>
Possible Values: 	<length> | <percentage>
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The padding-bottom property describes how much space to put between the bottom border and the content of the selector. The value is either a length or a percentage. Percentage values refer to the parent element''s width. Negative values are not permitted.
Left Padding
Syntax: 	padding-left: <value>
Possible Values: 	<length> | <percentage>
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The padding-left property describes how much space to put between the left border and the content of the selector. The value is either a length or a percentage. Percentage values refer to the parent element''s width. Negative values are not permitted.
Padding
Syntax: 	padding: <value>
Possible Values: 	[ <length> | <percentage> ]{1,4}
Initial Value: 	0
Applies to: 	All elements
Inherited: 	No

The padding property is a shorthand for the padding-top, padding-right, padding-bottom, and padding-left properties.

An element''s padding is the amount of space between the border and the content of the element. Between one and four values are given, where each value is either a length or a percentage. Percentage values refer to the parent element''s width. Negative values are not permitted.

If four values are given, they apply to top, right, bottom, and left padding, respectively. If one value is given, it applies to all sides. If two or three values are given, the missing values are taken from the opposite side.

For example, the following rule sets the top padding to 2em, the right padding to 4em, the bottom padding to 5em, and the left padding to 4em:

BLOCKQUOTE { padding: 2em 4em 5em }

Top Border Width
Syntax: 	border-top-width: <value>
Possible Values: 	thin | medium | thick | <length>
Initial Value: 	medium
Applies to: 	All elements
Inherited: 	No

The border-top-width property is used to specify the width of an element''s top border. The value may be one of three keywords, which are not affected by font size, or a length, which can be used to achieve relative widths. Negative values are not allowed.

One may also use the border-top, border-width, or border shorthand properties.
Right Border Width
Syntax: 	border-right-width: <value>
Possible Values: 	thin | medium | thick | <length>
Initial Value: 	medium
Applies to: 	All elements
Inherited: 	No

The border-right-width property is used to specify the width of an element''s right border. The value may be one of three keywords, which are not affected by font size, or a length, which can be used to achieve relative widths. Negative values are not allowed.

One may also use the border-right, border-width, or border shorthand properties.
Bottom Border Width
Syntax: 	border-bottom-width: <value>
Possible Values: 	thin | medium | thick | <length>
Initial Value: 	medium
Applies to: 	All elements
Inherited: 	No

The border-bottom-width property is used to specify the width of an element''s bottom border. The value may be one of three keywords, which are not affected by font size, or a length, which can be used to achieve relative widths. Negative values are not allowed.

One may also use the border-bottom, border-width, or border shorthand properties.
Left Border Width
Syntax: 	border-left-width: <value>
Possible Values: 	thin | medium | thick | <length>
Initial Value: 	medium
Applies to: 	All elements
Inherited: 	No

The border-left-width property is used to specify the width of an element''s left border. The value may be one of three keywords, which are not affected by font size, or a length, which can be used to achieve relative widths. Negative values are not allowed.

One may also use the border-left, border-width, or border shorthand properties.
Border Width
Syntax: 	border-width: <value>
Possible Values: 	[ thin | medium | thick | <length> ]{1,4}
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	No

The border-width property is used to set the border width of an element by specifying between one and four values, where each value is a keyword or a length. Negative lengths are not permitted.

If four values are given, they apply to top, right, bottom, and left border width, respectively. If one value is given, it applies to all sides. If two or three values are given, the missing values are taken from the opposite side.

This property is a shorthand for the border-top-width, border-right-width, border-bottom-width, and border-left-width properties.

One may also use the border shorthand property.
Border Color
Syntax: 	border-color: <value>
Possible Values: 	<color>{1,4}
Initial Value: 	The value of the color property
Applies to: 	All elements
Inherited: 	No

The border-color property sets the color of an element''s border. Between one and four color values are specified. If four values are given, they apply to top, right, bottom, and left border color, respectively. If one value is given, it applies to all sides. If two or three values are given, the missing values are taken from the opposite side.

One may also use the border shorthand property.
Border Style
Syntax: 	border-style: <value>
Possible Values: 	[ none | dotted | dashed | solid | double | groove | ridge | inset | outset ]{1,4}
Initial Value: 	none
Applies to: 	All elements
Inherited: 	No

The border-style property sets the style of an element''s border. This property must be specified for the border to be visible.

Between one and four keywords are specified. If four values are given, they apply to top, right, bottom, and left border style, respectively. If one value is given, it applies to all sides. If two or three values are given, the missing values are taken from the opposite side.

One may also use the border shorthand property.
Top Border
Syntax: 	border-top: <value>
Possible Values: 	<border-top-width> || <border-style> || <color>
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	No

The border-top property is a shorthand for setting the width, style, and color of an element''s top border.

Note that only one border-style value may be given.

One may also use the border shorthand property.
Right Border
Syntax: 	border-right: <value>
Possible Values: 	<border-right-width> || <border-style> || <color>
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	No

The border-right property is a shorthand for setting the width, style, and color of an element''s right border.

Note that only one border-style value may be given.

One may also use the border shorthand property.
Bottom Border
Syntax: 	border-bottom: <value>
Possible Values: 	<border-bottom-width> || <border-style> || <color>
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	No

The border-bottom property is a shorthand for setting the width, style, and color of an element''s bottom border.

Note that only one border-style value may be given.

One may also use the border shorthand property.
Left Border
Syntax: 	border-left: <value>
Possible Values: 	<border-left-width> || <border-style> || <color>
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	No

The border-left property is a shorthand for setting the width, style, and color of an element''s left border.

Note that only one border-style value may be given.

One may also use the border shorthand property.
Border
Syntax: 	border: <value>
Possible Values: 	<border-width> || <border-style> || <color>
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	No

The border property is a shorthand for setting the width, style, and color of an element''s border.

Examples of border declarations include:

H2        { border: groove 3em }
A:link    { border: solid blue }
A:visited { border: thin dotted #800080 }
A:active  { border: thick double red }

The border property can only set all four borders; only one border width and border style may be given. To give different values to an element''s four borders, an author must use one or more of the border-top, border-right, border-bottom, border-left, border-color, border-width, border-style, border-top-width, border-right-width, border-bottom-width, or border-left-width properties.
Width
Syntax: 	width: <value>
Possible Values: 	<length> | <percentage> | auto
Initial Value: 	auto
Applies to: 	Block-level and replaced elements
Inherited: 	No

Each block-level or replaced element can be given a width, specified as a length, a percentage, or as auto. (A replaced element is one for which only the intrinsic dimensions are known; HTML replaced elements include IMG, INPUT, TEXTAREA, SELECT, and OBJECT.) The initial value for the width property is auto, which results in the element''s intrinsic width (i.e., the width of the element itself, for example the width of an image). Percentages refer to the parent element''s width. Negative values are not allowed.

This property could be used to give common widths to some INPUT elements, such as submit and reset buttons:

INPUT.button { width: 10em }

Height
Syntax: 	height: <value>
Possible Values: 	<length> | auto
Initial Value: 	auto
Applies to: 	Block-level and replaced elements
Inherited: 	No

Each block-level or replaced element can be given a height, specified as a length or as auto. (A replaced element is one for which only the intrinsic dimensions are known; HTML replaced elements include IMG, INPUT, TEXTAREA, SELECT, and OBJECT.) The initial value for the height property is auto, which results in the element''s intrinsic height (i.e., the height of the element itself, for example the height of an image). Negative lengths are not allowed.

As with the width property, height can be used to scale an image:

IMG.foo { width: 40px; height: 40px }

In most cases, authors are advised to scale the image in an image editing program, since browsers will not likely scale images with high quality, and since scaling down causes the user to download an unnecessarily large file. However, scaling through the width and height properties is a useful option for user-defined style sheets in order to overcome vision problems.
Float
Syntax: 	float: <value>
Possible Values: 	left | right | none
Initial Value: 	none
Applies to: 	All elements
Inherited: 	No

The float property allows authors to wrap text around an element. This is identical in purpose to HTML 3.2''s ALIGN=left and ALIGN=right for the IMG element, but CSS1 allows all elements to "float," not just the images and tables that HTML 3.2 allows.
Clear
Syntax: 	clear: <value>
Possible Values: 	none | left | right | both
Initial Value: 	none
Applies to: 	All elements
Inherited: 	No

The clear property specifies if an element allows floating elements to its sides. A value of left moves the element below any floating element on its left; right acts similarly for floating elements on the right. Other values are none, which is the initial value, and both, which moves the element below floating elements on both of its sides. This property is similar in function to HTML 3.2''s <BR CLEAR=left|right|all|none>, but it can be applied to all elements.'! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'tb 12/15/2005 03:18'!
classNameFromProperty: aString

	| in s nm |
	
	in := ReadStream on: aString.
	nm := ''.

	[in atEnd] whileFalse:
	[
		s := in upTo: $-.
		s := ((s first: 1) asUppercase),(s last: (s size -1)).
		nm := nm,s. 
	].

	^'CSS',nm,'Property'
	! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'tb 12/12/2005 10:56'!
classificationRules

	^'Display
Syntax: 	display: <value>
Possible Values: 	block | inline | list-item | none
Initial Value: 	block
Applies to: 	All elements
Inherited: 	No

The display property is used to define an element with one of four values:

    * block (a line break before and after the element)
    * inline (no line break before and after the element)
    * list-item (same as block except a list-item marker is added)
    * none (no display)

Each element typically is given a default display value by the browser, based on suggested rendering in the HTML specification.

The display property can be dangerous because of its ability to display elements in what would otherwise be an improper format. The use of the value none will turn off display of the element to which it is assigned, including any children elements!!
Whitespace
Syntax: 	white-space: <value>
Possible Values: 	normal | pre | nowrap
Initial Value: 	normal
Applies to: 	All elements
Inherited: 	Yes

The white-space property will determine how spaces within the element are treated. This property takes one of three values:

    * normal (collapses multiple spaces into one)
    * pre (does not collapse multiple spaces)
    * nowrap (does not allow line wrapping without a <BR> tag)

List Style Type
Syntax: 	list-style-type: <value>
Possible Values: 	disc | circle | square | decimal | lower-roman | upper-roman | lower-alpha | upper-alpha | none
Initial Value: 	disc
Applies to: 	Elements with display value list-item
Inherited: 	Yes

The list-style-type property specifies the type of list-item marker, and is used if list-style-image is none or if image loading is turned off.

Examples:

LI.square { list-style-type: square }
UL.plain  { list-style-type: none }
OL        { list-style-type: upper-alpha }  /* A B C D E etc. */
OL OL     { list-style-type: decimal }      /* 1 2 3 4 5 etc. */
OL OL OL  { list-style-type: lower-roman }  /* i ii iii iv v etc. */

List Style Image
Syntax: 	list-style-image: <value>
Possible Values: 	<url> | none
Initial Value: 	none
Applies to: 	Elements with display value list-item
Inherited: 	Yes

The list-style-image property specifies the image that will be used as list-item marker when image loading is turned on, replacing the marker specified in the list-style-type property.

Examples:

UL.check { list-style-image: url(/LI-markers/checkmark.gif) }
UL LI.x  { list-style-image: url(x.png) }

List Style Position
Syntax: 	list-style-position: <value>
Possible Values: 	inside | outside
Initial Value: 	outside
Applies to: 	Elements with display value list-item
Inherited: 	Yes

The list-style-position property takes the value inside or outside, with outside being the default. This property determines where the marker is placed in regard to the list item. If the value inside is used, the lines will wrap under the marker instead of being indented. An example rendering is:

Outside rendering:
 * List item 1
   second line of list item

Inside rendering:
   * List item 1
   second line of list item

List Style
Syntax: 	list-style: <value>
Possible Values: 	<list-style-type> || <list-style-position> || <url>
Initial Value: 	Not defined
Applies to: 	Elements with display value list-item
Inherited: 	Yes

The list-style property is a shorthand for the list-style-type, list-style-position, and list-style-image properties.

Examples:

LI.square { list-style: square inside }
UL.plain  { list-style: none }
UL.check  { list-style: url(/LI-markers/checkmark.gif) circle }
OL        { list-style: upper-alpha }
OL OL     { list-style: lower-roman inside }'! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'tb 12/12/2005 09:20'!
colorAndBackgroundRules

	^'Color
Syntax: 	color: <color>
Initial Value: 	Determined by browser
Applies to: 	All elements
Inherited: 	Yes

The color property allows authors to specify the color of an element. See the Units section for color value descriptions. Some example color rules include:

H1 { color: blue }
H2 { color: #000080 }
H3 { color: #0c0 }

To help avoid conflicts with user style sheets, background and color properties should always be specified together.
Background Color
Syntax: 	background-color: <value>
Possible Values: 	<color> | transparent
Initial Value: 	transparent
Applies to: 	All elements
Inherited: 	No

The background-color property sets the background color of an element. For example:

BODY { background-color: white }
H1   { background-color: #000080 }

To help avoid conflicts with user style sheets, background-image should be specified whenever background-color is used. In most cases, background-image: none is suitable.

Authors may also use the shorthand background property, which is currently better supported than the background-color property.
Background Image
Syntax: 	background-image: <value>
Possible Values: 	<url> | none
Initial Value: 	none
Applies to: 	All elements
Inherited: 	No

The background-image property sets the background image of an element. For example:

BODY { background-image: url(/images/foo.gif) }
P    { background-image: url(http://www.htmlhelp.com/bg.png) }

When a background image is defined, a similar background color should also be defined for those not loading images.

Authors may also use the shorthand background property, which is currently better supported than the background-image property.
Background Repeat
Syntax: 	background-repeat: <value>
Possible Values: 	repeat | repeat-x | repeat-y | no-repeat
Initial Value: 	repeat
Applies to: 	All elements
Inherited: 	No

The background-repeat property determines how a specified background image is repeated. The repeat-x value will repeat the image horizontally while the repeat-y value will repeat the image vertically. For example:

BODY { background: white url(candybar.gif);
       background-repeat: repeat-x }

In the above example, the image will only be tiled horizontally.

Authors may also use the shorthand background property, which is currently better supported than the background-repeat property.
Background Attachment
Syntax: 	background-attachment: <value>
Possible Values: 	scroll | fixed
Initial Value: 	scroll
Applies to: 	All elements
Inherited: 	No

The background-attachment property determines if a specified background image will scroll with the content or be fixed with regard to the canvas. For example, the following specifies a fixed background image:

BODY { background: white url(candybar.gif);
       background-attachment: fixed }

Authors may also use the shorthand background property, which is currently better supported than the background-attachment property.
Background Position
Syntax: 	background-position: <value>
Possible Values: 	[<percentage> | <length>]{1,2} | [top | center | bottom] || [left | center | right]
Initial Value: 	0% 0%
Applies to: 	Block-level and replaced elements
Inherited: 	No

The background-position property gives the initial position of a specified background image. This property may only be applied to block-level elements and replaced elements. (A replaced element is one for which only the intrinsic dimensions are known; HTML replaced elements include IMG, INPUT, TEXTAREA, SELECT, and OBJECT.)

The easiest way to assign a background position is with keywords:

    * Horizontal keywords (left, center, right)
    * Vertical keywords (top, center, bottom)

Percentages and lengths may also be used to assign the position of the background image. Percentages are relative to the size of the element. Although lengths are allowed, they are not recommended due to their inherent weakness in dealing with differing display resolutions.

When using percentages or lengths, the horizontal position is specified first, followed by the vertical position. A value such as 20% 65% specifies that the point 20% across and 65% down the image be placed at the point 20% across and 65% down the element. A value such as 5px 10px specifies that the upper left corner of the image be placed 5 pixels to the right of and 10 pixels below the upper left corner of the element.

If only the horizontal value is given, the vertical position will be 50%. Combinations of lengths and percentages are allowed, as are negative positions. For example, 10% -2cm is permitted. However, percentages and lengths cannot be combined with keywords.

The keywords are interpreted as follows:

    * top left = left top = 0% 0%
    * top = top center = center top = 50% 0%
    * right top = top right = 100% 0%
    * left = left center = center left = 0% 50%
    * center = center center = 50% 50%
    * right = right center = center right = 100% 50%
    * bottom left = left bottom = 0% 100%
    * bottom = bottom center = center bottom = 50% 100%
    * bottom right = right bottom = 100% 100%

If the background image is fixed with regard to the canvas, the image is placed relative to the canvas instead of the element.

Authors may also use the shorthand background property, which is currently better supported than the background-position property.
Background
Syntax: 	background: <value>
Possible Values: 	<background-color> || <background-image> || <background-repeat> || <background-attachment> || <background-position>
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	No

The background property is a shorthand for the more specific background-related properties. Some examples of background declarations follow:

BODY       { background: white url(http://www.htmlhelp.com/foo.gif) }
BLOCKQUOTE { background: #7fffd4 }
P          { background: url(../backgrounds/pawn.png) #f0f8ff fixed }
TABLE      { background: #0c0 url(leaves.jpg) no-repeat bottom right }

A value not specified will receive its initial value. For example, in the first three rules above, the background-position property will be set to 0% 0%.

To help avoid conflicts with user style sheets, background and color properties should always be specified together.'! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'tb 12/12/2005 09:18'!
fontRules

	^'Syntax:  	font-family: [[<family-name> | <generic-family>],]* [<family-name> | <generic-family>]
Possible Values: 	<family-name>

    * Any font family name may be used

<generic-family>

    * serif (e.g., Times)
    * sans-serif (e.g., Arial or Helvetica)
    * cursive (e.g., Zapf-Chancery)
    * fantasy (e.g., Western)
    * monospace (e.g., Courier)

Initial Value: 	Determined by browser
Applies to: 	All elements
Inherited: 	Yes

Font families may be assigned by a specific font name or a generic font family. Obviously, defining a specific font will not be as likely to match as a generic font family. Multiple family assignments can be made, and if a specific font assignment is made it should be followed by a generic family name in case the first choice is not present.

A sample font-family declaration might look like this:

P { font-family: "New Century Schoolbook", Times, serif }

Notice that the first two assignments are specific type faces: New Century Schoolbook and Times. However, since both of them are serif fonts, the generic font family is listed as a backup in case the system does not have either of these but has another serif font which meets the qualifications.

Any font name containing whitespace must be quoted, with either single or double quotes.

The font family may also be given with the font property.
Font Style
Syntax: 	font-style: <value>
Possible Values: 	normal | italic | oblique
Initial Value: 	normal
Applies to: 	All elements
Inherited: 	Yes

The font-style property defines that the font be displayed in one of three ways: normal, italic or oblique (slanted). A sample style sheet with font-style declarations might look like this:

H1 { font-style: oblique }
P  { font-style: normal }

Font Variant
Syntax: 	font-variant: <value>
Possible Values: 	normal | small-caps
Initial Value: 	normal
Applies to: 	All elements
Inherited: 	Yes

The font-variant property determines if the font is to display in normal or small-caps. Small-caps are displayed when all the letters of the word are in capitals with uppercase characters slightly larger than lowercase. Later versions of CSS may support additional variants such as condensed, expanded, small-caps numerals or other custom variants. An example of a font-variant assignment would be:

SPAN { font-variant: small-caps }

Font Weight
Syntax: 	font-weight: <value>
Possible Values: 	normal | bold | bolder | lighter | 100 | 200 | 300 | 400 | 500 | 600 | 700 | 800 | 900
Initial Value: 	normal
Applies to: 	All elements
Inherited: 	Yes

The font-weight property is used to specify the weight of the font. The bolder and lighter values are relative to the inherited font weight, while the other values are absolute font weights.

Note: Since not all fonts have nine possible display weights some of the weights may be grouped together in assignment. If the specified weight is not available, an alternate will be chosen on the following basis:

    * 500 may be switched with 400, and vice versa
    * 100-300 may be assigned to the next lighter weight, if any, or the next darker weight otherwise
    * 600-900 may be assigned to the next darker weight, if any, or the next lighter weight otherwise

Some example font-weight assignments would be:

H1 { font-weight: 800 }
P  { font-weight: normal }

Font Size
Syntax: 	font-size: <absolute-size> | <relative-size> | <length> | <percentage>
Possible Values: 	

    * <absolute-size>
          o xx-small | x-small | small | medium | large | x-large | xx-large
    * <relative-size>
          o larger | smaller
    * <length>
    * <percentage> (in relation to parent element) 


Initial Value: 	medium
Applies to: 	All elements
Inherited: 	Yes

The font-size property is used to modify the size of the displayed font. Absolute lengths (using units like pt and in) should be used sparingly due to their weakness in adapting to different browsing environments. Fonts with absolute lengths can very easily be too small or too large for a user.

Some example size assignments would be:

H1     { font-size: large }
P      { font-size: 12pt }
LI     { font-size: 90% }
STRONG { font-size: larger }

Authors should be aware that Microsoft Internet Explorer 3.x incorrectly treats em and ex units as pixels, which can easily make text using these units unreadable. The browser also incorrectly applies percentage values relative to its default font size for the selector, rather than relative to the parent element''s font size. This makes rules like

H1 { font-size: 200% }

dangerous in that the size will be twice IE''s default font size for level-one headings, rather than twice the parent element''s font size. In this case, BODY would most likely be the parent element, and it would likely define a medium font size, whereas the default level-one heading font size imposed by IE would probably be considered xx-large.

Given these bugs, authors should take care in using percentage values for font-size, and should probably avoid em and ex units for this property.
Font
Syntax: 	font: <value>
Possible Values: 	[ <font-style> || <font-variant> || <font-weight> ]? <font-size> [ / <line-height> ]? <font-family>
Initial Value: 	Not defined
Applies to: 	All elements
Inherited: 	Yes

The font property may be used as a shorthand for the various font properties, as well as the line height. For example,

P { font: italic bold 12pt/14pt Times, serif }

specifies paragraphs with a bold and italic Times or serif font with a size of 12 points and a line height of 14 points.'! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'SebastianSastre 1/15/2012 16:39'!
fromString: aString

	| in sel attrString pairs rules |

	rules := OrderedCollection new.
	in := ReadStream on: aString.
	
	[in atEnd] whileFalse:
	[
		| rule |
		sel := (in upTo: ${) trimBoth.
		(in atEnd not and: [sel size > 0]) ifTrue:
		[
			rule:= self new selector: sel.
			rules add: rule.
			attrString := (in upTo: $}) trimBoth.
			pairs := attrString findTokens: ';'.	
			pairs do: 
			[:ea || k v p |  
				p := ea splitOn: ':'. 
				k := p first trimBoth.
				v := p last trimBoth.
				rule properties at: k put: v. 
			]
		]
	].
	^rules! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'SebastianSastre 1/15/2012 16:39'!
structuredRulesFrom: aString

	| current style dict | 

	dict := Dictionary new.

	aString linesDo: 
	[:ea |  
		(ea beginsWith: 'Syntax:') ifTrue: 
		[
			current := Dictionary new.
			style := (ea findTokens: ':') second trimBoth.
			dict at: style put: current.
		]
		ifFalse:
		[
			(ea beginsWith: 'Possible Values:') ifTrue: 
			[
				current at: 'possibleValues' 
					put: (ea last: (ea size - 'Possible Values:' size)) trimBoth.
			]
			ifFalse:
			[
				(ea beginsWith: 'Initial Value:') ifTrue:
				[
					current at: 'initialValue' 
						put: (ea last: (ea size - 'Initial Value:' size)) trimBoth.
				]
				ifFalse:
				[
					(ea beginsWith: 'Applies to:') ifTrue:
					[
						current at: 'appliesTo' 
							put: (ea last: (ea size - 'Applies to:' size)) trimBoth.
					]
					ifFalse:
					[
						(ea beginsWith: 'Inherited:') ifTrue: 
						[
							current at: 'inherited' 
								put: (ea last: (ea size - 'Inherited:' size)) trimBoth.
						]
					]
				]
			]
		]
	].
	^ dict
! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'tb 12/12/2005 09:22'!
textPropertyRules

	^'Word Spacing
Syntax: 	word-spacing: <value>
Possible Values: 	normal | <length>
Initial Value: 	normal
Applies to: 	All elements
Inherited: 	Yes

The word-spacing property defines an additional amount of space between words. The value must be in the length format; negative values are permitted.

Examples:

P EM   { word-spacing: 0.4em }
P.note { word-spacing: -0.2em }

Letter Spacing
Syntax: 	letter-spacing: <value>
Possible Values: 	normal | <length>
Initial Value: 	normal
Applies to: 	All elements
Inherited: 	Yes

The letter-spacing property defines an additional amount of space between characters. The value must be in the length format; negative values are permitted. A setting of 0 will prevent justification.

Examples:

H1     { letter-spacing: 0.1em }
P.note { letter-spacing: -0.1em }

Text Decoration
Syntax: 	text-decoration: <value>
Possible Values: 	none | [ underline || overline || line-through || blink ]
Initial Value: 	none
Applies to: 	All elements
Inherited: 	No

The text-decoration property allows text to be decorated through one of five properties: underline, overline, line-through, blink, or the default, none.

For example, one can suggest that links not be underlined with

A:link, A:visited, A:active { text-decoration: none }

Vertical Alignment
Syntax: 	vertical-align: <value>
Possible Values: 	baseline | sub | super | top | text-top | middle | bottom | text-bottom | <percentage>
Initial Value: 	baseline
Applies to: 	Inline elements
Inherited: 	No

The vertical-align property may be used to alter the vertical positioning of an inline element, relative to its parent element or to the element''s line. (An inline element is one which has no line break before and after it, for example, EM, A, and IMG in HTML.)

The value may be a percentage relative to the element''s line-height property, which would raise the element''s baseline the specified amount above the parent''s baseline. Negative values are permitted.

The value may also be a keyword. The following keywords affect the positioning relative to the parent element:

    * baseline (align baselines of element and parent)
    * middle (align vertical midpoint of element with baseline plus half the x-height--the height of the letter "x"--of the parent)
    * sub (subscript)
    * super (superscript)
    * text-top (align tops of element and parent''s font)
    * text-bottom (align bottoms of element and parent''s font)

The keywords affecting the positioning relative to the element''s line are

    * top (align top of element with tallest element on the line)
    * bottom (align bottom of element with lowest element on the line)

The vertical-align property is particularly useful for aligning images. Some examples follow:

IMG.middle { vertical-align: middle }
IMG        { vertical-align: 50% }
.exponent  { vertical-align: super }

Text Transformation
Syntax: 	text-transform: <value>
Possible Values: 	none | capitalize | uppercase | lowercase
Initial Value: 	none
Applies to: 	All elements
Inherited: 	Yes

The text-transform property allows text to be transformed by one of four properties:

    * capitalize (capitalizes first character of each word)
    * uppercase (capitalizes all characters of each word)
    * lowercase (uses small letters for all characters of each word)
    * none (the initial value)

Examples:

H1 { text-transform: uppercase }
H2 { text-transform: capitalize }

The text-transform property should only be used to express a stylistic desire. It would be inappropriate, for example, to use text-transform to capitalize a list of countries or names.
Text Alignment
Syntax: 	text-align: <value>
Possible Values: 	left | right | center | justify
Initial Value: 	Determined by browser
Applies to: 	Block-level elements
Inherited: 	Yes

The text-align property can be applied to block-level elements (P, H1, etc.) to give the alignment of the element''s text. This property is similar in function to HTML''s ALIGN attribute on paragraphs, headings, and divisions.

Some examples follow:

H1          { text-align: center }
P.newspaper { text-align: justify }

Text Indentation
Syntax: 	text-indent: <value>
Possible Values: 	<length> | <percentage>
Initial Value: 	0
Applies to: 	Block-level elements
Inherited: 	Yes

The text-indent property can be applied to block-level elements (P, H1, etc.) to define the amount of indentation that the first line of the element should receive. The value must be a length or a percentage; percentages refer to the parent element''s width. A common use of text-indent would be to indent a paragraph:

P { text-indent: 5em }

Line Height
Syntax: 	line-height: <value>
Possible Values: 	normal | <number> | <length> | <percentage>
Initial Value: 	normal
Applies to: 	All elements
Inherited: 	Yes

The line-height property will accept a value to control the spacing between baselines of text. When the value is a number, the line height is calculated by multiplying the element''s font size by the number. Percentage values are relative to the element''s font size. Negative values are not permitted.

Line height may also be given in the font property along with a font size.

The line-height property could be used to double space text:

P { line-height: 200% }

Microsoft Internet Explorer 3.x incorrectly treats number values and values with em or ex units as pixel values. This bug can easily make pages unreadable, and so authors should avoid provoking it wherever possible; percentage units are often a good choice.'! !

!CSSStyleRule class methodsFor: 'definitions' stamp: 'tb 12/12/2005 10:57'!
unitRules

	^'Length Units

A length value is formed by an optional + or -, followed by a number, followed by a two-letter abbreviation that indicates the unit. There are no spaces in a length value; e.g., 1.3 em is not a valid length value, but 1.3em is valid. A length of 0 does not require the two-letter unit identifier.

Both relative and absolute length units are supported in CSS1. Relative units give a length relative to another length property, and are preferred since they will better adjust to different media. The following relative units are available:

    * em (ems, the height of the element''s font)
    * ex (x-height, the height of the letter "x")
    * px (pixels, relative to the canvas resolution)

Absolute length units are highly dependent on the output medium, and so are less useful than relative units. The following absolute units are available:

    * in (inches; 1in=2.54cm)
    * cm (centimeters; 1cm=10mm)
    * mm (millimeters)
    * pt (points; 1pt=1/72in)
    * pc (picas; 1pc=12pt)

Percentage Units

A percentage value is formed by an optional + or -, followed by a number, followed by %. There are no spaces in a percentage value.

Percentage values are relative to other values, as defined for each property. Most often the percentage value is relative to the element''s font size.
Color Units

A color value is a keyword or a numerical RGB specification.

The 16 keywords are taken from the Windows VGA palette: aqua, black, blue, fuchsia, gray, green, lime, maroon, navy, olive, purple, red, silver, teal, white, and yellow.

RGB colors are given in one of four ways:

    * #rrggbb (e.g., #00cc00)
    * #rgb (e.g., #0c0)
    * rgb(x,x,x) where x is an integer between 0 and 255 inclusive (e.g., rgb(0,204,0))
    * rgb(y%,y%,y%) where y is a number between 0.0 and 100.0 inclusive (e.g., rgb(0%,80%,0%))

The examples above all specify the same color.

Douglas R. Jacobson has also developed a handy quick reference RGB Color Chart (61 kB).
URLs

A URL value is given by url(foo), where foo is the URL. The URL may be optionally quoted with either single ('') or double (") quotes and may contain whitespace before or after the (optionally quoted) URL.

Parentheses, commas, spaces, single quotes, or double quotes in the URL must be escaped with a backslash. Partial URLs are interpreted relative to the style sheet source, not to the HTML source.

Note that Netscape Navigator 4.x incorrectly interprets partial URLs relative to the HTML source. Given this bug, authors may wish to use full URLs where possible.

Examples:

BODY { background: url(stripe.gif) }
BODY { background: url(http://www.htmlhelp.com/stripe.gif) }
BODY { background: url( stripe.gif ) }
BODY { background: url("stripe.gif") }
BODY { background: url(\"Ulalume\".png) } /* quotes in URL escaped */'! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'j.c 6/20/2007 09:39'!
addProperty: anObject
	
	self properties: (self properties copyWith: anObject).
	anObject parent: self.
! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 1/18/2006 23:53'!
addToRuleSet: aRuleSet

	| keys |

	keys := self selector keySelectors.

	keys do:
	[:k |
		k idSelector isNil ifFalse:
		[
			(aRuleSet idSelectors at: k idSelector value ifAbsentPut: [OrderedCollection new]) 
				add: self.			
		]
		ifTrue:
		[
			| classes |
			classes := k classSelectors.
			classes isEmpty ifFalse:
			[
				classes do:
				[:cls |
					(aRuleSet classSelectors at: cls value ifAbsentPut: [OrderedCollection new])
						add: self.
				]
			]
			ifTrue:
			[
				k element isNil ifFalse:
				[
					(aRuleSet tagSelectors at: k element asLowercase ifAbsentPut: [OrderedCollection new])
						add: self.
				]
				ifTrue:
				[
					self selector isValid ifTrue: [aRuleSet otherSelectors add: self].
				]
			]
		]
	].

	(keys isEmpty and: [self selector isValid]) ifTrue: 
	[
		aRuleSet otherSelectors add: self
	]! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 12/18/2005 00:25'!
end
	"Answer the value of end"

	^ end! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:57'!
end: anObject
	"Set the value of end"

	end := anObject! !

!CSSStyleRule methodsFor: 'initialize-release' stamp: 'tb 12/7/2005 15:47'!
initialize

	super initialize.
	attributes := Dictionary new.! !

!CSSStyleRule methodsFor: 'testing' stamp: 'tb 12/30/2005 20:04'!
isStyleRule

	^true! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 1/3/2006 04:27'!
keySelector
	"Answer the value of keySelectors"

	keySelector isNil ifTrue:
	[
		| ks |

		ks := self selector keySelectors.
		ks size = 0 ifTrue: [keySelector := self selector]
		ifFalse:
		[
			ks size = 1 
				ifTrue: [keySelector := ks first]
				ifFalse: [keySelector := CSSOrSelector new selectors: ks asSet]
		]
	].
	^ keySelector! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 1/3/2006 02:10'!
matchCount
	"Answer the value of matchCount"

	^ matchCount! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:21'!
matchCount: anObject
	"Set the value of matchCount"

	matchCount := anObject! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'spd 12/7/2010 21:53'!
matchingNodes

	^self document nodesSelect:[:ea | selector matches: ea]! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 12/20/2005 10:28'!
media
	"Answer the value of media"

	^ media! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
media: anObject
	"Set the value of media"

	media := anObject! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'SebastianSastre 1/15/2012 16:39'!
parseContents: aStream

	| pString in pos |
	aStream skipSeparators.
	self selector: (CSSSelector fromString: (aStream upTo: ${) trimBoth).
	
	pos := aStream position.
	pString := aStream upTo: $}.

	in := ReadStream on: pString.
	in skipSeparators.
	[in atEnd] whileFalse:
	[
		| property npos |
		in skipSeparators.
		npos := in position.
		(in peek = $/ and: [in next. in peek = $*]) ifTrue: "Don't keep comments in properties"
		[
			"| child |
			in back.
			child := (CSSComment new start: in position).
			self addProperty: child.
			child parseContents: in."
			in upToAll: '*/'.			
		]
		ifFalse:
		[
			property := (in upTo: $:) trimBoth.
			(property notNil and: [property size > 0])
			ifTrue:
			[
				| child |
				child := (CSSProperty propertyNamed: property) start: pos + npos.
				self addProperty: child.
				child parseContents: in.
				child end: pos + in position.
			]
		] 
	].
	self end: aStream position.! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 1/3/2006 17:36'!
printOn: aStream

	aStream nextPutAll: self prefix; nextPutAll: self innerContents; nextPutAll: self suffix.
	! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 12/21/2005 15:42'!
properties
	"Answer the value of properties"

	^properties isNil ifTrue: [#()] ifFalse:[properties]
	! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
properties: anObject
	"Set the value of properties"

	properties := anObject! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 12/7/2005 15:48'!
selector
	"Answer the value of selector"

	^ selector! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
selector: anObject
	"Set the value of selector"

	selector := anObject! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 12/18/2005 00:25'!
start
	"Answer the value of start"

	^ start! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
start: anObject
	"Set the value of start"

	start := anObject! !

!CSSStyleRule methodsFor: 'accessing' stamp: 'tb 12/21/2005 16:14'!
tag

	^'css-style'! !

CSSElement subclass: #CSSStyleSheet
	instanceVariableNames: 'imports content url'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-CSS'!

!CSSStyleSheet commentStamp: 'tb 12/21/2005 11:45' prior: 0!
I represent a collection of CSS style rules!

!CSSStyleSheet class methodsFor: 'class initialization' stamp: 'tb 1/4/2006 08:18'!
fromUrl: aStringOrUrl

	| doc url |

	url := aStringOrUrl asString.

	doc := HTTPSocket httpGetDocument: url args: nil accept: 'text/css' request: ''.
	^doc isString ifFalse: "life is good"
	[
		self new content: doc content; start: 0; end: doc content size; url: url		
	]
	ifTrue:
	[
		nil
	]

! !

!CSSStyleSheet class methodsFor: 'class initialization' stamp: 'tb 1/4/2006 08:19'!
fromUrl: aStringOrUrl parent: aDOMNode

	| doc href url |

	href := aStringOrUrl asString.
	((href beginsWith: 'http:') or: [(href beginsWith: 'https:') or:[href beginsWith: 'ftp:']]) 
	ifFalse:
	[
		url := (Url combine: aDOMNode document responseUrl asString withRelative: href).
	]
	ifTrue: [url := href].

	doc := HTTPSocket httpGetDocument: url args: nil accept: 'text/css' request: ''.
	^doc isString ifFalse: "life is good"
	[
		self new 
			content: doc content; 
			start: 0; end: doc content size; 
			parent: aDOMNode; 
			url: url;
			parseContents: (ReadStream on: doc content)		
	]
	ifTrue:
	[
		nil
	]

! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 12/17/2005 23:43'!
addImportedStyleSheet: anObject
	"Set the value of imports"

	self imports: (self imports copyWith: anObject)! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 1/3/2006 17:07'!
addImportedUrl: aUrl

	| doc ss |

	doc := aUrl asUrl retrieveContents.
	ss := CSSStyleSheet new content: doc content; start: 0; end: doc content size.
	self addImportedStyleSheet: ss.
	ss parseContents: (ReadStream on: ss content).! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 12/17/2005 23:40'!
content
	"Answer the value of content"

	^content isNil 
		ifTrue: [self document content] 
		ifFalse: [content]! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:22'!
content: anObject
	"Set the value of content"

	content := anObject! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 1/4/2006 14:37'!
docType

	^'Content-Type: text/css'! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 1/4/2006 14:43'!
endTagMissing

	^false! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 12/17/2005 23:36'!
imports
	"Answer the value of imports"

	imports isNil ifTrue:[imports := #()].
	^ imports! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:22'!
imports: anObject
	"Set the value of imports"

	imports := anObject! !

!CSSStyleSheet methodsFor: 'testing' stamp: 'tb 12/20/2005 13:33'!
isComment

	^false! !

!CSSStyleSheet methodsFor: 'testing' stamp: 'tb 1/4/2006 08:23'!
isInline

	^url isNil! !

!CSSStyleSheet methodsFor: 'testing' stamp: 'tb 12/17/2005 22:25'!
isStyleSheet

	^true! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 12/30/2005 14:50'!
media

	^(self parent attributes at: 'media' ifAbsent: ['all']) findTokens: ', '! !

!CSSStyleSheet methodsFor: 'parsing' stamp: 'tb 1/6/2006 09:43'!
parseAtRule: aStream media: anArray

	| label child s |
	label := WriteStream on: String new.
	aStream peek = $@ 
	ifFalse:[self halt]
	ifTrue:
	[
		s := aStream position.
		aStream next.
		aStream skipSeparators.
		aStream peek isLetter 
			ifTrue: [label nextPut: aStream next]
			ifFalse: [self halt].
		[aStream peek isLetter or: [aStream peek isDigit or:['-_' includes: aStream peek]]] 
			whileTrue: [label nextPut: aStream next].

		label contents asLowercase = 'media' ifTrue:
		[
			| media array |

			media := aStream upTo: ${.
			array := (media findTokens: ', ') asArray.
			self parseStyles: aStream media: array.
		]
		ifFalse:
		[
			child := (CSSAtRule named: label contents) start: s; media: anArray.		
			self addChild: child.
			child parseContents: aStream.
		]
	]! !

!CSSStyleSheet methodsFor: 'parsing' stamp: 'tb 1/3/2006 17:04'!
parseContents: aStream

	| in  |
	
	aStream upToAll: '<!!--'.
	aStream position >= self end 
		ifTrue: 
		[
			in := ReadStream on: (aStream reset next: self end).
			in position: self start.
		]
		ifFalse: 
		[
			| styleEnd |
			aStream position: self end-1. 
			[(aStream position > start) and: [aStream peek ~= $>]] whileTrue: [aStream back].
			((aStream position > start) and: [aStream peek = $>]) ifTrue: [aStream back].
			[(aStream position > start) and: [aStream peek = $-]] whileTrue: [aStream back].
			styleEnd := aStream position.
			in := ReadStream on: (aStream reset next: styleEnd).
			in position: self start.
			in upToAll: '<!!--'.			
		].
	
	in skipSeparators. 
	self parseStyles: in media: self media.
	! !

!CSSStyleSheet methodsFor: 'parsing' stamp: 'tb 1/3/2006 17:07'!
parseStyles: aStream media: anArray

	aStream skipSeparators.
	[aStream atEnd] whileFalse: 
	[
		aStream skipSeparators.
		aStream peek = $} ifTrue: [aStream next. ^self]. "End of media block"
		(aStream peek = $/ and: [aStream next. aStream peek = $*]) 
		ifTrue: 
		[
			| child |
			aStream back. 
			child := CSSComment new start: aStream position.
			self addChild: child.
			child parseContents: aStream.
		]
		ifFalse:
		[	
			(aStream peek = $@) 
			ifTrue: "Read the imports"
			[
				self parseAtRule: aStream media: anArray.
			]
			ifFalse:
			[
				aStream atEnd ifFalse:
				[
					| child |

					child := CSSStyleRule new start: aStream position; media: anArray.
					self addChild: child.
					child parseContents: aStream.
				]
			]
		]
	]! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 12/20/2005 13:17'!
tag

	^'css-stylesheet'! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'tb 1/4/2006 08:22'!
url
	"Answer the value of url"

	^ url isNil ifTrue: [self document url] ifFalse: [url]! !

!CSSStyleSheet methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:23'!
url: anObject
	"Set the value of url"

	url := anObject! !

HtmlParseNode subclass: #HtmlDOMNode
	instanceVariableNames: 'attributes hasTrailingSlash tag isImplied dtd endTagMissing styles id classes'
	classVariableNames: 'DefaultRules HTML401'
	poolDictionaries: ''
	category: 'HTML-DOM'!

HtmlDOMNode subclass: #HtmlABBRNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlABBRNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'abbr'! !

HtmlDOMNode subclass: #HtmlACRONYMNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlACRONYMNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'acronym'! !

HtmlDOMNode subclass: #HtmlADDRESSNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlADDRESSNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'address'! !

HtmlDOMNode subclass: #HtmlANode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlANode methodsFor: 'as yet unclassified' stamp: 'gk 11/13/2007 14:39'!
href
	^self attributes at: 'href' ifAbsent: ''! !

!HtmlANode methodsFor: 'accessing' stamp: ''!
tag
	^ 'a'! !

HtmlDOMNode subclass: #HtmlAPPLETNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlAPPLETNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'applet'! !

HtmlDOMNode subclass: #HtmlAREANode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlAREANode methodsFor: 'accessing' stamp: ''!
tag
	^ 'area'! !

HtmlDOMNode subclass: #HtmlBASEFONTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBASEFONTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'basefont'! !

HtmlDOMNode subclass: #HtmlBASENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBASENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'base'! !

HtmlDOMNode subclass: #HtmlBDONode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBDONode methodsFor: 'accessing' stamp: ''!
tag
	^ 'bdo'! !

HtmlDOMNode subclass: #HtmlBIGNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBIGNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'big'! !

HtmlDOMNode subclass: #HtmlBLOCKQUOTENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBLOCKQUOTENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'blockquote'! !

HtmlDOMNode subclass: #HtmlBNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'b'! !

HtmlDOMNode subclass: #HtmlBODYNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBODYNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'body'! !

HtmlDOMNode subclass: #HtmlBRNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBRNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'br'! !

HtmlDOMNode subclass: #HtmlBUTTONNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlBUTTONNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'button'! !

HtmlDOMNode subclass: #HtmlCAPTIONNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlCAPTIONNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'caption'! !

HtmlDOMNode subclass: #HtmlCDATANode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlCDATANode class methodsFor: 'as yet unclassified' stamp: 'tb 12/8/2005 11:58'!
start: s end: e

	^self new start: s; end: e! !

!HtmlCDATANode methodsFor: 'accessing' stamp: 'tb 1/3/2006 17:36'!
asString

	^self innerContents

! !

!HtmlCDATANode methodsFor: 'accessing' stamp: 'tb 3/1/2006 15:36'!
checkForUnescapedCharacters

	| str |

	self parent isComment ifTrue: [^self].
	str := self innerContents.

	str withIndexDo: 	
	[:ea :idx |
		(HtmlUnescapedSpecialCharacterError specialCharacters includes: ea) ifTrue: 
		[
		
			ea = $& ifFalse:
			[
				self addError: 
					(HtmlUnescapedSpecialCharacterError new position: self start + idx)
			]
			ifTrue:
			[
				(str size > idx and: [(str at: idx+1) = $#]) ifFalse:
				[
					| strm s |

					strm := ReadStream on: str.
					strm position: idx.
					s := strm upTo: $;.
					((s includes: $ ) or: [s size > 6 or:[s size < 3]]) ifTrue:
					[
						(s = 'lt' or: [s = 'gt']) ifFalse:
						[
							self addError: 
							(HtmlUnescapedSpecialCharacterError new position: self start + idx)						
						]
					]					
				]
			]
		]	
	]! !

!HtmlCDATANode methodsFor: 'accessing' stamp: 'j.c 6/6/2007 14:02'!
initialize

	super initialize.
	endTagMissing := false.
	! !

!HtmlCDATANode methodsFor: 'accessing' stamp: 'tb 1/4/2006 15:05'!
innerContents

	^self rawContent! !

!HtmlCDATANode methodsFor: 'accessing' stamp: 'tb 12/2/2005 09:13'!
isCDATA

	^true! !

!HtmlCDATANode methodsFor: 'accessing' stamp: 'tb 12/8/2005 12:02'!
prefix

	^''! !

!HtmlCDATANode methodsFor: 'accessing' stamp: 'tb 12/8/2005 12:02'!
suffix

	^''! !

!HtmlCDATANode methodsFor: 'accessing' stamp: ''!
tag
	^ 'cdata'! !

HtmlDOMNode subclass: #HtmlCENTERNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlCENTERNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'center'! !

HtmlDOMNode subclass: #HtmlCITENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlCITENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'cite'! !

HtmlDOMNode subclass: #HtmlCODENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlCODENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'code'! !

HtmlDOMNode subclass: #HtmlCOLGROUPNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlCOLGROUPNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'colgroup'! !

HtmlDOMNode subclass: #HtmlCOLNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlCOLNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'col'! !

HtmlDOMNode subclass: #HtmlCommentNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlCommentNode methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 09:12'!
isComment

	^true! !

!HtmlCommentNode methodsFor: 'as yet unclassified' stamp: 'tb 1/14/2006 13:57'!
parseContents: aStream

	| c s allowBrackets |

	allowBrackets := false.
	aStream upTo: $!!.
	s := aStream position.
	[aStream atEnd] whileFalse:
	[
		((c := aStream next) = $- and: [aStream peek = $-]) ifTrue:
		[
			aStream next.
			allowBrackets := allowBrackets not.			
		]
		ifFalse:
		[
			(c = $> and: [allowBrackets not]) ifTrue:
			[
				self end: aStream position.
				self addChild: (HtmlCDATANode start: s end: aStream position -1).
				^self
			]
			ifFalse:
			[
				(c = $- and: [aStream peek = $>]) ifTrue:
				[
					self addChild: (HtmlCDATANode start: s end: aStream position).
					aStream next.
					self end: aStream position.
					self endTagMissing: allowBrackets.
					^self
				]	
			]
		]
	].
	self endTagMissing: true.
	self end: aStream position.! !

!HtmlCommentNode methodsFor: 'as yet unclassified' stamp: 'tb 11/29/2005 14:55'!
tag

	^'comment'! !

HtmlCommentNode subclass: #HtmlXMLCDataNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlXMLCDataNode methodsFor: 'as yet unclassified' stamp: 'tb 3/12/2006 15:08'!
parseContents: aStream

	| cdStart out |

	out := WriteStream on: String new.
	aStream skipSeparators.
	('<!![CDATA[' size) timesRepeat: [out nextPut: aStream next].

	(out contents beginsWith: '<!![CDATA[') 
		ifTrue: [cdStart := aStream position]
		ifFalse: [self halt]. "I musta messed up something here"

	[(out contents endsWith: ']]>') or: [aStream atEnd]] whileFalse: [out nextPut: aStream next].

	(out contents endsWith: ']]>') ifFalse: 
	[
		aStream position: cdStart.
		aStream upTo: $>.
		self addChild: (HtmlCDATANode start: cdStart end: aStream position -1).
		self endTagMissing: true.
	]
	ifTrue:
	[
		self end: aStream position.
		self addChild: (HtmlCDATANode start: cdStart end: aStream position - ( ']]>' size)).		
		self endTagMissing: false.
	]
! !

HtmlDOMNode subclass: #HtmlDDNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlDDNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'dd'! !

HtmlDOMNode subclass: #HtmlDELNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlDELNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'del'! !

HtmlDOMNode subclass: #HtmlDFNNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlDFNNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'dfn'! !

HtmlDOMNode subclass: #HtmlDIRNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlDIRNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'dir'! !

HtmlDOMNode subclass: #HtmlDIVNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlDIVNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'div'! !

HtmlDOMNode subclass: #HtmlDLNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlDLNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'dl'! !

!HtmlDOMNode class methodsFor: 'as yet unclassified' stamp: 'sebastian.sastre 9/5/2010 14:25'!
fromStream: aStream

	| s doctype html |

	doctype := nil.
	aStream skipSeparators.
	s := aStream upTo: $<.
	s := aStream upTo: $>.
	s size = 0 ifTrue: [aStream reset. ^nil].
	(s asUppercase beginsWith: '!!DOCTYPE') ifTrue:
	[
		doctype := s.		
		s := aStream upTo: $<.
		s := aStream upTo: $>.
	].
	
	html := (self tagContents: s) start: (aStream position - s size - 2); parseContents: aStream.			
	^html docType: doctype.! !

!HtmlDOMNode class methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:40'!
newForTag: aTag 

	| node className cls tag |

	tag := aTag trimBoth.
	node := nil.

	className := ('Html',tag asUppercase,'Node') asSymbol.

	cls := Smalltalk at: className ifAbsent: [HtmlDOMNode].

	node := cls new tag: tag;
			hasTrailingSlash: false; 
			attributes: Dictionary new.

	tag asLowercase = tag ifFalse: 
	[
		node addError: (HtmlTagNotLowerCaseError new tag: tag).
	].

	^node.	
! !

!HtmlDOMNode class methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:41'!
tagContents: aTag 

	| node className cls tag in out trimmedTag endsWithSlash |

	trimmedTag := aTag trimBoth.
	node := nil.
	in := ReadStream on: aTag.
	out := WriteStream on: (String new:20).

	in skipSeparators.
	[in atEnd not and: [in peek isAlphaNumeric]] whileTrue: [out nextPut: in next].

	tag := out contents.

	className := ('Html',tag asUppercase,'Node') asSymbol.

	cls := Smalltalk at: className ifAbsent: [HtmlDOMNode].

	endsWithSlash := (trimmedTag last = $/) ifTrue:
	[
		(trimmedTag detect:[:ea | ea isSeparator] ifNone: [nil]) isNil or:
			[(trimmedTag at: ((trimmedTag size -1))) isSeparator]
	]
	ifFalse: [false ].

	node := cls new tag: tag;
			hasTrailingSlash: endsWithSlash; 
			parseAttributes: (trimmedTag last: (trimmedTag size - tag size)).

	tag asLowercase = tag ifFalse: 
	[
		node addError: (HtmlTagNotLowerCaseError new tag: tag).
	].

	^node.	
! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'j.c 6/14/2007 12:24'!
addStyle: aStyleRule
	self styles: (self styles copyWith: aStyleRule)! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/19/2007 11:52'!
anchors
	"Find children tags of given class."

	^self tagsOfClass: HtmlANode! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 11/30/2005 12:26'!
attributes
	"Answer the value of attributes"

	^ attributes ifNil: [attributes:=Dictionary new]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 7/12/2007 11:04'!
attributes: anObject
	"Set the value of attributes"

	| cstr |

	attributes := anObject.
	id := (attributes at: 'id' ifAbsent: [nil]).
	cstr := attributes at: 'class' ifAbsent: [nil].
	classes := cstr isNil ifTrue: [#()] ifFalse: [cstr findTokens: ' '].! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/1/2005 09:11'!
childDTDTags

	^((self children collect: [:ea | ea dtdTag]) reject:[:ea | ea = 'COMMENT']) asArray! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/1/2005 09:12'!
childTags

	^((self children collect: [:ea | ea tag]) reject:[:ea | ea = 'comment']) asArray! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 1/4/2006 13:52'!
classes
	
	^classes isNil ifTrue: [#()] ifFalse: [classes]! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 09:30'!
closeTagIsOptional

	^self dtd tagCloseIsOptional: self dtdTag! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/3/2005 16:14'!
closeTagMissing
	

	^self endTagMissing! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/20/2005 10:21'!
content

	^self parent content! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 11/30/2005 12:39'!
createImpliedNodes

	self children do:[:ea | ea createImpliedNodes]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/2/2005 09:43'!
dtd

	dtd isNil ifTrue: [self parent notNil ifTrue: [dtd := self parent dtd]].
	dtd isNil ifTrue: [self halt].
	^dtd! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/1/2005 10:39'!
dtd: aDTD

	dtd := aDTD! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/5/2005 12:07'!
dtdTag

	^self dtd isXML ifFalse: [self tag asUppercase] ifTrue: [self tag]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/5/2005 12:02'!
endTagMissing
	"Answer the value of endTagMissing"

	^self hasTrailingSlash 
		ifTrue:[false] 
		ifFalse:
		[
			endTagMissing isNil ifTrue: [endTagMissing := false].
			endTagMissing
		]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:53'!
endTagMissing: anObject
	"Set the value of endTagMissing"

	endTagMissing := anObject! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 17:29'!
hasTrailingSlash

	^hasTrailingSlash isNil ifTrue: [false] ifFalse: [hasTrailingSlash] ! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:53'!
hasTrailingSlash: anObject
	"Set the value of hasTrailingSlash"

	hasTrailingSlash := anObject! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 7/12/2007 11:07'!
id

	^id! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 09:13'!
isCDATA

	^false! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 12/19/2005 23:03'!
isCSS

	^false! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 09:12'!
isComment

	^false! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/12/2007 12:35'!
isDOMNode

	^true! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 12:26'!
isDocument

	^false! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 11/30/2005 12:21'!
isImplied
	
	^isImplied isNil ifTrue: [^false] ifFalse: [isImplied]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 11/30/2005 12:21'!
isImplied: trueOrFalse
	
	isImplied := trueOrFalse! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/2/2005 09:34'!
isSelfContained

	^self hasTrailingSlash or: [self dtd tagIsEmpty: self dtdTag]! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 12/17/2005 22:26'!
isStyleSheet

	^false! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 1/8/2006 10:03'!
loader

	^self document loader! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 12/5/2005 22:08'!
mayContainTag: aTag

	^self isComment 
		ifTrue:[false] 
		ifFalse:[self dtd tag: self dtdTag canContainTag: aTag]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'gk 11/19/2007 23:24'!
name
	^self attributes at: 'name' ifAbsent: ['']! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:40'!
parseAttributes: attrString

	| attrs in s |
	attrs := Dictionary new. 

	s := attrString trimBoth.
	(s endsWith: '/') ifTrue: [ s:= s first: (s size-1) ].

	in := ReadStream on: s trimBoth.

	[in atEnd] whileFalse:
	[
		| k v pos |
		in skipSeparators.
		pos := in position.
		v := nil.
		k := in upTo: $=.
		k := k trimBoth.
		(k includes: $ ) ifTrue: "Looks like there is no value. "
		[			
			in position: pos. 
			k:= in upTo: $ . 
		]
		ifFalse:
		[
			in skipSeparators.
			in atEnd ifFalse:
			[
				(in next = $") ifTrue: [v := in upTo: $"]
				ifFalse:
				[
					in back.
					(in next = $') ifTrue: [v := in upTo: $']
					ifFalse:
					[
						in back.
						v := in upTo: $ .
						self addError: (HtmlAttributeValueNotQuotedError new attribute: k).
					]
				]		 
			]
		].

		attrs at: k ifPresent:
		[:val | 
			self addError: (HtmlAttributeMultiplySpecifiedError new 
				attribute: k; value: v value: val).
		].
		attrs at: k put: v.
	].
	self attributes: attrs.! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'SebastianSastre 1/15/2012 16:40'!
parseContents: aStream

	"tagParent: parent contents: aString stream: aStream"

	| s |

	self isSelfContained ifTrue: [^self end: aStream position; endTagMissing: true].

	[self end isNil and:[aStream atEnd not]] whileTrue:
	[
		aStream skipSeparators.

		s := aStream upTo: $<.
		s size > 0 ifTrue: 
		[
			self addChild: (HtmlCDATANode start: self startForNextElement 
								end: (aStream atEnd ifFalse:[aStream position-1] ifTrue: [aStream position]))

		].

		s := aStream upTo: $>.	

		(s size > 0 and:[s first = $!! or: [(s first = $?) & (s last = $?)]]) ifTrue: 
		[
			s first = $? ifTrue:
			[
				"blow off the processing directive"
				"aStream skip: (s size+2)negated. 
				self addChild: (HtmlDOMNode parseProcessingDirective: aStream)."
			]
			ifFalse:
			[
				| child |
				aStream skip: (s size+2)negated. 
				(self dtd isXML and: [s beginsWith: '<!![CDATA[']) ifFalse:
				[
					child := (HtmlCommentNode new) start: aStream position.
				]
				ifTrue:
				[
					child := (HtmlXMLCDataNode new) start: aStream position.
				].
				self addChild: child.
				child parseContents: aStream.
			]
		]
		ifFalse: "It is a tag"
		[
			(s includes: $<) ifTrue: [s := self recoverFromLTIn: s stream: aStream]. 			
			(s trimBoth beginsWith: '/') ifTrue: 
			[
				| closeTag |
				closeTag := (s findTokens: ('/',Character separators)) first asLowercase.
				closeTag = self tag ifTrue: 
				[
					self end: aStream position; endTagMissing: false.
					^self
				]
				ifFalse:
				[
					(self hasAncestor: closeTag) ifTrue:
					[
						self addError: (HtmlTagMissingEndError new).
						aStream skip: (s size + 2) negated.
						^self end: (aStream position); endTagMissing: true.
					]
					ifFalse:
					[
						| child |
						child := (HtmlDOMNode newForTag: s) 
							start: (aStream position - (s size) - 2); end: (aStream position).
							self addChild: child.
							child addError: (HtmlTagMissingStartError new 
								position: (aStream position - (s size) - 1); 
								tag: s).
					]
				]
			]
			ifFalse: "Nested Tag Found"
			[
				"Does the new tag imply termination?"
				| short |
				short := (s findTokens: ('/',Character separators)).
				short := short size > 0 ifTrue: [short first] ifFalse: [short].
				"The above hack and aStream atEnd guards against input underflow"
				(aStream atEnd or:[((self mayContainTag: short) not 
					and: [self closeTagIsOptional])
					and:
					[
						| likelyParent |
						likelyParent := self detectAncestor:[:p | p mayContainTag: short]. 
						(likelyParent level > 0 and: [(self level - likelyParent level) < 3])
							or: [short = self tag] "siblings"
					]
				])
				ifTrue:
				"Fake a termination"
				[
					self end: (aStream atEnd ifFalse: [aStream position - s size - 2] ifTrue: [aStream position]); 
						endTagMissing: true.
					aStream atEnd ifFalse: [aStream skip: (s size + 2) negated].
					^self
				]
				ifFalse:
				[
					(short size > 0 and: [short first isLetter]) ifTrue:
					[
						| child |
						child := (HtmlDOMNode tagContents: s) start: (aStream position - s size - 2).
						self addChild: child.
						child parseContents: aStream.
					]
					ifFalse:
					[
						self addChild: (HtmlCDATANode start: self startForNextElement 
								end: aStream position)
					]
				]
			]
		]
	].
	(aStream atEnd and:[end isNil]) ifTrue: 
	[
		self end: aStream position.
		self endTagMissing: true.
	].
! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/2/2005 20:33'!
rawTag

	^tag! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'tb 12/10/2005 21:41'!
recoverFromLTIn: aString stream: in

	| s  |

	in skip: (aString size - (aString lastIndexOf: $<)) negated-1.

	((self children isEmpty not) and: [self children last isCDATA]) ifTrue:
	[
		self children last end: in position -1.
	]
	ifFalse: 
	[
		self addChild:
				(HtmlCDATANode start: self startForNextElement
								end: in position-1)		
	].
	s := in upTo: $>.
	^s! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/30/2005 20:08'!
styles

	^styles isNil ifTrue: [#()] ifFalse: [styles]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/30/2005 20:08'!
styles: anArray

	styles := anArray! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/2/2005 10:47'!
tag
	"Answer the value of tag"

	^ tag isNil ifFalse: [tag asLowercase] ifTrue: [tag]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/2/2005 20:31'!
tag: anObject
	"Set the value of tag"

	tag := anObject.! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/19/2007 12:44'!
tagOfClass: aClass
	"Find children tags of given class."

	self nodesDo: [:ea | ea class = aClass ifTrue: [^ea]].
	^nil! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/19/2007 12:03'!
tagWithClass: aClass
	"Find children tags of given class."

	self nodesDo: [:ea | (ea isDOMNode and: [ea classes includes: aClass]) ifTrue: [^ea]].
	^nil! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/12/2007 23:35'!
tagWithId: anId
	"Find the first tag with given id."

	self nodesDo: [:ea | (ea isDOMNode and: [ea id = anId]) ifTrue: [^ea]].
	^nil! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/12/2007 23:35'!
tagWithId: anId class: aClass
	"Find the first tag with given id and class."

	self nodesDo: [:ea |
		(ea class = aClass and: [ea isDOMNode and: [ea id = anId]]) ifTrue: [^ea]].
	^nil! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/19/2007 21:54'!
tagWithName: aName
	"Find the first tag with given name."

	self nodesDo: [:ea | (ea isDOMNode and: [ea name = aName]) ifTrue: [^ea]].
	^nil! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/12/2007 12:39'!
tagsNamed: aName
	"Find all children tags with given name."

	^self nodesCollect: [:ea | ea tag = aName]! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/19/2007 11:51'!
tagsOfClass: aClass
	"Find children tags of given class."

	^self nodesCollect: [:ea | ea class = aClass]! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/19/2007 12:03'!
tagsWithClass: aClass
	"Find children tags of given class."

	^self nodesCollect: [:ea | ea isDOMNode and: [ea classes includes: aClass]]! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 11/12/2007 12:39'!
tagsWithId: anId
	"Find all children tags with given id."

	^self nodesCollect:[:ea | ea isDOMNode and: [ea id = anId]]! !

!HtmlDOMNode methodsFor: 'as yet unclassified' stamp: 'gk 7/3/2009 22:39'!
tagsWithIdBegins: aString
	"Find all children tags with given id prefix."

	^self nodesCollect:[:ea | ea isDOMNode and: [ea id notNil and: [ea id beginsWith: aString]]]! !

!HtmlDOMNode methodsFor: 'accessing' stamp: 'tb 12/3/2005 13:52'!
uniqueId

	^(self parent uniqueId, '.', ((self parent children indexOf: self) asString))! !

HtmlDOMNode subclass: #HtmlDTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlDTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'dt'! !

HtmlDOMNode subclass: #HtmlDocument
	instanceVariableNames: 'docType content requestUrl responseUrl dtdIsImplied loader'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlDocument methodsFor: 'parsing' stamp: 'tb 12/30/2005 20:49'!
allStyles

	| rules |

	rules := OrderedCollection new.
	self stylesheets do: [:sheet | rules addAll: (sheet children select:[:ea | ea isStyleRule])].
	^rules asArray! !

!HtmlDocument methodsFor: 'parsing' stamp: 'tb 7/12/2007 12:55'!
applyStyles

	"TODO: Optimize this based on information described at " 	"http://weblogs.mozillazine.org/hyatt/archives/2005_05.html#007507"

	| rules set |

	rules := self allStyles.
	rules do:[:ea | ea matchCount: 0].
	rules := rules reject:[:ea | ea selector isValid not].
	set := CSSRuleSet new.
	rules do: [:ea | ea addToRuleSet: set].

	self nodesDo:
	[:ea | 
		ea isCSS 
			ifFalse: 
			[
				(set candidateRulesForNode: ea) do: 
				[:rule | 
					(rule selector matches: ea) 
						ifTrue: [ea addStyle: rule. rule matchCount: rule matchCount + 1]
				].
				(ea attributes includesKey: 'style') ifTrue: 
				[
					|  styleRule style |
					style := (ea attributes at: 'style').
					styleRule := (CSSNodeAttributeStyleRule new content: style).
					ea addStyle: styleRule.
				].
			]
	]. ! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 12/2/2005 20:38'!
content
	"Answer the value of content"

	^ content! !

!HtmlDocument methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:31'!
content: anObject
	"Set the value of content"

	content := anObject! !

!HtmlDocument methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 00:24'!
createImpliedNodes

	(self children detect: [:ea | ea tag = 'html'] ifNone:[nil]) ifNil: 
	[self children: 
		(Array with: 
			(HtmlHTMLNode new children: self children; 
				start: self start; 
				end: self end; 
				isImplied: true))
	].
	super createImpliedNodes.! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 12/1/2005 10:45'!
docType
	"Answer the value of docType"

	docType isNil ifTrue: [docType := Html401Strict docType].
	^docType! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 12/3/2005 14:11'!
docType: anObject
	"Set the value of docType"

	docType := anObject.
	self dtd: (HtmlDTD forDocType: anObject).! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 12/2/2005 20:35'!
document

	^self! !

!HtmlDocument methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 09:39'!
dtd

	^dtd isNil 
		ifTrue: [dtd:=HtmlDTD forDocType: self docType] 
		ifFalse: [dtd]! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 1/4/2006 10:47'!
dtdIsImplied
	"Answer the value of dtdIsImplied"

	^ dtdIsImplied! !

!HtmlDocument methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:53'!
dtdIsImplied: anObject
	"Set the value of dtdIsImplied"

	dtdIsImplied := anObject! !

!HtmlDocument methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 12:27'!
isDocument

	^true! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 1/10/2006 08:50'!
loader
	"Answer the value of loader"

	^ loader isNil ifTrue: [loader := HtmlCache new] ifFalse: [loader]! !

!HtmlDocument methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:51'!
loader: anObject
	"Set the value of loader"

	loader := anObject! !

!HtmlDocument methodsFor: 'as yet unclassified' stamp: 'tb 12/2/2005 11:02'!
mayContainTag: aTag

	^true! !

!HtmlDocument methodsFor: 'parsing' stamp: 'SebastianSastre 1/15/2012 16:35'!
parseContents: aStream

	"tagParent: parent contents: aString stream: aStream"

	| s docTypeFound hasXMLDirective |

	docTypeFound := false.
	hasXMLDirective := false.

	self start: aStream position.

	content := aStream contents.

	[aStream atEnd not] whileTrue:
	[
		| p |
		p := aStream position.
		
		"If we skip separatos, we will be neglecting spaces WTF?!!
		aStream skipSeparators."
		
		s := aStream upTo: $<.
		s size > 0 ifTrue: 
		[
			| pos |
			pos := aStream atEnd ifTrue: [aStream position] ifFalse: [aStream position-1].
			self addChild: 
				(HtmlCDATANode start: p 
								end: (pos max: p))
		].
		s := aStream upTo: $>.	

		(s size > 0 and:[s first = $!! or: [(s first = $?) & (s last = $?)]]) ifTrue: 
		[
			s first = $? ifTrue:
			[
				hasXMLDirective := true.
				self parseProcessingDirective: s
			]
			ifFalse:
			[
				(s asUppercase beginsWith: '!!DOCTYPE') ifTrue:
				[
					self docType: s.
					docTypeFound := true.
				]
				ifFalse:
				[
					| child |
					aStream skip: (s size+2)negated. 
					child := (HtmlCommentNode new) start: aStream position.
					self addChild: child.
					child parseContents: aStream.
				]
			]
		]
		ifFalse: "It is a tag"
		[
			(docType isNil and: [hasXMLDirective]) ifTrue:
			[
				self docType: XHtml1Transitional docType.
			]
			ifFalse:
			[
				docType isNil ifTrue: [self docType: Html401Strict docType].
			].
			s := s trimBoth.
			(s size > 0) ifTrue:
			[
				(s first = $/) ifTrue: "stray close tag I think"
				[
					| child |
					child := (HtmlDOMNode newForTag: s) 
							start: (aStream position - (s size) - 2); end: (aStream position).
					self addChild: child.
					child addError: (HtmlTagMissingStartError new 
							position: (aStream position - (s size) - 1); 
							tag: s).
				]
				ifFalse: "Nested Tag Found"
				[
					| child |
					child := (HtmlDOMNode tagContents: s) start: (aStream position - s size - 2).
					self addChild: child.
					child parseContents: aStream.
				]
			]
		]
	].
	self end: aStream position.
	"self createImpliedNodes."
	self dtdIsImplied: docTypeFound not.
	docTypeFound ifFalse: 
	[
		self dtdIsImplied: true.
		self addError: HtmlMissingDocTypeError new
	].
	
! !

!HtmlDocument methodsFor: 'parsing' stamp: 'tb 12/11/2005 21:10'!
parseProcessingDirective: aString

	| in |
	in := ReadStream on: aString.
	in upTo: $?.
	in skipSeparators.
	[Character separators includes: in peek] whileFalse: [in next].
	in skipSeparators.
	self parseAttributes: (in upTo: $?).	
! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 12/17/2005 22:54'!
requestUrl
	"Answer the value of requestUrl"

	^ requestUrl! !

!HtmlDocument methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:51'!
requestUrl: anObject
	"Set the value of requestUrl"

	requestUrl := anObject! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 12/17/2005 22:54'!
responseUrl
	"Answer the value of responseUrl"

	^ responseUrl! !

!HtmlDocument methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
responseUrl: anObject
	"Set the value of responseUrl"

	responseUrl := anObject! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 12/30/2005 17:37'!
stylesheets

	| stylesheets |

	stylesheets := OrderedCollection new.

	self nodesDo:
	[:ea |
		((ea respondsTo: #stylesheet)and:[ea stylesheet notNil]) 
			ifTrue: [stylesheets add: ea stylesheet].
	].
	^stylesheets
	! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 11/30/2005 11:52'!
tag

	^'document'! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 12/3/2005 13:52'!
uniqueId

	^'D'! !

!HtmlDocument methodsFor: 'accessing' stamp: 'tb 1/4/2006 14:35'!
url

	^self responseUrl! !

HtmlDOMNode subclass: #HtmlEMNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlEMNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'em'! !

HtmlDOMNode subclass: #HtmlFIELDSETNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlFIELDSETNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'fieldset'! !

HtmlDOMNode subclass: #HtmlFONTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlFONTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'font'! !

HtmlDOMNode subclass: #HtmlFORMNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlFORMNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'form'! !

HtmlDOMNode subclass: #HtmlFRAMENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlFRAMENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'frame'! !

HtmlDOMNode subclass: #HtmlFRAMESETNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlFRAMESETNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'frameset'! !

HtmlDOMNode subclass: #HtmlH1Node
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlH1Node methodsFor: 'accessing' stamp: ''!
tag
	^ 'h1'! !

HtmlDOMNode subclass: #HtmlH2Node
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlH2Node methodsFor: 'accessing' stamp: ''!
tag
	^ 'h2'! !

HtmlDOMNode subclass: #HtmlH3Node
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlH3Node methodsFor: 'accessing' stamp: ''!
tag
	^ 'h3'! !

HtmlDOMNode subclass: #HtmlH4Node
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlH4Node methodsFor: 'accessing' stamp: ''!
tag
	^ 'h4'! !

HtmlDOMNode subclass: #HtmlH5Node
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlH5Node methodsFor: 'accessing' stamp: ''!
tag
	^ 'h5'! !

HtmlDOMNode subclass: #HtmlH6Node
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlH6Node methodsFor: 'accessing' stamp: ''!
tag
	^ 'h6'! !

HtmlDOMNode subclass: #HtmlHEADNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlHEADNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'head'! !

HtmlDOMNode subclass: #HtmlHRNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlHRNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'hr'! !

HtmlDOMNode subclass: #HtmlHTMLNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlHTMLNode methodsFor: 'as yet unclassified' stamp: 'tb 12/1/2005 09:12'!
createImpliedNodes

	| dtags |
	dtags := self childDTDTags.
	((dtags includes: 'HEAD') and: [dtags includes: 'BODY'])
	ifFalse: "one or both is missing"
	[
		(dtags includes: 'HEAD') ifTrue: "everything after head goes in body"
		[
			| head bodyParts |
			head := self children detect:[:ea | ea tag = 'head'].
			bodyParts := self children copyAfter: head.
			self children: (self children copyWithoutAll: bodyParts).
			self addChild: (HtmlBODYNode new children: bodyParts; 
				start: head end; 
				end: self end; 
				isImplied: true).
		]
		ifFalse: "No head - is there a body?"
		[
			(dtags includes: 'BODY') ifTrue:
			[
				| headParts body |
				body := self children detect:[:ea | ea tag = 'body'].
				headParts := self children copyUpTo: body.
				self children: (self children copyWithoutAll: headParts). 
				self children:
					(self children copyWithFirst:
						(HtmlHEADNode new children: headParts;
							start: self start;
							end: body start;
							isImplied: true)).
			]
			ifFalse: "Neither are found"
			[
				| kids |
				kids := self children.
				self children: #().
				self addChild: (HtmlHEADNode new 
							start: self start;
							end: self start;
							isImplied: true).
				self addChild: (HtmlBODYNode new children: kids; 
				start: self start; 
				end: self end; 
				isImplied: true).
			]
		]
	]! !

!HtmlHTMLNode methodsFor: 'accessing' stamp: 'tb 11/29/2005 09:19'!
tag

	^'html'! !

HtmlDOMNode subclass: #HtmlIFRAMENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlIFRAMENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'iframe'! !

HtmlDOMNode subclass: #HtmlIMGNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlIMGNode methodsFor: 'accessing' stamp: 'CAs 6/25/2012 10:16'!
src
	^self attributes at: 'src' ifAbsent: ''! !

!HtmlIMGNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'img'! !

HtmlDOMNode subclass: #HtmlINPUTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlINPUTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'input'! !

HtmlDOMNode subclass: #HtmlINSNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlINSNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'ins'! !

HtmlDOMNode subclass: #HtmlINode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlINode methodsFor: 'accessing' stamp: ''!
tag
	^ 'i'! !

HtmlDOMNode subclass: #HtmlISINDEXNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlISINDEXNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'isindex'! !

HtmlDOMNode subclass: #HtmlKBDNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlKBDNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'kbd'! !

HtmlDOMNode subclass: #HtmlLABELNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlLABELNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'label'! !

HtmlDOMNode subclass: #HtmlLEGENDNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlLEGENDNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'legend'! !

HtmlDOMNode subclass: #HtmlLINKNode
	instanceVariableNames: 'stylesheet href'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlLINKNode commentStamp: '<historical>' prior: 0!
<LINK rel="stylesheet" type="text/css" name="anyname" href="url">!

!HtmlLINKNode methodsFor: 'accessing' stamp: 'tb 1/10/2006 15:01'!
href
	"Answer the value of href"

	^ href! !

!HtmlLINKNode methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:32'!
href: anObject
	"Set the value of href"

	href := anObject! !

!HtmlLINKNode methodsFor: 'parsing' stamp: 'tb 1/10/2006 19:53'!
parseContents: aStream

	| rel |

	rel := self attributes keys detect: [:ea | ea asLowercase = 'rel'] ifNone: [nil].
	(rel notNil and: [((self attributes at: rel) asLowercase findTokens: ' ') includes: 'stylesheet']) 
	ifTrue: 
	[
		href := self attributes keys detect: [:ea | ea asLowercase = 'href'] ifNone: [nil].
		href isNil ifFalse: 
		[
			href := self attributes at: href.
			(href notNil and:[href size > 0]) ifTrue:
			[
				self loader queueUrl: href. "Start download in another thread"
			]
			ifFalse: [href := nil].
		]
	].
	super parseContents: aStream! !

!HtmlLINKNode methodsFor: 'accessing' stamp: 'tb 1/17/2008 15:01'!
stylesheet
	"Answer the value of stylesheet"

	| |

	(stylesheet isNil and: [href notNil])
	ifTrue:
	[
		| doc |
		doc := self loader documentForUrl: self href.
		stylesheet := CSSStyleSheet new parent: self; 
			start: 0; 
			content: doc content; 
			end: doc content size;
 			url: self href; 
			parseContents: (ReadStream on: doc content)
	].	
	^ stylesheet! !

!HtmlLINKNode methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:32'!
stylesheet: anObject
	"Set the value of stylesheet"

	stylesheet := anObject! !

!HtmlLINKNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'link'! !

HtmlDOMNode subclass: #HtmlLINode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlLINode methodsFor: 'accessing' stamp: ''!
tag
	^ 'li'! !

HtmlDOMNode subclass: #HtmlMAPNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlMAPNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'map'! !

HtmlDOMNode subclass: #HtmlMENUNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlMENUNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'menu'! !

HtmlDOMNode subclass: #HtmlMETANode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlMETANode methodsFor: 'accessing' stamp: ''!
tag
	^ 'meta'! !

HtmlDOMNode subclass: #HtmlNOBRNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlNOBRNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'nobr'! !

HtmlDOMNode subclass: #HtmlNOFRAMESNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlNOFRAMESNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'noframes'! !

HtmlDOMNode subclass: #HtmlNOSCRIPTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlNOSCRIPTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'noscript'! !

HtmlDOMNode subclass: #HtmlOBJECTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlOBJECTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'object'! !

HtmlDOMNode subclass: #HtmlOLNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlOLNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'ol'! !

HtmlDOMNode subclass: #HtmlOPTGROUPNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlOPTGROUPNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'optgroup'! !

HtmlDOMNode subclass: #HtmlOPTIONNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlOPTIONNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'option'! !

HtmlDOMNode subclass: #HtmlPARAMNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlPARAMNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'param'! !

HtmlDOMNode subclass: #HtmlPNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlPNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'p'! !

HtmlDOMNode subclass: #HtmlPRENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlPRENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'pre'! !

HtmlDOMNode subclass: #HtmlProcessingDirectiveNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

HtmlDOMNode subclass: #HtmlQNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlQNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'q'! !

HtmlDOMNode subclass: #HtmlSAMPNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSAMPNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'samp'! !

HtmlDOMNode subclass: #HtmlSCRIPTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSCRIPTNode methodsFor: 'accessing' stamp: 'CAs 7/9/2012 02:20'!
parseContents: aStream

	| str spos script |

	self hasTrailingSlash ifTrue: [^self end: aStream position; endTagMissing: true].

	spos := aStream position.
	str := aStream upTo: $>.
	
	[(str asLowercase endsWith: ('</',self tag)) or: [aStream atEnd]] 
		whileFalse: [str := (str, '>', (aStream upTo: $>))].

	self end: aStream position.
	script := (str first: (str size - (self tag size + 2))) trimBoth.

	(script trimBoth size > 0) ifTrue:
	[ 
		((script beginsWith: '<!!--') and: [script endsWith: '-->']) ifFalse:
		[
			(self dtd isXML and: [(script beginsWith: '<!![CDATA[') and:[script endsWith:']]>']]) 
			ifTrue:
			[
				| child  |
				self addChild: (child:= HtmlXMLCDataNode new start: spos).
				aStream position: spos.
				child parseContents: aStream.
				aStream upTo: $>.
				self end: aStream position. 
			]
			ifFalse:
			[
				self addChild: (HtmlCDATANode  
					start: spos end: (aStream position - (self tag size + 3))).
				self addError:(HtmlUncommentedScriptError new).
			]
		]
		ifTrue:
		[
			| child in cend |
			cend := (aStream position - (self tag size + 3)).
			in := ReadStream on: (aStream contents first: cend).
			in position: spos.
			child := HtmlCommentNode new start: spos.
			self addChild: child.
			child parseContents: in. 
		]
	].
	self end: aStream position.
	self endTagMissing: (str asLowercase endsWith: ('</',self tag)) not.
	! !

!HtmlSCRIPTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'script'! !

HtmlDOMNode subclass: #HtmlSELECTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSELECTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'select'! !

HtmlDOMNode subclass: #HtmlSMALLNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSMALLNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'small'! !

HtmlDOMNode subclass: #HtmlSNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSNode methodsFor: 'accessing' stamp: ''!
tag
	^ 's'! !

HtmlDOMNode subclass: #HtmlSPANNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSPANNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'span'! !

HtmlDOMNode subclass: #HtmlSTRIKENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSTRIKENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'strike'! !

HtmlDOMNode subclass: #HtmlSTRONGNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSTRONGNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'strong'! !

HtmlDOMNode subclass: #HtmlSTYLENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSTYLENode methodsFor: 'accessing' stamp: 'tb 12/20/2005 13:45'!
isCSS

	^true! !

!HtmlSTYLENode methodsFor: 'accessing' stamp: 'tb 1/3/2006 17:10'!
parseContents: aStream

	| str  child |

	str := aStream upTo: $>.
	
	[(str asLowercase endsWith: ('</',self tag)) or: [aStream atEnd]] 
		whileFalse: [str := (str, '>', (aStream upTo: $>))].

	self end: aStream position.
	self endTagMissing: (str asLowercase endsWith: ('</',self tag)) not.
	child := CSSStyleSheet start: self startForNextElement end: (aStream position - (self tag size + 3)).
	self addChild: child.
	aStream position: child start.
	child parseContents: aStream.
	aStream position: self end.
	
! !

!HtmlSTYLENode methodsFor: 'accessing' stamp: 'tb 12/30/2005 17:36'!
stylesheet

	^self children first! !

!HtmlSTYLENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'style'! !

!HtmlSTYLENode methodsFor: 'accessing' stamp: 'tb 12/20/2005 13:51'!
validateAgainstDTD: aDTD! !

HtmlDOMNode subclass: #HtmlSUBNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSUBNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'sub'! !

HtmlDOMNode subclass: #HtmlSUPNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlSUPNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'sup'! !

HtmlDOMNode subclass: #HtmlTABLENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTABLENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'table'! !

HtmlDOMNode subclass: #HtmlTBODYNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTBODYNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'tbody'! !

HtmlDOMNode subclass: #HtmlTDNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTDNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'td'! !

HtmlDOMNode subclass: #HtmlTEXTAREANode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTEXTAREANode methodsFor: 'accessing' stamp: ''!
tag
	^ 'textarea'! !

HtmlDOMNode subclass: #HtmlTFOOTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTFOOTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'tfoot'! !

HtmlDOMNode subclass: #HtmlTHEADNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTHEADNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'thead'! !

HtmlDOMNode subclass: #HtmlTHNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTHNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'th'! !

HtmlDOMNode subclass: #HtmlTITLENode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTITLENode methodsFor: 'accessing' stamp: ''!
tag
	^ 'title'! !

HtmlDOMNode subclass: #HtmlTRNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTRNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'tr'! !

HtmlDOMNode subclass: #HtmlTTNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlTTNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'tt'! !

HtmlDOMNode subclass: #HtmlULNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlULNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'ul'! !

HtmlDOMNode subclass: #HtmlUNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlUNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'u'! !

HtmlDOMNode subclass: #HtmlVARNode
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-DOM'!

!HtmlVARNode methodsFor: 'accessing' stamp: ''!
tag
	^ 'var'! !

!HtmlParseNode class methodsFor: 'as yet unclassified' stamp: 'tb 1/3/2006 16:47'!
start: s end: e

	^self new start: s; end: e! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
addChild: aNode

	self children: (self children copyWith: aNode).
	aNode parent: self.! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
addError: anError

	anError parent: self.
	self errors: (self errors copyWith: anError)! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
allErrors

	| collection |

	self errors do:[:ea | ea position isNil ifTrue: [ea position: self start]].
	collection := OrderedCollection new.
	collection addAll: (self errors).
	self children do:[:ea | collection addAll: ea allErrors].
	^collection asArray! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/4/2006 10:46'!
allErrorsMatching: aBlock

	| errs |

	errs := OrderedCollection new.

	self nodesDo: [:ea | ea errors do:[:er | (aBlock value: er) ifTrue: [errs add: er]]].

	^errs asArray! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 17:36'!
asString

	| |

	^(self prefix, self innerContents, self suffix)
		! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
children
	"Answer the value of children"

	^ children isNil ifTrue: [#()] ifFalse: [children]! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:53'!
children: anObject
	"Set the value of children"

	children := anObject.
	children do:[:ea | ea parent: self]! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 7/12/2007 11:48'!
content

	^self parent ifNil: [''] ifNotNil: [self parent content]! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
detectAncestor: aBlock

	| p |

	p := self.
	[p notNil] whileTrue: [(aBlock value: p) ifTrue: [^p]. p := p parent].
	^p

	! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
document

	^self parent document! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
end
	"Answer the value of end"

	^ end! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:53'!
end: anObject
	"Set the value of end"

	end := anObject! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
errors
	"Answer the value of errors"

	^ errors isNil ifTrue: [#()] ifFalse: [errors]! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:53'!
errors: anObject
	"Set the value of errors"

	errors := anObject! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
firstAncestor: aString

	| p |

	p := self parent.
	[p notNil] whileTrue: [p tag = aString ifTrue: [^p]. p := p parent].
	^p

	! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
hasAncestor: aString

	^(self firstAncestor: aString) notNil! !

!HtmlParseNode methodsFor: 'testing' stamp: 'tb 1/3/2006 17:35'!
innerContents

	| in | 

	^self children isEmpty 
		ifTrue: ['']
		ifFalse: 
		[
			in := ReadStream on: self content.
			in position: self children first start.
			in next: (self children last end - self children first start).
		]! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 17:30'!
innerId

	^(self uniqueId, '-inner')! !

!HtmlParseNode methodsFor: 'testing' stamp: 'tb 1/3/2006 16:47'!
isCDATA

	^false! !

!HtmlParseNode methodsFor: 'testing' stamp: 'tb 1/3/2006 16:47'!
isCSS

	^false! !

!HtmlParseNode methodsFor: 'testing' stamp: 'tb 1/3/2006 16:47'!
isComment

	^false! !

!HtmlParseNode methodsFor: 'testing' stamp: 'gk 11/12/2007 12:35'!
isDOMNode

	^false! !

!HtmlParseNode methodsFor: 'testing' stamp: 'tb 1/3/2006 16:47'!
isDocument

	^false! !

!HtmlParseNode methodsFor: 'testing' stamp: 'tb 1/3/2006 17:01'!
isHTML

	^false! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:57'!
isImplied
	
	^false! !

!HtmlParseNode methodsFor: 'testing' stamp: 'tb 1/3/2006 16:47'!
isStyleSheet

	^false! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
level

	| level p |
	level := 0.
	p := self.
	[p notNil] whileTrue: [p := p parent. p notNil ifTrue:[level := level + 1]].
	^level! !

!HtmlParseNode methodsFor: 'iterating' stamp: 'tb 1/3/2006 16:47'!
nodesCollect: aBlock

	| col |

	col := OrderedCollection new.
	self nodesDo:[:ea | (aBlock value: ea) ifTrue: [col add: ea]].
	^col asArray! !

!HtmlParseNode methodsFor: 'iterating' stamp: 'tb 1/3/2006 16:47'!
nodesDo: aBlock

	aBlock value: self.
	self children do:[:ea | ea nodesDo: aBlock].! !

!HtmlParseNode methodsFor: 'iterating' stamp: 'spd 12/7/2010 21:53'!
nodesSelect: aBlock

	| col |

	col := OrderedCollection new.
	self nodesDo:[:ea | (aBlock value: ea) ifTrue: [col add: ea]].
	^col asArray! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
parent
	"Answer the value of parent"

	^ parent isNil ifTrue: [nil] ifFalse: [parent first]! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:53'!
parent: anObject
	"Set the value of parent"

	parent := WeakArray with: anObject! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'gk 11/13/2007 14:45'!
parentWithClass: aClass
	"Find parent tag of given class."

	self parent ifNil: [^self].
	^self parent class = aClass
		ifTrue: [self parent]
		ifFalse: [self parent parentWithClass: aClass]
	! !

!HtmlParseNode methodsFor: 'parsing' stamp: 'tb 1/3/2006 17:15'!
parseContents: aStream

	self subclassResponsibility! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
prefix

	| in theEnd |

	in := ReadStream on: self content.
	in position: self start.	
	theEnd := self children isEmpty ifTrue:[self end] ifFalse:[self children first start].

	^in next: (theEnd - self start)! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
prefixId

	^self uniqueId! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
rawContent

	| in |

	in := ReadStream on: self document content.
	in position: self start.	
	^in next: (self end - self start)! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
start
	"Answer the value of start"

	^ start! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:52'!
start: anObject
	"Set the value of start"

	start := anObject! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
startForNextElement

	^self children isEmpty ifTrue:
	[
		| in |
		in := ReadStream on: self document content.
		in position: start.
		in upTo: $>.
		in position
	]
	ifFalse:
	[
		self children last end
	]! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/4/2006 17:50'!
suffix

	| in theStart |

	in := ReadStream on: self content.

	theStart := self children isEmpty ifTrue:[self end] ifFalse:[self children last end].
	in position: theStart.	
	^in next: (0 max: (self end - theStart))! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
suffixId

	^(self uniqueId, '-end')! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
tag
	"Answer the value of tag"

	^ tag isNil ifFalse: [tag asLowercase] ifTrue: [tag]! !

!HtmlParseNode methodsFor: 'accessing' stamp: 'tb 1/3/2006 16:47'!
uniqueId

	^(self parent uniqueId, '.', ((self parent children indexOf: self) asString))! !

Object subclass: #HtmlValidator
	instanceVariableNames: 'url document errors dom content loader'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'HTML-Validator'!

!HtmlValidator commentStamp: 'tb 4/14/2006 21:59' prior: 0!
I validate html for correctness.  I am the usual entry point into this library.  Typical usage for simply obtaining a DOM of a webpage might be (HtmlValidator onUrl: 'http://www.amazon.com' asUrl) dom.  If you want the DOM decorated with errors, then call validate.  

Author: Todd Blanchard, 2006.  tblanchard@mac.com

Free for any use that does not involve providing QA software packages or services for websites on a commercial or non-commercial basis.!

!HtmlValidator class methodsFor: 'as yet unclassified' stamp: 'tb 11/28/2005 17:34'!
on: aString

	^self new content: aString! !

!HtmlValidator class methodsFor: 'as yet unclassified' stamp: 'tb 11/28/2005 10:35'!
onUrl: aStringOrUrl

	^self new url: aStringOrUrl asUrl ! !

!HtmlValidator methodsFor: 'actions' stamp: 'tb 12/4/2005 23:01'!
bruteForceAnalyze

	"Maybe add more as I think of things"


	^self checkHtmlHeadBody! !

!HtmlValidator methodsFor: 'actions' stamp: 'tb 1/15/2006 19:26'!
checkHtmlHeadBody

	| structure in tags pos tag patterns |

	structure := OrderedCollection new.
	patterns := #( "These are the only valid combinations"
				#( '<html' '<head' '</head' '<body' '</body' '</html')
				#( '<html' '<body' '</body' '</html')
				#( '<html' '<head' '</head' '</html')
				#( '<html' '</html')
				#( '<body' '</body')
				#( '<head' '</head' )
				#( )
			).
	tags := 	patterns first.

	in := ReadStream on: self content asLowercase.

	[in atEnd] whileFalse:
	[
		| min locs |
		pos := in position.

		locs := (tags collect: [:ea | in position: pos; upToAll: ea; position]). 	
		(locs reject:[:ea | locs first = ea])	isEmpty ifFalse:
		[
			min := locs min.
			tag := tags at: (locs indexOf: min).

			structure add: tag.
			in position: min.
			in atEnd ifFalse: [in skip: 1].
		]
		ifTrue: [in upToEnd].
	].
	structure := structure asArray.
	((patterns detect: [:ea | structure = ea] ifNone: [nil]) notNil) ifFalse:  
	[
		self dom addError: (HtmlHeadBodyNestingError new nesting: structure asArray). 
	]
	ifTrue:
	[
		(dom children size > 1 and: 
			[(dom children detect:[:ea | ea tag = 'html'] ifNone:[nil]) notNil]) ifTrue:
		[
			dom children do:
			[:ea | 
				(ea isComment or: [ea tag = 'html']) ifFalse: 
				[
					ea addError:(HtmlTagImproperlyNestedError new).
				]
			]
		]
	]
	! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 11/28/2005 17:35'!
content
	"Answer the value of content"

	^ content isNil 
		ifFalse: [content]
		ifTrue: [content := self document content]
	! !

!HtmlValidator methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:51'!
content: anObject
	"Set the value of content"

	content := anObject! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 1/13/2006 18:42'!
cssErrors

	| list |
	list := OrderedCollection new.
	self stylesheets do:
	[:ss |
		list addAll: (ss allErrorsMatching: [:ea | ea isCSSError])
	].
	^list asArray! !

!HtmlValidator methodsFor: 'accessing' stamp: 'sebastian.sastre 10/18/2010 11:56'!
document
	"Answer the value of document"

	document isNil ifTrue: 
	[
		url isNil 
			ifTrue: [nil]
			ifFalse: 
			[
				document := self loader documentForUrl: self url.
				loader url: document url.
			] 
	].
	^document ! !

!HtmlValidator methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:35'!
document: anObject
	"Set the value of document"

	document := anObject! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 4/14/2006 21:55'!
dom
	"Answer the value of dom"

	dom isNil ifTrue: [self parse].
	^ dom! !

!HtmlValidator methodsFor: 'accessing' stamp: 'sebastian.sastre 9/20/2010 11:53'!
dom: anObject
	"Set the value of dom"

	dom := anObject! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 11/28/2005 10:38'!
errors
	"Answer the value of errors"

	^ errors! !

!HtmlValidator methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:36'!
errors: anObject
	"Set the value of errors"

	errors := anObject! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 1/4/2006 11:40'!
htmlErrors

	^self dom allErrorsMatching: [:ea | ea isCSSError not ].! !

!HtmlValidator methodsFor: 'initialize-release' stamp: 'tb 11/28/2005 10:37'!
initialize

	super initialize.
! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 1/10/2006 19:34'!
loader
	"Answer the value of loader"

	^ loader isNil ifTrue: [loader := HtmlCache new url: self url] ifFalse: [loader]! !

!HtmlValidator methodsFor: 'accessing' stamp: 'PK 10/18/2010 22:36'!
loader: anObject
	"Set the value of loader"

	loader := anObject! !

!HtmlValidator methodsFor: 'actions' stamp: 'tb 1/10/2006 19:48'!
parse

	| in |
	
	in := ReadStream on: self content.
	self dom: (HtmlDocument new loader: self loader;
			requestUrl: self url; 
			responseUrl: (self document isNil ifTrue: [nil] ifFalse: [self document url]); 
			parseContents: in)
! !

!HtmlValidator methodsFor: 'actions' stamp: 'tb 1/10/2006 19:50'!
parseStylesheets

	self dom stylesheets
! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 12/30/2005 18:52'!
stylesheets

	^self dom stylesheets! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 11/28/2005 10:37'!
url
	"Answer the value of url"

	^ url! !

!HtmlValidator methodsFor: 'accessing' stamp: 'tb 1/10/2006 15:07'!
url: anObject
	"Set the value of url"

	url := anObject asUrl.
! !

!HtmlValidator methodsFor: 'actions' stamp: 'tb 1/10/2006 19:47'!
validate

	| dtd |
	
	self parse.
	self checkHtmlHeadBody.
	dtd := HtmlDTD forDocType: self dom docType.
	self dom applyStyles.
	self validateAgainstDTD: dtd.
	self validateCSSImports.
	self validateCSSAgainstDTD: dtd! !

!HtmlValidator methodsFor: 'actions' stamp: 'tb 1/14/2006 14:04'!
validateAgainstDTD: dtd

	self dom nodesDo:
	[:ea |
		(ea isComment or: [ea isCDATA or: [ea isDocument or: [ea isCSS]]]) ifFalse:
		[
			(dtd tagIsReal: ea dtdTag) 
			ifFalse: [ea addError: (HtmlTagUnknownError new)]	
			ifTrue:
			[
				dtd validateChildrenOfNode: ea.
			 	dtd validateAttributesOfNode: ea.
			]
		]
		ifTrue:
		[
			(ea isCDATA) ifTrue: [ea checkForUnescapedCharacters]
			ifFalse: 
			[
				(ea isComment) ifTrue: 
				[
					ea endTagMissing ifTrue: 
					[
						ea addError: HtmlUnclosedCommentError new
					]
				]
			]
		]
	]! !

!HtmlValidator methodsFor: 'actions' stamp: 'tb 1/14/2006 21:39'!
validateCSSAgainstDTD: dtd

	| rules  color backgroundColor |

	rules := dom allStyles.
	rules do:
	[:ea |
		ea selector isValid ifFalse: [ea addError: CSSSelectorNotValidError new]
		ifTrue:
		[
			ea selector isCSS3 ifTrue: [ea addError: CSSVersion3ConstructFoundError new].
			ea selector typeSelectors do:
			[:sel |
				sel element = sel element asLowercase ifFalse:
				[
					ea addError: (CSSSelectorElementNotLowercaseError new element: sel element).
				].
				(dtd tagIsReal: (dtd isXML 
					ifTrue: [sel element asLowercase] 
					ifFalse:[sel element asUppercase])) 
				ifFalse:
				[
					ea addError: (CSSSelectorElementNotValidError new element: sel element)
				]
			].
			(ea matchCount = 0 and: [ea selector isValid]) ifTrue:
			[
				ea addError: CSSUnmatchedRuleError new.		
			].
			ea properties do:
			[:prop |
				| lcname |
				lcname := prop propertyName asLowercase.
				prop propertyName = lcname ifFalse: 
				[
					ea addError:
					(CSSPropertyNameNotLowercaseError new propertyName: prop propertyName).
					prop propertyName: lcname.
				].
 
				prop definition isNil ifTrue:
				[
					ea addError: 
					(CSSPropertyUnknownError new propertyName: prop propertyName)
				]
				ifFalse:
				[
					prop valueIsValid ifFalse:
					[
						ea addError: 
						(CSSPropertyValueNotValidError new propertyName: prop propertyName)					
					]
				]
			].

			color := backgroundColor := nil.

			ea properties do:
			[:prop |
				prop propertyName = 'color' ifTrue: [ color:= prop].
				prop propertyName = 'background-color' ifTrue: [backgroundColor := prop].

			].

			color = backgroundColor ifFalse:
			[
				color isNil ifTrue: 
				[
					ea addError: 
					(CSSColorAndBackgroundColorNotSet new propertyName: 'color')
				]
				ifFalse:
				[
					backgroundColor isNil ifTrue:
					[
						ea addError: 
					(CSSColorAndBackgroundColorNotSet new propertyName: 'background-color')
					]
				]
			]
		]
	]
	! !

!HtmlValidator methodsFor: 'actions' stamp: 'tb 1/6/2006 17:31'!
validateCSSImports

	| ss styleFound |
	ss := dom stylesheets.
	ss do:
	[:ea |
		styleFound := false.
		ea children do:
		[:rule |
			styleFound ifFalse:
			[
				rule isStyleRule ifTrue:[ styleFound := true]
			]
			ifTrue:
			[
				rule isAtImportRule ifTrue:
				[
					rule addError: (CSSImportAfterStyleRuleError new)
				]
			]
		]


	]
! !

Object subclass: #OSFileLock
	instanceVariableNames: 'fileStream exclusive'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!OSFileLock commentStamp: 'dtl 2/23/2004 19:36' prior: 0!
I describe the region representing the entire addressable space of an external file, including regions that have not yet been allocated for use by the file. On platforms that support file locking, an OSFileLock is used to describe a lock on the entire file.
!

!OSFileLock class methodsFor: 'instance creation' stamp: 'jf 2/22/2004 17:35'!
onFile: aFileStream exclusive: writeLockFlag
	"A region representing the whole file"

	^ super new onFile: aFileStream exclusive: writeLockFlag
! !

!OSFileLock methodsFor: 'comparing' stamp: 'dtl 11/28/2010 09:18'!
= aFileLock
	^ ((self species = aFileLock species)
		and: [self fileStream == aFileLock fileStream])
			and: [self exclusive == aFileLock exclusive]! !

!OSFileLock methodsFor: 'conflict testing' stamp: 'jf 2/22/2004 16:38'!
conflictsWith: otherFileLock
	^ (self isExclusive or: [otherFileLock isExclusive])
		and: [self overlaps: otherFileLock]! !

!OSFileLock methodsFor: 'accessing' stamp: 'jf 2/22/2004 16:00'!
exclusive
	"Answer the value of exclusive. Default is true, indicating a read-write lock
	as opposed to a shared read lock."

	^ exclusive ifNil: [exclusive := true].! !

!OSFileLock methodsFor: 'accessing' stamp: 'jf 2/22/2004 16:00'!
exclusive: anObject
	"Set the value of exclusive"

	exclusive := anObject! !

!OSFileLock methodsFor: 'accessing' stamp: 'jf 2/22/2004 16:00'!
fileStream
	"Answer the value of fileStream"

	^ fileStream! !

!OSFileLock methodsFor: 'accessing' stamp: 'jf 2/22/2004 16:00'!
fileStream: anObject
	"Set the value of fileStream"

	fileStream := anObject! !

!OSFileLock methodsFor: 'comparing' stamp: 'jf 2/22/2004 18:32'!
hash
	^ (self fileStream hash + self exclusive hash) hashMultiply! !

!OSFileLock methodsFor: 'testing' stamp: 'jf 2/22/2004 18:13'!
isActive
	^ self fileStream closed not! !

!OSFileLock methodsFor: 'testing' stamp: 'jf 2/22/2004 16:01'!
isExclusive
	^ self exclusive! !

!OSFileLock methodsFor: 'private' stamp: 'jf 2/22/2004 18:07'!
length
	^ 0! !

!OSFileLock methodsFor: 'system locking' stamp: 'dtl 5/8/2006 06:52'!
lock
	"Answer true on success"

	^ self lockIfFail: [false]
! !

!OSFileLock methodsFor: 'system locking' stamp: 'dtl 5/8/2006 06:57'!
lockIfFail: failBlock

	| handle result |
	handle := ThisOSProcess accessor handleFromFileStream: self fileStream.
	result := ThisOSProcess accessor
		primLockFileRegion: handle
		offset: self offset
		length: self length
		exclusive: self isExclusive.
	result == 0
		ifTrue: [^ true]
		ifFalse: [^ failBlock value]
! !

!OSFileLock methodsFor: 'private' stamp: 'jf 2/22/2004 18:02'!
offset
	^ 0! !

!OSFileLock methodsFor: 'initialize-release' stamp: 'jf 2/22/2004 16:02'!
onFile: aFileStream exclusive: writeLockFlag

	self fileStream: aFileStream.
	self exclusive: writeLockFlag
! !

!OSFileLock methodsFor: 'comparing' stamp: 'jf 2/22/2004 16:31'!
overlaps: aFileLock
	"Answer true if the receiver represents an addressable region that overlaps aFileLock"
	
	^ aFileLock fileStream fullName = self fileStream fullName! !

!OSFileLock methodsFor: 'comparing' stamp: 'jf 2/22/2004 16:35'!
overlapsRegion: aFileRegionLock
	^ aFileRegionLock fileStream fullName = self fileStream fullName! !

!OSFileLock methodsFor: 'system locking' stamp: 'dtl 5/8/2006 07:00'!
test
	"Answer true if this is a lockable file or region"

	| handle result |
	handle := ThisOSProcess accessor handleFromFileStream: self fileStream.
	result := ThisOSProcess accessor
		primTestLockableFileRegion: handle
		offset: self offset
		length: self length
		exclusive: self isExclusive.
		
	^ (result == -1 or: [result isNil])
		ifTrue: [false]
		ifFalse: [result first]! !

!OSFileLock methodsFor: 'system locking' stamp: 'dtl 5/8/2006 06:58'!
unlock
	"Answer true on success"

	^ self unlockIfFail: [false]
! !

!OSFileLock methodsFor: 'system locking' stamp: 'dtl 5/8/2006 06:57'!
unlockIfFail: failBlock

	| handle result |
	handle := ThisOSProcess accessor handleFromFileStream: self fileStream.
	result := ThisOSProcess accessor
		primUnlockFileRegion: handle
		offset: self offset
		length: self length.
	result == 0
		ifTrue: [^ true]
		ifFalse: [^ failBlock value]
! !

OSFileLock subclass: #OSFileRegionLock
	instanceVariableNames: 'interval'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!OSFileRegionLock commentStamp: 'jf 2/22/2004 19:50' prior: 0!
I describe an addressable region of contiguous bytes in an external file. On platforms that support file locking, an OSFileRegionLock is used to specify a portion of the file to be locked.
!

!OSFileRegionLock class methodsFor: 'instance creation' stamp: 'dtl 2/22/2004 13:58'!
onFile: aFileStream from: start to: end exclusive: writeLockFlag
	"A region representing part of a file"

	^ self onFile: aFileStream interval: (start to: end) exclusive: writeLockFlag
! !

!OSFileRegionLock class methodsFor: 'instance creation' stamp: 'dtl 2/22/2004 13:59'!
onFile: aFileStream interval: anInterval exclusive: writeLockFlag
	"A region representing part of a file"

	^ super new onFile: aFileStream interval: anInterval exclusive: writeLockFlag
! !

!OSFileRegionLock methodsFor: 'comparing' stamp: 'jf 2/22/2004 18:33'!
= aFileRegion

	^ (super = aFileRegion) and: [self interval = aFileRegion interval]
! !

!OSFileRegionLock methodsFor: 'comparing' stamp: 'jf 2/22/2004 18:33'!
hash

	^ (super hash + self interval hash) hashMultiply
! !

!OSFileRegionLock methodsFor: 'accessing' stamp: 'dtl 3/10/2005 20:23'!
interval
	"Answer the value of interval"

	^ interval! !

!OSFileRegionLock methodsFor: 'accessing' stamp: 'dtl 3/10/2005 20:23'!
interval: anObject
	"Set the value of interval"

	interval := anObject! !

!OSFileRegionLock methodsFor: 'private' stamp: 'jf 2/22/2004 18:02'!
length
	^ self interval size! !

!OSFileRegionLock methodsFor: 'private' stamp: 'jf 2/22/2004 18:02'!
offset
	^ self interval first! !

!OSFileRegionLock methodsFor: 'initialize-release' stamp: 'jf 2/22/2004 16:03'!
onFile: aFileStream interval: anInterval exclusive: writeLockFlag

	self onFile: aFileStream exclusive: writeLockFlag.
	self interval: anInterval.! !

!OSFileRegionLock methodsFor: 'comparing' stamp: 'jf 2/22/2004 19:47'!
overlaps: aFileLock
	"Call #overlapsRegion: on aFileLock since we know we're a region but we don't know whether aFileLock is"
	
	^ aFileLock overlapsRegion: self! !

!OSFileRegionLock methodsFor: 'comparing' stamp: 'jf 2/22/2004 18:56'!
overlapsRegion: aFileRegionLock
	^ (super overlapsRegion: aFileRegionLock) and: [(self interval intersection: aFileRegionLock interval) isEmpty not]! !

Object subclass: #OSProcess
	instanceVariableNames: 'pid'
	classVariableNames: 'UseIOHandle'
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!OSProcess commentStamp: '<historical>' prior: 0!
I represent an operating system process, such as the process in which the Squeak VM is currently running. My subclasses implement system specific features for Unix, Windows, MacOS, or other operating systems.
!

OSProcess subclass: #ExternalOSProcess
	instanceVariableNames: 'runState initialStdIn initialStdOut initialStdErr'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!ExternalOSProcess commentStamp: '<historical>' prior: 0!
I represent an OSProcess other than the process in which this Squeak is executing. I maintain information about the state of the external process during and after the lifetime of the process.!

ExternalOSProcess subclass: #ExternalMacOSProcess
	instanceVariableNames: 'ppid exitStatus'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Mac'!

!ExternalMacOSProcess commentStamp: '<historical>' prior: 0!
I represent an external MacOS process other than the process in which this Squeak is executing. I maintain information about the state of the external process during and after the lifetime of the process. In particular, I hold the exit status of the process after it completes execution. When the external process changes state (e.g. it exits), the VM signals a Squeak semaphore. A singleton MacOSProcessAccessor maintains a process which waits on the semaphore, and sends a changed: #childProcessStatus message to itself, thereby notifying its dependent MacOSProcess (a singleton) to check the status of all its ExternalMacOSProcess children, and #update: them accordingly.!

!ExternalMacOSProcess class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:45'!
isResponsibleForThisPlatform
	"Answer true if, for the current platform, this class is responsible for representing
	an OS process other than that in which the Squeak VM is currently running."

	^ self isNonUnixMac
! !

ExternalOSProcess subclass: #ExternalOS2Process
	instanceVariableNames: 'ppid exitStatus'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-OS2'!

!ExternalOS2Process commentStamp: '<historical>' prior: 0!
I represent an external OS2 process other than the process in which this Squeak is executing. I maintain information about the state of the external process during and after the lifetime of the process. In particular, I hold the exit status of the process after it completes execution. When the external process changes state (e.g. it exits), the VM signals a Squeak semaphore. A singleton OS2ProcessAccessor maintains a process which waits on the semaphore, and sends a changed: #childProcessStatus message to itself, thereby notifying its dependent OS2Process (a singleton) to check the status of all its ExternalOS2Process children, and #update: them accordingly.!

!ExternalOS2Process class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:47'!
isResponsibleForThisPlatform
	"Answer true if, for the current platform, this class is responsible for representing
	an OS process other than that in which the Squeak VM is currently running."

	^ self isOS2
! !

!ExternalOSProcess class methodsFor: 'concrete subclasses' stamp: 'dtl 3/5/2005 12:02'!
concreteClass

	"ExternalOSProcess concreteClass"

	^ self subclasses
		detect: [:c | c isResponsibleForThisPlatform]
		ifNone: [self notify: self printString,
					': No concrete class implementation available for system type ',
					self platformName printString.
				nil]

! !

!ExternalOSProcess class methodsFor: 'instance creation' stamp: 'dtl 1/16/2001 05:36'!
exec: programName
	"Run a program in an external OS process, and answer an instance of myself
	which represents the external process."

	"ExternalOSProcess exec: '/bin/ls'"

	^ self concreteClass exec: programName
! !

!ExternalOSProcess class methodsFor: 'shell' stamp: 'SeanDeNigris 12/27/2013 12:30'!
shellFlags

	^ '-c'.! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 11/8/2000 21:41'!
accessor
	"Answer an OSProcessAccessor which may be used to obtain information about
	the external OS process which I represent."

	^ OSProcess thisOSProcess processAccessor
! !

!ExternalOSProcess methodsFor: 'initialize - release' stamp: 'dtl 3/1/2002 06:41'!
closeInitialStdErr

	initialStdErr ifNotNil: [initialStdErr close]
! !

!ExternalOSProcess methodsFor: 'initialize - release' stamp: 'dtl 3/1/2002 06:42'!
closeInitialStdIn

	initialStdIn ifNotNil: [initialStdIn close]
! !

!ExternalOSProcess methodsFor: 'initialize - release' stamp: 'dtl 3/1/2002 06:42'!
closeInitialStdOut

	initialStdOut ifNotNil: [initialStdOut close]
! !

!ExternalOSProcess methodsFor: 'initialize - release' stamp: 'dtl 11/21/2006 14:04'!
closeStreams

	self closeInitialStdIn; closeInitialStdOut; closeInitialStdErr! !

!ExternalOSProcess methodsFor: 'setting run state' stamp: 'dtl 3/4/2001 18:55'!
complete
	"Process has exited and has been reaped. It no longer exists in the external operating system."

	self runState: #complete
! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 3/1/2002 06:38'!
initialStdErr
	"The stderr stream at the time the child process is invoked. If the same as
	stderr for the current Squeak process, it may change as a result of Squeak
	using its stderr stream. The child process may also modify its actual stderr;
	therefore this is not an accurate representation of the child process stderr
	during the life of the child process."

	^ initialStdErr! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 3/1/2002 06:38'!
initialStdErr: anExternalStream

	initialStdErr := anExternalStream! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 3/1/2002 06:39'!
initialStdIn
	"The stdin stream at the time the child process is invoked. If the same as
	stdin for the current Squeak process, it may change as a result of Squeak
	using its stdin stream. The child process may also modify its actual stdin;
	therefore this is not an accurate representation of the child process stdin
	during the life of the child process."

	^ initialStdIn! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 3/1/2002 06:39'!
initialStdIn: anExternalStream

	initialStdIn := anExternalStream! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 3/1/2002 06:39'!
initialStdOut
	"The stdout stream at the time the child process is invoked. If the same as
	stdout for the current Squeak process, it may change as a result of Squeak
	using its stdout stream. The child process may also modify its actual stdout;
	therefore this is not an accurate representation of the child process stdout
	during the life of the child process."

	^ initialStdOut! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 3/1/2002 06:40'!
initialStdOut: anExternalStream

	initialStdOut := anExternalStream! !

!ExternalOSProcess methodsFor: 'initialize - release' stamp: 'dtl 10/6/2000 21:01'!
initialize

	self notYetRunning! !

!ExternalOSProcess methodsFor: 'testing' stamp: 'dtl 10/7/2000 14:45'!
isAccessible

	^ self accessor canAccessChildProcess: self! !

!ExternalOSProcess methodsFor: 'testing' stamp: 'dtl 1/20/2001 11:40'!
isComplete

	^ self runState == #complete! !

!ExternalOSProcess methodsFor: 'testing' stamp: 'dtl 1/20/2001 11:40'!
isNotYetRunning

	^ self runState == #notYetRunning! !

!ExternalOSProcess methodsFor: 'testing' stamp: 'dtl 1/20/2001 11:40'!
isRunning

	^ self runState == #running! !

!ExternalOSProcess methodsFor: 'setting run state' stamp: 'dtl 3/26/2000 15:23'!
notYetRunning
	"Process has not yet entered running state."

	self runState: #notYetRunning
! !

!ExternalOSProcess methodsFor: 'printing' stamp: 'dtl 3/18/2000 14:07'!
printOn: aStream

	super printOn: aStream.
	self isComplete ifTrue: [ aStream nextPutAll: ' (', self runState, ' with status ', self exitStatus printString, ')' ]
		ifFalse: [ aStream nextPutAll: ' (', self runState asString, ')' ]! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 1/25/2004 11:01'!
runState

	^ runState ifNil: [self unknownRunState]
! !

!ExternalOSProcess methodsFor: 'accessing' stamp: 'dtl 3/18/2000 12:30'!
runState: aSymbol

	runState := aSymbol.
	self changed: #runState
! !

!ExternalOSProcess methodsFor: 'setting run state' stamp: 'dtl 3/26/2000 15:23'!
running
	"Process is actively running."

	self runState: #running
! !

!ExternalOSProcess methodsFor: 'testing' stamp: 'dtl 12/22/2001 18:32'!
succeeded
	"Answer true if my process completed successfully. Be optimistic here, and let
	my subclasses implement the details."

	^ self isComplete! !

!ExternalOSProcess methodsFor: 'setting run state' stamp: 'dtl 10/6/2000 20:59'!
unknownRunState
	"Unable to determine the current run state of the process, possibly because
	this is a stale reference to a process which no longer exists."

	self runState: #unknownRunState
! !

!ExternalOSProcess methodsFor: 'updating' stamp: 'dtl 2/27/2002 09:45'!
update: aParameter
	"Notify any dependents if my run state changes. My subclasses will do additional
	updating when the run state changes."

	aParameter == #runState ifTrue: [self changed: #runState]
! !

ExternalOSProcess subclass: #ExternalRiscOSProcess
	instanceVariableNames: 'ppid exitStatus'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-RiscOS'!

!ExternalRiscOSProcess commentStamp: '<historical>' prior: 0!
I represent an external RiscOS task other than the process in which this Squeak is executing. I maintain information about the state of the external task during and after the lifetime of the task. In particular, I hold the exit status of the task after it completes execution. When the external task changes state (e.g. it exits), the VM signals a Squeak semaphore. A singleton RiscOSProcessAccessor maintains a process which waits on the semaphore, and sends a changed: #childProcessStatus message to itself, thereby notifying its dependent RiscOSProcess (a singleton) to check the status of all its ExternalRiscOSProcess children, and #update: them accordingly.
!

!ExternalRiscOSProcess class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:48'!
isResponsibleForThisPlatform
	"Answer true if, for the current platform, this class is responsible for representing
	an OS process other than that in which the Squeak VM is currently running."

	^ self isRiscOS
! !

ExternalOSProcess subclass: #ExternalUnixOSProcess
	instanceVariableNames: 'ppid pwd exitStatus programName arguments initialEnvironment'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Unix'!

!ExternalUnixOSProcess commentStamp: '<historical>' prior: 0!
I represent an external Unix process other than the process in which this Squeak is executing. I maintain information about the state of the external process during and after the lifetime of the process. In particular, I hold the exit status of the process after it completes execution. When the external process changes state (e.g. it exits), the VM signals a Squeak semaphore. A singleton UnixProcessAccessor maintains a process which waits on the semaphore, and sends a changed: #childProcessStatus message to itself, thereby notifying its dependent UnixProcess (a singleton) to check the status of all its ExternalUnixProcess children, and #update: them accordingly.!

!ExternalUnixOSProcess class methodsFor: 'shells' stamp: 'dtl 1/23/2013 20:21'!
bashShellPath
	"A more full-featured shell from the Free Software Foundation"

	| path |
	path := '/bin/bash'.
	(OSProcess fileExists: path)
		ifTrue: [^ path]
		ifFalse: [self notify: path, ' not found']! !

!ExternalUnixOSProcess class methodsFor: 'instance creation' stamp: 'dtl 7/12/2003 11:38'!
command: aCommandString

	"ExternalUnixOSProcess command: 'ls -l /etc'"

	^ self forkAndExec: self defaultShellPath
		arguments: (Array with: '-c' with: aCommandString)
		environment: nil! !

!ExternalUnixOSProcess class methodsFor: 'shells' stamp: 'dtl 1/23/2013 20:30'!
defaultShellPath
	"Default shell to run"

	| path |
	path := '/bin/sh'.
	(OSProcess fileExists: path)
		ifTrue: [^ path]
		ifFalse: [self notify: path, ' not found']! !

!ExternalUnixOSProcess class methodsFor: 'instance creation' stamp: 'dtl 1/16/2001 05:33'!
exec: programName
	"Run a program in an external OS process, and answer an instance of myself
	which represents the external process."

	^ self forkAndExec: programName
! !

!ExternalUnixOSProcess class methodsFor: 'instance creation' stamp: 'dtl 2/27/2002 15:25'!
forkAndExec: executableFile

	"ExternalUnixOSProcess forkAndExec: '/bin/ls'"

	^ super new
		programName: executableFile;
		initialize;
		forkChild
! !

!ExternalUnixOSProcess class methodsFor: 'instance creation' stamp: 'dtl 2/27/2002 15:25'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary 
	"Run a program in an external OS process, and answer an instance of 
	myself which represents the external process."

	"ExternalUnixOSProcess forkAndExec: '/bin/ls' arguments: (Array with: '-l') environment: (UnixProcess env)"

	^ super new
		programName: executableFile;
		arguments: arrayOfStrings;
		initialEnvironment: stringDictionary;
		initialize;
		forkChild
! !

!ExternalUnixOSProcess class methodsFor: 'instance creation' stamp: 'dtl 2/27/2002 15:25'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams
	"Run a program in an external OS process, and answer an instance of myself
	which represents the external process."

	"ExternalUnixOSProcess
		forkAndExec: '/bin/ls'
		arguments: (Array with: '-l')
		environment: (UnixProcess env)
		descriptors: nil"

	| proc |
	proc := super new
		programName: executableFile;
		arguments: arrayOfStrings;
		initialEnvironment: stringDictionary.
	arrayOf3Streams ifNotNil:
		[proc initialStdIn: (arrayOf3Streams at: 1).
		proc initialStdOut: (arrayOf3Streams at: 2).
		proc initialStdErr: (arrayOf3Streams at: 3)].
	^ proc initialize forkChild
! !

!ExternalUnixOSProcess class methodsFor: 'instance creation' stamp: 'dtl 2/27/2002 15:25'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams workingDir: pathString
	"Run a program in an external OS process, and answer an instance of myself
	which represents the external process."

	"ExternalUnixOSProcess forkAndExec: '/bin/ls' arguments: nil environment: nil descriptors: nil workingDir: '/etc'"

	| proc |
	proc := super new
		programName: executableFile;
		arguments: arrayOfStrings;
		initialEnvironment: stringDictionary.
	arrayOf3Streams ifNotNil:
		[proc initialStdIn: (arrayOf3Streams at: 1).
		proc initialStdOut: (arrayOf3Streams at: 2).
		proc initialStdErr: (arrayOf3Streams at: 3)].
	pathString ifNotNil: [proc pwd: pathString].
	^ proc initialize forkChild
! !

!ExternalUnixOSProcess class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:48'!
isResponsibleForThisPlatform
	"Answer true if, for the current platform, this class is responsible for representing
	an OS process other than that in which the Squeak VM is currently running."

	^ self isUnix
! !

!ExternalUnixOSProcess class methodsFor: 'instance creation' stamp: 'dtl 2/28/2002 10:15'!
programName: executableFile arguments: arrayOfStrings initialEnvironment: stringDictionary
	"Answer an instance not yet running."

	^ super new
		programName: executableFile;
		arguments: arrayOfStrings;
		initialEnvironment: stringDictionary
! !

!ExternalUnixOSProcess class methodsFor: 'shells' stamp: 'dtl 1/23/2013 20:21'!
remoteShellPath
	"A remote shell processor. This may need to be edited for different systems."

	| path |
	path := '/usr/bin/rsh'.
	(OSProcess fileExists: path)
		ifTrue: [^ path]
		ifFalse: [self notify: path, ' not found']! !

!ExternalUnixOSProcess class methodsFor: 'shells' stamp: 'dtl 1/23/2013 20:21'!
tkShellPath
	"The wish command shell for Tk/Tcl. This tends to be installed in a wide variety
	of places, so you may need to edit this method for your system."

	| path |
	path := '/usr/bin/wish'.
	(OSProcess fileExists: path)
		ifTrue: [^ path]
		ifFalse: [self notify: path, ' not found']! !

!ExternalUnixOSProcess methodsFor: 'comparing' stamp: 'dtl 11/15/2013 10:31'!
= processProxy
	"True if this is a proxy that represents the same OS process as processProxy"

	^ ((processProxy isKindOf: ExternalUnixOSProcess)
			and: [pid = processProxy pid])
				and: [ppid = processProxy ppid]! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 1/20/2001 12:48'!
arguments

	^ arguments! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 1/20/2001 12:49'!
arguments: arrayOfArgumentStrings

	arguments := arrayOfArgumentStrings! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 7/3/1999 12:34'!
exitStatus

	^ exitStatus
! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 10/22/1999 22:21'!
exitStatus: anInteger

	exitStatus := anInteger
! !

!ExternalUnixOSProcess methodsFor: 'initialize - release' stamp: 'dtl 4/4/2006 21:16'!
forkChild
	"Start the external OS process. All instances variables except for pid  
	should have been set. The pid will be set following creation of the new 
	external process. Creating a child process is the responsibility of the
	currently executing OS process, so request it to do so on behalf of this
	instance of ExternalUnixOSProcess."

	^ OSProcess thisOSProcess
		processProxy: self
		forkAndExec: programName
		arguments: arguments
		environment: initialEnvironment
		descriptors: (Array
				with: initialStdIn
				with: initialStdOut
				with: initialStdErr)
! !

!ExternalUnixOSProcess methodsFor: 'comparing' stamp: 'dtl 6/28/2010 22:26'!
hash

	^ (pid ifNil: [self identityHash]) + (ppid ifNil: [self identityHash])
! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 3/11/2001 09:17'!
initialEnvironment

	^ initialEnvironment! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 1/21/2001 11:31'!
initialEnvironment: aDictionary

	initialEnvironment := aDictionary! !

!ExternalUnixOSProcess methodsFor: 'initialize - release' stamp: 'dtl 2/11/2001 19:07'!
initialize

	super initialize.
	^ self setDefaults
! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 7/3/1999 12:51'!
pid

	^ pid
! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 10/22/1999 22:21'!
pid: aPid

	pid := aPid
! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 7/3/1999 12:33'!
ppid

	^ ppid
! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 10/22/1999 22:21'!
ppid: aPid

	ppid := aPid
! !

!ExternalUnixOSProcess methodsFor: 'printing' stamp: 'dtl 8/23/2012 22:45'!
printOn: aStream

	self programName isNil
		ifTrue:
			[^ super printOn: aStream]
		ifFalse:
			[aStream
				nextPutAll: 'an ';
				nextPutAll: self class name, ' with pid ';
				nextPutAll: self pid printString;
				nextPutAll: ' on ';
				nextPutAll: programName;
				nextPutAll: ' (';
				nextPutAll: self runState asString.
			self isComplete
				ifTrue: [aStream nextPutAll: ', ';
					nextPutAll: (UnixProcessExitStatus for: exitStatus) printString].
			aStream nextPut: $)].
! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 1/20/2001 12:51'!
programName

	^ programName! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 1/20/2001 12:51'!
programName: fileName

	programName := fileName! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 3/31/2001 17:08'!
pwd

	pwd ifNil: [pwd := self accessor primGetCurrentWorkingDirectory].
	^ pwd! !

!ExternalUnixOSProcess methodsFor: 'accessing' stamp: 'dtl 3/31/2001 17:08'!
pwd: pathString

	pwd := pathString! !

!ExternalUnixOSProcess methodsFor: 'initialize - release' stamp: 'dtl 3/31/2001 17:10'!
setDefaults

	| this |
	this := OSProcess thisOSProcess.
	initialEnvironment ifNil: [self initialEnvironment: this environment].
	initialStdIn ifNil: [self initialStdIn: this stdIn].
	initialStdOut ifNil: [self initialStdOut: this stdOut].
	initialStdErr ifNil: [self initialStdErr: this stdErr].	
	self pwd

! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:44'!
sigabrt
	"Send a SIGABRT signal to the external process which I represent."

	OSProcess thisOSProcess sigabrt: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:52'!
sigalrm
	"Send a SIGALRM signal to the external process which I represent."

	OSProcess thisOSProcess sigalrm: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:52'!
sigchld
	"Send a SIGCHLD signal to the external process which I represent."

	OSProcess thisOSProcess sigchld: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:52'!
sigcont
	"Send a SIGCONT signal to the external process which I represent."

	OSProcess thisOSProcess sigcont: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:53'!
sighup
	"Send a SIGHUP signal to the external process which I represent."

	OSProcess thisOSProcess sighup: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:53'!
sigint
	"Send a SIGINT signal to the external process which I represent."

	OSProcess thisOSProcess sigint: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:54'!
sigkill
	"Send a SIGKILL signal to the external process which I represent."

	OSProcess thisOSProcess sigkill: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:54'!
sigpipe
	"Send a SIGPIPE signal to the external process which I represent."

	OSProcess thisOSProcess sigpipe: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:55'!
sigquit
	"Send a SIGQUIT signal to the external process which I represent."

	OSProcess thisOSProcess sigquit: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:55'!
sigstop
	"Send a SIGSTOP signal to the external process which I represent."

	OSProcess thisOSProcess sigstop: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:55'!
sigterm
	"Send a SIGTERM signal to the external process which I represent."

	OSProcess thisOSProcess sigterm: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:56'!
sigusr1
	"Send a SIGUSR1 signal to the external process which I represent."

	OSProcess thisOSProcess sigusr1: self! !

!ExternalUnixOSProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:56'!
sigusr2
	"Send a SIGUSR2 signal to the external process which I represent."

	OSProcess thisOSProcess sigusr2: self! !

!ExternalUnixOSProcess methodsFor: 'testing' stamp: 'dtl 12/22/2001 18:30'!
succeeded

	^ self isComplete and: [self exitStatus == 0]! !

!ExternalUnixOSProcess methodsFor: 'terminating child' stamp: 'dtl 2/11/2001 16:08'!
terminate
	"Kill the child process which I represent."

	self sigkill.
	self finalize

! !

!ExternalUnixOSProcess methodsFor: 'updating' stamp: 'dtl 7/6/2006 10:06'!
update: aParameter 

	| accessible |
	aParameter == #runState
		ifTrue: [| statusArray | 
			"Has the process exited?"
			statusArray := self accessor primGetChildExitStatus: self pid.
			statusArray notNil
				ifTrue: [self exitStatus: (statusArray at: 2).
						^ self complete]].
	aParameter == #accessibility
		ifTrue: ["Does the process still exist, and is it reachable?"
			((accessible := self isAccessible) notNil and: [accessible])
				ifFalse: [^ self isRunning ifTrue: [self unknownRunState]]].
	super update: aParameter
! !

!ExternalUnixOSProcess methodsFor: 'evaluating' stamp: 'dtl 11/21/2006 14:09'!
value
	"Start the external process"

	self isNotYetRunning ifTrue: [self forkChild; closeStreams]
! !

ExternalOSProcess subclass: #ExternalWindowsOSProcess
	instanceVariableNames: 'ppid exitStatus handle threads commandLine pwd'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Win32'!

!ExternalWindowsOSProcess commentStamp: '<historical>' prior: 0!
I represent an external Windows process other than the process in which this Squeak is executing. I maintain information about the state of the external process during and after the lifetime of the process. In particular, I hold the exit status of the process after it completes execution. When the external process changes state (e.g. it exits), the VM signals a Squeak semaphore. A singleton WindowsOSProcessAccessor maintains a process which waits on the semaphore, and sends a changed: #childProcessStatus message to itself, thereby notifying its dependent WindowsOSProcess (a singleton) to check the status of all its ExternalWindowsOSProcess children, and #update: them accordingly.!

!ExternalWindowsOSProcess class methodsFor: 'instance creation' stamp: 'dtl 2/28/2002 10:56'!
command: aCommandString

	"ExternalWindowsOSProcess command: 'SOL'"
	"ExternalWindowsOSProcess command: 'NoSuchProgram'"

	^ (self commandNoEvaluate: aCommandString) value
! !

!ExternalWindowsOSProcess class methodsFor: 'instance creation' stamp: 'dtl 2/28/2002 10:49'!
commandNoEvaluate: aCommandString
	"Answer an instance not yet running."

	"ExternalWindowsOSProcess commandNoEvaluate: 'SOL'"

	^ super new
		commandLine: aCommandString;
		ppid: OSProcess thisOSProcess pid;
		notYetRunning
! !

!ExternalWindowsOSProcess class methodsFor: 'shells' stamp: 'SeanDeNigris 12/27/2013 10:17'!
defaultShellPath

	^ 'C:\Windows\System32\cmd.exe'.! !

!ExternalWindowsOSProcess class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:50'!
isResponsibleForThisPlatform
	"Answer true if, for the current platform, this class is responsible for representing
	an OS process other than that in which the Squeak VM is currently running."

	^ self isWindows
! !

!ExternalWindowsOSProcess class methodsFor: 'instance creation' stamp: 'dtl 3/1/2002 07:02'!
programName: executableFile arguments: arrayOfStrings initialEnvironment: stringDictionary
	"This is for protocol compatibility with ExternalUnixOSProcess. For now, just reassemble
	a command line string and ignore the environment argument."

	| commandLine |
	commandLine := WriteStream on: String new.
	commandLine nextPutAll: executableFile.
	arrayOfStrings ifNotNil:
		[arrayOfStrings do: [:arg | commandLine nextPut: Character space; nextPutAll: arg]].
	^ self commandNoEvaluate: commandLine contents
! !

!ExternalWindowsOSProcess class methodsFor: 'shell' stamp: 'SeanDeNigris 12/27/2013 12:31'!
shellFlags

	^ '/c'.! !

!ExternalWindowsOSProcess methodsFor: 'comparing' stamp: 'dtl 11/15/2013 10:30'!
= processProxy
	"True if this is a proxy that represents the same OS process as processProxy"

	^ ((processProxy isKindOf: ExternalWindowsOSProcess)
			and: [pid = processProxy pid])
				and: [ppid = processProxy ppid]! !

!ExternalWindowsOSProcess methodsFor: 'initialize - release' stamp: 'dtl 3/1/2002 07:44'!
closeHandles
	"Clean up after process exits."

	self threads do: [:thread | thread closeHandle].
	handle ifNotNil:
		[OSProcess accessor primCloseHandle: handle.
		handle := nil]
! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/26/2002 10:09'!
commandLine

	^ commandLine! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/26/2002 10:09'!
commandLine: aCommandLineString

	commandLine := aCommandLineString! !

!ExternalWindowsOSProcess methodsFor: 'setting run state' stamp: 'dtl 6/24/2006 08:41'!
complete
	"Process has exited and has been reaped. It no longer exists in the external operating system."

	(threads notNil and: [threads size > 0]) ifTrue: [threads do: [:t | t complete]].
	self closeHandles.
	super complete
! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/25/2002 08:32'!
exitStatus

	^ exitStatus! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/25/2002 08:33'!
exitStatus: status

	exitStatus := status
! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/25/2002 07:24'!
handle
	"A Windows HANDLE for this OS process, represented as a ByteArray. The
	handle should be closed when the process exits."

	^ handle! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/25/2002 07:32'!
handle: aHandleObject
	"A Windows HANDLE for this OS process, represented as a ByteArray. The
	handle should be closed when the process exits."

	handle := aHandleObject! !

!ExternalWindowsOSProcess methodsFor: 'comparing' stamp: 'dtl 6/28/2010 22:27'!
hash

	^ (pid ifNil: [self identityHash]) + (ppid ifNil: [self identityHash])
! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/25/2002 07:29'!
ppid

	^ ppid
! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/25/2002 07:29'!
ppid: aProcessID

	ppid := aProcessID! !

!ExternalWindowsOSProcess methodsFor: 'printing' stamp: 'dtl 2/26/2002 10:42'!
printOn: aStream

	self commandLine isNil
		ifTrue:
			[^ super printOn: aStream]
		ifFalse:
			[aStream
				nextPutAll: 'a ';
				nextPutAll: self class name, ' with pid ';
				nextPutAll: self pid printString.
			(self isComplete and: [handle isNil])
				ifTrue:
					[aStream nextPutAll: ' (handle closed)']
				ifFalse:
					[aStream nextPutAll: ' handle ';
						nextPutAll: (handle isNil
							ifTrue: [handle printString]
							ifFalse: [handle asArray printString])].
			aStream
				nextPutAll: ' on ''';
				nextPutAll: commandLine;
				nextPutAll: ''' (';
				nextPutAll: self runState asString;
				nextPut: $)]

! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/28/2002 10:28'!
pwd
	"Answer the current working directory string."

	^ pwd! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/28/2002 10:28'!
pwd: pathString
	"The current working directory string."

	pwd := pathString! !

!ExternalWindowsOSProcess methodsFor: 'accessing' stamp: 'dtl 2/25/2002 07:09'!
threads
	"One or more threads of execution within the OS process"

	^ threads ifNil: [threads := OrderedCollection new]
! !

!ExternalWindowsOSProcess methodsFor: 'updating' stamp: 'dtl 2/27/2002 09:45'!
update: aParameter 

	| accessible stat |
	aParameter == #runState
		ifTrue:
			["Has the process exited?"
	 		stat := self accessor primGetExitStatusForHandle: self handle.
			stat ifNotNil: [self complete closeHandles exitStatus: stat]].
	aParameter == #accessibility
		ifTrue: ["Does the process still exist, and is it reachable?"
			((accessible := self isAccessible) notNil and: [accessible])
				ifFalse: [self isRunning ifTrue: [self unknownRunState]]].
	super update: aParameter
! !

!ExternalWindowsOSProcess methodsFor: 'evaluating' stamp: 'dtl 11/21/2006 14:05'!
value
	"Start the external process"

	| procInfo mainThread |
	self isNotYetRunning ifTrue:
		[procInfo := OSProcess accessor primCommand: self commandLine.
		procInfo isNil
			ifTrue:
				[self initialStdErr nextPutAll: 'cannot execute ', self commandLine; cr.
				self exitStatus: #cannotExecuteCommandLine.
				"FIXME: Close the OSPipes now, otherwise the image will block on a read"
				self closeStreams.
				[self complete] fork "defer execution so OSPipes stay in place for now"]
			ifFalse:
				[self pid: (procInfo at: 3).
				self handle: (procInfo at: 1).
				mainThread := WindowsThread
						threadID: (procInfo at: 4)
						handle: (procInfo at: 2)
						running: true.
				self threads add: mainThread.
 				self running.
				OSProcess thisOSProcess registerChildProcess: self.
				"FIXME: Close the initial pipe handles. For now, I have not implemented
				passing these to the child, and there is no support yet for nonblocking
				Windows OS pipes. Once those are available, this method needs to change
				to support."
				self closeStreams]].
! !

!OSProcess class methodsFor: 'external process access' stamp: 'dtl 11/8/2000 22:04'!
accessor
	"Answer an OSProcessAccessor for this OS process."

	^ self thisOSProcess processAccessor
! !

!OSProcess class methodsFor: 'sUnit' stamp: 'dtl 10/13/2013 11:02'!
allTestResults
	"Run all available sUnit tests and save the results in a file named
	'OSProcessTestResults-<platform>.txt'. Display results on the Transcript as
	tests are run. Answer a string with the test results."

	"OSProcess allTestResults"

	| resultsFileName writeStream resultString results fs problems result wordSize |
	self listLoadedModules
		detect: [:e | '*OSProcessPlugin*' match: e]
		ifNone: [self notify: 'no OSProcessPlugin loaded'].
	wordSize := (Smalltalk respondsTo: #wordSize)
		ifTrue: [Smalltalk wordSize]
		ifFalse: [4].
	resultsFileName := 'OSProcessTestResults-' , OSProcess platformName , '-' ,
		(wordSize * 8) asString , 'bitImage-' ,
		(OSProcess accessor sizeOfPointer * 8) asString , 'bitHw.txt'.
	[fs := FileStream newFileNamed: resultsFileName.
	problems := false.
	writeStream := WriteStream on: String new.
	writeStream nextPutAll: self authorInitials , ' running OSProcess unit tests ' , TimeStamp now asString; cr.
	writeStream nextPutAll: 'OSProcess platformName => ' , OSProcess platformName; cr.
	writeStream nextPutAll: 'OSProcess platformSubtype => ' , OSProcess platformSubtype; cr.
	writeStream nextPutAll: 'OSProcess osVersion => ' , OSProcess osVersion; cr.
	writeStream nextPutAll: 'OSProcess vmVersion => ' , OSProcess vmVersion; cr.
	writeStream nextPutAll: 'size of C int: OSProcess accessor sizeOfInt ==> ' , OSProcess accessor sizeOfInt asString; cr.
	writeStream nextPutAll: 'size of C pointer: OSProcess accessor sizeOfPointer ==> ' , OSProcess accessor sizeOfPointer asString; cr.
	writeStream nextPutAll: 'OSProcess versionInformation asString => ' , OSProcess versionInformation asString; cr.
	Transcript show: writeStream contents.
	(((Class allSubInstances select: [:class | #'OSProcess-Tests' = class category]) ,
		(Class allSubInstances select: [:class | #'CommandShell-Tests' = class category]))
			reject: [:class | 'Abstract*' match: class name])
		do: [:testCase | 
			writeStream nextPutAll: 'running tests in ' , testCase name; cr.
			Transcript show: 'running tests in ' , testCase name; cr.
			testCase allTestSelectors
				do: [:testSelector |
					OSProcess thisOSProcess stdOut nextPutAll:
						testCase name, '>>',testSelector, String lf; flush.
					resultString := (testSelector
								-> ([result := (testCase selector: testSelector) run.
									result hasPassed
										ifFalse: [problems := true].
									result]
										on: Error
										do: [:ex | problems := true.
											'caught exception ' , ex printString])) printString.
					writeStream nextPutAll: resultString; cr.
					Transcript show: resultString; cr]].
	problems
		ifTrue: [writeStream nextPutAll: 'one or more problems found'; cr.
			Transcript show: 'one or more problems were found, see results file'; cr]
		ifFalse: [writeStream nextPutAll: 'all tests completed without problems'; cr.
			Transcript show: 'all tests completed without problems'; cr].
	results := writeStream contents copyReplaceAll: String cr with: String lf.
	fs nextPutAll: results]
		ensure: [fs close].
	Transcript show: 'tests results have been saved in a file named ''' , resultsFileName , ''''; cr.
	^ results! !

!OSProcess class methodsFor: 'external system access' stamp: 'dtl 1/7/2001 12:32'!
arguments

	"OSProcess arguments"

	^ self thisOSProcess arguments
! !

!OSProcess class methodsFor: 'sUnit' stamp: 'ThierryGoubier 10/11/2013 11:16'!
authorInitials
	"Get the author initials from the system."

	((Smalltalk respondsTo: #metacelloPlatformAttributes)
		and: [ 
			(Smalltalk metacelloPlatformAttributes includes: #'pharo3.x')
				or: [ Smalltalk metacelloPlatformAttributes includes: #'pharo2.x' ] ])
		ifTrue: [ ^ (Smalltalk at: #Author) fullName select: [ :c | c isUppercase ] ].
	^ (Smalltalk at: #Utilities) authorInitials! !

!OSProcess class methodsFor: 'private' stamp: 'dtl 1/7/2001 13:13'!
classForThisOSProcess
	"Answer the class which represents the OS process in which Squeak runs."

	^ ThisOSProcess concreteClass! !

!OSProcess class methodsFor: 'external command processing' stamp: 'dtl 1/7/2001 12:09'!
command: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in the standard C library,
	except that aCommandString runs asynchronously in a child process."

	"OSProcess command: 'ls -l /etc'"

	^ self thisOSProcess command: aCommandString
! !

!OSProcess class methodsFor: 'debugging' stamp: 'dtl 4/2/2005 12:32'!
debugMessage: aString
	"Print aString on standard output. The debug message is prefixed with the
	identity of the process in which the method is being evaluated, and the
	identity of the object which received the message. Useful for debugging
	timing or deadlock problems."

	[self thisOSProcess stdOut
		"The process in which the traced message is being evaluated"
		nextPutAll: Processor activeProcess hash printString, ':';
		"The identity of the object being traced"
		nextPutAll: thisContext sender sender sender receiver hash printString, ':';
		"The debug message"
		nextPutAll: aString asString; nextPut: Character lf; flush]
			on: Error
			do: []! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 1/27/2013 21:48'!
defaultPathString

	^ self useFileSystem
		ifTrue: [(((Smalltalk at: #Path) perform: #workingDirectory) perform: #asFileReference) pathString]
		ifFalse: [(Smalltalk at: #FileDirectory) default pathName]
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 9/5/2013 21:11'!
deleteFileNamed: fileName
	"Delete the file with the given name."

	self useFileSystem
		ifTrue: [ | file |
			file := fileName perform: #asFileReference.
			( {	#ensureDelete "preferred selector" .
				#ensureDeleted "pharo 2, now deprecated in favor of ensureDelete" .
				#delete "fallback (earlier pharo versions?)" }
					detect: [ :e | file respondsTo: e ]
					ifNone: [ self error: 'do not know how to ensureDelete' ] )
				ifNotNilDo: [ :sel | file perform: sel ].
			(file respondsTo: #ensureDelete)
				ifTrue: [ file perform: #ensureDelete ]
				ifFalse: [ file perform: #ensureDeleted ] ]
		ifFalse: [ (Smalltalk at: #FileDirectory) default
				deleteFileNamed: fileName ]
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 1/27/2013 21:49'!
directoryEntryNames: path

	^ self useFileSystem
		ifTrue: [ (path perform: #asFileReference) children collect: [:e | e perform: #basename] ]
		ifFalse: [ ((Smalltalk at: #FileDirectory) on: path) entries collect: [:e | e name] ]
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 1/27/2013 21:49'!
directoryExists: path
	"Answer true if a directory of the given name exists. The given name may
	be either a full path name or a local directory within this directory."

	^ self useFileSystem
		ifTrue: [ (path perform: #asFileReference) exists ]
		ifFalse: [ (Smalltalk at: #FileDirectory) default directoryExists: path ]
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 1/27/2013 21:49'!
fileExists: path
	"Answer true if a file of the given name exists. The given name may be
	either a full path name or a local file within this directory."

	^ self useFileSystem
		ifTrue: [ (path perform: #asFileReference) exists ]
		ifFalse: [ (Smalltalk at: #FileDirectory) default fileExists: path ]
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 3/5/2005 11:32'!
getSystemAttribute: attributeID 
	"After Squeak version 3.6, #getSystemAttribute was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk getSystemAttribute: attributeID]) current getSystemAttribute: attributeID! !

!OSProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 21:52'!
helloStdErr
	"Write a message on the standard error stream of the OS process, normally
	the terminal or window from which Squeak is being run. Most operating
	systems implement stdin, stdout, and stderr in some manner, so this shown
	as an OSProcess example even though the implemention is in my subclasses."

	"OSProcess helloStdErr"

	| this |
	this := self thisOSProcess.
	this stdErr ifNil: [self noAccessorAvailable. ^ nil].
	^ this stdErr nextPutAll: 'Hello stderr'; nextPut: (Character lf); yourself! !

!OSProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 21:53'!
helloWorld
	"Write a message on the standard output stream of the OS process, normally
	the terminal or window from which Squeak is being run. Most operating
	systems implement stdin, stdout, and stderr in some manner, so this shown
	as an OSProcess example even though the implemention is in my subclasses."

	"OSProcess helloWorld"

	| this |
	this := self thisOSProcess.
	this stdOut ifNil: [self noAccessorAvailable. ^ nil].
	^ this stdOut nextPutAll: 'Hello world'; nextPut: Character lf; yourself! !

!OSProcess class methodsFor: 'initialize-release' stamp: 'dtl 10/19/2001 18:44'!
initialize

	"OSProcess initialize"

	UseIOHandle := (Smalltalk hasClassNamed: #IOHandle)
! !

!OSProcess class methodsFor: 'platform identification' stamp: 'dtl 3/5/2005 11:41'!
isNonUnixMac
	"True if the platform is Mac OS prior to OSX"

	"OSProcess isNonUnixMac"

	| numericOsVersion |
	numericOsVersion := self osVersion asInteger ifNil: [0].
	^ (self platformName = 'Mac OS') and: [numericOsVersion < 1000]
! !

!OSProcess class methodsFor: 'platform identification' stamp: 'dtl 3/5/2005 11:42'!
isOS2
	"True if the platform is OS2"
	"FIXME please - What is the correct platform name for OS2?"

	"OSProcess isOS2"

	^ self platformName = 'OS2'
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'ChristerAssmundsson 5/21/2014 10:28'!
isPharo2AndLater
	"True if this image is a pharo of major version 2 or greater"

	Smalltalk
		at: #SystemVersion
		ifPresent: [:cls | ((cls canUnderstand: #type) and: [ cls canUnderstand: #major ])
			ifTrue: [^ cls current type = 'Pharo' and: [ cls current major >= 2 ]]].
	^false
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 9/4/2013 08:35'!
isPharo3AndLater
	"True if this image is a pharo of major version 3 or greater"

	Smalltalk
		at: #SystemVersion
		ifPresent: [:cls | ((cls canUnderstand: #type) and: [ cls canUnderstand: #major ])
			ifTrue: [^ cls current type = 'Pharo' and: [ cls current major >= 3 ]]].
	^false
! !

!OSProcess class methodsFor: 'platform identification' stamp: 'dtl 8/24/2003 09:52'!
isResponsibleForThisPlatform
	"Answer true if this class has responsibilities for the platform on which the
	Squeak VM is currently running."

	^ self subclassResponsibility! !

!OSProcess class methodsFor: 'platform identification' stamp: 'dtl 3/5/2005 11:43'!
isRiscOS
	"True if the platform is RiscOS"

	"OSProcess isRiscOS"

	^ self platformName = 'RiscOS'! !

!OSProcess class methodsFor: 'platform identification' stamp: 'dtl 3/5/2005 11:43'!
isUnix
	"True if the platform is Unix (including Linux, Mac OS X, or other unix-like OS).
	Note: Keep this method in sync with UnixOSProcessPlugin>>isResponsibleForThisPlatform."

	"OSProcess isUnix"

	| numericOsVersion |

	^ (self platformName = 'unix') or:
		[numericOsVersion := self osVersion asInteger ifNil: [0].
		(self platformName = 'Mac OS') and: [numericOsVersion >= 1000]]
! !

!OSProcess class methodsFor: 'platform identification' stamp: 'dtl 3/5/2005 11:46'!
isUnixMac
	"True if the platform is Mac OS on OSX"

	"OSProcess isUnixMac"

	| osVersion numericOsVersion |
	osVersion := self osVersion.
	^ ('darwin*' match: osVersion "Ian's VM")
		or: [numericOsVersion := osVersion asInteger ifNil: [0].
			(self platformName = 'Mac OS') and: [numericOsVersion >= 1000] "John's VM"]
! !

!OSProcess class methodsFor: 'platform identification' stamp: 'dtl 3/5/2005 11:46'!
isWindows
	"True if the platform is an MS Windows OS"

	"OSProcess isWindows"

	^ self platformName = 'Win32'! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 10/27/2007 10:55'!
listLoadedModules
	"After Squeak version 3.6, #listLoadedModules was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk listLoadedModules]) current listLoadedModules! !

!OSProcess class methodsFor: 'utility' stamp: 'dtl 1/24/2013 19:59'!
makeVM
	"Rebuild the virtual machine and plugins, assuming that this Squeak
	is running from a home directory in the appropriate place in the
	source code tree. If the build is successful, save the image and
	restart using the new VM."

	"OSProcess makeVM"

	^ self makeVmIn: OSProcess defaultPathString, OSProcess pathSeparator, 'build'
! !

!OSProcess class methodsFor: 'utility' stamp: 'dtl 1/25/2013 18:58'!
makeVmIn: buildDirectoryPathName
	"Rebuild the virtual machine and plugins in the buildDirectoryPathName
	directory. If the build is successful, save the image and restart using the
	new VM. This assumes that the currently executing VM is either located in,
	or linked to, the buildDirectoryPathName directory."

	"OSProcess makeVmIn: OSProcess defaultPathString, OSProcess pathSeparator, 'build'"

	^ self classForThisOSProcess makeVmIn: buildDirectoryPathName

! !

!OSProcess class methodsFor: 'private' stamp: 'dtl 6/29/2005 14:18'!
noAccessorAvailable

	self notify: 'process accessor module not available'! !

!OSProcess class methodsFor: 'version dependent' stamp: 'ThierryGoubier 9/5/2013 11:13'!
osVersion
	"After Squeak version 3.6, #osVersion was moved to SmalltalkImage. Some
	versions of Pharo move this to OSPlatform and issue deprecation warnings
	about the other usages. Pharo3 deprecated OSPlatform direct access."

	"self osVersion"
	
	self isPharo3AndLater
		ifTrue: [ ^ (Smalltalk perform: #os) version ].
	^ (((Smalltalk hasClassNamed: #OSPlatform)
			and: [(Smalltalk at: #OSPlatform)
					respondsTo: #osVersion])
		ifTrue: [Smalltalk at: #OSPlatform]
		ifFalse: [((Smalltalk classNamed: 'SmalltalkImage')
				ifNil: [^ Smalltalk osVersion]) current]) osVersion! !

!OSProcess class methodsFor: 'version dependent' stamp: 'SeanDeNigris 2/13/2013 09:46'!
pathSeparator

	^ self useFileSystem
		ifTrue: [((Smalltalk at: #DiskStore) current perform: #delimiter) asString]
		ifFalse: [(Smalltalk at: #FileDirectory) slash]
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'ChristerAssmundsson 5/21/2014 10:41'!
platformSubtype 
	"After Squeak version 3.6, #platformSubtype was moved to SmalltalkImage "

	"OSProcess platformSubtype"

	self isPharo2AndLater
		ifTrue: [ ^ (Smalltalk perform: #os) perform: #subtype ].
	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk platformSubtype]) current platformSubtype! !

!OSProcess class methodsFor: 'external system access' stamp: 'dtl 1/7/2001 12:32'!
programName

	"OSProcess programName"

	^ self thisOSProcess programName
! !

!OSProcess class methodsFor: 'utility' stamp: 'dtl 1/7/2001 12:35'!
quitAndRestart
	"Save image, start a new instance from the saved image, and quit this instance.
	This is useful if the VM has been recompiled or if a new pluggable primitive
	has been added."

	"OSProcess quitAndRestart"

	| firstPid this |
	firstPid := OSProcess thisOSProcess pid.
	firstPid ifNil: 
			[self noAccessorAvailable.
			^ nil].
	Smalltalk saveSession.
	"Value of firstPid gets saved in the image"
	this := OSProcess thisOSProcess.
	this pid = firstPid ifTrue: [self squeak ifNotNil: [Smalltalk quitPrimitive]].
	^ this! !

!OSProcess class methodsFor: 'examples' stamp: 'dtl 11/8/2000 23:23'!
readFromStdIn
	"Type some text on the standard input terminal, followed by <return> or <enter>,
	then call this method. Any available input text in the stdin stream will be read.
	This method sets standard input for the Squeak OS process for non-blocking reads
	in order to prevent the Smalltalk image from blocking on the read. After the read,
	standard input is set back to its normal blocking I/O mode.

	Most operating systems implement stdin, stdout, and stderr in some manner, so this
	is shown as an OSProcess example even though the implemention is in my subclasses."

	"OSProcess readFromStdIn inspect"

	| input ioHandle resultString |
	input := self thisOSProcess stdIn.
	input ifNil: [self noAccessorAvailable. ^ nil].
	ioHandle := input ioHandle.
	self accessor setNonBlocking: ioHandle.
	resultString := self thisOSProcess stdIn next: 10000.
	self accessor setBlocking: ioHandle.
	^ resultString
! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 12/15/2013 14:53'!
snapshot: save andQuit: quit
	"After Squeak version 3.6, #snapshot:andQuit: was moved to SmalltalkImage "

	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk snapshot: save andQuit: quit embedded: false]) current snapshot: save andQuit: quit! !

!OSProcess class methodsFor: 'external command processing' stamp: 'dtl 1/7/2001 13:06'!
squeak
	"Start a new instance of Squeak running in a child OS process. The new instance
	will restart from the image file, so it is a clone of this image as it existed at the
	most recent image save. Note that subclasses can implement additional methods of
	starting Squeak images, especially for Unix systems."

	"OSProcess squeak"

	^ self thisOSProcess squeak

! !

!OSProcess class methodsFor: 'utility' stamp: 'dtl 3/5/2005 11:34'!
systemAttributes
	"Answer a Dictionary of all of the system attributes which can be obtained from
	SystemDictionary>>getSystemAttribute."

	"OSProcess systemAttributes"

	| args idx a |
	args := Dictionary new.

	idx := -1.
	[a := self getSystemAttribute: idx.
	a notNil and: [a size > 0]]
		whileTrue: [args at: idx put: a. idx := idx - 1].

	a := self getSystemAttribute: 0.
	(a notNil and: [a size > 0]) ifTrue: [args at: 0 put: a. idx := idx - 1].
	a := self getSystemAttribute: 1.
	(a notNil and: [a size > 0]) ifTrue: [args at: 1 put: a. idx := idx - 1].
	a := self getSystemAttribute: 2.
	(a notNil and: [a size > 0]) ifTrue: [args at: 2 put: a. idx := idx - 1].

	idx := 2.
	[a := self getSystemAttribute: idx.
	a notNil and: [a size > 0]]
		whileTrue: [args at: idx put: a. idx := idx + 1].

	idx := 1001.
	[a := self getSystemAttribute: idx.
	a notNil and: [a size > 0]]
		whileTrue: [args at: idx put: a. idx := idx + 1].

	^ args.

! !

!OSProcess class methodsFor: 'instance creation' stamp: 'dtl 1/7/2001 12:03'!
thisOSProcess
	"Answer the single instance of the class corresponding to the OS process in which
	this Smalltalk image is executing."

	"OSProcess thisOSProcess"

	^ ThisOSProcess thisOSProcess
! !

!OSProcess class methodsFor: 'debugging' stamp: 'dtl 4/2/2005 12:32'!
trace
	"Print the sender's context on standard output. The debug message is
	prefixed with the identity of the process in which the method is being
	evaluated, and the identity of the object which received the message.
	Useful for debugging timing or deadlock problems."

	[self thisOSProcess stdOut
		"The process in which the traced message is being evaluated"
		nextPutAll: Processor activeProcess hash printString, ':';
		"The identity of the object being traced"
		nextPutAll: thisContext sender sender sender receiver hash printString, ':';
		"The method context describing the method being evaluated"
		nextPutAll: thisContext sender sender sender printString;
		nextPut: Character lf;
		flush]
			on: Error
			do: []! !

!OSProcess class methodsFor: 'debugging' stamp: 'dtl 4/2/2005 12:30'!
trace: debugMessageString
	"Print trace information followed by a debug message"

	[self thisOSProcess stdOut
		"The process in which the traced message is being evaluated"
		nextPutAll: Processor activeProcess hash printString, ':';
		"The identity of the object being traced"
		nextPutAll: thisContext sender sender sender receiver hash printString, ':';
		"The method context describing the method being evaluated"
		nextPutAll: thisContext sender sender sender printString, ':';
		nextPutAll: debugMessageString;
		nextPut: Character lf;
		flush]
			on: Error
			do: []! !

!OSProcess class methodsFor: 'version dependent' stamp: 'dtl 1/27/2013 21:48'!
useFileSystem
	"If true use FileSystem, otherwise use traditional FileDirectory. See senders
	for methods with file system dependencies."

	^ Smalltalk hasClassNamed: #FileReference! !

!OSProcess class methodsFor: 'version testing' stamp: 'dtl 8/7/2003 07:28'!
versionInformation

	"OSProcess versionInformation"

	| osppVersion |
	osppVersion := (Smalltalk hasClassNamed: #OSProcessPlugin)
		ifTrue:
			[(Smalltalk at: #OSProcessPlugin) versionInformation]
		ifFalse:
			['(not installed in this image)'].
	^ Array
		with: (self name, ' version ', self versionString)
		with: ((Smalltalk hasClassNamed: #CommandShell)
				ifTrue:
					[((Smalltalk at: #CommandShell) respondsTo: #versionString)
						ifTrue:
							['CommandShell version ', (Smalltalk at: #CommandShell) versionString]
						ifFalse:
							['CommandShell installed (old version, no versionString)']]
				ifFalse:
					['CommandShell is not installed'])
		with:  osppVersion
! !

!OSProcess class methodsFor: 'version testing' stamp: 'dtl 4/12/2014 16:20'!
versionString

	"OSProcess versionString"

	^'4.5.11'! !

!OSProcess class methodsFor: 'version dependent' stamp: 'ChristerAssmundsson 5/21/2014 10:40'!
vmVersion 
	"After Squeak version 3.6, #vmVersion was moved to SmalltalkImage. Pharo3 deprecated vmVersion."

	self isPharo2AndLater
		ifTrue: [ ^ (Smalltalk perform: #vm) version ].
	^ ((Smalltalk classNamed: 'SmalltalkImage')
		ifNil: [^ Smalltalk vmVersion]) current vmVersion! !

!OSProcess class methodsFor: 'external command processing' stamp: 'dtl 1/7/2001 12:11'!
waitForCommand: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in the standard C library.
	The active Smalltalk process waits for completion of the external command process."

	"OSProcess waitForCommand: 'echo sleeping...; sleep 3; echo I just slept for three seconds'"

	^ self thisOSProcess waitForCommand: aCommandString
! !

!OSProcess methodsFor: 'initialize - release' stamp: 'dtl 5/31/1999 13:54'!
initialize

	self subclassResponsibility! !

!OSProcess methodsFor: 'testing' stamp: 'dtl 7/14/2001 21:12'!
isExternalProcess

	^ true! !

!OSProcess methodsFor: 'private' stamp: 'dtl 8/30/2003 10:11'!
noAccessorAvailable

	^ self class noAccessorAvailable
! !

!OSProcess methodsFor: 'accessing' stamp: 'dtl 3/18/2000 13:57'!
pid

	^ pid
! !

!OSProcess methodsFor: 'accessing' stamp: 'dtl 3/18/2000 13:58'!
pid: processIdentifier

	pid := processIdentifier
! !

!OSProcess methodsFor: 'printing' stamp: 'dtl 10/17/1999 21:12'!
printOn: aStream

	super printOn: aStream. aStream nextPutAll: ' with pid '. self pid printOn: aStream! !

OSProcess subclass: #ThisOSProcess
	instanceVariableNames: 'sessionID stdIn stdOut stdErr processAccessor allMyChildren accessProtect'
	classVariableNames: 'ThisInstance'
	poolDictionaries: ''
	category: 'OSProcess-Base'!

!ThisOSProcess commentStamp: '<historical>' prior: 0!
I represent the operating system process in which the Squeak VM is currently running. My subclasses implement system specific features for Unix, Windows, MacOS, or other operating systems by collaborating with corresponding subclasses of OSProcessAccessor to provide primitive access to the external operating system.
!

ThisOSProcess subclass: #MacProcess
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Mac'!

!MacProcess commentStamp: '<historical>' prior: 0!
I represent a Macintosh operating system process, such as the process in which the Squeak VM is currently running. I collaborate with an instance of MacOSProcessAccessor to provide primitive access to the external operating system. My instance variables are maintained as a convenience to allow inspection of an OSProcess. Access to these variables should always be done with my accessor methods, which update the instance variables by querying my MacOSProcessAccessor.!

!MacProcess class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:46'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for
	the Squeak VM running on the current platform."

	^ self isNonUnixMac
! !

!MacProcess methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:18'!
command: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in
	the standard C library, except that aCommandString runs asynchronously
	in a child process. Answer an instance of ExternalMacOSProcess which
	is a proxy for the new MacOS process."

	self notYetImplemented
! !

!MacProcess methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:18'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams
	"Use my processAccessor to call vfork() and execve() and create a
	new child task. Answer a proxy for the new task, an instance of
	ExternalMacOSProcess."

	self notYetImplemented
! !

!MacProcess methodsFor: 'environment' stamp: 'dtl 1/23/2013 21:10'!
getCwd
	"Not yet implemented - answer a reasonable default."

	^ OSProcess defaultPathString! !

!MacProcess methodsFor: 'initialize - release' stamp: 'dtl 10/14/2001 14:03'!
initialize
	"Set my instance variables to reflect the state of the OS process in which 
	this Smalltalk virtual machine is executing."
! !

!MacProcess methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:46'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ self class isNonUnixMac! !

ThisOSProcess subclass: #OS2Process
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-OS2'!

!OS2Process commentStamp: '<historical>' prior: 0!
I represent an OS2 operating system process, such as the process in which the Squeak VM is currently running. I collaborate with an instance of OS2OSProcessAccessor to provide primitive access to the external operating system. My instance variables are maintained as a convenience to allow inspection of an OSProcess. Access to these variables should always be done with my accessor methods, which update the instance variables by querying my OS2OSProcessAccessor.!

!OS2Process class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:47'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for
	the Squeak VM running on the current platform."

	^ self isOS2
! !

!OS2Process methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:18'!
command: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in
	the standard C library, except that aCommandString runs asynchronously
	in a child process. Answer an instance of ExternalOS2OSProcess which
	is a proxy for the new OS2 process."

	self notYetImplemented
! !

!OS2Process methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:18'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams
	"Use my processAccessor to call vfork() and execve() and create a
	new child task. Answer a proxy for the new task, an instance of
	ExternalOS2OSProcess."

	self notYetImplemented
! !

!OS2Process methodsFor: 'environment' stamp: 'dtl 1/23/2013 21:09'!
getCwd
	"Not yet implemented - answer a reasonable default."

	^ OSProcess defaultPathString! !

!OS2Process methodsFor: 'initialize - release' stamp: 'dtl 10/14/2001 14:03'!
initialize
	"Set my instance variables to reflect the state of the OS process in which 
	this Smalltalk virtual machine is executing."
! !

!OS2Process methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:47'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ self class isOS2
! !

ThisOSProcess subclass: #RiscOSProcess
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-RiscOS'!

!RiscOSProcess commentStamp: '<historical>' prior: 0!
I represent an Acorn RiscOS operating system task, such as the task in which the Squeak VM is currently running. I collaborate with a singleton instance of RiscOSProcessAccessor to provide primitive access to the external operating system. My instance variables are maintained as a convenience to allow inspection of a RiscOSProcess. Access to these variables should always be done with my accessor methods, which update the instance variables by querying my RiscOSProcessAccessor.
!

!RiscOSProcess class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:48'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for
	the Squeak VM running on the current platform."

	^ self isRiscOS
! !

!RiscOSProcess class methodsFor: 'utility' stamp: 'dtl 3/11/2001 11:47'!
makeVmIn: buildDirectoryPathName
	"Rebuild the virtual machine and plugins in the buildDirectoryPathName
	directory. If the build is successful, save the image and restart
	using the new VM. This assumes that the currently executing VM is
	either located in, or linked to, the buildDirectoryPathName directory."
! !

!RiscOSProcess methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:18'!
command: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in
	the standard C library, except that aCommandString runs asynchronously
	in a child process. Answer an instance of ExternalRiscOSProcess which
	is a proxy for the new RiscOS task."

	self notYetImplemented
! !

!RiscOSProcess methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:18'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams
	"Use my processAccessor to call vfork() and execve() and create a
	new child task. Answer a proxy for the new task, an instance of
	ExternalRiscOSProcess."

	self notYetImplemented
! !

!RiscOSProcess methodsFor: 'environment' stamp: 'dtl 1/23/2013 21:10'!
getCwd
	"Not yet implemented - answer a reasonable default."

	^ OSProcess defaultPathString! !

!RiscOSProcess methodsFor: 'initialize - release' stamp: 'dtl 10/14/2001 14:04'!
initialize
	"Set my instance variables to reflect the state of the OS process in which 
	this Smalltalk virtual machine is executing."
! !

!RiscOSProcess methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:48'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ self class isRiscOS
! !

!ThisOSProcess class methodsFor: 'concrete subclasses' stamp: 'dtl 3/5/2005 12:08'!
concreteClass

	"ThisOSProcess concreteClass"

	^ self subclasses
		detect: [:c | c isResponsibleForThisPlatform]
		ifNone: [self notify: self printString,
					': No concrete class implementation available for system type ',
					self platformName printString.
				nil]

! !

!ThisOSProcess class methodsFor: 'concrete subclasses' stamp: 'dtl 6/28/2010 21:14'!
concreteClassOrNil

	"ThisOSProcess concreteClassOrNil"

	^ self subclasses
		detect: [:c | c isResponsibleForThisPlatform]
		ifNone: [nil]

! !

!ThisOSProcess class methodsFor: 'initialize-release' stamp: 'dtl 5/20/2011 19:11'!
initialize

	"ThisOSProcess initialize"

	OSProcess initialize. "required to ensure the change sets file in smoothly"
	AttachableFileStream initialize.
	OSProcessAccessor initialize.
	self initializeThisOSProcess.
	Smalltalk addToStartUpList: ThisOSProcess.
	Smalltalk addToShutDownList: ThisOSProcess
! !

!ThisOSProcess class methodsFor: 'initialize-release' stamp: 'dtl 8/25/2010 22:58'!
initializeThisOSProcess
	"Initialize the singleton instance, creating a new instance only if the
	platform type has changed since shutdown (running on a different
	type of computer)."

	(ThisInstance isNil or: [ThisInstance isResponsibleForThisPlatform not])
		ifTrue: [ | cls |
			(cls := self concreteClassOrNil)
				ifNil: [ThisInstance := nil "no concrete class for this platform"]
				ifNotNil: [ThisInstance := cls basicNew]].
	ThisInstance initialize! !

!ThisOSProcess class methodsFor: 'initialize-release' stamp: 'dtl 7/5/2010 11:56'!
initializeThisOSProcessWithNotifier
	"Initialize the singleton instance, creating a new instance only if the
	platform type has changed since shutdown (running on a different
	type of computer). Warn if no implementation is available for this platform."

	(ThisInstance isNil or: [ThisInstance isResponsibleForThisPlatform not])
		ifTrue: [ThisInstance := self concreteClass basicNew].
	ThisInstance initialize! !

!ThisOSProcess class methodsFor: 'instance creation' stamp: 'dtl 11/5/2000 16:10'!
new

	self notify: self name, ': Only one instance of ThisOSProcess or any of its subclasses should exist in the image. Use #thisOSProcess to obtain the singleton instance.'.
	self shouldNotImplement! !

!ThisOSProcess class methodsFor: 'system startup' stamp: 'dtl 3/8/2013 18:42'!
shutDown: quitting
	"Break dependency on my OSProcessAccessor. This is done explicitly at
	shutDown time in order to prevent possible problems when an image is
	restarted on another platform type, in which case a new ThisOSProcess
	instance is created and the old instance could still have an unwanted
	dependency on an OSProcessAccessor."

	quitting ifTrue: [OSProcess accessor
				ifNotNilDo: [:acc | acc breakDependents]]! !

!ThisOSProcess class methodsFor: 'system startup' stamp: 'dtl 9/3/2010 16:41'!
startUp: resuming
	"Initialize my singleton instance, and the singleton instance of my
	OSProcessAccessor. On Unix, set the signal handler in my process
	accessor to respond to externally generated sigchld signals. This
	must be done after each image restart in order to call a primitive
	which informs the VM of the identity of the semaphore to signal.
	When not running on a Unix system, the primitive fails and the
	method has no effect. Notify dependents of the singleton instance
	if the image has restarted in a different OS process (this is not the
	case when #startUp is called after a simple image save). The
	notification is done in the initialization of my OSProcessAccessor."

	| aio |
	(aio := Smalltalk at: #AioEventHandler)
		ifNotNil: [ aio startUp: resuming ].
	self initializeThisOSProcess! !

!ThisOSProcess class methodsFor: 'instance creation' stamp: 'dtl 10/1/2006 08:38'!
thisOSProcess
	"Answer a single instance of the class corresponding to the OS process in 
	which this Smalltalk image is executing."

	"ThisOSProcess thisOSProcess"

	^ ThisInstance! !

!ThisOSProcess methodsFor: 'comparing' stamp: 'dtl 11/15/2013 10:30'!
= processProxy
	"True if this is a proxy that represents the same OS process as processProxy"

	^ ((processProxy isKindOf: ThisOSProcess)
			and: [pid = processProxy pid])
		and: [sessionID = processProxy sessionID]! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 14:55'!
activeChildren
	"Answer child processes which are currently believed to be running."

	"OSProcess thisOSProcess activeChildren"

	| children |
	accessProtect
		critical: [ children := allMyChildren select: [ :p | p isRunning ] ].
	^children
! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 11:07'!
allMyChildren

	^ allMyChildren! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/5/2003 21:59'!
canConnectToXDisplay: xDisplayName
	"Open and close a connection to displayName. It the connection was successfully
	opened, answer true; otherwise false. This is intended to check for the ability
	to open an X display prior to actually making the attempt."

	"self thisOSProcess canConnectToXDisplay: ':0.0' "
	"self thisOSProcess canConnectToXDisplay: ':1' "
	"self thisOSProcess canConnectToXDisplay: 'bogus:0' " "<-make sure network is running first!!"

	(xDisplayName isKindOf: String) ifFalse: [^ false].
	^ self processAccessor primCanConnectToXDisplay: xDisplayName
! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 14:59'!
childPids

	^ allMyChildren asArray asArray collect: [ :p | p pid ]
! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/6/2003 06:35'!
closeXDisplay
	"Become headless by closing the X session. All subsequent processing should involve
	no further display interaction."

	"self thisOSProcess closeXDisplay"

	| proc |
	proc := self processAccessor primKillDisplay.
	proc ifNil: [self noAccessorAvailable].
	^ proc
! !

!ThisOSProcess methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:15'!
command: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in
	the standard C library, except that aCommandString runs asynchronously
	in a child process. Answer an instance of ExternalMacOSProcess which
	is a proxy for the new OS process."

	self subclassResponsibility! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/5/2003 22:05'!
currentXDisplayName

	"self thisOSProcess currentXDisplayName"

	^ self processAccessor primGetXDisplayName! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/6/2003 06:35'!
decapitate
	"Become headless by closing the X session. All subsequent processing should involve
	no further display interaction."

	"self thisOSProcess decapitate"

	^ self closeXDisplay
! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 14:52'!
discardExitedChildren
	"Remove entries for completed child processed from dictionary."

	self updateAllMyChildren.
	accessProtect
		critical: [ allMyChildren := allMyChildren select: [ :p | p isComplete not ] ].
	^ allMyChildren
! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/6/2003 06:30'!
disconnectXDisplay
	"Disconnect from the X server, but do not close the existing Squeak window. A new
	display medium must be opened before further interaction with the display."

	"self thisOSProcess disconnectXDisplay"

	| proc |
	proc := self processAccessor primDisconnectDisplay.
	proc ifNil: [self noAccessorAvailable].
	^ proc
! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/5/2003 22:12'!
displayOnXServer: xDisplayName
	"Check if it is possible to open a display on the X server identified by
	xDisplayName. If so, close the current X display and reopen it on
	the new server. On success, answer the previous display name. On
	failure, answer a string with an error message. This method is expected
	to be called by a web server or other application which may wish to
	make use of the result string."

	"self thisOSProcess displayOnXServer: ':0.0' "
	"self thisOSProcess displayOnXServer: 'unix:0' "
	"self thisOSProcess displayOnXServer: ':1' "
	"self thisOSProcess displayOnXServer: 'noSuchMachine'"
	"self thisOSProcess displayOnXServer: 'noSuchMachine:0'" "<-make sure network is running first!!"

	| previousDisplayName |
	(xDisplayName isKindOf: String)
		ifFalse: [^ 'expected display name string'].
	previousDisplayName := self currentXDisplayName.
	(self canConnectToXDisplay: xDisplayName)
		ifTrue:
			[self decapitate.
			self setXDisplayName: xDisplayName.
			self recapitate.
			^ previousDisplayName]
		ifFalse:
			[^ 'cannot connect to display ', xDisplayName]
! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 14:54'!
exitedChildren
	"Answer child processes which have exited and are no longer running."

	"OSProcess thisOSProcess exitedChildren"

	| children |
	accessProtect
		critical: [ children := allMyChildren select: [ :p | p isComplete ]].
	^ children
! !

!ThisOSProcess methodsFor: 'finalization' stamp: 'dtl 11/4/2000 15:49'!
finalize

	processAccessor
		ifNotNil: 
			[processAccessor removeDependent: self.
			processAccessor finalize].
	processAccessor := nil! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 9/1/2003 13:15'!
flushXDisplay
	"Synchronize output to the X display."

	"self thisOSProcess flushXDisplay"

	^ self processAccessor primFlushXDisplay

! !

!ThisOSProcess methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:16'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams
	"Use my processAccessor to call vfork() and execve() and create a
	new child task. Answer a proxy for the new task, an instance of
	ExternalOSProcess."

	self subclassResponsibility! !

!ThisOSProcess methodsFor: 'environment' stamp: 'dtl 10/18/2001 20:10'!
getCwd

	self subclassResponsibility! !

!ThisOSProcess methodsFor: 'private - IOHandle' stamp: 'dtl 9/25/2005 16:23'!
handleFromAccessor: aByteArrayOrIOAccessor

	UseIOHandle
		ifTrue: [aByteArrayOrIOAccessor isNil
			ifTrue: [^ nil]
			ifFalse: [^ aByteArrayOrIOAccessor asSQFileStruct]]
		ifFalse: [^ aByteArrayOrIOAccessor]
! !

!ThisOSProcess methodsFor: 'private - IOHandle' stamp: 'dtl 9/25/2005 16:22'!
handleFromFileStream: aFileStream

	UseIOHandle
		ifTrue: [^ aFileStream ioHandle asSQFileStruct]
		ifFalse: [^ aFileStream fileID]
! !

!ThisOSProcess methodsFor: 'comparing' stamp: 'dtl 6/27/2010 22:38'!
hash

	^ (pid ifNil: [self identityHash])
		+ (sessionID
				ifNil: [0]
				ifNotNil: [sessionID sum])! !

!ThisOSProcess methodsFor: 'initialize - release' stamp: 'dtl 7/26/2010 13:51'!
initialize
	"Set my instance variables to reflect the state of the OS process in which 
	this Smalltalk virtual machine is. executing."

	accessProtect := Semaphore forMutualExclusion.
	self initializeAllMyChildren.
	processAccessor ifNotNil:
		[processAccessor breakDependents.
		processAccessor := nil].
	((self processAccessor notNil
		and: [processAccessor canAccessSystem])
			and: [pid ~= processAccessor primGetPid])
		ifTrue: [self resetChildProcessDictionary]
! !

!ThisOSProcess methodsFor: 'initialize - release' stamp: 'dtl 7/4/2010 11:18'!
initializeAllMyChildren
	"Use a Dictionary if process identifiers are unique. On Windows, the
	process ID is not unique, so use an OrderedCollection instead."

	self subclassResponsibility
! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/24/2003 10:12'!
isConnectedToXServer
	"Answer true if VM is currently connected to an X server."

	"self thisOSProcess isConnectedToXServer"

	^ self processAccessor primIsConnectedToXServer

! !

!ThisOSProcess methodsFor: 'platform identification' stamp: 'dtl 10/10/2001 21:24'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ self subclassResponsibility! !

!ThisOSProcess methodsFor: 'private - IOHandle' stamp: 'dtl 9/25/2005 16:23'!
isStdErr: anIOHandle
	"Answer true if anIOHandle represents stderr."

	| realHandle |
	anIOHandle ifNil: [^ false].
	realHandle := self processAccessor getStdErrHandle.
	realHandle ifNil: [^ false].
	UseIOHandle
		ifTrue: [^ anIOHandle handle = realHandle handle]
		ifFalse: [^ anIOHandle = realHandle]
! !

!ThisOSProcess methodsFor: 'private - IOHandle' stamp: 'dtl 9/25/2005 16:23'!
isStdIn: anIOHandle
	"Answer true if anIOHandle represents stdin."

	| realHandle |
	anIOHandle ifNil: [^ false].
	realHandle := self processAccessor getStdInHandle.
	realHandle ifNil: [^ false].
	UseIOHandle
		ifTrue: [^ anIOHandle handle = realHandle handle]
		ifFalse: [^ anIOHandle = realHandle]
! !

!ThisOSProcess methodsFor: 'private - IOHandle' stamp: 'dtl 9/25/2005 16:23'!
isStdOut: anIOHandle
	"Answer true if anIOHandle represents stdout."

	| realHandle |
	anIOHandle ifNil: [^ false].
	realHandle := self processAccessor getStdOutHandle.
	realHandle ifNil: [^ false].
	UseIOHandle
		ifTrue: [^ anIOHandle handle = realHandle handle]
		ifFalse: [^ anIOHandle = realHandle]
! !

!ThisOSProcess methodsFor: 'updating' stamp: 'dtl 2/26/2002 08:37'!
needsRefresh
	"Answer true if the sessionID variable is out of date with respect to the running
	OS Process. Subclasses should provide implementation, answer true as default."

	^ true! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/6/2003 06:36'!
openXDisplay
	"Restore headful display opening the X session."

	"self thisOSProcess closeXDisplay. (Delay forSeconds: 5) wait. self thisOSProcess openXDisplay."

	| proc |
	proc := self processAccessor primOpenXDisplay.
	proc ifNil: [self noAccessorAvailable].
	^ proc
! !

!ThisOSProcess methodsFor: 'accessing' stamp: 'dtl 2/28/2002 13:30'!
processAccessor

	| a |
	processAccessor
		ifNil: 
			[a := OSProcessAccessor forThisOSProcess.
			a isResponsibleForThisPlatform ifTrue:
				[processAccessor := a.
				processAccessor addDependent: self]].
	^ processAccessor! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/6/2003 06:34'!
recapitate
	"Restore headful display opening the X session."

	"self thisOSProcess decapitate. (Delay forSeconds: 5) wait. self thisOSProcess recapitate."

	^ self openXDisplay
! !

!ThisOSProcess methodsFor: 'updating' stamp: 'dtl 2/26/2002 08:32'!
refreshFromProcessAccessor
	"Set my instance variables to reflect the state of the OS process in which this Smalltalk
	virtual machine is executing."

	self subclassResponsibility! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 7/4/2010 16:32'!
registerChildProcess: anOSProcess

	self subclassResponsibility! !

!ThisOSProcess methodsFor: 'updating' stamp: 'dtl 7/4/2010 14:43'!
resetChildProcessDictionary
	"Forget all the entries in the allMyChildren dictionary. This method may be called
	when a new session is started, since the child processes of the previous session are
	probably no longer of any interest."

	self updateAllMyChildren. "Ensure contents are up to date before releasing them."
	self initializeAllMyChildren
! !

!ThisOSProcess methodsFor: 'accessing' stamp: 'dtl 9/26/2005 20:04'!
sessionID

	^ sessionID! !

!ThisOSProcess methodsFor: 'private' stamp: 'dtl 9/25/2005 13:28'!
setStdErr
	"If stdErr is nil, then set it. If not nil, check to see if it is has a valid connection to
	stderr. If not valid, then replace it, otherwise answer the existing valid stream.
	Obscure bug warning: If a valid AttachableFileStream on stderr is garbage collected,
	then stderr will be closed. It is advisable (but not necessary) to treat the stream
	on stderr as a singleton, but in any case, any extra instances attached to stderr
	should not be allowed to be garbage collected."

	| stdErrHandle |
	stdErr ifNotNil:
			[(self isStdErr: stdErr ioHandle) ifTrue: [^ stdErr]].
	stdErrHandle := self processAccessor getStdErrHandle.
	stdErrHandle ifNotNil:
			[stdErr := AttachableFileStream name: 'stderr' attachTo: stdErrHandle writable: true].
	^ stdErr
! !

!ThisOSProcess methodsFor: 'private' stamp: 'dtl 9/25/2005 13:28'!
setStdIn
	"If stdIn is nil, then set it. If not nil, check to see if it is has a valid connection to
	stdin. If not valid, then replace it, otherwise answer the existing valid stream.
	Obscure bug warning: If a valid AttachableFileStream on stdin is garbage collected,
	then stdin will be closed. It is advisable (but not necessary) to treat the stream
	on stdin as a singleton, but in any case, any extra instances attached to stdin
	should not be allowed to be garbage collected."

	| stdInHandle |
	stdIn ifNotNil:
			[(self isStdIn: stdIn ioHandle) ifTrue: [^ stdIn]].
	stdInHandle := self processAccessor getStdInHandle.
	stdInHandle ifNotNil:
			[stdIn := AttachableFileStream name: 'stdin' attachTo: stdInHandle writable: false].
	^ stdIn

! !

!ThisOSProcess methodsFor: 'private' stamp: 'dtl 9/25/2005 13:29'!
setStdOut
	"If stdOut is nil, then set it. If not nil, check to see if it is has a valid connection to
	stdout. If not valid, then replace it, otherwise answer the existing valid stream.
	Obscure bug warning: If a valid AttachableFileStream on stdout is garbage collected,
	then stdout will be closed. It is advisable (but not necessary) to treat the stream
	on stdout as a singleton, but in any case, any extra instances attached to stdout
	should not be allowed to be garbage collected."

	| stdOutHandle |
	stdOut ifNotNil:
			[(self isStdOut: stdOut ioHandle) ifTrue: [^ stdOut]].
	stdOutHandle := self processAccessor getStdOutHandle.
	stdOutHandle ifNotNil:
			[stdOut := AttachableFileStream name: 'stdout' attachTo: stdOutHandle writable: true].
	^ stdOut
! !

!ThisOSProcess methodsFor: 'display management' stamp: 'dtl 8/5/2003 22:16'!
setXDisplayName: xDisplayName
	"Set X display name for use by the next call to recapitate"

	"self thisOSProcess setXDisplayName: 'unix:0' "
	"self thisOSProcess setXDisplayName: ':1' "

	self processAccessor primSetXDisplayName: xDisplayName

! !

!ThisOSProcess methodsFor: 'accessing' stamp: 'dtl 9/25/2005 13:30'!
stdErr

	^ stdErr! !

!ThisOSProcess methodsFor: 'accessing' stamp: 'dtl 9/25/2005 13:30'!
stdIn

	^ stdIn! !

!ThisOSProcess methodsFor: 'accessing' stamp: 'dtl 9/25/2005 13:31'!
stdOut

	^ stdOut! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 7/5/2010 11:02'!
unregisterChildProcess: anOSProcess

	self subclassResponsibility
! !

!ThisOSProcess methodsFor: 'updating' stamp: 'dtl 10/15/2001 21:27'!
update: aParameter

	aParameter == #invalidProcessAccessor ifTrue: [processAccessor := nil].
	^ super update: aParameter! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 2/25/2002 23:23'!
updateActiveChildren
	"Test each active child for its completion status and update runState and exitStatus
	accordingly. This method may be called when a semaphore is set indicating that
	some child OSProcess has died. A better approach might be to use an event queue
	for death of child events; however, until event queues are part of Squeak image,
	this polling mechanism is sufficient."

	self activeChildren do:
		[:child | child update: #runState]
! !

!ThisOSProcess methodsFor: 'child process management' stamp: 'dtl 7/26/2010 11:34'!
updateAllMyChildren
	"Test each child to make sure that it is still accessible. If the child is believed to be
	running, check to see if it has exited, and update runState and exitStatus accordingly."

	| children |
	accessProtect critical: [children := allMyChildren asArray].
	children do:
		[:child |
		child update: #accessibility.
		child isRunning ifTrue: [child update: #runState]]
! !

ThisOSProcess subclass: #UnixProcess
	instanceVariableNames: 'ppid pthread path programName arguments environment'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Unix'!

!UnixProcess commentStamp: '<historical>' prior: 0!
I represent the Unix operating system process in which this Squeak session is running. I collaborate with an instance of UnixOSProcessAccessor to provide access to the external operating system. My instance variables are updated when my process accessor changes, allowing them to be monitored with a Smalltalk inspector.
!

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 3/10/2001 09:52'!
arguments

	"UnixProcess arguments"

	^ self thisOSProcess arguments
! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 2/24/2013 10:26'!
backgroundProcessInHeadlessSqueak
	"Demonstrate running a ''batch job'' in a low priority background Squeak,
	with output written to an OSPipe. This Squeak image reads data from the
	pipe and displays it on the Transcript. The background Squeak will write
	enough data to block an OSPipe, so the foreground Squeak loops while
	reading available data up to end of file.

	Open a Transcript, then '' inspect it '' on this method. Process runs
	indefinitely, so terminate the process to end demo."

	"UnixProcess backgroundProcessInHeadlessSqueak"

	^ [[| pipe backgroundJob ws s | 
	pipe := OSPipe new.
	backgroundJob := OSProcess thisOSProcess
	forkHeadlessSqueakAndDoThenQuit: [| beers | 
		OSProcess accessor nice: 1. "lower priority of background OS process"
		beers := [:i | (i < 1 ifTrue: ['no more'] ifFalse: [i asString])
					, ' bottle' , (i = 1 ifTrue: [''] ifFalse: ['s'])].
		(99 to: 1 by: -1) do: [:count | pipe nextPutAll:
			(beers value: count) , ' of beer on the wall, '
			, (beers value: count) , ' of beer' , String cr
			, 'take one down and pass it around, '
			, (beers value: count - 1) , ' of beer on the wall'; cr].
		pipe close].
	WorldState addDeferredUIMessage:
		[Transcript show: backgroundJob asString , ' started'; cr].
	pipe closeWriter. "don't need writer end, close it before the #upToEnd"
	"pipe writer blocks when pipe full, so we need to loop while reading to end "
	ws := WriteStream on: String new.
	[backgroundJob isComplete] whileFalse:
		[(Delay forMilliseconds: 200) wait.
		(s := pipe upToEnd) ifNotNil: [ws nextPutAll: s]].
	pipe close.
	WorldState addDeferredUIMessage:
		[Transcript show: backgroundJob asString
			, ' completed, display results in 2 seconds'; cr].
	(Delay forSeconds: 2) wait.
	WorldState addDeferredUIMessage:
		[Transcript show: ws contents.
		Transcript cr; show: 'delay 5 seconds before forking next Squeak job'; cr].
	(Delay forSeconds: 5) wait] repeat]
		forkAt: Processor userBackgroundPriority! !

!UnixProcess class methodsFor: 'X display - deprecated' stamp: 'dtl 8/5/2003 22:10'!
canConnectToXDisplay: xDisplayName
	"Deprecated. See ThisOSProcess>canConnectToXDisplay:"

	"self canConnectToXDisplay: ':0.0' "

	^ self thisOSProcess canConnectToXDisplay: xDisplayName
! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 04:17'!
catAFile
	"Copy contents of a file to standard output. This demonstrates reassigning stdin to an open FileStream."

	"UnixProcess catAFile"

	| in proc |
	in := FileStream readOnlyFileNamed: '/etc/hosts'.
	proc := self forkJob: '/bin/cat'
			arguments: nil
			environment: nil
			descriptors: (Array with: in with: nil with: nil).
	in close.
	proc ifNil: [self noAccessorAvailable].
	^ proc! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 04:28'!
catFromFileToFiles
	"Copy contents of a file to another file, with any error messages going 
	to a third file."

	"UnixProcess catFromFileToFiles"

	| in out err proc |
	in := FileStream readOnlyFileNamed: '/etc/hosts'.
	out := FileStream newFileNamed: '/tmp/deleteMe.out'.
	err := FileStream newFileNamed: '/tmp/deleteMe.err'.
	proc := UnixProcess
				forkJob: '/bin/cat'
				arguments: nil
				environment: nil
				descriptors: (Array
						with: in
						with: out
						with: err).
	in close.
	out close.
	err close.
	proc ifNil: [self noAccessorAvailable].
	^ proc! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 11/14/2000 00:07'!
clientServerDemo
	"Start a new headless instance of Squeak running in a child OS process. The new
	instance is a nearly identical copy of its parent, resuming execution at the same
	point as the parent. The child process will run a TCP server process in the headless
	image, then exit. Test results are written to the Transcript.

	Answer '127 0 0 1' in response to the host address dialog."

	"UnixProcess clientServerDemo"

	| proc |
	proc := self forkHeadlessSqueakAndDoThenQuit: [Socket remoteTestServerTCP].
	proc ifNotNil: [Socket remoteTestClientTCP].
	^ proc

! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 1/7/2001 12:55'!
clientServerDemo2
	"Start a new instance of Squeak running in a child OS process. The new 
	instance is a nearly identical copy of its parent, resuming execution at
	the same point as the parent. The child process will run a TCP server
	process, then exit. Test results are written to the Transcript. "

	"UnixProcess clientServerDemo2"

	| remotePort remoteHost serverBlock socket buffer n bytesToSend sendBuf
	receiveBuf done bytesSent bytesReceived packetsSent packetsReceived t |

	remotePort := 54321.
	remoteHost := 'localhost'.
	serverBlock := 
			[Transcript clear; cr; show: 'This is the server Squeak session'; cr.
			Socket initializeNetworkIfFail: [^ nil].
			socket := Socket newTCP.
			socket listenOn: remotePort.
			buffer := String new: 4000.
			socket waitForConnectionUntil: Socket standardDeadline.
			[socket isConnected]
				whileTrue: [socket dataAvailable
						ifTrue: 
							[n := socket receiveDataInto: buffer.
							socket sendData: buffer count: n]].
			socket closeAndDestroy].

	(UnixProcess forkSqueakAndDoThenQuit: serverBlock)
		ifNil: [self noAccessorAvailable. ^ nil].
	Transcript cr; show: 'This is the client Squeak session'; cr.
	Transcript show: 'starting client/server TCP test'; cr.
	Transcript show: 'initializing network ... '.
	Socket initializeNetworkIfFail: [^ Transcript show: 'failed'].
	Transcript show: 'ok'; cr.
	socket := Socket newTCP.
	socket connectTo: (NetNameResolver addressForName: remoteHost) port: remotePort.
	socket waitForConnectionUntil: Socket standardDeadline.
	Transcript show: 'client endpoint created'; cr.
	bytesToSend := 1000000.
	sendBuf := String new: 4000 withAll: $x.
	receiveBuf := String new: 50000.
	done := false.
	bytesSent := bytesReceived := packetsSent := packetsReceived := 0.
	t := Time
				millisecondsToRun: 
					[[done]
						whileFalse: 
							[(socket sendDone and: [bytesSent < bytesToSend])
								ifTrue: 
									[packetsSent := packetsSent + 1.
									bytesSent := bytesSent + (socket sendData: sendBuf)].
							socket dataAvailable
								ifTrue: 
									[packetsReceived := packetsReceived + 1.
									bytesReceived := bytesReceived + (socket receiveDataInto: receiveBuf)].
							done := bytesSent >= bytesToSend].
					[bytesReceived < bytesToSend]
						whileTrue: [socket dataAvailable
								ifTrue: 
									[packetsReceived := packetsReceived + 1.
									bytesReceived := bytesReceived + (socket receiveDataInto: receiveBuf)]]].
	socket closeAndDestroy.
	Transcript show: 'remoteClient TCP test done; time = ' , t printString; cr.
	Transcript show: packetsSent printString , ' packets, ' , bytesSent printString ,
		' bytes sent (' , (bytesSent * 1000 // t) printString , ' bytes/sec)'; cr.
	Transcript show: packetsReceived printString , ' packets, ' , bytesReceived printString ,
		' bytes received (' , (bytesReceived * 1000 // t) printString , ' bytes/sec)'; cr.
	^ bytesReceived! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 9/9/2000 16:00'!
cloneSqueak
	"Start a new instance of Squeak running in a child OS process. The new instance is a
	nearly identical copy of its parent, resuming execution at the same point as the
	parent, and differentiated only by the return value of this method."

	"UnixProcess cloneSqueak"

	| proc |
	proc := self forkSqueak.
	proc ifNil: [self noAccessorAvailable].
	^ proc
! !

!UnixProcess class methodsFor: 'X display - deprecated' stamp: 'dtl 8/5/2003 22:10'!
currentXDisplayName
	"Deprecated. See ThisOSProcess>>currentXDisplayName"

	"self currentXDisplayName"

	^ self thisOSProcess currentXDisplayName
! !

!UnixProcess class methodsFor: 'X display - deprecated' stamp: 'dtl 8/5/2003 22:09'!
decapitate
	"Deprecated. See ThisOSProcess>>decapitate"

	"self decapitate"

	^ OSProcess thisOSProcess decapitate! !

!UnixProcess class methodsFor: 'X display - deprecated' stamp: 'dtl 8/5/2003 22:18'!
displayOnXServer: xDisplayName
	"Deprecated. See ThisOSProcess>>displayOnXServer"

	"self displayOnXServer: ':0.0' "
	"self displayOnXServer: 'unix:0' "
	"self displayOnXServer: ':1' "
	"self displayOnXServer: 'noSuchMachine'"
	"self displayOnXServer: 'noSuchMachine:0'" "<-make sure network is running first!!"

	^ OSProcess thisOSProcess displayOnXServer: xDisplayName! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 04:31'!
eightLeafSqueakTree
	"Clone this squeak three times, resulting in a total of (2 raisedTo: 3) nearly identical
	squeaks. Have a look at the pidArray inspectors and to the debug messages on stdout
	in order to see what is going on. The tree of processes looks like this:

	111
	 +--------+--------+
	 |		 |		 |
	011		101		110
	 +---+	 |
	 |	 |	100
	010	001
		 |
		000
	"

	"UnixProcess eightLeafSqueakTree inspect"

	| depth this pidArray debugString |
	depth := 3.
	this := OSProcess thisOSProcess.
	this stdOut ifNil: [self noAccessorAvailable. ^ nil].
	pidArray := Array new: depth.
	(1 to: depth) do: [ :e | | p pid |
		p := this forkSqueak.
		pid := (p == this) ifTrue: [0] ifFalse: [p pid].	"Use Unix fork(2) convention"
		pidArray at: e put: pid].
	debugString := 'pid ', (this pid printString), ' ppid ', (this ppid printString), ' ',
		(pidArray printString), (Character lf asString).
	this stdOut nextPutAll: debugString.
	^ pidArray! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 12/27/2000 17:18'!
env
	"Note: The #environment selector has special meaning for classes, so use #env."

	"UnixProcess env"

	^ self thisOSProcess environment
! !

!UnixProcess class methodsFor: 'child process creation' stamp: 'dtl 12/27/2000 16:57'!
forkHeadlessSqueakAndDo: aBlock 
	"Start a new instance of Squeak running in a child OS process, and  
	execute aBlock in the child instance. The new instance is a clone of 
	this image, but without a connection to the X display. The child instance 
	executes aBlock, which hopefully does not involve interaction with the 
	X display; and the parent continues normally.  
	  
	The child should not depend on using existing connections to external  
	resources. For example, the child may lose its connections to stdin, 
	stdout, and stderr after its parent exits."

	"UnixProcess forkHeadlessSqueakAndDo: [UnixProcess helloWorld]"

	^ self thisOSProcess forkHeadlessSqueakAndDo: aBlock

! !

!UnixProcess class methodsFor: 'child process creation' stamp: 'dtl 12/27/2000 17:00'!
forkHeadlessSqueakAndDoThenQuit: aBlock 
	"Start a new instance of Squeak running in a child OS process, and 
	execute aBlock in the child instance. The new instance is a clone of
	this image, but without a connection to the X display. The child instance
	executes aBlock, which hopefully does not involve interaction with the
	X display; and the parent continues normally.

	The child should not depend on using existing connections to external 
	resources. For example, the child may lose its connections to stdin, stdout,
	and stderr after its parent exits."

	"UnixProcess forkHeadlessSqueakAndDoThenQuit: [UnixProcess helloWorld]"

	^ self thisOSProcess forkHeadlessSqueakAndDoThenQuit: aBlock
! !

!UnixProcess class methodsFor: 'child process creation' stamp: 'dtl 12/27/2000 16:48'!
forkJob: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams 
	"Call Unix vfork() and execve() to create a child process, and answer the 
	child process. Delegate this to the singleton OSProcess>>thisOSProcess."

	^ self thisOSProcess
		forkJob: executableFile
		arguments: arrayOfStrings
		environment: stringDictionary
		descriptors: arrayOf3Streams! !

!UnixProcess class methodsFor: 'child process creation' stamp: 'dtl 7/5/2000 07:19'!
forkSqueak
	"Start a new instance of Squeak running in a child OS process. The new instance is a
	clone of this image except for the return value of this method. It does not reload the
	image file from disk.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	"UnixProcess forkSqueak"

	^ self thisOSProcess forkSqueak.
! !

!UnixProcess class methodsFor: 'child process creation' stamp: 'dtl 10/8/2001 20:40'!
forkSqueakAndDo: aBlock
	"Start a new instance of Squeak running in a child OS process. The new instance is a
	clone of this image except for the return value of this method. It does not reload the
	image file from disk. The child image evaluates aBlock.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	"UnixProcess forkSqueakAndDo:
		[Object inform: 'Hi, I am the child Squeak process.']"

	^ self thisOSProcess forkSqueakAndDo: aBlock
! !

!UnixProcess class methodsFor: 'child process creation' stamp: 'dtl 10/8/2001 20:41'!
forkSqueakAndDoThenQuit: aBlock
	"Start a new instance of Squeak running in a child OS process. The new instance is a
	clone of this image except for the return value of this method. It does not reload the
	image file from disk. The child image evaluates aBlock.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	"UnixProcess forkSqueakAndDoThenQuit:
		[Object inform: 'Hi, I am the child Squeak process. Click OK to exit the child Squeak.']"

	^ self thisOSProcess forkSqueakAndDoThenQuit: aBlock
! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 05:11'!
headlessChild
	"Start a new instance of Squeak running in a child OS process. The new
	instance is a nearly identical copy of its parent, resuming execution at the
	same point as the parent, and differentiated only by the return value of
	this method. The child squeak will write a message to standard output,
	then exit."

	"UnixProcess headlessChild"

	| this childBlock |
	this := OSProcess thisOSProcess.
	childBlock :=
		[this stdOut nextPutAll: 'hello world from child process '.
		this pid printOn: OSProcess thisOSProcess stdOut.
		this stdOut nextPut: Character lf ].
	^ self forkHeadlessSqueakAndDoThenQuit: childBlock

! !

!UnixProcess class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:49'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for
	the Squeak VM running on the current platform."

	^ self isUnix
! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 04:34'!
listDirectory
	"Execute a simple command, sending output to standard output."

	"UnixProcess listDirectory"

	| proc |
	proc := self
			forkJob: '/bin/ls'
			arguments: nil
			environment: nil
			descriptors: nil.
	proc ifNil: [self noAccessorAvailable].
	^ proc! !

!UnixProcess class methodsFor: 'utility - VM building' stamp: 'dtl 10/19/2001 13:31'!
makeVmIn: buildDirectoryPathName
	"Rebuild the virtual machine and plugins in the buildDirectoryPathName
	directory. If the build is successful, save the image and restart using the
	new VM. This assumes that the currently executing VM is either located in,
	or linked to, the buildDirectoryPathName directory."

	"UnixProcess makeVmIn: '/usr/local/squeak/Squeak-2.8/build'"
	"UnixProcess makeVmIn: nil"

	^ self makeVmIn: (buildDirectoryPathName ifNil: ['']) interactive: true

! !

!UnixProcess class methodsFor: 'utility - VM building' stamp: 'dtl 10/19/2001 11:41'!
makeVmIn: buildDirectoryPathName interactive: aBoolean
	"Rebuild the virtual machine and plugins in the buildDirectoryPathName
	directory. If the build is successful, save the image and restart using the
	new VM. This assumes that the currently executing VM is either located in,
	or linked to, the buildDirectoryPathName directory. If aBoolean is true,
	display interactive dialogs, otherwise output goes only to the Transcript."

	"UnixProcess makeVmIn: '/usr/local/squeak/Squeak-2.8/build' interactive: true"

	| commandString result |
	commandString := 'cd ', buildDirectoryPathName, '; make'.
	Transcript cr; show: 'running external command "', commandString, '"'; cr.
	result := self waitForCommandOutputArray: commandString.
	((result at: 3) == 0 and: [(result at: 2) isEmpty])
		ifTrue:
			[Transcript show: 'make completed successfully, restarting VM'; cr.
			aBoolean ifTrue:
				[self inform: 'make completed successfully, restarting VM'].
			self quitAndRestart]
		ifFalse:
			[Transcript show: (result at: 2); cr.
			Transcript show: 'make did not succeed, VM will not be restarted'; cr.
			aBoolean ifTrue:
				[self inform: 'make did not succeed, VM will not be restarted'.
				self inform: (result at: 2)]].
	^ result
! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 12/27/2000 17:16'!
path
	"UnixProcess path"

	^ self thisOSProcess path
! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 3/10/2001 09:53'!
programName

	"UnixProcess programName"

	^ self thisOSProcess programName
! !

!UnixProcess class methodsFor: 'X display - deprecated' stamp: 'dtl 8/5/2003 22:15'!
recapitate
	"Deprecated. See ThisOSProcess>>recapitate"

	"self decapitate. (Delay forSeconds: 5) wait. self recapitate."

	^ OSProcess thisOSProcess recapitate! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 3/17/2002 13:38'!
restartVirtualMachine
	"Fork a new instance and quit this one. This moves the running VM into a new
	OS process, and starts a new X display for the new process. It does not reload the
	VM program text, so this cannot be used to restart the VM after rebuild."

	"UnixProcess restartVirtualMachine"

	| proc |
	proc := self forkSqueak.
	proc isNil
		ifTrue: 
			[self noAccessorAvailable.
			nil]
		ifFalse:
			[OSProcess thisOSProcess == proc
				ifFalse:
					["Quit if this is the parent process"
					Smalltalk quitPrimitive]].
	^ proc! !

!UnixProcess class methodsFor: 'unit tests' stamp: 'dtl 3/5/2005 14:19'!
runTests
	"Run a few tests to see if things are working correctly on Unix/Linux. 
	Output is on stdout, stderr, and the Squeak Transcript. One of the tests 
	requires input from stdin, so Squeak should be run from a shell command
	line and not as a background process.

	Warning: This test will crash your VM if your are using the -xshm command
	line option. For reasons which I do not quite understand, the X shared
	memory segment becomes invalid when the Squeak VM which initially
	opened the shared memory exits. The remaining Squeak children will crash
	when then then next try to update the display.

	Note: If you see 'select: Bad file descriptor' messages on your console standard
	output, these are occuring while running headless in the decapitate/recapitate
	tests.

	Important: Prior to evaluating this method, please type one line of text
	followed by a <cr> on the terminal standard input. This provides the
	input for the stdin test. Failing to provide this input prior to evaluating
	the tests will cause one of the test cases to fail."

	"UnixProcess runTests"

	| this s p failures result a |
	failures := 0.
	Transcript show: 'Begin OSProcess tests'; cr.
	Transcript show: 'Test for working ProcessAccessor ... '.
	this := OSProcess thisOSProcess.
	(this pid isKindOf: Integer)
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'Echo one line of text previously entered from stdin ... '.
	s := OSProcess readFromStdIn.
	s size > 0
		ifTrue: [Transcript show: 'OK'; cr; show: s; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'Message to stdout ... '.
	(OSProcess helloWorld isKindOf: AttachableFileStream)
		ifTrue: [Transcript show: 'OK'; cr; show: s; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'Message to stderr ... '.
	(OSProcess helloStdErr isKindOf: AttachableFileStream)
		ifTrue: [Transcript show: 'OK'; cr; show: s; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess cataFile ... '.
	p := UnixProcess catAFile.
	(Delay forSeconds: 1) wait.
	p exitStatus == 0
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess testEnvSet ... '.
	p := UnixProcess testEnvSet.
	(Delay forSeconds: 1) wait.
	p exitStatus == 0
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess catFromFileToFiles ... '.
	OSProcess command: 'rm /tmp/deleteMe.out'.
	OSProcess command: 'rm /tmp/deleteMe.err'.
	p := UnixProcess catFromFileToFiles.
	(Delay forSeconds: 1) wait.
	p exitStatus == 0
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess testRunCommand ... '.
	OSProcess command: 'rm /tmp/deleteMe.out'.
	OSProcess command: 'rm /tmp/deleteMe.err'.
	p := UnixProcess testRunCommand.
	(Delay forSeconds: 1) wait.
	p exitStatus == 256
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess testPipe ... '.
	UnixProcess testPipe = 'this is some text to write into the pipe'
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess testPipeLine ... '.
	UnixProcess testPipeLine = 'This is the text to write out through one pipe, copy through an external cat command, and then read back in through another pipe.'
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess spawnTenHeadlessChildren ... '.
	p := UnixProcess spawnTenHeadlessChildren.
	(p size == 10 and: [(p select: [:e | (e runState == #running) | (e exitStatus == 0)]) size == 10])
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess cloneSqueak ... '.
	p := UnixProcess cloneSqueak.
	(p isKindOf: UnixProcess)
		ifTrue: 
			[(Delay forSeconds: 1) wait.
			OSProcess snapshot: false andQuit: true].
	(Delay forSeconds: 5) wait.
	p exitStatus == 0
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess squeakSqueak ... '.
	p := UnixProcess squeakSqueak.
	(Delay forSeconds: 5) wait.
	OSProcess thisOSProcess sigkill: p.
	(Delay forSeconds: 1) wait.
	p exitStatus == 9
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess restartVirtualMachine ... '.
	p := UnixProcess restartVirtualMachine.
	p pid == OSProcess thisOSProcess pid
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess command: ''cat'' input: ''this is some test data'' ... '.
	p := (OSProcess thisOSProcess
		command: 'cat'
		input: 'this is some test data').
	(Delay forSeconds: 1) wait.
	p upToEnd = 'this is some test data'
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'UnixProcess waitForCommandOutput: ''echo sleeping...; sleep 1; echo I just slept for one second'' ... '.
	('*I just slept for one second*' match:
		(OSProcess thisOSProcess
			waitForCommandOutput: 'echo sleeping...; sleep 1; echo I just slept for one second'))
		ifTrue: [Transcript show: 'OK'; cr]
		ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1].
	Transcript show: 'Unix command pipeline with output and error returned in an array ... '.
	(Smalltalk hasClassNamed: #PipeableOSProcess)
		ifTrue:
			[a := (((Smalltalk at: #PipeableOSProcess) command: 'echo this is a test; BOGUS') | 'cut -c11-14') outputAndError.
			(((a isKindOf: Array)
				and: ['test*' match: (a at: 1)])
					and: ['*BOGUS*' match: (a at: 2)])
						ifTrue: [Transcript show: 'OK'; cr]
						ifFalse: [Transcript show: 'NFG'; cr. failures := failures + 1]]
		ifFalse:
			[Transcript show: 'skipping PipeableOSProcess test (requires CommandShell)'; cr].

	Transcript show: 'UnixProcess decapitate and recapatiate five times'; cr.
	5 timesRepeat:
		[UnixProcess decapitate.
		UnixProcess recapitate].

	failures == 1
		ifTrue: [result := 'OSProcess tests completed with ', failures printString, ' failure']
		ifFalse: [result := 'OSProcess tests completed with ', failures printString, ' failures'].
	Transcript show: result; cr.

	^ result
! !

!UnixProcess class methodsFor: 'utility - image save' stamp: 'dtl 7/23/2010 07:50'!
saveImageInBackground
	"Save image in a background OS process."

	"UnixProcess saveImageInBackground"

	^ self saveImageInBackground: self uniqueNameForSavedImage nice: false
! !

!UnixProcess class methodsFor: 'utility - image save' stamp: 'dtl 1/8/2008 22:39'!
saveImageInBackground: savedImageName nice: niceFlag
	"When Squeak is used as a server it is sometimes desirable to periodically
	save image snapshots. This method forks a headless Squeak to perform a
	snapshot without impacting the server Squeak. Very little additional memory
	is required to do this because Unix copy-on-write memory management allows
	the two Squeak images to share object memory while the save is performed.
	The saved image is given a time stamped name, and the image name of
	the main server Squeak remains unchanged. If niceFlag is true, the
	background OS process runs at lowered scheduling priority."

	^ self forkHeadlessSqueakAndDo:
		[| st |
		"Delay is required in the background process when forking a headless
		Squeak, probably necessary to permit X session stuff to settle down."
		(Delay forMilliseconds: 500) wait.
		(niceFlag notNil and: [niceFlag]) ifTrue:
			["lower priority of background OS process"
			OSProcess accessor nice: 1].
		st := SmalltalkImage current.
		(SourceFiles at: 2) ifNotNil:
			["ensure that copying the changes file will always work"
			st closeSourceFiles.
			st saveChangesInFileNamed: (st fullNameForChangesNamed: savedImageName)].
		st changeImageNameTo: savedImageName,'.image';
			closeSourceFiles;
			openSourceFiles;  "so SNAPSHOT appears in new changes file"
			saveImageSegments;
			snapshot: true andQuit: true]
! !

!UnixProcess class methodsFor: 'utility - image save' stamp: 'dtl 7/23/2010 07:50'!
saveImageInBackgroundNicely
	"Save image in a background OS process with lowered scheduling priority."

	"UnixProcess saveImageInBackgroundNicely"

	^ self saveImageInBackground: self uniqueNameForSavedImage nice: true
! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 3/10/2001 09:53'!
sessionID

	"UnixProcess sessionID"

	^ self thisOSProcess sessionID
! !

!UnixProcess class methodsFor: 'X display - deprecated' stamp: 'dtl 8/5/2003 22:18'!
setXDisplayName: xDisplayName
	"Deprecated. See ThisOSProcess>>setXDisplayName"

	"self setXDisplayName: 'unix:0' "
	"self setXDisplayName: ':1' "

	^ OSProcess thisOSProcess setXDisplayName: xDisplayName

! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 04:35'!
spawnTenHeadlessChildren
	"Spawn ten Squeak children, each of which writes a message to standard 
	output, then exits. Answer an array of pid values for the child 
	processes. "

	"UnixProcess spawnTenHeadlessChildren"

	| childBlock count children this |
	this := OSProcess thisOSProcess.
	this stdOut ifNil: [self noAccessorAvailable. ^ nil].
	count := 10.
	children := Array new: count.
	childBlock := 
			[this stdOut nextPutAll: 'hello world from child process '.
			this pid printOn: this stdOut.
			this stdOut nextPut: Character lf.
			this stdOut flush].
	(1 to: count)
		do: 
			[:e | 
			OSProcess thisOSProcess stdOut flush.
			children at: e put: (self forkHeadlessSqueakAndDoThenQuit: childBlock)].
	^ children! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 04:38'!
squeakSqueak
	"Start a new instance of Squeak running in a child OS process. The new instance
	will restart from the image file, so it is a clone of this image as it existed at the
	most recent image save. See cloneSqueak for an example of how to clone the
	running image without going back to the saved image file."

	"UnixProcess squeakSqueak"

	| proc |
	proc := self squeak.
	proc ifNil: [self noAccessorAvailable].
	^ proc
! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 7/13/2003 14:47'!
startSwiki: aSwiki onPort: num loggingTo: aFileName
	"Start a swiki in a headless Squeak image."

	"UnixProcess startSwiki: 'myswiki' onPort: 8081 loggingTo: 'log.txt'"

	| proc |
	(Smalltalk hasClassNamed: #SwikiAction)
		ifTrue:
			[proc := self forkSqueakAndDo:
				[(Smalltalk at: #SwikiAction) new restore: 'myswiki'.
				(Smalltalk at: #PWS) serveOnPort: num loggingTo: aFileName.
				UnixProcess decapitate].
			proc ifNil: [self noAccessorAvailable].
			^ proc]
		ifFalse:
		[self notify: 'PWS not installed in this image']
! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 12/27/2000 17:12'!
stdErr
	"UnixProcess stdErr"

	^ self thisOSProcess stdErr
! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 12/27/2000 17:11'!
stdIn
	"UnixProcess stdIn"

	^ self thisOSProcess stdIn
! !

!UnixProcess class methodsFor: 'utility' stamp: 'dtl 12/27/2000 17:11'!
stdOut
	"UnixProcess stdOut"

	^ self thisOSProcess stdOut
! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 04:41'!
testEnvSet
	"Set up a new environment for a child process. Exec a shell to show the 
	environment variables on the terminal standard output. Note that many 
	shells will set other environment variables in addition to those which we
	set up prior to executing the shell."

	"UnixProcess testEnvSet"

	| e proc |
	e := Dictionary new.
	e at: #KEY1 put: 'value1';
	 at: #KEY2 put: 'value2';
	 at: #KEY3 put: 'value3'.
	proc := self
			forkJob: '/bin/sh'
			arguments: #('-c' 'env' )
			environment: e
			descriptors: nil.
	proc ifNil: [self noAccessorAvailable].
	^ proc! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 2/11/2001 14:03'!
testPipe
	"Create an OS pipe, write some text to it, and read the text back from the 
	other end of the pipe."

	"UnixProcess testPipe inspect"

	| s p r |
	s := 'this is some text to write into the pipe'.
	p := OSPipe new.
	p ifNil: 
		[self noAccessorAvailable.
		^ p].
	p writer nextPutAll: s.
	p writer close.
	r := p reader next: s size.
	p reader close.
	^ r! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 2/24/2001 16:28'!
testPipeLine
	"Create two OS pipes, and a child OS process with its input connected to 
	one pipe and its  output connected to the other pipe. Write some text to
	the input pipe, and read the resulting output (just echoed back by the
	Unix cat command) back through the output pipe. Send a SIGHUP signal
	to the child process to tell it to exit.

	This test verifies the ability of Squeak to send text to an external OS process
	through a pipe, and read the output text back from another pipe.
	See ConnectedUnixProcess for a more useful implementation."

	"UnixProcess testPipeLine inspect"

	| testString pipe1 pipe2 input output src dest desc result child |
	OSProcess accessor canAccessSystem ifFalse: [^ nil].
	testString := 'This is the text to write out through one pipe, copy through an external cat command, and then read back in through another pipe.'.
	pipe1 := OSPipe blockingPipe.
	pipe2 := OSPipe nonBlockingPipe.
	input := pipe1 reader.
	output := pipe2 writer.
	src := pipe1 writer.
	dest := pipe2 reader.
	desc := Array
				with: input
				with: output
				with: nil.
	child := self
				forkJob: '/bin/cat'
				arguments: nil
				environment: nil
				descriptors: desc.
	input close.
	output close.
	src nextPutAll: testString.
	src close.
	(Delay forSeconds: 1) wait.
	result := dest next: testString size.
	dest close.
	child sigterm. "Tell the child to exit"
	^ result
! !

!UnixProcess class methodsFor: 'examples' stamp: 'dtl 7/12/2000 04:46'!
testRunCommand
	"Run the command 'ls -l /etc /etc/noSuchFile'. The output of the 
	command will be in the file '/tmp/deleteMe.out', and the error output
	will be in '/tmp/deleteMe.err'. "

	"UnixProcess testRunCommand"

	| out err desc args proc |
	out := FileStream newFileNamed: '/tmp/deleteMe.out'.
	err := FileStream newFileNamed: '/tmp/deleteMe.err'.
	desc := Array
				with: nil
				with: out
				with: err.
	args := Array
				with: '-l'
				with: '/etc'
				with: '/etc/noSuchFile'
				with: '/etc/anotherNonexistentFile'.
	proc := self
				forkJob: '/bin/ls'
				arguments: args
				environment: nil
				descriptors: desc.
	proc ifNil: [self noAccessorAvailable].
	out close.
	err close.
	^ proc! !

!UnixProcess class methodsFor: 'utility - image save' stamp: 'dtl 11/8/2005 11:39'!
uniqueNameForSavedImage
	"A time stamped image name that will sort in date order in a directory listing"

	"UnixProcess uniqueNameForSavedImage"

	| now month day hour minute second |
	now := DateAndTime now.
	month := now month asString.
	day := now dayOfMonth asString.
	hour := now hour24 asString.
	minute := now minute asString.
	second := now second asString.
	^ 'squeak-',
		now year asString,
		(month size < 2 ifTrue: ['0', month] ifFalse: [month]),
		(day size < 2 ifTrue: ['0', day] ifFalse: [day]),
		(hour size < 2 ifTrue: ['0', hour] ifFalse: [hour]),
		(minute size < 2 ifTrue: ['0', minute] ifFalse: [minute]),
		(second size < 2 ifTrue: ['0', second] ifFalse: [second])
! !

!UnixProcess class methodsFor: 'external command processing' stamp: 'dtl 1/18/2001 23:14'!
waitForCommandOutput: aCommandString 
	"Run a command in a shell process. Similar to the system(3) call in the 
	standard C library. The active Smalltalk process waits for completion of
	the external command process."

	"UnixProcess waitForCommandOutput: 'echo sleeping...; sleep 1; echo I just slept for one second'"
	"UnixProcess waitForCommandOutput: 'ThisIsABogusCommand'"

	^ self thisOSProcess waitForCommandOutput: aCommandString
! !

!UnixProcess class methodsFor: 'external command processing' stamp: 'dtl 1/18/2001 23:14'!
waitForCommandOutputArray: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in the standard C library.
	The active Smalltalk process waits for completion of the external command process."

	"UnixProcess waitForCommandOutputArray: 'echo Hello world!!; ls /NOSUCHFILE'"

	^ self thisOSProcess waitForCommandOutputArray: aCommandString! !

!UnixProcess methodsFor: 'private' stamp: 'djr 4/6/2009 14:31'!
argsAsFlatArrayAndOffsets: anArrayOfNullTerminatedStrings
	"Given anArrayOfNullTerminatedStrings, flatten the array into a string buffer,
	leaving space at the beginning of the buffer for a list of C pointers. Answer
	the string buffer and an array of address offsets. The address offsets may later
	be converted to C pointers and overlaid on the beginning of the buffer, resulting
	in a data structure which can be treated as a C array of strings suitable for
	use as a Unix argv vector."

	"OSProcess thisOSProcess argsAsFlatArrayAndOffsets:
		(UnixProcess arguments collect:
			[:e | e, (Character value: 0) asString])"

	| offset arraySize flatStringArray ws addresses |
	"Preserve offset space to hold address pointers."
	offset := (anArrayOfNullTerminatedStrings size + 1) * self processAccessor sizeOfPointer.
	"Allocate flat strings buffer"
	arraySize := (anArrayOfNullTerminatedStrings collect: [:e | e size]) sum
		+ ((anArrayOfNullTerminatedStrings size + 1) * self processAccessor sizeOfPointer).
	"Flatten the strings out into a buffer, leaving room at the
	beginning of the buffer for an array of addresses."
	flatStringArray := String new: arraySize.
	ws := WriteStream on: flatStringArray.
	offset timesRepeat: [ws nextPut: (Character value: 0)].
	anArrayOfNullTerminatedStrings do: [:e | ws nextPutAll: e].
	"Find address offsets to be used in creating the addresses for the strings."
	ws := WriteStream on: (Array new: anArrayOfNullTerminatedStrings size).
	(anArrayOfNullTerminatedStrings collect: [:e | e size])
		inject: offset into: [:p :e | ws nextPut: p. p + e].
	addresses := ws contents.
	"Results"
	^ Array with: flatStringArray with: addresses
! !

!UnixProcess methodsFor: 'accessing' stamp: 'dtl 11/7/2000 09:09'!
arguments

	^ arguments
! !

!UnixProcess methodsFor: 'environment' stamp: 'dtl 1/25/2013 19:00'!
chDir: pathString
	"Change current working directory, and update $PWD if it exists in the environment.
	Answer nil for success, or an error message."

	"OSProcess thisOSProcess chDir: '/tmp'"
	"OSProcess thisOSProcess chDir: '/no/such/path'"
	"OSProcess thisOSProcess chDir: OSProcess defaultPathString"

	| realPath result |
	realPath := self processAccessor realpath: pathString.
	realPath ifNil: [realPath := pathString].
	result := self processAccessor chDir: realPath.
	result isNil
		ifTrue:
			[(self environmentAt: #PWD)
				ifNotNil:
					[self environmentAt: #PWD put: realPath.
					^ nil]]
		ifFalse:
			[self inform: realPath, ': ', result.
			^ result]! !

!UnixProcess methodsFor: 'external command processing' stamp: 'dtl 2/27/2002 15:24'!
command: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in the standard C library,
	except that aCommandString runs asynchronously in a child process. The command is
	run by a ConnectedUnixProcess in order to facilitate command pipelines within Squeak."

	"UnixProcess thisOSProcess command: 'ls -l /etc'"

	| proc |
	pid isNil
		ifTrue:
			[self class noAccessorAvailable. ^nil]
		ifFalse:
			[proc := self
					forkJob: ExternalUnixOSProcess defaultShellPath
					arguments: (Array with: '-c' with: aCommandString)
					environment: nil
					descriptors: nil.
			proc ifNil: [self class noAccessorAvailable].
			^ proc]
! !

!UnixProcess methodsFor: 'external command processing' stamp: 'dtl 7/12/2003 11:51'!
command: aCommandString input: aStreamOrString
	"Run a command in a shell process. Similar to the system(3) call in the standard C library,
	except that aCommandString runs asynchronously in a child process."

	"OSProcess thisOSProcess
		command: 'cat'
		input: 'this is some test data'"

	"OSProcess thisOSProcess
		command: 'cat'
		input: (ReadStream on: 'this is some test data')"

	| proc |
	(Smalltalk hasClassNamed: #PipeableOSProcess)
		ifTrue:
			[proc := (Smalltalk at: #PipeableOSProcess) command: aCommandString.
			proc ifNil: [^ nil].
			proc nextPutAll: aStreamOrString contents.
			proc pipeToInput close.
			^ proc]
		ifFalse:
			[self notify: 'the #command:input: method requires CommandShell, using #command: instead'.
			^ self command: aStreamOrString contents]

! !

!UnixProcess methodsFor: 'private' stamp: 'djr 4/6/2009 14:31'!
envAsFlatArrayAndOffsets: anEnvironmentDictionary
	"Given anEnvironmentDictionary, flatten the dictionary into a string buffer,
	leaving space at the beginning of the buffer for a list of C pointers. Answer
	the string buffer and an array of address offsets. The address offsets may later
	be converted to C pointers and overlaid on the beginning of the buffer, resulting
	in a data structure which can be treated as a C array of strings suitable for
	use as a Unix process environment."

	"OSProcess thisOSProcess envAsFlatArrayAndOffsets: UnixProcess env"

	| offset envArray arraySize flatStringArray ws addresses |
	"Preserve offset space to hold address pointers."
	offset := (anEnvironmentDictionary size + 1) * self processAccessor sizeOfPointer.
	"Build collection of environment strings"
	envArray := OrderedCollection new.
	anEnvironmentDictionary keysAndValuesDo: [:k :v | envArray add: (k, '=', v)].
	"Allocate flat strings buffer"
	arraySize := ((envArray collect: [:e | e size + 1]) sum) + offset.
	"Flatten the strings out into a buffer, leaving room at the beginning of
	the buffer for an array of addresses."
	flatStringArray := String new: arraySize.
	ws := WriteStream on: flatStringArray.
	offset timesRepeat: [ws nextPut: (Character value: 0)].
	envArray do: [:e | ws nextPutAll: e; nextPut: (Character value: 0)].
	"Find address offsets to be used in creating the addresses for the strings."
	ws := WriteStream on: (Array new: anEnvironmentDictionary size).
	(envArray collect: [:e | e size + 1])
		inject: offset into: [:p :e | ws nextPut: p. p + e].
	addresses := ws contents.
	"Results"
	^ Array with: flatStringArray with: addresses
! !

!UnixProcess methodsFor: 'accessing' stamp: 'dtl 11/7/2000 09:09'!
environment

	^ environment
! !

!UnixProcess methodsFor: 'environment' stamp: 'dtl 3/6/2001 21:30'!
environmentAt: aSymbol
	"Answer an environment variable for the external OS process, and update the dictionary
	in this Smalltalk object."

	^ environment at: aSymbol asSymbol ifAbsent: []
! !

!UnixProcess methodsFor: 'environment' stamp: 'dtl 3/6/2001 21:30'!
environmentAt: aSymbol put: aString
	"Set an environment variable for the external OS process, and update the dictionary in
	this Smalltalk object."

	| s |
	self initialize.
	s := self processAccessor environmentAt: aSymbol put: aString.
	s ifNotNil: [ self environment at: aSymbol asSymbol put: aString ].
	^ s

! !

!UnixProcess methodsFor: 'finalization' stamp: 'dtl 12/14/2013 09:16'!
finalize
	"Use this to release any external resources prior to reinitializing."

	super finalize.
	stdIn := stdIn ifNotNil:
		[[stdIn close] on: Error do: [:ex | ].
		nil].
	stdIn := stdOut ifNotNil:
		[[stdOut close] on: Error do: [:ex | ].
		nil].
	stdIn := stdErr ifNotNil:
		[[stdErr close] on: Error do: [:ex | ].
		nil].
	sessionID := nil.
	ppid := nil.
	programName := nil.
	arguments := nil.
	path := nil.
	environment := nil.
	self updateAllMyChildren.
	allMyChildren := nil.
	processAccessor
		ifNotNil: 
			[processAccessor removeDependent: self.
			processAccessor := nil]
! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 2/27/2002 15:24'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams 

	"Call Unix vfork() and execve() to create a child process, and answer the 
	child process. This method is expected to be called by class side methods."

	^ ExternalUnixOSProcess
		forkAndExec: executableFile
		arguments: arrayOfStrings
		environment: stringDictionary
		descriptors: arrayOf3Streams! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 6/10/2011 14:27'!
forkHeadlessSqueak
	"Just like forkSqueak, except that the child Squeak continues headless."

	| thisPid childPid child connected |
	stdOut ifNil: [^ nil].
	self stdOut flush.
	self stdErr flush.
	thisPid := self pid.
	connected := self processAccessor canControlXDisplay
					and: [self flushXDisplay notNil].
	childPid := self processAccessor forkSqueak.
	^ childPid ifNotNil: [
		childPid == 0
			ifTrue:
				[connected ifTrue: [self disconnectXDisplay].
				OSProcess thisOSProcess processAccessor changed: #pid.
				^ self]
			ifFalse:
				[child := ExternalUnixOSProcess new.
				child pid: childPid.
				child ppid: thisPid.
				child programName: self programName.
				child initialStdIn: self stdIn.	
				child initialStdOut: self stdOut.
				child initialStdErr: self stdErr.
				child arguments: self arguments.
				child initialEnvironment: self environment.
				child notYetRunning.
				self registerChildProcess: child.
				child running.
				^ child]]
! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 1/7/2001 12:59'!
forkHeadlessSqueakAndDo: aBlock 
	"Start a new instance of Squeak running in a child OS process, and  
	execute aBlock in the child instance. The new instance is a clone of 
	this image, but without a connection to the X display. The child instance 
	executes aBlock, which hopefully does not involve interaction with the 
	X display; and the parent continues normally.  
	  
	The child should not depend on using existing connections to external  
	resources. For example, the child may lose its connections to stdin, 
	stdout, and stderr after its parent exits."

	| childOrThisProc |
	childOrThisProc := self forkHeadlessSqueak.
	childOrThisProc
		ifNil: 
			[self class noAccessorAvailable.
			^ nil].
	childOrThisProc == self
		ifTrue:
			["Child process"
			aBlock value].
	^ childOrThisProc
! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 8/7/2003 07:17'!
forkHeadlessSqueakAndDoThenQuit: aBlock 
	"Start a new instance of Squeak running in a child OS process, and 
	execute aBlock in the child instance. The new instance is a clone of
	this image, but without a connection to the X display. The child instance
	executes aBlock, which hopefully does not involve interaction with the
	X display; and the parent continues normally.

	The child should not depend on using existing connections to external 
	resources. For example, the child may lose its connections to stdin, stdout,
	and stderr after its parent exits."

	"self thisOSProcess forkHeadlessSqueakAndDoThenQuit:
		[OSProcess thisOSProcess stdOut nextPutAll: 'hello world!!'; nextPut: Character lf]"
	"self thisOSProcess forkHeadlessSqueakAndDoThenQuit: [OSProcess thisOSProcess command: 'xeyes']"

	| childOrThisProc |
	childOrThisProc := self forkHeadlessSqueak.
	childOrThisProc
		ifNil: 
			[self class noAccessorAvailable.
			^ nil].
	childOrThisProc == self
		ifTrue:
			["Child process"
			aBlock value.
			Smalltalk quitPrimitive].
	^ childOrThisProc! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 12/27/2000 16:47'!
forkJob: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams
	"Call Unix vfork() and execve() to create a child process, and answer the child process.
	Delegate this to the singleton OSProcess>>thisOSProcess."

	^ self forkAndExec: executableFile
		arguments: arrayOfStrings
		environment: stringDictionary
		descriptors: arrayOf3Streams
! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 4/27/2011 21:28'!
forkSqueak
	"Fork a child and continue running this Squeak image in both the parent and the child.
	Parent and child are distinguished by the pid returned by primForkSqueak.

	If continuing as the parent process, answer the ExternalUnixOSProcess which
	represents the child. This can be inspected to watch the run state of the child
	process from the parent.

	If continuing as the child process, answer OSProcess thisOSProcess. This can be
	inspected to watch the full state of the child process from the child. The
	child cannot directly view the state of its parent.

	Parent and child should be cautious about using shared connections to external
	resources."

	"self thisOSProcess forkSqueak"

	| thisPid childPid child connected |
	stdOut ifNil: [^ nil].
	self stdOut flush.
	self stdErr flush.
	thisPid := self pid.
	connected := self processAccessor canControlXDisplay
					and: [self flushXDisplay notNil].
	childPid := self processAccessor forkSqueak.
	^ childPid ifNotNil: [
		childPid == 0
			ifTrue:
				[connected ifTrue: [self disconnectXDisplay; recapitate].
				OSProcess thisOSProcess processAccessor changed: #pid.
				^ self]
			ifFalse:
				[child := ExternalUnixOSProcess new.
				child pid: childPid.
				child ppid: thisPid.
				child programName: self programName.
				child initialStdIn: self stdIn.	
				child initialStdOut: self stdOut.
				child initialStdErr: self stdErr.
				child arguments: self arguments.
				child initialEnvironment: self environment.
				child notYetRunning.
				self registerChildProcess: child.
				child running.
				^ child]]
! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 10/8/2001 19:56'!
forkSqueakAndDo: aBlock
	"Start a new instance of Squeak running in a child OS process. The new instance is a
	clone of this image except for the return value of this method. It does not reload the
	image file from disk. The child image evaluates aBlock.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	"UnixProcess thisOSProcess forkSqueakAndDo:
		[Object inform: 'Hi, I am the child Squeak process.']"

	| childOrThisProc |
	childOrThisProc := self forkSqueak.
	(childOrThisProc == self) ifTrue: [aBlock value]. "Child process"
	^ childOrThisProc! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 10/8/2001 19:56'!
forkSqueakAndDoThenQuit: aBlock
	"Start a new instance of Squeak running in a child OS process. The new instance is a
	clone of this image except for the return value of this method. It does not reload the
	image file from disk. The child image evaluates aBlock.

	The child should not depend on using existing connections to external resources. For
	example, the child may lose its connections to stdin, stdout, and stderr after its parent
	exits."

	"UnixProcess thisOSProcess forkSqueakAndDoThenQuit:
		[Object inform: 'Hi, I am the child Squeak process.']"

	| childOrThisProc |
	childOrThisProc := self forkSqueak.
	(childOrThisProc == self)
		ifTrue:
			[ aBlock value.
			Smalltalk quitPrimitive]. "Child process"
	^ childOrThisProc! !

!UnixProcess methodsFor: 'private' stamp: 'dtl 3/31/2001 14:16'!
getArgumentList
	"Answer the argument list using anOSProcessAccessor. For Unix, the first element of the list
	would be the program name. This element will not be treated as an argument; rather, it is
	stored as the programName instance variable."

	| index val list |
	list := OrderedCollection new.
	index := 2.
	[(val := processAccessor primArgumentAt: index) notNil] whileTrue: [
		list add: val.
		index := index + 1].
	^ list asArray

! !

!UnixProcess methodsFor: 'environment' stamp: 'dtl 3/22/2000 05:55'!
getCwd
	"Get current working directory. At image startup, this is equivalent to
	evaluating environmentAt: #PWD"

	"OSProcess thisOSProcess getCwd"

	^ self processAccessor primGetCurrentWorkingDirectory


! !

!UnixProcess methodsFor: 'private' stamp: 'dtl 3/31/2001 14:14'!
getEnvironmentDictionary
	"Answer an environment dictionary using an OSProcessAccessor."

	| index str key val env |
	env := Dictionary new.
	index := 1.
	[(str := processAccessor primEnvironmentAt: index) notNil] whileTrue: [
		key := (str copyUpTo: $=) asSymbol.
		val := (str copyFrom: ((str indexOf: $=) + 1) to: (str size)).
		env at: key put: val.
		index := index + 1].
	^ env


! !

!UnixProcess methodsFor: 'private' stamp: 'dtl 3/31/2001 14:16'!
getProgramName
	"Answer the name of the program which is being run by this OS process. Assume the
	Unix convention where the first element of (char **)argv is the program name."

	^ programName := processAccessor primArgumentAt: 1
! !

!UnixProcess methodsFor: 'initialize - release' stamp: 'dtl 7/6/2010 21:55'!
initialize
	"Set my instance variables to reflect the state of the OS process in which 
	this Smalltalk virtual machine is executing."

	super initialize.
	(self processAccessor notNil and: [processAccessor canAccessSystem])
		ifTrue: 
			[self refreshFromProcessAccessor]
		ifFalse: 
			[stdIn := nil.
			stdOut := nil.
			stdErr := nil].
	processAccessor restartChildWatcherProcess
! !

!UnixProcess methodsFor: 'initialize - release' stamp: 'dtl 7/4/2010 11:15'!
initializeAllMyChildren

	^ allMyChildren := Dictionary new
! !

!UnixProcess methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:49'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ self class isUnix
! !

!UnixProcess methodsFor: 'VM atexit' stamp: 'dtl 7/4/2010 14:44'!
killCurrentChildrenAtExit
	"Arrange for the currently active child processes to receive a SIGTERM
	signal then the Squeak VM exits. Each invokation of this method overrides
	the effects of any previous calls."

	"OSProcess thisOSProcess killCurrentChildrenAtExit"

	self processAccessor killOnVmExit: allMyChildren withSignal: nil
! !

!UnixProcess methodsFor: 'private' stamp: 'dtl 11/7/2000 09:26'!
needsRefresh
	"Answer true if the sessionID variable is out of date with respect to the running OS Process."

	^ ((sessionID ~= (self processAccessor primGetSession)) | (pid ~= (self processAccessor primGetPid)))
! !

!UnixProcess methodsFor: 'accessing' stamp: 'dtl 11/7/2000 09:10'!
path

	^ path := self pathString
! !

!UnixProcess methodsFor: 'accessing' stamp: 'dtl 11/7/2000 09:10'!
path: aPathString

	self environmentAt: #PATH put: aPathString.
	path := self pathString! !

!UnixProcess methodsFor: 'private' stamp: 'dtl 3/6/2001 21:46'!
pathString
	"Answer the path string from the environment. Assume Unix convention in which the
	path name is a colon delimited string stored in the PATH environment variable."

	^ self environment at: #PATH ifAbsent: [nil]! !

!UnixProcess methodsFor: 'accessing' stamp: 'dtl 10/18/2001 20:01'!
pid

	^ pid := self processAccessor primGetPid
! !

!UnixProcess methodsFor: 'accessing' stamp: 'dtl 11/7/2000 09:10'!
ppid
	"Always refresh ppid from the processAccessor, because it is possible for a child
	to be reparented when the parent exits. The child does not know about this, so
	we refresh ppid on every access."

	^ ppid := processAccessor primGetPPid
! !

!UnixProcess methodsFor: 'printing' stamp: 'dtl 7/7/2010 07:37'!
printOn: aStream
	"In English, say 'a Unix' rather than 'an Unix'. Therefore do not use super printOn, which
	treats $U as a vowel."

	self ppid.		"Force update of ppid, in case parent may have exited."
	aStream nextPutAll: 'a ';
			nextPutAll: self class name;
			nextPutAll: ' with pid ';
			nextPutAll: self pid printString
! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 1/25/2004 12:40'!
processProxy: anExternalProcess forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams
	"Call Unix vfork() and execve() to create a child process, and answer the child process.
	This method is expected to be called by class side methods. Prepare the arguments before
	calling the primitive, including null termination of all strings. anExternalProcess is an
	object which represents the new child process, and which responds to the #pid: message."

	| nullString progName args argVecAndOffsets argVec argOffsets
	envVecAndOffsets envVec envOffsets in out err childPid pwd |
	stdOut ifNil: [^ nil].
	nullString := (Character value: 0) asString.
	progName := executableFile, nullString.		"Null terminated string"
	arrayOfStrings isNil							"Should be a (possibly empty) array"
		ifTrue:
			[args := Array with: progName]		"First argument is the program name (Unix convention)"
		ifFalse:
			[args := (OrderedCollection new: arrayOfStrings size + 2)
						add: progName;
						addAll: (arrayOfStrings collect: [:e | e, nullString ]);	"Null terminate each string"
						yourself;
						asArray].
	argVecAndOffsets := self argsAsFlatArrayAndOffsets: args.
	argVec := argVecAndOffsets at: 1.
	argOffsets := argVecAndOffsets at: 2.
	(stringDictionary notNil and: [stringDictionary ~= (self environment)])
		ifTrue:
			[envVecAndOffsets := self envAsFlatArrayAndOffsets: stringDictionary.
			envVec := envVecAndOffsets at: 1.
			envOffsets := envVecAndOffsets at: 2]
		ifFalse:
			[envVec := nil.
			envOffsets := nil].	"Same as current environment, so just pass nil."
	arrayOf3Streams isNil
		ifTrue:
			[in := self handleFromAccessor: (self stdIn ioHandle).
			out := self handleFromAccessor: (self stdOut ioHandle).
			err := self handleFromAccessor: (self stdIn ioHandle)]
		ifFalse:
			[(arrayOf3Streams at: 1)
				isNil
					ifTrue: [in := self handleFromAccessor: (self stdIn ioHandle)]
					ifFalse: [in := self handleFromFileStream: (arrayOf3Streams at: 1)].			
			(arrayOf3Streams at: 2)
				isNil
					ifTrue: [ out := self handleFromAccessor: (self stdOut ioHandle)]
					ifFalse: [ out := self handleFromFileStream: (arrayOf3Streams at: 2)].
			(arrayOf3Streams at: 3)
				isNil
					ifTrue: [ err := self handleFromAccessor: (self stdErr ioHandle)]
					ifFalse: [ err := self handleFromFileStream: (arrayOf3Streams at: 3)]].
	pwd := anExternalProcess pwd.
	(pwd = self getCwd)
		ifTrue: [pwd := nil]
		ifFalse: [pwd := pwd, nullString].
	childPid := self processAccessor
				forkAndExec: progName
				stdIn: in
				stdOut: out
				stdErr: err
				argBuf: argVec
				argOffsets: argOffsets
				envBuf: envVec
				envOffsets: envOffsets
				workingDir: pwd.
	anExternalProcess pid: childPid.
	anExternalProcess ppid: self pid.
	((childPid == 0) or: [childPid isNil])
		ifTrue:
			[anExternalProcess unknownRunState]
		ifFalse:
			[anExternalProcess running.
			self registerChildProcess: anExternalProcess].
	^ anExternalProcess
! !

!UnixProcess methodsFor: 'accessing' stamp: 'dtl 11/7/2000 09:10'!
programName

	^ programName
! !

!UnixProcess methodsFor: 'accessing' stamp: 'dtl 3/17/2007 22:54'!
pthread
	"The identity of the pthread in which the interpreter executes.
	Always refresh pthread from the processAccessor, because it is possible for a child
	to be reparented when the parent exits. The child does not know about this, so
	we refresh pthread on every access."

	^ pthread := processAccessor getThreadID
! !

!UnixProcess methodsFor: 'IO redirection' stamp: 'dtl 5/18/2009 21:07'!
redirect: aFileStream to: newFileStream

	| oldfd newfd |
	oldfd := self processAccessor unixFileNumber: aFileStream fileID.
	newfd := self processAccessor unixFileNumber: newFileStream fileID.
	self processAccessor primUnixFileClose: oldfd.
	^ self processAccessor primDup: newfd to: oldfd
! !

!UnixProcess methodsFor: 'IO redirection' stamp: 'dtl 5/17/2009 23:12'!
redirectStdErrTo: fileStream
	"Redirect the standard error stream to another output stream."

	^ self redirect: self stdErr to: fileStream
! !

!UnixProcess methodsFor: 'IO redirection' stamp: 'dtl 5/17/2009 23:12'!
redirectStdInTo: fileStream
	"Redirect the standard input stream to another input stream."

	^ self redirect: self stdIn to: fileStream
! !

!UnixProcess methodsFor: 'IO redirection' stamp: 'dtl 5/17/2009 23:11'!
redirectStdOutTo: fileStream
	"Redirect the standard output stream to another output stream."

	"| pipe |
	pipe := OSPipe nonBlockingPipe.
	OSProcess thisOSProcess redirectStdOutTo: pipe writer.
	pipe"
	
	^ self redirect: self stdOut to: fileStream
! !

!UnixProcess methodsFor: 'updating' stamp: 'dtl 7/4/2010 22:35'!
refreshFromProcessAccessor
	"Set my instance variables to reflect the state of the OS process in which this Smalltalk
	virtual machine is executing."

	self needsRefresh ifTrue: [
		sessionID := self processAccessor getSessionIdentifier.
		pid := processAccessor primGetPid.
		ppid := processAccessor primGetPPid.
		pthread := processAccessor getThreadID.
		self setStdIn.
		self setStdOut.
		self setStdErr.
		programName := self getProgramName.
		arguments := self getArgumentList.
		environment := self getEnvironmentDictionary.
		path := self pathString.
		self updateAllMyChildren]
! !

!UnixProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 11:12'!
registerChildProcess: anOSProcess

	self processAccessor sigChldSemaphore.
	self processAccessor grimReaperProcess.	 "Start the reaper process if it is not running."
	accessProtect
		critical: [allMyChildren
				at: anOSProcess pid
				put: anOSProcess].
	^ anOSProcess
! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:49'!
sigabrt: anExternalOSProcess
	"Send a SIGABRT signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigabrtTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:49'!
sigalrm: anExternalOSProcess
	"Send a SIGALRM signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigalrmTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:48'!
sigchld: anExternalOSProcess
	"Send a SIGCHLD signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigchldTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:48'!
sigcont: anExternalOSProcess
	"Send a SIGCONT signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigcontTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:48'!
sighup: anExternalOSProcess
	"Send a SIGHUP signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSighupTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:48'!
sigint: anExternalOSProcess
	"Send a SIGINT signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigintTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:48'!
sigkill: anExternalOSProcess
	"Send a SIGKILL signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigkillTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:48'!
sigpipe: anExternalOSProcess
	"Send a SIGPIPE signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigpipeTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:49'!
sigquit: anExternalOSProcess
	"Send a SIGQUIT signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigquitTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:50'!
sigstop: anExternalOSProcess
	"Send a SIGSTOP signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigstopTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:50'!
sigterm: anExternalOSProcess
	"Send a SIGTERM signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigtermTo: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:50'!
sigusr1: anExternalOSProcess
	"Send a SIGUSR1 signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigusr1To: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'OS signal sending' stamp: 'dtl 7/6/2000 16:51'!
sigusr2: anExternalOSProcess
	"Send a SIGUSR2 signal to the external process represented by anExternalOSProcess."

	^ self processAccessor primSendSigusr2To: anExternalOSProcess pid! !

!UnixProcess methodsFor: 'child process creation' stamp: 'dtl 1/21/2001 11:47'!
squeak
	"Start a new instance of Squeak running in a child OS process. The new 
	instance will restart from the image file, so it is a clone of this image 
	as it existed at the most recent image save."

	"OSProcess thisOSProcess squeak"

	^ self
		forkJob: self programName
		arguments: self arguments
		environment: nil
		descriptors: nil! !

!UnixProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 11:10'!
unregisterChildProcess: anOSProcess

	accessProtect
		critical: [allMyChildren
				removeKey: self pid
				ifAbsent: []].
	^ anOSProcess
! !

!UnixProcess methodsFor: 'updating' stamp: 'dtl 10/20/2001 09:05'!
update: aParameter
	"Framework to update some or all of the instance variables based on external events,
	such as receipt of a sigchd signal when a child process exits."

	(aParameter == (self processAccessor)) ifTrue:
		[^ self refreshFromProcessAccessor; yourself].
	(aParameter == #pid) ifTrue:
		[^ self resetChildProcessDictionary. "Forget children of prior process"].
	(aParameter == #childProcessStatus) ifTrue:
		[^ self updateActiveChildren; changed; yourself].
	(aParameter == #startUp) ifTrue:
		[^ self update: #pid].
	aParameter == #invalidProcessAccessor ifTrue:
		[processAccessor := processAccessor ifNotNil: [processAccessor removeDependent: self. nil].
		^ self].
	self error: 'Unexpected update parameter'! !

!UnixProcess methodsFor: 'external command processing' stamp: 'dtl 2/27/2002 15:24'!
waitForCommand: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in the standard C library.
	The active Smalltalk process waits for completion of the external command process. This just
	uses a simple polling loop, which is not very elegant but works well enough for most purposes."

	"OSProcess thisOSProcess waitForCommand: 'echo sleeping...; sleep 3; echo I just slept for three seconds'"

	| proc d |
	d := Delay forMilliseconds: 50.
	proc := self
		forkJob: ExternalUnixOSProcess defaultShellPath
		arguments: (Array with: '-c' with: aCommandString)
		environment: nil
		descriptors: nil.
	proc ifNil: [self class noAccessorAvailable].
	[proc runState == #complete] whileFalse: [d wait].
	^ proc
! !

!UnixProcess methodsFor: 'external command processing' stamp: 'dtl 7/12/2003 12:06'!
waitForCommandOutput: aCommandString 
	"Run a command in a shell process. Similar to the system(3) call in the 
	standard C library. The active Smalltalk process waits for completion of
	the external command process."

	"OSProcess thisOSProcess waitForCommandOutput: 'echo sleeping...; sleep 1; echo I just slept for one second'"
	"OSProcess thisOSProcess waitForCommandOutput: 'ThisIsABogusCommand'"
	"OSProcess thisOSProcess waitForCommandOutput: '/bin/ls -l /etc /bin'"
	"OSProcess thisOSProcess waitForCommandOutput: 'echo Hello world!!; ls /NOSUCHFILE'"

	(Smalltalk hasClassNamed: #PipeableOSProcess)
		ifTrue:
			[^ ((Smalltalk at: #PipeableOSProcess) command: aCommandString) output]
		ifFalse:
			[self notify: 'the #waitForCommandOutput: method requires CommandShell'.
			^ '']
! !

!UnixProcess methodsFor: 'external command processing' stamp: 'dtl 7/12/2003 12:08'!
waitForCommandOutputArray: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in the standard C library.
	The active Smalltalk process waits for completion of the external command process."

	"OSProcess thisOSProcess waitForCommandOutputArray: 'echo Hello world!!; ls /NOSUCHFILE'"

	| proc |
	(Smalltalk hasClassNamed: #PipeableOSProcess)
		ifTrue:
			[proc := (Smalltalk at: #PipeableOSProcess) command: aCommandString.
			^ Array
				with: proc output
				with: proc errorUpToEnd
				with: proc processProxy exitStatus]
		ifFalse:
			[self notify: 'the #waitForCommandOutputArray: method requires CommandShell'.
			^ Array with: '' with: '' with: nil]
! !

ThisOSProcess subclass: #WindowsProcess
	instanceVariableNames: 'processHandle environment mainThread threads'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Win32'!

!WindowsProcess commentStamp: 'dtl 9/25/2005 16:31' prior: 0!
I represent a Windows operating system process, such as the process in which the Squeak VM is currently running. I collaborate with an instance of WindowsOSProcessAccessor to provide primitive access to the external operating system. My instance variables are maintained as a convenience to allow inspection of an OSProcess. Access to these variables should always be done with my accessor methods, which update the instance variables by querying my WindowsOSProcessAccessor.

My process ID and process handle (a Win32 HANDLE) are held by my pid and processHandle variables. The main thread for this process is held by my mainThread variable.

Standard input, output, and error streams are available, and my be used when the console is open (WindowsProcess>>openConsole). They can also be reassigned to file streams (WindowsOSProcessAccessor>>setStdOut:).

When external processes are created, they are added to my allMyChildren collection, and a thread is created to wait for any of them to exit. This thread is held by my childWatcherThread instance variable while the thread is active, and is also added to my threads collection.

Whenever a child process exits, the childWatcherThread will signal a Semaphore (a Smalltalk Semaphore, not a Windows semaphore), then exit. A Squeak process in my processAccessor waits on this Semaphore, and sends an 'update: #childProcessStatus' message to me. In response to this, I update the status of my active child processes, one or more of which will have exited. If any of my child processes are still active, I set a new childWatcherThread to wait for them to exit.

Note that some Windows applications will exit their main process after creating another application process. These applications will appear to Squeak as if they have exited immediately, even though the application is running.!

!WindowsProcess class methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:50'!
isResponsibleForThisPlatform
	"Answer true if this class is responsible for representing the OS process for
	the Squeak VM running on the current platform."

	^ self isWindows
! !

!WindowsProcess methodsFor: 'child process management' stamp: 'dtl 2/26/2002 16:08'!
activeHandles
	"Answer an Array of handles for all children that are believed to be running."

	^ (self activeChildren collect: [:c | c handle]) asArray
! !

!WindowsProcess methodsFor: 'console' stamp: 'dtl 9/7/2002 20:57'!
closeConsole
	"Close the console. The standard input, output and error streams will no longer be available."

	"OSProcess thisOSProcess closeConsole"

	self processAccessor primFreeConsole.
	self refreshFromProcessAccessor
! !

!WindowsProcess methodsFor: 'child process creation' stamp: 'dtl 2/28/2002 14:48'!
command: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in
	the standard C library, except that aCommandString runs asynchronously
	in a child process. Answer an instance of ExternalWindowsProcess which
	is a proxy for the new Windows process."

	"OSProcess command: 'SOL'"

	^ ExternalWindowsOSProcess command: aCommandString
! !

!WindowsProcess methodsFor: 'accessing' stamp: 'dtl 2/22/2002 22:02'!
environment

	^ environment
! !

!WindowsProcess methodsFor: 'environment' stamp: 'DamienCassou 5/3/2012 11:53'!
environmentAt: aSymbol
	"Answer an environment variable for the external OS process, and update the dictionary
	in this Smalltalk object."

	^ environment at: aSymbol asSymbol ifAbsent: []
! !

!WindowsProcess methodsFor: 'environment' stamp: 'DamienCassou 5/3/2012 11:53'!
environmentAt: aSymbol put: aString
	"Set an environment variable for the external OS process, and update the dictionary in
	this Smalltalk object."

	| s |
	self initialize.
	s := self processAccessor environmentAt: aSymbol put: aString.
	s ifNotNil: [ self environment at: aSymbol asSymbol put: aString ].
	^ s

! !

!WindowsProcess methodsFor: 'child process creation' stamp: 'dtl 10/18/2001 20:19'!
forkAndExec: executableFile arguments: arrayOfStrings environment: stringDictionary descriptors: arrayOf3Streams
	"Use my processAccessor to call vfork() and execve() and create a
	new child task. Answer a proxy for the new task, an instance of
	ExternalWindowsProcess."

	self notYetImplemented
! !

!WindowsProcess methodsFor: 'environment' stamp: 'dtl 1/24/2013 19:23'!
getCwd
	"Get current working directory. If this cannot be obtained from the
	environment, answer a reasonable default."

	"OSProcess thisOSProcess getCwd"

	^ self processAccessor primGetCurrentWorkingDirectory
		ifNil: [OSProcess defaultPathString]


! !

!WindowsProcess methodsFor: 'private' stamp: 'dtl 2/22/2002 22:00'!
getEnvironmentDictionary
	"Answer an environment dictionary using an OSProcessAccessor."

	"OSProcess thisOSProcess getEnvironmentDictionary"

	| strings env |
	strings := processAccessor primGetEnvironmentStrings.
	strings isNil
		ifTrue:
			[^ nil]
		ifFalse:
			[env := Dictionary new.
			strings do: [:s |
				env at: (s copyUpTo: $=) asSymbol put: (s copyAfterLast: $=)].
			^ env]
! !

!WindowsProcess methodsFor: 'initialize - release' stamp: 'dtl 8/25/2010 20:49'!
initialize
	"Set my instance variables to reflect the state of the OS process in which 
	this Smalltalk virtual machine is executing. On Windows, we cannot rely
	on the pid to have changed when the VM is restarted, so use a one-shot
	function to determine if we are restarting the VM (as opposed to returning
	from an image save)."

	super initialize.
	(self processAccessor canAccessSystem not or: [processAccessor primOneShot]) ifTrue:
		["Restarting the VM in a new process"
		self resetThreads.
		self threads; mainThread.
		processHandle := nil.
		self processHandle].
	self refreshFromProcessAccessor
! !

!WindowsProcess methodsFor: 'initialize - release' stamp: 'dtl 7/4/2010 11:16'!
initializeAllMyChildren
	"On Windows, the process ID is not unique. Use an OrderedCollection instead
	of a Dictionary to keep track of the child processes."

	^ allMyChildren := OrderedCollection new
! !

!WindowsProcess methodsFor: 'platform identification' stamp: 'dtl 8/30/2003 17:50'!
isResponsibleForThisPlatform
	"Answer true is this is an instance of the class which is responsible for representing
	the OS process for the Squeak VM running on the current platform. A false answer is
	usually the result of running the image on a different platform and VM."

	^ self class isWindows
! !

!WindowsProcess methodsFor: 'accessing' stamp: 'dtl 2/28/2002 07:15'!
mainThread
	"The main thread for this OS process. The handle for this thread is a
	pseudo-handle, and cannot be used to close the main thread."

	^ mainThread ifNil: [mainThread := processAccessor getMainThread]! !

!WindowsProcess methodsFor: 'console' stamp: 'dtl 9/7/2002 20:57'!
openConsole
	"Open a console. This makes the standard input, output and error streams available."

	"OSProcess thisOSProcess openConsole"

	self processAccessor primAllocConsole.
	self refreshFromProcessAccessor
! !

!WindowsProcess methodsFor: 'environment' stamp: 'dtl 9/25/2005 05:11'!
path
	"Newer versions of Windows mixed case"

	^ self environment
		at: #PATH
		ifAbsent: [environment
				at: #Path
				ifAbsent: ['']]! !

!WindowsProcess methodsFor: 'accessing' stamp: 'dtl 2/22/2002 16:43'!
pid

	^ pid := self processAccessor primGetPid
! !

!WindowsProcess methodsFor: 'accessing' stamp: 'dtl 2/28/2002 07:26'!
processHandle
	"The handle for this OS process. This is a pseudo-handle, a constant provided
	by Windows to represent the process. Note that the main thread handle is also
	represented by a pseudo-handle."

	^ processHandle ifNil: [processHandle := self processAccessor primGetPidHandle]
! !

!WindowsProcess methodsFor: 'updating' stamp: 'dtl 9/7/2002 20:47'!
refreshFromProcessAccessor
	"Set my instance variables to reflect the state of the OS process in which this Smalltalk
	virtual machine is executing."

	sessionID := self processAccessor getSessionIdentifier.
	pid := processAccessor primGetPid.
	self setStdIn.
	self setStdOut.
	self setStdErr.
	environment := self getEnvironmentDictionary
! !

!WindowsProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 15:00'!
registerChildProcess: anOSProcess
	"Register the external process and set an exit handler thread to signal when
	the process exits."

	accessProtect critical: [ allMyChildren add: anOSProcess ].
	self threads add: self restartChildWatcherThread.
	^ anOSProcess
! !

!WindowsProcess methodsFor: 'initialize - release' stamp: 'dtl 2/28/2002 07:16'!
resetThreads
	"If any cleanup is required, do it here."

	threads := nil.
	mainThread := nil
! !

!WindowsProcess methodsFor: 'child process management' stamp: 'dtl 1/13/2007 09:49'!
restartChildWatcherThread
	"Set an exit handler thread to signal when the process exits."


	^ self processAccessor restartChildWatcherThread: self activeHandles
! !

!WindowsProcess methodsFor: 'private' stamp: 'dtl 9/7/2002 20:55'!
setStdErr
	"Reset to nil if the console has been closed"

	^ self processAccessor getStdErr isNil
		ifTrue: [stdErr := nil]
		ifFalse: [super setStdErr]
! !

!WindowsProcess methodsFor: 'private' stamp: 'dtl 9/7/2002 20:55'!
setStdIn
	"Reset to nil if the console has been closed"

	^ self processAccessor getStdIn isNil
		ifTrue: [stdIn := nil]
		ifFalse: [super setStdIn]
! !

!WindowsProcess methodsFor: 'private' stamp: 'dtl 9/7/2002 20:55'!
setStdOut
	"Reset to nil if the console has been closed"

	^ self processAccessor getStdOut isNil
		ifTrue: [stdOut := nil]
		ifFalse: [super setStdOut]
! !

!WindowsProcess methodsFor: 'accessing' stamp: 'dtl 2/28/2002 08:04'!
threads
	"One or more threads of execution within the OS process. The main
	thread for the process is held by the mainThread variable and is not
	included in this collection. Threads are created to wait for the exit of
	child processes, so this collection grows as child processes are created."

	^ threads ifNil: [threads := OrderedCollection new]
! !

!WindowsProcess methodsFor: 'child process management' stamp: 'dtl 7/24/2010 15:00'!
unregisterChildProcess: anOSProcess

	| p |
	accessProtect critical: [ p := allMyChildren remove: anOSProcess pid ].
	^ p
! !

!WindowsProcess methodsFor: 'updating' stamp: 'dtl 2/25/2002 08:29'!
update: aParameter
	"Framework to update some or all of the instance variables based on external events,
	such as receipt of a sigchd signal when a child process exits."

	(aParameter == (self processAccessor)) ifTrue:
		[^ self refreshFromProcessAccessor; yourself].
	(aParameter == #pid) ifTrue:
		[^ self resetChildProcessDictionary. "Forget children of prior process"].
	(aParameter == #childProcessStatus) ifTrue:
		[^ self updateActiveChildren; changed; yourself].
	(aParameter == #startUp) ifTrue:
		[^ self update: #pid].
	aParameter == #invalidProcessAccessor ifTrue:
		[processAccessor := processAccessor ifNotNil: [processAccessor removeDependent: self. nil].
		^ self].
	self error: 'Unexpected update parameter'! !

!WindowsProcess methodsFor: 'child process management' stamp: 'dtl 2/28/2002 08:17'!
updateActiveChildren

	super updateActiveChildren.
	self restartChildWatcherThread
! !

!WindowsProcess methodsFor: 'child process creation' stamp: 'dtl 11/24/2008 17:47'!
waitForCommand: aCommandString
	"Run a command in a shell process. Similar to the system(3) call in the standard C library.
	The active Smalltalk process waits for completion of the external command process. This just
	uses a simple polling loop, which is not very elegant but works well enough for most purposes."

	| proc d |
	d := Delay forMilliseconds: 50.
	proc := self command: aCommandString.
	
	proc ifNil: [self class noAccessorAvailable].
	[proc runState == #complete] whileFalse: [d wait].
	^ proc
! !

Object subclass: #UnixProcessExitStatus
	instanceVariableNames: 'intValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Unix'!

!UnixProcessExitStatus commentStamp: 'dtl 8/23/2012 22:48' prior: 0!
A UnixProcessExitStatus represents the exit status of a unix process. This is an integer bit field answered by the wait() system call that contains information about exit status of the process. The meaning of the bit field varies according to the cause of process exit. 

Following a normal process exit, the status may be decoded to provide a small positive integer value in the range 0 - 255, which is the value that is presented by a unix shell as the exit status of a program. If terminated by a signal, the corresponding value is the signal number of the signal that caused process exit.

UnixExitStatus decodes the process exit status in a manner compatible with a typical GNU unix implementation. It is not guaranteed to be portable and may produce misleading results on other unix systems.
!

!UnixProcessExitStatus class methodsFor: 'instance creation' stamp: 'dtl 8/22/2012 21:24'!
for: integerValue

	^self new for: integerValue
! !

!UnixProcessExitStatus methodsFor: 'accessing' stamp: 'dtl 8/22/2012 23:10'!
exitStatus
	self isExited ifTrue: [^self statusIfExited].
	self isSignaled ifTrue: [^self statusIfSignaled].
	self isStopped ifTrue: [^self statusIfStopped].
	self notify: 'cannot decode exit status ', intValue asString.
	^intValue! !

!UnixProcessExitStatus methodsFor: 'initialize-release' stamp: 'dtl 8/22/2012 23:10'!
for: anInteger
	intValue := anInteger! !

!UnixProcessExitStatus methodsFor: 'testing' stamp: 'dtl 8/22/2012 22:32'!
isExited
	"
	/* Nonzero if STATUS indicates normal termination.  */
	#define __WIFEXITED(status)     (__WTERMSIG(status) == 0)
	"
	^ self statusIfSignaled = 0

! !

!UnixProcessExitStatus methodsFor: 'testing' stamp: 'dtl 8/22/2012 23:11'!
isSignaled
	"
	/* Nonzero if STATUS indicates termination by a signal.  */
	#define __WIFSIGNALED(status) \
	  (((signed char) (((status) & 0x7f) + 1) >> 1) > 0)
	"
	^(((intValue bitAnd: 16r7F) + 1) >> 1) > 0

! !

!UnixProcessExitStatus methodsFor: 'testing' stamp: 'dtl 8/22/2012 23:11'!
isStopped
	"
	/* Nonzero if STATUS indicates the child is stopped.  */
	#define __WIFSTOPPED(status)    (((status) & 0xff) == 0x7f)
	"
	^(intValue bitAnd: 16rFF) = 16r7F

! !

!UnixProcessExitStatus methodsFor: 'testing' stamp: 'dtl 8/22/2012 23:31'!
isSuccess
	^self exitStatus = 0! !

!UnixProcessExitStatus methodsFor: 'printing' stamp: 'dtl 8/22/2012 23:18'!
printOn: aStream

	self isExited
		ifTrue: [aStream nextPutAll: 'normal termination with status ';
				 nextPutAll: self statusIfExited asString].
	self isSignaled
		ifTrue: [aStream nextPutAll: 'exit due to signal ';
				nextPutAll: self statusIfSignaled asString].
	self isStopped
		ifTrue: [aStream nextPutAll: 'stopped due to signal ';
				nextPutAll: self statusIfStopped].! !

!UnixProcessExitStatus methodsFor: 'accessing' stamp: 'dtl 8/22/2012 23:11'!
statusIfExited
	"
	/* If WIFEXITED(STATUS), the low-order 8 bits of the status.  */
	#define __WEXITSTATUS(status)   (((status) & 0xff00) >> 8)
	"
	^(intValue bitAnd: 16rFF00) >> 8! !

!UnixProcessExitStatus methodsFor: 'accessing' stamp: 'dtl 8/22/2012 23:11'!
statusIfSignaled
	"
	/* If WIFSIGNALED(STATUS), the terminating signal.  */
	#define __WTERMSIG(status)      ((status) & 0x7f)
	"
	^intValue bitAnd: 16r7F! !

!UnixProcessExitStatus methodsFor: 'accessing' stamp: 'dtl 8/22/2012 22:28'!
statusIfStopped
	"
	/* If WIFSTOPPED(STATUS), the signal that stopped the child.  */
	#define __WSTOPSIG(status)      __WEXITSTATUS(status)
	"
	^self statusIfExited! !

!UnixProcessExitStatus methodsFor: 'accessing' stamp: 'dtl 8/22/2012 23:11'!
value
	^intValue! !

Object subclass: #WindowsThread
	instanceVariableNames: 'threadID handle runState'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'OSProcess-Win32'!

!WindowsThread commentStamp: '<historical>' prior: 0!
I represent a thread of execution within a Windows process. May threadID is a unique
identifier for the thread, and my handle is a Windows HANDLE to the thread. My handle
should be closed when the thread exits.!

!WindowsThread class methodsFor: 'instance creation' stamp: 'dtl 2/25/2002 07:00'!
threadID: anInteger handle: aHandleObject

	^ super new threadID: anInteger; handle: aHandleObject; initialize
! !

!WindowsThread class methodsFor: 'instance creation' stamp: 'dtl 2/26/2002 07:17'!
threadID: anInteger handle: aHandleObject running: trueOrFalse

	| thread |
	thread := super new threadID: anInteger; handle: aHandleObject; initialize.
	trueOrFalse ifTrue: [thread running].
	^ thread
! !

!WindowsThread methodsFor: 'initialize - release' stamp: 'dtl 2/25/2002 07:38'!
closeHandle
	"Clean up after thread exits."

	OSProcess accessor primCloseHandle: handle.
	handle := nil
! !

!WindowsThread methodsFor: 'setting run state' stamp: 'dtl 2/26/2002 10:45'!
complete
	"Thread has exited."

	self closeHandle; runState: #complete
! !

!WindowsThread methodsFor: 'accessing' stamp: 'dtl 2/25/2002 06:39'!
handle
	"A Windows HANDLE represented as a ByteArray."

	^ handle! !

!WindowsThread methodsFor: 'accessing' stamp: 'dtl 2/25/2002 06:38'!
handle: aHandleObject
	"A Windows HANDLE represented as a ByteArray."

	handle := aHandleObject! !

!WindowsThread methodsFor: 'initialize - release' stamp: 'dtl 2/25/2002 07:03'!
initialize

	self runState! !

!WindowsThread methodsFor: 'testing' stamp: 'dtl 2/25/2002 06:37'!
isComplete

	^ self runState == #complete! !

!WindowsThread methodsFor: 'testing' stamp: 'dtl 2/25/2002 06:37'!
isRunning

	^ self runState == #running! !

!WindowsThread methodsFor: 'printing' stamp: 'dtl 2/27/2002 12:02'!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' (', self threadID printString, ', ', self runState, ')'
! !

!WindowsThread methodsFor: 'accessing' stamp: 'dtl 1/25/2004 11:01'!
runState

	^ runState ifNil: [self unknownRunState]
! !

!WindowsThread methodsFor: 'accessing' stamp: 'dtl 2/25/2002 07:02'!
runState: aSymbol

	runState := aSymbol
! !

!WindowsThread methodsFor: 'setting run state' stamp: 'dtl 2/25/2002 06:36'!
running
	"Thread is scheduled to run."

	self runState: #running
! !

!WindowsThread methodsFor: 'initialize - release' stamp: 'dtl 2/27/2002 11:35'!
terminate
	"Force an exit. No cleanup is performed. Use with caution for a thread which
	is (for example) manipulating a mutex."

	(self isRunning and: [self handle notNil]) ifTrue:
		[OSProcess accessor primTerminateThread: self handle.
		self complete]
! !

!WindowsThread methodsFor: 'accessing' stamp: 'dtl 2/25/2002 06:40'!
threadID
	"A unique identifier for the thread."

	^ threadID
! !

!WindowsThread methodsFor: 'accessing' stamp: 'dtl 2/25/2002 06:40'!
threadID: anInteger
	"A unique identifier for the thread."

	threadID := anInteger
! !

!WindowsThread methodsFor: 'setting run state' stamp: 'dtl 2/25/2002 06:36'!
unknownRunState
	"Unable to determine the current run state of the thread, possibly because
	this is a stale reference to a thread which no longer exists."

	self runState: #unknownRunState
! !
HtmlUnescapedSpecialCharacterError initialize!
CSSPropertyDefinition initialize!
OSProcess initialize!
ThisOSProcess initialize!
OSProcessAccessor initialize!
HtmlPListSerializer initialize!
AttachableFileStream initialize!
